package com.simibubi.create.compat.jei.category;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import javax.annotation.ParametersAreNonnullByDefault;

import com.simibubi.create.compat.jei.category.BlockCuttingCategory.CondensedBlockCuttingRecipe;
import com.simibubi.create.compat.jei.category.animations.AnimatedSaw;
import com.simibubi.create.foundation.gui.AllGuiTextures;
import com.simibubi.create.foundation.item.ItemHelper;

import mezz.jei.api.gui.builder.IRecipeLayoutBuilder;
import mezz.jei.api.gui.ingredient.IRecipeSlotsView;
import mezz.jei.api.recipe.IFocusGroup;
import mezz.jei.api.recipe.RecipeIngredientRole;
import net.minecraft.class_1799;
import net.minecraft.class_1856;
import net.minecraft.class_1860;
import net.minecraft.class_2960;
import net.minecraft.class_332;
import net.minecraft.class_3975;

@ParametersAreNonnullByDefault
public class BlockCuttingCategory extends CreateRecipeCategory<CondensedBlockCuttingRecipe> {

	private final AnimatedSaw saw = new AnimatedSaw();

	public BlockCuttingCategory(Info<CondensedBlockCuttingRecipe> info) {
		super(info);
	}

	@Override
	public void setRecipe(IRecipeLayoutBuilder builder, CondensedBlockCuttingRecipe recipe, IFocusGroup focuses) {
		List<List<class_1799>> results = recipe.getCondensedOutputs();

		builder
				.addSlot(RecipeIngredientRole.INPUT, 5, 5)
				.setBackground(getRenderedSlot(), -1 , -1)
				.addItemStacks(Arrays.asList(recipe.method_8117().get(0).method_8105()));

		int i = 0;
		for (List<class_1799> itemStacks : results) {
			int xPos = 78 + (i % 5) * 19;
			int yPos = 48 + (i / 5) * -19;

			builder
					.addSlot(RecipeIngredientRole.OUTPUT, xPos, yPos)
					.setBackground(getRenderedSlot(), -1 , -1)
					.addItemStacks(itemStacks);
			i++;
		}
	}

	@Override
	public void draw(CondensedBlockCuttingRecipe recipe, IRecipeSlotsView recipeSlotsView, class_332 graphics, double mouseX, double mouseY) {
		AllGuiTextures.JEI_DOWN_ARROW.render(graphics, 31, 6);
		AllGuiTextures.JEI_SHADOW.render(graphics, 33 - 17, 37 + 13);
		saw.draw(graphics, 33, 37);
	}

	public static class CondensedBlockCuttingRecipe extends class_3975 {

		List<class_1799> outputs = new ArrayList<>();

		public CondensedBlockCuttingRecipe(class_1856 ingredient) {
			super(new class_2960(""), "", ingredient, class_1799.field_8037);
		}

		public void addOutput(class_1799 stack) {
			outputs.add(stack);
		}

		public List<class_1799> getOutputs() {
			return outputs;
		}

		public List<List<class_1799>> getCondensedOutputs() {
			List<List<class_1799>> result = new ArrayList<>();
			int index = 0;
			boolean firstPass = true;
			for (class_1799 itemStack : outputs) {
				if (firstPass)
					result.add(new ArrayList<>());
				result.get(index).add(itemStack);
				index++;
				if (index >= 15) {
					index = 0;
					firstPass = false;
				}
			}
			return result;
		}

		@Override
		public boolean method_8118() {
			return true;
		}

	}
	
	public static List<CondensedBlockCuttingRecipe> condenseRecipes(List<class_1860<?>> stoneCuttingRecipes) {
		List<CondensedBlockCuttingRecipe> condensed = new ArrayList<>();
		Recipes: for (class_1860<?> recipe : stoneCuttingRecipes) {
			class_1856 i1 = recipe.method_8117().get(0);
			for (CondensedBlockCuttingRecipe condensedRecipe : condensed) {
				if (ItemHelper.matchIngredients(i1, condensedRecipe.method_8117().get(0))) {
					condensedRecipe.addOutput(getResultItem(recipe));
					continue Recipes;
				}
			}
			CondensedBlockCuttingRecipe cr = new CondensedBlockCuttingRecipe(i1);
			cr.addOutput(getResultItem(recipe));
			condensed.add(cr);
		}
		return condensed;
	}

}
