package com.simibubi.create.compat.rei.category;

import java.util.List;
import java.util.function.Supplier;
import com.simibubi.create.AllBlocks;
import com.simibubi.create.AllPartialModels;
import com.simibubi.create.compat.rei.category.animations.AnimatedKinetics;
import com.simibubi.create.compat.rei.display.CreateDisplay;
import com.simibubi.create.content.processing.recipe.ProcessingOutput;
import com.simibubi.create.content.processing.recipe.ProcessingRecipe;
import com.simibubi.create.foundation.gui.AllGuiTextures;
import com.simibubi.create.foundation.utility.Lang;

import me.shedaniel.math.Point;
import me.shedaniel.rei.api.client.gui.widgets.Widget;
import me.shedaniel.rei.api.common.util.EntryIngredients;
import net.minecraft.class_1799;
import net.minecraft.class_1860;
import net.minecraft.class_332;
import net.minecraft.class_4587;
import net.minecraft.class_7833;

public abstract class ProcessingViaFanCategory<T extends class_1860<?>> extends CreateRecipeCategory<T> {

	protected static final int SCALE = 24;

	public ProcessingViaFanCategory(Info<T> info) {
		super(info);
	}

	public static Supplier<class_1799> getFan(String name) {
		return () -> AllBlocks.ENCASED_FAN.asStack()
			.method_7977(Lang.translateDirect("recipe." + name + ".fan").method_27694(style -> style.method_10978(false)));
	}

	@Override
	public void addWidgets(CreateDisplay<T> display, List<Widget> ingredients, Point origin) {
		ingredients.add(basicSlot(origin.x + 21, origin.y + 48)
				.markInput()
				.entries(display.getInputEntries().get(0)));
		ingredients.add(basicSlot(origin.x + 141, origin.y + 48)
				.markOutput()
				.entries(display.getOutputEntries().get(0)));
	}

	@Override
	public void draw(T recipe, class_332 graphics, double mouseX, double mouseY) {
		renderWidgets(graphics, recipe, mouseX, mouseY);

		class_4587 matrixStack = graphics.method_51448();
		matrixStack.method_22903();
		translateFan(matrixStack);
		matrixStack.method_22907(class_7833.field_40714.rotationDegrees(-12.5f));
		matrixStack.method_22907(class_7833.field_40716.rotationDegrees(22.5f));

		AnimatedKinetics.defaultBlockElement(AllPartialModels.ENCASED_FAN_INNER)
			.rotateBlock(180, 0, AnimatedKinetics.getCurrentAngle() * 16)
			.scale(SCALE)
			.render(graphics);

		AnimatedKinetics.defaultBlockElement(AllBlocks.ENCASED_FAN.getDefaultState())
			.rotateBlock(0, 180, 0)
			.atLocal(0, 0, 0)
			.scale(SCALE)
			.render(graphics);

		renderAttachedBlock(graphics);
		matrixStack.method_22909();
	}

	protected void renderWidgets(class_332 graphics, T recipe, double mouseX, double mouseY) {
		AllGuiTextures.JEI_SHADOW.render(graphics, 46, 29);
		getBlockShadow().render(graphics, 65, 39);
		AllGuiTextures.JEI_LONG_ARROW.render(graphics, 54, 51);
		AllGuiTextures.JEI_SLOT.render(graphics, 20, 47);
		AllGuiTextures.JEI_SLOT.render(graphics, 140, 47);
	}

	protected AllGuiTextures getBlockShadow() {
		return AllGuiTextures.JEI_SHADOW;
	}

	protected void translateFan(class_4587 matrixStack) {
		matrixStack.method_46416(56, 33, 0);
	}

	protected abstract void renderAttachedBlock(class_332 graphics);

	public static abstract class MultiOutput<T extends ProcessingRecipe<?>> extends ProcessingViaFanCategory<T> {

		public MultiOutput(Info<T> info) {
			super(info);
		}

		@Override
		public void addWidgets(CreateDisplay<T> display, List<Widget> ingredients, Point origin) {
			List<ProcessingOutput> results = display.getRecipe().getRollableResults();
			int xOffsetAmount = 1 - Math.min(3, results.size());

			ingredients.add(basicSlot(origin.x + 5 * xOffsetAmount + 21, origin.y + 48)
					.markInput()
					.entries(display.getInputEntries().get(0)));

			int xOffsetOutput = 9 * xOffsetAmount;
			boolean excessive = results.size() > 9;
			for (int outputIndex = 0; outputIndex < results.size(); outputIndex++) {
				int xOffset = (outputIndex % 3) * 19 + xOffsetOutput;
				int yOffset = (outputIndex / 3) * -19 + (excessive ? 8 : 0);

				ingredients.add(basicSlot(origin.x + 141 + xOffset, origin.y + 48 + yOffset)
						.markOutput()
						.entries(EntryIngredients.of(results.get(outputIndex).getStack())));
			}

			addStochasticTooltip(ingredients, results);
		}

		@Override
		protected void renderWidgets(class_332 graphics, T recipe, double mouseX, double mouseY) {
			int size = recipe.getRollableResultsAsItemStacks()
				.size();
			int xOffsetAmount = 1 - Math.min(3, size);

			AllGuiTextures.JEI_SHADOW.render(graphics, 46, 29);
			getBlockShadow().render(graphics, 65, 39);
			AllGuiTextures.JEI_LONG_ARROW.render(graphics, 7 * xOffsetAmount + 54, 51);
			AllGuiTextures.JEI_SLOT.render(graphics, 5 * xOffsetAmount + 20, 47);

			int xOffsetOutput = 9 * xOffsetAmount;
			boolean excessive = size > 9;
			for (int i = 0; i < size; i++) {
				int xOffset = (i % 3) * 19 + xOffsetOutput;
				int yOffset = (i / 3) * -19 + (excessive ? 8 : 0);
				getRenderedSlot(recipe, i).render(graphics, 140 + xOffset, 47 + yOffset);
			}
		}

	}

}
