package com.simibubi.create.compat.sandwichable;

import javax.annotation.Nullable;
import net.minecraft.class_1799;
import net.minecraft.class_1937;
import net.minecraft.class_2338;
import net.minecraft.class_2487;
import net.minecraft.class_3417;
import net.minecraft.class_3419;
import com.google.common.collect.ImmutableList;
import com.simibubi.create.Create;
import com.simibubi.create.content.kinetics.belt.BeltHelper;
import com.simibubi.create.content.kinetics.belt.behaviour.TransportedItemStackHandlerBehaviour;
import com.simibubi.create.content.kinetics.belt.behaviour.TransportedItemStackHandlerBehaviour.TransportedResult;
import com.simibubi.create.content.kinetics.belt.transport.TransportedItemStack;
import com.simibubi.create.content.kinetics.deployer.DeployerBlockEntity;
import com.simibubi.create.content.kinetics.deployer.DeployerFakePlayer;

import io.github.foundationgames.sandwichable.Sandwichable;
import io.github.foundationgames.sandwichable.items.ItemsRegistry;
import io.github.foundationgames.sandwichable.util.Sandwich;

public class SequencedSandwiching {
	/**
	 * Only allow sequenced sandwiching on sandwiches (or bread, becoming sandwiches),
	 * AND if the item-to-add can be put on a sandwich.
	 */
	public static boolean shouldSandwich(class_1799 handling, class_1799 held, class_1937 level) {
		boolean eligible = Sandwich.canAdd(held) && (Sandwichable.isBread(handling) || handling.method_31574(ItemsRegistry.SANDWICH));
		int max = level.method_8450().method_8356(Sandwichable.SANDWICH_SIZE_RULE);
		Sandwich sandwich = sandwichFromStack(handling);
		if (sandwich != null && max > 0) {
			return eligible && sandwich.getFoodList().size() < max;
		}
		return eligible;
	}

	/**
	 * Actually assemble a sandwich.
	 * @param transported the sandwich stack, passing below the deployer
	 * @param handler the deployer's handler
	 * @param deployer the block entity of the deployer
	 */
	public static void activateSandwich(TransportedItemStack transported, TransportedItemStackHandlerBehaviour handler,
										DeployerBlockEntity deployer) {

		TransportedItemStack transportedRemainder = transported.copy();
		DeployerFakePlayer player = deployer.getPlayer();
		transportedRemainder.stack.method_7934(1);
		class_1799 heldItem = player.method_6047();

		class_1799 newSandwich = stackOnSandwich(transported.stack, heldItem, deployer);
		if (newSandwich.method_7960())
			return;

		TransportedItemStack output = transported.copy();
		boolean centered = BeltHelper.isItemUpright(newSandwich);
		output.stack = newSandwich;
		output.angle = centered ? 180 : Create.RANDOM.nextInt(360);

		handler.handleProcessingOnItem(transported, TransportedResult
				.convertToAndLeaveHeld(ImmutableList.of(output), transportedRemainder));

		heldItem.method_7934(1);

		class_2338 pos = deployer.method_11016();
		class_1937 world = deployer.method_10997();
		if (heldItem.method_7960())
			world.method_8396(null, pos, class_3417.field_15075, class_3419.field_15245, .25f, 1);
		world.method_8396(null, pos, class_3417.field_15197, class_3419.field_15245, .25f, .75f);

		deployer.sendData();
	}

	/**
	 * Stacks 'toAdd' on top of the given sandwich.
	 * If the item to add is a spread (and has a remainder) it is added to the deployer's overflow items.
	 * @return the new sandwich, with the item stacked on top, or EMPTY if nothing stacked
	 */
	public static class_1799 stackOnSandwich(class_1799 sandwich, class_1799 toAdd, DeployerBlockEntity deployer) {
		class_1937 level = deployer.method_10997();
		Sandwich s = sandwichFromStack(sandwich);
		if (s != null) {
			// null - not added
			// empty - do not return an item
			// non-empty - return item
			class_1799 result = s.tryAddTopFoodFrom(level, toAdd.method_7972());
			if (result == null) {
				return sandwich;
			} else if (!result.method_7960()) {
				deployer.getOverflowItems().add(result.method_7972());
			}
			class_2487 newTag = s.writeToNbt(new class_2487());
			class_1799 newSandwich = sandwich.method_7972();
			newSandwich.method_7948().method_10566("BlockEntityTag", newTag);
			return newSandwich;
		} else if (Sandwichable.isBread(sandwich)) {
			s = new Sandwich();
			s.addTopFoodFrom(sandwich.method_7972());
			class_1799 result = s.tryAddTopFoodFrom(level, toAdd.method_7972());
			if (result == null) {
				return sandwich;
			} else if (!result.method_7960()) {
				deployer.getOverflowItems().add(result.method_7972());
			}
			class_2487 newTag = s.writeToNbt(new class_2487());
			class_1799 freshSandwich = ItemsRegistry.SANDWICH.method_7854();
			freshSandwich.method_7948().method_10566("BlockEntityTag", newTag);
			return freshSandwich;
		}
		return class_1799.field_8037;
	}

	@Nullable
	public static Sandwich sandwichFromStack(class_1799 stack) {
		if (stack.method_31574(ItemsRegistry.SANDWICH)) {
			class_2487 tag = stack.method_7969();
			if (tag != null && tag.method_10545("BlockEntityTag")) {
				tag = tag.method_10562("BlockEntityTag");
				Sandwich s = new Sandwich();
				s.addFromNbt(tag);
				return s;
			}
		}
		return null;
	}
}
