package com.simibubi.create.compat.sodium;

import java.lang.reflect.Method;
import java.util.function.Function;

import com.simibubi.create.Create;
import com.simibubi.create.compat.Mods;
import com.simibubi.create.foundation.utility.Components;

import me.jellysquid.mods.sodium.client.render.texture.SpriteUtil;
import net.fabricmc.fabric.api.client.networking.v1.ClientPlayConnectionEvents;
import net.fabricmc.fabric.api.client.rendering.v1.WorldRenderEvents;
import net.fabricmc.fabric.api.networking.v1.PacketSender;
import net.minecraft.class_1058;
import net.minecraft.class_124;
import net.minecraft.class_1723;
import net.minecraft.class_2558;
import net.minecraft.class_2564;
import net.minecraft.class_2568;
import net.minecraft.class_2960;
import net.minecraft.class_310;
import net.minecraft.class_5250;
import net.minecraft.class_634;

/**
 * Fixes the Mechanical Saw's sprite and lets players know when Indium isn't installed.
 */
public class SodiumCompat {
	public static final class_2960 SAW_TEXTURE = Create.asResource("block/saw_reversed");

	public static void init() {
		if (!Mods.INDIUM.isLoaded()) {
			ClientPlayConnectionEvents.JOIN.register(SodiumCompat::sendNoIndiumWarning);
		}
		if (spriteUtilWorks()) {
			class_310 mc = class_310.method_1551();
			WorldRenderEvents.START.register(ctx -> {
				Function<class_2960, class_1058> atlas = mc.method_1549(class_1723.field_21668);
				class_1058 sawSprite = atlas.apply(SAW_TEXTURE);
				SpriteUtil.markSpriteActive(sawSprite);
			});
		}
	}

	public static void sendNoIndiumWarning(class_634 handler, PacketSender sender, class_310 mc) {
		if (mc.field_1724 == null)
			return;

		class_5250 text = class_2564.method_10885(Components.literal("WARN"))
				.method_27692(class_124.field_1065)
				.method_10852(Components.literal(" Sodium is installed, but Indium is not. This will cause visual issues with Create!")
				)
				.method_27694(style -> style
						.method_10958(
								new class_2558(class_2558.class_2559.field_11749, "https://modrinth.com/mod/indium")
						)
						.method_10949(
								new class_2568(class_2568.class_5247.field_24342, Components.literal("Click here to open Indium's mod page"))
						)
				);

		mc.field_1724.method_7353(text, false);
	}

	private static boolean spriteUtilWorks() {
		try {
			// make sure class and method still exist, sodium is unstable
			Method method = SpriteUtil.class.getMethod("markSpriteActive", class_1058.class); // throws if missing
			if (method.getReturnType() != Void.TYPE)
				throw new IllegalStateException("markSpriteActive's signature has changed");
			return true;
		} catch (Throwable t) {
			Create.LOGGER.error("Create's Sodium compat errored and has been partially disabled. Report this!", t);
		}
		return false;
	}
}
