package com.simibubi.create.content.contraptions;

import java.util.Collection;
import java.util.function.Supplier;

import javax.annotation.Nullable;
import net.minecraft.class_2246;
import net.minecraft.class_2248;
import net.minecraft.class_2960;
import net.minecraft.class_3499;
import net.minecraft.class_7923;
import com.simibubi.create.foundation.utility.AttachedRegistry;
import com.simibubi.create.infrastructure.config.AllConfigs;

public enum ContraptionMovementSetting {
	MOVABLE, NO_PICKUP, UNMOVABLE;

	private static final AttachedRegistry<class_2248, Supplier<ContraptionMovementSetting>> SETTING_SUPPLIERS = new AttachedRegistry<>(class_7923.field_41175);

	public static void register(class_2960 block, Supplier<ContraptionMovementSetting> settingSupplier) {
		SETTING_SUPPLIERS.register(block, settingSupplier);
	}

	public static void register(class_2248 block, Supplier<ContraptionMovementSetting> settingSupplier) {
		SETTING_SUPPLIERS.register(block, settingSupplier);
	}

	@Nullable
	public static ContraptionMovementSetting get(class_2248 block) {
		if (block instanceof IMovementSettingProvider provider)
			return provider.getContraptionMovementSetting();
		Supplier<ContraptionMovementSetting> supplier = SETTING_SUPPLIERS.get(block);
		if (supplier == null)
			return null;
		return supplier.get();
	}

	public static boolean allAre(Collection<class_3499.class_3501> blocks, ContraptionMovementSetting are) {
		return blocks.stream().anyMatch(b -> get(b.comp_1342().method_26204()) == are);
	}

	public static boolean isNoPickup(Collection<class_3499.class_3501> blocks) {
		return allAre(blocks, ContraptionMovementSetting.NO_PICKUP);
	}

	public static void registerDefaults() {
		register(class_2246.field_10260, () -> AllConfigs.server().kinetics.spawnerMovement.get());
		register(class_2246.field_27160, () -> AllConfigs.server().kinetics.amethystMovement.get());
		register(class_2246.field_10540, () -> AllConfigs.server().kinetics.obsidianMovement.get());
		register(class_2246.field_22423, () -> AllConfigs.server().kinetics.obsidianMovement.get());
		register(class_2246.field_23152, () -> AllConfigs.server().kinetics.obsidianMovement.get());
		register(class_2246.field_38420, () -> AllConfigs.server().kinetics.reinforcedDeepslateMovement.get());
	}

	public interface IMovementSettingProvider /* extends Block */ {
		ContraptionMovementSetting getContraptionMovementSetting();
	}
}
