package com.simibubi.create.content.contraptions;

import java.util.List;

import com.simibubi.create.AllBlockEntityTypes;
import com.simibubi.create.AllTags.AllBlockTags;
import com.simibubi.create.content.kinetics.crafter.MechanicalCrafterBlockEntity;
import com.simibubi.create.content.logistics.crate.BottomlessItemHandler;
import com.simibubi.create.content.logistics.vault.ItemVaultBlockEntity;
import com.simibubi.create.content.processing.recipe.ProcessingInventory;
import com.simibubi.create.foundation.utility.NBTHelper;

import io.github.fabricators_of_create.porting_lib.transfer.TransferUtil;
import io.github.fabricators_of_create.porting_lib.transfer.item.ItemStackHandler;
import io.github.fabricators_of_create.porting_lib.util.NBTSerializer;
import net.fabricmc.fabric.api.transfer.v1.item.InventoryStorage;
import net.fabricmc.fabric.api.transfer.v1.item.ItemVariant;
import net.fabricmc.fabric.api.transfer.v1.storage.Storage;
import net.fabricmc.fabric.api.transfer.v1.storage.StorageView;
import net.fabricmc.fabric.api.transfer.v1.storage.base.SingleSlotStorage;
import net.fabricmc.fabric.api.transfer.v1.transaction.Transaction;
import net.minecraft.class_1799;
import net.minecraft.class_2487;
import net.minecraft.class_2586;
import net.minecraft.class_2595;
import net.minecraft.class_2627;
import net.minecraft.class_2680;
import net.minecraft.class_3719;
import net.minecraft.class_7923;

public class MountedStorage {

	ItemStackHandler handler;
	boolean noFuel;
	boolean valid;

	private class_2586 blockEntity;

	public static boolean canUseAsStorage(class_2586 be) {
		if (be == null)
			return false;
		if (be instanceof MechanicalCrafterBlockEntity)
			return false;
		if (AllBlockEntityTypes.CREATIVE_CRATE.is(be))
			return true;
		if (be instanceof class_2627)
			return true;
		if (be instanceof class_2595)
			return true;
		if (be instanceof class_3719)
			return true;
		if (be instanceof ItemVaultBlockEntity)
			return true;

		try {
			Storage<ItemVariant> handler = TransferUtil.getItemStorage(be);
			if (handler instanceof ItemStackHandler)
				return !(handler instanceof ProcessingInventory);
			return handler != null && canUseModdedInventory(be, handler);

		} catch (Exception e) {
			return false;
		}
	}

	public static boolean canUseModdedInventory(class_2586 be, Storage<ItemVariant> handler) {
		if (!handler.supportsExtraction() || !handler.supportsInsertion())
			return false;
		class_2680 blockState = be.method_11010();
		if (AllBlockTags.CONTRAPTION_INVENTORY_DENY.matches(blockState))
			return false;

		// There doesn't appear to be much of a standard for tagging chests/barrels
		String blockId = class_7923.field_41175.method_10221(blockState.method_26204())
			.method_12832();
		if (blockId.contains("ender"))
			return false;
		return blockId.endsWith("_chest") || blockId.endsWith("_barrel");
	}

	public MountedStorage(class_2586 be) {
		this.blockEntity = be;
		noFuel = be instanceof ItemVaultBlockEntity;
	}

	public void removeStorageFromWorld() {
		valid = false;
		if (blockEntity == null)
			return;

		if (blockEntity instanceof class_2595 chest) {
			class_2487 tag = blockEntity.method_38242();
			if (tag.method_10573("LootTable", 8))
				return;

			handler = new ItemStackHandler(chest.method_5439());
			for (int i = 0; i < handler.getSlotCount(); i++) {
				handler.setStackInSlot(i, chest.method_5438(i));
			}
			valid = true;
			return;
		}

		// multiblock vaults need to provide individual invs
		if (blockEntity instanceof ItemVaultBlockEntity) {
			handler = ((ItemVaultBlockEntity) blockEntity).getInventoryOfBlock();
			valid = true;
			return;
		}

		Storage<ItemVariant> beHandler = TransferUtil.getItemStorage(blockEntity);
		if (beHandler == null)
			return;

		// be uses ItemStackHandler
		if (beHandler instanceof ItemStackHandler) {
			handler = (ItemStackHandler) beHandler;
			valid = true;
			return;
		}

		// serialization not accessible -> fill into a serializable handler
		// fabric: only trust Containers, todo: maybe change
		if (beHandler instanceof InventoryStorage inv && beHandler.supportsInsertion() && beHandler.supportsExtraction()) {
			try (Transaction t = TransferUtil.getTransaction()) {
				List<SingleSlotStorage<ItemVariant>> slots = inv.getSlots();
				class_1799[] stacks = new class_1799[slots.size()];
				for (int i = 0; i < slots.size(); i++) {
					SingleSlotStorage<ItemVariant> slot = slots.get(i);
					if (slot.isResourceBlank()) {
						stacks[i] = class_1799.field_8037;
						continue;
					}
					long contained = slot.getAmount();
					ItemVariant variant = slot.getResource();
					long extracted = slot.extract(variant, contained, t);
					if (extracted != contained) return; // can't extract it all for whatever reason - that's bad, give up
					stacks[i] = variant.toStack((int) extracted);
				}
				handler = new ItemStackHandler(stacks);
				valid = true;
			}
		}
	}

	public void addStorageToWorld(class_2586 be) {
		// FIXME: More dynamic mounted storage in .4
		if (handler instanceof BottomlessItemHandler)
			return;

		if (be instanceof class_2595 chest) {
			for (int i = 0; i < chest.method_5439(); i++) {
				class_1799 stack = i < handler.getSlotCount() ? handler.getStackInSlot(i) : class_1799.field_8037;
				chest.method_5447(i, stack);
			}
			return;
		}

		if (be instanceof ItemVaultBlockEntity) {
			((ItemVaultBlockEntity) be).applyInventoryToBlock(handler);
			return;
		}

		Storage<ItemVariant> teHandler = TransferUtil.getItemStorage(be);
		if (teHandler != null && teHandler.supportsInsertion()) {
			try (Transaction t = TransferUtil.getTransaction()) {
				// we need to remove whatever is in there to fill with our modified contents
				TransferUtil.clearStorage(teHandler);
				for (StorageView<ItemVariant> view : handler.nonEmptyViews()) {
					teHandler.insert(view.getResource(), view.getAmount(), t);
				}
				t.commit();
			}
		}
	}

	public ItemStackHandler getItemHandler() {
		return handler;
	}

	public class_2487 serialize() {
		if (!valid)
			return null;

		class_2487 tag = handler.serializeNBT();
		if (noFuel)
			NBTHelper.putMarker(tag, "NoFuel");
		if (!(handler instanceof BottomlessItemHandler))
			return tag;

		NBTHelper.putMarker(tag, "Bottomless");
		tag.method_10566("ProvidedStack", NBTSerializer.serializeNBT(handler.getStackInSlot(0)));
		return tag;
	}

	public static MountedStorage deserialize(class_2487 nbt) {
		MountedStorage storage = new MountedStorage(null);
		storage.handler = new ItemStackHandler();
		if (nbt == null)
			return storage;
		storage.valid = true;
		storage.noFuel = nbt.method_10545("NoFuel");

		if (nbt.method_10545("Bottomless")) {
			class_1799 providedStack = class_1799.method_7915(nbt.method_10562("ProvidedStack"));
			storage.handler = new BottomlessItemHandler(() -> providedStack);
			return storage;
		}

		storage.handler.deserializeNBT(nbt);
		return storage;
	}

	public boolean isValid() {
		return valid;
	}

	public boolean canUseForFuel() {
		return isValid() && !noFuel;
	}

}
