package com.simibubi.create.content.contraptions;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.function.Supplier;
import java.util.stream.Collectors;

import com.simibubi.create.content.contraptions.Contraption.ContraptionInvWrapper;
import com.simibubi.create.content.fluids.tank.FluidTankBlockEntity;
import com.simibubi.create.foundation.fluid.CombinedTankWrapper;
import com.simibubi.create.foundation.utility.Components;
import com.simibubi.create.foundation.utility.NBTHelper;

import io.github.fabricators_of_create.porting_lib.transfer.TransferUtil;
import io.github.fabricators_of_create.porting_lib.transfer.fluid.FluidTank;
import io.github.fabricators_of_create.porting_lib.transfer.item.ItemStackHandler;
import io.github.fabricators_of_create.porting_lib.fluids.FluidStack;
import net.fabricmc.fabric.api.transfer.v1.fluid.FluidVariant;
import net.fabricmc.fabric.api.transfer.v1.item.ItemVariant;
import net.fabricmc.fabric.api.transfer.v1.storage.Storage;
import net.minecraft.class_1657;
import net.minecraft.class_2281;
import net.minecraft.class_2338;
import net.minecraft.class_2350;
import net.minecraft.class_243;
import net.minecraft.class_2487;
import net.minecraft.class_2499;
import net.minecraft.class_2512;
import net.minecraft.class_2520;
import net.minecraft.class_2561;
import net.minecraft.class_2586;
import net.minecraft.class_2745;
import net.minecraft.class_3417;
import net.minecraft.class_3419;
import net.minecraft.class_3499.class_3501;

public class MountedStorageManager {

	protected ContraptionInvWrapper inventory;
	protected ContraptionInvWrapper fuelInventory;
	protected CombinedTankWrapper fluidInventory;
	protected Map<class_2338, MountedStorage> storage;
	protected Map<class_2338, MountedFluidStorage> fluidStorage;

	public MountedStorageManager() {
		storage = new HashMap<>();
		fluidStorage = new HashMap<>();
	}

	public void entityTick(AbstractContraptionEntity entity) {
		fluidStorage.forEach((pos, mfs) -> mfs.tick(entity, pos, entity.method_37908().field_9236));
	}

	public void createHandlers() {
		Collection<MountedStorage> itemHandlers = storage.values();

		inventory = wrapItems(itemHandlers.stream()
			.filter(MountedStorage::isValid)
			.map(MountedStorage::getItemHandler)
			.toList(), false);

		fuelInventory = wrapItems(itemHandlers.stream()
			.filter(MountedStorage::canUseForFuel)
			.map(MountedStorage::getItemHandler)
			.toList(), true);

		fluidInventory = wrapFluids(fluidStorage.values()
			.stream()
			.map(MountedFluidStorage::getFluidHandler)
			.collect(Collectors.toList()));
	}

	protected ContraptionInvWrapper wrapItems(Collection<? extends Storage<ItemVariant>> list, boolean fuel) {
		return new ContraptionInvWrapper(Arrays.copyOf(list.toArray(), list.size(), Storage[].class));
	}

	protected CombinedTankWrapper wrapFluids(Collection<? extends Storage<FluidVariant>> list) {
		return new CombinedTankWrapper(Arrays.copyOf(list.toArray(), list.size(), Storage[].class));
	}

	public void addBlock(class_2338 localPos, class_2586 be) {
		if (be != null && MountedStorage.canUseAsStorage(be))
			storage.put(localPos, new MountedStorage(be));
		if (be != null && MountedFluidStorage.canUseAsStorage(be))
			fluidStorage.put(localPos, new MountedFluidStorage(be));
	}

	public void read(class_2487 nbt, Map<class_2338, class_2586> presentBlockEntities, boolean clientPacket) {
		storage.clear();
		NBTHelper.iterateCompoundList(nbt.method_10554("Storage", class_2520.field_33260), c -> storage
			.put(class_2512.method_10691(c.method_10562("Pos")), MountedStorage.deserialize(c.method_10562("Data"))));

		fluidStorage.clear();
		NBTHelper.iterateCompoundList(nbt.method_10554("FluidStorage", class_2520.field_33260), c -> fluidStorage
			.put(class_2512.method_10691(c.method_10562("Pos")), MountedFluidStorage.deserialize(c.method_10562("Data"))));

		if (clientPacket && presentBlockEntities != null)
			bindTanks(presentBlockEntities);

		List<Storage<ItemVariant>> handlers = new ArrayList<>();
		List<Storage<ItemVariant>> fuelHandlers = new ArrayList<>();
		for (MountedStorage mountedStorage : storage.values()) {
			Storage<ItemVariant> itemHandler = mountedStorage.getItemHandler();
			handlers.add(itemHandler);
			if (mountedStorage.canUseForFuel())
				fuelHandlers.add(itemHandler);
		}

		inventory = wrapItems(handlers, false);
		fuelInventory = wrapItems(fuelHandlers, true);
		fluidInventory = wrapFluids(fluidStorage.values()
			.stream()
			.map(MountedFluidStorage::getFluidHandler)
			.map(tank -> (Storage<FluidVariant>) tank)
			.toList());
	}

	public void bindTanks(Map<class_2338, class_2586> presentBlockEntities) {
		fluidStorage.forEach((pos, mfs) -> {
			class_2586 blockEntity = presentBlockEntities.get(pos);
			if (!(blockEntity instanceof FluidTankBlockEntity))
				return;
			FluidTankBlockEntity tank = (FluidTankBlockEntity) blockEntity;
			FluidTank tankInventory = tank.getTankInventory();
			if (tankInventory instanceof FluidTank)
				((FluidTank) tankInventory).setFluid(mfs.tank.getFluid());
			tank.getFluidLevel()
				.startWithValue(tank.getFillState());
			mfs.assignBlockEntity(tank);
		});
	}

	public void write(class_2487 nbt, boolean clientPacket) {
		class_2499 storageNBT = new class_2499();
		if (!clientPacket)
			for (class_2338 pos : storage.keySet()) {
				class_2487 c = new class_2487();
				MountedStorage mountedStorage = storage.get(pos);
				if (!mountedStorage.isValid())
					continue;
				c.method_10566("Pos", class_2512.method_10692(pos));
				c.method_10566("Data", mountedStorage.serialize());
				storageNBT.add(c);
			}

		class_2499 fluidStorageNBT = new class_2499();
		for (class_2338 pos : fluidStorage.keySet()) {
			class_2487 c = new class_2487();
			MountedFluidStorage mountedStorage = fluidStorage.get(pos);
			if (!mountedStorage.isValid())
				continue;
			c.method_10566("Pos", class_2512.method_10692(pos));
			c.method_10566("Data", mountedStorage.serialize());
			fluidStorageNBT.add(c);
		}

		nbt.method_10566("Storage", storageNBT);
		nbt.method_10566("FluidStorage", fluidStorageNBT);
	}

	public void removeStorageFromWorld() {
		storage.values()
			.forEach(MountedStorage::removeStorageFromWorld);
		fluidStorage.values()
			.forEach(MountedFluidStorage::removeStorageFromWorld);
	}

	public void addStorageToWorld(class_3501 block, class_2586 blockEntity) {
		if (storage.containsKey(block.comp_1341())) {
			MountedStorage mountedStorage = storage.get(block.comp_1341());
			if (mountedStorage.isValid())
				mountedStorage.addStorageToWorld(blockEntity);
		}

		if (fluidStorage.containsKey(block.comp_1341())) {
			MountedFluidStorage mountedStorage = fluidStorage.get(block.comp_1341());
			if (mountedStorage.isValid())
				mountedStorage.addStorageToWorld(blockEntity);
		}
	}

	public void clear() {
		for (Storage<ItemVariant> storage : inventory.parts) {
			if (!(storage instanceof ContraptionInvWrapper wrapper) || !wrapper.isExternal) {
				TransferUtil.clearStorage(storage);
			}
		}
		TransferUtil.clearStorage(fluidInventory);
	}

	public void updateContainedFluid(class_2338 localPos, FluidStack containedFluid) {
		MountedFluidStorage mountedFluidStorage = fluidStorage.get(localPos);
		if (mountedFluidStorage != null)
			mountedFluidStorage.updateFluid(containedFluid);
	}

	public void attachExternal(Storage<ItemVariant> externalStorage) {
		inventory = new ContraptionInvWrapper(externalStorage, inventory);
		fuelInventory = new ContraptionInvWrapper(externalStorage, fuelInventory);
	}

	public ContraptionInvWrapper getItems() {
		return inventory;
	}

	public ContraptionInvWrapper getFuelItems() {
		return fuelInventory;
	}

	public CombinedTankWrapper getFluids() {
		return fluidInventory;
	}

	public boolean handlePlayerStorageInteraction(Contraption contraption, class_1657 player, class_2338 localPos) {
		if (player.method_37908().method_8608()) {
			class_2586 localBE = contraption.presentBlockEntities.get(localPos);
			return MountedStorage.canUseAsStorage(localBE);
		}

		MountedStorageManager storageManager = contraption.getStorageForSpawnPacket();
		MountedStorage storage = storageManager.storage.get(localPos);
		if (storage == null || storage.getItemHandler() == null)
			return false;
		ItemStackHandler primary = storage.getItemHandler();
		ItemStackHandler secondary = null;

		class_3501 info = contraption.getBlocks()
			.get(localPos);
		if (info != null && info.comp_1342().method_28498(class_2281.field_10770)) {
			class_2745 chestType = info.comp_1342().method_11654(class_2281.field_10770);
			class_2350 facing = info.comp_1342().method_28500(class_2281.field_10768)
				.orElse(class_2350.field_11035);
			class_2350 connectedDirection =
				chestType == class_2745.field_12574 ? facing.method_10170() : facing.method_10160();

			if (chestType != class_2745.field_12569) {
				MountedStorage storage2 = storageManager.storage.get(localPos.method_10093(connectedDirection));
				if (storage2 != null && storage2.getItemHandler() != null) {
					secondary = storage2.getItemHandler();
					if (chestType == class_2745.field_12574) {
						// switcheroo
						ItemStackHandler temp = primary;
						primary = secondary;
						secondary = temp;
					}
				}
			}
		}

		int slotCount = primary.getSlotCount() + (secondary == null ? 0 : secondary.getSlotCount());
		if (slotCount == 0)
			return false;
		if (slotCount % 9 != 0)
			return false;

		Supplier<Boolean> stillValid = () -> contraption.entity.method_5805()
			&& player.method_5707(contraption.entity.toGlobalVector(class_243.method_24953(localPos), 0)) < 64;
		class_2561 name = info != null ? info.comp_1342().method_26204()
			.method_9518() : Components.literal("Container");
		player.method_17355(MountedStorageInteraction.createMenuProvider(name, primary, secondary, slotCount, stillValid));

		class_243 soundPos = contraption.entity.toGlobalVector(class_243.method_24953(localPos), 0);
		player.method_37908().method_8396(null, class_2338.method_49638(soundPos), class_3417.field_17604, class_3419.field_15245, 0.75f, 1f);
		return true;
	}

}
