package com.simibubi.create.content.contraptions.actors.harvester;

import com.jozufozu.flywheel.api.Material;
import com.jozufozu.flywheel.api.MaterialManager;
import com.jozufozu.flywheel.core.Materials;
import com.jozufozu.flywheel.core.PartialModel;
import com.jozufozu.flywheel.core.materials.model.ModelData;
import com.jozufozu.flywheel.core.virtual.VirtualRenderWorld;
import com.simibubi.create.AllPartialModels;
import com.simibubi.create.content.contraptions.behaviour.MovementContext;
import com.simibubi.create.content.contraptions.render.ActorInstance;
import com.simibubi.create.foundation.utility.AngleHelper;
import com.simibubi.create.foundation.utility.AnimationTickHolder;
import com.simibubi.create.foundation.utility.VecHelper;
import net.minecraft.class_2350;
import net.minecraft.class_243;
import net.minecraft.class_2680;
import net.minecraft.class_2741;

public class HarvesterActorInstance extends ActorInstance {
    static float originOffset = 1 / 16f;
    static class_243 rotOffset = new class_243(0.5f, -2 * originOffset + 0.5f, originOffset + 0.5f);

    protected ModelData harvester;
    private class_2350 facing;

    protected float horizontalAngle;

    private double rotation;
    private double previousRotation;

    public HarvesterActorInstance(MaterialManager materialManager, VirtualRenderWorld simulationWorld, MovementContext context) {
        super(materialManager, simulationWorld, context);

		Material<ModelData> material = materialManager.defaultCutout()
				.material(Materials.TRANSFORMED);

        class_2680 state = context.state;

        facing = state.method_11654(class_2741.field_12481);

        harvester = material.getModel(getRollingPartial(), state).createInstance();

        horizontalAngle = facing.method_10144() + ((facing.method_10166() == class_2350.class_2351.field_11048) ? 180 : 0);

		harvester.setBlockLight(localBlockLight());
	}

	protected PartialModel getRollingPartial() {
		return AllPartialModels.HARVESTER_BLADE;
	}
	
	protected class_243 getRotationOffset() {
		return rotOffset;
	}
	
	protected double getRadius() {
		return 6.5;
	}

	@Override
	public void tick() {
		super.tick();

		previousRotation = rotation;

		if (context.contraption.stalled || context.disabled
			|| VecHelper.isVecPointingTowards(context.relativeMotion, facing.method_10153()))
			return;

		double arcLength = context.motion.method_1033();

		double radians = arcLength * 16 / getRadius();

		float deg = AngleHelper.deg(radians);

		deg = (float) (((int) (deg * 3000)) / 3000);

		rotation += deg * 1.25;

		rotation %= 360;
	}

    @Override
    public void beginFrame() {
        harvester.loadIdentity()
				.translate(context.localPos)
				.centre()
				.rotateY(horizontalAngle)
				.unCentre()
				.translate(getRotationOffset())
				.rotateX(getRotation())
				.translateBack(getRotationOffset());
	}

    protected double getRotation() {
        return AngleHelper.angleLerp(AnimationTickHolder.getPartialTicks(), previousRotation, rotation);
    }
}
