package com.simibubi.create.content.contraptions.bearing;

import org.joml.Quaternionf;

import com.jozufozu.flywheel.api.MaterialManager;
import com.jozufozu.flywheel.core.Materials;
import com.jozufozu.flywheel.core.materials.oriented.OrientedData;
import com.jozufozu.flywheel.core.virtual.VirtualRenderWorld;
import com.simibubi.create.AllPartialModels;
import com.simibubi.create.content.contraptions.behaviour.MovementContext;
import com.simibubi.create.content.contraptions.render.ActorInstance;
import com.simibubi.create.content.kinetics.base.flwdata.RotatingData;
import com.simibubi.create.foundation.render.AllMaterialSpecs;
import com.simibubi.create.foundation.utility.AnimationTickHolder;
import net.minecraft.class_2350;
import net.minecraft.class_2680;
import net.minecraft.class_2741;
import net.minecraft.class_7833;

public class StabilizedBearingInstance extends ActorInstance {

	final OrientedData topInstance;
	final RotatingData shaft;

	final class_2350 facing;
	final class_7833 rotationAxis;
	final Quaternionf blockOrientation;

	public StabilizedBearingInstance(MaterialManager materialManager, VirtualRenderWorld simulationWorld, MovementContext context) {
		super(materialManager, simulationWorld, context);

		class_2680 blockState = context.state;

		facing = blockState.method_11654(class_2741.field_12525);
		rotationAxis = class_7833.method_46356(class_2350.method_10156(class_2350.class_2352.field_11056, facing.method_10166()).method_23955());

		blockOrientation = BearingInstance.getBlockStateOrientation(facing);

        topInstance = materialManager.defaultSolid()
                .material(Materials.ORIENTED)
                .getModel(AllPartialModels.BEARING_TOP, blockState)
				.createInstance();

		int blockLight = localBlockLight();
		topInstance.setPosition(context.localPos)
				.setRotation(blockOrientation)
				.setBlockLight(blockLight);

		shaft = materialManager.defaultSolid()
				.material(AllMaterialSpecs.ROTATING)
				.getModel(AllPartialModels.SHAFT_HALF, blockState, blockState.method_11654(class_2741.field_12525).method_10153())
				.createInstance();

		// not rotating so no need to set speed, axis, etc.
		shaft.setPosition(context.localPos)
				.setBlockLight(blockLight);
	}

	@Override
	public void beginFrame() {
		float counterRotationAngle = StabilizedBearingMovementBehaviour.getCounterRotationAngle(context, facing, AnimationTickHolder.getPartialTicks());

		Quaternionf rotation = rotationAxis.rotationDegrees(counterRotationAngle);

		rotation.mul(blockOrientation);

		topInstance.setRotation(rotation);
	}
}
