package com.simibubi.create.content.contraptions.bearing;

import java.util.List;
import net.minecraft.class_2338;
import net.minecraft.class_2487;
import net.minecraft.class_2591;
import net.minecraft.class_2680;
import net.minecraft.class_3532;
import com.simibubi.create.foundation.advancement.AllAdvancements;
import com.simibubi.create.foundation.blockEntity.behaviour.BlockEntityBehaviour;
import com.simibubi.create.foundation.blockEntity.behaviour.scrollValue.INamedIconOptions;
import com.simibubi.create.foundation.blockEntity.behaviour.scrollValue.ScrollOptionBehaviour;
import com.simibubi.create.foundation.gui.AllIcons;
import com.simibubi.create.foundation.utility.Lang;
import com.simibubi.create.infrastructure.config.AllConfigs;

public class WindmillBearingBlockEntity extends MechanicalBearingBlockEntity {

	protected ScrollOptionBehaviour<RotationDirection> movementDirection;
	protected float lastGeneratedSpeed;

	protected boolean queuedReassembly;

	public WindmillBearingBlockEntity(class_2591<?> type, class_2338 pos, class_2680 state) {
		super(type, pos, state);
	}

	@Override
	public void updateGeneratedRotation() {
		super.updateGeneratedRotation();
		lastGeneratedSpeed = getGeneratedSpeed();
		queuedReassembly = false;
	}

	@Override
	public void onSpeedChanged(float prevSpeed) {
		boolean cancelAssembly = assembleNextTick;
		super.onSpeedChanged(prevSpeed);
		assembleNextTick = cancelAssembly;
	}

	@Override
	public void tick() {
		super.tick();
		if (field_11863.method_8608())
			return;
		if (!queuedReassembly)
			return;
		queuedReassembly = false;
		if (!running)
			assembleNextTick = true;
	}
	
	public void disassembleForMovement() {
		if (!running)
			return;
		disassemble();
		queuedReassembly = true;
	}

	@Override
	public float getGeneratedSpeed() {
		if (!running)
			return 0;
		if (movedContraption == null)
			return lastGeneratedSpeed;
		int sails = ((BearingContraption) movedContraption.getContraption()).getSailBlocks()
			/ AllConfigs.server().kinetics.windmillSailsPerRPM.get();
		return class_3532.method_15340(sails, 1, 16) * getAngleSpeedDirection();
	}

	@Override
	protected boolean isWindmill() {
		return true;
	}

	protected float getAngleSpeedDirection() {
		RotationDirection rotationDirection = RotationDirection.values()[movementDirection.getValue()];
		return (rotationDirection == RotationDirection.CLOCKWISE ? 1 : -1);
	}

	@Override
	public void write(class_2487 compound, boolean clientPacket) {
		compound.method_10548("LastGenerated", lastGeneratedSpeed);
		compound.method_10556("QueueAssembly", queuedReassembly);
		super.write(compound, clientPacket);
	}

	@Override
	protected void read(class_2487 compound, boolean clientPacket) {
		if (!wasMoved)
			lastGeneratedSpeed = compound.method_10583("LastGenerated");
		queuedReassembly = compound.method_10577("QueueAssembly");
		super.read(compound, clientPacket);
	}

	@Override
	public void addBehaviours(List<BlockEntityBehaviour> behaviours) {
		super.addBehaviours(behaviours);
		behaviours.remove(movementMode);
		movementDirection = new ScrollOptionBehaviour<>(RotationDirection.class,
			Lang.translateDirect("contraptions.windmill.rotation_direction"), this, getMovementModeSlot());
		movementDirection.withCallback($ -> onDirectionChanged());
		behaviours.add(movementDirection);
		registerAwardables(behaviours, AllAdvancements.WINDMILL, AllAdvancements.WINDMILL_MAXED);
	}

	private void onDirectionChanged() {
		if (!running)
			return;
		if (!field_11863.field_9236)
			updateGeneratedRotation();
	}

	@Override
	public boolean isWoodenTop() {
		return true;
	}

	public static enum RotationDirection implements INamedIconOptions {

		CLOCKWISE(AllIcons.I_REFRESH), COUNTER_CLOCKWISE(AllIcons.I_ROTATE_CCW),

		;

		private String translationKey;
		private AllIcons icon;

		private RotationDirection(AllIcons icon) {
			this.icon = icon;
			translationKey = "generic." + Lang.asId(name());
		}

		@Override
		public AllIcons getIcon() {
			return icon;
		}

		@Override
		public String getTranslationKey() {
			return translationKey;
		}

	}

}
