package com.simibubi.create.content.contraptions.behaviour;

import javax.annotation.Nullable;

import com.jozufozu.flywheel.api.MaterialManager;
import com.jozufozu.flywheel.core.virtual.VirtualRenderWorld;
import com.simibubi.create.content.contraptions.render.ActorInstance;
import com.simibubi.create.content.contraptions.render.ContraptionMatrices;
import com.simibubi.create.infrastructure.config.AllConfigs;

import io.github.fabricators_of_create.porting_lib.transfer.TransferUtil;
import io.github.fabricators_of_create.porting_lib.transfer.item.ItemHandlerHelper;

import net.fabricmc.api.EnvType;
import net.fabricmc.api.Environment;
import net.fabricmc.fabric.api.transfer.v1.item.ItemVariant;
import net.fabricmc.fabric.api.transfer.v1.transaction.Transaction;
import net.minecraft.class_1542;
import net.minecraft.class_1799;
import net.minecraft.class_2248;
import net.minecraft.class_2338;
import net.minecraft.class_243;
import net.minecraft.class_4597;

public interface MovementBehaviour {

	default boolean isActive(MovementContext context) {
		return !context.disabled;
	}

	default void tick(MovementContext context) {}

	default void startMoving(MovementContext context) {}

	default void visitNewPosition(MovementContext context, class_2338 pos) {}

	default class_243 getActiveAreaOffset(MovementContext context) {
		return class_243.field_1353;
	}

	@Nullable
	default class_1799 canBeDisabledVia(MovementContext context) {
		class_2248 block = context.state.method_26204();
		if (block == null)
			return null;
		return new class_1799(block);
	}

	default void onDisabledByControls(MovementContext context) {
		cancelStall(context);
	}

	default boolean mustTickWhileDisabled() {
		return false;
	}

	default void dropItem(MovementContext context, class_1799 stack) {
		class_1799 remainder;
		if (AllConfigs.server().kinetics.moveItemsToStorage.get()) {
			try (Transaction t = TransferUtil.getTransaction()) {
				long inserted = context.contraption.getSharedInventory().insert(ItemVariant.of(stack), stack.method_7947(), t);
				remainder = stack.method_7972();
				remainder.method_7934((int) inserted);
				t.commit();
			}
		}
		else
			remainder = stack;
		if (remainder.method_7960())
			return;

		// Actors might void items if their positions is undefined
		class_243 vec = context.position;
		if (vec == null)
			return;
		
		class_1542 itemEntity = new class_1542(context.world, vec.field_1352, vec.field_1351, vec.field_1350, remainder);
		itemEntity.method_18799(context.motion.method_1031(0, 0.5f, 0)
			.method_1021(context.world.field_9229.method_43057() * .3f));
		context.world.method_8649(itemEntity);
	}

	default void onSpeedChanged(MovementContext context, class_243 oldMotion, class_243 motion) {}

	default void stopMoving(MovementContext context) {}

	default void cancelStall(MovementContext context) {
		context.stall = false;
	}

	default void writeExtraData(MovementContext context) {}

	default boolean renderAsNormalBlockEntity() {
		return false;
	}

	default boolean hasSpecialInstancedRendering() {
		return false;
	}

	@Environment(EnvType.CLIENT)
	default void renderInContraption(MovementContext context, VirtualRenderWorld renderWorld,
		ContraptionMatrices matrices, class_4597 buffer) {}

	@Environment(EnvType.CLIENT)
	@Nullable
	default ActorInstance createInstance(MaterialManager materialManager, VirtualRenderWorld simulationWorld,
		MovementContext context) {
		return null;
	}
}
