package com.simibubi.create.content.contraptions.behaviour.dispenser;

import java.util.ArrayList;
import java.util.List;
import com.simibubi.create.content.contraptions.behaviour.MovementBehaviour;
import com.simibubi.create.content.contraptions.behaviour.MovementContext;
import com.simibubi.create.foundation.item.ItemHelper;

import io.github.fabricators_of_create.porting_lib.transfer.TransferUtil;
import net.fabricmc.fabric.api.transfer.v1.item.ItemVariant;
import net.fabricmc.fabric.api.transfer.v1.storage.StorageUtil;
import net.fabricmc.fabric.api.transfer.v1.storage.StorageView;
import net.fabricmc.fabric.api.transfer.v1.storage.base.ResourceAmount;
import net.fabricmc.fabric.api.transfer.v1.transaction.Transaction;
import net.minecraft.class_1262;
import net.minecraft.class_1799;
import net.minecraft.class_1802;
import net.minecraft.class_2338;
import net.minecraft.class_2371;
import net.minecraft.class_5819;

public class DropperMovementBehaviour implements MovementBehaviour {
	protected static final MovedDefaultDispenseItemBehaviour DEFAULT_BEHAVIOUR =
		new MovedDefaultDispenseItemBehaviour();
	private static final class_5819 RNG = class_5819.method_43047();

	protected void activate(MovementContext context, class_2338 pos) {
		DispenseItemLocation location = getDispenseLocation(context);
		if (location.isEmpty()) {
			context.world.method_20290(1001, pos, 0);
		} else {
			setItemStackAt(location, DEFAULT_BEHAVIOUR.dispense(getItemStackAt(location, context), context, pos),
				context);
		}
	}

	@Override
	public void visitNewPosition(MovementContext context, class_2338 pos) {
		if (context.world.field_9236)
			return;
		collectItems(context);
		activate(context, pos);
	}

	private void collectItems(MovementContext context) {
		getStacks(context).stream()
			.filter(itemStack -> !itemStack.method_7960() && itemStack.method_7909() != class_1802.field_8162
				&& itemStack.method_7914() > itemStack.method_7947())
			.forEach(itemStack -> itemStack.method_7933(ItemHelper
				.extract(context.contraption.getSharedInventory(), ItemHelper.sameItemPredicate(itemStack),
					ItemHelper.ExtractionCountMode.UPTO, itemStack.method_7914() - itemStack.method_7947(), false)
				.method_7947()));
	}

	private void updateTemporaryData(MovementContext context) {
		if (!(context.temporaryData instanceof class_2371) && context.world != null) {
			class_2371<class_1799> stacks = class_2371.method_10213(getInvSize(), class_1799.field_8037);
			class_1262.method_5429(context.blockEntityData, stacks);
			context.temporaryData = stacks;
		}
	}

	@SuppressWarnings("unchecked")
	private class_2371<class_1799> getStacks(MovementContext context) {
		updateTemporaryData(context);
		return (class_2371<class_1799>) context.temporaryData;
	}

	private ArrayList<DispenseItemLocation> getUseableLocations(MovementContext context) {
		ArrayList<DispenseItemLocation> useable = new ArrayList<>();
		try (Transaction t = TransferUtil.getTransaction()) {
			for (int slot = 0; slot < getInvSize(); slot++) {
				DispenseItemLocation location = new DispenseItemLocation(slot);
				class_1799 testStack = getItemStackAt(location, context);
				if (testStack == null || testStack.method_7960())
					continue;
				if (testStack.method_7914() == 1) {
					ResourceAmount<ItemVariant> available = StorageUtil.findExtractableContent(context.contraption.getSharedInventory(), v -> v.matches(testStack), t);
					if (available != null) {
						location = new DispenseItemLocation(available);
						useable.add(location);
					}
				} else if (testStack.method_7947() >= 2)
					useable.add(location);
			}
			return useable;
		}
	}

	@Override
	public void writeExtraData(MovementContext context) {
		class_2371<class_1799> stacks = getStacks(context);
		if (stacks == null)
			return;
		class_1262.method_5426(context.blockEntityData, stacks);
	}

	@Override
	public void stopMoving(MovementContext context) {
		MovementBehaviour.super.stopMoving(context);
		writeExtraData(context);
	}

	protected DispenseItemLocation getDispenseLocation(MovementContext context) {
		int i = -1;
		int j = 1;
		List<DispenseItemLocation> useableLocations = getUseableLocations(context);
		for (int k = 0; k < useableLocations.size(); ++k) {
			if (RNG.method_43048(j++) == 0) {
				i = k;
			}
		}
		if (i < 0)
			return DispenseItemLocation.NONE;
		else
			return useableLocations.get(i);
	}

	protected class_1799 getItemStackAt(DispenseItemLocation location, MovementContext context) {
		if (location.isInternal()) {
			return getStacks(context).get(location.getSlot());
		} else {
			return location.getVariant().toStack(location.getCount());
		}
	}

	protected void setItemStackAt(DispenseItemLocation location, class_1799 stack, MovementContext context) {
		if (location.isInternal()) {
			getStacks(context).set(location.getSlot(), stack);
		} else {
			try (Transaction t = TransferUtil.getTransaction()) {
				context.contraption.getSharedInventory()
				.extract(location.getVariant(), location.getCount(), t);
				context.contraption.getSharedInventory().insert(ItemVariant.of(stack), stack.method_7947(), t);
				t.commit();
			}
		}
	}

	private static int getInvSize() {
		return 9;
	}
}
