package com.simibubi.create.content.contraptions.gantry;

import com.jozufozu.flywheel.api.MaterialManager;
import com.jozufozu.flywheel.api.instance.DynamicInstance;
import com.jozufozu.flywheel.core.materials.model.ModelData;
import com.simibubi.create.AllPartialModels;
import com.simibubi.create.content.kinetics.base.KineticBlockEntityRenderer;
import com.simibubi.create.content.kinetics.base.ShaftInstance;
import com.simibubi.create.foundation.utility.AngleHelper;
import com.simibubi.create.foundation.utility.Iterate;
import net.minecraft.class_2338;
import net.minecraft.class_2350;
import net.minecraft.class_3532;

public class GantryCarriageInstance extends ShaftInstance<GantryCarriageBlockEntity> implements DynamicInstance {

	private final ModelData gantryCogs;

	final class_2350 facing;
	final Boolean alongFirst;
	final class_2350.class_2351 rotationAxis;
	final float rotationMult;
	final class_2338 visualPos;

	private float lastAngle = Float.NaN;

	public GantryCarriageInstance(MaterialManager materialManager, GantryCarriageBlockEntity blockEntity) {
		super(materialManager, blockEntity);

		gantryCogs = getTransformMaterial()
								 .getModel(AllPartialModels.GANTRY_COGS, blockState)
								 .createInstance();

		facing = blockState.method_11654(GantryCarriageBlock.FACING);
		alongFirst = blockState.method_11654(GantryCarriageBlock.AXIS_ALONG_FIRST_COORDINATE);
		rotationAxis = KineticBlockEntityRenderer.getRotationAxisOf(blockEntity);

		rotationMult = getRotationMultiplier(getGantryAxis(), facing);

		visualPos = facing.method_10171() == class_2350.class_2352.field_11056 ? blockEntity.method_11016()
				: blockEntity.method_11016()
					  .method_10093(facing.method_10153());

		animateCogs(getCogAngle());
	}

	@Override
	public void beginFrame() {
		float cogAngle = getCogAngle();

		if (class_3532.method_15347(cogAngle, lastAngle)) return;

		animateCogs(cogAngle);
	}

	private float getCogAngle() {
		return GantryCarriageRenderer.getAngleForBE(blockEntity, visualPos, rotationAxis) * rotationMult;
	}

	private void animateCogs(float cogAngle) {
		gantryCogs.loadIdentity()
				.translate(getInstancePosition())
				.centre()
				.rotateY(AngleHelper.horizontalAngle(facing))
				.rotateX(facing == class_2350.field_11036 ? 0 : facing == class_2350.field_11033 ? 180 : 90)
				.rotateY(alongFirst ^ facing.method_10166() == class_2350.class_2351.field_11048 ? 0 : 90)
				.translate(0, -9 / 16f, 0)
				.rotateX(-cogAngle)
				.translate(0, 9 / 16f, 0)
				.unCentre();
	}

	static float getRotationMultiplier(class_2350.class_2351 gantryAxis, class_2350 facing) {
		float multiplier = 1;
		if (gantryAxis == class_2350.class_2351.field_11048)
			if (facing == class_2350.field_11036)
				multiplier *= -1;
		if (gantryAxis == class_2350.class_2351.field_11052)
			if (facing == class_2350.field_11043 || facing == class_2350.field_11034)
				multiplier *= -1;

		return multiplier;
	}

	private class_2350.class_2351 getGantryAxis() {
		class_2350.class_2351 gantryAxis = class_2350.class_2351.field_11048;
		for (class_2350.class_2351 axis : Iterate.axes)
			if (axis != rotationAxis && axis != facing.method_10166())
				gantryAxis = axis;
		return gantryAxis;
	}

	@Override
	public void updateLight() {
		relight(pos, gantryCogs, rotatingModel);
	}

	@Override
	public void remove() {
		super.remove();
		gantryCogs.delete();
	}
}
