package com.simibubi.create.content.contraptions.minecart.capability;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.UUID;

import javax.annotation.Nullable;
import net.minecraft.class_1297;
import net.minecraft.class_1542;
import net.minecraft.class_1688;
import net.minecraft.class_1923;
import net.minecraft.class_1937;
import net.minecraft.class_243;
import net.minecraft.class_2487;
import net.minecraft.class_2818;
import com.simibubi.create.AllItems;
import com.simibubi.create.content.contraptions.minecart.CouplingHandler;
import com.simibubi.create.foundation.utility.Iterate;
import com.simibubi.create.foundation.utility.WorldAttached;
import com.simibubi.create.foundation.utility.fabric.AbstractMinecartExtensions;

import io.github.fabricators_of_create.porting_lib.core.util.INBTSerializable;
import io.github.fabricators_of_create.porting_lib.util.LazyOptional;

import it.unimi.dsi.fastutil.objects.ObjectArrayList;
import it.unimi.dsi.fastutil.objects.ObjectLists;

public class CapabilityMinecartController implements INBTSerializable<class_2487> {

	/* Global map of loaded carts */

	public static WorldAttached<Map<UUID, MinecartController>> loadedMinecartsByUUID;
	public static WorldAttached<Set<UUID>> loadedMinecartsWithCoupling;
	static WorldAttached<List<class_1688>> queuedAdditions;
	static WorldAttached<List<UUID>> queuedUnloads;

// fabric: we just manually deal with removal to avoid the weirdness of adding 5 listeners but only ever calling accept on one.
//	/**
//	 * This callback wrapper ensures that the listeners map in the controller
//	 * capability only ever contains one instance
//	 */
//	public static class MinecartRemovalListener implements NonNullConsumer<LazyOptional<MinecartController>> {
//
//		private Level world;
//		private AbstractMinecart cart;
//
//		public MinecartRemovalListener(Level world, AbstractMinecart cart) {
//			this.world = world;
//			this.cart = cart;
//		}
//
//		@Override
//		public boolean equals(Object obj) {
//			return obj instanceof MinecartRemovalListener;
//		}
//
//		@Override
//		public int hashCode() {
//			return 100;
//		}
//
//		@Override
//		public void accept(LazyOptional<MinecartController> t) {
//			onCartRemoved(world, cart);
//		}
//
//	}

	static {
		loadedMinecartsByUUID = new WorldAttached<>($ -> new HashMap<>());
		loadedMinecartsWithCoupling = new WorldAttached<>($ -> new HashSet<>());
		queuedAdditions = new WorldAttached<>($ -> ObjectLists.synchronize(new ObjectArrayList<>()));
		queuedUnloads = new WorldAttached<>($ -> ObjectLists.synchronize(new ObjectArrayList<>()));
	}

	public static void tick(class_1937 world) {
		List<UUID> toRemove = new ArrayList<>();
		Map<UUID, MinecartController> carts = loadedMinecartsByUUID.get(world);
		List<class_1688> queued = queuedAdditions.get(world);
		List<UUID> queuedRemovals = queuedUnloads.get(world);
		Set<UUID> cartsWithCoupling = loadedMinecartsWithCoupling.get(world);
		Set<UUID> keySet = carts.keySet();

		keySet.removeAll(queuedRemovals);
		cartsWithCoupling.removeAll(queuedRemovals);

		for (class_1688 cart : queued) {
			UUID uniqueID = cart.method_5667();

			if (world.field_9236 && carts.containsKey(uniqueID)) {
				MinecartController minecartController = carts.get(uniqueID);
				if (minecartController != null) {
					class_1688 minecartEntity = minecartController.cart();
					if (minecartEntity != null && minecartEntity.method_5628() != cart.method_5628())
						continue; // Away with you, Fake Entities!
				}
			}

			cartsWithCoupling.remove(uniqueID);

			MinecartController controller = cart.create$getController();
//			capability.addListener(new MinecartRemovalListener(world, cart)); // fabric: handled via AbstractMinecartMixin
			carts.put(uniqueID, controller);

			if (controller.isLeadingCoupling())
				cartsWithCoupling.add(uniqueID);
			if (!world.field_9236 && controller != null)
				controller.sendData();
		}

		queuedRemovals.clear();
		queued.clear();

		for (Entry<UUID, MinecartController> entry : carts.entrySet()) {
			MinecartController controller = entry.getValue();
			if (controller != null) {
				if (controller.isPresent()) {
					controller.tick();
					continue;
				}
			}
			toRemove.add(entry.getKey());
		}

		cartsWithCoupling.removeAll(toRemove);
		keySet.removeAll(toRemove);
	}

	public static void onChunkUnloaded(class_1937 world, class_2818 chunk) {
		class_1923 chunkPos = chunk
			.method_12004();
		Map<UUID, MinecartController> carts = loadedMinecartsByUUID.get(world);
		for (MinecartController minecartController : carts.values()) {
			if (minecartController == null)
				continue;
			if (!minecartController.isPresent())
				continue;
			class_1688 cart = minecartController.cart();
			if (cart.method_31476()
				.equals(chunkPos))
				queuedUnloads.get(world)
					.add(cart.method_5667());
		}
	}

	public static void onCartRemoved(class_1937 world, class_1688 entity) {
		Map<UUID, MinecartController> carts = loadedMinecartsByUUID.get(world);
		List<UUID> unloads = queuedUnloads.get(world);
		UUID uniqueID = entity.method_5667();
		if (!carts.containsKey(uniqueID) || unloads.contains(uniqueID))
			return;
		if (world.field_9236)
			return;
		handleKilledMinecart(world, carts.get(uniqueID), entity.method_19538());
	}

	protected static void handleKilledMinecart(class_1937 world, MinecartController controller, class_243 removedPos) {
		if (controller == null)
			return;
		for (boolean forward : Iterate.trueAndFalse) {
			MinecartController next = CouplingHandler.getNextInCouplingChain(world, controller, forward);
			if (next == null || next == MinecartController.EMPTY)
				continue;

			next.removeConnection(!forward);
			if (controller.hasContraptionCoupling(forward))
				continue;
			class_1688 cart = next.cart();
			if (cart == null)
				continue;

			class_243 itemPos = cart.method_19538()
				.method_1019(removedPos)
				.method_1021(.5f);
			class_1542 itemEntity =
				new class_1542(world, itemPos.field_1352, itemPos.field_1351, itemPos.field_1350, AllItems.MINECART_COUPLING.asStack());
			itemEntity.method_6988();
			world.method_8649(itemEntity);
		}
	}

	@Nullable
	public static MinecartController getIfPresent(class_1937 world, UUID cartId) {
		Map<UUID, MinecartController> carts = loadedMinecartsByUUID.get(world);
		if (carts == null)
			return null;
		if (!carts.containsKey(cartId))
			return null;
		return carts.get(cartId);
	}

	public static void attach(class_1688 entity) {
		queuedAdditions.get(entity.method_5770())
			.add(entity);
	}

	public static void startTracking(class_1297 entity) {
		if (!(entity instanceof class_1688 cart))
			return;
		cart.create$getController().sendData();
	}
}
