package com.simibubi.create.content.contraptions.pulley;

import com.jozufozu.flywheel.api.Instancer;
import com.jozufozu.flywheel.api.MaterialManager;
import com.jozufozu.flywheel.api.instance.DynamicInstance;
import com.jozufozu.flywheel.core.instancing.ConditionalInstance;
import com.jozufozu.flywheel.core.instancing.GroupInstance;
import com.jozufozu.flywheel.core.instancing.SelectInstance;
import com.jozufozu.flywheel.core.materials.oriented.OrientedData;
import com.jozufozu.flywheel.light.LightPacking;
import com.jozufozu.flywheel.light.LightVolume;
import com.jozufozu.flywheel.light.TickingLightListener;
import com.jozufozu.flywheel.util.box.GridAlignedBB;
import com.jozufozu.flywheel.util.box.ImmutableBox;
import com.simibubi.create.content.kinetics.base.KineticBlockEntity;
import com.simibubi.create.content.kinetics.base.ShaftInstance;
import net.minecraft.class_1944;
import net.minecraft.class_2350;
import net.minecraft.class_3532;
import net.minecraft.class_7833;

public abstract class AbstractPulleyInstance<T extends KineticBlockEntity> extends ShaftInstance<T> implements DynamicInstance, TickingLightListener {

	final OrientedData coil;
	final SelectInstance<OrientedData> magnet;
	final GroupInstance<OrientedData> rope;
	final ConditionalInstance<OrientedData> halfRope;

	protected float offset;
	protected final class_2350 rotatingAbout;
	protected final class_7833 rotationAxis;

	private final GridAlignedBB volume = new GridAlignedBB();
	private final LightVolume light;

	public AbstractPulleyInstance(MaterialManager dispatcher, T blockEntity) {
		super(dispatcher, blockEntity);

		rotatingAbout = class_2350.method_10156(class_2350.class_2352.field_11056, axis);
		rotationAxis = class_7833.method_46356(rotatingAbout.method_23955());

		coil = getCoilModel().createInstance()
				.setPosition(getInstancePosition());

		magnet = new SelectInstance<>(this::getMagnetModelIndex);
		magnet.addModel(getMagnetModel())
				.addModel(getHalfMagnetModel());

		rope = new GroupInstance<>(getRopeModel());
		halfRope = new ConditionalInstance<>(getHalfRopeModel()).withCondition(this::shouldRenderHalfRope);

		updateOffset();
		updateVolume();

		light = new LightVolume(world, volume);
		light.initialize();
	}

	@Override
	public void beginFrame() {
		updateOffset();
		coil.setRotation(rotationAxis.rotationDegrees(offset * 180));

		int neededRopeCount = getNeededRopeCount();
		rope.resize(neededRopeCount);

		magnet.update()
				.get()
				.ifPresent(data -> {
					int i = Math.max(0, class_3532.method_15375(offset));
					short packed = light.getPackedLight(pos.method_10263(), pos.method_10264() - i, pos.method_10260());
					data.setPosition(getInstancePosition())
							.nudge(0, -offset, 0)
							.setBlockLight(LightPacking.getBlock(packed))
							.setSkyLight(LightPacking.getSky(packed));
				});

		halfRope.update()
				.get()
				.ifPresent(rope1 -> {
					float f = offset % 1;
					float halfRopeNudge = f > .75f ? f - 1 : f;

					short packed = light.getPackedLight(pos.method_10263(), pos.method_10264(), pos.method_10260());
					rope1.setPosition(getInstancePosition())
							.nudge(0, -halfRopeNudge, 0)
							.setBlockLight(LightPacking.getBlock(packed))
							.setSkyLight(LightPacking.getSky(packed));
				});

		if (isRunning()) {
			int size = rope.size();
			int bottomY = pos.method_10264() - size;
			for (int i = 0; i < size; i++) {
				short packed = light.getPackedLight(pos.method_10263(), bottomY + i, pos.method_10260());

				rope.get(i)
						.setPosition(getInstancePosition())
						.nudge(0, -offset + i + 1, 0)
						.setBlockLight(LightPacking.getBlock(packed))
						.setSkyLight(LightPacking.getSky(packed));
			}
		} else {
			rope.clear();
		}
	}

	@Override
	public void updateLight() {
		super.updateLight();
		relight(pos, coil);
	}

	@Override
	public void remove() {
		super.remove();
		coil.delete();
		magnet.delete();
		rope.clear();
		halfRope.delete();
		light.delete();
	}

	protected abstract Instancer<OrientedData> getRopeModel();

	protected abstract Instancer<OrientedData> getMagnetModel();

	protected abstract Instancer<OrientedData> getHalfMagnetModel();

	protected abstract Instancer<OrientedData> getCoilModel();

	protected abstract Instancer<OrientedData> getHalfRopeModel();

	protected abstract float getOffset();

	protected abstract boolean isRunning();

	@Override
	public boolean tickLightListener() {
		if (updateVolume()) {
			light.move(volume);
			return true;
		}
		return false;
	}

	private boolean updateVolume() {
		int length = class_3532.method_15386(offset) + 2;

		if (volume.sizeY() < length) {
			volume.assign(pos.method_10087(length), pos)
					.fixMinMax();
			return true;
		}
		return false;
	}

	private void updateOffset() {
		offset = getOffset();
	}

	private int getNeededRopeCount() {
		return Math.max(0, class_3532.method_15386(offset - 1.25f));
	}

	private boolean shouldRenderHalfRope() {
		float f = offset % 1;
		return offset > .75f && (f < .25f || f > .75f);
	}

	private int getMagnetModelIndex() {
		if (isRunning() || offset == 0) {
			return offset > .25f ? 0 : 1;
		} else {
			return -1;
		}
	}

	@Override
	public boolean decreaseFramerateWithDistance() {
		return false;
	}

	@Override
	public ImmutableBox getVolume() {
		return volume;
	}

	@Override
	public void onLightUpdate(class_1944 type, ImmutableBox changed) {
		super.onLightUpdate(type, changed);
		light.onLightUpdate(type, changed);
	}
}
