package com.simibubi.create.content.contraptions.render;

import java.util.Collection;

import org.apache.commons.lang3.tuple.Pair;

import com.jozufozu.flywheel.backend.Backend;
import com.jozufozu.flywheel.backend.gl.error.GlError;
import com.jozufozu.flywheel.config.BackendType;
import com.jozufozu.flywheel.core.model.ShadeSeparatedBufferedData;
import com.jozufozu.flywheel.core.model.WorldModelBuilder;
import com.jozufozu.flywheel.core.virtual.VirtualRenderWorld;
import com.jozufozu.flywheel.event.BeginFrameEvent;
import com.jozufozu.flywheel.event.GatherContextEvent;
import com.jozufozu.flywheel.event.ReloadRenderersEvent;
import com.jozufozu.flywheel.event.RenderLayerEvent;
import com.jozufozu.flywheel.util.WorldAttached;
import com.jozufozu.flywheel.util.transform.TransformStack;
import com.simibubi.create.AllMovementBehaviours;
import com.simibubi.create.content.contraptions.AbstractContraptionEntity;
import com.simibubi.create.content.contraptions.Contraption;
import com.simibubi.create.content.contraptions.ContraptionWorld;
import com.simibubi.create.content.contraptions.behaviour.MovementBehaviour;
import com.simibubi.create.content.contraptions.behaviour.MovementContext;
import com.simibubi.create.foundation.render.BlockEntityRenderHelper;
import com.simibubi.create.foundation.render.SuperByteBuffer;

import net.fabricmc.api.EnvType;
import net.fabricmc.api.Environment;
import net.minecraft.class_1921;
import net.minecraft.class_1937;
import net.minecraft.class_1944;
import net.minecraft.class_2338;
import net.minecraft.class_310;
import net.minecraft.class_3499;
import net.minecraft.class_4587;
import net.minecraft.class_4597;
import net.minecraft.class_761;
import net.minecraft.class_765;

@Environment(EnvType.CLIENT)
public class ContraptionRenderDispatcher {

	private static WorldAttached<ContraptionRenderingWorld<?>> WORLDS = new WorldAttached<>(SBBContraptionManager::new);

	/**
	 * Reset a contraption's renderer.
	 *
	 * @param contraption The contraption to invalidate.
	 * @return true if there was a renderer associated with the given contraption.
	 */
	public static boolean invalidate(Contraption contraption) {
		class_1937 level = contraption.entity.method_37908();

		return WORLDS.get(level)
			.invalidate(contraption);
	}

	public static void tick(class_1937 world) {
		if (class_310.method_1551()
			.method_1493())
			return;

		WORLDS.get(world)
			.tick();
	}

	public static void beginFrame(BeginFrameEvent event) {
		WORLDS.get(event.getWorld())
			.beginFrame(event);
	}

	public static void renderLayer(RenderLayerEvent event) {
		WORLDS.get(event.getWorld())
			.renderLayer(event);

		GlError.pollAndThrow(() -> "contraption layer: " + event.getLayer());
	}

	public static void onRendererReload(ReloadRenderersEvent event) {
		reset();
	}

	public static void gatherContext(GatherContextEvent e) {
		reset();
	}

	public static void renderFromEntity(AbstractContraptionEntity entity, Contraption contraption,
		class_4597 buffers) {
		class_1937 world = entity.method_37908();

		ContraptionRenderInfo renderInfo = WORLDS.get(world)
			.getRenderInfo(contraption);
		ContraptionMatrices matrices = renderInfo.getMatrices();

		// something went wrong with the other rendering
		if (!matrices.isReady())
			return;

		VirtualRenderWorld renderWorld = renderInfo.renderWorld;

		renderBlockEntities(world, renderWorld, contraption, matrices, buffers);

		if (buffers instanceof class_4597.class_4598)
			((class_4597.class_4598) buffers).method_22993();

		renderActors(world, renderWorld, contraption, matrices, buffers);
	}

	public static VirtualRenderWorld setupRenderWorld(class_1937 world, Contraption c) {
		ContraptionWorld contraptionWorld = c.getContraptionWorld();

		class_2338 origin = c.anchor;
		int minBuildHeight = contraptionWorld.method_31607();
		int height = contraptionWorld.method_31605();
		VirtualRenderWorld renderWorld = new VirtualRenderWorld(world, minBuildHeight, height, origin) {
			@Override
			public boolean supportsFlywheel() {
				return canInstance();
			}
		};

		renderWorld.setBlockEntities(c.presentBlockEntities.values());
		for (class_3499.class_3501 info : c.getBlocks()
			.values())
			// Skip individual lighting updates to prevent lag with large contraptions
			// FIXME 1.20 this '0' used to be Block.UPDATE_SUPPRESS_LIGHT, yet VirtualRenderWorld didn't actually parse the flags at all
			renderWorld.method_8652(info.comp_1341(), info.comp_1342(), 0);

		renderWorld.runLightEngine();
		return renderWorld;
	}

	public static void renderBlockEntities(class_1937 world, VirtualRenderWorld renderWorld, Contraption c,
		ContraptionMatrices matrices, class_4597 buffer) {
		BlockEntityRenderHelper.renderBlockEntities(world, renderWorld, c.getSpecialRenderedBEs(),
			matrices.getModelViewProjection(), matrices.getLight(), buffer);
	}

	protected static void renderActors(class_1937 world, VirtualRenderWorld renderWorld, Contraption c,
		ContraptionMatrices matrices, class_4597 buffer) {
		class_4587 m = matrices.getModel();

		for (Pair<class_3499.class_3501, MovementContext> actor : c.getActors()) {
			MovementContext context = actor.getRight();
			if (context == null)
				continue;
			if (context.world == null)
				context.world = world;
			class_3499.class_3501 blockInfo = actor.getLeft();

			MovementBehaviour movementBehaviour = AllMovementBehaviours.getBehaviour(blockInfo.comp_1342());
			if (movementBehaviour != null) {
				if (c.isHiddenInPortal(blockInfo.comp_1341()))
					continue;
				m.method_22903();
				TransformStack.cast(m)
					.translate(blockInfo.comp_1341());
				movementBehaviour.renderInContraption(context, renderWorld, matrices, buffer);
				m.method_22909();
			}
		}
	}

	public static SuperByteBuffer buildStructureBuffer(VirtualRenderWorld renderWorld, Contraption c,
		class_1921 layer) {
		Collection<class_3499.class_3501> values = c.getRenderedBlocks();
		ShadeSeparatedBufferedData data = new WorldModelBuilder(layer).withRenderWorld(renderWorld)
				.withBlocks(values)
				.build();
		SuperByteBuffer sbb = new SuperByteBuffer(data);
		data.release();
		return sbb;
	}

	public static int getLight(class_1937 world, float lx, float ly, float lz) {
		class_2338.class_2339 pos = new class_2338.class_2339();
		float block = 0, sky = 0;
		float offset = 1 / 8f;

		for (float zOffset = offset; zOffset >= -offset; zOffset -= 2 * offset)
			for (float yOffset = offset; yOffset >= -offset; yOffset -= 2 * offset)
				for (float xOffset = offset; xOffset >= -offset; xOffset -= 2 * offset) {
					pos.method_10102(lx + xOffset, ly + yOffset, lz + zOffset);
					block += world.method_8314(class_1944.field_9282, pos) / 8f;
					sky += world.method_8314(class_1944.field_9284, pos) / 8f;
				}

		return class_765.method_23687((int) block, (int) sky);
	}

	public static int getContraptionWorldLight(MovementContext context, VirtualRenderWorld renderWorld) {
		return class_761.method_23794(renderWorld, context.localPos);
	}

	public static void reset() {
		WORLDS.empty(ContraptionRenderingWorld::delete);

		if (Backend.isOn()) {
			WORLDS = new WorldAttached<>(FlwContraptionManager::new);
		} else {
			WORLDS = new WorldAttached<>(SBBContraptionManager::new);
		}
	}

	public static boolean canInstance() {
		return Backend.getBackendType() == BackendType.INSTANCING;
	}
}
