package com.simibubi.create.content.equipment.potatoCannon;

import java.util.HashMap;
import java.util.IdentityHashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.function.Supplier;

import com.google.gson.Gson;
import com.google.gson.JsonElement;
import com.google.gson.JsonObject;
import com.simibubi.create.AllItems;
import com.simibubi.create.Create;
import com.simibubi.create.AllPackets;
import com.simibubi.create.foundation.networking.SimplePacketBase;

import net.fabricmc.fabric.api.resource.IdentifiableResourceReloadListener;
import net.minecraft.class_1792;
import net.minecraft.class_1799;
import net.minecraft.class_2540;
import net.minecraft.class_2960;
import net.minecraft.class_3222;
import net.minecraft.class_3300;
import net.minecraft.class_3695;
import net.minecraft.class_4309;

public class PotatoProjectileTypeManager {

	private static final Map<class_2960, PotatoCannonProjectileType> BUILTIN_TYPE_MAP = new HashMap<>();
	private static final Map<class_2960, PotatoCannonProjectileType> CUSTOM_TYPE_MAP = new HashMap<>();
	private static final Map<class_1792, PotatoCannonProjectileType> ITEM_TO_TYPE_MAP = new IdentityHashMap<>();

	public static void registerBuiltinType(class_2960 id, PotatoCannonProjectileType type) {
		synchronized (BUILTIN_TYPE_MAP) {
			BUILTIN_TYPE_MAP.put(id, type);
		}
	}

	public static PotatoCannonProjectileType getBuiltinType(class_2960 id) {
		return BUILTIN_TYPE_MAP.get(id);
	}

	public static PotatoCannonProjectileType getCustomType(class_2960 id) {
		return CUSTOM_TYPE_MAP.get(id);
	}

	public static PotatoCannonProjectileType getTypeForItem(class_1792 item) {
		return ITEM_TO_TYPE_MAP.get(item);
	}

	public static Optional<PotatoCannonProjectileType> getTypeForStack(class_1799 item) {
		if (item.method_7960())
			return Optional.empty();
		return Optional.ofNullable(getTypeForItem(item.method_7909()));
	}

	public static void clear() {
		CUSTOM_TYPE_MAP.clear();
		ITEM_TO_TYPE_MAP.clear();
	}

	public static void fillItemMap() {
		for (Map.Entry<class_2960, PotatoCannonProjectileType> entry : BUILTIN_TYPE_MAP.entrySet()) {
			PotatoCannonProjectileType type = entry.getValue();
			for (Supplier<class_1792> delegate : type.getItems()) {
				ITEM_TO_TYPE_MAP.put(delegate.get(), type);
			}
		}
		for (Map.Entry<class_2960, PotatoCannonProjectileType> entry : CUSTOM_TYPE_MAP.entrySet()) {
			PotatoCannonProjectileType type = entry.getValue();
			for (Supplier<class_1792> delegate : type.getItems()) {
				ITEM_TO_TYPE_MAP.put(delegate.get(), type);
			}
		}
		ITEM_TO_TYPE_MAP.remove(AllItems.POTATO_CANNON.get());
	}

	public static void toBuffer(class_2540 buffer) {
		buffer.method_10804(CUSTOM_TYPE_MAP.size());
		for (Map.Entry<class_2960, PotatoCannonProjectileType> entry : CUSTOM_TYPE_MAP.entrySet()) {
			buffer.method_10812(entry.getKey());
			PotatoCannonProjectileType.toBuffer(entry.getValue(), buffer);
		}
	}

	public static void fromBuffer(class_2540 buffer) {
		clear();

		int size = buffer.method_10816();
		for (int i = 0; i < size; i++) {
			CUSTOM_TYPE_MAP.put(buffer.method_10810(), PotatoCannonProjectileType.fromBuffer(buffer));
		}

		fillItemMap();
	}

	public static void syncTo(class_3222 player) {
		AllPackets.getChannel().sendToClient(new SyncPacket(), player);
	}

	public static void syncToAll(List<class_3222> players) {
		AllPackets.getChannel().sendToClients(new SyncPacket(), players);
	}

	public static class ReloadListener extends class_4309 implements IdentifiableResourceReloadListener {

		private static final Gson GSON = new Gson();

		public static final ReloadListener INSTANCE = new ReloadListener();

		protected ReloadListener() {
			super(GSON, "potato_cannon_projectile_types");
		}

		@Override
		protected void apply(Map<class_2960, JsonElement> map, class_3300 resourceManager, class_3695 profiler) {
			clear();

			for (Map.Entry<class_2960, JsonElement> entry : map.entrySet()) {
				JsonElement element = entry.getValue();
				if (element.isJsonObject()) {
					class_2960 id = entry.getKey();
					JsonObject object = element.getAsJsonObject();
					PotatoCannonProjectileType type = PotatoCannonProjectileType.fromJson(object);
					CUSTOM_TYPE_MAP.put(id, type);
				}
			}

			fillItemMap();
		}

		@Override
		public class_2960 getFabricId() {
			return Create.asResource("potato_cannon_projectile_types");
		}
	}

	public static class SyncPacket extends SimplePacketBase {

		private class_2540 buffer;

		public SyncPacket() {
		}

		public SyncPacket(class_2540 buffer) {
			this.buffer = buffer;
		}

		@Override
		public void write(class_2540 buffer) {
			toBuffer(buffer);
		}

		@Override
		public boolean handle(Context context) {
			buffer.retain();
			context.enqueueWork(() -> {
				try {
					fromBuffer(buffer);
				} finally {
					buffer.release();
				}
			});
			return true;
		}

	}

}
