package com.simibubi.create.content.equipment.symmetryWand.mirror;

import java.util.HashMap;
import java.util.List;
import java.util.Map;

import com.google.common.collect.ImmutableList;
import com.jozufozu.flywheel.core.PartialModel;
import com.simibubi.create.foundation.utility.Lang;
import net.fabricmc.api.EnvType;
import net.fabricmc.api.Environment;
import net.minecraft.class_2338;
import net.minecraft.class_2350;
import net.minecraft.class_2415;
import net.minecraft.class_243;
import net.minecraft.class_2470;
import net.minecraft.class_2487;
import net.minecraft.class_2494;
import net.minecraft.class_2499;
import net.minecraft.class_2561;
import net.minecraft.class_2680;
import net.minecraft.class_2741;
import net.minecraft.class_2753;
import net.minecraft.class_2769;
import net.minecraft.class_3542;
import net.minecraft.class_4587;

public abstract class SymmetryMirror {

	public static final String EMPTY = "empty";
	public static final String PLANE = "plane";
	public static final String CROSS_PLANE = "cross_plane";
	public static final String TRIPLE_PLANE = "triple_plane";

	protected class_243 position;
	protected class_3542 orientation;
	protected int orientationIndex;
	public boolean enable;

	public SymmetryMirror(class_243 pos) {
		position = pos;
		enable = true;
		orientationIndex = 0;
	}

	public static List<class_2561> getMirrors() {
		return ImmutableList.of(Lang.translateDirect("symmetry.mirror.plane"), Lang.translateDirect("symmetry.mirror.doublePlane"),
			Lang.translateDirect("symmetry.mirror.triplePlane"));
	}

	public class_3542 getOrientation() {
		return orientation;
	}

	public class_243 getPosition() {
		return position;
	}

	public int getOrientationIndex() {
		return orientationIndex;
	}

	public void rotate(boolean forward) {
		orientationIndex += forward ? 1 : -1;
		setOrientation();
	}

	public void process(Map<class_2338, class_2680> blocks) {
		Map<class_2338, class_2680> result = new HashMap<>();
		for (class_2338 pos : blocks.keySet()) {
			result.putAll(process(pos, blocks.get(pos)));
		}
		blocks.putAll(result);
	}

	public abstract Map<class_2338, class_2680> process(class_2338 position, class_2680 block);

	protected abstract void setOrientation();

	public abstract void setOrientation(int index);

	public abstract String typeName();

	@Environment(EnvType.CLIENT)
	public abstract PartialModel getModel();

	public void applyModelTransform(class_4587 ms) {}

	private static final String $ORIENTATION = "direction";
	private static final String $POSITION = "pos";
	private static final String $TYPE = "type";
	private static final String $ENABLE = "enable";

	public class_2487 writeToNbt() {
		class_2487 nbt = new class_2487();
		nbt.method_10569($ORIENTATION, orientationIndex);

		class_2499 floatList = new class_2499();
		floatList.add(class_2494.method_23244((float) position.field_1352));
		floatList.add(class_2494.method_23244((float) position.field_1351));
		floatList.add(class_2494.method_23244((float) position.field_1350));
		nbt.method_10566($POSITION, floatList);
		nbt.method_10582($TYPE, typeName());
		nbt.method_10556($ENABLE, enable);

		return nbt;
	}

	public static SymmetryMirror fromNBT(class_2487 nbt) {
		class_2499 floatList = nbt.method_10554($POSITION, 5);
		class_243 pos = new class_243(floatList.method_10604(0), floatList.method_10604(1), floatList.method_10604(2));
		SymmetryMirror element;

		switch (nbt.method_10558($TYPE)) {
		case PLANE:
			element = new PlaneMirror(pos);
			break;
		case CROSS_PLANE:
			element = new CrossPlaneMirror(pos);
			break;
		case TRIPLE_PLANE:
			element = new TriplePlaneMirror(pos);
			break;
		default:
			element = new EmptyMirror(pos);
			break;
		}

		element.setOrientation(nbt.method_10550($ORIENTATION));
		element.enable = nbt.method_10577($ENABLE);

		return element;
	}

	protected class_243 getDiff(class_2338 position) {
		return this.position.method_1021(-1)
			.method_1031(position.method_10263(), position.method_10264(), position.method_10260());
	}

	protected class_2338 getIDiff(class_2338 position) {
		class_243 diff = getDiff(position);
		return new class_2338((int) diff.field_1352, (int) diff.field_1351, (int) diff.field_1350);
	}

	protected class_2680 flipX(class_2680 in) {
		return in.method_26185(class_2415.field_11301);
	}

	protected class_2680 flipY(class_2680 in) {
		for (class_2769<?> property : in.method_28501()) {

			if (property == class_2741.field_12518)
				return in.method_28493(property);
			// Directional Blocks
			if (property instanceof class_2753) {
				if (in.method_11654(property) == class_2350.field_11033) {
					return in.method_11657((class_2753) property, class_2350.field_11036);
				} else if (in.method_11654(property) == class_2350.field_11036) {
					return in.method_11657((class_2753) property, class_2350.field_11033);
				}
			}
		}
		return in;
	}

	protected class_2680 flipZ(class_2680 in) {
		return in.method_26185(class_2415.field_11300);
	}

	@SuppressWarnings("deprecation")
	protected class_2680 flipD1(class_2680 in) {
		return in.method_26186(class_2470.field_11465)
			.method_26185(class_2415.field_11301);
	}

	@SuppressWarnings("deprecation")
	protected class_2680 flipD2(class_2680 in) {
		return in.method_26186(class_2470.field_11465)
			.method_26185(class_2415.field_11300);
	}

	protected class_2338 flipX(class_2338 position) {
		class_2338 diff = getIDiff(position);
		return new class_2338(position.method_10263() - 2 * diff.method_10263(), position.method_10264(), position.method_10260());
	}

	protected class_2338 flipY(class_2338 position) {
		class_2338 diff = getIDiff(position);
		return new class_2338(position.method_10263(), position.method_10264() - 2 * diff.method_10264(), position.method_10260());
	}

	protected class_2338 flipZ(class_2338 position) {
		class_2338 diff = getIDiff(position);
		return new class_2338(position.method_10263(), position.method_10264(), position.method_10260() - 2 * diff.method_10260());
	}

	protected class_2338 flipD2(class_2338 position) {
		class_2338 diff = getIDiff(position);
		return new class_2338(position.method_10263() - diff.method_10263() + diff.method_10260(), position.method_10264(),
			position.method_10260() - diff.method_10260() + diff.method_10263());
	}

	protected class_2338 flipD1(class_2338 position) {
		class_2338 diff = getIDiff(position);
		return new class_2338(position.method_10263() - diff.method_10263() - diff.method_10260(), position.method_10264(),
			position.method_10260() - diff.method_10260() - diff.method_10263());
	}

	public void setPosition(class_243 pos3d) {
		this.position = pos3d;
	}

	public abstract List<class_2561> getAlignToolTips();

}
