package com.simibubi.create.content.equipment.toolbox;

import java.util.ArrayList;
import java.util.List;
import java.util.function.Consumer;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;

import com.simibubi.create.AllItems;
import com.simibubi.create.foundation.utility.NBTHelper;

import io.github.fabricators_of_create.porting_lib.transfer.callbacks.TransactionCallback;
import io.github.fabricators_of_create.porting_lib.transfer.item.ItemHandlerHelper;
import io.github.fabricators_of_create.porting_lib.transfer.item.ItemStackHandler;
import io.github.fabricators_of_create.porting_lib.transfer.item.ItemStackHandlerSlot;
import net.fabricmc.fabric.api.transfer.v1.item.ItemVariant;
import net.fabricmc.fabric.api.transfer.v1.transaction.TransactionContext;
import net.minecraft.class_1799;
import net.minecraft.class_2487;
import net.minecraft.class_2520;

public class ToolboxInventory extends ItemStackHandler {

	public static final int STACKS_PER_COMPARTMENT = 4;
	List<class_1799> filters;
	boolean settling;
	private ToolboxBlockEntity blockEntity;

	private boolean limitedMode;

	public ToolboxInventory(ToolboxBlockEntity be) {
		super(8 * STACKS_PER_COMPARTMENT);
		this.blockEntity = be;
		limitedMode = false;
		filters = new ArrayList<>();
		settling = false;
		for (int i = 0; i < 8; i++)
			filters.add(class_1799.field_8037);
	}

	public void inLimitedMode(Consumer<ToolboxInventory> action) {
		limitedMode = true;
		action.accept(this);
		limitedMode = false;
	}

	public void settle(int compartment) {
		int totalCount = 0;
		boolean valid = true;
		boolean shouldBeEmpty = false;
		class_1799 sample = class_1799.field_8037;

		for (int i = 0; i < STACKS_PER_COMPARTMENT; i++) {
			class_1799 stackInSlot = getStackInSlot(compartment * STACKS_PER_COMPARTMENT + i);
			totalCount += stackInSlot.method_7947();
			if (!shouldBeEmpty)
				shouldBeEmpty = stackInSlot.method_7960() || stackInSlot.method_7947() != stackInSlot.method_7914();
			else if (!stackInSlot.method_7960()) {
				valid = false;
				sample = stackInSlot;
			}
		}

		if (valid)
			return;

		settling = true;
		if (!sample.method_7946()) {
			for (int i = 0; i < STACKS_PER_COMPARTMENT; i++) {
				if (!getStackInSlot(compartment * STACKS_PER_COMPARTMENT + i).method_7960())
					continue;
				for (int j = i + 1; j < STACKS_PER_COMPARTMENT; j++) {
					class_1799 stackInSlot = getStackInSlot(compartment * STACKS_PER_COMPARTMENT + j);
					if (stackInSlot.method_7960())
						continue;
					setStackInSlot(compartment * STACKS_PER_COMPARTMENT + i, stackInSlot);
					setStackInSlot(compartment * STACKS_PER_COMPARTMENT + j, class_1799.field_8037);
					break;
				}
			}
		} else {
			for (int i = 0; i < STACKS_PER_COMPARTMENT; i++) {
				class_1799 copy = totalCount <= 0 ? class_1799.field_8037
					: ItemHandlerHelper.copyStackWithSize(sample, Math.min(totalCount, sample.method_7914()));
				setStackInSlot(compartment * STACKS_PER_COMPARTMENT + i, copy);
				totalCount -= copy.method_7947();
			}
		}
		settling = false;
		blockEntity.sendData();
	}

	@Override
	public boolean isItemValid(int slot, ItemVariant var) {
		class_1799 stack = var.toStack();
		if (!stack.method_7909().method_31568())
			return false;

		if (slot < 0 || slot >= getSlotCount())
			return false;
		int compartment = slot / STACKS_PER_COMPARTMENT;
		class_1799 filter = filters.get(compartment);
		if (limitedMode && filter.method_7960())
			return false;
		if (filter.method_7960() || ToolboxInventory.canItemsShareCompartment(filter, stack))
			return super.isItemValid(slot, var);
		return false;
	}

	@Override
	public void setStackInSlot(int slot, class_1799 stack) {
		super.setStackInSlot(slot, stack);
		updateCompartmentFilters(slot, stack, null);
	}

	private void updateCompartmentFilters(int slot, class_1799 stack, @Nullable TransactionContext ctx) {
		int compartment = slot / STACKS_PER_COMPARTMENT;
		if (!stack.method_7960() && filters.get(compartment)
				.method_7960()) {
			filters.set(compartment, ItemHandlerHelper.copyStackWithSize(stack, 1));
			if (ctx != null) TransactionCallback.onSuccess(ctx, blockEntity::sendData);
			else blockEntity.sendData();
		}
	}

	@Override
	public class_2487 serializeNBT() {
		class_2487 compound = super.serializeNBT();
		compound.method_10566("Compartments", NBTHelper.writeItemList(filters));
		return compound;
	}

	@Override
	protected void onContentsChanged(int slot) {
		if (!settling && !blockEntity.method_10997().field_9236)
			settle(slot / STACKS_PER_COMPARTMENT);
		blockEntity.sendData();
		blockEntity.method_5431();
		super.onContentsChanged(slot);
	}

	@Override
	public void deserializeNBT(class_2487 nbt) {
		filters = NBTHelper.readItemList(nbt.method_10554("Compartments", class_2520.field_33260));
		if (filters.size() != 8) {
			filters.clear();
			for (int i = 0; i < 8; i++)
				filters.add(class_1799.field_8037);
		}
		super.deserializeNBT(nbt);
	}

	public class_1799 distributeToCompartment(@Nonnull class_1799 stack, int compartment, TransactionContext ctx) {
		if (stack.method_7960())
			return stack;
		if (filters.get(compartment)
			.method_7960())
			return stack;

		int toInsert = stack.method_7947();
		int inserted = 0;
		ItemVariant variant = ItemVariant.of(stack);
		for (int i = STACKS_PER_COMPARTMENT - 1; i >= 0; i--) {
			int slot = compartment * STACKS_PER_COMPARTMENT + i;
			inserted += getSlot(slot).insert(variant, toInsert - inserted, ctx);
			if (inserted >= toInsert)
				break;
		}

		return ItemHandlerHelper.copyStackWithSize(stack, toInsert - inserted);
	}

	public class_1799 takeFromCompartment(int amount, int compartment, TransactionContext ctx) {
		if (amount == 0)
			return class_1799.field_8037;

		ItemVariant toExtract = null;
		int extracted = 0;
		for (int i = STACKS_PER_COMPARTMENT - 1; i >= 0; i--) {
			int slot = compartment * STACKS_PER_COMPARTMENT + i;
			ItemStackHandlerSlot handlerSlot = getSlot(slot);
			if (handlerSlot.isResourceBlank())
				continue;
			if (toExtract == null)
				toExtract = handlerSlot.getResource();
			extracted += handlerSlot.extract(toExtract, amount - extracted, ctx);
			if (extracted >= amount)
				break;
		}

		return toExtract == null || extracted == 0 ? class_1799.field_8037 : toExtract.toStack(extracted);
	}

	public static class_1799 cleanItemNBT(class_1799 stack) {
		if (AllItems.BELT_CONNECTOR.isIn(stack))
			stack.method_7983("FirstPulley");
		return stack;
	}

	public static boolean canItemsShareCompartment(class_1799 stack1, class_1799 stack2) {
		if (!stack1.method_7946() && !stack2.method_7946() && stack1.method_7963() && stack2.method_7963())
			return stack1.method_7909() == stack2.method_7909();
		if (AllItems.BELT_CONNECTOR.isIn(stack1) && AllItems.BELT_CONNECTOR.isIn(stack2))
			return true;
		return ItemHandlerHelper.canItemStacksStack(stack1, stack2);
	}
}
