package com.simibubi.create.content.fluids;

import static net.minecraft.class_2741.field_12508;

import java.util.ArrayList;
import java.util.List;

import com.simibubi.create.AllFluids;
import com.simibubi.create.content.fluids.pipes.VanillaFluidTargets;
import com.simibubi.create.content.fluids.potion.PotionFluidHandler;
import com.simibubi.create.foundation.advancement.AdvancementBehaviour;
import com.simibubi.create.foundation.advancement.AllAdvancements;
import com.simibubi.create.foundation.fluid.FluidHelper;
import com.simibubi.create.foundation.utility.BlockFace;
import com.simibubi.create.infrastructure.config.AllConfigs;

import io.github.fabricators_of_create.porting_lib.tags.Tags;
import io.github.fabricators_of_create.porting_lib.transfer.TransferUtil;
import io.github.fabricators_of_create.porting_lib.transfer.callbacks.TransactionCallback;
import io.github.fabricators_of_create.porting_lib.transfer.fluid.FluidTank;
import io.github.fabricators_of_create.porting_lib.fluids.FluidStack;
import io.github.tropheusj.milk.Milk;
import net.fabricmc.fabric.api.transfer.v1.fluid.FluidConstants;
import net.fabricmc.fabric.api.transfer.v1.fluid.FluidVariant;
import net.fabricmc.fabric.api.transfer.v1.storage.Storage;
import net.fabricmc.fabric.api.transfer.v1.transaction.Transaction;
import net.fabricmc.fabric.api.transfer.v1.transaction.TransactionContext;
import net.minecraft.class_1291;
import net.minecraft.class_1293;
import net.minecraft.class_1294;
import net.minecraft.class_1297;
import net.minecraft.class_1309;
import net.minecraft.class_1799;
import net.minecraft.class_1802;
import net.minecraft.class_1844;
import net.minecraft.class_1937;
import net.minecraft.class_2338;
import net.minecraft.class_2350;
import net.minecraft.class_238;
import net.minecraft.class_2404;
import net.minecraft.class_2487;
import net.minecraft.class_2680;
import net.minecraft.class_2741;
import net.minecraft.class_3417;
import net.minecraft.class_3419;
import net.minecraft.class_3481;
import net.minecraft.class_3486;
import net.minecraft.class_3610;
import net.minecraft.class_3612;
import net.minecraft.class_3922;
import net.minecraft.class_5540;

public class OpenEndedPipe extends FlowSource {

	private static final List<IEffectHandler> EFFECT_HANDLERS = new ArrayList<>();

	static {
		registerEffectHandler(new PotionEffectHandler());
		registerEffectHandler(new MilkEffectHandler());
		registerEffectHandler(new WaterEffectHandler());
		registerEffectHandler(new LavaEffectHandler());
		registerEffectHandler(new TeaEffectHandler());
	}

	private class_1937 world;
	private class_2338 pos;
	private class_238 aoe;

	private OpenEndFluidHandler fluidHandler;
	private class_2338 outputPos;
	private boolean wasPulling;

	private FluidStack cachedFluid;
	private List<class_1293> cachedEffects;

	public OpenEndedPipe(BlockFace face) {
		super(face);
		fluidHandler = new OpenEndFluidHandler();
		outputPos = face.getConnectedPos();
		pos = face.getPos();
		aoe = new class_238(outputPos).method_1012(0, -1, 0);
		if (face.getFace() == class_2350.field_11033)
			aoe = aoe.method_1012(0, -1, 0);
	}

	public static void registerEffectHandler(IEffectHandler handler) {
		EFFECT_HANDLERS.add(handler);
	}

	public class_1937 getWorld() {
		return world;
	}

	public class_2338 getPos() {
		return pos;
	}

	public class_2338 getOutputPos() {
		return outputPos;
	}

	public class_238 getAOE() {
		return aoe;
	}

	@Override
	public void manageSource(class_1937 world) {
		this.world = world;
	}

	@Override
	public Storage<FluidVariant> provideHandler() {
		return fluidHandler;
	}

	@Override
	public boolean isEndpoint() {
		return true;
	}

	public class_2487 serializeNBT() {
		class_2487 compound = new class_2487();
		fluidHandler.writeToNBT(compound);
		compound.method_10556("Pulling", wasPulling);
		compound.method_10566("Location", location.serializeNBT());
		return compound;
	}

	public static OpenEndedPipe fromNBT(class_2487 compound, class_2338 blockEntityPos) {
		BlockFace fromNBT = BlockFace.fromNBT(compound.method_10562("Location"));
		OpenEndedPipe oep = new OpenEndedPipe(new BlockFace(blockEntityPos, fromNBT.getFace()));
		oep.fluidHandler.readFromNBT(compound);
		oep.wasPulling = compound.method_10577("Pulling");
		return oep;
	}

	private FluidStack removeFluidFromSpace(TransactionContext ctx) {
		FluidStack empty = FluidStack.EMPTY;
		if (world == null)
			return empty;
		if (!world.method_8477(outputPos))
			return empty;

		class_2680 state = world.method_8320(outputPos);
		class_3610 fluidState = state.method_26227();
		boolean waterlog = state.method_28498(field_12508);

		FluidStack drainBlock = VanillaFluidTargets.drainBlock(world, outputPos, state, ctx);
		if (!drainBlock.isEmpty()) {
			if (state.method_28498(class_2741.field_20432)
				&& AllFluids.HONEY.is(drainBlock.getFluid()))
				TransactionCallback.onSuccess(ctx, () -> AdvancementBehaviour.tryAward(world, pos, AllAdvancements.HONEY_DRAIN));
			return drainBlock;
		}

		if (!waterlog && !state.method_45474())
			return empty;
		if (fluidState.method_15769() || !fluidState.method_15771())
			return empty;

		FluidStack stack = new FluidStack(fluidState.method_15772(), FluidConstants.BUCKET);

		if (FluidHelper.isWater(stack.getFluid()))
			AdvancementBehaviour.tryAward(world, pos, AllAdvancements.WATER_SUPPLY);

		world.updateSnapshots(ctx);
		if (waterlog) {
			world.method_8652(outputPos, state.method_11657(field_12508, false), 3);
			TransactionCallback.onSuccess(ctx, () -> world.method_39281(outputPos, class_3612.field_15910, 1));
			return stack;
		}
		world.method_8652(outputPos, fluidState.method_15759()
			.method_11657(class_2404.field_11278, 14), 3);
		return stack;
	}

	private boolean provideFluidToSpace(FluidStack fluid, TransactionContext ctx) {
		if (world == null)
			return false;
		if (!world.method_8477(outputPos))
			return false;

		class_2680 state = world.method_8320(outputPos);
		class_3610 fluidState = state.method_26227();
		boolean waterlog = state.method_28498(field_12508);

		if (!waterlog && !state.method_45474())
			return false;
		if (fluid.isEmpty())
			return false;
		if (!FluidHelper.hasBlockState(fluid.getFluid()) || fluid.getFluid().method_15791(Milk.MILK_FLUID_TAG)) // fabric: milk logic is different
			return true;

		// fabric: note - this is possibly prone to issues but follows what forge does.
		// collisions completely ignore simulation / transactions.
		if (!fluidState.method_15769() && fluidState.method_15772() != fluid.getFluid()) {
			FluidReactions.handlePipeSpillCollision(world, outputPos, fluid.getFluid(), fluidState);
			return false;
		}

		if (fluidState.method_15771())
			return false;
		if (waterlog && fluid.getFluid() != class_3612.field_15910)
			return false;

		if (world.method_8597()
			.comp_644() && FluidHelper.isTag(fluid, class_3486.field_15517)) {
			int i = outputPos.method_10263();
			int j = outputPos.method_10264();
			int k = outputPos.method_10260();
			TransactionCallback.onSuccess(ctx, () -> world.method_43128(null, i, j, k, class_3417.field_15102, class_3419.field_15245, 0.5F,
					2.6F + (world.field_9229.method_43057() - world.field_9229.method_43057()) * 0.8F));
			return true;
		}

		world.updateSnapshots(ctx);
		if (waterlog) {
			world.method_8652(outputPos, state.method_11657(field_12508, true), 3);
			TransactionCallback.onSuccess(ctx, () -> world.method_39281(outputPos, class_3612.field_15910, 1));
			return true;
		}

		if (!AllConfigs.server().fluids.pipesPlaceFluidSourceBlocks.get())
			return true;

		world.method_8652(outputPos, fluid.getFluid()
			.method_15785()
			.method_15759(), 3);
		return true;
	}

	private boolean canApplyEffects(FluidStack fluid) {
		for (IEffectHandler handler : EFFECT_HANDLERS) {
			if (handler.canApplyEffects(this, fluid)) {
				return true;
			}
		}
		return false;
	}

	private void applyEffects(FluidStack fluid) {
		for (IEffectHandler handler : EFFECT_HANDLERS) {
			if (handler.canApplyEffects(this, fluid)) {
				handler.applyEffects(this, fluid);
			}
		}
	}

	private class OpenEndFluidHandler extends FluidTank {

		public OpenEndFluidHandler() {
			super(FluidConstants.BUCKET);
		}

		@Override
		public long insert(FluidVariant resource, long maxAmount, TransactionContext transaction) {
			// Never allow being filled when a source is attached
			if (world == null)
				return 0;
			if (!world.method_8477(outputPos))
				return 0;
			if (resource.isBlank())
				return 0;
			FluidStack stack = new FluidStack(resource, 81);
			updateSnapshots(transaction);
			try (Transaction provideTest = transaction.openNested()) {
				if (!provideFluidToSpace(stack, provideTest))
					return 0;
			}

			FluidStack containedFluidStack = getFluid();
			boolean hasBlockState = FluidHelper.hasBlockState(containedFluidStack.getFluid());

			if (!containedFluidStack.isEmpty() && !containedFluidStack.canFill(resource))
				setFluid(FluidStack.EMPTY);
			if (wasPulling)
				wasPulling = false;

			if (canApplyEffects(stack) && !hasBlockState)
				maxAmount = 81; // fabric: deplete fluids 81 times faster to account for larger amounts
			long fill = super.insert(resource, maxAmount, transaction);
			if (!stack.isEmpty())
				TransactionCallback.onSuccess(transaction, () -> applyEffects(stack));
			if (getFluidAmount() == FluidConstants.BUCKET || (!FluidHelper.hasBlockState(containedFluidStack.getFluid()) || containedFluidStack.getFluid().method_15791(Milk.MILK_FLUID_TAG))) { // fabric: milk logic is different
				if (provideFluidToSpace(containedFluidStack, transaction))
					setFluid(FluidStack.EMPTY);
			}
			return fill;
		}

		@Override
		public long extract(FluidVariant extractedVariant, long maxAmount, TransactionContext transaction) {
			if (world == null)
				return 0;
			if (!world.method_8477(outputPos))
				return 0;
			if (maxAmount == 0)
				return 0;
			if (maxAmount > FluidConstants.BUCKET) {
				maxAmount = FluidConstants.BUCKET;
			}

			if (!wasPulling)
				wasPulling = true;

			updateSnapshots(transaction);
			long drainedFromInternal = super.extract(extractedVariant, maxAmount, transaction);
			if (drainedFromInternal != 0)
				return drainedFromInternal;

			FluidStack drainedFromWorld = removeFluidFromSpace(transaction);
			if (drainedFromWorld.isEmpty())
				return 0;
			if (!drainedFromWorld.canFill(extractedVariant))
				return 0;

			long remainder = drainedFromWorld.getAmount() - maxAmount;
			drainedFromWorld.setAmount(maxAmount);

			if (remainder > 0) {
				if (!getFluid().isEmpty() && !getFluid().isFluidEqual(drainedFromWorld))
					setFluid(FluidStack.EMPTY);
				super.insert(drainedFromWorld.getType(), remainder, transaction);
			}
			return drainedFromWorld.getAmount();
		}

		@Override
		public boolean isResourceBlank() {
			if (!super.isResourceBlank()) return false;
			return getResource().isBlank();
		}

		@Override
		public FluidVariant getResource() {
			if (!super.isResourceBlank()) return super.getResource();
			try (Transaction t = TransferUtil.getTransaction()) {
				FluidStack stack = removeFluidFromSpace(t);
				return stack.getType();
			}
		}

		@Override
		public long getAmount() {
			long amount = super.getAmount();
			if (amount != 0) return amount;
			return isResourceBlank() ? 0 : FluidConstants.BUCKET;
		}
	}

	public interface IEffectHandler {
		boolean canApplyEffects(OpenEndedPipe pipe, FluidStack fluid);

		void applyEffects(OpenEndedPipe pipe, FluidStack fluid);
	}

	public static class PotionEffectHandler implements IEffectHandler {
		@Override
		public boolean canApplyEffects(OpenEndedPipe pipe, FluidStack fluid) {
			return fluid.getFluid()
				.method_15780(AllFluids.POTION.get());
		}

		@Override
		public void applyEffects(OpenEndedPipe pipe, FluidStack fluid) {
			if (pipe.cachedFluid == null || pipe.cachedEffects == null || !fluid.isFluidEqual(pipe.cachedFluid)) {
				FluidStack copy = fluid.copy();
				copy.setAmount(FluidConstants.BOTTLE);
				class_1799 bottle = PotionFluidHandler.fillBottle(new class_1799(class_1802.field_8469), fluid);
				pipe.cachedEffects = class_1844.method_8067(bottle);
			}

			if (pipe.cachedEffects.isEmpty())
				return;

			List<class_1309> entities = pipe.getWorld()
				.method_8390(class_1309.class, pipe.getAOE(), class_1309::method_6086);
			for (class_1309 entity : entities) {
				for (class_1293 effectInstance : pipe.cachedEffects) {
					class_1291 effect = effectInstance.method_5579();
					if (effect.method_5561()) {
						effect.method_5564(null, null, entity, effectInstance.method_5578(), 0.5D);
					} else {
						entity.method_6092(new class_1293(effectInstance));
					}
				}
			}
		}
	}

	public static class MilkEffectHandler implements IEffectHandler {
		@Override
		public boolean canApplyEffects(OpenEndedPipe pipe, FluidStack fluid) {
			return FluidHelper.isTag(fluid, Tags.Fluids.MILK);
		}

		@Override
		public void applyEffects(OpenEndedPipe pipe, FluidStack fluid) {
			class_1937 world = pipe.getWorld();
			if (world.method_8510() % 5 != 0)
				return;
			List<class_1309> entities =
				world.method_8390(class_1309.class, pipe.getAOE(), class_1309::method_6086);
//			ItemStack curativeItem = new ItemStack(Items.MILK_BUCKET);
			for (class_1309 entity : entities)
				entity.method_6012();
		}
	}

	public static class WaterEffectHandler implements IEffectHandler {
		@Override
		public boolean canApplyEffects(OpenEndedPipe pipe, FluidStack fluid) {
			return FluidHelper.isTag(fluid, class_3486.field_15517);
		}

		@Override
		public void applyEffects(OpenEndedPipe pipe, FluidStack fluid) {
			class_1937 world = pipe.getWorld();
			if (world.method_8510() % 5 != 0)
				return;
			List<class_1297> entities = world.method_8333((class_1297) null, pipe.getAOE(), class_1297::method_5809);
			for (class_1297 entity : entities)
				entity.method_5646();
			class_2338.method_29715(pipe.getAOE())
				.forEach(pos -> dowseFire(world, pos));
		}

		// Adapted from ThrownPotion
		private static void dowseFire(class_1937 level, class_2338 pos) {
			class_2680 state = level.method_8320(pos);
			if (state.method_26164(class_3481.field_21952)) {
				level.method_8650(pos, false);
			} else if (class_5540.method_35245(state)) {
				class_5540.method_31614(null, state, level, pos);
			} else if (class_3922.method_23896(state)) {
				level.method_8444(null, 1009, pos, 0);
				class_3922.method_29288(null, level, pos, state);
				level.method_8501(pos, state.method_11657(class_3922.field_17352, false));
			}
		}
	}

	public static class LavaEffectHandler implements IEffectHandler {
		@Override
		public boolean canApplyEffects(OpenEndedPipe pipe, FluidStack fluid) {
			return FluidHelper.isTag(fluid, class_3486.field_15518);
		}

		@Override
		public void applyEffects(OpenEndedPipe pipe, FluidStack fluid) {
			class_1937 world = pipe.getWorld();
			if (world.method_8510() % 5 != 0)
				return;
			List<class_1297> entities = world.method_8333((class_1297) null, pipe.getAOE(), entity -> !entity.method_5753());
			for (class_1297 entity : entities)
				entity.method_5639(3);
		}
	}

	public static class TeaEffectHandler implements IEffectHandler {
		@Override
		public boolean canApplyEffects(OpenEndedPipe pipe, FluidStack fluid) {
			return fluid.getFluid().method_15780(AllFluids.TEA.get());
		}

		@Override
		public void applyEffects(OpenEndedPipe pipe, FluidStack fluid) {
			class_1937 world = pipe.getWorld();
			if (world.method_8510() % 5 != 0)
				return;
			List<class_1309> entities = world
					.method_8390(class_1309.class, pipe.getAOE(), class_1309::method_6086);
			for (class_1309 entity : entities) {
					entity.method_6092(new class_1293(class_1294.field_5917, 21, 0, false, false, false));
			}
		}
	}

}
