package com.simibubi.create.content.fluids;

import java.util.Arrays;
import java.util.function.Supplier;

import com.simibubi.create.AllPartialModels;
import com.simibubi.create.content.decoration.bracket.BracketedBlockEntityBehaviour;
import com.simibubi.create.content.fluids.FluidTransportBehaviour.AttachmentTypes;
import com.simibubi.create.content.fluids.FluidTransportBehaviour.AttachmentTypes.ComponentPartials;
import com.simibubi.create.content.fluids.pipes.FluidPipeBlock;
import com.simibubi.create.foundation.blockEntity.behaviour.BlockEntityBehaviour;
import com.simibubi.create.foundation.utility.Iterate;

import net.fabricmc.fabric.api.renderer.v1.model.FabricBakedModel;
import net.fabricmc.fabric.api.renderer.v1.model.ForwardingBakedModel;
import net.fabricmc.fabric.api.renderer.v1.render.RenderContext;
import net.fabricmc.fabric.api.rendering.data.v1.RenderAttachedBlockView;
import net.minecraft.class_1087;
import net.minecraft.class_1920;
import net.minecraft.class_2338;
import net.minecraft.class_2350;
import net.minecraft.class_2680;
import net.minecraft.class_310;
import net.minecraft.class_5819;

public class PipeAttachmentModel extends ForwardingBakedModel {

	public PipeAttachmentModel(class_1087 template) {
		wrapped = template;
	}

	@Override
	public boolean isVanillaAdapter() {
		return false;
	}

	@Override
	public void emitBlockQuads(class_1920 world, class_2680 state, class_2338 pos, Supplier<class_5819> randomSupplier, RenderContext context) {
		PipeModelData data = new PipeModelData();
		BracketedBlockEntityBehaviour bracket = BlockEntityBehaviour.get(world, pos, BracketedBlockEntityBehaviour.TYPE);

		RenderAttachedBlockView attachmentView = (RenderAttachedBlockView) world;
		Object attachment = attachmentView.getBlockEntityRenderAttachment(pos);
		if (attachment instanceof AttachmentTypes[] attachments) {
			for (int i = 0; i < attachments.length; i++) {
				data.putAttachment(Iterate.directions[i], attachments[i]);
			}
		}

		if (bracket != null)
			data.putBracket(bracket.getBracket());

		data.setEncased(FluidPipeBlock.shouldDrawCasing(world, pos, state));

		super.emitBlockQuads(world, state, pos, randomSupplier, context);

		addQuads(world, state, pos, randomSupplier, context, data);
	}

	// fabric: unnecessary
	// TODO: Update once MinecraftForge#9163 is merged
//	@SuppressWarnings("removal")
//	@Override
//	public ChunkRenderTypeSet getRenderTypes(@NotNull BlockState state, @NotNull RandomSource rand, @NotNull ModelData data) {
//		ChunkRenderTypeSet set = super.getRenderTypes(state, rand, data);
//		if (set.isEmpty()) {
//			return ItemBlockRenderTypes.getRenderLayers(state);
//		}
//		return set;
//	}

	private void addQuads(class_1920 world, class_2680 state, class_2338 pos, Supplier<class_5819> randomSupplier, RenderContext context,
		PipeModelData pipeData) {
		class_1087 bracket = pipeData.getBracket();
		if (bracket != null)
			((FabricBakedModel) bracket).emitBlockQuads(world, state, pos, randomSupplier, context);
		for (class_2350 d : Iterate.directions) {
			AttachmentTypes type = pipeData.getAttachment(d);
			for (ComponentPartials partial : type.partials) {
				((FabricBakedModel) AllPartialModels.PIPE_ATTACHMENTS.get(partial)
					.get(d)
					.get())
					.emitBlockQuads(world, state, pos, randomSupplier, context);
			}
		}
		if (pipeData.isEncased())
			((FabricBakedModel) AllPartialModels.FLUID_PIPE_CASING.get())
				.emitBlockQuads(world, state, pos, randomSupplier, context);
	}

	private static class PipeModelData {
		private AttachmentTypes[] attachments;
		private boolean encased;
		private class_1087 bracket;

		public PipeModelData() {
			attachments = new AttachmentTypes[6];
			Arrays.fill(attachments, AttachmentTypes.NONE);
		}

		public void putBracket(class_2680 state) {
			if (state != null) {
				this.bracket = class_310.method_1551()
					.method_1541()
					.method_3349(state);
			}
		}

		public class_1087 getBracket() {
			return bracket;
		}

		public void putAttachment(class_2350 face, AttachmentTypes rim) {
			attachments[face.method_10146()] = rim;
		}

		public AttachmentTypes getAttachment(class_2350 face) {
			return attachments[face.method_10146()];
		}

		public void setEncased(boolean encased) {
			this.encased = encased;
		}

		public boolean isEncased() {
			return encased;
		}
	}

}
