package com.simibubi.create.content.fluids.drain;

import java.util.IdentityHashMap;
import java.util.List;
import java.util.Map;

import org.jetbrains.annotations.Nullable;

import com.simibubi.create.content.equipment.goggles.IHaveGoggleInformation;
import com.simibubi.create.content.fluids.transfer.GenericItemEmptying;
import com.simibubi.create.content.kinetics.belt.behaviour.DirectBeltInputBehaviour;
import com.simibubi.create.content.kinetics.belt.transport.TransportedItemStack;
import com.simibubi.create.foundation.advancement.AllAdvancements;
import com.simibubi.create.foundation.blockEntity.SmartBlockEntity;
import com.simibubi.create.foundation.blockEntity.behaviour.BlockEntityBehaviour;
import com.simibubi.create.foundation.blockEntity.behaviour.fluid.SmartFluidTankBehaviour;
import com.simibubi.create.foundation.utility.BlockHelper;
import com.simibubi.create.foundation.utility.Iterate;
import com.simibubi.create.foundation.utility.Pair;
import com.simibubi.create.foundation.utility.VecHelper;

import io.github.fabricators_of_create.porting_lib.transfer.TransferUtil;
import io.github.fabricators_of_create.porting_lib.transfer.item.ItemHandlerHelper;
import io.github.fabricators_of_create.porting_lib.fluids.FluidStack;
import net.fabricmc.fabric.api.transfer.v1.fluid.FluidConstants;
import net.fabricmc.fabric.api.transfer.v1.fluid.FluidVariant;
import net.fabricmc.fabric.api.transfer.v1.item.ItemVariant;
import net.fabricmc.fabric.api.transfer.v1.storage.Storage;
import net.fabricmc.fabric.api.transfer.v1.storage.base.SidedStorageBlockEntity;
import net.fabricmc.fabric.api.transfer.v1.transaction.Transaction;
import net.fabricmc.fabric.api.transfer.v1.transaction.base.SnapshotParticipant;
import net.minecraft.class_1542;
import net.minecraft.class_1799;
import net.minecraft.class_2338;
import net.minecraft.class_2350;
import net.minecraft.class_243;
import net.minecraft.class_2487;
import net.minecraft.class_2561;
import net.minecraft.class_2591;
import net.minecraft.class_2680;

public class ItemDrainBlockEntity extends SmartBlockEntity implements IHaveGoggleInformation, SidedStorageBlockEntity {

	public static final int FILLING_TIME = 20;

	SmartFluidTankBehaviour internalTank;
	TransportedItemStack heldItem;
	protected int processingTicks;
	Map<class_2350, ItemDrainItemHandler> itemHandlers;

	SnapshotParticipant<TransportedItemStack> snapshotParticipant = new SnapshotParticipant<>() {
		@Override
		protected TransportedItemStack createSnapshot() {
			return heldItem == null ? TransportedItemStack.EMPTY : heldItem.fullCopy();
		}

		@Override
		protected void readSnapshot(TransportedItemStack snapshot) {
			heldItem = snapshot == TransportedItemStack.EMPTY ? null : snapshot;
		}

		@Override
		protected void onFinalCommit() {
			notifyUpdate();
		}
	};

	public ItemDrainBlockEntity(class_2591<?> type, class_2338 pos, class_2680 state) {
		super(type, pos, state);
		itemHandlers = new IdentityHashMap<>();
		for (class_2350 d : Iterate.horizontalDirections) {
			ItemDrainItemHandler itemDrainItemHandler = new ItemDrainItemHandler(this, d);
			itemHandlers.put(d, itemDrainItemHandler);
		}
	}

	@Override
	public void addBehaviours(List<BlockEntityBehaviour> behaviours) {
		behaviours.add(new DirectBeltInputBehaviour(this).allowingBeltFunnels()
			.setInsertionHandler(this::tryInsertingFromSide));
		behaviours.add(internalTank = SmartFluidTankBehaviour.single(this, (long) (FluidConstants.BUCKET * 1.5))
			.allowExtraction()
			.forbidInsertion());
		registerAwardables(behaviours, AllAdvancements.DRAIN, AllAdvancements.CHAINED_DRAIN);
	}

	private class_1799 tryInsertingFromSide(TransportedItemStack transportedStack, class_2350 side, boolean simulate) {
		class_1799 inserted = transportedStack.stack;
		class_1799 returned = class_1799.field_8037;

		if (!getHeldItemStack().method_7960())
			return inserted;

		if (inserted.method_7947() > 1 && GenericItemEmptying.canItemBeEmptied(field_11863, inserted)) {
			returned = ItemHandlerHelper.copyStackWithSize(inserted, inserted.method_7947() - 1);
			inserted = ItemHandlerHelper.copyStackWithSize(inserted, 1);
		}

		if (simulate)
			return returned;

		transportedStack = transportedStack.copy();
		transportedStack.stack = inserted.method_7972();
		transportedStack.beltPosition = side.method_10166()
			.method_10178() ? .5f : 0;
		transportedStack.prevSideOffset = transportedStack.sideOffset;
		transportedStack.prevBeltPosition = transportedStack.beltPosition;
		setHeldItem(transportedStack, side);
		method_5431();
		sendData();

		return returned;
	}

	public class_1799 getHeldItemStack() {
		return heldItem == null ? class_1799.field_8037 : heldItem.stack;
	}

	@Override
	public void tick() {
		super.tick();

		if (heldItem == null) {
			processingTicks = 0;
			return;
		}

		boolean onClient = field_11863.field_9236 && !isVirtual();

		if (processingTicks > 0) {
			heldItem.prevBeltPosition = .5f;
			boolean wasAtBeginning = processingTicks == FILLING_TIME;
			if (!onClient || processingTicks < FILLING_TIME)
				processingTicks--;
			if (!continueProcessing()) {
				processingTicks = 0;
				notifyUpdate();
				return;
			}
			if (wasAtBeginning != (processingTicks == FILLING_TIME))
				sendData();
			return;
		}

		heldItem.prevBeltPosition = heldItem.beltPosition;
		heldItem.prevSideOffset = heldItem.sideOffset;

		heldItem.beltPosition += itemMovementPerTick();
		if (heldItem.beltPosition > 1) {
			heldItem.beltPosition = 1;

			if (onClient)
				return;

			class_2350 side = heldItem.insertedFrom;

			class_1799 tryExportingToBeltFunnel = getBehaviour(DirectBeltInputBehaviour.TYPE)
				.tryExportingToBeltFunnel(heldItem.stack, side.method_10153(), false);
			if (tryExportingToBeltFunnel != null) {
				if (tryExportingToBeltFunnel.method_7947() != heldItem.stack.method_7947()) {
					if (tryExportingToBeltFunnel.method_7960())
						heldItem = null;
					else
						heldItem.stack = tryExportingToBeltFunnel;
					notifyUpdate();
					return;
				}
				if (!tryExportingToBeltFunnel.method_7960())
					return;
			}

			class_2338 nextPosition = field_11867.method_10093(side);
			DirectBeltInputBehaviour directBeltInputBehaviour =
				BlockEntityBehaviour.get(field_11863, nextPosition, DirectBeltInputBehaviour.TYPE);
			if (directBeltInputBehaviour == null) {
				if (!BlockHelper.hasBlockSolidSide(field_11863.method_8320(nextPosition), field_11863, nextPosition,
					side.method_10153())) {
					class_1799 ejected = heldItem.stack;
					class_243 outPos = VecHelper.getCenterOf(field_11867)
						.method_1019(class_243.method_24954(side.method_10163())
							.method_1021(.75));
					float movementSpeed = itemMovementPerTick();
					class_243 outMotion = class_243.method_24954(side.method_10163())
						.method_1021(movementSpeed)
						.method_1031(0, 1 / 8f, 0);
					outPos.method_1019(outMotion.method_1029());
					class_1542 entity = new class_1542(field_11863, outPos.field_1352, outPos.field_1351 + 6 / 16f, outPos.field_1350, ejected);
					entity.method_18799(outMotion);
					entity.method_6988();
					entity.field_6037 = true;
					field_11863.method_8649(entity);

					heldItem = null;
					notifyUpdate();
				}
				return;
			}

			if (!directBeltInputBehaviour.canInsertFromSide(side))
				return;

			class_1799 returned = directBeltInputBehaviour.handleInsertion(heldItem.copy(), side, false);

			if (returned.method_7960()) {
				if (field_11863.method_8321(nextPosition) instanceof ItemDrainBlockEntity)
					award(AllAdvancements.CHAINED_DRAIN);
				heldItem = null;
				notifyUpdate();
				return;
			}

			if (returned.method_7947() != heldItem.stack.method_7947()) {
				heldItem.stack = returned;
				notifyUpdate();
				return;
			}

			return;
		}

		if (heldItem.prevBeltPosition < .5f && heldItem.beltPosition >= .5f) {
			if (!GenericItemEmptying.canItemBeEmptied(field_11863, heldItem.stack))
				return;
			heldItem.beltPosition = .5f;
			if (onClient)
				return;
			processingTicks = FILLING_TIME;
			sendData();
		}

	}

	protected boolean continueProcessing() {
		if (field_11863.field_9236 && !isVirtual())
			return true;
		if (processingTicks < 5)
			return true;
		if (!GenericItemEmptying.canItemBeEmptied(field_11863, heldItem.stack))
			return false;

		Pair<FluidStack, class_1799> emptyItem = GenericItemEmptying.emptyItem(field_11863, heldItem.stack, true);
		FluidStack fluidFromItem = emptyItem.getFirst();

		try (Transaction t = TransferUtil.getTransaction()) {
			if (processingTicks > 5) {
				internalTank.allowInsertion();
				try (Transaction nested = t.openNested()) {
					if (!fluidFromItem.isEmpty()) {
						long inserted = internalTank.getPrimaryHandler().insert(fluidFromItem.getType(), fluidFromItem.getAmount(), nested);
						if (inserted != fluidFromItem.getAmount()) {
							internalTank.forbidInsertion();
							processingTicks = FILLING_TIME;
							return true;
						}
					}
				}
				internalTank.forbidInsertion();
				return true;
			}

			emptyItem = GenericItemEmptying.emptyItem(field_11863, heldItem.stack.method_7972(), false);
			award(AllAdvancements.DRAIN);

			// Process finished
			class_1799 out = emptyItem.getSecond();
			if (!out.method_7960())
				heldItem.stack = out;
			else
				heldItem = null;
			internalTank.allowInsertion();
			TransferUtil.insertFluid(internalTank.getPrimaryHandler(), fluidFromItem);
			t.commit();
			internalTank.forbidInsertion();
			notifyUpdate();
			return true;
		}
	}

	private float itemMovementPerTick() {
		return 1 / 8f;
	}

	@Override
	public void invalidate() {
		super.invalidate();
	}

	public void setHeldItem(TransportedItemStack heldItem, class_2350 insertedFrom) {
		this.heldItem = heldItem;
		this.heldItem.insertedFrom = insertedFrom;
	}

	@Override
	public void write(class_2487 compound, boolean clientPacket) {
		compound.method_10569("ProcessingTicks", processingTicks);
		if (heldItem != null)
			compound.method_10566("HeldItem", heldItem.serializeNBT());
		super.write(compound, clientPacket);
	}

	@Override
	protected void read(class_2487 compound, boolean clientPacket) {
		heldItem = null;
		processingTicks = compound.method_10550("ProcessingTicks");
		if (compound.method_10545("HeldItem"))
			heldItem = TransportedItemStack.read(compound.method_10562("HeldItem"));
		super.read(compound, clientPacket);
	}

	@Nullable
	@Override
	public Storage<FluidVariant> getFluidStorage(@Nullable class_2350 face) {
		if (face != class_2350.field_11036) {
			return internalTank.getCapability();
		}
		return null;
	}

	@Nullable
	@Override
	public Storage<ItemVariant> getItemStorage(@Nullable class_2350 face) {
		if (face != null && face.method_10166().method_10179()) {
			return itemHandlers.get(face);
		}
		return null;
	}

	@Override
	public boolean addToGoggleTooltip(List<class_2561> tooltip, boolean isPlayerSneaking) {
		return containedFluidTooltip(tooltip, isPlayerSneaking, getFluidStorage(null));
	}
}
