package com.simibubi.create.content.fluids.pipes.valve;

import com.jozufozu.flywheel.api.MaterialManager;
import com.jozufozu.flywheel.api.instance.DynamicInstance;
import com.jozufozu.flywheel.core.Materials;
import com.jozufozu.flywheel.core.materials.model.ModelData;
import com.simibubi.create.AllPartialModels;
import com.simibubi.create.content.kinetics.base.KineticBlockEntityRenderer;
import com.simibubi.create.content.kinetics.base.ShaftInstance;
import com.simibubi.create.foundation.utility.AngleHelper;
import com.simibubi.create.foundation.utility.AnimationTickHolder;
import net.minecraft.class_2350;
import net.minecraft.class_3532;

public class FluidValveInstance extends ShaftInstance<FluidValveBlockEntity> implements DynamicInstance {

	protected ModelData pointer;
	protected boolean settled;

    protected final double xRot;
    protected final double yRot;
    protected final int pointerRotationOffset;

    public FluidValveInstance(MaterialManager dispatcher, FluidValveBlockEntity blockEntity) {
        super(dispatcher, blockEntity);

        class_2350 facing = blockState.method_11654(FluidValveBlock.FACING);

        yRot = AngleHelper.horizontalAngle(facing);
        xRot = facing == class_2350.field_11036 ? 0 : facing == class_2350.field_11033 ? 180 : 90;

        class_2350.class_2351 pipeAxis = FluidValveBlock.getPipeAxis(blockState);
        class_2350.class_2351 shaftAxis = KineticBlockEntityRenderer.getRotationAxisOf(blockEntity);

        boolean twist = pipeAxis.method_10179() && shaftAxis == class_2350.class_2351.field_11048 || pipeAxis.method_10178();
        pointerRotationOffset = twist ? 90 : 0;
        settled = false;

        pointer = materialManager.defaultSolid()
                .material(Materials.TRANSFORMED)
                .getModel(AllPartialModels.FLUID_VALVE_POINTER, blockState).createInstance();

		transformPointer();
    }

	@Override
	public void beginFrame() {
		if (blockEntity.pointer.settled() && settled)
			return;

		transformPointer();
	}

	private void transformPointer() {
		float value = blockEntity.pointer.getValue(AnimationTickHolder.getPartialTicks());
		float pointerRotation = class_3532.method_16439(value, 0, -90);
		settled = (value == 0 || value == 1) && blockEntity.pointer.settled();

        pointer.loadIdentity()
				 .translate(getInstancePosition())
				 .centre()
				 .rotateY(yRot)
				 .rotateX(xRot)
				 .rotateY(pointerRotationOffset + pointerRotation)
				 .unCentre();
	}

    @Override
    public void updateLight() {
        super.updateLight();
        relight(pos, pointer);
    }

    @Override
    public void remove() {
        super.remove();
        pointer.delete();
    }
}
