package com.simibubi.create.content.fluids.potion;

import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import com.google.common.collect.Lists;
import com.simibubi.create.content.fluids.potion.PotionFluid.BottleType;
import com.simibubi.create.foundation.fluid.FluidHelper;
import com.simibubi.create.foundation.fluid.FluidIngredient;
import com.simibubi.create.foundation.utility.Components;
import com.simibubi.create.foundation.utility.NBTHelper;
import com.simibubi.create.foundation.utility.Pair;
import io.github.fabricators_of_create.porting_lib.fluids.FluidStack;

import net.fabricmc.api.EnvType;
import net.fabricmc.api.Environment;
import net.fabricmc.fabric.api.transfer.v1.fluid.FluidConstants;
import net.fabricmc.fabric.api.transfer.v1.fluid.FluidVariant;
import net.minecraft.class_124;
import net.minecraft.class_1291;
import net.minecraft.class_1292;
import net.minecraft.class_1293;
import net.minecraft.class_1320;
import net.minecraft.class_1322;
import net.minecraft.class_1792;
import net.minecraft.class_1799;
import net.minecraft.class_1802;
import net.minecraft.class_1842;
import net.minecraft.class_1844;
import net.minecraft.class_1847;
import net.minecraft.class_1935;
import net.minecraft.class_2487;
import net.minecraft.class_2561;
import net.minecraft.class_3545;
import net.minecraft.class_3612;
import net.minecraft.class_5250;

public class PotionFluidHandler {

	public static Pair<FluidStack, class_1799> emptyPotion(class_1799 stack, boolean simulate) {
		FluidStack fluid = getFluidFromPotionItem(stack);
		if (!simulate)
			stack.method_7934(1);
		return Pair.of(fluid, new class_1799(class_1802.field_8469));
	}

	public static FluidIngredient potionIngredient(class_1842 potion, long amount) {
		return FluidIngredient.fromFluidStack(FluidHelper.copyStackWithAmount(PotionFluidHandler
			.getFluidFromPotionItem(class_1844.method_8061(new class_1799(class_1802.field_8574), potion)), amount));
	}

	public static FluidStack getFluidFromPotionItem(class_1799 stack) {
		class_1842 potion = class_1844.method_8063(stack);
		List<class_1293> list = class_1844.method_8068(stack);
		BottleType bottleTypeFromItem = bottleTypeFromItem(stack.method_7909());
		if (potion == class_1847.field_8984)
			return FluidStack.EMPTY;
		if (potion == class_1847.field_8991 && list.isEmpty() && bottleTypeFromItem == BottleType.REGULAR)
			return new FluidStack(class_3612.field_15910, FluidConstants.BOTTLE);
		FluidStack fluid = PotionFluid.withEffects(FluidConstants.BOTTLE, potion, list);
		class_2487 tagInfo = fluid.getTag();
		NBTHelper.writeEnum(tagInfo, "Bottle", bottleTypeFromItem);
		FluidVariant variant = FluidVariant.of(fluid.getFluid(), tagInfo);
		return new FluidStack(variant, fluid.getAmount(), tagInfo);
	}

	public static FluidStack getFluidFromPotion(class_1842 potion, BottleType bottleType, long amount) {
		if (potion == class_1847.field_8991 && bottleType == BottleType.REGULAR)
			return new FluidStack(class_3612.field_15910, amount);
		FluidStack fluid = PotionFluid.of(amount, potion);
		NBTHelper.writeEnum(fluid.getOrCreateTag(), "Bottle", bottleType);
		return new FluidStack(fluid.getFluid(), fluid.getAmount(), fluid.getTag());
	}

	public static BottleType bottleTypeFromItem(class_1792 item) {
		if (item == class_1802.field_8150)
			return BottleType.LINGERING;
		if (item == class_1802.field_8436)
			return BottleType.SPLASH;
		return BottleType.REGULAR;
	}

	public static class_1935 itemFromBottleType(BottleType type) {
		switch (type) {
		case LINGERING:
			return class_1802.field_8150;
		case SPLASH:
			return class_1802.field_8436;
		case REGULAR:
		default:
			return class_1802.field_8574;
		}
	}

	public static long getRequiredAmountForFilledBottle(class_1799 stack, FluidStack availableFluid) {
		return FluidConstants.BOTTLE;
	}

	public static class_1799 fillBottle(class_1799 stack, FluidStack availableFluid) {
		class_2487 tag = availableFluid.getOrCreateTag();
		class_1799 potionStack = new class_1799(itemFromBottleType(NBTHelper.readEnum(tag, "Bottle", BottleType.class)));
		class_1844.method_8061(potionStack, class_1844.method_8057(tag));
		class_1844.method_8056(potionStack, class_1844.method_8060(tag));
		return potionStack;
	}

	// Modified version of PotionUtils#addPotionTooltip
	@Environment(EnvType.CLIENT)
	public static void addPotionTooltip(FluidStack fs, List<class_2561> tooltip, float p_185182_2_) {
		addPotionTooltip(fs.getType(), tooltip, p_185182_2_);
	}

	@Environment(EnvType.CLIENT)
	public static void addPotionTooltip(FluidVariant fs, List<class_2561> tooltip, float p_185182_2_) {
		List<class_1293> list = class_1844.method_8066(fs.getNbt());
		List<class_3545<String, class_1322>> list1 = Lists.newArrayList();
		if (list.isEmpty()) {
			tooltip.add((Components.translatable("effect.none")).method_27692(class_124.field_1080));
		} else {
			for (class_1293 effectinstance : list) {
				class_5250 textcomponent = Components.translatable(effectinstance.method_5586());
				class_1291 effect = effectinstance.method_5579();
				Map<class_1320, class_1322> map = effect.method_5565();
				if (!map.isEmpty()) {
					for (Entry<class_1320, class_1322> entry : map.entrySet()) {
						class_1322 attributemodifier = entry.getValue();
						class_1322 attributemodifier1 = new class_1322(attributemodifier.method_6185(),
							effect.method_5563(effectinstance.method_5578(), attributemodifier),
							attributemodifier.method_6182());
						list1.add(new class_3545<>(
							entry.getKey().method_26830(),
							attributemodifier1));
					}
				}

				if (effectinstance.method_5578() > 0) {
					textcomponent.method_27693(" ")
						.method_27693(Components.translatable("potion.potency." + effectinstance.method_5578()).getString());
				}

				if (effectinstance.method_5584() > 20) {
					textcomponent.method_27693(" (")
						.method_10852(class_1292.method_5577(effectinstance, p_185182_2_))
						.method_27693(")");
				}

				tooltip.add(textcomponent.method_27692(effect.method_18792()
					.method_18793()));
			}
		}

		if (!list1.isEmpty()) {
			tooltip.add(Components.immutableEmpty());
			tooltip.add((Components.translatable("potion.whenDrank")).method_27692(class_124.field_1064));

			for (class_3545<String, class_1322> tuple : list1) {
				class_1322 attributemodifier2 = tuple.method_15441();
				double d0 = attributemodifier2.method_6186();
				double d1;
				if (attributemodifier2.method_6182() != class_1322.class_1323.field_6330
					&& attributemodifier2.method_6182() != class_1322.class_1323.field_6331) {
					d1 = attributemodifier2.method_6186();
				} else {
					d1 = attributemodifier2.method_6186() * 100.0D;
				}

				if (d0 > 0.0D) {
					tooltip.add((Components.translatable(
						"attribute.modifier.plus." + attributemodifier2.method_6182()
							.method_6191(),
						class_1799.field_8029.format(d1),
						Components.translatable(tuple.method_15442())))
							.method_27692(class_124.field_1078));
				} else if (d0 < 0.0D) {
					d1 = d1 * -1.0D;
					tooltip.add((Components.translatable(
						"attribute.modifier.take." + attributemodifier2.method_6182()
							.method_6191(),
						class_1799.field_8029.format(d1),
						Components.translatable(tuple.method_15442())))
							.method_27692(class_124.field_1061));
				}
			}
		}

	}

}
