package com.simibubi.create.content.fluids.potion;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import com.simibubi.create.Create;
import com.simibubi.create.content.fluids.potion.PotionFluid.BottleType;
import com.simibubi.create.content.kinetics.mixer.MixingRecipe;
import com.simibubi.create.content.processing.recipe.HeatCondition;
import com.simibubi.create.content.processing.recipe.ProcessingRecipeBuilder;
import com.simibubi.create.foundation.fluid.FluidIngredient;

import io.github.fabricators_of_create.porting_lib.brewing.BrewingRecipe;
import io.github.fabricators_of_create.porting_lib.brewing.BrewingRecipeRegistry;
import io.github.fabricators_of_create.porting_lib.brewing.IBrewingRecipe;
import io.github.fabricators_of_create.porting_lib.mixin.accessors.common.accessor.PotionBrewing$MixAccessor;
import io.github.fabricators_of_create.porting_lib.mixin.accessors.common.accessor.PotionBrewing.MixAccessor;
import io.github.fabricators_of_create.porting_lib.mixin.accessors.common.accessor.PotionBrewingAccessor;
import io.github.fabricators_of_create.porting_lib.fluids.FluidStack;
import net.fabricmc.fabric.api.registry.FabricBrewingRecipeRegistry;
import net.fabricmc.fabric.api.transfer.v1.fluid.FluidConstants;
import net.minecraft.class_1792;
import net.minecraft.class_1799;
import net.minecraft.class_1802;
import net.minecraft.class_1842;
import net.minecraft.class_1845;
import net.minecraft.class_1847;
import net.minecraft.class_1856;
import net.minecraft.class_5321;
import net.minecraft.class_7923;

public class PotionMixingRecipes {

	public static final List<class_1792> SUPPORTED_CONTAINERS = List.of(class_1802.field_8574, class_1802.field_8436, class_1802.field_8150);

	public static final List<MixingRecipe> ALL = createRecipes();
	public static final Map<class_1792, List<MixingRecipe>> BY_ITEM = sortRecipesByItem(ALL);

	private static List<MixingRecipe> createRecipes() {
		List<MixingRecipe> mixingRecipes = new ArrayList<>();

		int recipeIndex = 0;

		List<class_1792> allowedSupportedContainers = new ArrayList<>();
		List<class_1799> supportedContainerStacks = new ArrayList<>();
		for (class_1792 container : SUPPORTED_CONTAINERS) {
			class_1799 stack = new class_1799(container);
			supportedContainerStacks.add(stack);
			if (PotionBrewingAccessor.port_lib$ALLOWED_CONTAINER().test(stack)) {
				allowedSupportedContainers.add(container);
			}
		}

		for (class_1792 container : allowedSupportedContainers) {
			BottleType bottleType = PotionFluidHandler.bottleTypeFromItem(container);
			for (class_1845.class_1846<class_1842> mix : PotionBrewingAccessor.port_lib$POTION_MIXES()) {
				MixAccessor<class_1842> access = (MixAccessor<class_1842>) mix;
				FluidStack fromFluid = PotionFluidHandler.getFluidFromPotion(access.port_lib$from(), bottleType, FluidConstants.BUCKET);
				FluidStack toFluid = PotionFluidHandler.getFluidFromPotion(access.port_lib$to(), bottleType, FluidConstants.BUCKET);

				mixingRecipes.add(createRecipe("potion_mixing_vanilla_" + recipeIndex++, access.port_lib$ingredient(), fromFluid, toFluid));
			}
		}

		for (class_1845.class_1846<class_1792> mix : PotionBrewingAccessor.port_lib$CONTAINER_MIXES()) {
			MixAccessor<class_1792> access = (MixAccessor<class_1792>) mix;
			class_1792 from = access.port_lib$from();
			if (!allowedSupportedContainers.contains(from)) {
				continue;
			}
			class_1792 to = access.port_lib$to();
			if (!allowedSupportedContainers.contains(to)) {
				continue;
			}
			BottleType fromBottleType = PotionFluidHandler.bottleTypeFromItem(from);
			BottleType toBottleType = PotionFluidHandler.bottleTypeFromItem(to);
			class_1856 ingredient = access.port_lib$ingredient();

			for (Entry<class_5321<class_1842>, class_1842> entry : class_7923.field_41179.method_29722()) {
				class_1842 potion = entry.getValue();
				if (potion == class_1847.field_8984) {
					continue;
				}

				FluidStack fromFluid = PotionFluidHandler.getFluidFromPotion(potion, fromBottleType, FluidConstants.BUCKET);
				FluidStack toFluid = PotionFluidHandler.getFluidFromPotion(potion, toBottleType, FluidConstants.BUCKET);

				mixingRecipes.add(createRecipe("potion_mixing_vanilla_" + recipeIndex++, ingredient, fromFluid, toFluid));
			}
		}

		recipeIndex = 0;
		for (IBrewingRecipe recipe : BrewingRecipeRegistry.getRecipes()) {
			if (recipe instanceof BrewingRecipe recipeImpl) {
				class_1799 output = recipeImpl.getOutput();
				if (!SUPPORTED_CONTAINERS.contains(output.method_7909())) {
					continue;
				}

				class_1856 input = recipeImpl.getInput();
				class_1856 ingredient = recipeImpl.getIngredient();
				FluidStack outputFluid = null;
				for (class_1799 stack : supportedContainerStacks) {
					if (input.method_8093(stack)) {
						FluidStack inputFluid = PotionFluidHandler.getFluidFromPotionItem(stack);
						if (outputFluid == null) {
							outputFluid = PotionFluidHandler.getFluidFromPotionItem(output);
						}
						mixingRecipes.add(createRecipe("potion_mixing_modded_" + recipeIndex++, ingredient, inputFluid, outputFluid));
					}
				}
			}
		}

		return mixingRecipes;
	}

	private static MixingRecipe createRecipe(String id, class_1856 ingredient, FluidStack fromFluid, FluidStack toFluid) {
		return new ProcessingRecipeBuilder<>(MixingRecipe::new,
				Create.asResource(id)).require(ingredient)
				.require(FluidIngredient.fromFluidStack(fromFluid))
				.output(toFluid)
				.requiresHeat(HeatCondition.HEATED)
				.build();
	}

	private static Map<class_1792, List<MixingRecipe>> sortRecipesByItem(List<MixingRecipe> all) {
		Map<class_1792, List<MixingRecipe>> byItem = new HashMap<>();
		Set<class_1792> processedItems = new HashSet<>();
		for (MixingRecipe recipe : all) {
			for (class_1856 ingredient : recipe.method_8117()) {
				for (class_1799 itemStack : ingredient.method_8105()) {
					class_1792 item = itemStack.method_7909();
					if (processedItems.add(item)) {
						byItem.computeIfAbsent(item, i -> new ArrayList<>())
								.add(recipe);
					}
				}
			}
			processedItems.clear();
		}
		return byItem;
	}

}
