package com.simibubi.create.content.fluids.tank;

import java.util.Arrays;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import com.simibubi.create.foundation.utility.Components;

import net.fabricmc.fabric.api.transfer.v1.fluid.FluidConstants;

import org.jetbrains.annotations.NotNull;

import com.simibubi.create.AllBlocks;
import com.simibubi.create.Create;
import com.simibubi.create.content.decoration.steamWhistle.WhistleBlock;
import com.simibubi.create.content.decoration.steamWhistle.WhistleBlockEntity;
import com.simibubi.create.content.equipment.goggles.IHaveGoggleInformation;
import com.simibubi.create.content.kinetics.BlockStressValues;
import com.simibubi.create.content.kinetics.steamEngine.SteamEngineBlock;
import com.simibubi.create.foundation.advancement.AdvancementBehaviour;
import com.simibubi.create.foundation.advancement.AllAdvancements;
import com.simibubi.create.foundation.fluid.FluidHelper;
import com.simibubi.create.foundation.fluid.SmartFluidTank;
import com.simibubi.create.foundation.utility.Iterate;
import com.simibubi.create.foundation.utility.Lang;
import com.simibubi.create.foundation.utility.animation.LerpedFloat;
import com.simibubi.create.foundation.utility.animation.LerpedFloat.Chaser;

import io.github.fabricators_of_create.porting_lib.fluids.FluidStack;
import joptsimple.internal.Strings;
import net.fabricmc.fabric.api.transfer.v1.fluid.FluidVariant;
import net.fabricmc.fabric.api.transfer.v1.transaction.TransactionContext;
import net.minecraft.class_124;
import net.minecraft.class_1937;
import net.minecraft.class_2338;
import net.minecraft.class_2350;
import net.minecraft.class_2487;
import net.minecraft.class_2561;
import net.minecraft.class_2680;
import net.minecraft.class_3532;
import net.minecraft.class_5250;

public class BoilerData {

	static final int SAMPLE_RATE = 5;

	private static final int waterSupplyPerLevel = 10 * 81; // fabric: values are 81 times greater
	private static final float passiveEngineEfficiency = 1 / 8f;

	// pooled water supply
	long gatheredSupply;
	double[] supplyOverTime = new double[10];
	int ticksUntilNextSample;
	int currentIndex;

	// heat score
	public boolean needsHeatLevelUpdate;
	public boolean passiveHeat;
	public int activeHeat;

	public double waterSupply;
	public int attachedEngines;
	public int attachedWhistles;

	// display
	private int maxHeatForSize = 0;
	private int maxHeatForWater = 0;
	private int minValue = 0;
	private int maxValue = 0;

	public LerpedFloat gauge = LerpedFloat.linear();

	public void tick(FluidTankBlockEntity controller) {
		if (!isActive())
			return;
		if (controller.method_10997().field_9236) {
			gauge.tickChaser();
			float current = gauge.getValue(1);
			if (current > 1 && Create.RANDOM.nextFloat() < 1 / 2f)
				gauge.setValueNoUpdate(current + Math.min(-(current - 1) * Create.RANDOM.nextFloat(), 0));
			return;
		}
		if (needsHeatLevelUpdate && updateTemperature(controller))
			controller.notifyUpdate();
		ticksUntilNextSample--;
		if (ticksUntilNextSample > 0)
			return;
		long capacity = controller.tankInventory.getCapacity();
		if (capacity == 0)
			return;

		ticksUntilNextSample = SAMPLE_RATE;
		supplyOverTime[currentIndex] = gatheredSupply / (float) SAMPLE_RATE;
		waterSupply = Math.max(waterSupply, supplyOverTime[currentIndex]);
		currentIndex = (currentIndex + 1) % supplyOverTime.length;
		gatheredSupply = 0;

		if (currentIndex == 0) {
			waterSupply = 0;
			for (double i : supplyOverTime)
				waterSupply = Math.max(i, waterSupply);
		}

		if (controller instanceof CreativeFluidTankBlockEntity)
			waterSupply = waterSupplyPerLevel * 20;

		if (getActualHeat(controller.getTotalTankSize()) == 18)
			controller.award(AllAdvancements.STEAM_ENGINE_MAXED);

		controller.notifyUpdate();
	}

	public int getTheoreticalHeatLevel() {
		return activeHeat;
	}

	public int getMaxHeatLevelForBoilerSize(int boilerSize) {
		return (int) Math.min(18, boilerSize / 4);
	}

	public int getMaxHeatLevelForWaterSupply() {
		return (int) Math.min(18, class_3532.method_15384(waterSupply) / waterSupplyPerLevel);
	}

	public boolean isPassive() {
		return passiveHeat && maxHeatForSize > 0 && maxHeatForWater > 0;
	}

	public boolean isPassive(int boilerSize) {
		calcMinMaxForSize(boilerSize);
		return isPassive();
	}

	public float getEngineEfficiency(int boilerSize) {
		if (isPassive(boilerSize))
			return passiveEngineEfficiency / attachedEngines;
		if (activeHeat == 0)
			return 0;
		int actualHeat = getActualHeat(boilerSize);
		return attachedEngines <= actualHeat ? 1 : (float) actualHeat / attachedEngines;
	}

	private int getActualHeat(int boilerSize) {
		int forBoilerSize = getMaxHeatLevelForBoilerSize(boilerSize);
		int forWaterSupply = getMaxHeatLevelForWaterSupply();
		int actualHeat = Math.min(activeHeat, Math.min(forWaterSupply, forBoilerSize));
		return actualHeat;
	}

	public boolean addToGoggleTooltip(List<class_2561> tooltip, boolean isPlayerSneaking, int boilerSize) {
		if (!isActive())
			return false;

		class_2561 indent = Components.literal(IHaveGoggleInformation.spacing);
		class_2561 indent2 = Components.literal(IHaveGoggleInformation.spacing + " ");

		calcMinMaxForSize(boilerSize);

		tooltip.add(indent.method_27662()
			.method_10852(
				Lang.translateDirect("boiler.status", getHeatLevelTextComponent().method_27692(class_124.field_1060))));
		tooltip.add(indent2.method_27662()
			.method_10852(getSizeComponent(true, false)));
		tooltip.add(indent2.method_27662()
			.method_10852(getWaterComponent(true, false)));
		tooltip.add(indent2.method_27662()
			.method_10852(getHeatComponent(true, false)));

		if (attachedEngines == 0)
			return true;
int boilerLevel = Math.min(activeHeat, Math.min(maxHeatForWater, maxHeatForSize));
		double totalSU = getEngineEfficiency(boilerSize) * 16 * Math.max(boilerLevel, attachedEngines)
			* BlockStressValues.getCapacity(AllBlocks.STEAM_ENGINE.get());

		tooltip.add(Components.immutableEmpty());

		if (attachedEngines > 0 && maxHeatForSize > 0 && maxHeatForWater == 0 && (passiveHeat ? 1 : activeHeat) > 0) {
			Lang.translate("boiler.water_input_rate")
				.style(class_124.field_1080)
				.forGoggles(tooltip);
			Lang.number(waterSupply)
				.style(class_124.field_1078)
				.add(Lang.translate("generic.unit.millibuckets"))
				.add(Lang.text(" / ")
					.style(class_124.field_1080))
				.add(Lang.translate("boiler.per_tick", Lang.number(waterSupplyPerLevel)
					.add(Lang.translate("generic.unit.millibuckets")))
					.style(class_124.field_1063))
				.forGoggles(tooltip, 1);
			return true;
		}

		Lang.translate("tooltip.capacityProvided")
			.style(class_124.field_1080)
			.forGoggles(tooltip);

		Lang.number(totalSU)
			.translate("generic.unit.stress")
			.style(class_124.field_1075)
			.space()
			.add((attachedEngines == 1 ? Lang.translate("boiler.via_one_engine")
				: Lang.translate("boiler.via_engines", attachedEngines)).style(class_124.field_1063))
			.forGoggles(tooltip, 1);

		return true;
	}

	public void calcMinMaxForSize(int boilerSize) {
		maxHeatForSize = getMaxHeatLevelForBoilerSize(boilerSize);
		maxHeatForWater = getMaxHeatLevelForWaterSupply();

		minValue = Math.min(passiveHeat ? 1 : activeHeat, Math.min(maxHeatForWater, maxHeatForSize));
		maxValue = Math.max(passiveHeat ? 1 : activeHeat, Math.max(maxHeatForWater, maxHeatForSize));
	}

	@NotNull
	public class_5250 getHeatLevelTextComponent() {
		int boilerLevel = Math.min(activeHeat, Math.min(maxHeatForWater, maxHeatForSize));

		return isPassive() ? Lang.translateDirect("boiler.passive")
			: (boilerLevel == 0 ? Lang.translateDirect("boiler.idle")
				: boilerLevel == 18 ? Lang.translateDirect("boiler.max_lvl")
					: Lang.translateDirect("boiler.lvl", String.valueOf(boilerLevel)));
	}

	public class_5250 getSizeComponent(boolean forGoggles, boolean useBlocksAsBars, class_124... styles) {
		return componentHelper("size", maxHeatForSize, forGoggles, useBlocksAsBars, styles);
	}

	public class_5250 getWaterComponent(boolean forGoggles, boolean useBlocksAsBars, class_124... styles) {
		return componentHelper("water", maxHeatForWater, forGoggles, useBlocksAsBars, styles);
	}

	public class_5250 getHeatComponent(boolean forGoggles, boolean useBlocksAsBars, class_124... styles) {
		return componentHelper("heat", passiveHeat ? 1 : activeHeat, forGoggles, useBlocksAsBars, styles);
	}

	private class_5250 componentHelper(String label, int level, boolean forGoggles, boolean useBlocksAsBars,
		class_124... styles) {
		class_5250 base = useBlocksAsBars ? blockComponent(level) : barComponent(level);

		if (!forGoggles)
			return base;

		class_124 style1 = styles.length >= 1 ? styles[0] : class_124.field_1080;
		class_124 style2 = styles.length >= 2 ? styles[1] : class_124.field_1063;

		return Lang.translateDirect("boiler." + label)
			.method_27692(style1)
			.method_10852(Lang.translateDirect("boiler." + label + "_dots")
				.method_27692(style2))
			.method_10852(base);
	}

	private class_5250 blockComponent(int level) {
		return Components.literal(
			"" + "\u2588".repeat(minValue) + "\u2592".repeat(level - minValue) + "\u2591".repeat(maxValue - level));
	}

	private class_5250 barComponent(int level) {
		return Components.empty()
			.method_10852(bars(Math.max(0, minValue - 1), class_124.field_1077))
			.method_10852(bars(minValue > 0 ? 1 : 0, class_124.field_1060))
			.method_10852(bars(Math.max(0, level - minValue), class_124.field_1077))
			.method_10852(bars(Math.max(0, maxValue - level), class_124.field_1079))
			.method_10852(bars(Math.max(0, Math.min(18 - maxValue, ((maxValue / 5 + 1) * 5) - maxValue)),
				class_124.field_1063));

	}

	private class_5250 bars(int level, class_124 format) {
		return Components.literal(Strings.repeat('|', level))
			.method_27692(format);
	}

	public boolean evaluate(FluidTankBlockEntity controller) {
		class_2338 controllerPos = controller.method_11016();
		class_1937 level = controller.method_10997();
		int prevEngines = attachedEngines;
		int prevWhistles = attachedWhistles;
		attachedEngines = 0;
		attachedWhistles = 0;

		for (int yOffset = 0; yOffset < controller.height; yOffset++) {
			for (int xOffset = 0; xOffset < controller.width; xOffset++) {
				for (int zOffset = 0; zOffset < controller.width; zOffset++) {

					class_2338 pos = controllerPos.method_10069(xOffset, yOffset, zOffset);
					class_2680 blockState = level.method_8320(pos);
					if (!FluidTankBlock.isTank(blockState))
						continue;
					for (class_2350 d : Iterate.directions) {
						class_2338 attachedPos = pos.method_10093(d);
						class_2680 attachedState = level.method_8320(attachedPos);
						if (AllBlocks.STEAM_ENGINE.has(attachedState) && SteamEngineBlock.getFacing(attachedState) == d)
							attachedEngines++;
						if (AllBlocks.STEAM_WHISTLE.has(attachedState)
							&& WhistleBlock.getAttachedDirection(attachedState)
								.method_10153() == d)
							attachedWhistles++;
					}
				}
			}
		}

		needsHeatLevelUpdate = true;
		return prevEngines != attachedEngines || prevWhistles != attachedWhistles;
	}

	public void checkPipeOrganAdvancement(FluidTankBlockEntity controller) {
		if (!controller.getBehaviour(AdvancementBehaviour.TYPE)
			.isOwnerPresent())
			return;

		class_2338 controllerPos = controller.method_11016();
		class_1937 level = controller.method_10997();
		Set<Integer> whistlePitches = new HashSet<>();

		for (int yOffset = 0; yOffset < controller.height; yOffset++) {
			for (int xOffset = 0; xOffset < controller.width; xOffset++) {
				for (int zOffset = 0; zOffset < controller.width; zOffset++) {

					class_2338 pos = controllerPos.method_10069(xOffset, yOffset, zOffset);
					class_2680 blockState = level.method_8320(pos);
					if (!FluidTankBlock.isTank(blockState))
						continue;
					for (class_2350 d : Iterate.directions) {
						class_2338 attachedPos = pos.method_10093(d);
						class_2680 attachedState = level.method_8320(attachedPos);
						if (AllBlocks.STEAM_WHISTLE.has(attachedState)
							&& WhistleBlock.getAttachedDirection(attachedState)
								.method_10153() == d) {
							if (level.method_8321(attachedPos) instanceof WhistleBlockEntity wbe)
								whistlePitches.add(wbe.getPitchId());
						}
					}
				}
			}
		}

		if (whistlePitches.size() >= 12)
			controller.award(AllAdvancements.PIPE_ORGAN);
	}

	public boolean updateTemperature(FluidTankBlockEntity controller) {
		class_2338 controllerPos = controller.method_11016();
		class_1937 level = controller.method_10997();
		needsHeatLevelUpdate = false;

		boolean prevPassive = passiveHeat;
		int prevActive = activeHeat;
		passiveHeat = false;
		activeHeat = 0;

		for (int xOffset = 0; xOffset < controller.width; xOffset++) {
			for (int zOffset = 0; zOffset < controller.width; zOffset++) {
				class_2338 pos = controllerPos.method_10069(xOffset, -1, zOffset);
				class_2680 blockState = level.method_8320(pos);
				float heat = BoilerHeaters.getActiveHeat(level, pos, blockState);
				if (heat == 0) {
					passiveHeat = true;
				} else if (heat > 0) {
					activeHeat += heat;
				}
			}
		}

		passiveHeat &= activeHeat == 0;

		return prevActive != activeHeat || prevPassive != passiveHeat;
	}

	public boolean isActive() {
		return attachedEngines > 0 || attachedWhistles > 0;
	}

	public void clear() {
		waterSupply = 0;
		activeHeat = 0;
		passiveHeat = false;
		attachedEngines = 0;
		Arrays.fill(supplyOverTime, 0);
	}

	public class_2487 write() {
		class_2487 nbt = new class_2487();
		nbt.method_10549("Supply", waterSupply);
		nbt.method_10569("ActiveHeat", activeHeat);
		nbt.method_10556("PassiveHeat", passiveHeat);
		nbt.method_10569("Engines", attachedEngines);
		nbt.method_10569("Whistles", attachedWhistles);
		nbt.method_10556("Update", needsHeatLevelUpdate);
		return nbt;
	}

	public void read(class_2487 nbt, int boilerSize) {
		waterSupply = nbt.method_10574("Supply");
		activeHeat = nbt.method_10550("ActiveHeat");
		passiveHeat = nbt.method_10577("PassiveHeat");
		attachedEngines = nbt.method_10550("Engines");
		attachedWhistles = nbt.method_10550("Whistles");
		needsHeatLevelUpdate = nbt.method_10577("Update");
		Arrays.fill(supplyOverTime, waterSupply);

		int forBoilerSize = getMaxHeatLevelForBoilerSize(boilerSize);
		int forWaterSupply = getMaxHeatLevelForWaterSupply();
		int actualHeat = Math.min(activeHeat, Math.min(forWaterSupply, forBoilerSize));
		float target = isPassive(boilerSize) ? 1 / 8f : forBoilerSize == 0 ? 0 : actualHeat / (forBoilerSize * 1f);
		gauge.chase(target, 0.125f, Chaser.EXP);
	}

	public BoilerFluidHandler createHandler() {
		return new BoilerFluidHandler();
	}

	public class BoilerFluidHandler extends SmartFluidTank {

		public BoilerFluidHandler() {
			super(10 * FluidConstants.BUCKET, (f) -> {});
			setFluid(FluidStack.EMPTY);
		}

		public boolean isFluidValid(FluidVariant stack) {
			return FluidHelper.isWater(stack.getFluid());
		}

		@Override
		public long insert(FluidVariant insertedVariant, long maxAmount, TransactionContext transaction) {
			if (!isFluidValid(insertedVariant))
				return 0;
			gatheredSupply += maxAmount;
			return maxAmount;
		}

		@Override
		public long extract(FluidVariant extractedVariant, long maxAmount, TransactionContext transaction) {
			return 0;
		}
	}

}
