package com.simibubi.create.content.fluids.tank;

import java.util.ArrayList;
import java.util.List;
import net.minecraft.class_1937;
import net.minecraft.class_2248;
import net.minecraft.class_2338;
import net.minecraft.class_2680;
import net.minecraft.class_2960;
import net.minecraft.class_7923;
import org.jetbrains.annotations.Nullable;

import com.simibubi.create.AllBlocks;
import com.simibubi.create.AllTags.AllBlockTags;
import com.simibubi.create.content.processing.burner.BlazeBurnerBlock;
import com.simibubi.create.content.processing.burner.BlazeBurnerBlock.HeatLevel;
import com.simibubi.create.foundation.utility.AttachedRegistry;
import com.simibubi.create.foundation.utility.BlockHelper;

public class BoilerHeaters {
	private static final AttachedRegistry<class_2248, Heater> BLOCK_HEATERS = new AttachedRegistry<>(class_7923.field_41175);
	private static final List<HeaterProvider> GLOBAL_HEATERS = new ArrayList<>();

	public static void registerHeater(class_2960 block, Heater heater) {
		BLOCK_HEATERS.register(block, heater);
	}

	public static void registerHeater(class_2248 block, Heater heater) {
		BLOCK_HEATERS.register(block, heater);
	}

	public static void registerHeaterProvider(HeaterProvider provider) {
		GLOBAL_HEATERS.add(provider);
	}

	/**
	 * A return value of {@code -1} represents no heat.
	 * A return value of {@code 0} represents passive heat.
	 * All other positive values are used as the amount of active heat.
	 */
	public static float getActiveHeat(class_1937 level, class_2338 pos, class_2680 state) {
		Heater heater = BLOCK_HEATERS.get(state.method_26204());
		if (heater != null) {
			return heater.getActiveHeat(level, pos, state);
		}

		for (HeaterProvider provider : GLOBAL_HEATERS) {
			heater = provider.getHeater(level, pos, state);
			if (heater != null) {
				return heater.getActiveHeat(level, pos, state);
			}
		}

		return -1;
	}

	public static void registerDefaults() {
		registerHeater(AllBlocks.BLAZE_BURNER.get(), (level, pos, state) -> {
			HeatLevel value = state.method_11654(BlazeBurnerBlock.HEAT_LEVEL);
			if (value == HeatLevel.NONE) {
				return -1;
			}
			if (value == HeatLevel.SEETHING) {
				return 2;
			}
			if (value.isAtLeast(HeatLevel.FADING)) {
				return 1;
			}
			return 0;
		});

		registerHeaterProvider((level, pos, state) -> {
			if (AllBlockTags.PASSIVE_BOILER_HEATERS.matches(state) && BlockHelper.isNotUnheated(state)) {
				return (level1, pos1, state1) -> 0;
			}
			return null;
		});
	}

	public interface Heater {
		/**
		 * A return value of {@code -1} represents no heat.
		 * A return value of {@code 0} represents passive heat.
		 * All other positive values are used as the amount of active heat.
		 */
		float getActiveHeat(class_1937 level, class_2338 pos, class_2680 state);
	}

	public interface HeaterProvider {
		@Nullable
		Heater getHeater(class_1937 level, class_2338 pos, class_2680 state);
	}
}
