package com.simibubi.create.content.fluids.tank;

import static java.lang.Math.abs;

import java.util.List;

import javax.annotation.Nullable;

import com.simibubi.create.api.connectivity.ConnectivityHandler;
import com.simibubi.create.content.equipment.goggles.IHaveGoggleInformation;
import com.simibubi.create.content.fluids.tank.FluidTankBlock.Shape;
import com.simibubi.create.foundation.advancement.AllAdvancements;
import com.simibubi.create.foundation.blockEntity.IMultiBlockEntityContainer;
import com.simibubi.create.foundation.blockEntity.SmartBlockEntity;
import com.simibubi.create.foundation.blockEntity.behaviour.BlockEntityBehaviour;
import com.simibubi.create.foundation.fluid.SmartFluidTank;
import com.simibubi.create.foundation.utility.animation.LerpedFloat;
import com.simibubi.create.foundation.utility.animation.LerpedFloat.Chaser;
import com.simibubi.create.infrastructure.config.AllConfigs;

import io.github.fabricators_of_create.porting_lib.block.CustomRenderBoundingBoxBlockEntity;
import io.github.fabricators_of_create.porting_lib.transfer.TransferUtil;
import io.github.fabricators_of_create.porting_lib.transfer.fluid.FluidTank;
import io.github.fabricators_of_create.porting_lib.fluids.FluidStack;
import net.fabricmc.fabric.api.transfer.v1.fluid.FluidConstants;
import net.fabricmc.fabric.api.transfer.v1.fluid.FluidVariant;
import net.fabricmc.fabric.api.transfer.v1.fluid.FluidVariantAttributeHandler;
import net.fabricmc.fabric.api.transfer.v1.fluid.FluidVariantAttributes;
import net.fabricmc.fabric.api.transfer.v1.storage.Storage;
import net.fabricmc.fabric.api.transfer.v1.storage.base.SidedStorageBlockEntity;
import net.fabricmc.fabric.api.transfer.v1.transaction.Transaction;
import net.minecraft.class_2338;
import net.minecraft.class_2350;
import net.minecraft.class_238;
import net.minecraft.class_2487;
import net.minecraft.class_2512;
import net.minecraft.class_2561;
import net.minecraft.class_2586;
import net.minecraft.class_2591;
import net.minecraft.class_2680;

public class FluidTankBlockEntity extends SmartBlockEntity implements IHaveGoggleInformation, IMultiBlockEntityContainer.Fluid, CustomRenderBoundingBoxBlockEntity, SidedStorageBlockEntity {

	private static final int MAX_SIZE = 3;

	protected boolean forceFluidLevelUpdate;
	protected SmartFluidTank tankInventory;
	protected FluidTank exposedTank;
	protected class_2338 controller;
	protected class_2338 lastKnownPos;
	protected boolean updateConnectivity;
	protected boolean window;
	protected int luminosity;
	protected int width;
	protected int height;

	public BoilerData boiler;

	private static final int SYNC_RATE = 8;
	protected int syncCooldown;
	protected boolean queuedSync;

	// For rendering purposes only
	private LerpedFloat fluidLevel;

	public FluidTankBlockEntity(class_2591<?> type, class_2338 pos, class_2680 state) {
		super(type, pos, state);
		tankInventory = createInventory();
		forceFluidLevelUpdate = true;
		updateConnectivity = false;
		window = true;
		height = 1;
		width = 1;
		boiler = new BoilerData();
//		refreshCapability(); // fabric: lazy init to prevent access too early
	}

	protected SmartFluidTank createInventory() {
		return new SmartFluidTank(getCapacityMultiplier(), this::onFluidStackChanged);
	}

	protected void updateConnectivity() {
		updateConnectivity = false;
		if (field_11863.field_9236)
			return;
		if (!isController())
			return;
		ConnectivityHandler.formMulti(this);
	}

	@Override
	public void tick() {
		super.tick();
		if (syncCooldown > 0) {
			syncCooldown--;
			if (syncCooldown == 0 && queuedSync)
				sendData();
		}

		if (lastKnownPos == null)
			lastKnownPos = method_11016();
		else if (!lastKnownPos.equals(field_11867) && field_11867 != null) {
			onPositionChanged();
			return;
		}

		if (updateConnectivity)
			updateConnectivity();
		if (fluidLevel != null)
			fluidLevel.tickChaser();
		if (isController())
			boiler.tick(this);
	}

	@Override
	public class_2338 getLastKnownPos() {
		return lastKnownPos;
	}

	@Override
	public boolean isController() {
		return controller == null || field_11867.method_10263() == controller.method_10263()
			&& field_11867.method_10264() == controller.method_10264() && field_11867.method_10260() == controller.method_10260();
	}

	@Override
	public void initialize() {
		super.initialize();
		sendData();
		if (field_11863.field_9236)
			invalidateRenderBoundingBox();
	}

	private void onPositionChanged() {
		removeController(true);
		lastKnownPos = field_11867;
	}

	protected void onFluidStackChanged(FluidStack newFluidStack) {
		if (!method_11002())
			return;

		FluidVariantAttributeHandler handler = FluidVariantAttributes.getHandlerOrDefault(newFluidStack.getFluid());
		FluidVariant variant = newFluidStack.getType();
		int luminosity = (int) (handler.getLuminance(variant) / 1.2f);
		boolean reversed = handler.isLighterThanAir(variant);
		int maxY = (int) ((getFillState() * height) + 1);

		for (int yOffset = 0; yOffset < height; yOffset++) {
			boolean isBright = reversed ? (height - yOffset <= maxY) : (yOffset < maxY);
			int actualLuminosity = isBright ? luminosity : luminosity > 0 ? 1 : 0;

			for (int xOffset = 0; xOffset < width; xOffset++) {
				for (int zOffset = 0; zOffset < width; zOffset++) {
					class_2338 pos = this.field_11867.method_10069(xOffset, yOffset, zOffset);
					FluidTankBlockEntity tankAt = ConnectivityHandler.partAt(method_11017(), field_11863, pos);
					if (tankAt == null)
						continue;
					field_11863.method_8455(pos, tankAt.method_11010()
						.method_26204());
					if (tankAt.luminosity == actualLuminosity)
						continue;
					tankAt.setLuminosity(actualLuminosity);
				}
			}
		}

		if (!field_11863.field_9236) {
			method_5431();
			sendData();
		}

		if (isVirtual()) {
			if (fluidLevel == null)
				fluidLevel = LerpedFloat.linear()
					.startWithValue(getFillState());
			fluidLevel.chase(getFillState(), .5f, Chaser.EXP);
		}
	}

	protected void setLuminosity(int luminosity) {
		if (field_11863.field_9236)
			return;
		if (this.luminosity == luminosity)
			return;
		this.luminosity = luminosity;
		updateStateLuminosity();
	}

	protected void updateStateLuminosity() {
		if (field_11863.field_9236)
			return;
		int actualLuminosity = luminosity;
		FluidTankBlockEntity controllerBE = getControllerBE();
		if (controllerBE == null || !controllerBE.window)
			actualLuminosity = 0;
		refreshBlockState();
		class_2680 state = method_11010();
		if (state.method_11654(FluidTankBlock.LIGHT_LEVEL) != actualLuminosity) {
			field_11863.method_8652(field_11867, state.method_11657(FluidTankBlock.LIGHT_LEVEL, actualLuminosity), 23);
		}
	}

	@SuppressWarnings("unchecked")
	@Override
	public FluidTankBlockEntity getControllerBE() {
		if (isController())
			return this;
		class_2586 blockEntity = field_11863.method_8321(controller);
		if (blockEntity instanceof FluidTankBlockEntity)
			return (FluidTankBlockEntity) blockEntity;
		return null;
	}

	public void applyFluidTankSize(int blocks) {
		tankInventory.setCapacity(blocks * getCapacityMultiplier());
		long overflow = tankInventory.getFluidAmount() - tankInventory.getCapacity();
		if (overflow > 0)
			TransferUtil.extract(tankInventory, tankInventory.variant, overflow);
		forceFluidLevelUpdate = true;
	}

	public void removeController(boolean keepFluids) {
		if (field_11863.field_9236)
			return;
		updateConnectivity = true;
		if (!keepFluids)
			applyFluidTankSize(1);
		controller = null;
		width = 1;
		height = 1;
		boiler.clear();
		onFluidStackChanged(tankInventory.getFluid());

		class_2680 state = method_11010();
		if (FluidTankBlock.isTank(state)) {
			state = state.method_11657(FluidTankBlock.BOTTOM, true);
			state = state.method_11657(FluidTankBlock.TOP, true);
			state = state.method_11657(FluidTankBlock.SHAPE, window ? Shape.WINDOW : Shape.PLAIN);
			method_10997().method_8652(field_11867, state, 23);
		}

		refreshCapability();
		method_5431();
		sendData();
	}

	public void toggleWindows() {
		FluidTankBlockEntity be = getControllerBE();
		if (be == null)
			return;
		if (be.boiler.isActive())
			return;
		be.setWindows(!be.window);
	}

	public void updateBoilerTemperature() {
		FluidTankBlockEntity be = getControllerBE();
		if (be == null)
			return;
		if (!be.boiler.isActive())
			return;
		be.boiler.needsHeatLevelUpdate = true;
	}

	public void sendDataImmediately() {
		syncCooldown = 0;
		queuedSync = false;
		sendData();
	}

	@Override
	public void sendData() {
		if (syncCooldown > 0) {
			queuedSync = true;
			return;
		}
		super.sendData();
		queuedSync = false;
		syncCooldown = SYNC_RATE;
	}

	public void setWindows(boolean window) {
		this.window = window;
		for (int yOffset = 0; yOffset < height; yOffset++) {
			for (int xOffset = 0; xOffset < width; xOffset++) {
				for (int zOffset = 0; zOffset < width; zOffset++) {

					class_2338 pos = this.field_11867.method_10069(xOffset, yOffset, zOffset);
					class_2680 blockState = field_11863.method_8320(pos);
					if (!FluidTankBlock.isTank(blockState))
						continue;

					Shape shape = Shape.PLAIN;
					if (window) {
						// SIZE 1: Every tank has a window
						if (width == 1)
							shape = Shape.WINDOW;
						// SIZE 2: Every tank has a corner window
						if (width == 2)
							shape = xOffset == 0 ? zOffset == 0 ? Shape.WINDOW_NW : Shape.WINDOW_SW
								: zOffset == 0 ? Shape.WINDOW_NE : Shape.WINDOW_SE;
						// SIZE 3: Tanks in the center have a window
						if (width == 3 && abs(abs(xOffset) - abs(zOffset)) == 1)
							shape = Shape.WINDOW;
					}

					field_11863.method_8652(pos, blockState.method_11657(FluidTankBlock.SHAPE, shape), 23);
					class_2586 be = field_11863.method_8321(pos);
					if (be instanceof FluidTankBlockEntity tankAt)
						tankAt.updateStateLuminosity();
					field_11863.method_8398()
							.method_12130()
							.method_15513(pos);
				}
			}
		}
	}

	public void updateBoilerState() {
		if (!isController())
			return;

		boolean wasBoiler = boiler.isActive();
		boolean changed = boiler.evaluate(this);

		if (wasBoiler != boiler.isActive()) {
			if (boiler.isActive())
				setWindows(false);

			for (int yOffset = 0; yOffset < height; yOffset++)
				for (int xOffset = 0; xOffset < width; xOffset++)
					for (int zOffset = 0; zOffset < width; zOffset++)
						if (field_11863.method_8321(
							field_11867.method_10069(xOffset, yOffset, zOffset)) instanceof FluidTankBlockEntity fbe)
							fbe.refreshCapability();
		}

		if (changed) {
			notifyUpdate();
			boiler.checkPipeOrganAdvancement(this);
		}
	}

	@Override
	public void setController(class_2338 controller) {
		if (field_11863.field_9236 && !isVirtual())
			return;
		if (controller.equals(this.controller))
			return;
		this.controller = controller;
		refreshCapability();
		method_5431();
		sendData();
	}

	private void refreshCapability() {
		exposedTank = handlerForCapability();
	}

	private FluidTank handlerForCapability() {
		return isController() ? boiler.isActive() ? boiler.createHandler() : tankInventory
				: getControllerBE() != null ? getControllerBE().handlerForCapability() : new FluidTank(0);
	}

	@Override
	public class_2338 getController() {
		return isController() ? field_11867 : controller;
	}

	@Override
	protected class_238 createRenderBoundingBox() {
		if (isController())
			return super.createRenderBoundingBox().method_1012(width - 1, height - 1, width - 1);
		else
			return super.createRenderBoundingBox();
	}

	@Nullable
	public FluidTankBlockEntity getOtherFluidTankBlockEntity(class_2350 direction) {
		class_2586 otherBE = field_11863.method_8321(field_11867.method_10093(direction));
		if (otherBE instanceof FluidTankBlockEntity)
			return (FluidTankBlockEntity) otherBE;
		return null;
	}

	@Override
	public boolean addToGoggleTooltip(List<class_2561> tooltip, boolean isPlayerSneaking) {
		FluidTankBlockEntity controllerBE = getControllerBE();
		if (controllerBE == null)
			return false;
		if (controllerBE.boiler.addToGoggleTooltip(tooltip, isPlayerSneaking, controllerBE.getTotalTankSize()))
			return true;
		return containedFluidTooltip(tooltip, isPlayerSneaking,
				controllerBE.getFluidStorage(null));
	}

	@Override
	protected void read(class_2487 compound, boolean clientPacket) {
		super.read(compound, clientPacket);

		class_2338 controllerBefore = controller;
		int prevSize = width;
		int prevHeight = height;
		int prevLum = luminosity;

		updateConnectivity = compound.method_10545("Uninitialized");
		luminosity = compound.method_10550("Luminosity");
		controller = null;
		lastKnownPos = null;

		if (compound.method_10545("LastKnownPos"))
			lastKnownPos = class_2512.method_10691(compound.method_10562("LastKnownPos"));
		if (compound.method_10545("Controller"))
			controller = class_2512.method_10691(compound.method_10562("Controller"));

		if (isController()) {
			window = compound.method_10577("Window");
			width = compound.method_10550("Size");
			height = compound.method_10550("Height");
			tankInventory.setCapacity(getTotalTankSize() * getCapacityMultiplier());
			tankInventory.readFromNBT(compound.method_10562("TankContent"));
			if (tankInventory.getSpace() < 0) {
				try (Transaction t = TransferUtil.getTransaction()) {
					tankInventory.extract(tankInventory.variant, -tankInventory.getSpace(), t);
					t.commit();
				}
			}
		}

		boiler.read(compound.method_10562("Boiler"), width * width * height);

		if (compound.method_10545("ForceFluidLevel") || fluidLevel == null)
			fluidLevel = LerpedFloat.linear()
				.startWithValue(getFillState());

		if (!clientPacket)
			return;

		boolean changeOfController =
			controllerBefore == null ? controller != null : !controllerBefore.equals(controller);
		if (changeOfController || prevSize != width || prevHeight != height) {
			if (method_11002())
				field_11863.method_8413(method_11016(), method_11010(), method_11010(), 16);
			if (isController())
				tankInventory.setCapacity(getCapacityMultiplier() * getTotalTankSize());
			invalidateRenderBoundingBox();
		}
		if (isController()) {
			float fillState = getFillState();
			if (compound.method_10545("ForceFluidLevel") || fluidLevel == null)
				fluidLevel = LerpedFloat.linear()
					.startWithValue(fillState);
			fluidLevel.chase(fillState, 0.5f, Chaser.EXP);
		}
		if (luminosity != prevLum && method_11002())
			field_11863.method_8398()
				.method_12130()
				.method_15513(field_11867);

		if (compound.method_10545("LazySync"))
			fluidLevel.chase(fluidLevel.getChaseTarget(), 0.125f, Chaser.EXP);
	}

	public float getFillState() {
		return (float) tankInventory.getFluidAmount() / tankInventory.getCapacity();
	}

	@Override
	public void write(class_2487 compound, boolean clientPacket) {
		if (updateConnectivity)
			compound.method_10556("Uninitialized", true);
		compound.method_10566("Boiler", boiler.write());
		if (lastKnownPos != null)
			compound.method_10566("LastKnownPos", class_2512.method_10692(lastKnownPos));
		if (!isController())
			compound.method_10566("Controller", class_2512.method_10692(controller));
		if (isController()) {
			compound.method_10556("Window", window);
			compound.method_10566("TankContent", tankInventory.writeToNBT(new class_2487()));
			compound.method_10569("Size", width);
			compound.method_10569("Height", height);
		}
		compound.method_10569("Luminosity", luminosity);
		super.write(compound, clientPacket);

		if (!clientPacket)
			return;
		if (forceFluidLevelUpdate)
			compound.method_10556("ForceFluidLevel", true);
		if (queuedSync)
			compound.method_10556("LazySync", true);
		forceFluidLevelUpdate = false;
	}

	@Override
	public void invalidate() {
		super.invalidate();
	}

	@Override
	public void addBehaviours(List<BlockEntityBehaviour> behaviours) {
		registerAwardables(behaviours, AllAdvancements.STEAM_ENGINE_MAXED, AllAdvancements.PIPE_ORGAN);
	}

	public FluidTank getTankInventory() {
		return tankInventory;
	}

	public int getTotalTankSize() {
		return width * width * height;
	}

	public static int getMaxSize() {
		return MAX_SIZE;
	}

	public static long getCapacityMultiplier() {
		return AllConfigs.server().fluids.fluidTankCapacity.get() * FluidConstants.BUCKET;
	}

	public static int getMaxHeight() {
		return AllConfigs.server().fluids.fluidTankMaxHeight.get();
	}

	public LerpedFloat getFluidLevel() {
		return fluidLevel;
	}

	public void setFluidLevel(LerpedFloat fluidLevel) {
		this.fluidLevel = fluidLevel;
	}

	@Override
	public void preventConnectivityUpdate() {
		updateConnectivity = false;
	}

	// fabric: see comment in FluidTankItem
	public void queueConnectivityUpdate() {
		updateConnectivity = true;
	}

	@Override
	public void notifyMultiUpdated() {
		class_2680 state = this.method_11010();
		if (FluidTankBlock.isTank(state)) { // safety
			state = state.method_11657(FluidTankBlock.BOTTOM, getController().method_10264() == method_11016().method_10264());
			state = state.method_11657(FluidTankBlock.TOP, getController().method_10264() + height - 1 == method_11016().method_10264());
			field_11863.method_8652(method_11016(), state, 6);
		}
		if (isController())
			setWindows(window);
		onFluidStackChanged(tankInventory.getFluid());
		updateBoilerState();
		method_5431();
	}

	@Override
	public void setExtraData(@Nullable Object data) {
		if (data instanceof Boolean)
			window = (boolean) data;
	}

	@Override
	@Nullable
	public Object getExtraData() {
		return window;
	}

	@Override
	public Object modifyExtraData(Object data) {
		if (data instanceof Boolean windows) {
			windows |= window;
			return windows;
		}
		return data;
	}

	@Override
	public class_2350.class_2351 getMainConnectionAxis() {
		return class_2350.class_2351.field_11052;
	}

	@Override
	public int getMaxLength(class_2350.class_2351 longAxis, int width) {
		if (longAxis == class_2350.class_2351.field_11052)
			return getMaxHeight();
		return getMaxWidth();
	}

	@Override
	public int getMaxWidth() {
		return MAX_SIZE;
	}

	@Override
	public int getHeight() {
		return height;
	}

	@Override
	public void setHeight(int height) {
		this.height = height;
	}

	@Override
	public int getWidth() {
		return width;
	}

	@Override
	public void setWidth(int width) {
		this.width = width;
	}

	@Override
	public boolean hasTank() {
		return true;
	}

	@Override
	public long getTankSize(int tank) {
		return getCapacityMultiplier();
	}

	@Override
	public void setTankSize(int tank, int blocks) {
		applyFluidTankSize(blocks);
	}

	@Override
	public FluidTank getTank(int tank) {
		return tankInventory;
	}

	@Override
	public FluidStack getFluid(int tank) {
		return tankInventory.getFluid()
			.copy();
	}

	@Nullable
	@Override
	public Storage<FluidVariant> getFluidStorage(@Nullable class_2350 direction) {
		if (exposedTank == null)
			refreshCapability();
		return exposedTank;
	}
}
