package com.simibubi.create.content.fluids.tank;

import org.jetbrains.annotations.ApiStatus.Internal;

import com.simibubi.create.AllBlockEntityTypes;
import com.simibubi.create.AllBlocks;
import com.simibubi.create.api.connectivity.ConnectivityHandler;

import io.github.fabricators_of_create.porting_lib.fluids.FluidStack;
import net.minecraft.class_1269;
import net.minecraft.class_1657;
import net.minecraft.class_1747;
import net.minecraft.class_1750;
import net.minecraft.class_1799;
import net.minecraft.class_1937;
import net.minecraft.class_2248;
import net.minecraft.class_2338;
import net.minecraft.class_2350;
import net.minecraft.class_2487;
import net.minecraft.class_2680;
import net.minecraft.server.MinecraftServer;

public class FluidTankItem extends class_1747 {
	// fabric: (#690) because of ordering differences, we need to delay connection by a tick when placing multiblocks with NBT.
	// If the item has NBT, it needs to be applied to a controller. However, ordering is different on fabric.
	// on forge, the block is placed, the data is set, and the tanks connect.
	// on fabric, the block is placed, the tanks connect, and the data is set.
	// However, now that the tank is not a controller, nothing happens.
	// solution: hacky static state storage. If we're placing NBT, delay connection until next tick.
	@Internal
	public static boolean IS_PLACING_NBT = false;

	public FluidTankItem(class_2248 p_i48527_1_, class_1793 p_i48527_2_) {
		super(p_i48527_1_, p_i48527_2_);
	}

	@Override
	public class_1269 method_7712(class_1750 ctx) {
		IS_PLACING_NBT = FluidTankItem.checkPlacingNbt(ctx);
		class_1269 initialResult = super.method_7712(ctx);
		IS_PLACING_NBT = false;
		if (!initialResult.method_23665())
			return initialResult;
		tryMultiPlace(ctx);
		return initialResult;
	}

	@Override
	protected boolean method_7710(class_2338 p_195943_1_, class_1937 p_195943_2_, class_1657 p_195943_3_,
		class_1799 p_195943_4_, class_2680 p_195943_5_) {
		MinecraftServer minecraftserver = p_195943_2_.method_8503();
		if (minecraftserver == null)
			return false;
		class_2487 nbt = p_195943_4_.method_7941("BlockEntityTag");
		if (nbt != null) {
			nbt.method_10551("Luminosity");
			nbt.method_10551("Size");
			nbt.method_10551("Height");
			nbt.method_10551("Controller");
			nbt.method_10551("LastKnownPos");
			if (nbt.method_10545("TankContent")) {
				FluidStack fluid = FluidStack.loadFluidStackFromNBT(nbt.method_10562("TankContent"));
				if (!fluid.isEmpty()) {
					fluid.setAmount(Math.min(FluidTankBlockEntity.getCapacityMultiplier(), fluid.getAmount()));
					nbt.method_10566("TankContent", fluid.writeToNBT(new class_2487()));
				}
			}
		}
		return super.method_7710(p_195943_1_, p_195943_2_, p_195943_3_, p_195943_4_, p_195943_5_);
	}

	private void tryMultiPlace(class_1750 ctx) {
		class_1657 player = ctx.method_8036();
		if (player == null)
			return;
		if (player.method_5715())
			return;
		class_2350 face = ctx.method_8038();
		if (!face.method_10166()
			.method_10178())
			return;
		class_1799 stack = ctx.method_8041();
		class_1937 world = ctx.method_8045();
		class_2338 pos = ctx.method_8037();
		class_2338 placedOnPos = pos.method_10093(face.method_10153());
		class_2680 placedOnState = world.method_8320(placedOnPos);

		if (!FluidTankBlock.isTank(placedOnState))
			return;
		boolean creative = method_7711().equals(AllBlocks.CREATIVE_FLUID_TANK.get());
		FluidTankBlockEntity tankAt = ConnectivityHandler.partAt(
			creative ? AllBlockEntityTypes.CREATIVE_FLUID_TANK.get() : AllBlockEntityTypes.FLUID_TANK.get(), world, placedOnPos
		);
		if (tankAt == null)
			return;
		FluidTankBlockEntity controllerBE = tankAt.getControllerBE();
		if (controllerBE == null)
			return;

		int width = controllerBE.width;
		if (width == 1)
			return;

		int tanksToPlace = 0;
		class_2338 startPos = face == class_2350.field_11033 ? controllerBE.method_11016()
			.method_10074()
			: controllerBE.method_11016()
				.method_10086(controllerBE.height);

		if (startPos.method_10264() != pos.method_10264())
			return;

		for (int xOffset = 0; xOffset < width; xOffset++) {
			for (int zOffset = 0; zOffset < width; zOffset++) {
				class_2338 offsetPos = startPos.method_10069(xOffset, 0, zOffset);
				class_2680 blockState = world.method_8320(offsetPos);
				if (FluidTankBlock.isTank(blockState))
					continue;
				if (!blockState.method_45474())
					return;
				tanksToPlace++;
			}
		}

		if (!player.method_7337() && stack.method_7947() < tanksToPlace)
			return;

		for (int xOffset = 0; xOffset < width; xOffset++) {
			for (int zOffset = 0; zOffset < width; zOffset++) {
				class_2338 offsetPos = startPos.method_10069(xOffset, 0, zOffset);
				class_2680 blockState = world.method_8320(offsetPos);
				if (FluidTankBlock.isTank(blockState))
					continue;
				class_1750 context = class_1750.method_16355(ctx, offsetPos, face);
				player.getCustomData()
					.method_10556("SilenceTankSound", true);
				IS_PLACING_NBT = checkPlacingNbt(context);
				super.method_7712(context);
				IS_PLACING_NBT = false;
				player.getCustomData()
					.method_10551("SilenceTankSound");
			}
		}
	}

	public static boolean checkPlacingNbt(class_1750 ctx) {
		class_1799 item = ctx.method_8041();
		return class_1747.method_38072(item) != null;
	}
}
