package com.simibubi.create.content.fluids.transfer;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Objects;
import java.util.Set;

import javax.annotation.Nullable;

import com.simibubi.create.foundation.advancement.AllAdvancements;
import com.simibubi.create.foundation.blockEntity.SmartBlockEntity;
import com.simibubi.create.foundation.blockEntity.behaviour.BehaviourType;
import com.simibubi.create.foundation.fluid.FluidHelper;
import com.simibubi.create.foundation.utility.BBHelper;
import com.simibubi.create.foundation.utility.Iterate;

import io.github.fabricators_of_create.porting_lib.mixin.accessors.common.accessor.LiquidBlockAccessor;
import io.github.fabricators_of_create.porting_lib.transfer.TransferUtil;
import io.github.fabricators_of_create.porting_lib.transfer.callbacks.TransactionCallback;
import io.github.fabricators_of_create.porting_lib.fluids.FluidStack;
import net.fabricmc.fabric.api.transfer.v1.fluid.FluidConstants;
import net.fabricmc.fabric.api.transfer.v1.transaction.Transaction;
import net.fabricmc.fabric.api.transfer.v1.transaction.TransactionContext;
import net.fabricmc.fabric.api.transfer.v1.transaction.base.SnapshotParticipant;
import net.minecraft.class_1937;
import net.minecraft.class_2246;
import net.minecraft.class_2338;
import net.minecraft.class_2350;
import net.minecraft.class_2404;
import net.minecraft.class_2487;
import net.minecraft.class_2680;
import net.minecraft.class_2741;
import net.minecraft.class_3341;
import net.minecraft.class_3611;
import net.minecraft.class_3612;
import net.minecraft.class_3726;
import net.minecraft.class_4706;

public class FluidDrainingBehaviour extends FluidManipulationBehaviour {

	public static final BehaviourType<FluidDrainingBehaviour> TYPE = new BehaviourType<>();

	class_3611 fluid;

	// Execution
	Set<class_2338> validationSet;
	// fabric: we need to save the queue for snapshots, so it must be a copyable type.
	class_4706<BlockPosEntry> queue;
	boolean isValid;

	// Validation
	List<BlockPosEntry> validationFrontier;
	Set<class_2338> validationVisited;
	Set<class_2338> newValidationSet;

	SnapshotParticipant<Data> snapshotParticipant = new SnapshotParticipant<>() {
		@Override
		protected Data createSnapshot() {
			FluidDrainingBehaviour b = FluidDrainingBehaviour.this;
			class_2338 rootPos = b.rootPos == null ? null : b.rootPos.method_10062();
			class_3341 box = b.affectedArea == null ? null : new class_3341(
					affectedArea.method_35415(), affectedArea.method_35416(), affectedArea.method_35417(),
					affectedArea.method_35418(), affectedArea.method_35419(), affectedArea.method_35420()
			);

			return new Data(
					fluid, isValid, rootPos, new ArrayList<>(validationFrontier), new HashSet<>(validationVisited),
					new HashSet<>(newValidationSet), revalidateIn, box, copySet(queue)
			);
		}

		@Override
		protected void readSnapshot(Data snapshot) {
			fluid = snapshot.fluid;
			isValid = snapshot.valid;
			validationFrontier = snapshot.validationFrontier;
			validationVisited = snapshot.validationVisited;
			newValidationSet = snapshot.newValidationSet;
			revalidateIn = snapshot.revalidateIn;
			affectedArea = snapshot.affectedArea;
			rootPos = snapshot.rootPos;
			queue = snapshot.queue;
		}
	};

	@Override
	protected SnapshotParticipant<?> snapshotParticipant() {
		return snapshotParticipant;
	}

	record Data(class_3611 fluid, boolean valid, class_2338 rootPos, List<BlockPosEntry> validationFrontier,
				Set<class_2338> validationVisited, Set<class_2338> newValidationSet, int revalidateIn,
				class_3341 affectedArea, class_4706<BlockPosEntry> queue) {
	}

	public FluidDrainingBehaviour(SmartBlockEntity be) {
		super(be);
		validationVisited = new HashSet<>();
		validationFrontier = new ArrayList<>();
		validationSet = new HashSet<>();
		newValidationSet = new HashSet<>();
		queue = class_4706.method_34959(this::comparePositions);
	}

	@Nullable
	public boolean pullNext(class_2338 root, TransactionContext ctx) {
		if (!frontier.isEmpty())
			return false;
		if (!Objects.equals(root, rootPos)) {
			rebuildContext(root, ctx);
			return false;
		}

		if (counterpartActed) {
			counterpartActed = false;
			softReset(root, ctx);
			return false;
		}

		if (affectedArea == null)
			affectedArea = class_3341.method_34390(root, root);

		class_1937 world = getWorld();
		if (!queue.isEmpty() && !isValid) {
			rebuildContext(root, ctx);
			return false;
		}

		snapshotParticipant.updateSnapshots(ctx);
		if (validationFrontier.isEmpty() && !queue.isEmpty() && revalidateIn == 0)
			revalidate(root);

		if (infinite) {
			blockEntity.award(AllAdvancements.HOSE_PULLEY);
			if (FluidHelper.isLava(fluid))
				blockEntity.award(AllAdvancements.HOSE_PULLEY_LAVA);

			playEffect(world, root, fluid, true);
			return true;
		}

		while (!queue.isEmpty()) {
			// Dont dequeue here, so we can decide not to dequeue a valid entry when
			// simulating
			class_2338 currentPos = queue.method_23865()
				.pos();

			class_2680 blockState = world.method_8320(currentPos);
			class_2680 emptied = blockState;
			class_3611 fluid = class_3612.field_15906;

			if (blockState.method_28498(class_2741.field_12508)
				&& blockState.method_11654(class_2741.field_12508)) {
				emptied = blockState.method_11657(class_2741.field_12508, Boolean.valueOf(false));
				fluid = class_3612.field_15910;
			} else if (blockState.method_26204() instanceof class_2404) {
				class_2404 flowingFluid = (class_2404) blockState.method_26204();
				emptied = class_2246.field_10124.method_9564();
				if (blockState.method_11654(class_2404.field_11278) == 0)
					fluid = ((LiquidBlockAccessor) flowingFluid).port_lib$getFluid().method_15751();
				else {
					affectedArea = BBHelper.encapsulate(affectedArea, class_3341.method_34390(currentPos, currentPos));
					if (!blockEntity.isVirtual())
						world.method_8652(currentPos, emptied, 2 | 16);
					dequeue(queue);
					if (queue.isEmpty()) {
						isValid = checkValid(world, rootPos);
						reset(ctx);
					}
					continue;
				}
			} else if (blockState.method_26227()
				.method_15772() != class_3612.field_15906 && blockState.method_26194(world, currentPos, class_3726.method_16194())
					.method_1110()) {
				fluid = blockState.method_26227()
					.method_15772();
				emptied = class_2246.field_10124.method_9564();
			}

			if (this.fluid == null)
				this.fluid = fluid;

			if (!this.fluid.method_15780(fluid)) {
				dequeue(queue);
				if (queue.isEmpty()) {
					isValid = checkValid(world, rootPos);
					reset(ctx);
				}
				continue;
			}

			class_3611 finalFluid = fluid;
			TransactionCallback.onSuccess(ctx, () -> {
				playEffect(world, currentPos, finalFluid, true);
				blockEntity.award(AllAdvancements.HOSE_PULLEY);
				if (infinite && FluidHelper.isLava(finalFluid))
					blockEntity.award(AllAdvancements.HOSE_PULLEY_LAVA);
			});

			if (infinite) {
				return true;
			}

			if (!blockEntity.isVirtual()) {
				world.updateSnapshots(ctx);
				world.method_8652(currentPos, emptied, 2 | 16);
			}
			affectedArea = BBHelper.encapsulate(affectedArea, currentPos);

			dequeue(queue);
			if (queue.isEmpty()) {
				isValid = checkValid(world, rootPos);
				reset(ctx);
			} else if (!validationSet.contains(currentPos)) {
				reset(ctx);
			}
			return true;
		}

		if (rootPos == null)
			return false;

		if (isValid)
			rebuildContext(root, ctx);

		return false;
	}

	protected void softReset(class_2338 root, TransactionContext ctx) {
		queue.clear();
		validationSet.clear();
		newValidationSet.clear();
		validationFrontier.clear();
		validationVisited.clear();
		visited.clear();
		infinite = false;
		setValidationTimer();
		frontier.add(new BlockPosEntry(root, 0));
		TransactionCallback.onSuccess(ctx, blockEntity::sendData);
	}

	protected boolean checkValid(class_1937 world, class_2338 root) {
		class_2338 currentPos = root;
		for (int timeout = 1000; timeout > 0 && !root.equals(blockEntity.method_11016()); timeout--) {
			FluidBlockType canPullFluidsFrom = canPullFluidsFrom(world.method_8320(currentPos), currentPos);
			if (canPullFluidsFrom == FluidBlockType.FLOWING) {
				for (class_2350 d : Iterate.directions) {
					class_2338 side = currentPos.method_10093(d);
					if (canPullFluidsFrom(world.method_8320(side), side) == FluidBlockType.SOURCE)
						return true;
				}
				currentPos = currentPos.method_10084();
				continue;
			}
			if (canPullFluidsFrom == FluidBlockType.SOURCE)
				return true;
			break;
		}
		return false;
	}

	enum FluidBlockType {
		NONE, SOURCE, FLOWING;
	}

	@Override
	public void read(class_2487 nbt, boolean clientPacket) {
		super.read(nbt, clientPacket);
		if (!clientPacket && affectedArea != null)
			frontier.add(new BlockPosEntry(rootPos, 0));
	}

	protected FluidBlockType canPullFluidsFrom(class_2680 blockState, class_2338 pos) {
		if (blockState.method_28498(class_2741.field_12508)
			&& blockState.method_11654(class_2741.field_12508))
			return FluidBlockType.SOURCE;
		if (blockState.method_26204() instanceof class_2404)
			return blockState.method_11654(class_2404.field_11278) == 0 ? FluidBlockType.SOURCE : FluidBlockType.FLOWING;
		if (blockState.method_26227()
			.method_15772() != class_3612.field_15906 && blockState.method_26194(getWorld(), pos, class_3726.method_16194())
				.method_1110())
			return FluidBlockType.SOURCE;
		return FluidBlockType.NONE;
	}

	@Override
	public void tick() {
		super.tick();
		if (rootPos != null)
			isValid = checkValid(getWorld(), rootPos);
		if (!frontier.isEmpty()) {
			continueSearch();
			return;
		}
		if (!validationFrontier.isEmpty()) {
			continueValidation();
			return;
		}
		if (revalidateIn > 0)
			revalidateIn--;
	}

	@Override
	public void lazyTick() {
		super.lazyTick();
	}

	public void rebuildContext(class_2338 root, TransactionContext ctx) {
		reset(ctx);
		rootPos = root;
		affectedArea = class_3341.method_34390(rootPos, rootPos);
		if (isValid)
			frontier.add(new BlockPosEntry(root, 0));
	}

	public void revalidate(class_2338 root) {
		validationFrontier.clear();
		validationVisited.clear();
		newValidationSet.clear();
		validationFrontier.add(new BlockPosEntry(root, 0));
		setValidationTimer();
	}

	private void continueSearch() {
		try {
			fluid = search(fluid, frontier, visited, (e, d) -> {
				BlockPosEntry entry = new BlockPosEntry(e, d);
				queue.add(entry);
				validationSet.add(e);
			}, false);
		} catch (ChunkNotLoadedException e) {
			blockEntity.sendData();
			frontier.clear();
			visited.clear();
		}

		int maxBlocks = maxBlocks();
		if (visited.size() > maxBlocks && canDrainInfinitely(fluid) && !queue.isEmpty()) {
			infinite = true;
			class_2338 firstValid = queue.method_23865()
				.pos();
			frontier.clear();
			visited.clear();
			queue.clear();
			queue.add(new BlockPosEntry(firstValid, 0));
			blockEntity.sendData();
			return;
		}

		if (!frontier.isEmpty())
			return;

		blockEntity.sendData();
		visited.clear();
	}

	private void continueValidation() {
		try {
			search(fluid, validationFrontier, validationVisited, (e, d) -> newValidationSet.add(e), false);
		} catch (ChunkNotLoadedException e) {
			validationFrontier.clear();
			validationVisited.clear();
			setLongValidationTimer();
			return;
		}

		int maxBlocks = maxBlocks();
		if (validationVisited.size() > maxBlocks && canDrainInfinitely(fluid)) {
			if (!infinite)
				reset(null);
			validationFrontier.clear();
			setLongValidationTimer();
			return;
		}

		if (!validationFrontier.isEmpty())
			return;
		if (infinite) {
			reset(null);
			return;
		}

		validationSet = newValidationSet;
		newValidationSet = new HashSet<>();
		validationVisited.clear();
	}

	@Override
	public void reset(@Nullable TransactionContext ctx) {
		super.reset(ctx);

		fluid = null;
		rootPos = null;
		queue.clear();
		validationSet.clear();
		newValidationSet.clear();
		validationFrontier.clear();
		validationVisited.clear();
		if (ctx != null) TransactionCallback.onSuccess(ctx, blockEntity::sendData);
		else blockEntity.sendData();
	}

	@Override
	public BehaviourType<?> getType() {
		return TYPE;
	}

	protected boolean isSearching() {
		return !frontier.isEmpty();
	}

	public FluidStack getDrainableFluid(class_2338 rootPos) {
		try (Transaction t = TransferUtil.getTransaction()) { // simulate pullNext
			if (fluid == null || isSearching() || !pullNext(rootPos, t)) {
				return FluidStack.EMPTY;
			}
		}
		return new FluidStack(fluid, FluidConstants.BUCKET);
	}
}
