package com.simibubi.create.content.fluids.transfer;

import java.util.ArrayList;
import java.util.Comparator;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.function.BiConsumer;
import java.util.function.Predicate;

import org.jetbrains.annotations.Nullable;

import com.simibubi.create.AllPackets;
import com.simibubi.create.AllTags.AllFluidTags;
import com.simibubi.create.foundation.blockEntity.SmartBlockEntity;
import com.simibubi.create.foundation.blockEntity.behaviour.BlockEntityBehaviour;
import com.simibubi.create.foundation.fluid.FluidHelper;
import com.simibubi.create.foundation.mixin.fabric.SortedArraySetAccessor;
import com.simibubi.create.foundation.utility.Iterate;
import com.simibubi.create.foundation.utility.NBTHelper;
import com.simibubi.create.foundation.utility.VecHelper;
import com.simibubi.create.infrastructure.config.AllConfigs;

import io.github.fabricators_of_create.porting_lib.fluids.FluidStack;
import net.fabricmc.fabric.api.transfer.v1.fluid.FluidVariant;
import net.fabricmc.fabric.api.transfer.v1.fluid.FluidVariantAttributes;
import net.fabricmc.fabric.api.transfer.v1.transaction.TransactionContext;
import net.fabricmc.fabric.api.transfer.v1.transaction.base.SnapshotParticipant;
import net.minecraft.class_1937;
import net.minecraft.class_2338;
import net.minecraft.class_2350;
import net.minecraft.class_2350.class_2351;
import net.minecraft.class_243;
import net.minecraft.class_2487;
import net.minecraft.class_2512;
import net.minecraft.class_3218;
import net.minecraft.class_3341;
import net.minecraft.class_3414;
import net.minecraft.class_3419;
import net.minecraft.class_3610;
import net.minecraft.class_3611;
import net.minecraft.class_4706;
import org.jetbrains.annotations.Nullable;

public abstract class FluidManipulationBehaviour extends BlockEntityBehaviour {

	public static record BlockPosEntry(class_2338 pos, int distance) {
	};

	public static class ChunkNotLoadedException extends Exception {
		private static final long serialVersionUID = 1L;
	}

	class_3341 affectedArea;
	class_2338 rootPos;
	boolean infinite;
	protected boolean counterpartActed;

	// Search
	static final int searchedPerTick = 1024;
	static final int validationTimerMin = 160;
	List<BlockPosEntry> frontier;
	Set<class_2338> visited;

	int revalidateIn;

	public FluidManipulationBehaviour(SmartBlockEntity be) {
		super(be);
		setValidationTimer();
		infinite = false;
		visited = new HashSet<>();
		frontier = new ArrayList<>();
	}

	public boolean isInfinite() {
		return infinite;
	}

	public void counterpartActed(TransactionContext ctx) {
		snapshotParticipant().updateSnapshots(ctx);
		counterpartActed = true;
	}

	protected abstract SnapshotParticipant<?> snapshotParticipant();

	protected int validationTimer() {
		int maxBlocks = maxBlocks();
		// Allow enough time for the server's infinite block threshold to be reached
		return maxBlocks < 0 ? validationTimerMin : Math.max(validationTimerMin, maxBlocks / searchedPerTick + 1);
	}

	protected int setValidationTimer() {
		return revalidateIn = validationTimer();
	}

	protected int setLongValidationTimer() {
		return revalidateIn = validationTimer() * 2;
	}

	protected int maxRange() {
		return AllConfigs.server().fluids.hosePulleyRange.get();
	}

	protected int maxBlocks() {
		return AllConfigs.server().fluids.hosePulleyBlockThreshold.get();
	}

	protected boolean fillInfinite() {
		return AllConfigs.server().fluids.fillInfinite.get();
	}

	public void reset(@Nullable TransactionContext ctx) {
		if (affectedArea != null)
			scheduleUpdatesInAffectedArea();
		affectedArea = null;
		setValidationTimer();
		frontier.clear();
		visited.clear();
		infinite = false;
	}

	@Override
	public void destroy() {
		reset(null);
		super.destroy();
	}

	protected void scheduleUpdatesInAffectedArea() {
		class_1937 world = getWorld();
		class_2338
			.method_20437(
				new class_2338(affectedArea.method_35415() - 1, affectedArea.method_35416() - 1, affectedArea.method_35417() - 1),
				new class_2338(affectedArea.method_35418() + 1, affectedArea.method_35419() + 1, affectedArea.method_35420() + 1))
			.forEach(pos -> {
				class_3610 nextFluidState = world.method_8316(pos);
				if (nextFluidState.method_15769())
					return;
				world.method_39281(pos, nextFluidState.method_15772(), world.method_8409()
					.method_43048(5));
			});
	}

	protected int comparePositions(BlockPosEntry e1, BlockPosEntry e2) {
		class_243 centerOfRoot = VecHelper.getCenterOf(rootPos);
		class_2338 pos2 = e2.pos;
		class_2338 pos1 = e1.pos;
		if (pos1.method_10264() != pos2.method_10264())
			return Integer.compare(pos2.method_10264(), pos1.method_10264());
		int compareDistance = Integer.compare(e2.distance, e1.distance);
		if (compareDistance != 0)
			return compareDistance;
		int distanceCompared = Double.compare(VecHelper.getCenterOf(pos2)
						.method_1025(centerOfRoot),
				VecHelper.getCenterOf(pos1)
						.method_1025(centerOfRoot));
		// fabric: since we're using a set for the queue, we need to only have them equal if they're really equal.
		if (distanceCompared != 0)
			return distanceCompared;
		// equidistant, go by X and Z
		int xCompared = Integer.compare(pos2.method_10263(), pos1.method_10263());
		if (xCompared != 0)
			return xCompared;
		return Integer.compare(pos2.method_10260(), pos1.method_10260());
	}

	protected class_3611 search(class_3611 fluid, List<BlockPosEntry> frontier, Set<class_2338> visited,
		BiConsumer<class_2338, Integer> add, boolean searchDownward) throws ChunkNotLoadedException {
		class_1937 world = getWorld();
		int maxBlocks = maxBlocks();
		int maxRange = canDrainInfinitely(fluid) ? maxRange() : maxRange() / 2;
		int maxRangeSq = maxRange * maxRange;
		int i;

		for (i = 0; i < searchedPerTick && !frontier.isEmpty()
			&& (visited.size() <= maxBlocks || !canDrainInfinitely(fluid)); i++) {
			BlockPosEntry entry = frontier.remove(0);
			class_2338 currentPos = entry.pos;
			if (visited.contains(currentPos))
				continue;
			visited.add(currentPos);

			if (!world.method_8477(currentPos))
				throw new ChunkNotLoadedException();

			class_3610 fluidState = world.method_8316(currentPos);
			if (fluidState.method_15769())
				continue;

			class_3611 currentFluid = FluidHelper.convertToStill(fluidState.method_15772());
			if (fluid == null)
				fluid = currentFluid;
			if (!currentFluid.method_15780(fluid))
				continue;

			add.accept(currentPos, entry.distance);

			for (class_2350 side : Iterate.directions) {
				if (!searchDownward && side == class_2350.field_11033)
					continue;

				class_2338 offsetPos = currentPos.method_10093(side);
				if (!world.method_8477(offsetPos))
					throw new ChunkNotLoadedException();
				if (visited.contains(offsetPos))
					continue;
				if (offsetPos.method_10262(rootPos) > maxRangeSq)
					continue;

				class_3610 nextFluidState = world.method_8316(offsetPos);
				if (nextFluidState.method_15769())
					continue;
				class_3611 nextFluid = nextFluidState.method_15772();
				if (nextFluid == FluidHelper.convertToFlowing(nextFluid) && side == class_2350.field_11036
					&& !VecHelper.onSameAxis(rootPos, offsetPos, class_2351.field_11052))
					continue;

				frontier.add(new BlockPosEntry(offsetPos, entry.distance + 1));
			}
		}

		return fluid;
	}

	protected void playEffect(class_1937 world, class_2338 pos, class_3611 fluid, boolean fillSound) {
		if (fluid == null)
			return;

		class_2338 splooshPos = pos == null ? blockEntity.method_11016() : pos;
		FluidStack stack = new FluidStack(fluid, 1);

		FluidVariant variant = FluidVariant.of(fluid);
		class_3414 soundevent = fillSound
				? FluidVariantAttributes.getFillSound(variant)
				: FluidVariantAttributes.getEmptySound(variant);

		world.method_8396(null, splooshPos, soundevent, class_3419.field_15245, 0.3F, 1.0F);
		if (world instanceof class_3218)
			AllPackets.sendToNear(world, splooshPos, 10, new FluidSplashPacket(splooshPos, stack));
	}

	protected boolean canDrainInfinitely(class_3611 fluid) {
		if (fluid == null)
			return false;
		return maxBlocks() != -1 && AllConfigs.server().fluids.bottomlessFluidMode.get()
			.test(fluid);
	}

	@Override
	public void write(class_2487 nbt, boolean clientPacket) {
		if (infinite)
			NBTHelper.putMarker(nbt, "Infinite");
		if (rootPos != null)
			nbt.method_10566("LastPos", class_2512.method_10692(rootPos));
		if (affectedArea != null) {
			nbt.method_10566("AffectedAreaFrom",
				class_2512.method_10692(new class_2338(affectedArea.method_35415(), affectedArea.method_35416(), affectedArea.method_35417())));
			nbt.method_10566("AffectedAreaTo",
				class_2512.method_10692(new class_2338(affectedArea.method_35418(), affectedArea.method_35419(), affectedArea.method_35420())));
		}
		super.write(nbt, clientPacket);
	}

	@Override
	public void read(class_2487 nbt, boolean clientPacket) {
		infinite = nbt.method_10545("Infinite");
		if (nbt.method_10545("LastPos"))
			rootPos = class_2512.method_10691(nbt.method_10562("LastPos"));
		if (nbt.method_10545("AffectedAreaFrom") && nbt.method_10545("AffectedAreaTo"))
			affectedArea = class_3341.method_34390(class_2512.method_10691(nbt.method_10562("AffectedAreaFrom")),
				class_2512.method_10691(nbt.method_10562("AffectedAreaTo")));
		super.read(nbt, clientPacket);
	}

	public enum BottomlessFluidMode implements Predicate<class_3611> {
		ALLOW_ALL(fluid -> true),
		DENY_ALL(fluid -> false),
		ALLOW_BY_TAG(fluid -> AllFluidTags.BOTTOMLESS_ALLOW.matches(fluid)),
		DENY_BY_TAG(fluid -> !AllFluidTags.BOTTOMLESS_DENY.matches(fluid));

		private final Predicate<class_3611> predicate;

		BottomlessFluidMode(Predicate<class_3611> predicate) {
			this.predicate = predicate;
		}

		@Override
		public boolean test(class_3611 fluid) {
			return predicate.test(fluid);
		}
	}


	/**
	 * Quickly copy the given set.
	 * This is a shallow copy, so entries must be immutable.
	 */
	public static <T> class_4706<T> copySet(class_4706<T> set) {
		int size = set.size();
		SortedArraySetAccessor<T> access = (SortedArraySetAccessor<T>) set;
		Comparator<T> comparator = access.create$getComparator();
		T[] contents = access.create$getContents();
		T[] copiedContents = (T[]) new Object[size];
		System.arraycopy(contents, 0, copiedContents, 0, size);
		class_4706<T> copy = class_4706.method_34960(comparator, size);
		SortedArraySetAccessor<T> copyAccess = ((SortedArraySetAccessor<T>) copy);
		copyAccess.create$setContents(copiedContents);
		copyAccess.create$setSize(size);
		return copy;
	}

	/**
	 * Remove the first entry from the given set.
	 * identical to {@code set.remove(set.first())}
	 */
	public static <T> void dequeue(class_4706<T> set) {
		((SortedArraySetAccessor<T>) set).create$callRemoveInternal(0);
	}

}
