package com.simibubi.create.content.fluids.transfer;

import com.simibubi.create.AllFluids;
import com.simibubi.create.AllItems;
import com.simibubi.create.content.fluids.potion.PotionFluidHandler;
import com.simibubi.create.foundation.fluid.FluidHelper;
import io.github.fabricators_of_create.porting_lib.transfer.TransferUtil;
import io.github.fabricators_of_create.porting_lib.fluids.FluidStack;

import net.fabricmc.fabric.api.transfer.v1.context.ContainerItemContext;
import net.fabricmc.fabric.api.transfer.v1.fluid.FluidConstants;
import net.fabricmc.fabric.api.transfer.v1.fluid.FluidStorage;
import net.fabricmc.fabric.api.transfer.v1.fluid.FluidVariant;
import net.fabricmc.fabric.api.transfer.v1.storage.Storage;
import net.fabricmc.fabric.api.transfer.v1.transaction.Transaction;
import net.minecraft.class_1799;
import net.minecraft.class_1802;
import net.minecraft.class_1844;
import net.minecraft.class_1847;
import net.minecraft.class_1937;
import net.minecraft.class_3611;
import net.minecraft.class_3612;

public class GenericItemFilling {

	/**
	 * Checks if an ItemStack's IFluidHandlerItem is valid. Ideally, this check would
	 * not be necessary. Unfortunately, some mods that copy the functionality of the
	 * MilkBucketItem copy the FluidBucketWrapper capability that is patched in by
	 * Forge without looking into what it actually does. In all cases this is
	 * incorrect because having a non-bucket item turn into a bucket item does not
	 * make sense.
	 *
	 * <p>This check is only necessary for filling since a FluidBucketWrapper will be
	 * empty if it is initialized with a non-bucket item.
	 *
	 * @param stack The ItemStack.
	 * @param fluidHandler The IFluidHandlerItem instance retrieved from the ItemStack.
	 * @return If the IFluidHandlerItem is valid for the passed ItemStack.
	 */
	public static boolean isFluidHandlerValid(class_1799 stack, Storage<FluidVariant> fluidHandler) {
		// Not instanceof in case a correct subclass is made
//		if (fluidHandler.getClass() == FluidBucketWrapper.class) {
//			Item item = stack.getItem();
//			// Forge does not patch the FluidBucketWrapper onto subclasses of BucketItem
//			if (item.getClass() != BucketItem.class && !(item instanceof MilkBucketItem)) {
//				return false;
//			}
//		}
		return true;
	}

	public static boolean canItemBeFilled(class_1937 world, class_1799 stack) {
		if (stack.method_7909() == class_1802.field_8469)
			return true;
		if (stack.method_7909() == class_1802.field_8103)
			return false;

		Storage<FluidVariant> tank = FluidStorage.ITEM.find(stack, ContainerItemContext.withConstant(stack));
		if (tank == null)
			return false;
		if (!isFluidHandlerValid(stack, tank))
			return false;
		return tank.supportsInsertion();
	}

	public static long getRequiredAmountForItem(class_1937 world, class_1799 stack, FluidStack availableFluid) {
		if (stack.method_7909() == class_1802.field_8469 && canFillGlassBottleInternally(availableFluid))
			return PotionFluidHandler.getRequiredAmountForFilledBottle(stack, availableFluid);
		if (stack.method_7909() == class_1802.field_8550 && canFillBucketInternally(availableFluid))
			return FluidConstants.BUCKET;

		Storage<FluidVariant> tank = FluidStorage.ITEM.find(stack, ContainerItemContext.withConstant(stack));
		if (tank == null)
			return -1;

//		if (tank instanceof FluidBucketWrapper) {
//			Item filledBucket = availableFluid.getFluid()
//				.getBucket();
//			if (filledBucket == null || filledBucket == Items.AIR)
//				return -1;
//			if (!((FluidBucketWrapper) tank).getFluid()
//				.isEmpty())
//				return -1;
//			return FluidConstants.BUCKET;
//		}

		try (Transaction t = TransferUtil.getTransaction()) {
			long filled = tank.insert(availableFluid.getType(), availableFluid.getAmount(), t);
			return filled == 0 ? -1 : filled;
		}
	}

	private static boolean canFillGlassBottleInternally(FluidStack availableFluid) {
		class_3611 fluid = availableFluid.getFluid();
		if (fluid.method_15780(class_3612.field_15910))
			return true;
		if (fluid.method_15780(AllFluids.POTION.get()))
			return true;
		if (fluid.method_15780(AllFluids.TEA.get()))
			return true;
		return false;
	}

	private static boolean canFillBucketInternally(FluidStack availableFluid) {
		return true; // fabric: this is false on forge for reasons I'm not entirely sure of. we need it true here to catch buckets.
	}

	public static class_1799 fillItem(class_1937 world, long requiredAmount, class_1799 stack, FluidStack availableFluid) {
		FluidStack toFill = availableFluid.copy();
		toFill.setAmount(requiredAmount);
		availableFluid.shrink(requiredAmount);

		if (stack.method_7909() == class_1802.field_8469 && canFillGlassBottleInternally(toFill)) {
			class_1799 fillBottle = class_1799.field_8037;
			class_3611 fluid = toFill.getFluid();
			if (FluidHelper.isWater(fluid))
				fillBottle = class_1844.method_8061(new class_1799(class_1802.field_8574), class_1847.field_8991);
			else if (fluid.method_15780(AllFluids.TEA.get()))
				fillBottle = AllItems.BUILDERS_TEA.asStack();
			else
				fillBottle = PotionFluidHandler.fillBottle(stack, toFill);
			stack.method_7934(1);
			return fillBottle;
		}

		class_1799 split = stack.method_7972();
		split.method_7939(1);
		ContainerItemContext ctx = ContainerItemContext.withConstant(split);
		Storage<FluidVariant> tank = FluidStorage.ITEM.find(split, ctx);
		if (tank == null)
			return class_1799.field_8037;
		try (Transaction t = TransferUtil.getTransaction()) {
			tank.insert(toFill.getType(), toFill.getAmount(), t);
			t.commit();

			class_1799 container = ctx.getItemVariant().toStack((int) ctx.getAmount());
			stack.method_7934(1);
			return container;
		}
	}

}
