package com.simibubi.create.content.kinetics;

import java.util.HashMap;
import java.util.Map;

import javax.annotation.Nullable;
import net.minecraft.class_2248;
import net.minecraft.class_2960;
import com.simibubi.create.foundation.utility.Couple;
import com.simibubi.create.foundation.utility.RegisteredObjects;

public class BlockStressValues {

	private static final Map<String, IStressValueProvider> PROVIDERS = new HashMap<>();

	public static void registerProvider(String namespace, IStressValueProvider provider) {
		PROVIDERS.put(namespace, provider);
	}

	@Nullable
	public static IStressValueProvider getProvider(String namespace) {
		return PROVIDERS.get(namespace);
	}

	@Nullable
	public static IStressValueProvider getProvider(class_2248 block) {
		return getProvider(RegisteredObjects.getKeyOrThrow(block)
			.method_12836());
	}

	public static double getImpact(class_2248 block) {
		class_2960 blockId = RegisteredObjects.getKeyOrThrow(block);
		IStressValueProvider provider = getProvider(blockId.method_12836());
		if (provider != null) {
			return provider.getImpact(block);
		}
		Double defaultImpact = BlockStressDefaults.DEFAULT_IMPACTS.get(blockId);
		if (defaultImpact != null) {
			return defaultImpact;
		}
		return 0;
	}

	public static double getCapacity(class_2248 block) {
		class_2960 blockId = RegisteredObjects.getKeyOrThrow(block);
		IStressValueProvider provider = getProvider(blockId.method_12836());
		if (provider != null) {
			return provider.getCapacity(block);
		}
		Double defaultCapacity = BlockStressDefaults.DEFAULT_CAPACITIES.get(blockId);
		if (defaultCapacity != null) {
			return defaultCapacity;
		}
		return 0;
	}

	public static boolean hasImpact(class_2248 block) {
		class_2960 blockId = RegisteredObjects.getKeyOrThrow(block);
		IStressValueProvider provider = getProvider(blockId.method_12836());
		if (provider != null) {
			return provider.hasImpact(block);
		}
		return BlockStressDefaults.DEFAULT_IMPACTS.containsKey(blockId);
	}

	public static boolean hasCapacity(class_2248 block) {
		class_2960 blockId = RegisteredObjects.getKeyOrThrow(block);
		IStressValueProvider provider = getProvider(blockId.method_12836());
		if (provider != null) {
			return provider.hasCapacity(block);
		}
		return BlockStressDefaults.DEFAULT_CAPACITIES.containsKey(blockId);
	}

	@Nullable
	public static Couple<Integer> getGeneratedRPM(class_2248 block) {
		class_2960 blockId = RegisteredObjects.getKeyOrThrow(block);
		IStressValueProvider provider = getProvider(blockId.method_12836());
		if (provider != null) {
			return provider.getGeneratedRPM(block);
		}
		return null;
	}

	public interface IStressValueProvider {
		/**
		 * Gets the stress impact of a block.
		 *
		 * @param block The block.
		 * @return the stress impact value of the block, or 0 if it does not have one.
		 */
		double getImpact(class_2248 block);

		/**
		 * Gets the stress capacity of a block.
		 *
		 * @param block The block.
		 * @return the stress capacity value of the block, or 0 if it does not have one.
		 */
		double getCapacity(class_2248 block);

		boolean hasImpact(class_2248 block);

		boolean hasCapacity(class_2248 block);

		/**
		 *
		 * @param block
		 * @return min, max generated RPM; null if block does not have a stress capacity
		 */
		@Nullable
		Couple<Integer> getGeneratedRPM(class_2248 block);
	}

}
