package com.simibubi.create.content.kinetics.base;

import com.simibubi.create.content.equipment.wrench.IWrenchable;
import com.simibubi.create.foundation.item.TooltipHelper;
import com.simibubi.create.foundation.utility.Lang;
import com.simibubi.create.foundation.utility.LangBuilder;
import com.simibubi.create.infrastructure.config.AllConfigs;
import net.minecraft.class_124;
import net.minecraft.class_2338;
import net.minecraft.class_2350;
import net.minecraft.class_2350.class_2351;
import net.minecraft.class_2680;
import net.minecraft.class_4538;

public interface IRotate extends IWrenchable {

	enum SpeedLevel {
		NONE(class_124.field_1063, 0x000000, 0),
		SLOW(class_124.field_1060, 0x22FF22, 10),
		MEDIUM(class_124.field_1075, 0x0084FF, 20),
		FAST(class_124.field_1076, 0xFF55FF, 30);

		private final class_124 textColor;
		private final int color;
		private final int particleSpeed;

		SpeedLevel(class_124 textColor, int color, int particleSpeed) {
			this.textColor = textColor;
			this.color = color;
			this.particleSpeed = particleSpeed;
		}

		public class_124 getTextColor() {
			return textColor;
		}

		public int getColor() {
			return color;
		}

		public int getParticleSpeed() {
			return particleSpeed;
		}

		public float getSpeedValue() {
			switch (this) {
			case FAST:
				return AllConfigs.server().kinetics.fastSpeed.get()
					.floatValue();
			case MEDIUM:
				return AllConfigs.server().kinetics.mediumSpeed.get()
					.floatValue();
			case SLOW:
				return 1;
			case NONE:
			default:
				return 0;
			}
		}

		public static SpeedLevel of(float speed) {
			speed = Math.abs(speed);

			if (speed >= AllConfigs.server().kinetics.fastSpeed.get())
				return FAST;
			if (speed >= AllConfigs.server().kinetics.mediumSpeed.get())
				return MEDIUM;
			if (speed >= 1)
				return SLOW;
			return NONE;
		}

		public static LangBuilder getFormattedSpeedText(float speed, boolean overstressed) {
			SpeedLevel speedLevel = of(speed);
			LangBuilder builder = Lang.text(TooltipHelper.makeProgressBar(3, speedLevel.ordinal()));

			builder.translate("tooltip.speedRequirement." + Lang.asId(speedLevel.name()))
				.space()
				.text("(")
				.add(Lang.number(Math.abs(speed)))
				.space()
				.translate("generic.unit.rpm")
				.text(")")
				.space();

			if (overstressed)
				builder.style(class_124.field_1063)
					.style(class_124.field_1055);
			else
				builder.style(speedLevel.getTextColor());

			return builder;
		}

	}

	enum StressImpact {
		LOW(class_124.field_1054, class_124.field_1060),
		MEDIUM(class_124.field_1065, class_124.field_1054),
		HIGH(class_124.field_1061, class_124.field_1065),
		OVERSTRESSED(class_124.field_1061, class_124.field_1061);

		private final class_124 absoluteColor;
		private final class_124 relativeColor;

		StressImpact(class_124 absoluteColor, class_124 relativeColor) {
			this.absoluteColor = absoluteColor;
			this.relativeColor = relativeColor;
		}

		public class_124 getAbsoluteColor() {
			return absoluteColor;
		}

		public class_124 getRelativeColor() {
			return relativeColor;
		}

		public static StressImpact of(double stressPercent) {
			if (stressPercent > 1)
				return StressImpact.OVERSTRESSED;
			if (stressPercent > .75d)
				return StressImpact.HIGH;
			if (stressPercent > .5d)
				return StressImpact.MEDIUM;
			return StressImpact.LOW;
		}

		public static boolean isEnabled() {
			return !AllConfigs.server().kinetics.disableStress.get();
		}

		public static LangBuilder getFormattedStressText(double stressPercent) {
			StressImpact stressLevel = of(stressPercent);
			return Lang.text(TooltipHelper.makeProgressBar(3, Math.min(stressLevel.ordinal() + 1, 3)))
				.translate("tooltip.stressImpact." + Lang.asId(stressLevel.name()))
				.text(String.format(" (%s%%) ", (int) (stressPercent * 100)))
				.style(stressLevel.getRelativeColor());
		}
	}

	public boolean hasShaftTowards(class_4538 world, class_2338 pos, class_2680 state, class_2350 face);

	public class_2351 getRotationAxis(class_2680 state);

	public default SpeedLevel getMinimumRequiredSpeedLevel() {
		return SpeedLevel.NONE;
	}

	public default boolean hideStressImpact() {
		return false;
	}

	public default boolean showCapacityWithAnnotation() {
		return false;
	}

}
