package com.simibubi.create.content.kinetics.base;

import static net.minecraft.class_124.field_1065;
import static net.minecraft.class_124.field_1080;

import java.util.List;

import javax.annotation.Nullable;

import com.jozufozu.flywheel.backend.instancing.InstancedRenderDispatcher;
import com.simibubi.create.Create;
import com.simibubi.create.content.equipment.goggles.IHaveGoggleInformation;
import com.simibubi.create.content.equipment.goggles.IHaveHoveringInformation;
import com.simibubi.create.content.kinetics.BlockStressValues;
import com.simibubi.create.content.kinetics.KineticNetwork;
import com.simibubi.create.content.kinetics.RotationPropagator;
import com.simibubi.create.content.kinetics.base.IRotate.SpeedLevel;
import com.simibubi.create.content.kinetics.base.IRotate.StressImpact;
import com.simibubi.create.content.kinetics.gearbox.GearboxBlock;
import com.simibubi.create.content.kinetics.simpleRelays.ICogWheel;
import com.simibubi.create.content.kinetics.transmission.sequencer.SequencedGearshiftBlockEntity.SequenceContext;
import com.simibubi.create.foundation.blockEntity.SmartBlockEntity;
import com.simibubi.create.foundation.blockEntity.behaviour.BlockEntityBehaviour;
import com.simibubi.create.foundation.item.TooltipHelper;
import com.simibubi.create.foundation.item.TooltipHelper.Palette;
import com.simibubi.create.foundation.sound.SoundScapes;
import com.simibubi.create.foundation.sound.SoundScapes.AmbienceGroup;
import com.simibubi.create.foundation.utility.Lang;
import com.simibubi.create.infrastructure.config.AllConfigs;

import io.github.fabricators_of_create.porting_lib.block.CustomRenderBoundingBoxBlockEntity;
import io.github.fabricators_of_create.porting_lib.util.EnvExecutor;
import net.fabricmc.api.EnvType;
import net.fabricmc.api.Environment;
import net.minecraft.class_1074;
import net.minecraft.class_124;
import net.minecraft.class_1937;
import net.minecraft.class_2248;
import net.minecraft.class_2338;
import net.minecraft.class_2350;
import net.minecraft.class_2350.class_2351;
import net.minecraft.class_2350.class_2352;
import net.minecraft.class_2487;
import net.minecraft.class_2512;
import net.minecraft.class_2561;
import net.minecraft.class_2586;
import net.minecraft.class_2591;
import net.minecraft.class_2680;
import net.minecraft.class_3532;
import net.minecraft.class_5250;

public class KineticBlockEntity extends SmartBlockEntity implements IHaveGoggleInformation, IHaveHoveringInformation, CustomRenderBoundingBoxBlockEntity {

	public @Nullable Long network;
	public @Nullable class_2338 source;
	public boolean networkDirty;
	public boolean updateSpeed;
	public int preventSpeedUpdate;

	protected KineticEffectHandler effects;
	protected float speed;
	protected float capacity;
	protected float stress;
	protected boolean overStressed;
	protected boolean wasMoved;

	private int flickerTally;
	private int networkSize;
	private int validationCountdown;
	protected float lastStressApplied;
	protected float lastCapacityProvided;

	public SequenceContext sequenceContext;

	public KineticBlockEntity(class_2591<?> typeIn, class_2338 pos, class_2680 state) {
		super(typeIn, pos, state);
		effects = new KineticEffectHandler(this);
		updateSpeed = true;
	}

	@Override
	public void initialize() {
		if (hasNetwork() && !field_11863.field_9236) {
			KineticNetwork network = getOrCreateNetwork();
			if (!network.initialized)
				network.initFromTE(capacity, stress, networkSize);
			network.addSilently(this, lastCapacityProvided, lastStressApplied);
		}

		super.initialize();
	}

	@Override
	public void tick() {
		if (!field_11863.field_9236 && needsSpeedUpdate())
			attachKinetics();

		super.tick();
		effects.tick();

		preventSpeedUpdate = 0;

		if (field_11863.field_9236) {
			EnvExecutor.runWhenOn(EnvType.CLIENT, () -> () -> this.tickAudio());
			return;
		}

		if (validationCountdown-- <= 0) {
			validationCountdown = AllConfigs.server().kinetics.kineticValidationFrequency.get();
			validateKinetics();
		}

		if (getFlickerScore() > 0)
			flickerTally = getFlickerScore() - 1;

		if (networkDirty) {
			if (hasNetwork())
				getOrCreateNetwork().updateNetwork();
			networkDirty = false;
		}
	}

	private void validateKinetics() {
		if (hasSource()) {
			if (!hasNetwork()) {
				removeSource();
				return;
			}

			if (!field_11863.method_8477(source))
				return;

			class_2586 blockEntity = field_11863.method_8321(source);
			KineticBlockEntity sourceBE =
				blockEntity instanceof KineticBlockEntity ? (KineticBlockEntity) blockEntity : null;
			if (sourceBE == null || sourceBE.speed == 0) {
				removeSource();
				detachKinetics();
				return;
			}

			return;
		}

		if (speed != 0) {
			if (getGeneratedSpeed() == 0)
				speed = 0;
		}
	}

	public void updateFromNetwork(float maxStress, float currentStress, int networkSize) {
		networkDirty = false;
		this.capacity = maxStress;
		this.stress = currentStress;
		this.networkSize = networkSize;
		boolean overStressed = maxStress < currentStress && StressImpact.isEnabled();
		method_5431();

		if (overStressed != this.overStressed) {
			float prevSpeed = getSpeed();
			this.overStressed = overStressed;
			onSpeedChanged(prevSpeed);
			sendData();
		}
	}

	protected class_2248 getStressConfigKey() {
		return method_11010().method_26204();
	}

	public float calculateStressApplied() {
		float impact = (float) BlockStressValues.getImpact(getStressConfigKey());
		this.lastStressApplied = impact;
		return impact;
	}

	public float calculateAddedStressCapacity() {
		float capacity = (float) BlockStressValues.getCapacity(getStressConfigKey());
		this.lastCapacityProvided = capacity;
		return capacity;
	}

	public void onSpeedChanged(float previousSpeed) {
		boolean fromOrToZero = (previousSpeed == 0) != (getSpeed() == 0);
		boolean directionSwap = !fromOrToZero && Math.signum(previousSpeed) != Math.signum(getSpeed());
		if (fromOrToZero || directionSwap)
			flickerTally = getFlickerScore() + 5;
		method_5431();
	}

	@Override
	public void remove() {
		if (!field_11863.field_9236) {
			if (hasNetwork())
				getOrCreateNetwork().remove(this);
			detachKinetics();
		}
		super.remove();
	}

	@Override
	protected void write(class_2487 compound, boolean clientPacket) {
		compound.method_10548("Speed", speed);
		if (sequenceContext != null && (!clientPacket || syncSequenceContext()))
			compound.method_10566("Sequence", sequenceContext.serializeNBT());

		if (needsSpeedUpdate())
			compound.method_10556("NeedsSpeedUpdate", true);

		if (hasSource())
			compound.method_10566("Source", class_2512.method_10692(source));

		if (hasNetwork()) {
			class_2487 networkTag = new class_2487();
			networkTag.method_10544("Id", this.network);
			networkTag.method_10548("Stress", stress);
			networkTag.method_10548("Capacity", capacity);
			networkTag.method_10569("Size", networkSize);

			if (lastStressApplied != 0)
				networkTag.method_10548("AddedStress", lastStressApplied);
			if (lastCapacityProvided != 0)
				networkTag.method_10548("AddedCapacity", lastCapacityProvided);

			compound.method_10566("Network", networkTag);
		}

		super.write(compound, clientPacket);
	}

	public boolean needsSpeedUpdate() {
		return updateSpeed;
	}

	@Override
	protected void read(class_2487 compound, boolean clientPacket) {
		boolean overStressedBefore = overStressed;
		clearKineticInformation();

		// DO NOT READ kinetic information when placed after movement
		if (wasMoved) {
			super.read(compound, clientPacket);
			return;
		}

		speed = compound.method_10583("Speed");
		sequenceContext = SequenceContext.fromNBT(compound.method_10562("Sequence"));

		if (compound.method_10545("Source"))
			source = class_2512.method_10691(compound.method_10562("Source"));

		if (compound.method_10545("Network")) {
			class_2487 networkTag = compound.method_10562("Network");
			network = networkTag.method_10537("Id");
			stress = networkTag.method_10583("Stress");
			capacity = networkTag.method_10583("Capacity");
			networkSize = networkTag.method_10550("Size");
			lastStressApplied = networkTag.method_10583("AddedStress");
			lastCapacityProvided = networkTag.method_10583("AddedCapacity");
			overStressed = capacity < stress && StressImpact.isEnabled();
		}

		super.read(compound, clientPacket);

		if (clientPacket && overStressedBefore != overStressed && speed != 0)
			effects.triggerOverStressedEffect();

		if (clientPacket)
			EnvExecutor.runWhenOn(EnvType.CLIENT, () -> () -> InstancedRenderDispatcher.enqueueUpdate(this));
	}

	public float getGeneratedSpeed() {
		return 0;
	}

	public boolean isSource() {
		return getGeneratedSpeed() != 0;
	}

	public float getSpeed() {
		if (overStressed)
			return 0;
		return getTheoreticalSpeed();
	}

	public float getTheoreticalSpeed() {
		return speed;
	}

	public void setSpeed(float speed) {
		this.speed = speed;
	}

	public boolean hasSource() {
		return source != null;
	}

	public void setSource(class_2338 source) {
		this.source = source;
		if (field_11863 == null || field_11863.field_9236)
			return;

		class_2586 blockEntity = field_11863.method_8321(source);
		if (!(blockEntity instanceof KineticBlockEntity sourceBE)) {
			removeSource();
			return;
		}

		setNetwork(sourceBE.network);
		copySequenceContextFrom(sourceBE);
	}

	protected void copySequenceContextFrom(KineticBlockEntity sourceBE) {
		sequenceContext = sourceBE.sequenceContext;
	}

	public void removeSource() {
		float prevSpeed = getSpeed();

		speed = 0;
		source = null;
		setNetwork(null);
		sequenceContext = null;

		onSpeedChanged(prevSpeed);
	}

	public void setNetwork(@Nullable Long networkIn) {
		if (network == networkIn)
			return;
		if (network != null)
			getOrCreateNetwork().remove(this);

		network = networkIn;
		method_5431();

		if (networkIn == null)
			return;

		network = networkIn;
		KineticNetwork network = getOrCreateNetwork();
		network.initialized = true;
		network.add(this);
	}

	public KineticNetwork getOrCreateNetwork() {
		return Create.TORQUE_PROPAGATOR.getOrCreateNetworkFor(this);
	}

	public boolean hasNetwork() {
		return network != null;
	}

	public void attachKinetics() {
		updateSpeed = false;
		RotationPropagator.handleAdded(field_11863, field_11867, this);
	}

	public void detachKinetics() {
		RotationPropagator.handleRemoved(field_11863, field_11867, this);
	}

	public boolean isSpeedRequirementFulfilled() {
		class_2680 state = method_11010();
		if (!(method_11010().method_26204() instanceof IRotate))
			return true;
		IRotate def = (IRotate) state.method_26204();
		SpeedLevel minimumRequiredSpeedLevel = def.getMinimumRequiredSpeedLevel();
		return Math.abs(getSpeed()) >= minimumRequiredSpeedLevel.getSpeedValue();
	}

	public static void switchToBlockState(class_1937 world, class_2338 pos, class_2680 state) {
		if (world.field_9236)
			return;

		class_2586 blockEntity = world.method_8321(pos);
		class_2680 currentState = world.method_8320(pos);
		boolean isKinetic = blockEntity instanceof KineticBlockEntity;

		if (currentState == state)
			return;
		if (blockEntity == null || !isKinetic) {
			world.method_8652(pos, state, 3);
			return;
		}

		KineticBlockEntity kineticBlockEntity = (KineticBlockEntity) blockEntity;
		if (state.method_26204() instanceof KineticBlock
			&& !((KineticBlock) state.method_26204()).areStatesKineticallyEquivalent(currentState, state)) {
			if (kineticBlockEntity.hasNetwork())
				kineticBlockEntity.getOrCreateNetwork()
					.remove(kineticBlockEntity);
			kineticBlockEntity.detachKinetics();
			kineticBlockEntity.removeSource();
		}

		world.method_8652(pos, state, 3);
	}

	@Override
	public void addBehaviours(List<BlockEntityBehaviour> behaviours) {}

	@Override
	public boolean addToTooltip(List<class_2561> tooltip, boolean isPlayerSneaking) {
		boolean notFastEnough = !isSpeedRequirementFulfilled() && getSpeed() != 0;

		if (overStressed && AllConfigs.client().enableOverstressedTooltip.get()) {
			Lang.translate("gui.stressometer.overstressed")
				.style(field_1065)
				.forGoggles(tooltip);
			class_2561 hint = Lang.translateDirect("gui.contraptions.network_overstressed");
			List<class_2561> cutString = TooltipHelper.cutTextComponent(hint, Palette.GRAY_AND_WHITE);
			for (int i = 0; i < cutString.size(); i++)
				Lang.builder()
					.add(cutString.get(i)
						.method_27661())
					.forGoggles(tooltip);
			return true;
		}

		if (notFastEnough) {
			Lang.translate("tooltip.speedRequirement")
				.style(field_1065)
				.forGoggles(tooltip);
			class_5250 hint =
				Lang.translateDirect("gui.contraptions.not_fast_enough", class_1074.method_4662(method_11010().method_26204()
					.method_9539()));
			List<class_2561> cutString = TooltipHelper.cutTextComponent(hint, Palette.GRAY_AND_WHITE);
			for (int i = 0; i < cutString.size(); i++)
				Lang.builder()
					.add(cutString.get(i)
						.method_27661())
					.forGoggles(tooltip);
			return true;
		}

		return false;
	}

	@Override
	public boolean addToGoggleTooltip(List<class_2561> tooltip, boolean isPlayerSneaking) {
		boolean added = false;

		if (!StressImpact.isEnabled())
			return added;
		float stressAtBase = calculateStressApplied();
		if (class_3532.method_15347(stressAtBase, 0))
			return added;

		Lang.translate("gui.goggles.kinetic_stats")
			.forGoggles(tooltip);

		addStressImpactStats(tooltip, stressAtBase);

		return true;

	}

	protected void addStressImpactStats(List<class_2561> tooltip, float stressAtBase) {
		Lang.translate("tooltip.stressImpact")
			.style(field_1080)
			.forGoggles(tooltip);

		float stressTotal = stressAtBase * Math.abs(getTheoreticalSpeed());

		Lang.number(stressTotal)
			.translate("generic.unit.stress")
			.style(class_124.field_1075)
			.space()
			.add(Lang.translate("gui.goggles.at_current_speed")
				.style(class_124.field_1063))
			.forGoggles(tooltip, 1);
	}

	public void clearKineticInformation() {
		speed = 0;
		source = null;
		network = null;
		overStressed = false;
		stress = 0;
		capacity = 0;
		lastStressApplied = 0;
		lastCapacityProvided = 0;
	}

	public void warnOfMovement() {
		wasMoved = true;
	}

	public int getFlickerScore() {
		return flickerTally;
	}

	public static float convertToDirection(float axisSpeed, class_2350 d) {
		return d.method_10171() == class_2352.field_11056 ? axisSpeed : -axisSpeed;
	}

	public static float convertToLinear(float speed) {
		return speed / 512f;
	}

	public static float convertToAngular(float speed) {
		return speed * 3 / 10f;
	}

	public boolean isOverStressed() {
		return overStressed;
	}

	// Custom Propagation

	/**
	 * Specify ratio of transferred rotation from this kinetic component to a
	 * specific other.
	 *
	 * @param target           other Kinetic BE to transfer to
	 * @param stateFrom        this BE's blockstate
	 * @param stateTo          other BE's blockstate
	 * @param diff             difference in position (to.pos - from.pos)
	 * @param connectedViaAxes whether these kinetic blocks are connected via mutual
	 *                         IRotate.hasShaftTowards()
	 * @param connectedViaCogs whether these kinetic blocks are connected via mutual
	 *                         IRotate.hasIntegratedCogwheel()
	 * @return factor of rotation speed from this BE to other. 0 if no rotation is
	 *         transferred, or the standard rules apply (integrated shafts/cogs)
	 */
	public float propagateRotationTo(KineticBlockEntity target, class_2680 stateFrom, class_2680 stateTo, class_2338 diff,
		boolean connectedViaAxes, boolean connectedViaCogs) {
		return 0;
	}

	/**
	 * Specify additional locations the rotation propagator should look for
	 * potentially connected components. Neighbour list contains offset positions in
	 * all 6 directions by default.
	 *
	 * @param block
	 * @param state
	 * @param neighbours
	 * @return
	 */
	public List<class_2338> addPropagationLocations(IRotate block, class_2680 state, List<class_2338> neighbours) {
		if (!canPropagateDiagonally(block, state))
			return neighbours;

		class_2351 axis = block.getRotationAxis(state);
		class_2338.method_20437(new class_2338(-1, -1, -1), new class_2338(1, 1, 1))
			.forEach(offset -> {
				if (axis.method_10173(offset.method_10263(), offset.method_10264(), offset.method_10260()) != 0)
					return;
				if (offset.method_10262(class_2338.field_10980) != 2)
					return;
				neighbours.add(field_11867.method_10081(offset));
			});
		return neighbours;
	}

	/**
	 * Specify whether this component can propagate speed to the other in any
	 * circumstance. Shaft and cogwheel connections are already handled by internal
	 * logic. Does not have to be specified on both ends, it is assumed that this
	 * relation is symmetrical.
	 *
	 * @param other
	 * @param state
	 * @param otherState
	 * @return true if this and the other component should check their propagation
	 *         factor and are not already connected via integrated cogs or shafts
	 */
	public boolean isCustomConnection(KineticBlockEntity other, class_2680 state, class_2680 otherState) {
		return false;
	}

	protected boolean canPropagateDiagonally(IRotate block, class_2680 state) {
		return ICogWheel.isSmallCog(state);
	}

//	@Override
	public void requestModelDataUpdate() {
//		super.requestModelDataUpdate();
		if (!this.field_11865)
			EnvExecutor.runWhenOn(EnvType.CLIENT, () -> () -> InstancedRenderDispatcher.enqueueUpdate(this));
	}

	@Environment(EnvType.CLIENT)
	public void tickAudio() {
		float componentSpeed = Math.abs(getSpeed());
		if (componentSpeed == 0)
			return;
		float pitch = class_3532.method_15363((componentSpeed / 256f) + .45f, .85f, 1f);

		if (isNoisy())
			SoundScapes.play(AmbienceGroup.KINETIC, field_11867, pitch);

		class_2248 block = method_11010().method_26204();
		if (ICogWheel.isSmallCog(block) || ICogWheel.isLargeCog(block) || block instanceof GearboxBlock)
			SoundScapes.play(AmbienceGroup.COG, field_11867, pitch);
	}

	protected boolean isNoisy() {
		return true;
	}

	public int getRotationAngleOffset(class_2351 axis) {
		return 0;
	}

	protected boolean syncSequenceContext() {
		return false;
	}

}
