package com.simibubi.create.content.kinetics.base;

import com.jozufozu.flywheel.backend.Backend;
import com.simibubi.create.AllBlocks;
import com.simibubi.create.content.kinetics.KineticDebugger;
import com.simibubi.create.content.kinetics.simpleRelays.ICogWheel;
import com.simibubi.create.foundation.blockEntity.renderer.SafeBlockEntityRenderer;
import com.simibubi.create.foundation.render.CachedBufferer;
import com.simibubi.create.foundation.render.SuperByteBuffer;
import com.simibubi.create.foundation.render.SuperByteBufferCache;
import com.simibubi.create.foundation.utility.AnimationTickHolder;
import com.simibubi.create.foundation.utility.Color;
import net.minecraft.class_1921;
import net.minecraft.class_2338;
import net.minecraft.class_2350;
import net.minecraft.class_2350.class_2351;
import net.minecraft.class_2350.class_2352;
import net.minecraft.class_2680;
import net.minecraft.class_2741;
import net.minecraft.class_4587;
import net.minecraft.class_4588;
import net.minecraft.class_4597;
import net.minecraft.class_4696;
import net.minecraft.class_5614;

public class KineticBlockEntityRenderer<T extends KineticBlockEntity> extends SafeBlockEntityRenderer<T> {

	public static final SuperByteBufferCache.Compartment<class_2680> KINETIC_BLOCK = new SuperByteBufferCache.Compartment<>();
	public static boolean rainbowMode = false;

	public KineticBlockEntityRenderer(class_5614.class_5615 context) {
	}

	@Override
	protected void renderSafe(T be, float partialTicks, class_4587 ms, class_4597 buffer,
		int light, int overlay) {
		if (Backend.canUseInstancing(be.method_10997())) return;

		class_2680 state = getRenderedBlockState(be);
		class_1921 type = getRenderType(be, state);
		if (type != null)
			renderRotatingBuffer(be, getRotatedModel(be, state), ms, buffer.getBuffer(type), light);
	}

	protected class_2680 getRenderedBlockState(T be) {
		return be.method_11010();
	}

	protected class_1921 getRenderType(T be, class_2680 state) {
		return class_4696.method_23679(state);
	}

	protected SuperByteBuffer getRotatedModel(T be, class_2680 state) {
		return CachedBufferer.block(KINETIC_BLOCK, state);
	}

	public static void renderRotatingKineticBlock(KineticBlockEntity be, class_2680 renderedState, class_4587 ms,
		class_4588 buffer, int light) {
		SuperByteBuffer superByteBuffer = CachedBufferer.block(KINETIC_BLOCK, renderedState);
		renderRotatingBuffer(be, superByteBuffer, ms, buffer, light);
	}

	public static void renderRotatingBuffer(KineticBlockEntity be, SuperByteBuffer superBuffer, class_4587 ms,
		class_4588 buffer, int light) {
		standardKineticRotationTransform(superBuffer, be, light).renderInto(ms, buffer);
	}

	public static float getAngleForTe(KineticBlockEntity be, final class_2338 pos, class_2351 axis) {
		float time = AnimationTickHolder.getRenderTime(be.method_10997());
		float offset = getRotationOffsetForPosition(be, pos, axis);
		float angle = ((time * be.getSpeed() * 3f / 10 + offset) % 360) / 180 * (float) Math.PI;
		return angle;
	}

	public static SuperByteBuffer standardKineticRotationTransform(SuperByteBuffer buffer, KineticBlockEntity be,
		int light) {
		final class_2338 pos = be.method_11016();
		class_2351 axis = ((IRotate) be.method_11010()
			.method_26204()).getRotationAxis(be.method_11010());
		return kineticRotationTransform(buffer, be, axis, getAngleForTe(be, pos, axis), light);
	}

	public static SuperByteBuffer kineticRotationTransform(SuperByteBuffer buffer, KineticBlockEntity be, class_2351 axis,
		float angle, int light) {
		buffer.light(light);
		buffer.rotateCentered(class_2350.method_10156(class_2352.field_11056, axis), angle);

		if (KineticDebugger.isActive()) {
			rainbowMode = true;
			buffer.color(be.hasNetwork() ? Color.generateFromLong(be.network) : Color.WHITE);
		} else {
			float overStressedEffect = be.effects.overStressedEffect;
			if (overStressedEffect != 0)
				if (overStressedEffect > 0)
					buffer.color(Color.WHITE.mixWith(Color.RED, overStressedEffect));
				else
					buffer.color(Color.WHITE.mixWith(Color.SPRING_GREEN, -overStressedEffect));
			else
				buffer.color(Color.WHITE);
		}

		return buffer;
	}

	public static float getRotationOffsetForPosition(KineticBlockEntity be, final class_2338 pos, final class_2351 axis) {
		float offset = ICogWheel.isLargeCog(be.method_11010()) ? 11.25f : 0;
		double d = (((axis == class_2351.field_11048) ? 0 : pos.method_10263()) + ((axis == class_2351.field_11052) ? 0 : pos.method_10264())
			+ ((axis == class_2351.field_11051) ? 0 : pos.method_10260())) % 2;
		if (d == 0)
			offset = 22.5f;
		return offset + be.getRotationAngleOffset(axis);
	}

	public static class_2680 shaft(class_2351 axis) {
		return AllBlocks.SHAFT.getDefaultState()
			.method_11657(class_2741.field_12496, axis);
	}

	public static class_2351 getRotationAxisOf(KineticBlockEntity be) {
		return ((IRotate) be.method_11010()
			.method_26204()).getRotationAxis(be.method_11010());
	}

}
