package com.simibubi.create.content.kinetics.belt;

import static com.simibubi.create.content.kinetics.belt.BeltPart.MIDDLE;
import static com.simibubi.create.content.kinetics.belt.BeltSlope.HORIZONTAL;
import static net.minecraft.class_2350.class_2352.field_11060;
import static net.minecraft.class_2350.class_2352.field_11056;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.function.Function;

import org.jetbrains.annotations.Nullable;

import com.jozufozu.flywheel.light.LightListener;
import com.jozufozu.flywheel.light.LightUpdater;
import com.jozufozu.flywheel.util.box.GridAlignedBB;
import com.jozufozu.flywheel.util.box.ImmutableBox;
import com.simibubi.create.AllBlocks;
import com.simibubi.create.content.kinetics.base.IRotate;
import com.simibubi.create.content.kinetics.base.KineticBlockEntity;
import com.simibubi.create.content.kinetics.belt.behaviour.DirectBeltInputBehaviour;
import com.simibubi.create.content.kinetics.belt.behaviour.TransportedItemStackHandlerBehaviour;
import com.simibubi.create.content.kinetics.belt.behaviour.TransportedItemStackHandlerBehaviour.TransportedResult;
import com.simibubi.create.content.kinetics.belt.transport.BeltInventory;
import com.simibubi.create.content.kinetics.belt.transport.BeltMovementHandler;
import com.simibubi.create.content.kinetics.belt.transport.BeltMovementHandler.TransportedEntityInfo;
import com.simibubi.create.content.kinetics.belt.transport.BeltTunnelInteractionHandler;
import com.simibubi.create.content.kinetics.belt.transport.ItemHandlerBeltSegment;
import com.simibubi.create.content.kinetics.belt.transport.TransportedItemStack;
import com.simibubi.create.content.logistics.tunnel.BrassTunnelBlockEntity;
import com.simibubi.create.foundation.blockEntity.behaviour.BlockEntityBehaviour;
import com.simibubi.create.foundation.utility.NBTHelper;
import com.tterrag.registrate.fabric.EnvExecutor;

import net.fabricmc.api.EnvType;
import net.fabricmc.api.Environment;
import net.fabricmc.fabric.api.rendering.data.v1.RenderAttachmentBlockEntity;
import net.fabricmc.fabric.api.transfer.v1.item.ItemVariant;
import net.fabricmc.fabric.api.transfer.v1.storage.Storage;
import net.fabricmc.fabric.api.transfer.v1.storage.base.SidedStorageBlockEntity;
import net.minecraft.class_1297;
import net.minecraft.class_1767;
import net.minecraft.class_1799;
import net.minecraft.class_1944;
import net.minecraft.class_2248;
import net.minecraft.class_2338;
import net.minecraft.class_2338.class_2339;
import net.minecraft.class_2350;
import net.minecraft.class_2350.class_2351;
import net.minecraft.class_238;
import net.minecraft.class_2382;
import net.minecraft.class_243;
import net.minecraft.class_2487;
import net.minecraft.class_2512;
import net.minecraft.class_2586;
import net.minecraft.class_2591;
import net.minecraft.class_2680;
import net.minecraft.class_2741;
import net.minecraft.class_765;

public class BeltBlockEntity extends KineticBlockEntity implements SidedStorageBlockEntity, RenderAttachmentBlockEntity {

	public Map<class_1297, TransportedEntityInfo> passengers;
	public Optional<class_1767> color;
	public int beltLength;
	public int index;
	public class_2350 lastInsert;
	public CasingType casing;
	public boolean covered;

	protected class_2338 controller;
	protected BeltInventory inventory;
	protected Storage<ItemVariant> itemHandler;

	public class_2487 trackerUpdateTag;

	@Environment(EnvType.CLIENT)
	public BeltLighter lighter;

	public static enum CasingType {
		NONE, ANDESITE, BRASS;
	}

	public BeltBlockEntity(class_2591<?> type, class_2338 pos, class_2680 state) {
		super(type, pos, state);
		controller = class_2338.field_10980;
		itemHandler = null;
		casing = CasingType.NONE;
		color = Optional.empty();
	}

	@Override
	public void addBehaviours(List<BlockEntityBehaviour> behaviours) {
		super.addBehaviours(behaviours);
		behaviours.add(new DirectBeltInputBehaviour(this).onlyInsertWhen(this::canInsertFrom)
			.setInsertionHandler(this::tryInsertingFromSide).considerOccupiedWhen(this::isOccupied));
		behaviours.add(new TransportedItemStackHandlerBehaviour(this, this::applyToAllItems)
			.withStackPlacement(this::getWorldPositionOf));
	}

	@Override
	public void tick() {
		// Init belt
		if (beltLength == 0)
			BeltBlock.initBelt(field_11863, field_11867);

		super.tick();

		if (!AllBlocks.BELT.has(field_11863.method_8320(field_11867)))
			return;

		initializeItemHandler();

		// Move Items
		if (!isController())
			return;

		EnvExecutor.runWhenOn(EnvType.CLIENT, () -> () -> {
			if (beltLength > 0 && lighter == null) {
				lighter = new BeltLighter();
			}
		});
		invalidateRenderBoundingBox();

		getInventory().tick();

		if (getSpeed() == 0)
			return;

		// Move Entities
		if (passengers == null)
			passengers = new HashMap<>();

		List<class_1297> toRemove = new ArrayList<>();
		passengers.forEach((entity, info) -> {
			boolean canBeTransported = BeltMovementHandler.canBeTransported(entity);
			boolean leftTheBelt =
				info.getTicksSinceLastCollision() > ((method_11010().method_11654(BeltBlock.SLOPE) != HORIZONTAL) ? 3 : 1);
			if (!canBeTransported || leftTheBelt) {
				toRemove.add(entity);
				return;
			}

			info.tick();
			BeltMovementHandler.transportEntity(this, entity, info);
		});
		toRemove.forEach(passengers::remove);
	}

	@Override
	public float calculateStressApplied() {
		if (!isController())
			return 0;
		return super.calculateStressApplied();
	}

	@Override
	public class_238 createRenderBoundingBox() {
		if (!isController())
			return super.createRenderBoundingBox();
		else
			return super.createRenderBoundingBox().method_1014(beltLength + 1);
	}

	protected void initializeItemHandler() {
		if (field_11863.field_9236 || itemHandler != null)
			return;
		if (beltLength == 0 || controller == null)
			return;
		if (!field_11863.method_8477(controller))
			return;
		class_2586 be = field_11863.method_8321(controller);
		if (be == null || !(be instanceof BeltBlockEntity))
			return;
		BeltInventory inventory = ((BeltBlockEntity) be).getInventory();
		if (inventory == null)
			return;
		itemHandler = new ItemHandlerBeltSegment(inventory, index);
	}

	@Nullable
	@Override
	public Storage<ItemVariant> getItemStorage(@Nullable class_2350 direction) {
		if (!method_11015() && itemHandler == null)
			initializeItemHandler();
		return itemHandler;
	}

	@Override
	public void destroy() {
		super.destroy();
		if (isController())
			getInventory().ejectAll();
	}

	@Override
	public void invalidate() {
		super.invalidate();
		itemHandler = null;
	}

	@Override
	public void write(class_2487 compound, boolean clientPacket) {
		if (controller != null)
			compound.method_10566("Controller", class_2512.method_10692(controller));
		compound.method_10556("IsController", isController());
		compound.method_10569("Length", beltLength);
		compound.method_10569("Index", index);
		NBTHelper.writeEnum(compound, "Casing", casing);
		compound.method_10556("Covered", covered);

		if (color.isPresent())
			NBTHelper.writeEnum(compound, "Dye", color.get());

		if (isController())
			compound.method_10566("Inventory", getInventory().write());
		super.write(compound, clientPacket);
	}

	@Override
	protected void read(class_2487 compound, boolean clientPacket) {
		int prevBeltLength = beltLength;
		super.read(compound, clientPacket);

		if (compound.method_10577("IsController"))
			controller = field_11867;

		color = compound.method_10545("Dye") ? Optional.of(NBTHelper.readEnum(compound, "Dye", class_1767.class))
			: Optional.empty();

		if (!wasMoved) {
			if (!isController())
				controller = class_2512.method_10691(compound.method_10562("Controller"));
			trackerUpdateTag = compound;
			index = compound.method_10550("Index");
			beltLength = compound.method_10550("Length");
			if (prevBeltLength != beltLength) {
				EnvExecutor.runWhenOn(EnvType.CLIENT, () -> () -> {
					if (lighter != null) {
						lighter.initializeLight();
					}
				});
			}
		}

		if (isController())
			getInventory().read(compound.method_10562("Inventory"));

		CasingType casingBefore = casing;
		boolean coverBefore = covered;
		casing = NBTHelper.readEnum(compound, "Casing", CasingType.class);
		covered = compound.method_10577("Covered");

		if (!clientPacket)
			return;

		if (casingBefore == casing && coverBefore == covered)
			return;
		if (!isVirtual())
			requestModelDataUpdate();
		if (method_11002())
			field_11863.method_8413(method_11016(), method_11010(), method_11010(), 16);
	}

	@Override
	public void clearKineticInformation() {
		super.clearKineticInformation();
		beltLength = 0;
		index = 0;
		controller = null;
		trackerUpdateTag = new class_2487();
	}

	public boolean applyColor(class_1767 colorIn) {
		if (colorIn == null) {
			if (!color.isPresent())
				return false;
		} else if (color.isPresent() && color.get() == colorIn)
			return false;
		if (field_11863.method_8608())
			return true;

		for (class_2338 blockPos : BeltBlock.getBeltChain(field_11863, getController())) {
			BeltBlockEntity belt = BeltHelper.getSegmentBE(field_11863, blockPos);
			if (belt == null)
				continue;
			belt.color = Optional.ofNullable(colorIn);
			belt.method_5431();
			belt.sendData();
		}

		return true;
	}

	public BeltBlockEntity getControllerBE() {
		if (controller == null)
			return null;
		if (!field_11863.method_8477(controller))
			return null;
		class_2586 be = field_11863.method_8321(controller);
		if (be == null || !(be instanceof BeltBlockEntity))
			return null;
		return (BeltBlockEntity) be;
	}

	public void setController(class_2338 controller) {
		this.controller = controller;
	}

	public class_2338 getController() {
		return controller == null ? field_11867 : controller;
	}

	public boolean isController() {
		return controller != null && field_11867.method_10263() == controller.method_10263()
			&& field_11867.method_10264() == controller.method_10264() && field_11867.method_10260() == controller.method_10260();
	}

	public float getBeltMovementSpeed() {
		return getSpeed() / 480f;
	}

	public float getDirectionAwareBeltMovementSpeed() {
		int offset = getBeltFacing().method_10171()
			.method_10181();
		if (getBeltFacing().method_10166() == class_2351.field_11048)
			offset *= -1;
		return getBeltMovementSpeed() * offset;
	}

	public boolean hasPulley() {
		if (!AllBlocks.BELT.has(method_11010()))
			return false;
		return method_11010().method_11654(BeltBlock.PART) != MIDDLE;
	}

	protected boolean isLastBelt() {
		if (getSpeed() == 0)
			return false;

		class_2350 direction = getBeltFacing();
		if (method_11010().method_11654(BeltBlock.SLOPE) == BeltSlope.VERTICAL)
			return false;

		BeltPart part = method_11010().method_11654(BeltBlock.PART);
		if (part == MIDDLE)
			return false;

		boolean movingPositively = (getSpeed() > 0 == (direction.method_10171()
			.method_10181() == 1)) ^ direction.method_10166() == class_2351.field_11048;
		return part == BeltPart.START ^ movingPositively;
	}

	public class_2382 getMovementDirection(boolean firstHalf) {
		return this.getMovementDirection(firstHalf, false);
	}

	public class_2382 getBeltChainDirection() {
		return this.getMovementDirection(true, true);
	}

	protected class_2382 getMovementDirection(boolean firstHalf, boolean ignoreHalves) {
		if (getSpeed() == 0)
			return class_2338.field_10980;

		final class_2680 blockState = method_11010();
		final class_2350 beltFacing = blockState.method_11654(class_2741.field_12481);
		final BeltSlope slope = blockState.method_11654(BeltBlock.SLOPE);
		final BeltPart part = blockState.method_11654(BeltBlock.PART);
		final class_2351 axis = beltFacing.method_10166();

		class_2350 movementFacing = class_2350.method_10156(axis == class_2351.field_11048 ? field_11060 : field_11056, axis);
		boolean notHorizontal = blockState.method_11654(BeltBlock.SLOPE) != HORIZONTAL;
		if (getSpeed() < 0)
			movementFacing = movementFacing.method_10153();
		class_2382 movement = movementFacing.method_10163();

		boolean slopeBeforeHalf = (part == BeltPart.END) == (beltFacing.method_10171() == field_11056);
		boolean onSlope = notHorizontal && (part == MIDDLE || slopeBeforeHalf == firstHalf || ignoreHalves);
		boolean movingUp = onSlope && slope == (movementFacing == beltFacing ? BeltSlope.UPWARD : BeltSlope.DOWNWARD);

		if (!onSlope)
			return movement;

		return new class_2382(movement.method_10263(), movingUp ? 1 : -1, movement.method_10260());
	}

	public class_2350 getMovementFacing() {
		class_2351 axis = getBeltFacing().method_10166();
		return class_2350.method_10169(axis, getBeltMovementSpeed() < 0 ^ axis == class_2351.field_11048 ? field_11060 : field_11056);
	}

	protected class_2350 getBeltFacing() {
		return method_11010().method_11654(class_2741.field_12481);
	}

	public BeltInventory getInventory() {
		if (!isController()) {
			BeltBlockEntity controllerBE = getControllerBE();
			if (controllerBE != null)
				return controllerBE.getInventory();
			return null;
		}
		if (inventory == null) {
			inventory = new BeltInventory(this);
		}
		return inventory;
	}

	private void applyToAllItems(float maxDistanceFromCenter,
		Function<TransportedItemStack, TransportedResult> processFunction) {
		BeltBlockEntity controller = getControllerBE();
		if (controller == null)
			return;
		BeltInventory inventory = controller.getInventory();
		if (inventory != null)
			inventory.applyToEachWithin(index + .5f, maxDistanceFromCenter, processFunction);
	}

	private class_243 getWorldPositionOf(TransportedItemStack transported) {
		BeltBlockEntity controllerBE = getControllerBE();
		if (controllerBE == null)
			return class_243.field_1353;
		return BeltHelper.getVectorForOffset(controllerBE, transported.beltPosition);
	}

	public void setCasingType(CasingType type) {
		if (casing == type)
			return;

		class_2680 blockState = method_11010();
		boolean shouldBlockHaveCasing = type != CasingType.NONE;

		if (field_11863.field_9236) {
			casing = type;
			field_11863.method_8652(field_11867, blockState.method_11657(BeltBlock.CASING, shouldBlockHaveCasing), 0);
			requestModelDataUpdate();
			field_11863.method_8413(field_11867, method_11010(), method_11010(), 16);
			return;
		}

		if (casing != CasingType.NONE)
			field_11863.method_20290(2001, field_11867,
				class_2248.method_9507(casing == CasingType.ANDESITE ? AllBlocks.ANDESITE_CASING.getDefaultState()
					: AllBlocks.BRASS_CASING.getDefaultState()));
		if (blockState.method_11654(BeltBlock.CASING) != shouldBlockHaveCasing)
			KineticBlockEntity.switchToBlockState(field_11863, field_11867,
				blockState.method_11657(BeltBlock.CASING, shouldBlockHaveCasing));
		casing = type;
		method_5431();
		sendData();
	}

	private boolean canInsertFrom(class_2350 side) {
		if (getSpeed() == 0)
			return false;
		class_2680 state = method_11010();
		if (state.method_28498(BeltBlock.SLOPE) && (state.method_11654(BeltBlock.SLOPE) == BeltSlope.SIDEWAYS
			|| state.method_11654(BeltBlock.SLOPE) == BeltSlope.VERTICAL))
			return false;
		return getMovementFacing() != side.method_10153();
	}
	
	private boolean isOccupied(class_2350 side) {
		BeltBlockEntity nextBeltController = getControllerBE();
		if (nextBeltController == null)
			return true;
		BeltInventory nextInventory = nextBeltController.getInventory();
		if (nextInventory == null)
			return true;
		if (getSpeed() == 0)
			return true;
		if (getMovementFacing() == side.method_10153())
			return true;
		if (!nextInventory.canInsertAtFromSide(index, side))
			return true;
		return false;
	}

	private class_1799 tryInsertingFromSide(TransportedItemStack transportedStack, class_2350 side, boolean simulate) {
		BeltBlockEntity nextBeltController = getControllerBE();
		class_1799 inserted = transportedStack.stack;
		class_1799 empty = class_1799.field_8037;

		if (nextBeltController == null)
			return inserted;
		BeltInventory nextInventory = nextBeltController.getInventory();
		if (nextInventory == null)
			return inserted;

		class_2586 teAbove = field_11863.method_8321(field_11867.method_10084());
		if (teAbove instanceof BrassTunnelBlockEntity) {
			BrassTunnelBlockEntity tunnelBE = (BrassTunnelBlockEntity) teAbove;
			if (tunnelBE.hasDistributionBehaviour()) {
				if (!tunnelBE.getStackToDistribute()
					.method_7960())
					return inserted;
				if (!tunnelBE.testFlapFilter(side.method_10153(), inserted))
					return inserted;
				if (!simulate) {
					BeltTunnelInteractionHandler.flapTunnel(nextInventory, index, side.method_10153(), true);
					tunnelBE.setStackToDistribute(inserted, side.method_10153(), null);
				}
				return empty;
			}
		}

		if (isOccupied(side))
			return inserted;
		if (simulate)
			return empty;

		transportedStack = transportedStack.copy();
		transportedStack.beltPosition = index + .5f - Math.signum(getDirectionAwareBeltMovementSpeed()) / 16f;

		class_2350 movementFacing = getMovementFacing();
		if (!side.method_10166()
			.method_10178()) {
			if (movementFacing != side) {
				transportedStack.sideOffset = side.method_10171()
					.method_10181() * .35f;
				if (side.method_10166() == class_2351.field_11048)
					transportedStack.sideOffset *= -1;
			} else
				transportedStack.beltPosition = getDirectionAwareBeltMovementSpeed() > 0 ? index : index + 1;
		}

		transportedStack.prevSideOffset = transportedStack.sideOffset;
		transportedStack.insertedAt = index;
		transportedStack.insertedFrom = side;
		transportedStack.prevBeltPosition = transportedStack.beltPosition;

		BeltTunnelInteractionHandler.flapTunnel(nextInventory, index, side.method_10153(), true);

		nextInventory.addItem(transportedStack);
		nextBeltController.method_5431();
		nextBeltController.sendData();
		return empty;
	}

	@Override
	public RenderData getRenderAttachmentData() {
		return new RenderData(casing, covered);
	}

	@Override
	protected boolean canPropagateDiagonally(IRotate block, class_2680 state) {
		return state.method_28498(BeltBlock.SLOPE) && (state.method_11654(BeltBlock.SLOPE) == BeltSlope.UPWARD
			|| state.method_11654(BeltBlock.SLOPE) == BeltSlope.DOWNWARD);
	}

	@Override
	public float propagateRotationTo(KineticBlockEntity target, class_2680 stateFrom, class_2680 stateTo, class_2338 diff,
		boolean connectedViaAxes, boolean connectedViaCogs) {
		if (target instanceof BeltBlockEntity && !connectedViaAxes)
			return getController().equals(((BeltBlockEntity) target).getController()) ? 1 : 0;
		return 0;
	}

	public void invalidateItemHandler() {
		itemHandler = null;
	}

	public boolean shouldRenderNormally() {
		if (field_11863 == null)
			return isController();
		class_2680 state = method_11010();
		return state != null && state.method_28498(BeltBlock.PART) && state.method_11654(BeltBlock.PART) == BeltPart.START;
	}

	/**
	 * Hide this behavior in an inner class to avoid loading LightListener on servers.
	 */
	@Environment(EnvType.CLIENT)
	class BeltLighter implements LightListener {
		private byte[] light;

		public BeltLighter() {
			initializeLight();
			LightUpdater.get(field_11863)
					.addListener(this);
		}

		/**
		 * Get the number of belt segments represented by the lighter.
		 * @return The number of segments.
		 */
		public int lightSegments() {
			return light == null ? 0 : light.length / 2;
		}

		/**
		 * Get the light value for a given segment.
		 * @param segment The segment to get the light value for.
		 * @return The light value.
		 */
		public int getPackedLight(int segment) {
			return light == null ? 0 : class_765.method_23687(light[segment * 2], light[segment * 2 + 1]);
		}

		@Override
		public GridAlignedBB getVolume() {
			class_2338 endPos = BeltHelper.getPositionForOffset(BeltBlockEntity.this, beltLength - 1);
			GridAlignedBB bb = GridAlignedBB.from(field_11867, endPos);
			bb.fixMinMax();
			return bb;
		}

		@Override
		public boolean isListenerInvalid() {
			return field_11865;
		}

		@Override
		public void onLightUpdate(class_1944 type, ImmutableBox changed) {
			if (field_11865)
				return;
			if (field_11863 == null)
				return;

			GridAlignedBB beltVolume = getVolume();

			if (beltVolume.intersects(changed)) {
				if (type == class_1944.field_9282)
					updateBlockLight();

				if (type == class_1944.field_9284)
					updateSkyLight();
			}
		}

		private void initializeLight() {
			light = new byte[beltLength * 2];

			class_2382 vec = getBeltFacing().method_10163();
			BeltSlope slope = method_11010().method_11654(BeltBlock.SLOPE);
			int verticality = slope == BeltSlope.DOWNWARD ? -1 : slope == BeltSlope.UPWARD ? 1 : 0;

			class_2339 pos = new class_2339(controller.method_10263(), controller.method_10264(), controller.method_10260());
			for (int i = 0; i < beltLength * 2; i += 2) {
				light[i] = (byte) field_11863.method_8314(class_1944.field_9282, pos);
				light[i + 1] = (byte) field_11863.method_8314(class_1944.field_9284, pos);
				pos.method_10100(vec.method_10263(), verticality, vec.method_10260());
			}
		}

		private void updateBlockLight() {
			class_2382 vec = getBeltFacing().method_10163();
			BeltSlope slope = method_11010().method_11654(BeltBlock.SLOPE);
			int verticality = slope == BeltSlope.DOWNWARD ? -1 : slope == BeltSlope.UPWARD ? 1 : 0;

			class_2339 pos = new class_2339(controller.method_10263(), controller.method_10264(), controller.method_10260());
			for (int i = 0; i < beltLength * 2; i += 2) {
				light[i] = (byte) field_11863.method_8314(class_1944.field_9282, pos);

				pos.method_10100(vec.method_10263(), verticality, vec.method_10260());
			}
		}

		private void updateSkyLight() {
			class_2382 vec = getBeltFacing().method_10163();
			BeltSlope slope = method_11010().method_11654(BeltBlock.SLOPE);
			int verticality = slope == BeltSlope.DOWNWARD ? -1 : slope == BeltSlope.UPWARD ? 1 : 0;

			class_2339 pos = new class_2339(controller.method_10263(), controller.method_10264(), controller.method_10260());
			for (int i = 1; i < beltLength * 2; i += 2) {
				light[i] = (byte) field_11863.method_8314(class_1944.field_9284, pos);

				pos.method_10100(vec.method_10263(), verticality, vec.method_10260());
			}
		}
	}

	public void setCovered(boolean blockCoveringBelt) {
		if (blockCoveringBelt == covered)
			return;
		covered = blockCoveringBelt;
		notifyUpdate();
	}

	public record RenderData(CasingType casingType, boolean covered) {
	}
}
