package com.simibubi.create.content.kinetics.belt;

import java.util.function.Supplier;

import com.simibubi.create.AllPartialModels;
import com.simibubi.create.AllSpriteShifts;
import com.simibubi.create.content.kinetics.belt.BeltBlockEntity.CasingType;
import com.simibubi.create.content.kinetics.belt.BeltBlockEntity.RenderData;
import com.simibubi.create.foundation.block.render.SpriteShiftEntry;

import io.github.fabricators_of_create.porting_lib.models.CustomParticleIconModel;
import net.fabricmc.fabric.api.renderer.v1.model.FabricBakedModel;
import net.fabricmc.fabric.api.renderer.v1.model.ForwardingBakedModel;
import net.fabricmc.fabric.api.renderer.v1.model.SpriteFinder;
import net.fabricmc.fabric.api.renderer.v1.render.RenderContext;
import net.fabricmc.fabric.api.rendering.data.v1.RenderAttachedBlockView;
import net.minecraft.class_1058;
import net.minecraft.class_1087;
import net.minecraft.class_1723;
import net.minecraft.class_1920;
import net.minecraft.class_2338;
import net.minecraft.class_2350.class_2351;
import net.minecraft.class_2680;
import net.minecraft.class_310;
import net.minecraft.class_5819;

public class BeltModel extends ForwardingBakedModel implements CustomParticleIconModel {

	private static final SpriteShiftEntry SPRITE_SHIFT = AllSpriteShifts.ANDESIDE_BELT_CASING;

	public BeltModel(class_1087 template) {
		wrapped = template;
	}

	@Override
	public class_1058 getParticleIcon(Object data) {
		if (data instanceof RenderData renderData) {
			CasingType type = renderData.casingType();
			if (type == CasingType.NONE || type == CasingType.BRASS)
				return CustomParticleIconModel.super.getParticleIcon(data);
		}

		return AllSpriteShifts.ANDESITE_CASING.getOriginal();
	}

	@Override
	public boolean isVanillaAdapter() {
		return false;
	}

	@Override
	public void emitBlockQuads(class_1920 blockView, class_2680 state, class_2338 pos, Supplier<class_5819> randomSupplier, RenderContext context) {
		if (!(blockView instanceof RenderAttachedBlockView attachmentView
				&& attachmentView.getBlockEntityRenderAttachment(pos) instanceof RenderData data)) {
			super.emitBlockQuads(blockView, state, pos, randomSupplier, context);
			return;
		}

		boolean cover = data.covered();
		CasingType type = data.casingType();
		boolean brassCasing = type == CasingType.BRASS;

		if (type == CasingType.NONE || brassCasing && !cover) {
			super.emitBlockQuads(blockView, state, pos, randomSupplier, context);
			return;
		}

		if (!brassCasing) {
			SpriteFinder spriteFinder = SpriteFinder.get(class_310.method_1551().method_1554().method_24153(class_1723.field_21668));
			context.pushTransform(quad -> {
				class_1058 sprite = spriteFinder.find(quad, 0);
				if (sprite == SPRITE_SHIFT.getOriginal()) {
					for (int vertex = 0; vertex < 4; vertex++) {
						float u = quad.spriteU(vertex, 0);
						float v = quad.spriteV(vertex, 0);
						quad.sprite(vertex, 0,
								SPRITE_SHIFT.getTargetU(u),
								SPRITE_SHIFT.getTargetV(v)
						);
					}
				}
				return true;
			});
		}

		super.emitBlockQuads(blockView, state, pos, randomSupplier, context);

		if (cover) {
			boolean alongX = state.method_11654(BeltBlock.HORIZONTAL_FACING)
				.method_10166() == class_2351.field_11048;
			class_1087 coverModel =
				(brassCasing ? alongX ? AllPartialModels.BRASS_BELT_COVER_X : AllPartialModels.BRASS_BELT_COVER_Z
					: alongX ? AllPartialModels.ANDESITE_BELT_COVER_X : AllPartialModels.ANDESITE_BELT_COVER_Z).get();
			((FabricBakedModel) coverModel).emitBlockQuads(blockView, state, pos, randomSupplier, context);
		}

		if (!brassCasing) {
			context.popTransform();
		}
	}

}
