package com.simibubi.create.content.kinetics.belt.transport;

import java.util.Random;
import net.minecraft.class_1799;
import net.minecraft.class_2350;
import net.minecraft.class_2487;
import com.simibubi.create.content.kinetics.belt.BeltHelper;
import com.simibubi.create.content.kinetics.fan.processing.FanProcessingType;

import io.github.fabricators_of_create.porting_lib.util.NBTSerializer;

public class TransportedItemStack implements Comparable<TransportedItemStack> {

	private static Random R = new Random();

	// fabric: can't use null in some places, have a static empty stack
	public static final TransportedItemStack EMPTY = new TransportedItemStack(class_1799.field_8037);

	public class_1799 stack;
	public float beltPosition;
	public float sideOffset;
	public int angle;
	public int insertedAt;
	public class_2350 insertedFrom;
	public boolean locked;
	public boolean lockedExternally;

	public float prevBeltPosition;
	public float prevSideOffset;

	public FanProcessingType processedBy;
	public int processingTime;

	public TransportedItemStack(class_1799 stack) {
		this.stack = stack;
		boolean centered = BeltHelper.isItemUpright(stack);
		angle = centered ? 180 : R.nextInt(360);
		sideOffset = prevSideOffset = getTargetSideOffset();
		insertedFrom = class_2350.field_11036;
	}

	public float getTargetSideOffset() {
		return (angle - 180) / (360 * 3f);
	}

	@Override
	public int compareTo(TransportedItemStack o) {
		return beltPosition < o.beltPosition ? 1 : beltPosition > o.beltPosition ? -1 : 0;
	}

	public TransportedItemStack getSimilar() {
		TransportedItemStack copy = new TransportedItemStack(stack.method_7972());
		copy.beltPosition = beltPosition;
		copy.insertedAt = insertedAt;
		copy.insertedFrom = insertedFrom;
		copy.prevBeltPosition = prevBeltPosition;
		copy.prevSideOffset = prevSideOffset;
		copy.processedBy = processedBy;
		copy.processingTime = processingTime;
		return copy;
	}

	public TransportedItemStack copy() {
		TransportedItemStack copy = getSimilar();
		copy.angle = angle;
		copy.sideOffset = sideOffset;
		return copy;
	}

	// fabric: full copy, used for snapshots
	public TransportedItemStack fullCopy() {
		TransportedItemStack copy = copy();
		copy.locked = locked;
		copy.lockedExternally = lockedExternally;
		return copy;
	}

	public class_2487 serializeNBT() {
		class_2487 nbt = new class_2487();
		nbt.method_10566("Item", NBTSerializer.serializeNBT(stack));
		nbt.method_10548("Pos", beltPosition);
		nbt.method_10548("PrevPos", prevBeltPosition);
		nbt.method_10548("Offset", sideOffset);
		nbt.method_10548("PrevOffset", prevSideOffset);
		nbt.method_10569("InSegment", insertedAt);
		nbt.method_10569("Angle", angle);
		nbt.method_10569("InDirection", insertedFrom.method_10146());
		if (locked)
			nbt.method_10556("Locked", locked);
		if (lockedExternally)
			nbt.method_10556("LockedExternally", lockedExternally);
		return nbt;
	}

	public static TransportedItemStack read(class_2487 nbt) {
		TransportedItemStack stack = new TransportedItemStack(class_1799.method_7915(nbt.method_10562("Item")));
		stack.beltPosition = nbt.method_10583("Pos");
		stack.prevBeltPosition = nbt.method_10583("PrevPos");
		stack.sideOffset = nbt.method_10583("Offset");
		stack.prevSideOffset = nbt.method_10583("PrevOffset");
		stack.insertedAt = nbt.method_10550("InSegment");
		stack.angle = nbt.method_10550("Angle");
		stack.insertedFrom = class_2350.method_10143(nbt.method_10550("InDirection"));
		stack.locked = nbt.method_10577("Locked");
		stack.lockedExternally = nbt.method_10577("LockedExternally");
		return stack;
	}

	// fabric: a proper equals() is required for correct functionality with transfer snapshots. (#903)
	@Override
	public boolean equals(Object o) {
		if (this == o) return true;
		if (o == null || getClass() != o.getClass()) return false;

		TransportedItemStack that = (TransportedItemStack) o;

		if (Float.compare(that.beltPosition, beltPosition) != 0) return false;
		if (Float.compare(that.sideOffset, sideOffset) != 0) return false;
		if (angle != that.angle) return false;
		if (insertedAt != that.insertedAt) return false;
		if (locked != that.locked) return false;
		if (lockedExternally != that.lockedExternally) return false;
		if (Float.compare(that.prevBeltPosition, prevBeltPosition) != 0) return false;
		if (Float.compare(that.prevSideOffset, prevSideOffset) != 0) return false;
		if (processingTime != that.processingTime) return false;
		if (!class_1799.method_7973(stack, that.stack)) return false;
		if (insertedFrom != that.insertedFrom) return false;
		return processedBy == that.processedBy;
	}
}
