package com.simibubi.create.content.kinetics.clock;

import java.util.List;
import net.minecraft.class_1282;
import net.minecraft.class_1937.class_7867;
import net.minecraft.class_2338;
import net.minecraft.class_2398;
import net.minecraft.class_243;
import net.minecraft.class_2487;
import net.minecraft.class_2591;
import net.minecraft.class_2680;
import net.minecraft.class_3414;
import net.minecraft.class_3417;
import net.minecraft.class_3419;
import net.minecraft.class_3532;
import com.simibubi.create.content.kinetics.base.KineticBlockEntity;
import com.simibubi.create.foundation.advancement.AllAdvancements;
import com.simibubi.create.foundation.blockEntity.behaviour.BlockEntityBehaviour;
import com.simibubi.create.foundation.damageTypes.CreateDamageSources;
import com.simibubi.create.foundation.utility.AnimationTickHolder;
import com.simibubi.create.foundation.utility.NBTHelper;
import com.simibubi.create.foundation.utility.VecHelper;
import com.simibubi.create.foundation.utility.animation.LerpedFloat;
import com.simibubi.create.foundation.utility.animation.LerpedFloat.Chaser;

public class CuckooClockBlockEntity extends KineticBlockEntity {

	public LerpedFloat hourHand = LerpedFloat.angular();
	public LerpedFloat minuteHand = LerpedFloat.angular();
	public LerpedFloat animationProgress = LerpedFloat.linear();
	public Animation animationType;
	private boolean sendAnimationUpdate;

	enum Animation {
		PIG, CREEPER, SURPRISE, NONE;
	}

	public CuckooClockBlockEntity(class_2591<?> type, class_2338 pos, class_2680 state) {
		super(type, pos, state);
		animationType = Animation.NONE;
	}

	@Override
	public void addBehaviours(List<BlockEntityBehaviour> behaviours) {
		super.addBehaviours(behaviours);
		registerAwardables(behaviours, AllAdvancements.CUCKOO_CLOCK);
	}

	@Override
	protected void read(class_2487 compound, boolean clientPacket) {
		super.read(compound, clientPacket);
		if (clientPacket && compound.method_10545("Animation")) {
			animationType = NBTHelper.readEnum(compound, "Animation", Animation.class);
			animationProgress.startWithValue(0);
		}
	}

	@Override
	public void write(class_2487 compound, boolean clientPacket) {
		if (clientPacket && sendAnimationUpdate)
			NBTHelper.writeEnum(compound, "Animation", animationType);
		sendAnimationUpdate = false;
		super.write(compound, clientPacket);
	}

	@Override
	public void tick() {
		super.tick();
		if (getSpeed() == 0)
			return;


		boolean isNatural = field_11863.method_8597().comp_645();
		int dayTime = (int) ((field_11863.method_8532() * (isNatural ? 1 : 24)) % 24000);
		int hours = (dayTime / 1000 + 6) % 24;
		int minutes = (dayTime % 1000) * 60 / 1000;

		if (!isNatural) {
			if (field_11863.field_9236) {
				moveHands(hours, minutes);

				if (AnimationTickHolder.getTicks() % 6 == 0)
					playSound(class_3417.field_15204.comp_349(), 1 / 16f, 2f);
				else if (AnimationTickHolder.getTicks() % 3 == 0)
					playSound(class_3417.field_15204.comp_349(), 1 / 16f, 1.5f);
			}
			return;
		}

		if (!field_11863.field_9236) {
			if (animationType == Animation.NONE) {
				if (hours == 12 && minutes < 5)
					startAnimation(Animation.PIG);
				if (hours == 18 && minutes < 36 && minutes > 31)
					startAnimation(Animation.CREEPER);
			} else {
				float value = animationProgress.getValue();
				animationProgress.setValue(value + 1);
				if (value > 100)
					animationType = Animation.NONE;

				if (animationType == Animation.SURPRISE && class_3532.method_15347(animationProgress.getValue(), 50)) {
					class_243 center = VecHelper.getCenterOf(field_11867);
					field_11863.method_22352(field_11867, false);
					class_1282 damageSource = CreateDamageSources.cuckooSurprise(field_11863);
					field_11863.method_8454(null, damageSource, null, center.field_1352, center.field_1351, center.field_1350, 3, false,
						class_7867.field_40889);
				}

			}
		}

		if (field_11863.field_9236) {
			moveHands(hours, minutes);

			if (animationType == Animation.NONE) {
				if (AnimationTickHolder.getTicks() % 32 == 0)
					playSound(class_3417.field_15204.comp_349(), 1 / 16f, 2f);
				else if (AnimationTickHolder.getTicks() % 16 == 0)
					playSound(class_3417.field_15204.comp_349(), 1 / 16f, 1.5f);
			} else {

				boolean isSurprise = animationType == Animation.SURPRISE;
				float value = animationProgress.getValue();
				animationProgress.setValue(value + 1);
				if (value > 100)
					animationType = null;

				// sounds

				if (value == 1)
					playSound(class_3417.field_14725.comp_349(), 2, .5f);
				if (value == 21)
					playSound(class_3417.field_14725.comp_349(), 2, 0.793701f);

				if (value > 30 && isSurprise) {
					class_243 pos = VecHelper.offsetRandomly(VecHelper.getCenterOf(this.field_11867), field_11863.field_9229, .5f);
					field_11863.method_8406(class_2398.field_11237, pos.field_1352, pos.field_1351, pos.field_1350, 0, 0, 0);
				}
				if (value == 40 && isSurprise)
					playSound(class_3417.field_15079, 1f, 1f);

				int step = isSurprise ? 3 : 15;
				for (int phase = 30; phase <= 60; phase += step) {
					if (value == phase - step / 3)
						playSound(class_3417.field_14982, 1 / 16f, 2f);
					if (value == phase) {
						if (animationType == Animation.PIG)
							playSound(class_3417.field_14615, 1 / 4f, 1f);
						else
							playSound(class_3417.field_15192, 1 / 4f, 3f);
					}
					if (value == phase + step / 3)
						playSound(class_3417.field_14823, 1 / 16f, 2f);

				}

			}

			return;
		}
	}

	public void startAnimation(Animation animation) {
		animationType = animation;
		if (animation != null && CuckooClockBlock.containsSurprise(method_11010()))
			animationType = Animation.SURPRISE;
		animationProgress.startWithValue(0);
		sendAnimationUpdate = true;

		if (animation == Animation.CREEPER)
			awardIfNear(AllAdvancements.CUCKOO_CLOCK, 32);

		sendData();
	}

	public void moveHands(int hours, int minutes) {
		float hourTarget = (float) (360 / 12 * (hours % 12));
		float minuteTarget = (float) (360 / 60 * minutes);

		hourHand.chase(hourTarget, .2f, Chaser.EXP);
		minuteHand.chase(minuteTarget, .2f, Chaser.EXP);

		hourHand.tickChaser();
		minuteHand.tickChaser();
	}

	private void playSound(class_3414 sound, float volume, float pitch) {
		class_243 vec = VecHelper.getCenterOf(field_11867);
		field_11863.method_8486(vec.field_1352, vec.field_1351, vec.field_1350, sound, class_3419.field_15245, volume, pitch, false);
	}
}
