package com.simibubi.create.content.kinetics.crafter;

import static com.simibubi.create.content.kinetics.base.HorizontalKineticBlock.HORIZONTAL_FACING;

import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;
import java.util.Map.Entry;

import org.apache.commons.lang3.tuple.Pair;
import org.jetbrains.annotations.Nullable;

import com.simibubi.create.AllBlocks;
import com.simibubi.create.AllItems;
import com.simibubi.create.AllSoundEvents;
import com.simibubi.create.content.kinetics.base.KineticBlockEntity;
import com.simibubi.create.content.kinetics.belt.behaviour.DirectBeltInputBehaviour;
import com.simibubi.create.content.kinetics.crafter.ConnectedInputHandler.ConnectedInput;
import com.simibubi.create.content.kinetics.crafter.RecipeGridHandler.GroupedItems;
import com.simibubi.create.foundation.advancement.AllAdvancements;
import com.simibubi.create.foundation.blockEntity.behaviour.BlockEntityBehaviour;
import com.simibubi.create.foundation.blockEntity.behaviour.edgeInteraction.EdgeInteractionBehaviour;
import com.simibubi.create.foundation.blockEntity.behaviour.inventory.InvManipulationBehaviour;
import com.simibubi.create.foundation.item.SmartInventory;
import com.simibubi.create.foundation.utility.BlockFace;
import com.simibubi.create.foundation.utility.Pointing;
import com.simibubi.create.foundation.utility.VecHelper;

import io.github.fabricators_of_create.porting_lib.transfer.callbacks.TransactionCallback;
import net.fabricmc.fabric.api.transfer.v1.item.ItemVariant;
import net.fabricmc.fabric.api.transfer.v1.storage.Storage;
import net.fabricmc.fabric.api.transfer.v1.storage.base.SidedStorageBlockEntity;
import net.fabricmc.fabric.api.transfer.v1.transaction.TransactionContext;
import net.minecraft.class_1542;
import net.minecraft.class_1799;
import net.minecraft.class_1937;
import net.minecraft.class_2338;
import net.minecraft.class_2350;
import net.minecraft.class_2392;
import net.minecraft.class_2398;
import net.minecraft.class_243;
import net.minecraft.class_2487;
import net.minecraft.class_2591;
import net.minecraft.class_2680;
import net.minecraft.class_3417;
import net.minecraft.class_3419;
import net.minecraft.class_3532;

public class MechanicalCrafterBlockEntity extends KineticBlockEntity implements SidedStorageBlockEntity {

	enum Phase {
		IDLE, ACCEPTING, ASSEMBLING, EXPORTING, WAITING, CRAFTING, INSERTING;
	}

	public static class Inventory extends SmartInventory {

		private MechanicalCrafterBlockEntity blockEntity;

		public Inventory(MechanicalCrafterBlockEntity blockEntity) {
			super(1, blockEntity, 1, false);
			this.blockEntity = blockEntity;
			forbidExtraction();
			whenContentsChanged(() -> {
				if (getStackInSlot(0).method_7960()) // fabric: only has one slot, this is safe
					return;
				if (blockEntity.phase == Phase.IDLE)
					blockEntity.checkCompletedRecipe(false);
			});
		}

		@Override
		public long insert(ItemVariant resource, long maxAmount, TransactionContext transaction) {
			if (blockEntity.phase != Phase.IDLE)
				return 0;
			if (blockEntity.covered)
				return 0;
			long inserted = super.insert(resource, maxAmount, transaction);
			if (inserted != 0)
				TransactionCallback.onSuccess(transaction, () -> blockEntity.method_10997()
						.method_8396(null, blockEntity.method_11016(), class_3417.field_14667, class_3419.field_15245, .25f,
								.5f));
			return inserted;
		}

	}

	protected Inventory inventory;
	protected GroupedItems groupedItems = new GroupedItems();
	protected ConnectedInput input = new ConnectedInput();
	protected boolean reRender;
	protected Phase phase;
	protected int countDown;
	protected boolean covered;
	protected boolean wasPoweredBefore;

	protected GroupedItems groupedItemsBeforeCraft; // for rendering on client
	private InvManipulationBehaviour inserting;
	private EdgeInteractionBehaviour connectivity;

	private class_1799 scriptedResult = class_1799.field_8037;

	public MechanicalCrafterBlockEntity(class_2591<?> type, class_2338 pos, class_2680 state) {
		super(type, pos, state);
		setLazyTickRate(20);
		phase = Phase.IDLE;
		groupedItemsBeforeCraft = new GroupedItems();
		inventory = new Inventory(this);

		// Does not get serialized due to active checking in tick
		wasPoweredBefore = true;
	}

	@Override
	public void addBehaviours(List<BlockEntityBehaviour> behaviours) {
		super.addBehaviours(behaviours);
		inserting = new InvManipulationBehaviour(this, this::getTargetFace);
		connectivity = new EdgeInteractionBehaviour(this, ConnectedInputHandler::toggleConnection)
			.connectivity(ConnectedInputHandler::shouldConnect)
			.require(AllItems.WRENCH.get());
		behaviours.add(inserting);
		behaviours.add(connectivity);
		registerAwardables(behaviours, AllAdvancements.CRAFTER, AllAdvancements.CRAFTER_LAZY);
	}

	@Override
	public void onSpeedChanged(float previousSpeed) {
		super.onSpeedChanged(previousSpeed);
		if (!class_3532.method_15347(getSpeed(), 0)) {
			award(AllAdvancements.CRAFTER);
			if (Math.abs(getSpeed()) < 5)
				award(AllAdvancements.CRAFTER_LAZY);
		}
	}

	public void blockChanged() {
		removeBehaviour(InvManipulationBehaviour.TYPE);
		inserting = new InvManipulationBehaviour(this, this::getTargetFace);
		attachBehaviourLate(inserting);
	}

	public BlockFace getTargetFace(class_1937 world, class_2338 pos, class_2680 state) {
		return new BlockFace(pos, MechanicalCrafterBlock.getTargetDirection(state));
	}

	public class_2350 getTargetDirection() {
		return MechanicalCrafterBlock.getTargetDirection(method_11010());
	}

	@Override
	public void writeSafe(class_2487 compound) {
		super.writeSafe(compound);
		if (input == null)
			return;

		class_2487 inputNBT = new class_2487();
		input.write(inputNBT);
		compound.method_10566("ConnectedInput", inputNBT);
	}

	@Override
	public void write(class_2487 compound, boolean clientPacket) {
		compound.method_10566("Inventory", inventory.serializeNBT());

		class_2487 inputNBT = new class_2487();
		input.write(inputNBT);
		compound.method_10566("ConnectedInput", inputNBT);

		class_2487 groupedItemsNBT = new class_2487();
		groupedItems.write(groupedItemsNBT);
		compound.method_10566("GroupedItems", groupedItemsNBT);

		compound.method_10582("Phase", phase.name());
		compound.method_10569("CountDown", countDown);
		compound.method_10556("Cover", covered);

		super.write(compound, clientPacket);

		if (clientPacket && reRender) {
			compound.method_10556("Redraw", true);
			reRender = false;
		}
	}

	@Override
	protected void read(class_2487 compound, boolean clientPacket) {
		Phase phaseBefore = phase;
		GroupedItems before = this.groupedItems;

		inventory.deserializeNBT(compound.method_10562("Inventory"));
		input.read(compound.method_10562("ConnectedInput"));
		groupedItems = GroupedItems.read(compound.method_10562("GroupedItems"));
		phase = Phase.IDLE;
		String name = compound.method_10558("Phase");
		for (Phase phase : Phase.values())
			if (phase.name()
				.equals(name))
				this.phase = phase;
		countDown = compound.method_10550("CountDown");
		covered = compound.method_10577("Cover");
		super.read(compound, clientPacket);
		if (!clientPacket)
			return;
		if (compound.method_10545("Redraw"))
			field_11863.method_8413(method_11016(), method_11010(), method_11010(), 16);
		if (phaseBefore != phase && phase == Phase.CRAFTING)
			groupedItemsBeforeCraft = before;
		if (phaseBefore == Phase.EXPORTING && phase == Phase.WAITING) {
			class_2350 facing = method_11010().method_11654(MechanicalCrafterBlock.HORIZONTAL_FACING);
			class_243 vec = class_243.method_24954(facing.method_10163())
				.method_1021(.75)
				.method_1019(VecHelper.getCenterOf(field_11867));
			class_2350 targetDirection = MechanicalCrafterBlock.getTargetDirection(method_11010());
			vec = vec.method_1019(class_243.method_24954(targetDirection.method_10163())
				.method_1021(1));
			field_11863.method_8406(class_2398.field_11205, vec.field_1352, vec.field_1351, vec.field_1350, 0, 0, 0);
		}
	}

	@Override
	public void invalidate() {
		super.invalidate();
	}

	public int getCountDownSpeed() {
		if (getSpeed() == 0)
			return 0;
		return class_3532.method_15340((int) Math.abs(getSpeed()), 4, 250);
	}

	@Override
	public void tick() {
		super.tick();

		if (phase == Phase.ACCEPTING)
			return;

		boolean onClient = field_11863.field_9236;
		boolean runLogic = !onClient || isVirtual();

		if (wasPoweredBefore != field_11863.method_49803(field_11867)) {
			wasPoweredBefore = field_11863.method_49803(field_11867);
			if (wasPoweredBefore) {
				if (!runLogic)
					return;
				checkCompletedRecipe(true);
			}
		}

		if (phase == Phase.ASSEMBLING) {
			countDown -= getCountDownSpeed();
			if (countDown < 0) {
				countDown = 0;
				if (!runLogic)
					return;
				if (RecipeGridHandler.getTargetingCrafter(this) != null) {
					phase = Phase.EXPORTING;
					countDown = 1000;
					sendData();
					return;
				}

				class_1799 result =
					isVirtual() ? scriptedResult : RecipeGridHandler.tryToApplyRecipe(field_11863, groupedItems);

				if (result != null) {
					List<class_1799> containers = new ArrayList<>();
					groupedItems.grid.values()
						.forEach(stack -> {
							class_1799 remaining = stack.getRecipeRemainder();
							if (!remaining.method_7960())
								containers.add(remaining);
						});

					if (isVirtual())
						groupedItemsBeforeCraft = groupedItems;

					groupedItems = new GroupedItems(result);
					for (int i = 0; i < containers.size(); i++) {
						class_1799 stack = containers.get(i);
						GroupedItems container = new GroupedItems();
						container.grid.put(Pair.of(i, 0), stack);
						container.mergeOnto(groupedItems, Pointing.LEFT);
					}

					phase = Phase.CRAFTING;
					countDown = 2000;
					sendData();
					return;
				}
				ejectWholeGrid();
				return;
			}
		}

		if (phase == Phase.EXPORTING) {
			countDown -= getCountDownSpeed();

			if (countDown < 0) {
				countDown = 0;
				if (!runLogic)
					return;

				MechanicalCrafterBlockEntity targetingCrafter = RecipeGridHandler.getTargetingCrafter(this);
				if (targetingCrafter == null) {
					ejectWholeGrid();
					return;
				}

				Pointing pointing = method_11010().method_11654(MechanicalCrafterBlock.POINTING);
				groupedItems.mergeOnto(targetingCrafter.groupedItems, pointing);
				groupedItems = new GroupedItems();

				float pitch = targetingCrafter.groupedItems.grid.size() * 1 / 16f + .5f;
				AllSoundEvents.CRAFTER_CLICK.playOnServer(field_11863, field_11867, 1, pitch);

				phase = Phase.WAITING;
				countDown = 0;
				sendData();
				targetingCrafter.continueIfAllPrecedingFinished();
				targetingCrafter.sendData();
				return;
			}
		}

		if (phase == Phase.CRAFTING) {

			if (onClient) {
				class_2350 facing = method_11010().method_11654(MechanicalCrafterBlock.HORIZONTAL_FACING);
				float progress = countDown / 2000f;
				class_243 facingVec = class_243.method_24954(facing.method_10163());
				class_243 vec = facingVec.method_1021(.65)
					.method_1019(VecHelper.getCenterOf(field_11867));
				class_243 offset = VecHelper.offsetRandomly(class_243.field_1353, field_11863.field_9229, .125f)
					.method_18806(VecHelper.axisAlingedPlaneOf(facingVec))
					.method_1029()
					.method_1021(progress * .5f)
					.method_1019(vec);
				if (progress > .5f)
					field_11863.method_8406(class_2398.field_11205, offset.field_1352, offset.field_1351, offset.field_1350, 0, 0, 0);

				if (!groupedItemsBeforeCraft.grid.isEmpty() && progress < .5f) {
					if (groupedItems.grid.containsKey(Pair.of(0, 0))) {
						class_1799 stack = groupedItems.grid.get(Pair.of(0, 0));
						groupedItemsBeforeCraft = new GroupedItems();

						for (int i = 0; i < 10; i++) {
							class_243 randVec = VecHelper.offsetRandomly(class_243.field_1353, field_11863.field_9229, .125f)
								.method_18806(VecHelper.axisAlingedPlaneOf(facingVec))
								.method_1029()
								.method_1021(.25f);
							class_243 offset2 = randVec.method_1019(vec);
							randVec = randVec.method_1021(.35f);
							field_11863.method_8406(new class_2392(class_2398.field_11218, stack), offset2.field_1352, offset2.field_1351,
								offset2.field_1350, randVec.field_1352, randVec.field_1351, randVec.field_1350);
						}
					}
				}
			}

			int prev = countDown;
			countDown -= getCountDownSpeed();

			if (countDown < 1000 && prev >= 1000) {
				AllSoundEvents.CRAFTER_CLICK.playOnServer(field_11863, field_11867, 1, 2);
				AllSoundEvents.CRAFTER_CRAFT.playOnServer(field_11863, field_11867);
			}

			if (countDown < 0) {
				countDown = 0;
				if (!runLogic)
					return;
				tryInsert();
				return;
			}
		}

		if (phase == Phase.INSERTING) {
			if (runLogic && isTargetingBelt())
				tryInsert();
			return;
		}
	}

	protected boolean isTargetingBelt() {
		DirectBeltInputBehaviour behaviour = getTargetingBelt();
		return behaviour != null && behaviour.canInsertFromSide(getTargetDirection());
	}

	protected DirectBeltInputBehaviour getTargetingBelt() {
		class_2338 targetPos = field_11867.method_10093(getTargetDirection());
		return BlockEntityBehaviour.get(field_11863, targetPos, DirectBeltInputBehaviour.TYPE);
	}

	public void tryInsert() {
		if (!inserting.hasInventory() && !isTargetingBelt()) {
			ejectWholeGrid();
			return;
		}

		boolean chagedPhase = phase != Phase.INSERTING;
		final List<Pair<Integer, Integer>> inserted = new LinkedList<>();

		DirectBeltInputBehaviour behaviour = getTargetingBelt();
		for (Entry<Pair<Integer, Integer>, class_1799> entry : groupedItems.grid.entrySet()) {
			Pair<Integer, Integer> pair = entry.getKey();
			class_1799 stack = entry.getValue();
			BlockFace face = getTargetFace(field_11863, field_11867, method_11010());

			class_1799 remainder = behaviour == null ? inserting.insert(stack.method_7972())
				: behaviour.handleInsertion(stack, face.getFace(), false);
			if (!remainder.method_7960()) {
				stack.method_7939(remainder.method_7947());
				continue;
			}

			inserted.add(pair);
		}

		inserted.forEach(groupedItems.grid::remove);
		if (groupedItems.grid.isEmpty())
			ejectWholeGrid();
		else
			phase = Phase.INSERTING;
		if (!inserted.isEmpty() || chagedPhase)
			sendData();
	}

	public void ejectWholeGrid() {
		List<MechanicalCrafterBlockEntity> chain = RecipeGridHandler.getAllCraftersOfChain(this);
		if (chain == null)
			return;
		chain.forEach(MechanicalCrafterBlockEntity::eject);
	}

	public void eject() {
		class_2680 blockState = method_11010();
		boolean present = AllBlocks.MECHANICAL_CRAFTER.has(blockState);
		class_243 vec = present ? class_243.method_24954(blockState.method_11654(HORIZONTAL_FACING)
			.method_10163())
			.method_1021(.75f) : class_243.field_1353;
		class_243 ejectPos = VecHelper.getCenterOf(field_11867)
			.method_1019(vec);
		groupedItems.grid.forEach((pair, stack) -> dropItem(ejectPos, stack));
		if (!inventory.method_5438(0)
			.method_7960())
			dropItem(ejectPos, inventory.method_5438(0));
		phase = Phase.IDLE;
		groupedItems = new GroupedItems();
		inventory.setStackInSlot(0, class_1799.field_8037);
		sendData();
	}

	public void dropItem(class_243 ejectPos, class_1799 stack) {
		class_1542 itemEntity = new class_1542(field_11863, ejectPos.field_1352, ejectPos.field_1351, ejectPos.field_1350, stack);
		itemEntity.method_6988();
		field_11863.method_8649(itemEntity);
	}

	@Override
	public void lazyTick() {
		super.lazyTick();
		if (field_11863.field_9236 && !isVirtual())
			return;
		if (phase == Phase.IDLE && craftingItemPresent())
			checkCompletedRecipe(false);
		if (phase == Phase.INSERTING)
			tryInsert();
	}

	public boolean craftingItemPresent() {
		return !inventory.method_5438(0)
			.method_7960();
	}

	public boolean craftingItemOrCoverPresent() {
		return !inventory.method_5438(0)
			.method_7960() || covered;
	}

	protected void checkCompletedRecipe(boolean poweredStart) {
		if (getSpeed() == 0)
			return;
		if (field_11863.field_9236 && !isVirtual())
			return;
		List<MechanicalCrafterBlockEntity> chain = RecipeGridHandler.getAllCraftersOfChainIf(this,
			poweredStart ? MechanicalCrafterBlockEntity::craftingItemPresent
				: MechanicalCrafterBlockEntity::craftingItemOrCoverPresent,
			poweredStart);
		if (chain == null)
			return;
		chain.forEach(MechanicalCrafterBlockEntity::begin);
	}

	protected void begin() {
		phase = Phase.ACCEPTING;
		groupedItems = new GroupedItems(inventory.method_5438(0));
		inventory.setStackInSlot(0, class_1799.field_8037);
		if (RecipeGridHandler.getPrecedingCrafters(this)
			.isEmpty()) {
			phase = Phase.ASSEMBLING;
			countDown = 500;
		}
		sendData();
	}

	protected void continueIfAllPrecedingFinished() {
		List<MechanicalCrafterBlockEntity> preceding = RecipeGridHandler.getPrecedingCrafters(this);
		if (preceding == null) {
			ejectWholeGrid();
			return;
		}

		for (MechanicalCrafterBlockEntity blockEntity : preceding)
			if (blockEntity.phase != Phase.WAITING)
				return;

		phase = Phase.ASSEMBLING;
		countDown = Math.max(100, getCountDownSpeed() + 1);
	}

	@Nullable
	@Override
	public Storage<ItemVariant> getItemStorage(@Nullable class_2350 face) {
		return input.getItemHandler(field_11863, field_11867);
	}

	public void connectivityChanged() {
		reRender = true;
		sendData();
	}

	public Inventory getInventory() {
		return inventory;
	}

	public void setScriptedResult(class_1799 scriptedResult) {
		this.scriptedResult = scriptedResult;
	}

}
