package com.simibubi.create.content.kinetics.crafter;

import static com.simibubi.create.content.kinetics.base.HorizontalKineticBlock.HORIZONTAL_FACING;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.function.Predicate;
import net.minecraft.class_1799;
import net.minecraft.class_1851;
import net.minecraft.class_1937;
import net.minecraft.class_2338;
import net.minecraft.class_2350;
import net.minecraft.class_2487;
import net.minecraft.class_2499;
import net.minecraft.class_2520;
import net.minecraft.class_2680;
import net.minecraft.class_3955;
import net.minecraft.class_3956;
import net.minecraft.class_5455;
import net.minecraft.class_8566;
import org.apache.commons.lang3.tuple.Pair;

import com.google.common.base.Predicates;
import com.simibubi.create.AllBlocks;
import com.simibubi.create.AllRecipeTypes;
import com.simibubi.create.foundation.utility.Iterate;
import com.simibubi.create.foundation.utility.Pointing;
import com.simibubi.create.infrastructure.config.AllConfigs;

import io.github.fabricators_of_create.porting_lib.util.NBTSerializer;

public class RecipeGridHandler {

	public static List<MechanicalCrafterBlockEntity> getAllCraftersOfChain(MechanicalCrafterBlockEntity root) {
		return getAllCraftersOfChainIf(root, Predicates.alwaysTrue());
	}

	public static List<MechanicalCrafterBlockEntity> getAllCraftersOfChainIf(MechanicalCrafterBlockEntity root,
		Predicate<MechanicalCrafterBlockEntity> test) {
		return getAllCraftersOfChainIf(root, test, false);
	}

	public static List<MechanicalCrafterBlockEntity> getAllCraftersOfChainIf(MechanicalCrafterBlockEntity root,
		Predicate<MechanicalCrafterBlockEntity> test, boolean poweredStart) {
		List<MechanicalCrafterBlockEntity> crafters = new ArrayList<>();
		List<Pair<MechanicalCrafterBlockEntity, MechanicalCrafterBlockEntity>> frontier = new ArrayList<>();
		Set<MechanicalCrafterBlockEntity> visited = new HashSet<>();
		frontier.add(Pair.of(root, null));

		boolean powered = false;
		boolean empty = false;
		boolean allEmpty = true;

		while (!frontier.isEmpty()) {
			Pair<MechanicalCrafterBlockEntity, MechanicalCrafterBlockEntity> pair = frontier.remove(0);
			MechanicalCrafterBlockEntity current = pair.getKey();
			MechanicalCrafterBlockEntity last = pair.getValue();

			if (visited.contains(current))
				return null;
			if (!(test.test(current)))
				empty = true;
			else
				allEmpty = false;
			if (poweredStart && current.method_10997()
				.method_49803(current.method_11016()))
				powered = true;

			crafters.add(current);
			visited.add(current);

			MechanicalCrafterBlockEntity target = getTargetingCrafter(current);
			if (target != last && target != null)
				frontier.add(Pair.of(target, current));
			for (MechanicalCrafterBlockEntity preceding : getPrecedingCrafters(current))
				if (preceding != last)
					frontier.add(Pair.of(preceding, current));
		}

		return empty && !powered || allEmpty ? null : crafters;
	}

	public static MechanicalCrafterBlockEntity getTargetingCrafter(MechanicalCrafterBlockEntity crafter) {
		class_2680 state = crafter.method_11010();
		if (!isCrafter(state))
			return null;

		class_2338 targetPos = crafter.method_11016()
			.method_10093(MechanicalCrafterBlock.getTargetDirection(state));
		MechanicalCrafterBlockEntity targetBE = CrafterHelper.getCrafter(crafter.method_10997(), targetPos);
		if (targetBE == null)
			return null;

		class_2680 targetState = targetBE.method_11010();
		if (!isCrafter(targetState))
			return null;
		if (state.method_11654(HORIZONTAL_FACING) != targetState.method_11654(HORIZONTAL_FACING))
			return null;
		return targetBE;
	}

	public static List<MechanicalCrafterBlockEntity> getPrecedingCrafters(MechanicalCrafterBlockEntity crafter) {
		class_2338 pos = crafter.method_11016();
		class_1937 world = crafter.method_10997();
		List<MechanicalCrafterBlockEntity> crafters = new ArrayList<>();
		class_2680 blockState = crafter.method_11010();
		if (!isCrafter(blockState))
			return crafters;

		class_2350 blockFacing = blockState.method_11654(HORIZONTAL_FACING);
		class_2350 blockPointing = MechanicalCrafterBlock.getTargetDirection(blockState);
		for (class_2350 facing : Iterate.directions) {
			if (blockFacing.method_10166() == facing.method_10166())
				continue;
			if (blockPointing == facing)
				continue;

			class_2338 neighbourPos = pos.method_10093(facing);
			class_2680 neighbourState = world.method_8320(neighbourPos);
			if (!isCrafter(neighbourState))
				continue;
			if (MechanicalCrafterBlock.getTargetDirection(neighbourState) != facing.method_10153())
				continue;
			if (blockFacing != neighbourState.method_11654(HORIZONTAL_FACING))
				continue;
			MechanicalCrafterBlockEntity be = CrafterHelper.getCrafter(world, neighbourPos);
			if (be == null)
				continue;

			crafters.add(be);
		}

		return crafters;
	}

	private static boolean isCrafter(class_2680 state) {
		return AllBlocks.MECHANICAL_CRAFTER.has(state);
	}

	public static class_1799 tryToApplyRecipe(class_1937 world, GroupedItems items) {
		items.calcStats();
		class_8566 craftinginventory = new MechanicalCraftingInventory(items);
		class_1799 result = null;
		class_5455 registryAccess = world.method_30349();
		if (AllConfigs.server().recipes.allowRegularCraftingInCrafter.get())
			result = world.method_8433()
				.method_8132(class_3956.field_17545, craftinginventory, world)
				.filter(r -> isRecipeAllowed(r, craftinginventory))
				.map(r -> r.method_8116(craftinginventory, registryAccess))
				.orElse(null);
		if (result == null)
			result = AllRecipeTypes.MECHANICAL_CRAFTING.find(craftinginventory, world)
				.map(r -> r.method_8116(craftinginventory, registryAccess))
				.orElse(null);
		return result;
	}

	public static boolean isRecipeAllowed(class_3955 recipe, class_8566 inventory) {
		if (recipe instanceof class_1851) {
			int numItems = 0;
			for (int i = 0; i < inventory.method_5439(); i++) {
				if (!inventory.method_5438(i).method_7960()) {
					numItems++;
				}
			}
			if (numItems > AllConfigs.server().recipes.maxFireworkIngredientsInCrafter.get()) {
				return false;
			}
		}
		if (AllRecipeTypes.shouldIgnoreInAutomation(recipe))
			return false;
		return true;
	}

	public static class GroupedItems {
		Map<Pair<Integer, Integer>, class_1799> grid = new HashMap<>();
		int minX, minY, maxX, maxY, width, height;
		boolean statsReady;

		public GroupedItems() {}

		public GroupedItems(class_1799 stack) {
			grid.put(Pair.of(0, 0), stack);
		}

		public void mergeOnto(GroupedItems other, Pointing pointing) {
			int xOffset = pointing == Pointing.LEFT ? 1 : pointing == Pointing.RIGHT ? -1 : 0;
			int yOffset = pointing == Pointing.DOWN ? 1 : pointing == Pointing.UP ? -1 : 0;
			grid.forEach(
				(pair, stack) -> other.grid.put(Pair.of(pair.getKey() + xOffset, pair.getValue() + yOffset), stack));
			other.statsReady = false;
		}

		public void write(class_2487 nbt) {
			class_2499 gridNBT = new class_2499();
			grid.forEach((pair, stack) -> {
				class_2487 entry = new class_2487();
				entry.method_10569("x", pair.getKey());
				entry.method_10569("y", pair.getValue());
				entry.method_10566("item", NBTSerializer.serializeNBT(stack));
				gridNBT.add(entry);
			});
			nbt.method_10566("Grid", gridNBT);
		}

		public static GroupedItems read(class_2487 nbt) {
			GroupedItems items = new GroupedItems();
			class_2499 gridNBT = nbt.method_10554("Grid", class_2520.field_33260);
			gridNBT.forEach(inbt -> {
				class_2487 entry = (class_2487) inbt;
				int x = entry.method_10550("x");
				int y = entry.method_10550("y");
				class_1799 stack = class_1799.method_7915(entry.method_10562("item"));
				items.grid.put(Pair.of(x, y), stack);
			});
			return items;
		}

		public void calcStats() {
			if (statsReady)
				return;
			statsReady = true;

			minX = 0;
			minY = 0;
			maxX = 0;
			maxY = 0;

			for (Pair<Integer, Integer> pair : grid.keySet()) {
				int x = pair.getKey();
				int y = pair.getValue();
				minX = Math.min(minX, x);
				minY = Math.min(minY, y);
				maxX = Math.max(maxX, x);
				maxY = Math.max(maxY, y);
			}

			width = maxX - minX + 1;
			height = maxY - minY + 1;
		}

	}

}
