package com.simibubi.create.content.kinetics.crank;

import com.jozufozu.flywheel.api.Instancer;
import com.jozufozu.flywheel.api.Material;
import com.jozufozu.flywheel.core.materials.model.ModelData;
import com.simibubi.create.AllBlocks;
import com.simibubi.create.AllPartialModels;
import com.simibubi.create.AllSoundEvents;
import com.simibubi.create.content.kinetics.base.GeneratingKineticBlockEntity;
import com.simibubi.create.foundation.render.CachedBufferer;
import com.simibubi.create.foundation.render.SuperByteBuffer;
import com.simibubi.create.foundation.utility.AnimationTickHolder;

import net.fabricmc.api.EnvType;
import net.fabricmc.api.Environment;
import net.minecraft.class_2248;
import net.minecraft.class_2338;
import net.minecraft.class_2350;
import net.minecraft.class_2487;
import net.minecraft.class_2591;
import net.minecraft.class_2680;

public class HandCrankBlockEntity extends GeneratingKineticBlockEntity {

	public int inUse;
	public boolean backwards;
	public float independentAngle;
	public float chasingVelocity;

	public HandCrankBlockEntity(class_2591<?> type, class_2338 pos, class_2680 state) {
		super(type, pos, state);
	}

	public void turn(boolean back) {
		boolean update = false;

		if (getGeneratedSpeed() == 0 || back != backwards)
			update = true;

		inUse = 10;
		this.backwards = back;
		if (update && !field_11863.field_9236)
			updateGeneratedRotation();
	}

	public float getIndependentAngle(float partialTicks) {
		return (independentAngle + partialTicks * chasingVelocity) / 360;
	}

	@Override
	public float getGeneratedSpeed() {
		class_2248 block = method_11010().method_26204();
		if (!(block instanceof HandCrankBlock))
			return 0;
		HandCrankBlock crank = (HandCrankBlock) block;
		int speed = (inUse == 0 ? 0 : clockwise() ? -1 : 1) * crank.getRotationSpeed();
		return convertToDirection(speed, method_11010().method_11654(HandCrankBlock.FACING));
	}

	protected boolean clockwise() {
		return backwards;
	}

	@Override
	public void write(class_2487 compound, boolean clientPacket) {
		compound.method_10569("InUse", inUse);
		compound.method_10556("Backwards", backwards);
		super.write(compound, clientPacket);
	}

	@Override
	protected void read(class_2487 compound, boolean clientPacket) {
		inUse = compound.method_10550("InUse");
		backwards = compound.method_10577("Backwards");
		super.read(compound, clientPacket);
	}

	@Override
	public void tick() {
		super.tick();

		float actualSpeed = getSpeed();
		chasingVelocity += ((actualSpeed * 10 / 3f) - chasingVelocity) * .25f;
		independentAngle += chasingVelocity;

		if (inUse > 0) {
			inUse--;

			if (inUse == 0 && !field_11863.field_9236) {
				sequenceContext = null;
				updateGeneratedRotation();
			}
		}
	}

	@Environment(EnvType.CLIENT)
	public SuperByteBuffer getRenderedHandle() {
		class_2680 blockState = method_11010();
		class_2350 facing = blockState.method_28500(HandCrankBlock.FACING)
			.orElse(class_2350.field_11036);
		return CachedBufferer.partialFacing(AllPartialModels.HAND_CRANK_HANDLE, blockState, facing.method_10153());
	}

	@Environment(EnvType.CLIENT)
	public Instancer<ModelData> getRenderedHandleInstance(Material<ModelData> material) {
		class_2680 blockState = method_11010();
		class_2350 facing = blockState.method_28500(HandCrankBlock.FACING)
			.orElse(class_2350.field_11036);
		return material.getModel(AllPartialModels.HAND_CRANK_HANDLE, blockState, facing.method_10153());
	}

	@Environment(EnvType.CLIENT)
	public boolean shouldRenderShaft() {
		return true;
	}

	@Override
	protected class_2248 getStressConfigKey() {
		return AllBlocks.HAND_CRANK.has(method_11010()) ? AllBlocks.HAND_CRANK.get()
			: AllBlocks.COPPER_VALVE_HANDLE.get();
	}

	@Override
	@Environment(EnvType.CLIENT)
	public void tickAudio() {
		super.tickAudio();
		if (inUse > 0 && AnimationTickHolder.getTicks() % 10 == 0) {
			if (!AllBlocks.HAND_CRANK.has(method_11010()))
				return;
			AllSoundEvents.CRANKING.playAt(field_11863, field_11867, (inUse) / 2.5f, .65f + (10 - inUse) / 10f, true);
		}
	}

}
