package com.simibubi.create.content.kinetics.deployer;

import static com.simibubi.create.content.kinetics.base.DirectionalKineticBlock.FACING;

import java.util.List;
import java.util.stream.Collectors;
import net.minecraft.class_1268;
import net.minecraft.class_1799;
import net.minecraft.class_1860;
import net.minecraft.class_1937;
import net.minecraft.class_2338;
import net.minecraft.class_2350;
import net.minecraft.class_2680;
import net.minecraft.class_3417;
import net.minecraft.class_3419;
import com.simibubi.create.AllBlocks;
import com.simibubi.create.AllSoundEvents;
import com.simibubi.create.Create;
import com.simibubi.create.compat.Mods;
import com.simibubi.create.compat.sandwichable.SequencedSandwiching;
import com.simibubi.create.content.equipment.sandPaper.SandPaperPolishingRecipe;
import com.simibubi.create.content.kinetics.belt.BeltHelper;
import com.simibubi.create.content.kinetics.belt.behaviour.BeltProcessingBehaviour.ProcessingResult;
import com.simibubi.create.content.kinetics.belt.behaviour.TransportedItemStackHandlerBehaviour;
import com.simibubi.create.content.kinetics.belt.behaviour.TransportedItemStackHandlerBehaviour.TransportedResult;
import com.simibubi.create.content.kinetics.belt.transport.TransportedItemStack;
import com.simibubi.create.content.kinetics.deployer.DeployerBlockEntity.Mode;
import com.simibubi.create.content.kinetics.deployer.DeployerBlockEntity.State;
import com.simibubi.create.foundation.advancement.AllAdvancements;
import com.simibubi.create.foundation.advancement.CreateAdvancement;
import com.simibubi.create.foundation.recipe.RecipeApplier;
import io.github.fabricators_of_create.porting_lib.transfer.item.ItemHandlerHelper;

public class BeltDeployerCallbacks {

	public static ProcessingResult onItemReceived(TransportedItemStack s, TransportedItemStackHandlerBehaviour i,
		DeployerBlockEntity blockEntity) {

		if (blockEntity.getSpeed() == 0)
			return ProcessingResult.PASS;
		if (blockEntity.mode == Mode.PUNCH)
			return ProcessingResult.PASS;
		class_2680 blockState = blockEntity.method_11010();
		if (!blockState.method_28498(FACING) || blockState.method_11654(FACING) != class_2350.field_11033)
			return ProcessingResult.PASS;
		if (blockEntity.state != State.WAITING)
			return ProcessingResult.HOLD;
		if (blockEntity.redstoneLocked)
			return ProcessingResult.PASS;

		DeployerFakePlayer player = blockEntity.getPlayer();
		class_1799 held = player == null ? class_1799.field_8037 : player.method_6047();

		if (held.method_7960())
			return ProcessingResult.HOLD;
		if (blockEntity.getRecipe(s.stack) == null) {
			if (Mods.SANDWICHABLE.isLoaded()) {
				if (!SequencedSandwiching.shouldSandwich(s.stack, held, blockEntity.method_10997()))
					return ProcessingResult.PASS;
			} else {
				return ProcessingResult.PASS;
			}
		}

		blockEntity.start();
		return ProcessingResult.HOLD;
	}

	public static ProcessingResult whenItemHeld(TransportedItemStack s, TransportedItemStackHandlerBehaviour i,
		DeployerBlockEntity blockEntity) {

		if (blockEntity.getSpeed() == 0)
			return ProcessingResult.PASS;
		class_2680 blockState = blockEntity.method_11010();
		if (!blockState.method_28498(FACING) || blockState.method_11654(FACING) != class_2350.field_11033)
			return ProcessingResult.PASS;

		DeployerFakePlayer player = blockEntity.getPlayer();
		class_1799 held = player == null ? class_1799.field_8037 : player.method_6047();
		if (held.method_7960())
			return ProcessingResult.HOLD;

		class_1860<?> recipe = blockEntity.getRecipe(s.stack);
		boolean shouldSandwich = Mods.SANDWICHABLE.isLoaded() && SequencedSandwiching.shouldSandwich(s.stack, held, blockEntity.method_10997());
		if (recipe == null && !shouldSandwich) {
			return ProcessingResult.PASS;
		}

		if (blockEntity.state == State.RETRACTING && blockEntity.timer == 1000) {
			if (recipe != null)
				activate(s, i, blockEntity, recipe);
			else
				SequencedSandwiching.activateSandwich(s, i, blockEntity);
			return ProcessingResult.HOLD;
		}

		if (blockEntity.state == State.WAITING) {
			if (blockEntity.redstoneLocked)
				return ProcessingResult.PASS;
			blockEntity.start();
		}

		return ProcessingResult.HOLD;
	}

	public static void activate(TransportedItemStack transported, TransportedItemStackHandlerBehaviour handler,
		DeployerBlockEntity blockEntity, class_1860<?> recipe) {
		List<TransportedItemStack> collect =
			RecipeApplier.applyRecipeOn(blockEntity.method_10997(), ItemHandlerHelper.copyStackWithSize(transported.stack, 1), recipe)
				.stream()
				.map(stack -> {
					TransportedItemStack copy = transported.copy();
					boolean centered = BeltHelper.isItemUpright(stack);
					copy.stack = stack;
					copy.locked = true;
					copy.angle = centered ? 180 : Create.RANDOM.nextInt(360);
					return copy;
				})
				.map(t -> {
					t.locked = false;
					return t;
				})
				.collect(Collectors.toList());

		blockEntity.award(AllAdvancements.DEPLOYER);
		TransportedItemStack left = transported.copy();
		blockEntity.player.spawnedItemEffects = transported.stack.method_7972();
		left.stack.method_7934(1);
		class_1799 resultItem = null;

		if (collect.isEmpty()) {
			resultItem = left.stack.method_7972();
			handler.handleProcessingOnItem(transported, TransportedResult.convertTo(left));
		} else {
			resultItem = collect.get(0).stack.method_7972();
			handler.handleProcessingOnItem(transported, TransportedResult.convertToAndLeaveHeld(collect, left));
		}

		class_1799 heldItem = blockEntity.player.method_6047();
		boolean unbreakable = heldItem.method_7985() && heldItem.method_7969()
			.method_10577("Unbreakable");
		boolean keepHeld =
			recipe instanceof ItemApplicationRecipe && ((ItemApplicationRecipe) recipe).shouldKeepHeldItem();

		if (!unbreakable && !keepHeld) {
			if (heldItem.method_7963())
				heldItem.method_7956(1, blockEntity.player,
					s -> s.method_20236(class_1268.field_5808));
			else
				heldItem.method_7934(1);
		}

		if (resultItem != null && !resultItem.method_7960())
			awardAdvancements(blockEntity, resultItem);

		class_2338 pos = blockEntity.method_11016();
		class_1937 world = blockEntity.method_10997();
		if (heldItem.method_7960())
			world.method_8396(null, pos, class_3417.field_15075, class_3419.field_15245, .25f, 1);
		world.method_8396(null, pos, class_3417.field_15197, class_3419.field_15245, .25f, .75f);
		if (recipe instanceof SandPaperPolishingRecipe)
			AllSoundEvents.SANDING_SHORT.playOnServer(world, pos, .35f, 1f);

		blockEntity.sendData();
	}

	private static void awardAdvancements(DeployerBlockEntity blockEntity, class_1799 created) {
		CreateAdvancement advancement = null;

		if (AllBlocks.ANDESITE_CASING.isIn(created))
			advancement = AllAdvancements.ANDESITE_CASING;
		else if (AllBlocks.BRASS_CASING.isIn(created))
			advancement = AllAdvancements.BRASS_CASING;
		else if (AllBlocks.COPPER_CASING.isIn(created))
			advancement = AllAdvancements.COPPER_CASING;
		else if (AllBlocks.RAILWAY_CASING.isIn(created))
			advancement = AllAdvancements.TRAIN_CASING;
		else
			return;

		blockEntity.award(advancement);
	}

}
