package com.simibubi.create.content.kinetics.deployer;

import static com.simibubi.create.content.kinetics.base.DirectionalAxisKineticBlock.AXIS_ALONG_FIRST_COORDINATE;
import static com.simibubi.create.content.kinetics.base.DirectionalKineticBlock.FACING;

import com.jozufozu.flywheel.api.Material;
import com.jozufozu.flywheel.api.MaterialManager;
import com.jozufozu.flywheel.core.Materials;
import com.jozufozu.flywheel.core.PartialModel;
import com.jozufozu.flywheel.core.materials.model.ModelData;
import com.jozufozu.flywheel.core.virtual.VirtualRenderWorld;
import com.jozufozu.flywheel.util.transform.TransformStack;
import com.simibubi.create.AllPartialModels;
import com.simibubi.create.content.contraptions.behaviour.MovementContext;
import com.simibubi.create.content.contraptions.render.ActorInstance;
import com.simibubi.create.content.kinetics.base.IRotate;
import com.simibubi.create.content.kinetics.base.KineticBlockEntityInstance;
import com.simibubi.create.content.kinetics.base.flwdata.RotatingData;
import com.simibubi.create.foundation.render.AllMaterialSpecs;
import com.simibubi.create.foundation.utility.AngleHelper;
import com.simibubi.create.foundation.utility.AnimationTickHolder;
import com.simibubi.create.foundation.utility.NBTHelper;
import com.simibubi.create.foundation.utility.VecHelper;
import net.minecraft.class_2338;
import net.minecraft.class_2350;
import net.minecraft.class_243;
import net.minecraft.class_2680;
import net.minecraft.class_3532;
import net.minecraft.class_4587;

public class DeployerActorInstance extends ActorInstance {

	private final class_4587 stack = new class_4587();
	class_2350 facing;
    boolean stationaryTimer;

    float yRot;
    float xRot;
    float zRot;

    ModelData pole;
    ModelData hand;
    RotatingData shaft;

	public DeployerActorInstance(MaterialManager materialManager, VirtualRenderWorld simulationWorld, MovementContext context) {
        super(materialManager, simulationWorld, context);

		Material<ModelData> mat = materialManager.defaultSolid()
				.material(Materials.TRANSFORMED);

        class_2680 state = context.state;
        DeployerBlockEntity.Mode mode = NBTHelper.readEnum(context.blockEntityData, "Mode", DeployerBlockEntity.Mode.class);
        PartialModel handPose = DeployerRenderer.getHandPose(mode);

        stationaryTimer = context.data.method_10545("StationaryTimer");
        facing = state.method_11654(FACING);

        boolean rotatePole = state.method_11654(AXIS_ALONG_FIRST_COORDINATE) ^ facing.method_10166() == class_2350.class_2351.field_11051;
        yRot = AngleHelper.horizontalAngle(facing);
        xRot = facing == class_2350.field_11036 ? 270 : facing == class_2350.field_11033 ? 90 : 0;
        zRot = rotatePole ? 90 : 0;

        pole = mat.getModel(AllPartialModels.DEPLOYER_POLE, state).createInstance();
        hand = mat.getModel(handPose, state).createInstance();

        class_2350.class_2351 axis = ((IRotate) state.method_26204()).getRotationAxis(state);
        shaft = materialManager.defaultSolid()
                .material(AllMaterialSpecs.ROTATING)
				.getModel(KineticBlockEntityInstance.shaft(axis))
				.createInstance();

        int blockLight = localBlockLight();

        shaft.setRotationAxis(axis)
                .setPosition(context.localPos)
                .setBlockLight(blockLight);

        pole.setBlockLight(blockLight);
        hand.setBlockLight(blockLight);
    }

    @Override
    public void beginFrame() {
        double factor;
        if (context.disabled) {
        	factor = 0;
        } else if (context.contraption.stalled || context.position == null || context.data.method_10545("StationaryTimer")) {
            factor = class_3532.method_15374(AnimationTickHolder.getRenderTime() * .5f) * .25f + .25f;
        } else {
        	class_243 center = VecHelper.getCenterOf(class_2338.method_49638(context.position));
            double distance = context.position.method_1022(center);
            double nextDistance = context.position.method_1019(context.motion)
                                                  .method_1022(center);
            factor = .5f - class_3532.method_15350(class_3532.method_16436(AnimationTickHolder.getPartialTicks(), distance, nextDistance), 0, 1);
        }

        class_243 offset = class_243.method_24954(facing.method_10163()).method_1021(factor);

        TransformStack tstack = TransformStack.cast(stack);
        stack.method_34426();
        tstack.translate(context.localPos)
				.translate(offset);

        transformModel(stack, pole, hand, yRot, xRot, zRot);
    }

    static void transformModel(class_4587 stack, ModelData pole, ModelData hand, float yRot, float xRot, float zRot) {
        TransformStack tstack = TransformStack.cast(stack);

        tstack.centre();
        tstack.rotate(class_2350.field_11036, (float) ((yRot) / 180 * Math.PI));
        tstack.rotate(class_2350.field_11034, (float) ((xRot) / 180 * Math.PI));

        stack.method_22903();
        tstack.rotate(class_2350.field_11035, (float) ((zRot) / 180 * Math.PI));
        tstack.unCentre();
        pole.setTransform(stack);
        stack.method_22909();

        tstack.unCentre();

        hand.setTransform(stack);
    }
}
