package com.simibubi.create.content.kinetics.deployer;

import static com.simibubi.create.content.kinetics.base.DirectionalKineticBlock.FACING;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.UUID;

import javax.annotation.Nullable;

import com.jozufozu.flywheel.core.PartialModel;
import com.simibubi.create.AllBlocks;
import com.simibubi.create.AllPartialModels;
import com.simibubi.create.AllRecipeTypes;
import com.simibubi.create.content.equipment.sandPaper.SandPaperItem;
import com.simibubi.create.content.equipment.sandPaper.SandPaperPolishingRecipe.SandPaperInv;
import com.simibubi.create.content.kinetics.base.IRotate.StressImpact;
import com.simibubi.create.content.kinetics.base.KineticBlockEntity;
import com.simibubi.create.content.kinetics.belt.behaviour.BeltProcessingBehaviour;
import com.simibubi.create.content.kinetics.belt.behaviour.TransportedItemStackHandlerBehaviour;
import com.simibubi.create.content.processing.sequenced.SequencedAssemblyRecipe;
import com.simibubi.create.foundation.advancement.AllAdvancements;
import com.simibubi.create.foundation.blockEntity.behaviour.BlockEntityBehaviour;
import com.simibubi.create.foundation.blockEntity.behaviour.filtering.FilteringBehaviour;
import com.simibubi.create.foundation.item.TooltipHelper;
import com.simibubi.create.foundation.utility.Components;
import com.simibubi.create.foundation.utility.Lang;
import com.simibubi.create.foundation.utility.NBTHelper;
import com.simibubi.create.foundation.utility.VecHelper;
import com.simibubi.create.foundation.utility.animation.LerpedFloat;

import io.github.fabricators_of_create.porting_lib.transfer.item.ItemStackHandlerContainer;
import io.github.fabricators_of_create.porting_lib.util.NBTSerializer;
import net.fabricmc.api.EnvType;
import net.fabricmc.api.Environment;
import net.fabricmc.fabric.api.transfer.v1.item.ItemVariant;
import net.fabricmc.fabric.api.transfer.v1.storage.Storage;
import net.fabricmc.fabric.api.transfer.v1.storage.base.SidedStorageBlockEntity;
import net.fabricmc.fabric.api.transfer.v1.transaction.base.SnapshotParticipant;
import net.minecraft.class_124;
import net.minecraft.class_1263;
import net.minecraft.class_1268;
import net.minecraft.class_1661;
import net.minecraft.class_1799;
import net.minecraft.class_1860;
import net.minecraft.class_2338;
import net.minecraft.class_2350;
import net.minecraft.class_238;
import net.minecraft.class_243;
import net.minecraft.class_2487;
import net.minecraft.class_2499;
import net.minecraft.class_2520;
import net.minecraft.class_2561;
import net.minecraft.class_2586;
import net.minecraft.class_2591;
import net.minecraft.class_2680;
import net.minecraft.class_3218;
import net.minecraft.class_3417;
import net.minecraft.class_3419;
import net.minecraft.class_3532;
import net.minecraft.class_3959;
import net.minecraft.class_3959.class_242;
import net.minecraft.class_3959.class_3960;
import net.minecraft.class_3965;

public class DeployerBlockEntity extends KineticBlockEntity implements SidedStorageBlockEntity {

	protected State state;
	protected Mode mode;
	protected class_1799 heldItem = class_1799.field_8037;
	protected DeployerFakePlayer player;
	protected int timer;
	protected float reach;
	protected boolean fistBump = false;
	protected List<class_1799> overflowItems = new ArrayList<>();
	protected FilteringBehaviour filtering;
	protected boolean redstoneLocked;
	protected UUID owner;
	private DeployerItemHandler invHandler;
	private class_2499 deferredInventoryList;

	private LerpedFloat animatedOffset;

	public BeltProcessingBehaviour processingBehaviour;

	// list holds the held item in index 0, followed by overflow items
	public SnapshotParticipant<List<class_1799>> snapshotParticipant = new SnapshotParticipant<>() {
		@Override
		protected List<class_1799> createSnapshot() {
			List<class_1799> stacks = new ArrayList<>();
			stacks.add(player == null ? class_1799.field_8037 : player.method_6047());
			stacks.addAll(overflowItems);
			return stacks;
		}

		@Override
		protected void readSnapshot(List<class_1799> snapshot) {
			class_1799 held = snapshot.remove(0);
			if (player != null)
				player.method_6122(class_1268.field_5808, held);
			overflowItems = snapshot;
		}
	};

	enum State {
		WAITING, EXPANDING, RETRACTING, DUMPING;
	}

	enum Mode {
		PUNCH, USE
	}

	public DeployerBlockEntity(class_2591<?> type, class_2338 pos, class_2680 state) {
		super(type, pos, state);
		this.state = State.WAITING;
		mode = Mode.USE;
		heldItem = class_1799.field_8037;
		redstoneLocked = false;
		animatedOffset = LerpedFloat.linear()
			.startWithValue(0);
	}

	@Override
	public void addBehaviours(List<BlockEntityBehaviour> behaviours) {
		super.addBehaviours(behaviours);
		filtering = new FilteringBehaviour(this, new DeployerFilterSlot());
		behaviours.add(filtering);
		processingBehaviour =
			new BeltProcessingBehaviour(this).whenItemEnters((s, i) -> BeltDeployerCallbacks.onItemReceived(s, i, this))
				.whileItemHeld((s, i) -> BeltDeployerCallbacks.whenItemHeld(s, i, this));
		behaviours.add(processingBehaviour);

		registerAwardables(behaviours, AllAdvancements.TRAIN_CASING, AllAdvancements.ANDESITE_CASING,
			AllAdvancements.BRASS_CASING, AllAdvancements.COPPER_CASING, AllAdvancements.FIST_BUMP,
			AllAdvancements.DEPLOYER, AllAdvancements.SELF_DEPLOYING);
	}

	@Override
	public void initialize() {
		super.initialize();
		initHandler();
	}

	private void initHandler() {
		if (invHandler != null)
			return;
		if (field_11863 instanceof class_3218 sLevel) {
			player = new DeployerFakePlayer(sLevel, owner);
			if (deferredInventoryList != null) {
				player.method_31548()
					.method_7397(deferredInventoryList);
				deferredInventoryList = null;
				heldItem = player.method_6047();
				sendData();
			}
			class_243 initialPos = VecHelper.getCenterOf(field_11867.method_10093(method_11010().method_11654(FACING)));
			player.method_5814(initialPos.field_1352, initialPos.field_1351, initialPos.field_1350);
		}
		invHandler = this.createHandler();
	}

	protected void onExtract(class_1799 stack) {
		player.method_6122(class_1268.field_5808, stack.method_7972());
		sendData();
		method_5431();
	}

	protected int getTimerSpeed() {
		return (int) (getSpeed() == 0 ? 0 : class_3532.method_15363(Math.abs(getSpeed() * 2), 8, 512));
	}

	@Override
	public void tick() {
		super.tick();

		if (!overflowItems.isEmpty()) {
			overflowItems.removeIf(class_1799::method_7960);
		}

		if (getSpeed() == 0)
			return;
		if (!field_11863.field_9236 && player != null && player.blockBreakingProgress != null) {
			if (field_11863.method_22347(player.blockBreakingProgress.getKey())) {
				field_11863.method_8517(player.method_5628(), player.blockBreakingProgress.getKey(), -1);
				player.blockBreakingProgress = null;
			}
		}
		if (timer > 0) {
			timer -= getTimerSpeed();
			return;
		}
		if (field_11863.field_9236)
			return;

		class_1799 stack = player.method_6047();
		if (state == State.WAITING) {
			if (!overflowItems.isEmpty()) {
				timer = getTimerSpeed() * 10;
				return;
			}

			boolean changed = false;
			class_1661 inventory = player.method_31548();
			for (int i = 0; i < inventory.method_5439(); i++) {
				if (overflowItems.size() > 10)
					break;
				class_1799 item = inventory.method_5438(i);
				if (item.method_7960())
					continue;
				if (item != stack || !filtering.test(item)) {
					overflowItems.add(item);
					inventory.method_5447(i, class_1799.field_8037);
					changed = true;
				}
			}

			if (changed) {
				sendData();
				timer = getTimerSpeed() * 10;
				return;
			}

			class_2350 facing = method_11010().method_11654(FACING);
			if (mode == Mode.USE
				&& !DeployerHandler.shouldActivate(stack, field_11863, field_11867.method_10079(facing, 2), facing)) {
				timer = getTimerSpeed() * 10;
				return;
			}

			// Check for advancement conditions
			if (mode == Mode.PUNCH && !fistBump && startFistBump(facing))
				return;
			if (redstoneLocked)
				return;

			start();
			return;
		}

		if (state == State.EXPANDING) {
			if (fistBump)
				triggerFistBump();
			activate();

			state = State.RETRACTING;
			timer = 1000;
			sendData();
			return;
		}

		if (state == State.RETRACTING) {
			state = State.WAITING;
			timer = 500;
			sendData();
			return;
		}

	}

	protected void start() {
		state = State.EXPANDING;
		class_243 movementVector = getMovementVector();
		class_243 rayOrigin = VecHelper.getCenterOf(field_11867)
			.method_1019(movementVector.method_1021(3 / 2f));
		class_243 rayTarget = VecHelper.getCenterOf(field_11867)
			.method_1019(movementVector.method_1021(5 / 2f));
		class_3959 rayTraceContext = new class_3959(rayOrigin, rayTarget, class_3960.field_17559, class_242.field_1348, player);
		class_3965 result = field_11863.method_17742(rayTraceContext);
		reach = (float) (.5f + Math.min(result.method_17784()
			.method_1020(rayOrigin)
			.method_1033(), .75f));
		timer = 1000;
		sendData();
	}

	public boolean startFistBump(class_2350 facing) {
		int i = 0;
		DeployerBlockEntity partner = null;

		for (i = 2; i < 5; i++) {
			class_2338 otherDeployer = field_11867.method_10079(facing, i);
			if (!field_11863.method_8477(otherDeployer))
				return false;
			class_2586 other = field_11863.method_8321(otherDeployer);
			if (other instanceof DeployerBlockEntity dpe) {
				partner = dpe;
				break;
			}
		}

		if (partner == null)
			return false;

		if (field_11863.method_8320(partner.method_11016())
			.method_11654(FACING)
			.method_10153() != facing || partner.mode != Mode.PUNCH)
			return false;
		if (partner.getSpeed() == 0)
			return false;

		for (DeployerBlockEntity be : Arrays.asList(this, partner)) {
			be.fistBump = true;
			be.reach = ((i - 2)) * .5f;
			be.timer = 1000;
			be.state = State.EXPANDING;
			be.sendData();
		}

		return true;
	}

	public void triggerFistBump() {
		int i = 0;
		DeployerBlockEntity deployerBlockEntity = null;
		for (i = 2; i < 5; i++) {
			class_2338 pos = field_11867.method_10079(method_11010().method_11654(FACING), i);
			if (!field_11863.method_8477(pos))
				return;
			if (field_11863.method_8321(pos) instanceof DeployerBlockEntity dpe) {
				deployerBlockEntity = dpe;
				break;
			}
		}

		if (deployerBlockEntity == null)
			return;
		if (!deployerBlockEntity.fistBump || deployerBlockEntity.state != State.EXPANDING)
			return;
		if (deployerBlockEntity.timer > 0)
			return;

		fistBump = false;
		deployerBlockEntity.fistBump = false;
		deployerBlockEntity.state = State.RETRACTING;
		deployerBlockEntity.timer = 1000;
		deployerBlockEntity.sendData();
		award(AllAdvancements.FIST_BUMP);

		class_2338 soundLocation = class_2338.method_49638(class_243.method_24953(field_11867)
			.method_1019(class_243.method_24953(deployerBlockEntity.method_11016()))
			.method_1021(.5f));
		field_11863.method_8396(null, soundLocation, class_3417.field_14914, class_3419.field_15245, .75f, .75f);
	}

	protected void activate() {
		class_243 movementVector = getMovementVector();
		class_2350 direction = method_11010().method_11654(FACING);
		class_243 center = VecHelper.getCenterOf(field_11867);
		class_2338 clickedPos = field_11867.method_10079(direction, 2);
		player.method_36457(direction == class_2350.field_11036 ? -90 : direction == class_2350.field_11033 ? 90 : 0);
		player.method_36456(direction.method_10144());

		if (direction == class_2350.field_11033
			&& BlockEntityBehaviour.get(field_11863, clickedPos, TransportedItemStackHandlerBehaviour.TYPE) != null)
			return; // Belt processing handled in BeltDeployerCallbacks

		DeployerHandler.activate(player, center, clickedPos, movementVector, mode);
		award(AllAdvancements.DEPLOYER);

		if (player != null)
			heldItem = player.method_6047();
	}

	protected class_243 getMovementVector() {
		if (!AllBlocks.DEPLOYER.has(method_11010()))
			return class_243.field_1353;
		return class_243.method_24954(method_11010().method_11654(FACING)
			.method_10163());
	}

	@Override
	protected void read(class_2487 compound, boolean clientPacket) {
		state = NBTHelper.readEnum(compound, "State", State.class);
		mode = NBTHelper.readEnum(compound, "Mode", Mode.class);
		timer = compound.method_10550("Timer");
		redstoneLocked = compound.method_10577("Powered");
		if (compound.method_10545("Owner"))
			owner = compound.method_25926("Owner");

		deferredInventoryList = compound.method_10554("Inventory", class_2520.field_33260);
		overflowItems = NBTHelper.readItemList(compound.method_10554("Overflow", class_2520.field_33260));
		if (compound.method_10545("HeldItem"))
			heldItem = class_1799.method_7915(compound.method_10562("HeldItem"));
		super.read(compound, clientPacket);

		if (!clientPacket)
			return;
		fistBump = compound.method_10577("Fistbump");
		reach = compound.method_10583("Reach");
		if (compound.method_10545("Particle")) {
			class_1799 particleStack = class_1799.method_7915(compound.method_10562("Particle"));
			SandPaperItem.spawnParticles(VecHelper.getCenterOf(field_11867)
				.method_1019(getMovementVector().method_1021(reach + 1)), particleStack, this.field_11863);
		}
	}

	@Override
	public void write(class_2487 compound, boolean clientPacket) {
		NBTHelper.writeEnum(compound, "Mode", mode);
		NBTHelper.writeEnum(compound, "State", state);
		compound.method_10569("Timer", timer);
		compound.method_10556("Powered", redstoneLocked);
		if (owner != null)
			compound.method_25927("Owner", owner);

		if (player != null) {
			class_2499 invNBT = new class_2499();
			player.method_31548()
				.method_7384(invNBT);
			compound.method_10566("Inventory", invNBT);
			compound.method_10566("HeldItem", NBTSerializer.serializeNBT(player.method_6047()));
			compound.method_10566("Overflow", NBTHelper.writeItemList(overflowItems));
		} else if (deferredInventoryList != null) {
			compound.method_10566("Inventory", deferredInventoryList);
		}

		super.write(compound, clientPacket);

		if (!clientPacket)
			return;
		compound.method_10556("Fistbump", fistBump);
		compound.method_10548("Reach", reach);
		if (player == null)
			return;
		compound.method_10566("HeldItem", NBTSerializer.serializeNBT(player.method_6047()));
		if (player.spawnedItemEffects != null) {
			compound.method_10566("Particle", NBTSerializer.serializeNBT(player.spawnedItemEffects));
			player.spawnedItemEffects = null;
		}
	}

	@Override
	public void writeSafe(class_2487 tag) {
		NBTHelper.writeEnum(tag, "Mode", mode);
		super.writeSafe(tag);
	}

	private DeployerItemHandler createHandler() {
		return new DeployerItemHandler(this);
	}

	public void redstoneUpdate() {
		if (field_11863.field_9236)
			return;
		boolean blockPowered = field_11863.method_49803(field_11867);
		if (blockPowered == redstoneLocked)
			return;
		redstoneLocked = blockPowered;
		sendData();
	}

	@Environment(EnvType.CLIENT)
	public PartialModel getHandPose() {
		return mode == Mode.PUNCH ? AllPartialModels.DEPLOYER_HAND_PUNCHING
			: heldItem.method_7960() ? AllPartialModels.DEPLOYER_HAND_POINTING : AllPartialModels.DEPLOYER_HAND_HOLDING;
	}

	@Override
	protected class_238 createRenderBoundingBox() {
		return super.createRenderBoundingBox().method_1014(3);
	}

	public void discardPlayer() {
		if (player == null)
			return;
		player.method_31548()
			.method_7388();
		overflowItems.forEach(itemstack -> player.method_7329(itemstack, true, false));
		player.method_31472();
		player = null;
	}

	@Override
	public void invalidate() {
		super.invalidate();
	}

	public void changeMode() {
		mode = mode == Mode.PUNCH ? Mode.USE : Mode.PUNCH;
		method_5431();
		sendData();
	}

	@Nullable
	@Override
	public Storage<ItemVariant> getItemStorage(@Nullable class_2350 face) {
		if (invHandler == null)
			initHandler();
		return invHandler;
	}

	@Override
	public boolean addToTooltip(List<class_2561> tooltip, boolean isPlayerSneaking) {
		if (super.addToTooltip(tooltip, isPlayerSneaking))
			return true;
		if (getSpeed() == 0)
			return false;
		if (overflowItems.isEmpty())
			return false;
		TooltipHelper.addHint(tooltip, "hint.full_deployer");
		return true;
	}

	@Override
	public boolean addToGoggleTooltip(List<class_2561> tooltip, boolean isPlayerSneaking) {
		Lang.translate("tooltip.deployer.header")
			.forGoggles(tooltip);

		Lang.translate("tooltip.deployer." + (mode == Mode.USE ? "using" : "punching"))
			.style(class_124.field_1054)
			.forGoggles(tooltip);

		if (!heldItem.method_7960())
			Lang.translate("tooltip.deployer.contains", Components.translatable(heldItem.method_7922())
				.getString(), heldItem.method_7947())
				.style(class_124.field_1060)
				.forGoggles(tooltip);

		float stressAtBase = calculateStressApplied();
		if (StressImpact.isEnabled() && !class_3532.method_15347(stressAtBase, 0)) {
			tooltip.add(Components.immutableEmpty());
			addStressImpactStats(tooltip, stressAtBase);
		}

		return true;
	}

	@Environment(EnvType.CLIENT)
	public float getHandOffset(float partialTicks) {
		if (isVirtual())
			return animatedOffset.getValue(partialTicks);

		float progress = 0;
		int timerSpeed = getTimerSpeed();
		PartialModel handPose = getHandPose();

		if (state == State.EXPANDING) {
			progress = 1 - (timer - partialTicks * timerSpeed) / 1000f;
			if (fistBump)
				progress *= progress;
		}
		if (state == State.RETRACTING)
			progress = (timer - partialTicks * timerSpeed) / 1000f;
		float handLength = handPose == AllPartialModels.DEPLOYER_HAND_POINTING ? 0
			: handPose == AllPartialModels.DEPLOYER_HAND_HOLDING ? 4 / 16f : 3 / 16f;
		float distance = Math.min(class_3532.method_15363(progress, 0, 1) * (reach + handLength), 21 / 16f);

		return distance;
	}

	public void setAnimatedOffset(float offset) {
		animatedOffset.setValue(offset);
	}

	ItemStackHandlerContainer recipeInv = new ItemStackHandlerContainer(2);
	SandPaperInv sandpaperInv = new SandPaperInv(class_1799.field_8037);

	@Nullable
	public class_1860<? extends class_1263> getRecipe(class_1799 stack) {
		if (player == null || field_11863 == null)
			return null;

		class_1799 heldItemMainhand = player.method_6047();
		if (heldItemMainhand.method_7909() instanceof SandPaperItem) {
			sandpaperInv.method_5447(0, stack);
			return AllRecipeTypes.SANDPAPER_POLISHING.find(sandpaperInv, field_11863)
				.orElse(null);
		}

		recipeInv.method_5447(0, stack);
		recipeInv.method_5447(1, heldItemMainhand);

		DeployerRecipeSearchEvent event = new DeployerRecipeSearchEvent(this, recipeInv);

		event.addRecipe(() -> SequencedAssemblyRecipe.getRecipe(field_11863, event.getInventory(),
			AllRecipeTypes.DEPLOYING.getType(), DeployerApplicationRecipe.class), 100);
		event.addRecipe(() -> AllRecipeTypes.DEPLOYING.find(event.getInventory(), field_11863), 50);
		event.addRecipe(() -> AllRecipeTypes.ITEM_APPLICATION.find(event.getInventory(), field_11863), 50);

		DeployerRecipeSearchEvent.EVENT.invoker().handle(event);
		return event.getRecipe();
	}

	public DeployerFakePlayer getPlayer() {
		return player;
	}

	public List<class_1799> getOverflowItems() {
		return overflowItems;
	}
}
