package com.simibubi.create.content.kinetics.deployer;

import java.util.Collection;
import java.util.Objects;
import java.util.OptionalInt;
import java.util.UUID;

import javax.annotation.Nullable;

import io.github.fabricators_of_create.porting_lib.entity.events.EntityEvents;
import io.github.fabricators_of_create.porting_lib.util.UsernameCache;

import net.fabricmc.fabric.api.entity.FakePlayer;
import net.minecraft.class_1282;
import net.minecraft.class_1293;
import net.minecraft.class_1297;
import net.minecraft.class_1304;
import net.minecraft.class_1308;
import net.minecraft.class_1309;
import net.minecraft.class_1542;
import net.minecraft.class_1548;
import net.minecraft.class_1657;
import net.minecraft.class_1799;
import net.minecraft.class_1937;
import net.minecraft.class_2338;
import net.minecraft.class_243;
import net.minecraft.class_2535;
import net.minecraft.class_2561;
import net.minecraft.class_2598;
import net.minecraft.class_3218;
import net.minecraft.class_3908;
import net.minecraft.class_4050;
import org.apache.commons.lang3.tuple.Pair;

import com.mojang.authlib.GameProfile;
import com.simibubi.create.foundation.utility.Lang;
import com.simibubi.create.infrastructure.config.AllConfigs;
import com.simibubi.create.infrastructure.config.CKinetics;

import net.fabricmc.api.EnvType;
import net.fabricmc.api.Environment;

public class DeployerFakePlayer extends FakePlayer {

	private static final class_2535 NETWORK_MANAGER = new class_2535(class_2598.field_11942);
	public static final UUID fallbackID = UUID.fromString("9e2faded-cafe-4ec2-c314-dad129ae971d");
	Pair<class_2338, Float> blockBreakingProgress;
	class_1799 spawnedItemEffects;
	public boolean placedTracks;
	public boolean onMinecartContraption;
	private UUID owner;

	public DeployerFakePlayer(class_3218 world, @Nullable UUID owner) {
		super(world, new DeployerGameProfile(fallbackID, "Deployer", owner));
		// fabric: use the default FakePacketListener
//		connection = new FakePlayNetHandler(world.getServer(), this);
		this.owner = owner;
	}

	@Override
	public OptionalInt method_17355(class_3908 menuProvider) {
		return OptionalInt.empty();
	}

	@Override
	public class_2561 method_5476() {
		return Lang.translateDirect("block.deployer.damage_source_name");
	}

	@Override
	@Environment(EnvType.CLIENT)
	public float method_18381(class_4050 poseIn) {
		return 0;
	}

	@Override
	public class_243 method_19538() {
		return new class_243(method_23317(), method_23318(), method_23321());
	}

	@Override
	public float method_7279() {
		return 1 / 64f;
	}

	@Override
	public boolean method_7332(boolean ignoreHunger) {
		return false;
	}

	@Override
	public class_1799 method_18866(class_1937 world, class_1799 stack) {
		stack.method_7934(1);
		return stack;
	}

	@Override
	public boolean method_6049(class_1293 pEffectInstance) {
		return false;
	}

	@Override
	public UUID method_5667() {
		return owner == null ? super.method_5667() : owner;
	}

	public static void deployerHasEyesOnHisFeet(EntityEvents.Size event) {
		if (event.getEntity() instanceof DeployerFakePlayer)
			event.setNewEyeHeight(0);
	}

	public static boolean deployerCollectsDropsFromKilledEntities(class_1309 target, class_1282 source, Collection<class_1542> drops, int lootingLevel, boolean recentlyHit) {
		class_1297 trueSource = source.method_5529();
		if (trueSource != null && trueSource instanceof DeployerFakePlayer) {
			DeployerFakePlayer fakePlayer = (DeployerFakePlayer) trueSource;
			drops
				.forEach(stack -> fakePlayer.method_31548()
					.method_7398(stack.method_6983()));
			return true;
		}
		return false;
	}

	@Override
	protected boolean method_44201(class_1304 p_217035_) {
		return false;
	}

	@Override
	public void method_5650(class_5529 p_150097_) {
		if (blockBreakingProgress != null && !method_37908().field_9236)
			method_37908().method_8517(method_5628(), blockBreakingProgress.getKey(), -1);
		super.method_5650(p_150097_);
	}

	public static int deployerKillsDoNotSpawnXP(int i, class_1657 player, class_1309 entity) {
		if (player instanceof DeployerFakePlayer)
			return 0;
		return i;
	}

	public static boolean entitiesDontRetaliate(class_1309 target, class_1282 source, float amount) {
		if (!(target instanceof DeployerFakePlayer))
			return false;
		if (!(source.method_5529() instanceof class_1308))
			return false;
		class_1308 mob = (class_1308) source.method_5529();

		CKinetics.DeployerAggroSetting setting = AllConfigs.server().kinetics.ignoreDeployerAttacks.get();

		switch (setting) {
		case ALL:
			mob.method_5980(null);
			break;
		case CREEPERS:
			if (mob instanceof class_1548)
				mob.method_5980(null);
			break;
		case NONE:
		default:
		}
		return false; // true would short-circuit the event
	}

	// Credit to Mekanism for this approach. Helps fake players get past claims and
	// protection by other mods
	private static class DeployerGameProfile extends GameProfile {

		private UUID owner;

		public DeployerGameProfile(UUID id, String name, UUID owner) {
			super(id, name);
			this.owner = owner;
		}

		@Override
		public UUID getId() {
			return owner == null ? super.getId() : owner;
		}

		@Override
		public String getName() {
			if (owner == null)
				return super.getName();
			String lastKnownUsername = UsernameCache.getLastKnownUsername(owner);
			return lastKnownUsername == null ? super.getName() : lastKnownUsername;
		}

		@Override
		public boolean equals(final Object o) {
			if (this == o)
				return true;
			if (!(o instanceof GameProfile otherProfile))
				return false;
			return Objects.equals(getId(), otherProfile.getId()) && Objects.equals(getName(), otherProfile.getName());
		}

		@Override
		public int hashCode() {
			UUID id = getId();
			String name = getName();
			int result = id == null ? 0 : id.hashCode();
			result = 31 * result + (name == null ? 0 : name.hashCode());
			return result;
		}
	}

	// fabric: FakePlayNetHandler removed, unused

}
