package com.simibubi.create.content.kinetics.deployer;

import java.util.Iterator;
import java.util.function.Consumer;
import java.util.function.Predicate;
import java.util.function.Supplier;

import com.simibubi.create.foundation.blockEntity.behaviour.filtering.FilteringBehaviour;

import io.github.fabricators_of_create.porting_lib.transfer.item.ItemHandlerHelper;
import net.fabricmc.fabric.api.transfer.v1.item.ItemVariant;
import net.fabricmc.fabric.api.transfer.v1.storage.Storage;
import net.fabricmc.fabric.api.transfer.v1.storage.StorageView;
import net.fabricmc.fabric.api.transfer.v1.transaction.TransactionContext;
import net.fabricmc.fabric.api.transfer.v1.transaction.base.SnapshotParticipant;
import net.minecraft.class_1268;
import net.minecraft.class_1799;
import net.minecraft.class_3902;

public class DeployerItemHandler extends SnapshotParticipant<class_3902> implements Storage<ItemVariant> {

	private DeployerBlockEntity be;
	private DeployerFakePlayer player;

	public DeployerItemHandler(DeployerBlockEntity be) {
		this.be = be;
		this.player = be.player;
	}

	public class_1799 getHeld() {
		if (player == null)
			return class_1799.field_8037;
		return player.method_6047();
	}

	public void set(class_1799 stack) {
		if (player == null)
			return;
		if (be.method_10997().field_9236)
			return;
		player.method_6122(class_1268.field_5808, stack);
	}

	@Override
	public long insert(ItemVariant resource, long maxAmount, TransactionContext transaction) {
		int maxInsert = Math.min((int) maxAmount, resource.getItem().method_7882());
		class_1799 stack = resource.toStack(maxInsert);
		if (!isItemValid(stack))
			return 0;

		class_1799 held = getHeld();
		if (held.method_7960()) {
			be.snapshotParticipant.updateSnapshots(transaction);
			set(stack);
			return maxInsert;
		}

		if (!ItemHandlerHelper.canItemStacksStack(held, stack))
			return 0;

		int space = held.method_7914() - held.method_7947();
		if (space == 0)
			return 0;

		int toInsert = Math.min(maxInsert, space);
		class_1799 newStack = held.method_7972();
		newStack.method_7933(toInsert);
		be.snapshotParticipant.updateSnapshots(transaction);
		set(newStack);
		return toInsert;
	}

	@Override
	public long extract(ItemVariant resource, long maxAmount, TransactionContext transaction) {
		long extracted = extractFromOverflow(resource, maxAmount, transaction);
		if (extracted == maxAmount)
			return extracted;
		extracted += extractFromHeld(resource, maxAmount - extracted, transaction);
		return extracted;
	}

	/**
	 * @return number of items extracted from the TE's overflow slots
	 */
	public long extractFromOverflow(ItemVariant resource, long maxAmount, TransactionContext transaction) {
		long extracted = 0;
		for (int i = 0; i < be.overflowItems.size(); i++) {
			class_1799 itemStack = be.overflowItems.get(i);
			if (itemStack.method_7960())
				continue;
			if (!resource.matches(itemStack))
				continue;
			int toExtract = (int) Math.min(maxAmount - extracted, itemStack.method_7947());
			if (extracted == 0)
				be.snapshotParticipant.updateSnapshots(transaction);
			extracted += toExtract;
			class_1799 newStack = itemStack.method_7972();
			newStack.method_7934(toExtract);
			be.overflowItems.set(i, newStack);
		}
		return extracted;
	}

	/**
	 * @return the number of items extracted from the TE's held stack
	 */
	public long extractFromHeld(ItemVariant resource, long maxAmount, TransactionContext transaction) {
		class_1799 held = getHeld();
		if (held.method_7960() || !resource.matches(held))
			return 0;
		if (!be.filtering.getFilter().method_7960() && be.filtering.test(held))
			return 0;
		int toExtract = (int) Math.min(maxAmount, held.method_7947());
		be.snapshotParticipant.updateSnapshots(transaction);
		class_1799 newStack = held.method_7972();
		newStack.method_7934(toExtract);
		set(newStack);
		return toExtract;
	}

	@Override
	protected class_3902 createSnapshot() {
		return class_3902.field_17274;
	}

	@Override
	protected void readSnapshot(class_3902 snapshot) {
	}

	@Override
	protected void onFinalCommit() {
		super.onFinalCommit();
		be.method_5431();
		be.sendData();
	}

	@Override
	public Iterator<StorageView<ItemVariant>> iterator() {
		return new DeployerItemHandlerIterator();
	}

	public boolean isItemValid(class_1799 stack) {
		FilteringBehaviour filteringBehaviour = be.getBehaviour(FilteringBehaviour.TYPE);
		return filteringBehaviour == null || filteringBehaviour.test(stack);
	}

	public class DeployerItemHandlerIterator implements Iterator<StorageView<ItemVariant>> {
		private int index; // -1 means held

		public DeployerItemHandlerIterator() {
			this.index = be.overflowItems.size() != 0 ? 0 : -1;
		}

		@Override
		public boolean hasNext() {
			return index < be.overflowItems.size();
		}

		@Override
		public StorageView<ItemVariant> next() {
			final int indexFinal = this.index;
			Supplier<class_1799> heldGetter = () -> be.overflowItems.get(indexFinal);
			Consumer<class_1799> heldSetter = (stack) -> be.overflowItems.set(indexFinal, stack);
			Predicate<class_1799> mayExtract = stack -> true;
			if (index == -1) {
				heldGetter = player::method_6047;
				heldSetter = s -> player.method_6122(class_1268.field_5808, s);
				mayExtract = s -> be.filtering.getFilter().method_7960() || !be.filtering.test(s);
				index = be.overflowItems.size(); // hasNext will be false now
			}
			index++;
			if (index == be.overflowItems.size())
				index = -1;
			return new DeployerSlotView(heldGetter, heldSetter, mayExtract);
		}

	}

	public class DeployerSlotView extends SnapshotParticipant<class_3902> implements StorageView<ItemVariant> {
		public final Supplier<class_1799> heldGetter;
		public final Consumer<class_1799> heldSetter;
		public final Predicate<class_1799> mayExtract;

		public DeployerSlotView(Supplier<class_1799> heldGetter, Consumer<class_1799> heldSetter, Predicate<class_1799> mayExtract) {
			this.heldGetter = heldGetter;
			this.heldSetter = heldSetter;
			this.mayExtract = mayExtract;
		}

		@Override
		public long extract(ItemVariant resource, long maxAmount, TransactionContext transaction) {
			class_1799 stack = getStack();
			if (stack.method_7960() || !resource.matches(stack) || !mayExtract.test(stack))
				return 0;
			int toExtract = (int) Math.min(maxAmount, stack.method_7947());
			updateSnapshots(transaction);
			class_1799 newStack = ItemHandlerHelper.copyStackWithSize(stack, stack.method_7947() - toExtract);
			heldSetter.accept(newStack);
			return toExtract;
		}

		@Override
		public boolean isResourceBlank() {
			return getResource().isBlank();
		}

		@Override
		public ItemVariant getResource() {
			class_1799 stack = getStack();
			return stack.method_7960() ? ItemVariant.blank() : ItemVariant.of(stack);
		}

		@Override
		public long getAmount() {
			return getStack().method_7947();
		}

		@Override
		public long getCapacity() {
			return getStack().method_7914();
		}

		public class_1799 getStack() {
			return heldGetter.get();
		}

		@Override
		public void updateSnapshots(TransactionContext transaction) {
			super.updateSnapshots(transaction);
			be.snapshotParticipant.updateSnapshots(transaction);
		}

		@Override
		protected class_3902 createSnapshot() {
			return class_3902.field_17274;
		}

		@Override
		protected void readSnapshot(class_3902 snapshot) {
		}

		@Override
		protected void onFinalCommit() {
			super.onFinalCommit();
			DeployerItemHandler.this.onFinalCommit();
		}
	}
}
