package com.simibubi.create.content.kinetics.deployer;

import java.util.Arrays;
import java.util.List;
import java.util.UUID;

import javax.annotation.Nullable;

import org.apache.commons.lang3.tuple.Pair;

import com.jozufozu.flywheel.api.MaterialManager;
import com.jozufozu.flywheel.core.virtual.VirtualRenderWorld;
import com.simibubi.create.AllBlocks;
import com.simibubi.create.AllItems;
import com.simibubi.create.content.contraptions.AbstractContraptionEntity;
import com.simibubi.create.content.contraptions.OrientedContraptionEntity;
import com.simibubi.create.content.contraptions.behaviour.MovementBehaviour;
import com.simibubi.create.content.contraptions.behaviour.MovementContext;
import com.simibubi.create.content.contraptions.mounted.MountedContraption;
import com.simibubi.create.content.contraptions.render.ActorInstance;
import com.simibubi.create.content.contraptions.render.ContraptionMatrices;
import com.simibubi.create.content.contraptions.render.ContraptionRenderDispatcher;
import com.simibubi.create.content.kinetics.deployer.DeployerBlockEntity.Mode;
import com.simibubi.create.content.logistics.filter.FilterItemStack;
import com.simibubi.create.content.schematics.SchematicInstances;
import com.simibubi.create.content.schematics.SchematicWorld;
import com.simibubi.create.content.schematics.requirement.ItemRequirement;
import com.simibubi.create.content.trains.entity.CarriageContraption;
import com.simibubi.create.content.trains.entity.CarriageContraptionEntity;
import com.simibubi.create.foundation.advancement.AllAdvancements;
import com.simibubi.create.foundation.item.ItemHelper;
import com.simibubi.create.foundation.item.ItemHelper.ExtractionCountMode;
import com.simibubi.create.foundation.utility.BlockHelper;
import com.simibubi.create.foundation.utility.NBTHelper;
import com.simibubi.create.foundation.utility.VecHelper;

import io.github.fabricators_of_create.porting_lib.transfer.TransferUtil;
import io.github.fabricators_of_create.porting_lib.util.NBTSerializer;
import net.fabricmc.fabric.api.transfer.v1.item.ItemVariant;
import net.fabricmc.fabric.api.transfer.v1.storage.Storage;
import net.fabricmc.fabric.api.transfer.v1.storage.base.ResourceAmount;
import net.fabricmc.fabric.api.transfer.v1.transaction.Transaction;
import net.minecraft.class_1268;
import net.minecraft.class_1661;
import net.minecraft.class_1799;
import net.minecraft.class_1937;
import net.minecraft.class_2338;
import net.minecraft.class_2350.class_2351;
import net.minecraft.class_243;
import net.minecraft.class_2487;
import net.minecraft.class_2499;
import net.minecraft.class_2520;
import net.minecraft.class_2680;
import net.minecraft.class_3218;
import net.minecraft.class_4597;

public class DeployerMovementBehaviour implements MovementBehaviour {

	@Override
	public class_243 getActiveAreaOffset(MovementContext context) {
		return class_243.method_24954(context.state.method_11654(DeployerBlock.FACING)
			.method_10163())
			.method_1021(2);
	}

	@Override
	public void visitNewPosition(MovementContext context, class_2338 pos) {
		if (context.world.field_9236)
			return;

		tryGrabbingItem(context);
		DeployerFakePlayer player = getPlayer(context);
		Mode mode = getMode(context);
		if (mode == Mode.USE && !DeployerHandler.shouldActivate(player.method_6047(), context.world, pos, null))
			return;

		activate(context, pos, player, mode);
		tryDisposeOfExcess(context);
		context.stall = player.blockBreakingProgress != null;
	}

	public void activate(MovementContext context, class_2338 pos, DeployerFakePlayer player, Mode mode) {
		class_1937 world = context.world;

		FilterItemStack filter = context.getFilterFromBE();
		if (AllItems.SCHEMATIC.isIn(filter.item()))
			activateAsSchematicPrinter(context, pos, player, world, filter.item());

		class_243 facingVec = class_243.method_24954(context.state.method_11654(DeployerBlock.FACING)
			.method_10163());
		facingVec = context.rotation.apply(facingVec);
		class_243 vec = context.position.method_1020(facingVec.method_1021(2));

		float xRot = AbstractContraptionEntity.pitchFromVector(facingVec) - 90;
		if (Math.abs(xRot) > 89) {
			class_243 initial = new class_243(0, 0, 1);
			if (context.contraption.entity instanceof OrientedContraptionEntity oce)
				initial = VecHelper.rotate(initial, oce.getInitialYaw(), class_2351.field_11052);
			if (context.contraption.entity instanceof CarriageContraptionEntity cce)
				initial = VecHelper.rotate(initial, 90, class_2351.field_11052);
			facingVec = context.rotation.apply(initial);
		}

		player.method_36456(AbstractContraptionEntity.yawFromVector(facingVec));
		player.method_36457(xRot);
		player.placedTracks = false;

		DeployerHandler.activate(player, vec, pos, facingVec, mode);

		if ((context.contraption instanceof MountedContraption || context.contraption instanceof CarriageContraption)
			&& player.placedTracks && context.blockEntityData != null && context.blockEntityData.method_10545("Owner"))
			AllAdvancements.SELF_DEPLOYING.awardTo(world.method_18470(context.blockEntityData.method_25926("Owner")));
	}

	protected void activateAsSchematicPrinter(MovementContext context, class_2338 pos, DeployerFakePlayer player,
		class_1937 world, class_1799 filter) {
		if (!filter.method_7985())
			return;
		if (!world.method_8320(pos)
			.method_45474())
			return;

		class_2487 tag = filter.method_7969();
		if (!tag.method_10577("Deployed"))
			return;
		SchematicWorld schematicWorld = SchematicInstances.get(world, filter);
		if (schematicWorld == null)
			return;
		if (!schematicWorld.getBounds()
			.method_14662(pos.method_10059(schematicWorld.anchor)))
			return;
		class_2680 blockState = schematicWorld.method_8320(pos);
		ItemRequirement requirement = ItemRequirement.of(blockState, schematicWorld.method_8321(pos));
		if (requirement.isInvalid() || requirement.isEmpty())
			return;
		if (AllBlocks.BELT.has(blockState))
			return;

		List<ItemRequirement.StackRequirement> requiredItems = requirement.getRequiredItems();
		class_1799 contextStack = requiredItems.isEmpty() ? class_1799.field_8037 : requiredItems.get(0).stack;

		if (!context.contraption.hasUniversalCreativeCrate) {
			Storage<ItemVariant> itemHandler = context.contraption.getSharedInventory();
			try (Transaction t = TransferUtil.getTransaction()) {
				for (ItemRequirement.StackRequirement required : requiredItems) {
					int count = required.stack.method_7947();
					ResourceAmount<ItemVariant> resource = TransferUtil.extractMatching(itemHandler, required::matches, count, t);
					if (resource == null || resource.amount() != count)
						return; // didn't extract what we needed, skip
				}
				// if we get here all requirements were met
				t.commit();
			}
			for (ItemRequirement.StackRequirement required : requiredItems)
				contextStack = ItemHelper.extract(itemHandler, required::matches,
					ExtractionCountMode.EXACTLY, required.stack.method_7947(), false);
		}

		class_2487 data = BlockHelper.prepareBlockEntityData(blockState, schematicWorld.method_8321(pos));
//		BlockSnapshot blocksnapshot = BlockSnapshot.create(world.dimension(), world, pos);
		BlockHelper.placeSchematicBlock(world, blockState, pos, contextStack, data);
//		if (ForgeEventFactory.onBlockPlace(player, blocksnapshot, Direction.UP))
//			blocksnapshot.restore(true, false);
	}

	@Override
	public void tick(MovementContext context) {
		if (context.world.field_9236)
			return;
		if (!context.stall)
			return;

		DeployerFakePlayer player = getPlayer(context);
		Mode mode = getMode(context);

		Pair<class_2338, Float> blockBreakingProgress = player.blockBreakingProgress;
		if (blockBreakingProgress != null) {
			int timer = context.data.method_10550("Timer");
			if (timer < 20) {
				timer++;
				context.data.method_10569("Timer", timer);
				return;
			}

			context.data.method_10551("Timer");
			activate(context, blockBreakingProgress.getKey(), player, mode);
			tryDisposeOfExcess(context);
		}

		context.stall = player.blockBreakingProgress != null;
	}

	@Override
	public void cancelStall(MovementContext context) {
		if (context.world.field_9236)
			return;

		MovementBehaviour.super.cancelStall(context);
		DeployerFakePlayer player = getPlayer(context);
		if (player == null)
			return;
		if (player.blockBreakingProgress == null)
			return;
		context.world.method_8517(player.method_5628(), player.blockBreakingProgress.getKey(), -1);
		player.blockBreakingProgress = null;
	}

	@Override
	public void stopMoving(MovementContext context) {
		if (context.world.field_9236)
			return;

		DeployerFakePlayer player = getPlayer(context);
		if (player == null)
			return;

		cancelStall(context);
		context.blockEntityData.method_10566("Inventory", player.method_31548()
			.method_7384(new class_2499()));
		player.method_31472();
	}

	private void tryGrabbingItem(MovementContext context) {
		DeployerFakePlayer player = getPlayer(context);
		if (player == null)
			return;
		if (player.method_6047()
			.method_7960()) {
			FilterItemStack filter = context.getFilterFromBE();
			if (AllItems.SCHEMATIC.isIn(filter.item()))
				return;
			class_1799 held = ItemHelper.extract(context.contraption.getSharedInventory(),
				stack -> filter.test(context.world, stack), 1, false);
			player.method_6122(class_1268.field_5808, held);
		}
	}

	private void tryDisposeOfExcess(MovementContext context) {
		DeployerFakePlayer player = getPlayer(context);
		if (player == null)
			return;
		class_1661 inv = player.method_31548();
		FilterItemStack filter = context.getFilterFromBE();

		for (List<class_1799> list : Arrays.asList(inv.field_7548, inv.field_7544, inv.field_7547)) {
			for (int i = 0; i < list.size(); ++i) {
				class_1799 itemstack = list.get(i);
				if (itemstack.method_7960())
					continue;

				if (list == inv.field_7547 && i == inv.field_7545 && filter.test(context.world, itemstack))
					continue;

				dropItem(context, itemstack);
				list.set(i, class_1799.field_8037);
			}
		}
	}

	@Override
	public void writeExtraData(MovementContext context) {
		DeployerFakePlayer player = getPlayer(context);
		if (player == null)
			return;
		context.data.method_10566("HeldItem", NBTSerializer.serializeNBT(player.method_6047()));
	}

	private DeployerFakePlayer getPlayer(MovementContext context) {
		if (!(context.temporaryData instanceof DeployerFakePlayer) && context.world instanceof class_3218) {
			UUID owner = context.blockEntityData.method_10545("Owner") ? context.blockEntityData.method_25926("Owner") : null;
			DeployerFakePlayer deployerFakePlayer = new DeployerFakePlayer((class_3218) context.world, owner);
			deployerFakePlayer.onMinecartContraption = context.contraption instanceof MountedContraption;
			deployerFakePlayer.method_31548()
				.method_7397(context.blockEntityData.method_10554("Inventory", class_2520.field_33260));
			if (context.data.method_10545("HeldItem"))
				deployerFakePlayer.method_6122(class_1268.field_5808,
					class_1799.method_7915(context.data.method_10562("HeldItem")));
			context.blockEntityData.method_10551("Inventory");
			context.temporaryData = deployerFakePlayer;
		}
		return (DeployerFakePlayer) context.temporaryData;
	}

	private Mode getMode(MovementContext context) {
		return NBTHelper.readEnum(context.blockEntityData, "Mode", Mode.class);
	}

	@Override
	public void renderInContraption(MovementContext context, VirtualRenderWorld renderWorld,
		ContraptionMatrices matrices, class_4597 buffers) {
		if (!ContraptionRenderDispatcher.canInstance())
			DeployerRenderer.renderInContraption(context, renderWorld, matrices, buffers);
	}

	@Override
	public boolean hasSpecialInstancedRendering() {
		return true;
	}

	@Nullable
	@Override
	public ActorInstance createInstance(MaterialManager materialManager, VirtualRenderWorld simulationWorld,
		MovementContext context) {
		return new DeployerActorInstance(materialManager, simulationWorld, context);
	}
}
