package com.simibubi.create.content.kinetics.fan.processing;

import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import net.minecraft.class_1293;
import net.minecraft.class_1294;
import net.minecraft.class_1297;
import net.minecraft.class_1299;
import net.minecraft.class_1309;
import net.minecraft.class_1498;
import net.minecraft.class_1506;
import net.minecraft.class_1560;
import net.minecraft.class_1799;
import net.minecraft.class_1874;
import net.minecraft.class_1937;
import net.minecraft.class_2246;
import net.minecraft.class_2338;
import net.minecraft.class_2388;
import net.minecraft.class_2390;
import net.minecraft.class_2398;
import net.minecraft.class_243;
import net.minecraft.class_2487;
import net.minecraft.class_2680;
import net.minecraft.class_3417;
import net.minecraft.class_3419;
import net.minecraft.class_3481;
import net.minecraft.class_3610;
import net.minecraft.class_3859;
import net.minecraft.class_3861;
import net.minecraft.class_3862;
import net.minecraft.class_3922;
import net.minecraft.class_3956;
import net.minecraft.class_5455;
import net.minecraft.class_5819;
import org.jetbrains.annotations.Nullable;
import org.joml.Vector3f;

import com.simibubi.create.AllRecipeTypes;
import com.simibubi.create.AllTags.AllBlockTags;
import com.simibubi.create.AllTags.AllFluidTags;
import com.simibubi.create.Create;
import com.simibubi.create.content.kinetics.fan.processing.HauntingRecipe.HauntingWrapper;
import com.simibubi.create.content.kinetics.fan.processing.SplashingRecipe.SplashingWrapper;
import com.simibubi.create.content.processing.burner.BlazeBurnerBlock;
import com.simibubi.create.content.processing.burner.LitBlazeBurnerBlock;
import com.simibubi.create.foundation.damageTypes.CreateDamageSources;
import com.simibubi.create.foundation.recipe.RecipeApplier;
import com.simibubi.create.foundation.utility.Color;
import com.simibubi.create.foundation.utility.VecHelper;

import io.github.fabricators_of_create.porting_lib.transfer.item.ItemStackHandler;
import io.github.fabricators_of_create.porting_lib.transfer.item.RecipeWrapper;
import io.github.fabricators_of_create.porting_lib.util.NBTSerializer;
import it.unimi.dsi.fastutil.objects.Object2ReferenceOpenHashMap;

public class AllFanProcessingTypes {
	public static final NoneType NONE = register("none", new NoneType());
	public static final BlastingType BLASTING = register("blasting", new BlastingType());
	public static final HauntingType HAUNTING = register("haunting", new HauntingType());
	public static final SmokingType SMOKING = register("smoking", new SmokingType());
	public static final SplashingType SPLASHING = register("splashing", new SplashingType());

	private static final Map<String, FanProcessingType> LEGACY_NAME_MAP;

	static {
		Object2ReferenceOpenHashMap<String, FanProcessingType> map = new Object2ReferenceOpenHashMap<>();
		map.put("NONE", NONE);
		map.put("BLASTING", BLASTING);
		map.put("HAUNTING", HAUNTING);
		map.put("SMOKING", SMOKING);
		map.put("SPLASHING", SPLASHING);
		map.trim();
		LEGACY_NAME_MAP = map;
	}

	private static <T extends FanProcessingType> T register(String id, T type) {
		FanProcessingTypeRegistry.register(Create.asResource(id), type);
		return type;
	}

	@Nullable
	public static FanProcessingType ofLegacyName(String name) {
		return LEGACY_NAME_MAP.get(name);
	}

	public static void register() {
	}

	public static FanProcessingType parseLegacy(String str) {
		FanProcessingType type = ofLegacyName(str);
		if (type != null) {
			return type;
		}
		return FanProcessingType.parse(str);
	}

	public static class NoneType implements FanProcessingType {
		@Override
		public boolean isValidAt(class_1937 level, class_2338 pos) {
			return true;
		}

		@Override
		public int getPriority() {
			return -1000000;
		}

		@Override
		public boolean canProcess(class_1799 stack, class_1937 level) {
			return false;
		}

		@Override
		@Nullable
		public List<class_1799> process(class_1799 stack, class_1937 level) {
			return null;
		}

		@Override
		public void spawnProcessingParticles(class_1937 level, class_243 pos) {
		}

		@Override
		public void morphAirFlow(AirFlowParticleAccess particleAccess, class_5819 random) {
		}

		@Override
		public void affectEntity(class_1297 entity, class_1937 level) {
		}
	}

	public static class BlastingType implements FanProcessingType {
		private static final RecipeWrapper RECIPE_WRAPPER = new RecipeWrapper(new ItemStackHandler(1));

		@Override
		public boolean isValidAt(class_1937 level, class_2338 pos) {
			class_3610 fluidState = level.method_8316(pos);
			if (AllFluidTags.FAN_PROCESSING_CATALYSTS_BLASTING.matches(fluidState)) {
				return true;
			}
			class_2680 blockState = level.method_8320(pos);
			if (AllBlockTags.FAN_PROCESSING_CATALYSTS_BLASTING.matches(blockState)) {
				if (blockState.method_28498(BlazeBurnerBlock.HEAT_LEVEL) && !blockState.method_11654(BlazeBurnerBlock.HEAT_LEVEL).isAtLeast(BlazeBurnerBlock.HeatLevel.FADING)) {
					return false;
				}
				return true;
			}
			return false;
		}

		@Override
		public int getPriority() {
			return 100;
		}

		@Override
		public boolean canProcess(class_1799 stack, class_1937 level) {
			RECIPE_WRAPPER.method_5447(0, stack);
			Optional<class_3861> smeltingRecipe = level.method_8433()
				.method_8132(class_3956.field_17546, RECIPE_WRAPPER, level);

			if (smeltingRecipe.isPresent())
				return true;

			RECIPE_WRAPPER.method_5447(0, stack);
			Optional<class_3859> blastingRecipe = level.method_8433()
				.method_8132(class_3956.field_17547, RECIPE_WRAPPER, level);

			if (blastingRecipe.isPresent())
				return true;

			return !stack.method_7909()
				.method_24358();
		}

		@Override
		@Nullable
		public List<class_1799> process(class_1799 stack, class_1937 level) {
			RECIPE_WRAPPER.method_5447(0, stack);
			Optional<class_3862> smokingRecipe = level.method_8433()
				.method_8132(class_3956.field_17548, RECIPE_WRAPPER, level);

			RECIPE_WRAPPER.method_5447(0, stack);
			Optional<? extends class_1874> smeltingRecipe = level.method_8433()
				.method_8132(class_3956.field_17546, RECIPE_WRAPPER, level);
			if (!smeltingRecipe.isPresent()) {
				RECIPE_WRAPPER.method_5447(0, stack);
				smeltingRecipe = level.method_8433()
					.method_8132(class_3956.field_17547, RECIPE_WRAPPER, level);
			}

			if (smeltingRecipe.isPresent()) {
				class_5455 registryAccess = level.method_30349();
				if (!smokingRecipe.isPresent() || !class_1799.method_7984(smokingRecipe.get()
					.method_8110(registryAccess),
					smeltingRecipe.get()
						.method_8110(registryAccess))) {
					return RecipeApplier.applyRecipeOn(level, stack, smeltingRecipe.get());
				}
			}

			return Collections.emptyList();
		}

		@Override
		public void spawnProcessingParticles(class_1937 level, class_243 pos) {
			if (level.field_9229.method_43048(8) != 0)
				return;
			level.method_8406(class_2398.field_11237, pos.field_1352, pos.field_1351 + .25f, pos.field_1350, 0, 1 / 16f, 0);
		}

		@Override
		public void morphAirFlow(AirFlowParticleAccess particleAccess, class_5819 random) {
			particleAccess.setColor(Color.mixColors(0xFF4400, 0xFF8855, random.method_43057()));
			particleAccess.setAlpha(.5f);
			if (random.method_43057() < 1 / 32f)
				particleAccess.spawnExtraParticle(class_2398.field_11240, .25f);
			if (random.method_43057() < 1 / 16f)
				particleAccess.spawnExtraParticle(new class_2388(class_2398.field_11217, class_2246.field_10164.method_9564()), .25f);
		}

		@Override
		public void affectEntity(class_1297 entity, class_1937 level) {
			if (level.field_9236)
				return;

			if (!entity.method_5753()) {
				entity.method_5639(10);
				entity.method_5643(CreateDamageSources.fanLava(level), 4);
			}
		}
	}

	public static class HauntingType implements FanProcessingType {
		private static final HauntingWrapper HAUNTING_WRAPPER = new HauntingWrapper();

		@Override
		public boolean isValidAt(class_1937 level, class_2338 pos) {
			class_3610 fluidState = level.method_8316(pos);
			if (AllFluidTags.FAN_PROCESSING_CATALYSTS_HAUNTING.matches(fluidState)) {
				return true;
			}
			class_2680 blockState = level.method_8320(pos);
			if (AllBlockTags.FAN_PROCESSING_CATALYSTS_HAUNTING.matches(blockState)) {
				if (blockState.method_26164(class_3481.field_23799) && blockState.method_28498(class_3922.field_17352) && !blockState.method_11654(class_3922.field_17352)) {
					return false;
				}
				if (blockState.method_28498(LitBlazeBurnerBlock.FLAME_TYPE) && blockState.method_11654(LitBlazeBurnerBlock.FLAME_TYPE) != LitBlazeBurnerBlock.FlameType.SOUL) {
					return false;
				}
				return true;
			}
			return false;
		}

		@Override
		public int getPriority() {
			return 300;
		}

		@Override
		public boolean canProcess(class_1799 stack, class_1937 level) {
			HAUNTING_WRAPPER.method_5447(0, stack);
			Optional<HauntingRecipe> recipe = AllRecipeTypes.HAUNTING.find(HAUNTING_WRAPPER, level);
			return recipe.isPresent();
		}

		@Override
		@Nullable
		public List<class_1799> process(class_1799 stack, class_1937 level) {
			HAUNTING_WRAPPER.method_5447(0, stack);
			Optional<HauntingRecipe> recipe = AllRecipeTypes.HAUNTING.find(HAUNTING_WRAPPER, level);
			if (recipe.isPresent())
				return RecipeApplier.applyRecipeOn(level, stack, recipe.get());
			return null;
		}

		@Override
		public void spawnProcessingParticles(class_1937 level, class_243 pos) {
			if (level.field_9229.method_43048(8) != 0)
				return;
			pos = pos.method_1019(VecHelper.offsetRandomly(class_243.field_1353, level.field_9229, 1)
				.method_18805(1, 0.05f, 1)
				.method_1029()
				.method_1021(0.15f));
			level.method_8406(class_2398.field_22246, pos.field_1352, pos.field_1351 + .45f, pos.field_1350, 0, 0, 0);
			if (level.field_9229.method_43048(2) == 0)
				level.method_8406(class_2398.field_11251, pos.field_1352, pos.field_1351 + .25f, pos.field_1350, 0, 0, 0);
		}

		@Override
		public void morphAirFlow(AirFlowParticleAccess particleAccess, class_5819 random) {
			particleAccess.setColor(Color.mixColors(0x0, 0x126568, random.method_43057()));
			particleAccess.setAlpha(1f);
			if (random.method_43057() < 1 / 128f)
				particleAccess.spawnExtraParticle(class_2398.field_22246, .125f);
			if (random.method_43057() < 1 / 32f)
				particleAccess.spawnExtraParticle(class_2398.field_11251, .125f);
		}

		@Override
		public void affectEntity(class_1297 entity, class_1937 level) {
			if (level.field_9236) {
				if (entity instanceof class_1498) {
					class_243 p = entity.method_30950(0);
					class_243 v = p.method_1031(0, 0.5f, 0)
						.method_1019(VecHelper.offsetRandomly(class_243.field_1353, level.field_9229, 1)
							.method_18805(1, 0.2f, 1)
							.method_1029()
							.method_1021(1f));
					level.method_8406(class_2398.field_22246, v.field_1352, v.field_1351, v.field_1350, 0, 0.1f, 0);
					if (level.field_9229.method_43048(3) == 0)
						level.method_8406(class_2398.field_11237, p.field_1352, p.field_1351 + .5f, p.field_1350,
							(level.field_9229.method_43057() - .5f) * .5f, 0.1f, (level.field_9229.method_43057() - .5f) * .5f);
				}
				return;
			}

			if (entity instanceof class_1309 livingEntity) {
				livingEntity.method_6092(new class_1293(class_1294.field_5919, 30, 0, false, false));
				livingEntity.method_6092(new class_1293(class_1294.field_5909, 20, 1, false, false));
			}
			if (entity instanceof class_1498 horse) {
				int progress = horse.getCustomData()
					.method_10550("CreateHaunting");
				if (progress < 100) {
					if (progress % 10 == 0) {
						level.method_8396(null, entity.method_24515(), class_3417.field_23060, class_3419.field_15254,
							1f, 1.5f * progress / 100f);
					}
					horse.getCustomData()
						.method_10569("CreateHaunting", progress + 1);
					return;
				}

				level.method_8396(null, entity.method_24515(), class_3417.field_15222,
					class_3419.field_15254, 1.25f, 0.65f);

				class_1506 skeletonHorse = class_1299.field_6075.method_5883(level);
				class_2487 serializeNBT = horse.method_5647(new class_2487());
				serializeNBT.method_10551("UUID");
				if (!horse.method_6786()
					.method_7960())
					horse.method_5775(horse.method_6786());

				NBTSerializer.deserializeNBT(skeletonHorse, serializeNBT);
				skeletonHorse.method_33574(horse.method_30950(0));
				level.method_8649(skeletonHorse);
				horse.method_31472();
			}
		}
	}

	public static class SmokingType implements FanProcessingType {
		private static final RecipeWrapper RECIPE_WRAPPER = new RecipeWrapper(new ItemStackHandler(1));

		@Override
		public boolean isValidAt(class_1937 level, class_2338 pos) {
			class_3610 fluidState = level.method_8316(pos);
			if (AllFluidTags.FAN_PROCESSING_CATALYSTS_SMOKING.matches(fluidState)) {
				return true;
			}
			class_2680 blockState = level.method_8320(pos);
			if (AllBlockTags.FAN_PROCESSING_CATALYSTS_SMOKING.matches(blockState)) {
				if (blockState.method_26164(class_3481.field_23799) && blockState.method_28498(class_3922.field_17352) && !blockState.method_11654(class_3922.field_17352)) {
					return false;
				}
				if (blockState.method_28498(LitBlazeBurnerBlock.FLAME_TYPE) && blockState.method_11654(LitBlazeBurnerBlock.FLAME_TYPE) != LitBlazeBurnerBlock.FlameType.REGULAR) {
					return false;
				}
				if (blockState.method_28498(BlazeBurnerBlock.HEAT_LEVEL) && blockState.method_11654(BlazeBurnerBlock.HEAT_LEVEL) != BlazeBurnerBlock.HeatLevel.SMOULDERING) {
					return false;
				}
				return true;
			}
			return false;
		}

		@Override
		public int getPriority() {
			return 200;
		}

		@Override
		public boolean canProcess(class_1799 stack, class_1937 level) {
			RECIPE_WRAPPER.method_5447(0, stack);
			Optional<class_3862> recipe = level.method_8433()
				.method_8132(class_3956.field_17548, RECIPE_WRAPPER, level);
			return recipe.isPresent();
		}

		@Override
		@Nullable
		public List<class_1799> process(class_1799 stack, class_1937 level) {
			RECIPE_WRAPPER.method_5447(0, stack);
			Optional<class_3862> smokingRecipe = level.method_8433()
				.method_8132(class_3956.field_17548, RECIPE_WRAPPER, level);

			if (smokingRecipe.isPresent())
				return RecipeApplier.applyRecipeOn(level, stack, smokingRecipe.get());

			return null;
		}

		@Override
		public void spawnProcessingParticles(class_1937 level, class_243 pos) {
			if (level.field_9229.method_43048(8) != 0)
				return;
			level.method_8406(class_2398.field_11203, pos.field_1352, pos.field_1351 + .25f, pos.field_1350, 0, 1 / 16f, 0);
		}

		@Override
		public void morphAirFlow(AirFlowParticleAccess particleAccess, class_5819 random) {
			particleAccess.setColor(Color.mixColors(0x0, 0x555555, random.method_43057()));
			particleAccess.setAlpha(1f);
			if (random.method_43057() < 1 / 32f)
				particleAccess.spawnExtraParticle(class_2398.field_11251, .125f);
			if (random.method_43057() < 1 / 32f)
				particleAccess.spawnExtraParticle(class_2398.field_11237, .125f);
		}

		@Override
		public void affectEntity(class_1297 entity, class_1937 level) {
			if (level.field_9236)
				return;

			if (!entity.method_5753()) {
				entity.method_5639(2);
				entity.method_5643(CreateDamageSources.fanFire(level), 2);
			}
		}
	}

	public static class SplashingType implements FanProcessingType {
		private static final SplashingWrapper SPLASHING_WRAPPER = new SplashingWrapper();

		@Override
		public boolean isValidAt(class_1937 level, class_2338 pos) {
			class_3610 fluidState = level.method_8316(pos);
			if (AllFluidTags.FAN_PROCESSING_CATALYSTS_SPLASHING.matches(fluidState)) {
				return true;
			}
			class_2680 blockState = level.method_8320(pos);
			if (AllBlockTags.FAN_PROCESSING_CATALYSTS_SPLASHING.matches(blockState)) {
				return true;
			}
			return false;
		}

		@Override
		public int getPriority() {
			return 400;
		}

		@Override
		public boolean canProcess(class_1799 stack, class_1937 level) {
			SPLASHING_WRAPPER.method_5447(0, stack);
			Optional<SplashingRecipe> recipe = AllRecipeTypes.SPLASHING.find(SPLASHING_WRAPPER, level);
			return recipe.isPresent();
		}

		@Override
		@Nullable
		public List<class_1799> process(class_1799 stack, class_1937 level) {
			SPLASHING_WRAPPER.method_5447(0, stack);
			Optional<SplashingRecipe> recipe = AllRecipeTypes.SPLASHING.find(SPLASHING_WRAPPER, level);
			if (recipe.isPresent())
				return RecipeApplier.applyRecipeOn(level, stack, recipe.get());
			return null;
		}

		@Override
		public void spawnProcessingParticles(class_1937 level, class_243 pos) {
			if (level.field_9229.method_43048(8) != 0)
				return;
			Vector3f color = new Color(0x0055FF).asVectorF();
			level.method_8406(new class_2390(color, 1), pos.field_1352 + (level.field_9229.method_43057() - .5f) * .5f,
				pos.field_1351 + .5f, pos.field_1350 + (level.field_9229.method_43057() - .5f) * .5f, 0, 1 / 8f, 0);
			level.method_8406(class_2398.field_11228, pos.field_1352 + (level.field_9229.method_43057() - .5f) * .5f, pos.field_1351 + .5f,
				pos.field_1350 + (level.field_9229.method_43057() - .5f) * .5f, 0, 1 / 8f, 0);
		}

		@Override
		public void morphAirFlow(AirFlowParticleAccess particleAccess, class_5819 random) {
			particleAccess.setColor(Color.mixColors(0x4499FF, 0x2277FF, random.method_43057()));
			particleAccess.setAlpha(1f);
			if (random.method_43057() < 1 / 32f)
				particleAccess.spawnExtraParticle(class_2398.field_11247, .125f);
			if (random.method_43057() < 1 / 32f)
				particleAccess.spawnExtraParticle(class_2398.field_11241, .125f);
		}

		@Override
		public void affectEntity(class_1297 entity, class_1937 level) {
			if (level.field_9236)
				return;

			if (entity instanceof class_1560 || entity.method_5864() == class_1299.field_6047
				|| entity.method_5864() == class_1299.field_6099) {
				entity.method_5643(entity.method_48923().method_48824(), 2);
			}
			if (entity.method_5809()) {
				entity.method_5646();
				level.method_8396(null, entity.method_24515(), class_3417.field_15222,
					class_3419.field_15254, 0.7F, 1.6F + (level.field_9229.method_43057() - level.field_9229.method_43057()) * 0.4F);
			}
		}
	}
}
