package com.simibubi.create.content.kinetics.mechanicalArm;

import java.util.Optional;
import java.util.function.Function;

import org.apache.commons.lang3.mutable.MutableBoolean;

import com.simibubi.create.AllBlocks;
import com.simibubi.create.Create;
import com.simibubi.create.content.kinetics.base.KineticBlockEntity;
import com.simibubi.create.content.kinetics.belt.BeltBlockEntity;
import com.simibubi.create.content.kinetics.belt.BeltHelper;
import com.simibubi.create.content.kinetics.belt.behaviour.TransportedItemStackHandlerBehaviour;
import com.simibubi.create.content.kinetics.belt.behaviour.TransportedItemStackHandlerBehaviour.TransportedResult;
import com.simibubi.create.content.kinetics.crafter.MechanicalCrafterBlock;
import com.simibubi.create.content.kinetics.crafter.MechanicalCrafterBlockEntity;
import com.simibubi.create.content.kinetics.deployer.DeployerBlock;
import com.simibubi.create.content.kinetics.saw.SawBlock;
import com.simibubi.create.content.logistics.chute.AbstractChuteBlock;
import com.simibubi.create.content.logistics.funnel.AbstractFunnelBlock;
import com.simibubi.create.content.logistics.funnel.BeltFunnelBlock;
import com.simibubi.create.content.logistics.funnel.BeltFunnelBlock.Shape;
import com.simibubi.create.content.logistics.funnel.FunnelBlock;
import com.simibubi.create.content.logistics.funnel.FunnelBlockEntity;
import com.simibubi.create.content.logistics.tunnel.BeltTunnelBlock;
import com.simibubi.create.content.processing.burner.BlazeBurnerBlock;
import com.simibubi.create.foundation.blockEntity.behaviour.BlockEntityBehaviour;
import com.simibubi.create.foundation.blockEntity.behaviour.filtering.FilteringBehaviour;
import com.simibubi.create.foundation.blockEntity.behaviour.inventory.InvManipulationBehaviour;
import com.simibubi.create.foundation.item.SmartInventory;
import com.simibubi.create.foundation.utility.VecHelper;

import io.github.fabricators_of_create.porting_lib.transfer.callbacks.TransactionCallback;
import io.github.fabricators_of_create.porting_lib.transfer.item.ItemHandlerHelper;
import net.fabricmc.fabric.api.transfer.v1.transaction.TransactionContext;
import net.minecraft.class_1264;
import net.minecraft.class_1271;
import net.minecraft.class_1792;
import net.minecraft.class_1799;
import net.minecraft.class_1802;
import net.minecraft.class_1813;
import net.minecraft.class_1937;
import net.minecraft.class_2246;
import net.minecraft.class_2338;
import net.minecraft.class_2350;
import net.minecraft.class_2382;
import net.minecraft.class_2387;
import net.minecraft.class_243;
import net.minecraft.class_2586;
import net.minecraft.class_2619;
import net.minecraft.class_2680;
import net.minecraft.class_2741;
import net.minecraft.class_2960;
import net.minecraft.class_3920;
import net.minecraft.class_3922;
import net.minecraft.class_3924;
import net.minecraft.class_4969;

public class AllArmInteractionPointTypes {

	public static final BasinType BASIN = register("basin", BasinType::new);
	public static final BeltType BELT = register("belt", BeltType::new);
	public static final BlazeBurnerType BLAZE_BURNER = register("blaze_burner", BlazeBurnerType::new);
	public static final ChuteType CHUTE = register("chute", ChuteType::new);
	public static final CrafterType CRAFTER = register("crafter", CrafterType::new);
	public static final CrushingWheelsType CRUSHING_WHEELS = register("crushing_wheels", CrushingWheelsType::new);
	public static final DeployerType DEPLOYER = register("deployer", DeployerType::new);
	public static final DepotType DEPOT = register("depot", DepotType::new);
	public static final FunnelType FUNNEL = register("funnel", FunnelType::new);
	public static final MillstoneType MILLSTONE = register("millstone", MillstoneType::new);
	public static final SawType SAW = register("saw", SawType::new);

	public static final CampfireType CAMPFIRE = register("campfire", CampfireType::new);
	public static final ComposterType COMPOSTER = register("composter", ComposterType::new);
	public static final JukeboxType JUKEBOX = register("jukebox", JukeboxType::new);
	public static final RespawnAnchorType RESPAWN_ANCHOR = register("respawn_anchor", RespawnAnchorType::new);

	private static <T extends ArmInteractionPointType> T register(String id, Function<class_2960, T> factory) {
		T type = factory.apply(Create.asResource(id));
		ArmInteractionPointType.register(type);
		return type;
	}

	public static void register() {}

	//

	public static class BasinType extends ArmInteractionPointType {
		public BasinType(class_2960 id) {
			super(id);
		}

		@Override
		public boolean canCreatePoint(class_1937 level, class_2338 pos, class_2680 state) {
			return AllBlocks.BASIN.has(state);
		}

		@Override
		public ArmInteractionPoint createPoint(class_1937 level, class_2338 pos, class_2680 state) {
			return new ArmInteractionPoint(this, level, pos, state);
		}
	}

	public static class BeltType extends ArmInteractionPointType {
		public BeltType(class_2960 id) {
			super(id);
		}

		@Override
		public boolean canCreatePoint(class_1937 level, class_2338 pos, class_2680 state) {
			return AllBlocks.BELT.has(state) && !(level.method_8320(pos.method_10084())
				.method_26204() instanceof BeltTunnelBlock);
		}

		@Override
		public ArmInteractionPoint createPoint(class_1937 level, class_2338 pos, class_2680 state) {
			return new BeltPoint(this, level, pos, state);
		}
	}

	public static class BlazeBurnerType extends ArmInteractionPointType {
		public BlazeBurnerType(class_2960 id) {
			super(id);
		}

		@Override
		public boolean canCreatePoint(class_1937 level, class_2338 pos, class_2680 state) {
			return AllBlocks.BLAZE_BURNER.has(state);
		}

		@Override
		public ArmInteractionPoint createPoint(class_1937 level, class_2338 pos, class_2680 state) {
			return new BlazeBurnerPoint(this, level, pos, state);
		}
	}

	public static class ChuteType extends ArmInteractionPointType {
		public ChuteType(class_2960 id) {
			super(id);
		}

		@Override
		public boolean canCreatePoint(class_1937 level, class_2338 pos, class_2680 state) {
			return AbstractChuteBlock.isChute(state);
		}

		@Override
		public ArmInteractionPoint createPoint(class_1937 level, class_2338 pos, class_2680 state) {
			return new TopFaceArmInteractionPoint(this, level, pos, state);
		}
	}

	public static class CrafterType extends ArmInteractionPointType {
		public CrafterType(class_2960 id) {
			super(id);
		}

		@Override
		public boolean canCreatePoint(class_1937 level, class_2338 pos, class_2680 state) {
			return AllBlocks.MECHANICAL_CRAFTER.has(state);
		}

		@Override
		public ArmInteractionPoint createPoint(class_1937 level, class_2338 pos, class_2680 state) {
			return new CrafterPoint(this, level, pos, state);
		}
	}

	public static class CrushingWheelsType extends ArmInteractionPointType {
		public CrushingWheelsType(class_2960 id) {
			super(id);
		}

		@Override
		public boolean canCreatePoint(class_1937 level, class_2338 pos, class_2680 state) {
			return AllBlocks.CRUSHING_WHEEL_CONTROLLER.has(state);
		}

		@Override
		public ArmInteractionPoint createPoint(class_1937 level, class_2338 pos, class_2680 state) {
			return new TopFaceArmInteractionPoint(this, level, pos, state);
		}
	}

	public static class DeployerType extends ArmInteractionPointType {
		public DeployerType(class_2960 id) {
			super(id);
		}

		@Override
		public boolean canCreatePoint(class_1937 level, class_2338 pos, class_2680 state) {
			return AllBlocks.DEPLOYER.has(state);
		}

		@Override
		public ArmInteractionPoint createPoint(class_1937 level, class_2338 pos, class_2680 state) {
			return new DeployerPoint(this, level, pos, state);
		}
	}

	public static class DepotType extends ArmInteractionPointType {
		public DepotType(class_2960 id) {
			super(id);
		}

		@Override
		public boolean canCreatePoint(class_1937 level, class_2338 pos, class_2680 state) {
			return AllBlocks.DEPOT.has(state) || AllBlocks.WEIGHTED_EJECTOR.has(state)
				|| AllBlocks.TRACK_STATION.has(state);
		}

		@Override
		public ArmInteractionPoint createPoint(class_1937 level, class_2338 pos, class_2680 state) {
			return new DepotPoint(this, level, pos, state);
		}
	}

	public static class FunnelType extends ArmInteractionPointType {
		public FunnelType(class_2960 id) {
			super(id);
		}

		@Override
		public boolean canCreatePoint(class_1937 level, class_2338 pos, class_2680 state) {
			return state.method_26204() instanceof AbstractFunnelBlock
				&& !(state.method_28498(FunnelBlock.EXTRACTING) && state.method_11654(FunnelBlock.EXTRACTING))
				&& !(state.method_28498(BeltFunnelBlock.SHAPE)
					&& state.method_11654(BeltFunnelBlock.SHAPE) == Shape.PUSHING);
		}

		@Override
		public ArmInteractionPoint createPoint(class_1937 level, class_2338 pos, class_2680 state) {
			return new FunnelPoint(this, level, pos, state);
		}
	}

	public static class MillstoneType extends ArmInteractionPointType {
		public MillstoneType(class_2960 id) {
			super(id);
		}

		@Override
		public boolean canCreatePoint(class_1937 level, class_2338 pos, class_2680 state) {
			return AllBlocks.MILLSTONE.has(state);
		}

		@Override
		public ArmInteractionPoint createPoint(class_1937 level, class_2338 pos, class_2680 state) {
			return new ArmInteractionPoint(this, level, pos, state);
		}
	}

	public static class SawType extends ArmInteractionPointType {
		public SawType(class_2960 id) {
			super(id);
		}

		@Override
		public boolean canCreatePoint(class_1937 level, class_2338 pos, class_2680 state) {
			return AllBlocks.MECHANICAL_SAW.has(state) && state.method_11654(SawBlock.FACING) == class_2350.field_11036
				&& ((KineticBlockEntity) level.method_8321(pos)).getSpeed() != 0;
		}

		@Override
		public ArmInteractionPoint createPoint(class_1937 level, class_2338 pos, class_2680 state) {
			return new DepotPoint(this, level, pos, state);
		}
	}

	public static class CampfireType extends ArmInteractionPointType {
		public CampfireType(class_2960 id) {
			super(id);
		}

		@Override
		public boolean canCreatePoint(class_1937 level, class_2338 pos, class_2680 state) {
			return state.method_26204() instanceof class_3922;
		}

		@Override
		public ArmInteractionPoint createPoint(class_1937 level, class_2338 pos, class_2680 state) {
			return new CampfirePoint(this, level, pos, state);
		}
	}

	public static class ComposterType extends ArmInteractionPointType {
		public ComposterType(class_2960 id) {
			super(id);
		}

		@Override
		public boolean canCreatePoint(class_1937 level, class_2338 pos, class_2680 state) {
			return state.method_27852(class_2246.field_17563);
		}

		@Override
		public ArmInteractionPoint createPoint(class_1937 level, class_2338 pos, class_2680 state) {
			return new ComposterPoint(this, level, pos, state);
		}
	}

	public static class JukeboxType extends ArmInteractionPointType {
		public JukeboxType(class_2960 id) {
			super(id);
		}

		@Override
		public boolean canCreatePoint(class_1937 level, class_2338 pos, class_2680 state) {
			return state.method_27852(class_2246.field_10223);
		}

		@Override
		public ArmInteractionPoint createPoint(class_1937 level, class_2338 pos, class_2680 state) {
			return new JukeboxPoint(this, level, pos, state);
		}
	}

	public static class RespawnAnchorType extends ArmInteractionPointType {
		public RespawnAnchorType(class_2960 id) {
			super(id);
		}

		@Override
		public boolean canCreatePoint(class_1937 level, class_2338 pos, class_2680 state) {
			return state.method_27852(class_2246.field_23152);
		}

		@Override
		public ArmInteractionPoint createPoint(class_1937 level, class_2338 pos, class_2680 state) {
			return new RespawnAnchorPoint(this, level, pos, state);
		}
	}

	//

	public static class DepositOnlyArmInteractionPoint extends ArmInteractionPoint {
		public DepositOnlyArmInteractionPoint(ArmInteractionPointType type, class_1937 level, class_2338 pos,
			class_2680 state) {
			super(type, level, pos, state);
		}

		@Override
		public void cycleMode() {}

		@Override
		public class_1799 extract(int amount, TransactionContext ctx) {
			return class_1799.field_8037;
		}
	}

	public static class TopFaceArmInteractionPoint extends ArmInteractionPoint {
		public TopFaceArmInteractionPoint(ArmInteractionPointType type, class_1937 level, class_2338 pos, class_2680 state) {
			super(type, level, pos, state);
		}

		@Override
		protected class_243 getInteractionPositionVector() {
			return class_243.method_24954(pos)
				.method_1031(.5f, 1, .5f);
		}
	}

	public static class BeltPoint extends DepotPoint {
		public BeltPoint(ArmInteractionPointType type, class_1937 level, class_2338 pos, class_2680 state) {
			super(type, level, pos, state);
		}

		@Override
		public void keepAlive() {
			super.keepAlive();
			BeltBlockEntity beltBE = BeltHelper.getSegmentBE(level, pos);
			if (beltBE == null)
				return;
			TransportedItemStackHandlerBehaviour transport =
				beltBE.getBehaviour(TransportedItemStackHandlerBehaviour.TYPE);
			if (transport == null)
				return;
			MutableBoolean found = new MutableBoolean(false);
			transport.handleProcessingOnAllItems(tis -> {
				if (found.isTrue())
					return TransportedResult.doNothing();
				tis.lockedExternally = true;
				found.setTrue();
				return TransportedResult.doNothing();
			});
		}
	}

	public static class BlazeBurnerPoint extends DepositOnlyArmInteractionPoint {
		public BlazeBurnerPoint(ArmInteractionPointType type, class_1937 level, class_2338 pos, class_2680 state) {
			super(type, level, pos, state);
		}

		@Override
		public class_1799 insert(class_1799 stack, TransactionContext ctx) {
			class_1799 input = stack.method_7972();
			class_1271<class_1799> res =
				BlazeBurnerBlock.tryInsert(cachedState, level, pos, input, false, false, ctx);
			class_1799 remainder = res.method_5466();
			if (input.method_7960()) {
				return remainder;
			} else {
				TransactionCallback.onSuccess(ctx, () ->
						class_1264.method_5449(level, pos.method_10263(), pos.method_10264(), pos.method_10260(), remainder));
				return input;
			}
		}
	}

	public static class CrafterPoint extends ArmInteractionPoint {
		public CrafterPoint(ArmInteractionPointType type, class_1937 level, class_2338 pos, class_2680 state) {
			super(type, level, pos, state);
		}

		@Override
		protected class_2350 getInteractionDirection() {
			return cachedState.method_28500(MechanicalCrafterBlock.HORIZONTAL_FACING)
				.orElse(class_2350.field_11035)
				.method_10153();
		}

		@Override
		protected class_243 getInteractionPositionVector() {
			return super.getInteractionPositionVector().method_1019(class_243.method_24954(getInteractionDirection().method_10163())
				.method_1021(.5f));
		}

		@Override
		public void updateCachedState() {
			class_2680 oldState = cachedState;
			super.updateCachedState();
			if (oldState != cachedState)
				cachedAngles = null;
		}

		@Override
		public class_1799 extract(int amount, TransactionContext ctx) {
			class_2586 be = level.method_8321(pos);
			if (!(be instanceof MechanicalCrafterBlockEntity))
				return class_1799.field_8037;
			MechanicalCrafterBlockEntity crafter = (MechanicalCrafterBlockEntity) be;
			SmartInventory inventory = crafter.getInventory();
			inventory.allowExtraction();
			class_1799 extract = super.extract(amount, ctx);
			inventory.forbidExtraction();
			return extract;
		}
	}

	public static class DeployerPoint extends ArmInteractionPoint {
		public DeployerPoint(ArmInteractionPointType type, class_1937 level, class_2338 pos, class_2680 state) {
			super(type, level, pos, state);
		}

		@Override
		protected class_2350 getInteractionDirection() {
			return cachedState.method_28500(DeployerBlock.FACING)
				.orElse(class_2350.field_11036)
				.method_10153();
		}

		@Override
		protected class_243 getInteractionPositionVector() {
			return super.getInteractionPositionVector().method_1019(class_243.method_24954(getInteractionDirection().method_10163())
				.method_1021(.65f));
		}

		@Override
		public void updateCachedState() {
			class_2680 oldState = cachedState;
			super.updateCachedState();
			if (oldState != cachedState)
				cachedAngles = null;
		}
	}

	public static class DepotPoint extends ArmInteractionPoint {
		public DepotPoint(ArmInteractionPointType type, class_1937 level, class_2338 pos, class_2680 state) {
			super(type, level, pos, state);
		}

		@Override
		protected class_243 getInteractionPositionVector() {
			return class_243.method_24954(pos)
				.method_1031(.5f, 14 / 16f, .5f);
		}
	}

	public static class FunnelPoint extends DepositOnlyArmInteractionPoint {
		public FunnelPoint(ArmInteractionPointType type, class_1937 level, class_2338 pos, class_2680 state) {
			super(type, level, pos, state);
		}

		@Override
		protected class_243 getInteractionPositionVector() {
			class_2350 funnelFacing = FunnelBlock.getFunnelFacing(cachedState);
			class_2382 normal = funnelFacing != null ? funnelFacing.method_10163() : class_2382.field_11176;
			return VecHelper.getCenterOf(pos)
				.method_1019(class_243.method_24954(normal)
					.method_1021(-.15f));
		}

		@Override
		protected class_2350 getInteractionDirection() {
			class_2350 funnelFacing = FunnelBlock.getFunnelFacing(cachedState);
			return funnelFacing != null ? funnelFacing.method_10153() : class_2350.field_11036;
		}

		@Override
		public void updateCachedState() {
			class_2680 oldState = cachedState;
			super.updateCachedState();
			if (oldState != cachedState)
				cachedAngles = null;
		}

		@Override
		public class_1799 insert(class_1799 stack, TransactionContext ctx) {
			FilteringBehaviour filtering = BlockEntityBehaviour.get(level, pos, FilteringBehaviour.TYPE);
			InvManipulationBehaviour inserter = BlockEntityBehaviour.get(level, pos, InvManipulationBehaviour.TYPE);
			if (cachedState.method_28500(class_2741.field_12484)
				.orElse(false))
				return stack;
			if (inserter == null)
				return stack;
			if (filtering != null && !filtering.test(stack))
				return stack;

			// fabric: this is already wrapped in a transaction, no need to simulate
			class_1799 insert = inserter.insert(stack);
			if (insert.method_7947() != stack.method_7947()) {
				class_2586 blockEntity = level.method_8321(pos);
				if (blockEntity instanceof FunnelBlockEntity) {
					FunnelBlockEntity funnelBlockEntity = (FunnelBlockEntity) blockEntity;
					TransactionCallback.onSuccess(ctx, () -> {
						funnelBlockEntity.onTransfer(stack);
						if (funnelBlockEntity.hasFlap())
							funnelBlockEntity.flap(true);
					});
				}
			}
			return insert;
		}
	}

	public static class CampfirePoint extends DepositOnlyArmInteractionPoint {
		public CampfirePoint(ArmInteractionPointType type, class_1937 level, class_2338 pos, class_2680 state) {
			super(type, level, pos, state);
		}

		@Override
		public class_1799 insert(class_1799 stack, TransactionContext ctx) {
			class_2586 blockEntity = level.method_8321(pos);
			if (!(blockEntity instanceof class_3924 campfireBE))
				return stack;
			Optional<class_3920> recipe = campfireBE.method_17502(stack);
			if (recipe.isEmpty())
				return stack;
			boolean hasSpace = false;
			for (class_1799 campfireStack : campfireBE.method_17505()) {
				if (campfireStack.method_7960()) {
					hasSpace = true;
					break;
				}
			}
			if (!hasSpace)
				return stack;
			class_1799 inserted = ItemHandlerHelper.copyStackWithSize(stack, 1);
			TransactionCallback.onSuccess(ctx, () -> campfireBE.method_17503(null, inserted, recipe.get().method_8167()));
			class_1799 remainder = stack.method_7972();
			remainder.method_7934(1);
			return remainder;
		}
	}

	public static class ComposterPoint extends ArmInteractionPoint {
		public ComposterPoint(ArmInteractionPointType type, class_1937 level, class_2338 pos, class_2680 state) {
			super(type, level, pos, state);
		}

		@Override
		protected class_243 getInteractionPositionVector() {
			return class_243.method_24954(pos)
				.method_1031(.5f, 13 / 16f, .5f);
		}

		// fabric: should not be needed since FAPI handles wrapping Containers
//		@Override
//		public void updateCachedState() {
//			BlockState oldState = cachedState;
//			super.updateCachedState();
//			if (oldState != cachedState)
//				cachedHandler.invalidate();
//		}
//
//		@Nullable
//		@Override
//		protected IItemHandler getHandler() {
//			return null;
//		}
//
//		protected WorldlyContainer getContainer() {
//			ComposterBlock composterBlock = (ComposterBlock) Blocks.COMPOSTER;
//			return composterBlock.getContainer(cachedState, level, pos);
//		}
//
//		@Override
//		public ItemStack insert(ItemStack stack, boolean simulate) {
//			IItemHandler handler = new SidedInvWrapper(getContainer(), Direction.UP);
//			return ItemHandlerHelper.insertItem(handler, stack, simulate);
//		}
//
//		@Override
//		public ItemStack extract(int slot, int amount, boolean simulate) {
//			IItemHandler handler = new SidedInvWrapper(getContainer(), Direction.DOWN);
//			return handler.extractItem(slot, amount, simulate);
//		}
//
//		@Override
//		public int getSlotCount() {
//			return 2;
//		}
	}

	public static class JukeboxPoint extends TopFaceArmInteractionPoint {
		public JukeboxPoint(ArmInteractionPointType type, class_1937 level, class_2338 pos, class_2680 state) {
			super(type, level, pos, state);
		}

		@Override
		public class_1799 insert(class_1799 stack, TransactionContext ctx) {
			class_1792 item = stack.method_7909();
			if (!(item instanceof class_1813))
				return stack;
			if (cachedState.method_28500(class_2387.field_11180)
				.orElse(true))
				return stack;
			class_2586 blockEntity = level.method_8321(pos);
			if (!(blockEntity instanceof class_2619 jukeboxBE))
				return stack;
			if (!jukeboxBE.method_49274()
				.method_7960())
				return stack;
			class_1799 remainder = stack.method_7972();
			class_1799 toInsert = remainder.method_7971(1);
			level.updateSnapshots(ctx);
			level.method_8652(pos, cachedState.method_11657(class_2387.field_11180, true), 2);
			TransactionCallback.onSuccess(ctx, () -> {
				jukeboxBE.method_49275(toInsert);
				level.method_8444(null, 1010, pos, class_1792.method_7880(item));
			});
			return remainder;
		}

		@Override
		public class_1799 extract(int amount, TransactionContext ctx) {
			if (!cachedState.method_28500(class_2387.field_11180)
				.orElse(false))
				return class_1799.field_8037;
			class_2586 blockEntity = level.method_8321(pos);
			if (!(blockEntity instanceof class_2619 jukeboxBE))
				return class_1799.field_8037;
			class_1799 record = jukeboxBE.method_49274();
			if (record.method_7960())
				return class_1799.field_8037;
			level.updateSnapshots(ctx);
			level.method_8652(pos, cachedState.method_11657(class_2387.field_11180, false), 2);
			TransactionCallback.onSuccess(ctx, () -> {
				level.method_20290(1010, pos, 0);
				jukeboxBE.method_5448();
			});
			return record;
		}
	}

	public static class RespawnAnchorPoint extends DepositOnlyArmInteractionPoint {
		public RespawnAnchorPoint(ArmInteractionPointType type, class_1937 level, class_2338 pos, class_2680 state) {
			super(type, level, pos, state);
		}

		@Override
		protected class_243 getInteractionPositionVector() {
			return class_243.method_24954(pos)
				.method_1031(.5f, 1, .5f);
		}

		@Override
		public class_1799 insert(class_1799 stack, TransactionContext ctx) {
			if (!stack.method_31574(class_1802.field_8801))
				return stack;
			if (cachedState.method_28500(class_4969.field_23153)
				.orElse(4) == 4)
				return stack;
			TransactionCallback.onSuccess(ctx, () -> class_4969.method_26382(null, level, pos, cachedState));
			class_1799 remainder = stack.method_7972();
			remainder.method_7934(1);
			return remainder;
		}
	}

}
