package com.simibubi.create.content.kinetics.mechanicalArm;

import java.util.ArrayList;
import java.util.List;

import javax.annotation.Nullable;

import com.jozufozu.flywheel.backend.instancing.InstancedRenderDispatcher;
import com.simibubi.create.Create;
import com.simibubi.create.content.contraptions.ITransformableBlockEntity;
import com.simibubi.create.content.contraptions.StructureTransform;
import com.simibubi.create.content.kinetics.base.KineticBlockEntity;
import com.simibubi.create.content.kinetics.mechanicalArm.AllArmInteractionPointTypes.JukeboxPoint;
import com.simibubi.create.content.kinetics.mechanicalArm.ArmInteractionPoint.Mode;
import com.simibubi.create.foundation.advancement.AllAdvancements;
import com.simibubi.create.foundation.blockEntity.behaviour.BlockEntityBehaviour;
import com.simibubi.create.foundation.blockEntity.behaviour.CenteredSideValueBoxTransform;
import com.simibubi.create.foundation.blockEntity.behaviour.scrollValue.INamedIconOptions;
import com.simibubi.create.foundation.blockEntity.behaviour.scrollValue.ScrollOptionBehaviour;
import com.simibubi.create.foundation.gui.AllIcons;
import com.simibubi.create.foundation.item.TooltipHelper;
import com.simibubi.create.foundation.utility.AngleHelper;
import com.simibubi.create.foundation.utility.Lang;
import com.simibubi.create.foundation.utility.NBTHelper;
import com.simibubi.create.foundation.utility.VecHelper;
import com.simibubi.create.foundation.utility.animation.LerpedFloat;
import com.simibubi.create.infrastructure.config.AllConfigs;

import io.github.fabricators_of_create.porting_lib.transfer.TransferUtil;
import io.github.fabricators_of_create.porting_lib.util.EnvExecutor;
import io.github.fabricators_of_create.porting_lib.util.ItemStackUtil;
import io.github.fabricators_of_create.porting_lib.util.NBTSerializer;

import net.fabricmc.api.EnvType;
import net.fabricmc.fabric.api.transfer.v1.transaction.Transaction;
import net.minecraft.class_1799;
import net.minecraft.class_1937;
import net.minecraft.class_2248;
import net.minecraft.class_2338;
import net.minecraft.class_2350;
import net.minecraft.class_238;
import net.minecraft.class_2387;
import net.minecraft.class_243;
import net.minecraft.class_2487;
import net.minecraft.class_2499;
import net.minecraft.class_2520;
import net.minecraft.class_2561;
import net.minecraft.class_2591;
import net.minecraft.class_2680;
import net.minecraft.class_2802;
import net.minecraft.class_3417;
import net.minecraft.class_3419;
import net.minecraft.class_3532;
import net.minecraft.class_4076;

public class ArmBlockEntity extends KineticBlockEntity implements ITransformableBlockEntity {

	// Server
	List<ArmInteractionPoint> inputs;
	List<ArmInteractionPoint> outputs;
	class_2499 interactionPointTag;

	// Both
	float chasedPointProgress;
	int chasedPointIndex;
	class_1799 heldItem;
	Phase phase;
	boolean goggles;

	// Client
	ArmAngleTarget previousTarget;
	LerpedFloat lowerArmAngle;
	LerpedFloat upperArmAngle;
	LerpedFloat baseAngle;
	LerpedFloat headAngle;
	LerpedFloat clawAngle;
	float previousBaseAngle;
	boolean updateInteractionPoints;
	int tooltipWarmup;

	//
	protected ScrollOptionBehaviour<SelectionMode> selectionMode;
	protected int lastInputIndex = -1;
	protected int lastOutputIndex = -1;
	protected boolean redstoneLocked;

	public enum Phase {
		SEARCH_INPUTS, MOVE_TO_INPUT, SEARCH_OUTPUTS, MOVE_TO_OUTPUT, DANCING
	}

	public ArmBlockEntity(class_2591<?> typeIn, class_2338 pos, class_2680 state) {
		super(typeIn, pos, state);
		inputs = new ArrayList<>();
		outputs = new ArrayList<>();
		interactionPointTag = null;
		heldItem = class_1799.field_8037;
		phase = Phase.SEARCH_INPUTS;
		previousTarget = ArmAngleTarget.NO_TARGET;
		baseAngle = LerpedFloat.angular();
		baseAngle.startWithValue(previousTarget.baseAngle);
		lowerArmAngle = LerpedFloat.angular();
		lowerArmAngle.startWithValue(previousTarget.lowerArmAngle);
		upperArmAngle = LerpedFloat.angular();
		upperArmAngle.startWithValue(previousTarget.upperArmAngle);
		headAngle = LerpedFloat.angular();
		headAngle.startWithValue(previousTarget.headAngle);
		clawAngle = LerpedFloat.angular();
		previousBaseAngle = previousTarget.baseAngle;
		updateInteractionPoints = true;
		redstoneLocked = false;
		tooltipWarmup = 15;
		goggles = false;
	}

	@Override
	public void addBehaviours(List<BlockEntityBehaviour> behaviours) {
		super.addBehaviours(behaviours);

		selectionMode = new ScrollOptionBehaviour<SelectionMode>(SelectionMode.class,
				Lang.translateDirect("logistics.when_multiple_outputs_available"), this, new SelectionModeValueBox());
		behaviours.add(selectionMode);

		registerAwardables(behaviours, AllAdvancements.ARM_BLAZE_BURNER, AllAdvancements.ARM_MANY_TARGETS,
				AllAdvancements.MECHANICAL_ARM, AllAdvancements.MUSICAL_ARM);
	}

	@Override
	public void tick() {
		super.tick();
		initInteractionPoints();
		boolean targetReached = tickMovementProgress();

		if (tooltipWarmup > 0)
			tooltipWarmup--;
		if (chasedPointProgress < 1) {
			if (phase == Phase.MOVE_TO_INPUT) {
				ArmInteractionPoint point = getTargetedInteractionPoint();
				if (point != null)
					point.keepAlive();
			}
			return;
		}
		if (field_11863.field_9236)
			return;

		if (phase == Phase.MOVE_TO_INPUT)
			collectItem();
		else if (phase == Phase.MOVE_TO_OUTPUT)
			depositItem();
		else if (phase == Phase.SEARCH_INPUTS || phase == Phase.DANCING)
			searchForItem();

		if (targetReached)
			lazyTick();
	}

	@Override
	public void lazyTick() {
		super.lazyTick();

		if (field_11863.field_9236)
			return;
		if (chasedPointProgress < .5f)
			return;
		if (phase == Phase.SEARCH_INPUTS || phase == Phase.DANCING)
			checkForMusic();
		if (phase == Phase.SEARCH_OUTPUTS)
			searchForDestination();
	}

	private void checkForMusic() {
		boolean hasMusic = checkForMusicAmong(inputs) || checkForMusicAmong(outputs);
		if (hasMusic != (phase == Phase.DANCING)) {
			phase = hasMusic ? Phase.DANCING : Phase.SEARCH_INPUTS;
			method_5431();
			sendData();
		}
	}

	@Override
	protected class_238 createRenderBoundingBox() {
		return super.createRenderBoundingBox().method_1014(3);
	}

	private boolean checkForMusicAmong(List<ArmInteractionPoint> list) {
		for (ArmInteractionPoint armInteractionPoint : list) {
			if (!(armInteractionPoint instanceof AllArmInteractionPointTypes.JukeboxPoint))
				continue;
			class_2680 state = field_11863.method_8320(armInteractionPoint.getPos());
			if (state.method_28500(class_2387.field_11180)
					.orElse(false))
				return true;
		}
		return false;
	}

	private boolean tickMovementProgress() {
		boolean targetReachedPreviously = chasedPointProgress >= 1;
		chasedPointProgress += Math.min(256, Math.abs(getSpeed())) / 1024f;
		if (chasedPointProgress > 1)
			chasedPointProgress = 1;
		if (!field_11863.field_9236)
			return !targetReachedPreviously && chasedPointProgress >= 1;

		ArmInteractionPoint targetedInteractionPoint = getTargetedInteractionPoint();
		ArmAngleTarget previousTarget = this.previousTarget;
		ArmAngleTarget target = targetedInteractionPoint == null ? ArmAngleTarget.NO_TARGET
				: targetedInteractionPoint.getTargetAngles(field_11867, isOnCeiling());

		baseAngle.setValue(AngleHelper.angleLerp(chasedPointProgress, previousBaseAngle,
				target == ArmAngleTarget.NO_TARGET ? previousBaseAngle : target.baseAngle));

		// Arm's angles first backup to resting position and then continue
		if (chasedPointProgress < .5f)
			target = ArmAngleTarget.NO_TARGET;
		else
			previousTarget = ArmAngleTarget.NO_TARGET;
		float progress = chasedPointProgress == 1 ? 1 : (chasedPointProgress % .5f) * 2;

		lowerArmAngle.setValue(class_3532.method_16439(progress, previousTarget.lowerArmAngle, target.lowerArmAngle));
		upperArmAngle.setValue(class_3532.method_16439(progress, previousTarget.upperArmAngle, target.upperArmAngle));
		headAngle.setValue(AngleHelper.angleLerp(progress, previousTarget.headAngle % 360, target.headAngle % 360));

		return false;
	}

	protected boolean isOnCeiling() {
		class_2680 state = method_11010();
		return method_11002() && state.method_28500(ArmBlock.CEILING)
				.orElse(false);
	}

	@Override
	public void destroy() {
		super.destroy();
		if (!heldItem.method_7960())
			class_2248.method_9577(field_11863, field_11867, heldItem);
	}

	@Nullable
	private ArmInteractionPoint getTargetedInteractionPoint() {
		if (chasedPointIndex == -1)
			return null;
		if (phase == Phase.MOVE_TO_INPUT && chasedPointIndex < inputs.size())
			return inputs.get(chasedPointIndex);
		if (phase == Phase.MOVE_TO_OUTPUT && chasedPointIndex < outputs.size())
			return outputs.get(chasedPointIndex);
		return null;
	}

	protected void searchForItem() {
		if (redstoneLocked)
			return;

		boolean foundInput = false;
		// for round robin, we start looking after the last used index, for default we
		// start at 0;
		int startIndex = selectionMode.get() == SelectionMode.PREFER_FIRST ? 0 : lastInputIndex + 1;

		// if we enforce round robin, only look at the next input in the list,
		// otherwise, look at all inputs
		int scanRange = selectionMode.get() == SelectionMode.FORCED_ROUND_ROBIN ? lastInputIndex + 2 : inputs.size();
		if (scanRange > inputs.size())
			scanRange = inputs.size();

		InteractionPoints:
		for (int i = startIndex; i < scanRange; i++) {
			ArmInteractionPoint armInteractionPoint = inputs.get(i);
			if (!armInteractionPoint.isValid())
				continue;

			if (getDistributableAmount(armInteractionPoint) == 0)
				continue;

			selectIndex(true, i);
			foundInput = true;
			break InteractionPoints;
		}
		if (!foundInput && selectionMode.get() == SelectionMode.ROUND_ROBIN) {
			// if we didn't find an input, but don't want to enforce round robin, reset the
			// last index
			lastInputIndex = -1;
		}
		if (lastInputIndex == inputs.size() - 1) {
			// if we reached the last input in the list, reset the last index
			lastInputIndex = -1;
		}
	}

	protected void searchForDestination() {
		class_1799 held = heldItem.method_7972();

		boolean foundOutput = false;
		// for round robin, we start looking after the last used index, for default we
		// start at 0;
		int startIndex = selectionMode.get() == SelectionMode.PREFER_FIRST ? 0 : lastOutputIndex + 1;

		// if we enforce round robin, only look at the next index in the list,
		// otherwise, look at all
		int scanRange = selectionMode.get() == SelectionMode.FORCED_ROUND_ROBIN ? lastOutputIndex + 2 : outputs.size();
		if (scanRange > outputs.size())
			scanRange = outputs.size();

		try (Transaction t = TransferUtil.getTransaction()) {
			for (int i = startIndex; i < scanRange; i++) {
				ArmInteractionPoint armInteractionPoint = outputs.get(i);
				if (!armInteractionPoint.isValid())
					continue;

				class_1799 remainder = armInteractionPoint.insert(held, t);
				if (class_1799.method_7973(remainder, heldItem))
					continue;

				selectIndex(false, i);
				foundOutput = true;
				break;
			}
		}

		if (!foundOutput && selectionMode.get() == SelectionMode.ROUND_ROBIN) {
			// if we didn't find an input, but don't want to enforce round robin, reset the
			// last index
			lastOutputIndex = -1;
		}
		if (lastOutputIndex == outputs.size() - 1) {
			// if we reached the last input in the list, reset the last index
			lastOutputIndex = -1;
		}
	}

	// input == true => select input, false => select output
	private void selectIndex(boolean input, int index) {
		phase = input ? Phase.MOVE_TO_INPUT : Phase.MOVE_TO_OUTPUT;
		chasedPointIndex = index;
		chasedPointProgress = 0;
		if (input)
			lastInputIndex = index;
		else
			lastOutputIndex = index;
		sendData();
		method_5431();
	}

	protected int getDistributableAmount(ArmInteractionPoint armInteractionPoint) {
		try (Transaction t = TransferUtil.getTransaction()) {
			class_1799 stack = armInteractionPoint.extract(t);

			class_1799 remainder = stack.method_7960() ? stack : simulateInsertion(stack);
			if (class_1799.method_7984(stack, remainder)) {
				return stack.method_7947() - remainder.method_7947();
			} else {
				return stack.method_7947();
			}
		}
	}

	private class_1799 simulateInsertion(class_1799 stack) {
		try (Transaction t = TransferUtil.getTransaction()) {
			for (ArmInteractionPoint armInteractionPoint : outputs) {
				if (armInteractionPoint.isValid())
					stack = armInteractionPoint.insert(stack, t);
				if (stack.method_7960())
					break;
			}
			return stack;
		}
	}

	protected void depositItem() {
		ArmInteractionPoint armInteractionPoint = getTargetedInteractionPoint();
		if (armInteractionPoint != null && armInteractionPoint.isValid()) {
			class_1799 toInsert = heldItem.method_7972();
			try (Transaction t = TransferUtil.getTransaction()) {
				class_1799 remainder = armInteractionPoint.insert(toInsert, t);
				t.commit();
				heldItem = remainder;

				if (armInteractionPoint instanceof JukeboxPoint && remainder.method_7960())
					award(AllAdvancements.MUSICAL_ARM);
			}
		}

		phase = heldItem.method_7960() ? Phase.SEARCH_INPUTS : Phase.SEARCH_OUTPUTS;
		chasedPointProgress = 0;
		chasedPointIndex = -1;
		sendData();
		method_5431();

		if (!field_11863.field_9236)
			award(AllAdvancements.MECHANICAL_ARM);
	}

	protected void collectItem() {
		ArmInteractionPoint armInteractionPoint = getTargetedInteractionPoint();
		if (armInteractionPoint != null && armInteractionPoint.isValid()) {
			try (Transaction t = TransferUtil.getTransaction()) {
				int amountExtracted = getDistributableAmount(armInteractionPoint);
				if (amountExtracted == 0)
					return;
				class_1799 prevHeld = heldItem;
				heldItem = armInteractionPoint.extract(amountExtracted, t);
				phase = Phase.SEARCH_OUTPUTS;
				chasedPointProgress = 0;
				chasedPointIndex = -1;
				sendData();
				method_5431();

				if (!class_1799.method_7984(heldItem, prevHeld))
					field_11863.method_8396(null, field_11867, class_3417.field_15197, class_3419.field_15245, .125f,
							.5f + Create.RANDOM.nextFloat() * .25f);
				t.commit();
				return;
			}
		}

		phase = Phase.SEARCH_INPUTS;
		chasedPointProgress = 0;
		chasedPointIndex = -1;
		sendData();
		method_5431();
	}

	public void redstoneUpdate() {
		if (field_11863.field_9236)
			return;
		boolean blockPowered = field_11863.method_49803(field_11867);
		if (blockPowered == redstoneLocked)
			return;
		redstoneLocked = blockPowered;
		sendData();
		if (!redstoneLocked)
			searchForItem();
	}

	@Override
	public void transform(StructureTransform transform) {
		if (interactionPointTag == null)
			return;

		for (class_2520 tag : interactionPointTag) {
			ArmInteractionPoint.transformPos((class_2487) tag, transform);
		}

		notifyUpdate();
	}

	// ClientLevel#hasChunk (and consequently #isAreaLoaded) always returns true,
	// so manually check the ChunkSource to avoid weird behavior on the client side
	protected boolean isAreaActuallyLoaded(class_2338 center, int range) {
		if (!field_11863.method_22343(center.method_10069(-range, -range, -range), center.method_10069(range, range, range))) {
			return false;
		}
		if (field_11863.field_9236) {
			int minY = center.method_10264() - range;
			int maxY = center.method_10264() + range;
			if (maxY < field_11863.method_31607() || minY >= field_11863.method_31600()) {
				return false;
			}

			int minX = center.method_10263() - range;
			int minZ = center.method_10260() - range;
			int maxX = center.method_10263() + range;
			int maxZ = center.method_10260() + range;

			int minChunkX = class_4076.method_18675(minX);
			int maxChunkX = class_4076.method_18675(maxX);
			int minChunkZ = class_4076.method_18675(minZ);
			int maxChunkZ = class_4076.method_18675(maxZ);

			class_2802 chunkSource = field_11863.method_8398();
			for (int chunkX = minChunkX; chunkX <= maxChunkX; ++chunkX) {
				for (int chunkZ = minChunkZ; chunkZ <= maxChunkZ; ++chunkZ) {
					if (!chunkSource.method_12123(chunkX, chunkZ)) {
						return false;
					}
				}
			}
		}
		return true;
	}

	protected void initInteractionPoints() {
		if (!updateInteractionPoints || interactionPointTag == null)
			return;
		if (!isAreaActuallyLoaded(field_11867, getRange() + 1))
			return;
		inputs.clear();
		outputs.clear();

		boolean hasBlazeBurner = false;
		for (class_2520 tag : interactionPointTag) {
			ArmInteractionPoint point = ArmInteractionPoint.deserialize((class_2487) tag, field_11863, field_11867);
			if (point == null)
				continue;
			if (point.getMode() == Mode.DEPOSIT)
				outputs.add(point);
			else if (point.getMode() == Mode.TAKE)
				inputs.add(point);
			hasBlazeBurner |= point instanceof AllArmInteractionPointTypes.BlazeBurnerPoint;
		}

		if (!field_11863.field_9236) {
			if (outputs.size() >= 10)
				award(AllAdvancements.ARM_MANY_TARGETS);
			if (hasBlazeBurner)
				award(AllAdvancements.ARM_BLAZE_BURNER);
		}

		updateInteractionPoints = false;
		sendData();
		method_5431();
	}

	public void writeInteractionPoints(class_2487 compound) {
		if (updateInteractionPoints && interactionPointTag != null) {
			compound.method_10566("InteractionPoints", interactionPointTag);
		} else {
			class_2499 pointsNBT = new class_2499();
			inputs.stream()
					.map(aip -> aip.serialize(field_11867))
					.forEach(pointsNBT::add);
			outputs.stream()
					.map(aip -> aip.serialize(field_11867))
					.forEach(pointsNBT::add);
			compound.method_10566("InteractionPoints", pointsNBT);
		}
	}

	@Override
	public void write(class_2487 compound, boolean clientPacket) {
		super.write(compound, clientPacket);

		writeInteractionPoints(compound);

		NBTHelper.writeEnum(compound, "Phase", phase);
		compound.method_10556("Powered", redstoneLocked);
		compound.method_10556("Goggles", goggles);
		compound.method_10566("HeldItem", NBTSerializer.serializeNBT(heldItem));
		compound.method_10569("TargetPointIndex", chasedPointIndex);
		compound.method_10548("MovementProgress", chasedPointProgress);
	}

	@Override
	public void writeSafe(class_2487 compound) {
		super.writeSafe(compound);

		writeInteractionPoints(compound);
	}

	@Override
	protected void read(class_2487 compound, boolean clientPacket) {
		int previousIndex = chasedPointIndex;
		Phase previousPhase = phase;
		class_2499 interactionPointTagBefore = interactionPointTag;

		super.read(compound, clientPacket);
		heldItem = class_1799.method_7915(compound.method_10562("HeldItem"));
		phase = NBTHelper.readEnum(compound, "Phase", Phase.class);
		chasedPointIndex = compound.method_10550("TargetPointIndex");
		chasedPointProgress = compound.method_10583("MovementProgress");
		interactionPointTag = compound.method_10554("InteractionPoints", class_2520.field_33260);
		redstoneLocked = compound.method_10577("Powered");

		boolean hadGoggles = goggles;
		goggles = compound.method_10577("Goggles");

		if (!clientPacket)
			return;

		if (hadGoggles != goggles)
			EnvExecutor.runWhenOn(EnvType.CLIENT, () -> () -> InstancedRenderDispatcher.enqueueUpdate(this));

		boolean ceiling = isOnCeiling();
		if (interactionPointTagBefore == null || interactionPointTagBefore.size() != interactionPointTag.size())
			updateInteractionPoints = true;
		if (previousIndex != chasedPointIndex || (previousPhase != phase)) {
			ArmInteractionPoint previousPoint = null;
			if (previousPhase == Phase.MOVE_TO_INPUT && previousIndex < inputs.size())
				previousPoint = inputs.get(previousIndex);
			if (previousPhase == Phase.MOVE_TO_OUTPUT && previousIndex < outputs.size())
				previousPoint = outputs.get(previousIndex);
			previousTarget = previousPoint == null ? ArmAngleTarget.NO_TARGET
					: previousPoint.getTargetAngles(field_11867, ceiling);
			if (previousPoint != null)
				previousBaseAngle = previousTarget.baseAngle;

			ArmInteractionPoint targetedPoint = getTargetedInteractionPoint();
			if (targetedPoint != null)
				targetedPoint.updateCachedState();
		}
	}

	public static int getRange() {
		return AllConfigs.server().logistics.mechanicalArmRange.get();
	}

	@Override
	public boolean addToTooltip(List<class_2561> tooltip, boolean isPlayerSneaking) {
		if (super.addToTooltip(tooltip, isPlayerSneaking))
			return true;
		if (isPlayerSneaking)
			return false;
		if (tooltipWarmup > 0)
			return false;
		if (!inputs.isEmpty())
			return false;
		if (!outputs.isEmpty())
			return false;

		TooltipHelper.addHint(tooltip, "hint.mechanical_arm_no_targets");
		return true;
	}

	public void method_31662(class_1937 level) {
		super.method_31662(level);
		for (ArmInteractionPoint input : inputs) {
			input.setLevel(level);
		}
		for (ArmInteractionPoint output : outputs) {
			output.setLevel(level);
		}
	}

	private class SelectionModeValueBox extends CenteredSideValueBoxTransform {

		public SelectionModeValueBox() {
			super((blockState, direction) -> !direction.method_10166()
					.method_10178());
		}

		@Override
		public class_243 getLocalOffset(class_2680 state) {
			int yPos = state.method_11654(ArmBlock.CEILING) ? 16 - 3 : 3;
			class_243 location = VecHelper.voxelSpace(8, yPos, 15.5);
			location = VecHelper.rotateCentered(location, AngleHelper.horizontalAngle(getSide()), class_2350.class_2351.field_11052);
			return location;
		}

		@Override
		public float getScale() {
			return super.getScale();
		}

	}

	public enum SelectionMode implements INamedIconOptions {
		ROUND_ROBIN(AllIcons.I_ARM_ROUND_ROBIN),
		FORCED_ROUND_ROBIN(AllIcons.I_ARM_FORCED_ROUND_ROBIN),
		PREFER_FIRST(AllIcons.I_ARM_PREFER_FIRST),

		;

		private final String translationKey;
		private final AllIcons icon;

		SelectionMode(AllIcons icon) {
			this.icon = icon;
			this.translationKey = "mechanical_arm.selection_mode." + Lang.asId(name());
		}

		@Override
		public AllIcons getIcon() {
			return icon;
		}

		@Override
		public String getTranslationKey() {
			return translationKey;
		}
	}

}
