package com.simibubi.create.content.kinetics.mechanicalArm;

import javax.annotation.Nullable;

import com.simibubi.create.content.contraptions.StructureTransform;
import com.simibubi.create.foundation.item.ItemHelper;
import com.simibubi.create.foundation.utility.NBTHelper;
import com.simibubi.create.foundation.utility.VecHelper;

import io.github.fabricators_of_create.porting_lib.transfer.TransferUtil;
import io.github.fabricators_of_create.porting_lib.transfer.item.ItemHandlerHelper;
import io.github.fabricators_of_create.porting_lib.util.StorageProvider;
import net.fabricmc.fabric.api.transfer.v1.item.ItemVariant;
import net.fabricmc.fabric.api.transfer.v1.storage.Storage;
import net.fabricmc.fabric.api.transfer.v1.transaction.TransactionContext;
import net.minecraft.class_1799;
import net.minecraft.class_1937;
import net.minecraft.class_2338;
import net.minecraft.class_2350;
import net.minecraft.class_243;
import net.minecraft.class_2487;
import net.minecraft.class_2512;
import net.minecraft.class_2680;
import net.minecraft.class_2960;

public class ArmInteractionPoint {

	protected final ArmInteractionPointType type;
	protected class_1937 level;
	protected final class_2338 pos;
	protected Mode mode = Mode.DEPOSIT;

	protected class_2680 cachedState;
	protected StorageProvider<ItemVariant> handlerProvider = null;
	protected ArmAngleTarget cachedAngles;

	public ArmInteractionPoint(ArmInteractionPointType type, class_1937 level, class_2338 pos, class_2680 state) {
		this.type = type;
		this.level = level;
		this.pos = pos;
		this.cachedState = state;
	}

	public ArmInteractionPointType getType() {
		return type;
	}

	public class_1937 getLevel() {
		return level;
	}

	public void setLevel(class_1937 level) {
		this.level = level;
	}

	public class_2338 getPos() {
		return pos;
	}

	public Mode getMode() {
		return mode;
	}

	public void cycleMode() {
		mode = mode == Mode.DEPOSIT ? Mode.TAKE : Mode.DEPOSIT;
	}

	protected class_243 getInteractionPositionVector() {
		return VecHelper.getCenterOf(pos);
	}

	protected class_2350 getInteractionDirection() {
		return class_2350.field_11033;
	}

	public ArmAngleTarget getTargetAngles(class_2338 armPos, boolean ceiling) {
		if (cachedAngles == null)
			cachedAngles =
				new ArmAngleTarget(armPos, getInteractionPositionVector(), getInteractionDirection(), ceiling);

		return cachedAngles;
	}

	public void updateCachedState() {
		cachedState = level.method_8320(pos);
	}

	public boolean isValid() {
		updateCachedState();
		return type.canCreatePoint(level, pos, cachedState);
	}

	public void keepAlive() {}

	@Nullable
	protected Storage<ItemVariant> getHandler() {
		if (handlerProvider == null) {
			handlerProvider = StorageProvider.createForItems(level, pos);
		}
		return handlerProvider.get(class_2350.field_11036);
	}

	public class_1799 insert(class_1799 stack, TransactionContext ctx) {
		Storage<ItemVariant> handler = getHandler();
		if (handler == null)
			return stack;
		long inserted = handler.insert(ItemVariant.of(stack), stack.method_7947(), ctx);
		return ItemHandlerHelper.copyStackWithSize(stack, ItemHelper.truncateLong(stack.method_7947() - inserted));
	}

	public class_1799 extract(int amount, TransactionContext ctx) {
		Storage<ItemVariant> handler = getHandler();
		if (handler == null)
			return class_1799.field_8037;
		return TransferUtil.extractAnyItem(handler, amount);
	}

	public class_1799 extract(TransactionContext ctx) {
		return extract(64, ctx);
	}

	protected void serialize(class_2487 nbt, class_2338 anchor) {
		NBTHelper.writeEnum(nbt, "Mode", mode);
	}

	protected void deserialize(class_2487 nbt, class_2338 anchor) {
		mode = NBTHelper.readEnum(nbt, "Mode", Mode.class);
	}

	public final class_2487 serialize(class_2338 anchor) {
		class_2487 nbt = new class_2487();
		nbt.method_10582("Type", type.getId().toString());
		nbt.method_10566("Pos", class_2512.method_10692(pos.method_10059(anchor)));
		serialize(nbt, anchor);
		return nbt;
	}

	@Nullable
	public static ArmInteractionPoint deserialize(class_2487 nbt, class_1937 level, class_2338 anchor) {
		class_2960 id = class_2960.method_12829(nbt.method_10558("Type"));
		if (id == null)
			return null;
		ArmInteractionPointType type = ArmInteractionPointType.get(id);
		if (type == null)
			return null;
		class_2338 pos = class_2512.method_10691(nbt.method_10562("Pos")).method_10081(anchor);
		class_2680 state = level.method_8320(pos);
		if (!type.canCreatePoint(level, pos, state))
			return null;
		ArmInteractionPoint point = type.createPoint(level, pos, state);
		if (point == null)
			return null;
		point.deserialize(nbt, anchor);
		return point;
	}

	public static void transformPos(class_2487 nbt, StructureTransform transform) {
		class_2338 pos = class_2512.method_10691(nbt.method_10562("Pos"));
		pos = transform.applyWithoutOffset(pos);
		nbt.method_10566("Pos", class_2512.method_10692(pos));
	}

	public static boolean isInteractable(class_1937 level, class_2338 pos, class_2680 state) {
		return ArmInteractionPointType.getPrimaryType(level, pos, state) != null;
	}

	@Nullable
	public static ArmInteractionPoint create(class_1937 level, class_2338 pos, class_2680 state) {
		ArmInteractionPointType type = ArmInteractionPointType.getPrimaryType(level, pos, state);
		if (type == null)
			return null;
		return type.createPoint(level, pos, state);
	}

	public enum Mode {
		DEPOSIT("mechanical_arm.deposit_to", 0xDDC166),
		TAKE("mechanical_arm.extract_from", 0x7FCDE0);

		private final String translationKey;
		private final int color;

		Mode(String translationKey, int color) {
			this.translationKey = translationKey;
			this.color = color;
		}

		public String getTranslationKey() {
			return translationKey;
		}

		public int getColor() {
			return color;
		}
	}

}
