package com.simibubi.create.content.kinetics.mechanicalArm;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.function.Consumer;

import javax.annotation.Nullable;
import net.minecraft.class_1937;
import net.minecraft.class_2338;
import net.minecraft.class_2680;
import net.minecraft.class_2960;

public abstract class ArmInteractionPointType {

	private static final Map<class_2960, ArmInteractionPointType> TYPES = new HashMap<>();
	private static final List<ArmInteractionPointType> SORTED_TYPES = new ArrayList<>();

	protected final class_2960 id;

	public ArmInteractionPointType(class_2960 id) {
		this.id = id;
	}

	public static void register(ArmInteractionPointType type) {
		class_2960 id = type.getId();
		if (TYPES.containsKey(id))
			throw new IllegalArgumentException("Tried to override ArmInteractionPointType registration for id '" + id + "'. This is not supported!");
		TYPES.put(id, type);
		SORTED_TYPES.add(type);
		SORTED_TYPES.sort((t1, t2) -> t2.getPriority() - t1.getPriority());
	}

	@Nullable
	public static ArmInteractionPointType get(class_2960 id) {
		return TYPES.get(id);
	}

	public static void forEach(Consumer<ArmInteractionPointType> action) {
		SORTED_TYPES.forEach(action);
	}

	@Nullable
	public static ArmInteractionPointType getPrimaryType(class_1937 level, class_2338 pos, class_2680 state) {
		for (ArmInteractionPointType type : SORTED_TYPES)
			if (type.canCreatePoint(level, pos, state))
				return type;
		return null;
	}

	public final class_2960 getId() {
		return id;
	}

	public abstract boolean canCreatePoint(class_1937 level, class_2338 pos, class_2680 state);

	@Nullable
	public abstract ArmInteractionPoint createPoint(class_1937 level, class_2338 pos, class_2680 state);

	public int getPriority() {
		return 0;
	}

}
