package com.simibubi.create.content.kinetics.millstone;

import java.util.Iterator;
import java.util.List;
import java.util.Optional;

import io.github.fabricators_of_create.porting_lib.transfer.item.ItemStackHandlerSlot;

import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import com.simibubi.create.AllRecipeTypes;
import com.simibubi.create.content.kinetics.base.KineticBlockEntity;
import com.simibubi.create.content.kinetics.belt.behaviour.DirectBeltInputBehaviour;
import com.simibubi.create.foundation.advancement.AllAdvancements;
import com.simibubi.create.foundation.blockEntity.behaviour.BlockEntityBehaviour;
import com.simibubi.create.foundation.item.ItemHelper;
import com.simibubi.create.foundation.sound.SoundScapes;
import com.simibubi.create.foundation.sound.SoundScapes.AmbienceGroup;
import com.simibubi.create.foundation.utility.VecHelper;

import io.github.fabricators_of_create.porting_lib.transfer.TransferUtil;
import io.github.fabricators_of_create.porting_lib.transfer.ViewOnlyWrappedStorageView;
import io.github.fabricators_of_create.porting_lib.transfer.item.ItemStackHandler;
import io.github.fabricators_of_create.porting_lib.transfer.item.ItemStackHandlerContainer;
import net.fabricmc.api.EnvType;
import net.fabricmc.api.Environment;
import net.fabricmc.fabric.api.transfer.v1.item.ItemVariant;
import net.fabricmc.fabric.api.transfer.v1.storage.Storage;
import net.fabricmc.fabric.api.transfer.v1.storage.StorageView;
import net.fabricmc.fabric.api.transfer.v1.storage.base.CombinedStorage;
import net.fabricmc.fabric.api.transfer.v1.storage.base.SidedStorageBlockEntity;
import net.fabricmc.fabric.api.transfer.v1.transaction.Transaction;
import net.fabricmc.fabric.api.transfer.v1.transaction.TransactionContext;
import net.minecraft.class_1799;
import net.minecraft.class_2338;
import net.minecraft.class_2350;
import net.minecraft.class_2350.class_2351;
import net.minecraft.class_2392;
import net.minecraft.class_2398;
import net.minecraft.class_243;
import net.minecraft.class_2487;
import net.minecraft.class_2591;
import net.minecraft.class_2680;
import net.minecraft.class_3532;

public class MillstoneBlockEntity extends KineticBlockEntity implements SidedStorageBlockEntity {

	public ItemStackHandlerContainer inputInv;
	public ItemStackHandler outputInv;
	public MillstoneInventoryHandler capability;
	public int timer;
	private MillingRecipe lastRecipe;

	public MillstoneBlockEntity(class_2591<?> type, class_2338 pos, class_2680 state) {
		super(type, pos, state);
		inputInv = new ItemStackHandlerContainer(1);
		outputInv = new ItemStackHandler(9);
		capability = new MillstoneInventoryHandler();
	}

	@Override
	public void addBehaviours(List<BlockEntityBehaviour> behaviours) {
		behaviours.add(new DirectBeltInputBehaviour(this));
		super.addBehaviours(behaviours);
		registerAwardables(behaviours, AllAdvancements.MILLSTONE);
	}

	@Override
	@Environment(EnvType.CLIENT)
	public void tickAudio() {
		super.tickAudio();

		if (getSpeed() == 0)
			return;
		if (inputInv.getStackInSlot(0)
			.method_7960())
			return;

		float pitch = class_3532.method_15363((Math.abs(getSpeed()) / 256f) + .45f, .85f, 1f);
		SoundScapes.play(AmbienceGroup.MILLING, field_11867, pitch);
	}

	@Override
	public void tick() {
		super.tick();

		if (getSpeed() == 0)
			return;
		for (int i = 0; i < outputInv.getSlotCount(); i++)
			if (outputInv.getStackInSlot(i)
				.method_7947() == outputInv.getSlotLimit(i))
				return;

		if (timer > 0) {
			timer -= getProcessingSpeed();

			if (field_11863.field_9236) {
				spawnParticles();
				return;
			}
			if (timer <= 0)
				process();
			return;
		}

		if (inputInv.getStackInSlot(0)
			.method_7960())
			return;

		if (lastRecipe == null || !lastRecipe.method_8115(inputInv, field_11863)) {
			Optional<MillingRecipe> recipe = AllRecipeTypes.MILLING.find(inputInv, field_11863);
			if (!recipe.isPresent()) {
				timer = 100;
				sendData();
			} else {
				lastRecipe = recipe.get();
				timer = lastRecipe.getProcessingDuration();
				sendData();
			}
			return;
		}

		timer = lastRecipe.getProcessingDuration();
		sendData();
	}

	@Override
	public void invalidate() {
		super.invalidate();
	}

	@Override
	public void destroy() {
		super.destroy();
		ItemHelper.dropContents(field_11863, field_11867, inputInv);
		ItemHelper.dropContents(field_11863, field_11867, outputInv);
	}

	private void process() {
		if (lastRecipe == null || !lastRecipe.method_8115(inputInv, field_11863)) {
			Optional<MillingRecipe> recipe = AllRecipeTypes.MILLING.find(inputInv, field_11863);
			if (!recipe.isPresent())
				return;
			lastRecipe = recipe.get();
		}

		try (Transaction t = TransferUtil.getTransaction()) {
			ItemStackHandlerSlot slot = inputInv.getSlot(0);
			slot.extract(slot.getResource(), 1, t);
			lastRecipe.rollResults().forEach(stack -> outputInv.insert(ItemVariant.of(stack), stack.method_7947(), t));
			t.commit();
		}
		award(AllAdvancements.MILLSTONE);

		sendData();
		method_5431();
	}

	public void spawnParticles() {
		class_1799 stackInSlot = inputInv.getStackInSlot(0);
		if (stackInSlot.method_7960())
			return;

		class_2392 data = new class_2392(class_2398.field_11218, stackInSlot);
		float angle = field_11863.field_9229.method_43057() * 360;
		class_243 offset = new class_243(0, 0, 0.5f);
		offset = VecHelper.rotate(offset, angle, class_2351.field_11052);
		class_243 target = VecHelper.rotate(offset, getSpeed() > 0 ? 25 : -25, class_2351.field_11052);

		class_243 center = offset.method_1019(VecHelper.getCenterOf(field_11867));
		target = VecHelper.offsetRandomly(target.method_1020(offset), field_11863.field_9229, 1 / 128f);
		field_11863.method_8406(data, center.field_1352, center.field_1351, center.field_1350, target.field_1352, target.field_1351, target.field_1350);
	}

	@Override
	public void write(class_2487 compound, boolean clientPacket) {
		compound.method_10569("Timer", timer);
		compound.method_10566("InputInventory", inputInv.serializeNBT());
		compound.method_10566("OutputInventory", outputInv.serializeNBT());
		super.write(compound, clientPacket);
	}

	@Override
	protected void read(class_2487 compound, boolean clientPacket) {
		timer = compound.method_10550("Timer");
		inputInv.deserializeNBT(compound.method_10562("InputInventory"));
		outputInv.deserializeNBT(compound.method_10562("OutputInventory"));
		super.read(compound, clientPacket);
	}

	public int getProcessingSpeed() {
		return class_3532.method_15340((int) Math.abs(getSpeed() / 16f), 1, 512);
	}

	@Nullable
	@Override
	public Storage<ItemVariant> getItemStorage(@Nullable class_2350 direction) {
		return capability;
	}

	private boolean canProcess(class_1799 stack) {
		ItemStackHandlerContainer tester = new ItemStackHandlerContainer(1);
		tester.setStackInSlot(0, stack);

		if (lastRecipe != null && lastRecipe.method_8115(tester, field_11863))
			return true;
		return AllRecipeTypes.MILLING.find(tester, field_11863)
			.isPresent();
	}

	private class MillstoneInventoryHandler extends CombinedStorage<ItemVariant, ItemStackHandler> {

		public MillstoneInventoryHandler() {
			super(List.of(inputInv, outputInv));
		}

		@Override
		public long insert(ItemVariant resource, long maxAmount, TransactionContext transaction) {
			if (canProcess(resource.toStack()))
				return inputInv.insert(resource, maxAmount, transaction);
			return 0;
		}

		@Override
		public long extract(ItemVariant resource, long maxAmount, TransactionContext transaction) {
			return outputInv.extract(resource, maxAmount, transaction);
		}

		@Override
		public @NotNull Iterator<StorageView<ItemVariant>> iterator() {
			return new MillstoneInventoryHandlerIterator();
		}

		private class MillstoneInventoryHandlerIterator implements Iterator<StorageView<ItemVariant>> {
			private boolean output = true;
			private Iterator<StorageView<ItemVariant>> wrapped;

			public MillstoneInventoryHandlerIterator() {
				wrapped = outputInv.iterator();
			}

			@Override
			public boolean hasNext() {
				return wrapped.hasNext();
			}

			@Override
			public StorageView<ItemVariant> next() {
				StorageView<ItemVariant> view = wrapped.next();
				if (!output) view = new ViewOnlyWrappedStorageView<>(view);
				if (output && !hasNext()) {
					wrapped = inputInv.iterator();
					output = false;
				}
				return view;
			}
		}
	}

}
