package com.simibubi.create.content.kinetics.saw;

import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;
import java.util.Optional;
import java.util.function.Predicate;
import java.util.function.Supplier;
import java.util.stream.Collectors;

import javax.annotation.ParametersAreNonnullByDefault;

import org.jetbrains.annotations.Nullable;

import com.google.common.base.Suppliers;
import com.google.common.collect.ImmutableList;
import com.simibubi.create.AllRecipeTypes;
import com.simibubi.create.AllSoundEvents;
import com.simibubi.create.content.kinetics.base.BlockBreakingKineticBlockEntity;
import com.simibubi.create.content.kinetics.belt.behaviour.DirectBeltInputBehaviour;
import com.simibubi.create.content.processing.recipe.ProcessingInventory;
import com.simibubi.create.content.processing.sequenced.SequencedAssemblyRecipe;
import com.simibubi.create.foundation.advancement.AllAdvancements;
import com.simibubi.create.foundation.blockEntity.behaviour.BlockEntityBehaviour;
import com.simibubi.create.foundation.blockEntity.behaviour.filtering.FilteringBehaviour;
import com.simibubi.create.foundation.item.ItemHelper;
import com.simibubi.create.foundation.recipe.RecipeConditions;
import com.simibubi.create.foundation.recipe.RecipeFinder;
import com.simibubi.create.foundation.utility.AbstractBlockBreakQueue;
import com.simibubi.create.foundation.utility.TreeCutter;
import com.simibubi.create.foundation.utility.VecHelper;
import com.simibubi.create.infrastructure.config.AllConfigs;

import io.github.fabricators_of_create.porting_lib.transfer.TransferUtil;
import io.github.fabricators_of_create.porting_lib.transfer.item.ItemHandlerHelper;
import io.github.fabricators_of_create.porting_lib.util.ItemStackUtil;
import io.github.fabricators_of_create.porting_lib.util.NBTSerializer;
import net.fabricmc.api.EnvType;
import net.fabricmc.api.Environment;
import net.fabricmc.fabric.api.transfer.v1.item.ItemVariant;
import net.fabricmc.fabric.api.transfer.v1.storage.Storage;
import net.fabricmc.fabric.api.transfer.v1.storage.base.SidedStorageBlockEntity;
import net.fabricmc.fabric.api.transfer.v1.transaction.Transaction;
import net.minecraft.class_1542;
import net.minecraft.class_1747;
import net.minecraft.class_1792;
import net.minecraft.class_1799;
import net.minecraft.class_1860;
import net.minecraft.class_2211;
import net.minecraft.class_2248;
import net.minecraft.class_2266;
import net.minecraft.class_2283;
import net.minecraft.class_2338;
import net.minecraft.class_2350;
import net.minecraft.class_238;
import net.minecraft.class_2388;
import net.minecraft.class_2391;
import net.minecraft.class_2392;
import net.minecraft.class_2393;
import net.minecraft.class_2394;
import net.minecraft.class_2398;
import net.minecraft.class_243;
import net.minecraft.class_2487;
import net.minecraft.class_2498;
import net.minecraft.class_2511;
import net.minecraft.class_2523;
import net.minecraft.class_2591;
import net.minecraft.class_2680;
import net.minecraft.class_2960;
import net.minecraft.class_3481;
import net.minecraft.class_3532;
import net.minecraft.class_3956;
import net.minecraft.class_3975;
import net.minecraft.class_5819;
import net.minecraft.class_6328;
import net.minecraft.class_7923;

@ParametersAreNonnullByDefault
@class_6328
public class SawBlockEntity extends BlockBreakingKineticBlockEntity implements SidedStorageBlockEntity {

	private static final Object cuttingRecipesKey = new Object();
	public static final Supplier<class_3956<?>> woodcuttingRecipeType =
		Suppliers.memoize(() -> class_7923.field_41188.method_10223(new class_2960("druidcraft", "woodcutting")));

	public ProcessingInventory inventory;
	private int recipeIndex;
	private FilteringBehaviour filtering;

	private class_1799 playEvent;

	public SawBlockEntity(class_2591<?> type, class_2338 pos, class_2680 state) {
		super(type, pos, state);
		inventory = new ProcessingInventory(this::start).withSlotLimit(!AllConfigs.server().recipes.bulkCutting.get());
		inventory.remainingTime = -1;
		recipeIndex = 0;
		playEvent = class_1799.field_8037;
	}

	@Override
	public void addBehaviours(List<BlockEntityBehaviour> behaviours) {
		super.addBehaviours(behaviours);
		filtering = new FilteringBehaviour(this, new SawFilterSlot()).forRecipes();
		behaviours.add(filtering);
		behaviours.add(new DirectBeltInputBehaviour(this).allowingBeltFunnelsWhen(this::canProcess));
		registerAwardables(behaviours, AllAdvancements.SAW_PROCESSING);
	}

	@Override
	public void write(class_2487 compound, boolean clientPacket) {
		compound.method_10566("Inventory", inventory.serializeNBT());
		compound.method_10569("RecipeIndex", recipeIndex);
		super.write(compound, clientPacket);

		if (!clientPacket || playEvent.method_7960())
			return;
		compound.method_10566("PlayEvent", NBTSerializer.serializeNBT(playEvent));
		playEvent = class_1799.field_8037;
	}

	@Override
	protected void read(class_2487 compound, boolean clientPacket) {
		super.read(compound, clientPacket);
		inventory.deserializeNBT(compound.method_10562("Inventory"));
		recipeIndex = compound.method_10550("RecipeIndex");
		if (compound.method_10545("PlayEvent"))
			playEvent = class_1799.method_7915(compound.method_10562("PlayEvent"));
	}

	@Override
	protected class_238 createRenderBoundingBox() {
		return new class_238(field_11867).method_1014(.125f);
	}

	@Override
	@Environment(EnvType.CLIENT)
	public void tickAudio() {
		super.tickAudio();
		if (getSpeed() == 0)
			return;

		if (!playEvent.method_7960()) {
			boolean isWood = false;
			class_1792 item = playEvent.method_7909();
			if (item instanceof class_1747) {
				class_2248 block = ((class_1747) item).method_7711();
				isWood = block.method_9573(block.method_9564()) == class_2498.field_11547;
			}
			spawnEventParticles(playEvent);
			playEvent = class_1799.field_8037;
			if (!isWood)
				AllSoundEvents.SAW_ACTIVATE_STONE.playAt(field_11863, field_11867, 3, 1, true);
			else
				AllSoundEvents.SAW_ACTIVATE_WOOD.playAt(field_11863, field_11867, 3, 1, true);
			return;
		}
	}

	@Override
	public void tick() {
		if (shouldRun() && ticksUntilNextProgress < 0)
			destroyNextTick();
		super.tick();

		if (!canProcess())
			return;
		if (getSpeed() == 0)
			return;
		if (inventory.remainingTime == -1) {
			if (!inventory.method_5442() && !inventory.appliedRecipe)
				start(inventory.getStackInSlot(0));
			return;
		}

		float processingSpeed = class_3532.method_15363(Math.abs(getSpeed()) / 24, 1, 128);
		inventory.remainingTime -= processingSpeed;

		if (inventory.remainingTime > 0)
			spawnParticles(inventory.getStackInSlot(0));

		if (inventory.remainingTime < 5 && !inventory.appliedRecipe) {
			if (field_11863.field_9236 && !isVirtual())
				return;
			playEvent = inventory.getStackInSlot(0);
			applyRecipe();
			inventory.appliedRecipe = true;
			inventory.recipeDuration = 20;
			inventory.remainingTime = 20;
			sendData();
			return;
		}

		class_243 itemMovement = getItemMovementVec();
		class_2350 itemMovementFacing = class_2350.method_10142(itemMovement.field_1352, itemMovement.field_1351, itemMovement.field_1350);
		if (inventory.remainingTime > 0)
			return;
		inventory.remainingTime = 0;

		for (int slot = 0; slot < inventory.getSlotCount(); slot++) {
			class_1799 stack = inventory.getStackInSlot(slot);
			if (stack.method_7960())
				continue;
			class_1799 tryExportingToBeltFunnel = getBehaviour(DirectBeltInputBehaviour.TYPE)
				.tryExportingToBeltFunnel(stack, itemMovementFacing.method_10153(), false);
			if (tryExportingToBeltFunnel != null) {
				if (tryExportingToBeltFunnel.method_7947() != stack.method_7947()) {
					inventory.setStackInSlot(slot, tryExportingToBeltFunnel);
					notifyUpdate();
					return;
				}
				if (!tryExportingToBeltFunnel.method_7960())
					return;
			}
		}

		class_2338 nextPos = field_11867.method_10081(class_2338.method_49638(itemMovement));
		DirectBeltInputBehaviour behaviour = BlockEntityBehaviour.get(field_11863, nextPos, DirectBeltInputBehaviour.TYPE);
		if (behaviour != null) {
			boolean changed = false;
			if (!behaviour.canInsertFromSide(itemMovementFacing))
				return;
			if (field_11863.field_9236 && !isVirtual())
				return;
			for (int slot = 0; slot < inventory.getSlotCount(); slot++) {
				class_1799 stack = inventory.getStackInSlot(slot);
				if (stack.method_7960())
					continue;
				class_1799 remainder = behaviour.handleInsertion(stack, itemMovementFacing, false);
				if (class_1799.method_7973(remainder, stack))
					continue;
				inventory.setStackInSlot(slot, remainder);
				changed = true;
			}
			if (changed) {
				method_5431();
				sendData();
			}
			return;
		}

		// Eject Items
		class_243 outPos = VecHelper.getCenterOf(field_11867)
			.method_1019(itemMovement.method_1021(.5f)
				.method_1031(0, .5, 0));
		class_243 outMotion = itemMovement.method_1021(.0625)
			.method_1031(0, .125, 0);
		for (int slot = 0; slot < inventory.getSlotCount(); slot++) {
			class_1799 stack = inventory.getStackInSlot(slot);
			if (stack.method_7960())
				continue;
			class_1542 entityIn = new class_1542(field_11863, outPos.field_1352, outPos.field_1351, outPos.field_1350, stack);
			entityIn.method_18799(outMotion);
			field_11863.method_8649(entityIn);
		}
		inventory.clear();
		field_11863.method_8455(field_11867, method_11010().method_26204());
		inventory.remainingTime = -1;
		sendData();
	}

	@Override
	public void invalidate() {
		super.invalidate();
	}

	@Override
	public void destroy() {
		super.destroy();
		ItemHelper.dropContents(field_11863, field_11867, inventory);
	}

	@Nullable
	@Override
	public Storage<ItemVariant> getItemStorage(@Nullable class_2350 face) {
		return face == class_2350.field_11033 ? null : inventory;
	}

	protected void spawnEventParticles(class_1799 stack) {
		if (stack == null || stack.method_7960())
			return;

		class_2394 particleData = null;
		if (stack.method_7909() instanceof class_1747)
			particleData = new class_2388(class_2398.field_11217, ((class_1747) stack.method_7909()).method_7711()
				.method_9564());
		else
			particleData = new class_2392(class_2398.field_11218, stack);

		class_5819 r = field_11863.field_9229;
		class_243 v = VecHelper.getCenterOf(this.field_11867)
			.method_1031(0, 5 / 16f, 0);
		for (int i = 0; i < 10; i++) {
			class_243 m = VecHelper.offsetRandomly(new class_243(0, 0.25f, 0), r, .125f);
			field_11863.method_8406(particleData, v.field_1352, v.field_1351, v.field_1350, m.field_1352, m.field_1351, m.field_1351);
		}
	}

	protected void spawnParticles(class_1799 stack) {
		if (stack == null || stack.method_7960())
			return;

		class_2394 particleData = null;
		float speed = 1;
		if (stack.method_7909() instanceof class_1747)
			particleData = new class_2388(class_2398.field_11217, ((class_1747) stack.method_7909()).method_7711()
				.method_9564());
		else {
			particleData = new class_2392(class_2398.field_11218, stack);
			speed = .125f;
		}

		class_5819 r = field_11863.field_9229;
		class_243 vec = getItemMovementVec();
		class_243 pos = VecHelper.getCenterOf(this.field_11867);
		float offset = inventory.recipeDuration != 0 ? (float) (inventory.remainingTime) / inventory.recipeDuration : 0;
		offset /= 2;
		if (inventory.appliedRecipe)
			offset -= .5f;
		field_11863.method_8406(particleData, pos.method_10216() + -vec.field_1352 * offset, pos.method_10214() + .45f, pos.method_10215() + -vec.field_1350 * offset,
			-vec.field_1352 * speed, r.method_43057() * speed, -vec.field_1350 * speed);
	}

	public class_243 getItemMovementVec() {
		boolean alongX = !method_11010().method_11654(SawBlock.AXIS_ALONG_FIRST_COORDINATE);
		int offset = getSpeed() < 0 ? -1 : 1;
		return new class_243(offset * (alongX ? 1 : 0), 0, offset * (alongX ? 0 : -1));
	}

	private void applyRecipe() {
		List<? extends class_1860<?>> recipes = getRecipes();
		if (recipes.isEmpty())
			return;
		if (recipeIndex >= recipes.size())
			recipeIndex = 0;

		class_1860<?> recipe = recipes.get(recipeIndex);

		int rolls = inventory.getStackInSlot(0)
			.method_7947();
		inventory.clear();

		List<class_1799> list = new ArrayList<>();
		for (int roll = 0; roll < rolls; roll++) {
			List<class_1799> results = new LinkedList<class_1799>();
			if (recipe instanceof CuttingRecipe)
				results = ((CuttingRecipe) recipe).rollResults();
			else if (recipe instanceof class_3975 || recipe.method_17716() == woodcuttingRecipeType.get())
				results.add(recipe.method_8110(field_11863.method_30349())
					.method_7972());

			for (int i = 0; i < results.size(); i++) {
				class_1799 stack = results.get(i);
				ItemHelper.addToList(stack, list);
			}
		}

		for (int slot = 0; slot < list.size() && slot + 1 < inventory.getSlotCount(); slot++)
			inventory.setStackInSlot(slot + 1, list.get(slot));

		award(AllAdvancements.SAW_PROCESSING);
	}

	private List<? extends class_1860<?>> getRecipes() {
		Optional<CuttingRecipe> assemblyRecipe = SequencedAssemblyRecipe.getRecipe(field_11863, inventory.getStackInSlot(0),
			AllRecipeTypes.CUTTING.getType(), CuttingRecipe.class);
		if (assemblyRecipe.isPresent() && filtering.test(assemblyRecipe.get()
			.method_8110(field_11863.method_30349())))
			return ImmutableList.of(assemblyRecipe.get());

		Predicate<class_1860<?>> types = RecipeConditions.isOfType(AllRecipeTypes.CUTTING.getType(),
			AllConfigs.server().recipes.allowStonecuttingOnSaw.get() ? class_3956.field_17641 : null,
			AllConfigs.server().recipes.allowWoodcuttingOnSaw.get() ? woodcuttingRecipeType.get() : null);

		List<class_1860<?>> startedSearch = RecipeFinder.get(cuttingRecipesKey, field_11863, types);
		return startedSearch.stream()
			.filter(RecipeConditions.outputMatchesFilter(filtering))
			.filter(RecipeConditions.firstIngredientMatches(inventory.getStackInSlot(0)))
			.filter(r -> !AllRecipeTypes.shouldIgnoreInAutomation(r))
			.collect(Collectors.toList());
	}

	public void insertItem(class_1542 entity) {
		if (!canProcess())
			return;
		if (!inventory.method_5442())
			return;
		if (!entity.method_5805())
			return;
		if (field_11863.field_9236)
			return;

		inventory.clear();
		try (Transaction t = TransferUtil.getTransaction()) {
			class_1799 contained = entity.method_6983();
			long inserted = inventory.insert(ItemVariant.of(contained), contained.method_7947(), t);
			if (contained.method_7947() == inserted)
				entity.method_31472();
			else
				entity.method_6979(ItemHandlerHelper.copyStackWithSize(contained, (int) (contained.method_7947() - inserted)));
			t.commit();
		}
	}

	public void start(class_1799 inserted) {
		if (!canProcess())
			return;
		if (inventory.method_5442())
			return;
		if (field_11863.field_9236 && !isVirtual())
			return;

		List<? extends class_1860<?>> recipes = getRecipes();
		boolean valid = !recipes.isEmpty();
		int time = 50;

		if (recipes.isEmpty()) {
			inventory.remainingTime = inventory.recipeDuration = 10;
			inventory.appliedRecipe = false;
			sendData();
			return;
		}

		if (valid) {
			recipeIndex++;
			if (recipeIndex >= recipes.size())
				recipeIndex = 0;
		}

		class_1860<?> recipe = recipes.get(recipeIndex);
		if (recipe instanceof CuttingRecipe) {
			time = ((CuttingRecipe) recipe).getProcessingDuration();
		}

		inventory.remainingTime = time * Math.max(1, (inserted.method_7947() / 5));
		inventory.recipeDuration = inventory.remainingTime;
		inventory.appliedRecipe = false;
		sendData();
	}

	protected boolean canProcess() {
		return method_11010().method_11654(SawBlock.FACING) == class_2350.field_11036;
	}

	// Block Breaker

	@Override
	protected boolean shouldRun() {
		return method_11010().method_11654(SawBlock.FACING)
			.method_10166()
			.method_10179();
	}

	@Override
	protected class_2338 getBreakingPos() {
		return method_11016().method_10093(method_11010().method_11654(SawBlock.FACING));
	}

	@Override
	public void onBlockBroken(class_2680 stateToBreak) {
		Optional<AbstractBlockBreakQueue> dynamicTree =
			TreeCutter.findDynamicTree(stateToBreak.method_26204(), breakingPos);
		if (dynamicTree.isPresent()) {
			dynamicTree.get()
				.destroyBlocks(field_11863, null, this::dropItemFromCutTree);
			return;
		}

		super.onBlockBroken(stateToBreak);
		TreeCutter.findTree(field_11863, breakingPos)
			.destroyBlocks(field_11863, null, this::dropItemFromCutTree);
	}

	public void dropItemFromCutTree(class_2338 pos, class_1799 stack) {
		float distance = (float) Math.sqrt(pos.method_10262(breakingPos));
		class_243 dropPos = VecHelper.getCenterOf(pos);
		class_1542 entity = new class_1542(field_11863, dropPos.field_1352, dropPos.field_1351, dropPos.field_1350, stack);
		entity.method_18799(class_243.method_24954(breakingPos.method_10059(this.field_11867))
			.method_1021(distance / 20f));
		field_11863.method_8649(entity);
	}

	@Override
	public boolean canBreak(class_2680 stateToBreak, float blockHardness) {
		boolean sawable = isSawable(stateToBreak);
		return super.canBreak(stateToBreak, blockHardness) && sawable;
	}

	public static boolean isSawable(class_2680 stateToBreak) {
		if (stateToBreak.method_26164(class_3481.field_15462))
			return false;
		if (TreeCutter.isLog(stateToBreak) || (stateToBreak.method_26164(class_3481.field_15503)))
			return true;
		if (TreeCutter.isRoot(stateToBreak))
			return true;
		class_2248 block = stateToBreak.method_26204();
		if (block instanceof class_2211)
			return true;
		if (block instanceof class_2511)
			return true;
		if (block instanceof class_2266)
			return true;
		if (block instanceof class_2523)
			return true;
		if (block instanceof class_2391)
			return true;
		if (block instanceof class_2393)
			return true;
		if (block instanceof class_2283)
			return true;
		if (TreeCutter.canDynamicTreeCutFrom(block))
			return true;
		return false;
	}

}
