package com.simibubi.create.content.kinetics.saw;

import static net.minecraft.class_2741.field_12525;

import com.jozufozu.flywheel.backend.Backend;
import com.jozufozu.flywheel.core.PartialModel;
import com.jozufozu.flywheel.core.virtual.VirtualRenderWorld;
import com.simibubi.create.AllPartialModels;
import com.simibubi.create.content.contraptions.behaviour.MovementContext;
import com.simibubi.create.content.contraptions.render.ContraptionMatrices;
import com.simibubi.create.content.contraptions.render.ContraptionRenderDispatcher;
import com.simibubi.create.content.kinetics.base.KineticBlockEntity;
import com.simibubi.create.content.kinetics.base.KineticBlockEntityRenderer;
import com.simibubi.create.foundation.blockEntity.behaviour.filtering.FilteringRenderer;
import com.simibubi.create.foundation.blockEntity.renderer.SafeBlockEntityRenderer;
import com.simibubi.create.foundation.render.CachedBufferer;
import com.simibubi.create.foundation.render.SuperByteBuffer;
import com.simibubi.create.foundation.utility.AngleHelper;
import com.simibubi.create.foundation.utility.VecHelper;
import net.minecraft.class_1087;
import net.minecraft.class_1799;
import net.minecraft.class_1921;
import net.minecraft.class_2350;
import net.minecraft.class_243;
import net.minecraft.class_2470;
import net.minecraft.class_2680;
import net.minecraft.class_310;
import net.minecraft.class_3532;
import net.minecraft.class_4587;
import net.minecraft.class_4597;
import net.minecraft.class_5614;
import net.minecraft.class_7833;
import net.minecraft.class_811;
import net.minecraft.class_918;

public class SawRenderer extends SafeBlockEntityRenderer<SawBlockEntity> {

	public SawRenderer(class_5614.class_5615 context) {
	}

	@Override
	protected void renderSafe(SawBlockEntity be, float partialTicks, class_4587 ms, class_4597 buffer, int light,
		int overlay) {
		renderBlade(be, ms, buffer, light);
		renderItems(be, partialTicks, ms, buffer, light, overlay);
		FilteringRenderer.renderOnBlockEntity(be, partialTicks, ms, buffer, light, overlay);

		if (Backend.canUseInstancing(be.method_10997()))
			return;

		renderShaft(be, ms, buffer, light, overlay);
	}

	protected void renderBlade(SawBlockEntity be, class_4587 ms, class_4597 buffer, int light) {
		class_2680 blockState = be.method_11010();
		PartialModel partial;
		float speed = be.getSpeed();
		boolean rotate = false;

		if (SawBlock.isHorizontal(blockState)) {
			if (speed > 0) {
				partial = AllPartialModels.SAW_BLADE_HORIZONTAL_ACTIVE;
			} else if (speed < 0) {
				partial = AllPartialModels.SAW_BLADE_HORIZONTAL_REVERSED;
			} else {
				partial = AllPartialModels.SAW_BLADE_HORIZONTAL_INACTIVE;
			}
		} else {
			if (be.getSpeed() > 0) {
				partial = AllPartialModels.SAW_BLADE_VERTICAL_ACTIVE;
			} else if (speed < 0) {
				partial = AllPartialModels.SAW_BLADE_VERTICAL_REVERSED;
			} else {
				partial = AllPartialModels.SAW_BLADE_VERTICAL_INACTIVE;
			}

			if (blockState.method_11654(SawBlock.AXIS_ALONG_FIRST_COORDINATE))
				rotate = true;
		}

		SuperByteBuffer superBuffer = CachedBufferer.partialFacing(partial, blockState);
		if (rotate) {
			superBuffer.rotateCentered(class_2350.field_11036, AngleHelper.rad(90));
		}
		superBuffer.color(0xFFFFFF)
			.light(light)
			.renderInto(ms, buffer.getBuffer(class_1921.method_23579()));
	}

	protected void renderShaft(SawBlockEntity be, class_4587 ms, class_4597 buffer, int light, int overlay) {
		KineticBlockEntityRenderer.renderRotatingBuffer(be, getRotatedModel(be), ms,
			buffer.getBuffer(class_1921.method_23577()), light);
	}

	protected void renderItems(SawBlockEntity be, float partialTicks, class_4587 ms, class_4597 buffer,
		int light, int overlay) {
		boolean processingMode = be.method_11010()
			.method_11654(SawBlock.FACING) == class_2350.field_11036;
		if (processingMode && !be.inventory.method_5442()) {
			boolean alongZ = !be.method_11010()
				.method_11654(SawBlock.AXIS_ALONG_FIRST_COORDINATE);
			ms.method_22903();

			boolean moving = be.inventory.recipeDuration != 0;
			float offset = moving ? (float) (be.inventory.remainingTime) / be.inventory.recipeDuration : 0;
			float processingSpeed = class_3532.method_15363(Math.abs(be.getSpeed()) / 32, 1, 128);
			if (moving) {
				offset = class_3532
					.method_15363(offset + ((-partialTicks + .5f) * processingSpeed) / be.inventory.recipeDuration, 0.125f, 1f);
				if (!be.inventory.appliedRecipe)
					offset += 1;
				offset /= 2;
			}

			if (be.getSpeed() == 0)
				offset = .5f;
			if (be.getSpeed() < 0 ^ alongZ)
				offset = 1 - offset;

			for (int i = 0; i < be.inventory.getSlotCount(); i++) {
				class_1799 stack = be.inventory.getStackInSlot(i);
				if (stack.method_7960())
					continue;

				class_918 itemRenderer = class_310.method_1551()
					.method_1480();
				class_1087 modelWithOverrides = itemRenderer.method_4019(stack, be.method_10997(), null, 0);
				boolean blockItem = modelWithOverrides.method_4712();

				ms.method_22904(alongZ ? offset : .5, blockItem ? .925f : 13f / 16f, alongZ ? .5 : offset);

				ms.method_22905(.5f, .5f, .5f);
				if (alongZ)
					ms.method_22907(class_7833.field_40716.rotationDegrees(90));
				ms.method_22907(class_7833.field_40714.rotationDegrees(90));
				itemRenderer.method_23178(stack, class_811.field_4319, light, overlay, ms, buffer, be.method_10997(), 0);
				break;
			}

			ms.method_22909();
		}
	}

	protected SuperByteBuffer getRotatedModel(KineticBlockEntity be) {
		class_2680 state = be.method_11010();
		if (state.method_11654(field_12525)
			.method_10166()
			.method_10179())
			return CachedBufferer.partialFacing(AllPartialModels.SHAFT_HALF,
				state.method_26186(class_2470.field_11464));
		return CachedBufferer.block(KineticBlockEntityRenderer.KINETIC_BLOCK,
			getRenderedBlockState(be));
	}

	protected class_2680 getRenderedBlockState(KineticBlockEntity be) {
		return KineticBlockEntityRenderer.shaft(KineticBlockEntityRenderer.getRotationAxisOf(be));
	}

	public static void renderInContraption(MovementContext context, VirtualRenderWorld renderWorld,
		ContraptionMatrices matrices, class_4597 buffer) {
		class_2680 state = context.state;
		class_2350 facing = state.method_11654(SawBlock.FACING);

		class_243 facingVec = class_243.method_24954(context.state.method_11654(SawBlock.FACING)
			.method_10163());
		facingVec = context.rotation.apply(facingVec);

		class_2350 closestToFacing = class_2350.method_10142(facingVec.field_1352, facingVec.field_1351, facingVec.field_1350);

		boolean horizontal = closestToFacing.method_10166()
			.method_10179();
		boolean backwards = VecHelper.isVecPointingTowards(context.relativeMotion, facing.method_10153());
		boolean moving = context.getAnimationSpeed() != 0;
		boolean shouldAnimate =
			(context.contraption.stalled && horizontal) || (!context.contraption.stalled && !backwards && moving);

		SuperByteBuffer superBuffer;
		if (SawBlock.isHorizontal(state)) {
			if (shouldAnimate)
				superBuffer = CachedBufferer.partial(AllPartialModels.SAW_BLADE_HORIZONTAL_ACTIVE, state);
			else
				superBuffer = CachedBufferer.partial(AllPartialModels.SAW_BLADE_HORIZONTAL_INACTIVE, state);
		} else {
			if (shouldAnimate)
				superBuffer = CachedBufferer.partial(AllPartialModels.SAW_BLADE_VERTICAL_ACTIVE, state);
			else
				superBuffer = CachedBufferer.partial(AllPartialModels.SAW_BLADE_VERTICAL_INACTIVE, state);
		}

		superBuffer.transform(matrices.getModel())
			.centre()
			.rotateY(AngleHelper.horizontalAngle(facing))
			.rotateX(AngleHelper.verticalAngle(facing));

		if (!SawBlock.isHorizontal(state)) {
			superBuffer.rotateZ(state.method_11654(SawBlock.AXIS_ALONG_FIRST_COORDINATE) ? 90 : 0);
		}

		superBuffer.unCentre()
			.light(matrices.getWorld(), ContraptionRenderDispatcher.getContraptionWorldLight(context, renderWorld))
			.renderInto(matrices.getViewProjection(), buffer.getBuffer(class_1921.method_23579()));
	}

}
