package com.simibubi.create.content.kinetics.simpleRelays;

import static com.simibubi.create.content.kinetics.base.RotatedPillarKineticBlock.AXIS;

import java.util.List;
import java.util.function.Predicate;
import net.minecraft.class_1269;
import net.minecraft.class_1657;
import net.minecraft.class_1747;
import net.minecraft.class_1799;
import net.minecraft.class_1838;
import net.minecraft.class_1937;
import net.minecraft.class_2338;
import net.minecraft.class_2350;
import net.minecraft.class_2350.class_2351;
import net.minecraft.class_2680;
import net.minecraft.class_3965;
import net.minecraft.class_6328;
import com.simibubi.create.AllShapes;
import com.simibubi.create.content.kinetics.base.DirectionalKineticBlock;
import com.simibubi.create.content.kinetics.base.HorizontalKineticBlock;
import com.simibubi.create.content.kinetics.base.IRotate;
import com.simibubi.create.content.kinetics.base.RotatedPillarKineticBlock;
import com.simibubi.create.foundation.placement.IPlacementHelper;
import com.simibubi.create.foundation.placement.PlacementHelpers;
import com.simibubi.create.foundation.placement.PlacementOffset;
import com.simibubi.create.foundation.utility.Iterate;

import io.github.fabricators_of_create.porting_lib.item.UseFirstBehaviorItem;

public class CogwheelBlockItem extends class_1747 implements UseFirstBehaviorItem {

	boolean large;

	private final int placementHelperId;
	private final int integratedCogHelperId;

	public CogwheelBlockItem(CogWheelBlock block, class_1793 builder) {
		super(block, builder);
		large = block.isLarge;

		placementHelperId = PlacementHelpers.register(large ? new LargeCogHelper() : new SmallCogHelper());
		integratedCogHelperId =
			PlacementHelpers.register(large ? new IntegratedLargeCogHelper() : new IntegratedSmallCogHelper());
	}

	@Override
	public class_1269 onItemUseFirst(class_1799 stack, class_1838 context) {
		class_1937 world = context.method_8045();
		class_2338 pos = context.method_8037();
		class_2680 state = world.method_8320(pos);

		IPlacementHelper helper = PlacementHelpers.get(placementHelperId);
		class_1657 player = context.method_8036();
		class_3965 ray = new class_3965(context.method_17698(), context.method_8038(), pos, true);
		if (helper.matchesState(state) && player != null && !player.method_5715()) {
			return helper.getOffset(player, world, state, pos, ray)
				.placeInWorld(world, this, player, context.method_20287(), ray);
		}

		if (integratedCogHelperId != -1) {
			helper = PlacementHelpers.get(integratedCogHelperId);

			if (helper.matchesState(state) && player != null && !player.method_5715()) {
				return helper.getOffset(player, world, state, pos, ray)
					.placeInWorld(world, this, player, context.method_20287(), ray);
			}
		}

		return class_1269.field_5811;
	}

	@class_6328
	private static class SmallCogHelper extends DiagonalCogHelper {

		@Override
		public Predicate<class_1799> getItemPredicate() {
			return ((Predicate<class_1799>) ICogWheel::isSmallCogItem).and(ICogWheel::isDedicatedCogItem);
		}

		@Override
		public PlacementOffset getOffset(class_1657 player, class_1937 world, class_2680 state, class_2338 pos,
			class_3965 ray) {
			if (hitOnShaft(state, ray))
				return PlacementOffset.fail();

			if (!ICogWheel.isLargeCog(state)) {
				class_2351 axis = ((IRotate) state.method_26204()).getRotationAxis(state);
				List<class_2350> directions = IPlacementHelper.orderedByDistanceExceptAxis(pos, ray.method_17784(), axis);

				for (class_2350 dir : directions) {
					class_2338 newPos = pos.method_10093(dir);

					if (!CogWheelBlock.isValidCogwheelPosition(false, world, newPos, axis))
						continue;

					if (!world.method_8320(newPos)
						.method_45474())
						continue;

					return PlacementOffset.success(newPos, s -> s.method_11657(AXIS, axis));

				}

				return PlacementOffset.fail();
			}

			return super.getOffset(player, world, state, pos, ray);
		}
	}

	@class_6328
	private static class LargeCogHelper extends DiagonalCogHelper {

		@Override
		public Predicate<class_1799> getItemPredicate() {
			return ((Predicate<class_1799>) ICogWheel::isLargeCogItem).and(ICogWheel::isDedicatedCogItem);
		}

		@Override
		public PlacementOffset getOffset(class_1657 player, class_1937 world, class_2680 state, class_2338 pos,
			class_3965 ray) {
			if (hitOnShaft(state, ray))
				return PlacementOffset.fail();

			if (ICogWheel.isLargeCog(state)) {
				class_2351 axis = ((IRotate) state.method_26204()).getRotationAxis(state);
				class_2350 side = IPlacementHelper.orderedByDistanceOnlyAxis(pos, ray.method_17784(), axis)
					.get(0);
				List<class_2350> directions = IPlacementHelper.orderedByDistanceExceptAxis(pos, ray.method_17784(), axis);
				for (class_2350 dir : directions) {
					class_2338 newPos = pos.method_10093(dir)
						.method_10093(side);

					if (!CogWheelBlock.isValidCogwheelPosition(true, world, newPos, dir.method_10166()))
						continue;

					if (!world.method_8320(newPos)
						.method_45474())
						continue;

					return PlacementOffset.success(newPos, s -> s.method_11657(AXIS, dir.method_10166()));
				}

				return PlacementOffset.fail();
			}

			return super.getOffset(player, world, state, pos, ray);
		}
	}

	@class_6328
	public abstract static class DiagonalCogHelper implements IPlacementHelper {

		@Override
		public Predicate<class_2680> getStatePredicate() {
			return s -> ICogWheel.isSmallCog(s) || ICogWheel.isLargeCog(s);
		}

		@Override
		public PlacementOffset getOffset(class_1657 player, class_1937 world, class_2680 state, class_2338 pos,
			class_3965 ray) {
			// diagonal gears of different size
			class_2351 axis = ((IRotate) state.method_26204()).getRotationAxis(state);
			class_2350 closest = IPlacementHelper.orderedByDistanceExceptAxis(pos, ray.method_17784(), axis)
				.get(0);
			List<class_2350> directions = IPlacementHelper.orderedByDistanceExceptAxis(pos, ray.method_17784(), axis,
				d -> d.method_10166() != closest.method_10166());

			for (class_2350 dir : directions) {
				class_2338 newPos = pos.method_10093(dir)
					.method_10093(closest);
				if (!world.method_8320(newPos)
					.method_45474())
					continue;

				if (!CogWheelBlock.isValidCogwheelPosition(ICogWheel.isLargeCog(state), world, newPos, axis))
					continue;

				return PlacementOffset.success(newPos, s -> s.method_11657(AXIS, axis));
			}

			return PlacementOffset.fail();
		}

		protected boolean hitOnShaft(class_2680 state, class_3965 ray) {
			return AllShapes.SIX_VOXEL_POLE.get(((IRotate) state.method_26204()).getRotationAxis(state))
				.method_1107()
				.method_1014(0.001)
				.method_1006(ray.method_17784()
					.method_1020(ray.method_17784()
						.method_1032(Iterate.axisSet)));
		}
	}

	@class_6328
	public static class IntegratedLargeCogHelper implements IPlacementHelper {

		@Override
		public Predicate<class_1799> getItemPredicate() {
			return ((Predicate<class_1799>) ICogWheel::isLargeCogItem).and(ICogWheel::isDedicatedCogItem);
		}

		@Override
		public Predicate<class_2680> getStatePredicate() {
			return s -> !ICogWheel.isDedicatedCogWheel(s.method_26204()) && ICogWheel.isSmallCog(s);
		}

		@Override
		public PlacementOffset getOffset(class_1657 player, class_1937 world, class_2680 state, class_2338 pos,
			class_3965 ray) {
			class_2350 face = ray.method_17780();
			class_2351 newAxis;

			if (state.method_28498(HorizontalKineticBlock.HORIZONTAL_FACING))
				newAxis = state.method_11654(HorizontalKineticBlock.HORIZONTAL_FACING)
					.method_10166();
			else if (state.method_28498(DirectionalKineticBlock.FACING))
				newAxis = state.method_11654(DirectionalKineticBlock.FACING)
					.method_10166();
			else if (state.method_28498(RotatedPillarKineticBlock.AXIS))
				newAxis = state.method_11654(RotatedPillarKineticBlock.AXIS);
			else
				newAxis = class_2351.field_11052;

			if (face.method_10166() == newAxis)
				return PlacementOffset.fail();

			List<class_2350> directions =
				IPlacementHelper.orderedByDistanceExceptAxis(pos, ray.method_17784(), face.method_10166(), newAxis);

			for (class_2350 d : directions) {
				class_2338 newPos = pos.method_10093(face)
					.method_10093(d);

				if (!world.method_8320(newPos)
					.method_45474())
					continue;

				if (!CogWheelBlock.isValidCogwheelPosition(false, world, newPos, newAxis))
					return PlacementOffset.fail();

				return PlacementOffset.success(newPos, s -> s.method_11657(CogWheelBlock.AXIS, newAxis));
			}

			return PlacementOffset.fail();
		}

	}

	@class_6328
	public static class IntegratedSmallCogHelper implements IPlacementHelper {

		@Override
		public Predicate<class_1799> getItemPredicate() {
			return ((Predicate<class_1799>) ICogWheel::isSmallCogItem).and(ICogWheel::isDedicatedCogItem);
		}

		@Override
		public Predicate<class_2680> getStatePredicate() {
			return s -> !ICogWheel.isDedicatedCogWheel(s.method_26204()) && ICogWheel.isSmallCog(s);
		}

		@Override
		public PlacementOffset getOffset(class_1657 player, class_1937 world, class_2680 state, class_2338 pos,
			class_3965 ray) {
			class_2350 face = ray.method_17780();
			class_2351 newAxis;

			if (state.method_28498(HorizontalKineticBlock.HORIZONTAL_FACING))
				newAxis = state.method_11654(HorizontalKineticBlock.HORIZONTAL_FACING)
					.method_10166();
			else if (state.method_28498(DirectionalKineticBlock.FACING))
				newAxis = state.method_11654(DirectionalKineticBlock.FACING)
					.method_10166();
			else if (state.method_28498(RotatedPillarKineticBlock.AXIS))
				newAxis = state.method_11654(RotatedPillarKineticBlock.AXIS);
			else
				newAxis = class_2351.field_11052;

			if (face.method_10166() == newAxis)
				return PlacementOffset.fail();

			List<class_2350> directions = IPlacementHelper.orderedByDistanceExceptAxis(pos, ray.method_17784(), newAxis);

			for (class_2350 d : directions) {
				class_2338 newPos = pos.method_10093(d);

				if (!world.method_8320(newPos)
					.method_45474())
					continue;

				if (!CogWheelBlock.isValidCogwheelPosition(false, world, newPos, newAxis))
					return PlacementOffset.fail();

				return PlacementOffset.success()
					.at(newPos)
					.withTransform(s -> s.method_11657(CogWheelBlock.AXIS, newAxis));
			}

			return PlacementOffset.fail();
		}

	}
}
