package com.simibubi.create.content.kinetics.steamEngine;

import java.util.List;
import net.minecraft.class_2248;
import net.minecraft.class_2338;
import net.minecraft.class_2350.class_2351;
import net.minecraft.class_2487;
import net.minecraft.class_2512;
import net.minecraft.class_2561;
import net.minecraft.class_2591;
import net.minecraft.class_2680;
import net.minecraft.class_2960;
import net.minecraft.class_3532;
import net.minecraft.class_7923;
import com.simibubi.create.content.kinetics.BlockStressValues;
import com.simibubi.create.content.kinetics.base.GeneratingKineticBlockEntity;
import com.simibubi.create.foundation.utility.RegisteredObjects;

public class PoweredShaftBlockEntity extends GeneratingKineticBlockEntity {

	public class_2338 enginePos;
	public float engineEfficiency;
	public int movementDirection;
	public int initialTicks;
	public class_2248 capacityKey;

	public PoweredShaftBlockEntity(class_2591<?> typeIn, class_2338 pos, class_2680 state) {
		super(typeIn, pos, state);
		movementDirection = 1;
		initialTicks = 3;
	}

	@Override
	public void tick() {
		super.tick();
		if (initialTicks > 0)
			initialTicks--;
	}

	public void update(class_2338 sourcePos, int direction, float efficiency) {
		class_2338 key = field_11867.method_10059(sourcePos);
		enginePos = key;
		float prev = engineEfficiency;
		engineEfficiency = efficiency;
		int prevDirection = this.movementDirection;
		if (class_3532.method_15347(efficiency, prev) && prevDirection == direction)
			return;

		capacityKey = field_11863.method_8320(sourcePos)
				.method_26204();
		this.movementDirection = direction;
		updateGeneratedRotation();
	}

	public void remove(class_2338 sourcePos) {
		if (!isPoweredBy(sourcePos))
			return;

		enginePos = null;
		engineEfficiency = 0;
		movementDirection = 0;
		capacityKey = null;
		updateGeneratedRotation();
	}

	public boolean canBePoweredBy(class_2338 globalPos) {
		return initialTicks == 0 && (enginePos == null || isPoweredBy(globalPos));
	}

	public boolean isPoweredBy(class_2338 globalPos) {
		class_2338 key = field_11867.method_10059(globalPos);
		return key.equals(enginePos);
	}

	@Override
	protected void write(class_2487 compound, boolean clientPacket) {
		compound.method_10569("Direction", movementDirection);
		if (initialTicks > 0)
			compound.method_10569("Warmup", initialTicks);
		if (enginePos != null && capacityKey != null) {
			compound.method_10566("EnginePos", class_2512.method_10692(enginePos));
			compound.method_10548("EnginePower", engineEfficiency);
			compound.method_10582("EngineType", RegisteredObjects.getKeyOrThrow(capacityKey)
					.toString());
		}
		super.write(compound, clientPacket);
	}

	@Override
	protected void read(class_2487 compound, boolean clientPacket) {
		super.read(compound, clientPacket);
		movementDirection = compound.method_10550("Direction");
		initialTicks = compound.method_10550("Warmup");
		enginePos = null;
		engineEfficiency = 0;
		if (compound.method_10545("EnginePos")) {
			enginePos = class_2512.method_10691(compound.method_10562("EnginePos"));
			engineEfficiency = compound.method_10583("EnginePower");
			capacityKey = class_7923.field_41175.method_10223(new class_2960(compound.method_10558("EngineType")));
		}
	}

	@Override
	public float getGeneratedSpeed() {
		return getCombinedCapacity() > 0 ? movementDirection * 16 * getSpeedModifier() : 0;
	}

	private float getCombinedCapacity() {
		return capacityKey == null ? 0 : (float) (engineEfficiency * BlockStressValues.getCapacity(capacityKey));
	}

	private int getSpeedModifier() {
		return (int) (1 + (engineEfficiency >= 1 ? 3 : Math.min(2, Math.floor(engineEfficiency * 4))));
	}

	@Override
	public float calculateAddedStressCapacity() {
		float capacity = getCombinedCapacity() / getSpeedModifier();
		this.lastCapacityProvided = capacity;
		return capacity;
	}

	@Override
	public int getRotationAngleOffset(class_2351 axis) {
		int combinedCoords = axis.method_10173(field_11867.method_10263(), field_11867.method_10264(), field_11867.method_10260());
		return super.getRotationAngleOffset(axis) + (combinedCoords % 2 == 0 ? 180 : 0);
	}

	@Override
	public boolean addToGoggleTooltip(List<class_2561> tooltip, boolean isPlayerSneaking) {
		return false;
	}

	public boolean addToEngineTooltip(List<class_2561> tooltip, boolean isPlayerSneaking) {
		return super.addToGoggleTooltip(tooltip, isPlayerSneaking);
	}

}
