package com.simibubi.create.content.kinetics.steamEngine;

import java.lang.ref.WeakReference;
import java.util.List;

import javax.annotation.Nullable;

import com.simibubi.create.AllBlocks;
import com.simibubi.create.AllSoundEvents;
import com.simibubi.create.content.contraptions.bearing.WindmillBearingBlockEntity.RotationDirection;
import com.simibubi.create.content.equipment.goggles.IHaveGoggleInformation;
import com.simibubi.create.content.fluids.tank.FluidTankBlockEntity;
import com.simibubi.create.content.kinetics.base.GeneratingKineticBlockEntity;
import com.simibubi.create.content.kinetics.base.IRotate;
import com.simibubi.create.content.kinetics.base.KineticBlockEntityRenderer;
import com.simibubi.create.foundation.advancement.AllAdvancements;
import com.simibubi.create.foundation.blockEntity.SmartBlockEntity;
import com.simibubi.create.foundation.blockEntity.behaviour.BlockEntityBehaviour;
import com.simibubi.create.foundation.blockEntity.behaviour.scrollValue.ScrollOptionBehaviour;
import com.simibubi.create.foundation.utility.AngleHelper;
import com.simibubi.create.foundation.utility.Lang;
import com.simibubi.create.foundation.utility.VecHelper;

import com.tterrag.registrate.fabric.EnvExecutor;
import net.fabricmc.api.EnvType;
import net.fabricmc.api.Environment;
import net.minecraft.class_2338;
import net.minecraft.class_2350;
import net.minecraft.class_2350.class_2351;
import net.minecraft.class_2350.class_2352;
import net.minecraft.class_238;
import net.minecraft.class_243;
import net.minecraft.class_2561;
import net.minecraft.class_2586;
import net.minecraft.class_2591;
import net.minecraft.class_2680;
import net.minecraft.class_3417;
import net.minecraft.class_3419;
import net.minecraft.class_3532;

public class SteamEngineBlockEntity extends SmartBlockEntity implements IHaveGoggleInformation {

	protected ScrollOptionBehaviour<RotationDirection> movementDirection;

	public WeakReference<PoweredShaftBlockEntity> target;
	public WeakReference<FluidTankBlockEntity> source;

	public SteamEngineBlockEntity(class_2591<?> type, class_2338 pos, class_2680 state) {
		super(type, pos, state);
		source = new WeakReference<>(null);
		target = new WeakReference<>(null);
	}

	@Override
	public void addBehaviours(List<BlockEntityBehaviour> behaviours) {
		movementDirection = new ScrollOptionBehaviour<>(RotationDirection.class,
			Lang.translateDirect("contraptions.windmill.rotation_direction"), this, new SteamEngineValueBox());
		movementDirection.onlyActiveWhen(() -> {
			PoweredShaftBlockEntity shaft = getShaft();
			return shaft == null || !shaft.hasSource();
		});
		movementDirection.withCallback($ -> onDirectionChanged());
		behaviours.add(movementDirection);

		registerAwardables(behaviours, AllAdvancements.STEAM_ENGINE);
	}

	private void onDirectionChanged() {}

	@Override
	public void tick() {
		super.tick();
		FluidTankBlockEntity tank = getTank();
		PoweredShaftBlockEntity shaft = getShaft();

		if (tank == null || shaft == null) {
			if (field_11863.method_8608())
				return;
			if (shaft == null)
				return;
			if (!shaft.method_11016()
				.method_10059(field_11867)
				.equals(shaft.enginePos))
				return;
			if (shaft.engineEfficiency == 0)
				return;
			class_2350 facing = SteamEngineBlock.getFacing(method_11010());
			if (field_11863.method_8477(field_11867.method_10093(facing.method_10153())))
				shaft.update(field_11867, 0, 0);
			return;
		}

		boolean verticalTarget = false;
		class_2680 shaftState = shaft.method_11010();
		class_2351 targetAxis = class_2351.field_11048;
		if (shaftState.method_26204()instanceof IRotate ir)
			targetAxis = ir.getRotationAxis(shaftState);
		verticalTarget = targetAxis == class_2351.field_11052;

		class_2680 blockState = method_11010();
		if (!AllBlocks.STEAM_ENGINE.has(blockState))
			return;
		class_2350 facing = SteamEngineBlock.getFacing(blockState);
		if (facing.method_10166() == class_2351.field_11052)
			facing = blockState.method_11654(SteamEngineBlock.field_11177);

		float efficiency = class_3532.method_15363(tank.boiler.getEngineEfficiency(tank.getTotalTankSize()), 0, 1);
		if (efficiency > 0)

			award(AllAdvancements.STEAM_ENGINE);

		int conveyedSpeedLevel =
			efficiency == 0 ? 1 : verticalTarget ? 1 : (int) GeneratingKineticBlockEntity.convertToDirection(1, facing);
		if (targetAxis == class_2351.field_11051)
			conveyedSpeedLevel *= -1;
		if (movementDirection.get() == RotationDirection.COUNTER_CLOCKWISE)
			conveyedSpeedLevel *= -1;

		float shaftSpeed = shaft.getTheoreticalSpeed();
		if (shaft.hasSource() && shaftSpeed != 0 && conveyedSpeedLevel != 0
			&& (shaftSpeed > 0) != (conveyedSpeedLevel > 0)) {
			movementDirection.setValue(1 - movementDirection.get()
				.ordinal());
			conveyedSpeedLevel *= -1;
		}

		shaft.update(field_11867, conveyedSpeedLevel, efficiency);

		if (!field_11863.field_9236)
			return;

		EnvExecutor.runWhenOn(EnvType.CLIENT, () -> this::spawnParticles);
	}

	@Override
	public void remove() {
		PoweredShaftBlockEntity shaft = getShaft();
		if (shaft != null)
			shaft.remove(field_11867);
		super.remove();
	}

	@Override
	@Environment(EnvType.CLIENT)
	protected class_238 createRenderBoundingBox() {
		return super.createRenderBoundingBox().method_1014(2);
	}

	public PoweredShaftBlockEntity getShaft() {
		PoweredShaftBlockEntity shaft = target.get();
		if (shaft == null || shaft.method_11015() || !shaft.canBePoweredBy(field_11867)) {
			if (shaft != null)
				target = new WeakReference<>(null);
			class_2350 facing = SteamEngineBlock.getFacing(method_11010());
			class_2586 anyShaftAt = field_11863.method_8321(field_11867.method_10079(facing, 2));
			if (anyShaftAt instanceof PoweredShaftBlockEntity ps && ps.canBePoweredBy(field_11867))
				target = new WeakReference<>(shaft = ps);
		}
		return shaft;
	}

	public FluidTankBlockEntity getTank() {
		FluidTankBlockEntity tank = source.get();
		if (tank == null || tank.method_11015()) {
			if (tank != null)
				source = new WeakReference<>(null);
			class_2350 facing = SteamEngineBlock.getFacing(method_11010());
			class_2586 be = field_11863.method_8321(field_11867.method_10093(facing.method_10153()));
			if (be instanceof FluidTankBlockEntity tankBe)
				source = new WeakReference<>(tank = tankBe);
		}
		if (tank == null)
			return null;
		return tank.getControllerBE();
	}

	float prevAngle = 0;

	@Environment(EnvType.CLIENT)
	private void spawnParticles() {
		Float targetAngle = getTargetAngle();
		PoweredShaftBlockEntity ste = target.get();
		if (ste == null)
			return;
		if (!ste.isPoweredBy(field_11867) || ste.engineEfficiency == 0)
			return;
		if (targetAngle == null)
			return;

		float angle = AngleHelper.deg(targetAngle);
		angle += (angle < 0) ? -180 + 75 : 360 - 75;
		angle %= 360;

		PoweredShaftBlockEntity shaft = getShaft();
		if (shaft == null || shaft.getSpeed() == 0)
			return;

		if (angle >= 0 && !(prevAngle > 180 && angle < 180)) {
			prevAngle = angle;
			return;
		}
		if (angle < 0 && !(prevAngle < -180 && angle > -180)) {
			prevAngle = angle;
			return;
		}

		FluidTankBlockEntity sourceBE = source.get();
		if (sourceBE != null) {
			FluidTankBlockEntity controller = sourceBE.getControllerBE();
			if (controller != null && controller.boiler != null) {
				float volume = 3f / Math.max(2, controller.boiler.attachedEngines / 6);
				float pitch = 1.18f - field_11863.field_9229.method_43057() * .25f;
				field_11863.method_8486(field_11867.method_10263(), field_11867.method_10264(), field_11867.method_10260(),
					class_3417.field_26955, class_3419.field_15245, volume, pitch, false);
				AllSoundEvents.STEAM.playAt(field_11863, field_11867, volume / 16, .8f, false);
			}
		}

		class_2350 facing = SteamEngineBlock.getFacing(method_11010());

		class_243 offset = VecHelper.rotate(new class_243(0, 0, 1).method_1019(VecHelper.offsetRandomly(class_243.field_1353, field_11863.field_9229, 1)
			.method_18805(1, 1, 0)
			.method_1029()
			.method_1021(.5f)), AngleHelper.verticalAngle(facing), class_2351.field_11048);
		offset = VecHelper.rotate(offset, AngleHelper.horizontalAngle(facing), class_2351.field_11052);
		class_243 v = offset.method_1021(.5f)
			.method_1019(class_243.method_24953(field_11867));
		class_243 m = offset.method_1020(class_243.method_24954(facing.method_10163())
			.method_1021(.75f));
		field_11863.method_8406(new SteamJetParticleData(1), v.field_1352, v.field_1351, v.field_1350, m.field_1352, m.field_1351, m.field_1350);

		prevAngle = angle;
	}

	@Nullable
	@Environment(EnvType.CLIENT)
	public Float getTargetAngle() {
		float angle = 0;
		class_2680 blockState = method_11010();
		if (!AllBlocks.STEAM_ENGINE.has(blockState))
			return null;

		class_2350 facing = SteamEngineBlock.getFacing(blockState);
		PoweredShaftBlockEntity shaft = getShaft();
		class_2351 facingAxis = facing.method_10166();
		class_2351 axis = class_2351.field_11052;

		if (shaft == null)
			return null;

		axis = KineticBlockEntityRenderer.getRotationAxisOf(shaft);
		angle = KineticBlockEntityRenderer.getAngleForTe(shaft, shaft.method_11016(), axis);

		if (axis == facingAxis)
			return null;
		if (axis.method_10179() && (facingAxis == class_2351.field_11048 ^ facing.method_10171() == class_2352.field_11056))
			angle *= -1;
		if (axis == class_2351.field_11048 && facing == class_2350.field_11033)
			angle *= -1;
		return angle;
	}

	@Override
	public boolean addToGoggleTooltip(List<class_2561> tooltip, boolean isPlayerSneaking) {
		PoweredShaftBlockEntity shaft = getShaft();
		return shaft == null ? false : shaft.addToEngineTooltip(tooltip, isPlayerSneaking);
	}

}
