package com.simibubi.create.content.kinetics.steamEngine;

import com.jozufozu.flywheel.api.MaterialManager;
import com.jozufozu.flywheel.api.instance.DynamicInstance;
import com.jozufozu.flywheel.backend.instancing.blockentity.BlockEntityInstance;
import com.jozufozu.flywheel.core.Materials;
import com.jozufozu.flywheel.core.materials.model.ModelData;
import com.simibubi.create.AllPartialModels;
import com.simibubi.create.content.kinetics.base.KineticBlockEntityRenderer;
import com.simibubi.create.foundation.utility.AngleHelper;
import net.minecraft.class_2350;
import net.minecraft.class_2350.class_2351;
import net.minecraft.class_3532;

public class SteamEngineInstance extends BlockEntityInstance<SteamEngineBlockEntity> implements DynamicInstance {

	protected final ModelData piston;
	protected final ModelData linkage;
	protected final ModelData connector;

	public SteamEngineInstance(MaterialManager materialManager, SteamEngineBlockEntity blockEntity) {
		super(materialManager, blockEntity);

		piston = materialManager.defaultSolid()
				.material(Materials.TRANSFORMED)
				.getModel(AllPartialModels.ENGINE_PISTON, blockState)
				.createInstance();
		linkage = materialManager.defaultSolid()
				.material(Materials.TRANSFORMED)
				.getModel(AllPartialModels.ENGINE_LINKAGE, blockState)
				.createInstance();
		connector = materialManager.defaultSolid()
				.material(Materials.TRANSFORMED)
				.getModel(AllPartialModels.ENGINE_CONNECTOR, blockState)
				.createInstance();
	}

	@Override
	public void beginFrame() {
		Float angle = blockEntity.getTargetAngle();
		if (angle == null) {
			piston.setEmptyTransform();
			linkage.setEmptyTransform();
			connector.setEmptyTransform();
			return;
		}

		class_2350 facing = SteamEngineBlock.getFacing(blockState);
		class_2351 facingAxis = facing.method_10166();
		class_2351 axis = class_2351.field_11052;

		PoweredShaftBlockEntity shaft = blockEntity.getShaft();
		if (shaft != null)
			axis = KineticBlockEntityRenderer.getRotationAxisOf(shaft);

		boolean roll90 = facingAxis.method_10179() && axis == class_2351.field_11052 || facingAxis.method_10178() && axis == class_2351.field_11051;
		float sine = class_3532.method_15374(angle);
		float sine2 = class_3532.method_15374(angle - class_3532.field_29845);
		float piston = ((1 - sine) / 4) * 24 / 16f;

		transformed(this.piston, facing, roll90)
			.translate(0, piston, 0);

		transformed(linkage, facing, roll90)
			.centre()
			.translate(0, 1, 0)
			.unCentre()
			.translate(0, piston, 0)
			.translate(0, 4 / 16f, 8 / 16f)
			.rotateX(sine2 * 23f)
			.translate(0, -4 / 16f, -8 / 16f);

		transformed(connector, facing, roll90)
			.translate(0, 2, 0)
			.centre()
			.rotateXRadians(-angle + class_3532.field_29845)
			.unCentre();
	}

	protected ModelData transformed(ModelData modelData, class_2350 facing, boolean roll90) {
		return modelData.loadIdentity()
			.translate(getInstancePosition())
			.centre()
			.rotateY(AngleHelper.horizontalAngle(facing))
			.rotateX(AngleHelper.verticalAngle(facing) + 90)
			.rotateY(roll90 ? -90 : 0)
			.unCentre();
	}

	@Override
	public void updateLight() {
		relight(pos, piston, linkage, connector);
	}

	@Override
	protected void remove() {
		piston.delete();
		linkage.delete();
		connector.delete();
	}

}
