package com.simibubi.create.content.kinetics.transmission.sequencer;

import java.util.List;
import java.util.Vector;
import net.minecraft.class_2338;
import net.minecraft.class_2350;
import net.minecraft.class_2487;
import net.minecraft.class_2520;
import net.minecraft.class_2591;
import net.minecraft.class_2680;
import com.simibubi.create.compat.computercraft.AbstractComputerBehaviour;
import com.simibubi.create.compat.computercraft.ComputerCraftProxy;
import com.simibubi.create.content.kinetics.base.KineticBlockEntity;
import com.simibubi.create.content.kinetics.transmission.SplitShaftBlockEntity;
import com.simibubi.create.foundation.blockEntity.behaviour.BlockEntityBehaviour;
import com.simibubi.create.foundation.utility.NBTHelper;

public class SequencedGearshiftBlockEntity extends SplitShaftBlockEntity {

	Vector<Instruction> instructions;
	int currentInstruction;
	int currentInstructionDuration;
	float currentInstructionProgress;
	int timer;
	boolean poweredPreviously;

	public AbstractComputerBehaviour computerBehaviour;

	public record SequenceContext(SequencerInstructions instruction, double relativeValue) {

		public static SequenceContext fromGearshift(SequencerInstructions instruction, double kineticSpeed,
			int absoluteValue) {
			return instruction.needsPropagation()
				? new SequenceContext(instruction, kineticSpeed == 0 ? 0 : absoluteValue / kineticSpeed)
				: null;
		}

		public double getEffectiveValue(double speedAtTarget) {
			return Math.abs(relativeValue * speedAtTarget);
		}

		public class_2487 serializeNBT() {
			class_2487 nbt = new class_2487();
			NBTHelper.writeEnum(nbt, "Mode", instruction);
			nbt.method_10549("Value", relativeValue);
			return nbt;
		}

		public static SequenceContext fromNBT(class_2487 nbt) {
			if (nbt.method_33133())
				return null;
			return new SequenceContext(NBTHelper.readEnum(nbt, "Mode", SequencerInstructions.class),
				nbt.method_10574("Value"));
		}

	}

	public SequencedGearshiftBlockEntity(class_2591<?> type, class_2338 pos, class_2680 state) {
		super(type, pos, state);
		instructions = Instruction.createDefault();
		currentInstruction = -1;
		currentInstructionDuration = -1;
		currentInstructionProgress = 0;
		timer = 0;
		poweredPreviously = false;
	}

	@Override
	public void addBehaviours(List<BlockEntityBehaviour> behaviours) {
		super.addBehaviours(behaviours);
		behaviours.add(computerBehaviour = ComputerCraftProxy.behaviour(this));
	}

	@Override
	public void tick() {
		super.tick();

		if (isIdle())
			return;
		if (field_11863.field_9236)
			return;
		if (currentInstructionDuration < 0)
			return;
		if (timer < currentInstructionDuration) {
			timer++;
			currentInstructionProgress += getInstruction(currentInstruction).getTickProgress(speed);
			return;
		}
		run(currentInstruction + 1);
	}

	@Override
	public void onSpeedChanged(float previousSpeed) {
		super.onSpeedChanged(previousSpeed);
		if (isIdle())
			return;
		float currentSpeed = Math.abs(speed);
		if (Math.abs(previousSpeed) == currentSpeed)
			return;
		Instruction instruction = getInstruction(currentInstruction);
		if (instruction == null)
			return;
		if (getSpeed() == 0)
			run(-1);

		// Update instruction time with regards to new speed
		currentInstructionDuration = instruction.getDuration(currentInstructionProgress, getTheoreticalSpeed());
		timer = 0;
	}

	public boolean isIdle() {
		return currentInstruction == -1;
	}

	public void onRedstoneUpdate(boolean isPowered, boolean isRunning) {
		if (computerBehaviour.hasAttachedComputer())
			return;
		if (!poweredPreviously && isPowered)
			risingFlank();
		poweredPreviously = isPowered;
		if (!isIdle())
			return;
		if (isPowered == isRunning)
			return;
		if (!field_11863.method_49803(field_11867)) {
			field_11863.method_8652(field_11867, method_11010().method_11657(SequencedGearshiftBlock.STATE, 0), 3);
			return;
		}
		if (getSpeed() == 0)
			return;
		run(0);
	}

	public void risingFlank() {
		Instruction instruction = getInstruction(currentInstruction);
		if (instruction == null)
			return;
		if (poweredPreviously)
			return;
		poweredPreviously = true;

		switch (instruction.onRedstonePulse()) {
		case CONTINUE:
			run(currentInstruction + 1);
			break;
		default:
			break;
		}
	}

	public void run(int instructionIndex) {
		Instruction instruction = getInstruction(instructionIndex);
		if (instruction == null || instruction.instruction == SequencerInstructions.END) {
			if (getModifier() != 0)
				detachKinetics();
			currentInstruction = -1;
			currentInstructionDuration = -1;
			currentInstructionProgress = 0;
			sequenceContext = null;
			timer = 0;
			if (!field_11863.method_49803(field_11867))
				field_11863.method_8652(field_11867, method_11010().method_11657(SequencedGearshiftBlock.STATE, 0), 3);
			else
				sendData();
			return;
		}

		detachKinetics();
		currentInstructionDuration = instruction.getDuration(0, getTheoreticalSpeed());
		currentInstruction = instructionIndex;
		currentInstructionProgress = 0;
		sequenceContext = SequenceContext.fromGearshift(instruction.instruction, getTheoreticalSpeed() * getModifier(),
			instruction.value);
		timer = 0;
		field_11863.method_8652(field_11867, method_11010().method_11657(SequencedGearshiftBlock.STATE, instructionIndex + 1), 3);
	}

	public Instruction getInstruction(int instructionIndex) {
		return instructionIndex >= 0 && instructionIndex < instructions.size() ? instructions.get(instructionIndex)
			: null;
	}

	@Override
	protected void copySequenceContextFrom(KineticBlockEntity sourceBE) {}

	@Override
	public void write(class_2487 compound, boolean clientPacket) {
		compound.method_10569("InstructionIndex", currentInstruction);
		compound.method_10569("InstructionDuration", currentInstructionDuration);
		compound.method_10548("InstructionProgress", currentInstructionProgress);
		compound.method_10569("Timer", timer);
		compound.method_10556("PrevPowered", poweredPreviously);
		compound.method_10566("Instructions", Instruction.serializeAll(instructions));
		super.write(compound, clientPacket);
	}

	@Override
	protected void read(class_2487 compound, boolean clientPacket) {
		currentInstruction = compound.method_10550("InstructionIndex");
		currentInstructionDuration = compound.method_10550("InstructionDuration");
		currentInstructionProgress = compound.method_10583("InstructionProgress");
		poweredPreviously = compound.method_10577("PrevPowered");
		timer = compound.method_10550("Timer");
		instructions = Instruction.deserializeAll(compound.method_10554("Instructions", class_2520.field_33260));
		super.read(compound, clientPacket);
	}

	@Override
	public float getRotationSpeedModifier(class_2350 face) {
		if (isVirtual())
			return 1;
		return (!hasSource() || face == getSourceFacing()) ? 1 : getModifier();
	}

	public int getModifier() {
		if (currentInstruction >= instructions.size())
			return 0;
		return isIdle() ? 0
			: instructions.get(currentInstruction)
				.getSpeedModifier();
	}

	public Vector<Instruction> getInstructions() {
		return this.instructions;
	}

}
