package com.simibubi.create.content.kinetics.waterwheel;

import java.util.EnumMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import net.minecraft.class_1269;
import net.minecraft.class_1747;
import net.minecraft.class_1799;
import net.minecraft.class_2246;
import net.minecraft.class_2248;
import net.minecraft.class_2258;
import net.minecraft.class_2338;
import net.minecraft.class_2350;
import net.minecraft.class_2350.class_2351;
import net.minecraft.class_2350.class_2352;
import net.minecraft.class_238;
import net.minecraft.class_2382;
import net.minecraft.class_243;
import net.minecraft.class_2487;
import net.minecraft.class_2512;
import net.minecraft.class_2591;
import net.minecraft.class_2680;
import net.minecraft.class_3481;
import net.minecraft.class_3532;
import net.minecraft.class_3610;
import com.simibubi.create.content.kinetics.base.GeneratingKineticBlockEntity;
import com.simibubi.create.content.kinetics.base.IRotate;
import com.simibubi.create.foundation.advancement.AllAdvancements;
import com.simibubi.create.foundation.blockEntity.behaviour.BlockEntityBehaviour;
import com.simibubi.create.foundation.fluid.FluidHelper;
import com.simibubi.create.foundation.utility.Iterate;
import com.simibubi.create.foundation.utility.VecHelper;

public class WaterWheelBlockEntity extends GeneratingKineticBlockEntity {

	public static final Map<class_2351, Set<class_2338>> SMALL_OFFSETS = new EnumMap<>(class_2351.class);
	public static final Map<class_2351, Set<class_2338>> LARGE_OFFSETS = new EnumMap<>(class_2351.class);

	static {
		for (class_2351 axis : Iterate.axes) {
			HashSet<class_2338> offsets = new HashSet<>();
			for (class_2350 d : Iterate.directions)
				if (d.method_10166() != axis)
					offsets.add(class_2338.field_10980.method_10093(d));
			SMALL_OFFSETS.put(axis, offsets);

			offsets = new HashSet<>();
			for (class_2350 d : Iterate.directions) {
				if (d.method_10166() == axis)
					continue;
				class_2338 centralOffset = class_2338.field_10980.method_10079(d, 2);
				offsets.add(centralOffset);
				for (class_2350 d2 : Iterate.directions) {
					if (d2.method_10166() == axis)
						continue;
					if (d2.method_10166() == d.method_10166())
						continue;
					offsets.add(centralOffset.method_10093(d2));
				}
			}
			LARGE_OFFSETS.put(axis, offsets);
		}
	}

	public int flowScore;
	public class_2680 material;

	public WaterWheelBlockEntity(class_2591<?> type, class_2338 pos, class_2680 state) {
		super(type, pos, state);
		material = class_2246.field_9975.method_9564();
		setLazyTickRate(60);
	}

	protected int getSize() {
		return 1;
	}

	protected Set<class_2338> getOffsetsToCheck() {
		return (getSize() == 1 ? SMALL_OFFSETS : LARGE_OFFSETS).get(getAxis());
	}

	public class_1269 applyMaterialIfValid(class_1799 stack) {
		if (!(stack.method_7909()instanceof class_1747 blockItem))
			return class_1269.field_5811;
		class_2680 material = blockItem.method_7711()
			.method_9564();
		if (material == this.material)
			return class_1269.field_5811;
		if (!material.method_26164(class_3481.field_15471))
			return class_1269.field_5811;
		if (field_11863.method_8608() && !isVirtual())
			return class_1269.field_5812;
		this.material = material;
		notifyUpdate();
		field_11863.method_20290(2001, field_11867, class_2248.method_9507(material));
		return class_1269.field_5812;
	}

	protected class_2351 getAxis() {
		class_2351 axis = class_2351.field_11048;
		class_2680 blockState = method_11010();
		if (blockState.method_26204()instanceof IRotate irotate)
			axis = irotate.getRotationAxis(blockState);
		return axis;
	}

	@Override
	public void lazyTick() {
		super.lazyTick();

		// Water can change flow direction without notifying neighbours
		determineAndApplyFlowScore();
	}

	public void determineAndApplyFlowScore() {
		class_243 wheelPlane =
			class_243.method_24954(new class_2382(1, 1, 1).method_35852(class_2350.method_10156(class_2352.field_11056, getAxis())
				.method_10163()));

		int flowScore = 0;
		boolean lava = false;
		for (class_2338 blockPos : getOffsetsToCheck()) {
			class_2338 targetPos = blockPos.method_10081(field_11867);
			class_243 flowAtPos = getFlowVectorAtPosition(targetPos).method_18806(wheelPlane);
			lava |= FluidHelper.isLava(field_11863.method_8316(targetPos)
				.method_15772());

			if (flowAtPos.method_1027() == 0)
				continue;

			flowAtPos = flowAtPos.method_1029();
			class_243 normal = class_243.method_24954(blockPos)
				.method_1029();

			class_243 positiveMotion = VecHelper.rotate(normal, 90, getAxis());
			double dot = flowAtPos.method_1026(positiveMotion);
			if (Math.abs(dot) > .5)
				flowScore += Math.signum(dot);
		}

		if (flowScore != 0 && !field_11863.method_8608())
			award(lava ? AllAdvancements.LAVA_WHEEL : AllAdvancements.WATER_WHEEL);

		setFlowScoreAndUpdate(flowScore);
	}

	public class_243 getFlowVectorAtPosition(class_2338 pos) {
		class_3610 fluid = field_11863.method_8316(pos);
		class_243 vec = fluid.method_15758(field_11863, pos);
		class_2680 blockState = field_11863.method_8320(pos);
		if (blockState.method_26204() == class_2246.field_10422)
			vec = new class_243(0, blockState.method_11654(class_2258.field_10680) ? -1 : 1, 0);
		return vec;
	}

	public void setFlowScoreAndUpdate(int score) {
		if (flowScore == score)
			return;
		flowScore = score;
		updateGeneratedRotation();
		method_5431();
	}

	private void redraw() {
		if (!isVirtual())
			requestModelDataUpdate();
		if (method_11002()) {
			field_11863.method_8413(method_11016(), method_11010(), method_11010(), 16);
			field_11863.method_8398()
				.method_12130()
				.method_15513(field_11867);
		}
	}

	@Override
	public void addBehaviours(List<BlockEntityBehaviour> behaviours) {
		super.addBehaviours(behaviours);
		registerAwardables(behaviours, AllAdvancements.LAVA_WHEEL, AllAdvancements.WATER_WHEEL);
	}

	@Override
	protected void read(class_2487 compound, boolean clientPacket) {
		super.read(compound, clientPacket);
		flowScore = compound.method_10550("FlowScore");

		class_2680 prevMaterial = material;
		if (!compound.method_10545("Material"))
			return;

		material = class_2512.method_10681(blockHolderGetter(), compound.method_10562("Material"));
		if (material.method_26215())
			material = class_2246.field_9975.method_9564();

		if (clientPacket && prevMaterial != material)
			redraw();
	}

	@Override
	public void write(class_2487 compound, boolean clientPacket) {
		super.write(compound, clientPacket);
		compound.method_10569("FlowScore", flowScore);
		compound.method_10566("Material", class_2512.method_10686(material));
	}

	@Override
	protected class_238 createRenderBoundingBox() {
		return new class_238(field_11867).method_1014(getSize());
	}

	@Override
	public float getGeneratedSpeed() {
		return class_3532.method_15340(flowScore, -1, 1) * 8 / getSize();
	}

}
