package com.simibubi.create.content.logistics.crate;

import java.util.Collections;
import java.util.Iterator;
import java.util.function.Supplier;

import javax.annotation.ParametersAreNonnullByDefault;

import io.github.fabricators_of_create.porting_lib.transfer.item.ItemStackHandler;
import io.github.fabricators_of_create.porting_lib.transfer.item.ItemStackHandlerSlot;
import net.fabricmc.fabric.api.transfer.v1.item.ItemVariant;
import net.fabricmc.fabric.api.transfer.v1.storage.StorageView;
import net.fabricmc.fabric.api.transfer.v1.storage.base.SingleSlotStorage;
import net.fabricmc.fabric.api.transfer.v1.transaction.TransactionContext;
import net.minecraft.class_1799;
import net.minecraft.class_6328;

@class_6328
@ParametersAreNonnullByDefault
public class BottomlessItemHandler extends ItemStackHandler implements SingleSlotStorage<ItemVariant> { // must extend ItemStackHandler for mounted storages

	private Supplier<class_1799> suppliedItemStack;

	public BottomlessItemHandler(Supplier<class_1799> suppliedItemStack) {
		super(0);
		this.suppliedItemStack = suppliedItemStack;
		setSize(1); // create slot after setting supplier
	}

	@Override
	protected ItemStackHandlerSlot makeSlot(int index, class_1799 stack) {
		return new BottomlessSlot();
	}

	@Override
	public long insert(ItemVariant resource, long maxAmount, TransactionContext transaction) {
		return maxAmount;
	}

	@Override
	public long extract(ItemVariant resource, long maxAmount, TransactionContext transaction) {
		class_1799 stack = getStack();
		if (!resource.matches(stack))
			return 0;
		if (!stack.method_7960())
			return Math.min(stack.method_7914(), maxAmount);
		return 0;
	}

	protected class_1799 getStack() {
		class_1799 stack = suppliedItemStack.get();
		return stack == null || stack.method_7960() ? class_1799.field_8037 : stack;
	}

	@Override
	public boolean isResourceBlank() {
		return getStack().method_7960();
	}

	@Override
	public ItemVariant getResource() {
		return ItemVariant.of(getStack());
	}

	@Override
	public long getAmount() {
		return Long.MAX_VALUE;
	}

	@Override
	public long getCapacity() {
		return Long.MAX_VALUE;
	}

	// shortcuts

	@Override
	public Iterator<StorageView<ItemVariant>> iterator() {
		return SingleSlotStorage.super.iterator(); // singleton iterator on this
	}

	@Override
	public Iterable<StorageView<ItemVariant>> nonEmptyViews() {
		return this::nonEmptyIterator;
	}

	@Override
	public Iterator<StorageView<ItemVariant>> nonEmptyIterator() {
		return isResourceBlank() ? Collections.emptyIterator() : iterator();
	}

	private class BottomlessSlot extends ItemStackHandlerSlot {
		public BottomlessSlot() {
			super(0, BottomlessItemHandler.this, class_1799.field_8037);
		}

		@Override
		public class_1799 getStack() {
			return BottomlessItemHandler.this.getStack();
		}

		@Override
		public ItemVariant getResource() {
			return BottomlessItemHandler.this.getResource();
		}

		@Override
		public long getAmount() {
			return BottomlessItemHandler.this.getAmount();
		}

		@Override
		protected void setStack(class_1799 stack) {
		}

		@Override
		protected void onFinalCommit() {
		}
	}
}
