package com.simibubi.create.content.logistics.depot;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.function.Consumer;
import java.util.function.Function;
import java.util.function.Predicate;
import java.util.function.Supplier;

import com.simibubi.create.AllSoundEvents;
import com.simibubi.create.content.kinetics.belt.BeltHelper;
import com.simibubi.create.content.kinetics.belt.behaviour.BeltProcessingBehaviour;
import com.simibubi.create.content.kinetics.belt.behaviour.BeltProcessingBehaviour.ProcessingResult;
import com.simibubi.create.content.kinetics.belt.behaviour.DirectBeltInputBehaviour;
import com.simibubi.create.content.kinetics.belt.behaviour.TransportedItemStackHandlerBehaviour;
import com.simibubi.create.content.kinetics.belt.behaviour.TransportedItemStackHandlerBehaviour.TransportedResult;
import com.simibubi.create.content.kinetics.belt.transport.TransportedItemStack;
import com.simibubi.create.content.logistics.funnel.AbstractFunnelBlock;
import com.simibubi.create.foundation.blockEntity.SmartBlockEntity;
import com.simibubi.create.foundation.blockEntity.behaviour.BehaviourType;
import com.simibubi.create.foundation.blockEntity.behaviour.BlockEntityBehaviour;
import com.simibubi.create.foundation.item.ItemHelper;
import com.simibubi.create.foundation.utility.NBTHelper;
import com.simibubi.create.foundation.utility.VecHelper;

import io.github.fabricators_of_create.porting_lib.transfer.TransferUtil;
import io.github.fabricators_of_create.porting_lib.transfer.callbacks.TransactionCallback;
import io.github.fabricators_of_create.porting_lib.transfer.item.ItemHandlerHelper;
import io.github.fabricators_of_create.porting_lib.transfer.item.ItemStackHandler;
import io.github.fabricators_of_create.porting_lib.util.ItemStackUtil;
import net.fabricmc.fabric.api.transfer.v1.item.ItemVariant;
import net.fabricmc.fabric.api.transfer.v1.transaction.Transaction;
import net.fabricmc.fabric.api.transfer.v1.transaction.TransactionContext;
import net.fabricmc.fabric.api.transfer.v1.transaction.base.SnapshotParticipant;
import net.minecraft.class_1264;
import net.minecraft.class_1799;
import net.minecraft.class_1937;
import net.minecraft.class_2248;
import net.minecraft.class_2338;
import net.minecraft.class_2350;
import net.minecraft.class_243;
import net.minecraft.class_2487;
import net.minecraft.class_2499;
import net.minecraft.class_2520;
import net.minecraft.class_2680;

public class DepotBehaviour extends BlockEntityBehaviour {

	public static final BehaviourType<DepotBehaviour> TYPE = new BehaviourType<>();

	TransportedItemStack heldItem;
	List<TransportedItemStack> incoming;
	ItemStackHandler processingOutputBuffer;
	public DepotItemHandler itemHandler;
	TransportedItemStackHandlerBehaviour transportedHandler;
	Supplier<Integer> maxStackSize;
	Supplier<Boolean> canAcceptItems;
	Predicate<class_2350> canFunnelsPullFrom;
	Consumer<class_1799> onHeldInserted;
	Predicate<class_1799> acceptedItems;
	boolean allowMerge;

	SnapshotParticipant<Data> snapshotParticipant = new SnapshotParticipant<>() {
		@Override
		protected Data createSnapshot() {
			// incoming stacks are not mutated during transfer, no need to deep copy
			return new Data(new ArrayList<>(incoming), heldItem == null ? null : heldItem.fullCopy());
		}

		@Override
		protected void readSnapshot(Data snapshot) {
			incoming = snapshot.incoming;
			heldItem = snapshot.held;
		}

		@Override
		protected void onFinalCommit() {
			blockEntity.notifyUpdate();
		}
	};

	record Data(List<TransportedItemStack> incoming, TransportedItemStack held) {
	}

	public DepotBehaviour(SmartBlockEntity be) {
		super(be);
		maxStackSize = () -> 64;
		canAcceptItems = () -> true;
		canFunnelsPullFrom = $ -> true;
		acceptedItems = $ -> true;
		onHeldInserted = $ -> {
		};
		incoming = new ArrayList<>();
		itemHandler = new DepotItemHandler(this);
		processingOutputBuffer = new ItemStackHandler(8) {
			protected void onContentsChanged(int slot) {
				be.notifyUpdate();
			};
		};
	}

	public void enableMerging() {
		allowMerge = true;
	}

	public DepotBehaviour withCallback(Consumer<class_1799> changeListener) {
		onHeldInserted = changeListener;
		return this;
	}

	public DepotBehaviour onlyAccepts(Predicate<class_1799> filter) {
		acceptedItems = filter;
		return this;
	}

	@Override
	public void tick() {
		super.tick();

		class_1937 world = blockEntity.method_10997();

		for (Iterator<TransportedItemStack> iterator = incoming.iterator(); iterator.hasNext();) {
			TransportedItemStack ts = iterator.next();
			if (!tick(ts))
				continue;
			if (world.field_9236 && !blockEntity.isVirtual())
				continue;
			if (heldItem == null) {
				heldItem = ts;
			} else {
				if (!ItemHelper.canItemStackAmountsStack(heldItem.stack, ts.stack)) {
					class_243 vec = VecHelper.getCenterOf(blockEntity.method_11016());
					class_1264.method_5449(blockEntity.method_10997(), vec.field_1352, vec.field_1351 + .5f, vec.field_1350, ts.stack);
				} else {
					heldItem.stack.method_7933(ts.stack.method_7947());
				}
			}
			iterator.remove();
			blockEntity.notifyUpdate();
		}

		if (heldItem == null)
			return;
		if (!tick(heldItem))
			return;

		class_2338 pos = blockEntity.method_11016();

		if (world.field_9236)
			return;
		if (handleBeltFunnelOutput())
			return;

		BeltProcessingBehaviour processingBehaviour =
			BlockEntityBehaviour.get(world, pos.method_10086(2), BeltProcessingBehaviour.TYPE);
		if (processingBehaviour == null)
			return;
		if (!heldItem.locked && BeltProcessingBehaviour.isBlocked(world, pos))
			return;

		class_1799 previousItem = heldItem.stack;
		boolean wasLocked = heldItem.locked;
		ProcessingResult result = wasLocked ? processingBehaviour.handleHeldItem(heldItem, transportedHandler)
			: processingBehaviour.handleReceivedItem(heldItem, transportedHandler);
		if (result == ProcessingResult.REMOVE) {
			heldItem = null;
			blockEntity.sendData();
			return;
		}

		// fabric: might be set to null in processing
		if (heldItem == null) {
			blockEntity.sendData();
			return;
		}

		heldItem.locked = result == ProcessingResult.HOLD;
		if (heldItem.locked != wasLocked || !class_1799.method_7973(previousItem, heldItem.stack))
			blockEntity.sendData();
	}

	protected boolean tick(TransportedItemStack heldItem) {
		heldItem.prevBeltPosition = heldItem.beltPosition;
		heldItem.prevSideOffset = heldItem.sideOffset;
		float diff = .5f - heldItem.beltPosition;
		if (diff > 1 / 512f) {
			if (diff > 1 / 32f && !BeltHelper.isItemUpright(heldItem.stack))
				heldItem.angle += 1;
			heldItem.beltPosition += diff / 4f;
		}
		return diff < 1 / 16f;
	}

	private boolean handleBeltFunnelOutput() {
		class_2680 funnel = getWorld().method_8320(getPos().method_10084());
		class_2350 funnelFacing = AbstractFunnelBlock.getFunnelFacing(funnel);
		if (funnelFacing == null || !canFunnelsPullFrom.test(funnelFacing.method_10153()))
			return false;

		for (int slot = 0; slot < processingOutputBuffer.getSlotCount(); slot++) {
			class_1799 previousItem = processingOutputBuffer.getStackInSlot(slot);
			if (previousItem.method_7960())
				continue;
			class_1799 afterInsert = blockEntity.getBehaviour(DirectBeltInputBehaviour.TYPE)
				.tryExportingToBeltFunnel(previousItem, null, false);
			if (afterInsert == null)
				return false;
			if (previousItem.method_7947() != afterInsert.method_7947()) {
				processingOutputBuffer.setStackInSlot(slot, afterInsert);
				blockEntity.notifyUpdate();
				return true;
			}
		}

		class_1799 previousItem = heldItem.stack;
		if (previousItem.method_7960()) { // fabric: this is not allowed
			return false;
		}
		class_1799 afterInsert = blockEntity.getBehaviour(DirectBeltInputBehaviour.TYPE)
			.tryExportingToBeltFunnel(previousItem, null, false);
		if (afterInsert == null)
			return false;
		if (previousItem.method_7947() != afterInsert.method_7947()) {
			if (afterInsert.method_7960())
				heldItem = null;
			else
				heldItem.stack = afterInsert;
			blockEntity.notifyUpdate();
			return true;
		}

		return false;
	}

	@Override
	public void destroy() {
		super.destroy();
		class_1937 level = getWorld();
		class_2338 pos = getPos();
		ItemHelper.dropContents(level, pos, processingOutputBuffer);
		for (TransportedItemStack transportedItemStack : incoming)
			class_2248.method_9577(level, pos, transportedItemStack.stack);
		if (!getHeldItemStack().method_7960())
			class_2248.method_9577(level, pos, getHeldItemStack());
	}

	@Override
	public void unload() {
		itemHandler = null;
	}

	@Override
	public void write(class_2487 compound, boolean clientPacket) {
		if (heldItem != null)
			compound.method_10566("HeldItem", heldItem.serializeNBT());
		compound.method_10566("OutputBuffer", processingOutputBuffer.serializeNBT());
		if (canMergeItems() && !incoming.isEmpty())
			compound.method_10566("Incoming", NBTHelper.writeCompoundList(incoming, TransportedItemStack::serializeNBT));
	}

	@Override
	public void read(class_2487 compound, boolean clientPacket) {
		heldItem = null;
		if (compound.method_10545("HeldItem"))
			heldItem = TransportedItemStack.read(compound.method_10562("HeldItem"));
		processingOutputBuffer.deserializeNBT(compound.method_10562("OutputBuffer"));
		if (canMergeItems()) {
			class_2499 list = compound.method_10554("Incoming", class_2520.field_33260);
			incoming = NBTHelper.readCompoundList(list, TransportedItemStack::read);
		}
	}

	public void addSubBehaviours(List<BlockEntityBehaviour> behaviours) {
		behaviours.add(new DirectBeltInputBehaviour(blockEntity).allowingBeltFunnels()
			.setInsertionHandler(this::tryInsertingFromSide).considerOccupiedWhen(this::isOccupied));
		transportedHandler = new TransportedItemStackHandlerBehaviour(blockEntity, this::applyToAllItems)
			.withStackPlacement(this::getWorldPositionOf);
		behaviours.add(transportedHandler);
	}

	public class_1799 getHeldItemStack() {
		return heldItem == null ? class_1799.field_8037 : heldItem.stack;
	}

	public boolean canMergeItems() {
		return allowMerge;
	}

	public int getPresentStackSize() {
		int cumulativeStackSize = 0;
		cumulativeStackSize += getHeldItemStack().method_7947();
		for (int slot = 0; slot < processingOutputBuffer.getSlotCount(); slot++)
			cumulativeStackSize += processingOutputBuffer.getStackInSlot(slot)
				.method_7947();
		return cumulativeStackSize;
	}

	public int getRemainingSpace() {
		int cumulativeStackSize = getPresentStackSize();
		for (TransportedItemStack transportedItemStack : incoming)
			cumulativeStackSize += transportedItemStack.stack.method_7947();
		int fromGetter = maxStackSize.get();
		return (fromGetter == 0 ? 64 : fromGetter) - cumulativeStackSize;
	}

	public class_1799 insert(TransportedItemStack heldItem, TransactionContext ctx) {
		if (!canAcceptItems.get())
			return heldItem.stack;
		if (!acceptedItems.test(heldItem.stack))
			return heldItem.stack;

		if (canMergeItems()) {
			int remainingSpace = getRemainingSpace();
			class_1799 inserted = heldItem.stack;
			if (remainingSpace <= 0)
				return inserted;
			if (this.heldItem != null && !ItemHelper.canItemStackAmountsStack(this.heldItem.stack, inserted))
				return inserted;

			class_1799 returned = class_1799.field_8037;
			snapshotParticipant.updateSnapshots(ctx);
			if (remainingSpace < inserted.method_7947()) {
				returned = ItemHandlerHelper.copyStackWithSize(heldItem.stack, inserted.method_7947() - remainingSpace);
				TransportedItemStack copy = heldItem.copy();
				copy.stack.method_7939(remainingSpace);
				if (this.heldItem != null)
					incoming.add(copy);
				else
					this.heldItem = copy;
			} else {
				if (this.heldItem != null)
					incoming.add(heldItem);
				else
					this.heldItem = heldItem;
			}
			return returned;
		}

		if (this.isEmpty()) {
			if (heldItem.insertedFrom.method_10166()
					.method_10179())
				TransactionCallback.onSuccess(ctx, () -> AllSoundEvents.DEPOT_SLIDE.playOnServer(getWorld(), getPos()));
			else
				TransactionCallback.onSuccess(ctx, () -> AllSoundEvents.DEPOT_PLOP.playOnServer(getWorld(), getPos()));
		}
		snapshotParticipant.updateSnapshots(ctx);
		this.heldItem = heldItem;
		TransactionCallback.onSuccess(ctx, () -> onHeldInserted.accept(heldItem.stack));
		return class_1799.field_8037;
	}

	public void setHeldItem(TransportedItemStack heldItem) {
		this.heldItem = heldItem;
	}

	public void removeHeldItem() {
		this.heldItem = null;
	}

	public void setCenteredHeldItem(TransportedItemStack heldItem) {
		this.heldItem = heldItem;
		this.heldItem.beltPosition = 0.5f;
		this.heldItem.prevBeltPosition = 0.5f;
	}

//	public <T> LazyOptional<T> getItemCapability(Capability<T> cap, Direction side) {
//		return lazyItemHandler.cast();
//	}

	private boolean isOccupied(class_2350 side) {
		if (!getHeldItemStack().method_7960() && !canMergeItems())
			return true;
		if (!isOutputEmpty() && !canMergeItems())
			return true;
		if (!canAcceptItems.get())
			return true;
		return false;
	}
	
	private class_1799 tryInsertingFromSide(TransportedItemStack transportedStack, class_2350 side, boolean simulate) {
		class_1799 inserted = transportedStack.stack;

		if (isOccupied(side))
			return inserted;

		int size = transportedStack.stack.method_7947();
		transportedStack = transportedStack.copy();
		transportedStack.beltPosition = side.method_10166()
			.method_10178() ? .5f : 0;
		transportedStack.insertedFrom = side;
		transportedStack.prevSideOffset = transportedStack.sideOffset;
		transportedStack.prevBeltPosition = transportedStack.beltPosition;
		try (Transaction t = TransferUtil.getTransaction()) {
			snapshotParticipant.updateSnapshots(t);
			class_1799 remainder = insert(transportedStack, t);
			if (remainder.method_7947() != size)
				blockEntity.notifyUpdate();
			if (!simulate)
				t.commit();

			return remainder;
		}
	}

	private void applyToAllItems(float maxDistanceFromCentre,
		Function<TransportedItemStack, TransportedResult> processFunction) {
		if (heldItem == null)
			return;
		if (.5f - heldItem.beltPosition > maxDistanceFromCentre)
			return;

		boolean dirty = false;
		TransportedItemStack transportedItemStack = heldItem;
		class_1799 stackBefore = transportedItemStack.stack.method_7972();
		TransportedResult result = processFunction.apply(transportedItemStack);
		if (result == null || result.didntChangeFrom(stackBefore))
			return;

		dirty = true;
		heldItem = null;
		if (result.hasHeldOutput())
			setCenteredHeldItem(result.getHeldOutput());

		for (TransportedItemStack added : result.getOutputs()) {
			if (getHeldItemStack().method_7960()) {
				setCenteredHeldItem(added);
				continue;
			}
			try (Transaction t = TransferUtil.getTransaction()) {
				long inserted = processingOutputBuffer.insert(ItemVariant.of(added.stack), added.stack.method_7947(), t);
				t.commit();
				class_1799 remainder = added.stack.method_7972();
				remainder.method_7939(ItemHelper.truncateLong(added.stack.method_7947() - inserted));
				class_243 vec = VecHelper.getCenterOf(blockEntity.method_11016());
				class_1264.method_5449(blockEntity.method_10997(), vec.field_1352, vec.field_1351 + .5f, vec.field_1350, remainder);
			}
		}

		if (dirty)
			blockEntity.notifyUpdate();
	}

	public boolean isEmpty() {
		return heldItem == null && isOutputEmpty();
	}

	public boolean isOutputEmpty() {
		for (int i = 0; i < processingOutputBuffer.getSlotCount(); i++)
			if (!processingOutputBuffer.getStackInSlot(i)
				.method_7960())
				return false;
		return true;
	}

	private class_243 getWorldPositionOf(TransportedItemStack transported) {
		return VecHelper.getCenterOf(blockEntity.method_11016());
	}

	@Override
	public BehaviourType<?> getType() {
		return TYPE;
	}

	public boolean isItemValid(class_1799 stack) {
		return acceptedItems.test(stack);
	}

}
