package com.simibubi.create.content.logistics.depot;

import java.util.Iterator;

import com.google.common.collect.Iterators;
import com.simibubi.create.content.kinetics.belt.transport.TransportedItemStack;
import com.simibubi.create.foundation.item.ItemHelper;

import net.fabricmc.fabric.api.transfer.v1.item.ItemVariant;
import net.fabricmc.fabric.api.transfer.v1.storage.Storage;
import net.fabricmc.fabric.api.transfer.v1.storage.StorageView;
import net.fabricmc.fabric.api.transfer.v1.transaction.TransactionContext;
import net.fabricmc.fabric.api.transfer.v1.transaction.base.SnapshotParticipant;
import net.minecraft.class_1799;
import net.minecraft.class_3902;

public class DepotItemHandler extends SnapshotParticipant<class_3902> implements Storage<ItemVariant> {

	private static final int MAIN_SLOT = 0;
	private DepotBehaviour behaviour;

	public DepotItemHandler(DepotBehaviour behaviour) {
		this.behaviour = behaviour;
	}

	@Override
	public long insert(ItemVariant resource, long maxAmount, TransactionContext transaction) {
		if (!behaviour.getHeldItemStack().method_7960() && !behaviour.canMergeItems())
			return 0;
		if (!behaviour.isOutputEmpty() && !behaviour.canMergeItems())
			return 0;
		int toInsert = Math.min(ItemHelper.truncateLong(maxAmount), resource.getItem().method_7882());
		class_1799 stack = resource.toStack(toInsert);
		if (!behaviour.isItemValid(stack))
			return 0;
		class_1799 remainder = behaviour.insert(new TransportedItemStack(stack), transaction);
		return stack.method_7947() - remainder.method_7947();
	}

	@Override
	public long extract(ItemVariant resource, long maxAmount, TransactionContext transaction) {
		long extracted = behaviour.processingOutputBuffer.extract(resource, maxAmount, transaction);
		if (extracted == maxAmount)
			return extracted;
		extracted += extractFromMain(resource, maxAmount - extracted, transaction);
		return extracted;
	}

	public long extractFromMain(ItemVariant resource, long maxAmount, TransactionContext transaction) {
		TransportedItemStack held = behaviour.heldItem;
		if (held == null)
			return 0;
		class_1799 stack = held.stack;
		if (!resource.matches(stack))
			return 0;
		int toExtract = Math.min(ItemHelper.truncateLong(maxAmount), Math.min(stack.method_7947(), behaviour.maxStackSize.get()));
		stack = stack.method_7972();
		stack.method_7934(toExtract);
		if (stack.method_7960())
			stack = class_1799.field_8037;
		behaviour.snapshotParticipant.updateSnapshots(transaction);
		behaviour.heldItem.stack = stack;
		if (stack.method_7960())
			behaviour.heldItem = null;
		return toExtract;
	}

	@Override
	public Iterator<StorageView<ItemVariant>> iterator() {
		return Iterators.concat(Iterators.singletonIterator(new MainSlotView()), behaviour.processingOutputBuffer.iterator());
	}

	@Override
	protected class_3902 createSnapshot() {
		return class_3902.field_17274;
	}

	@Override
	protected void readSnapshot(class_3902 snapshot) {
	}

	@Override
	protected void onFinalCommit() {
		super.onFinalCommit();
		behaviour.blockEntity.notifyUpdate();
	}

	public class MainSlotView implements StorageView<ItemVariant> {

		@Override
		public long extract(ItemVariant resource, long maxAmount, TransactionContext transaction) {
			return extractFromMain(resource, maxAmount, transaction);
		}

		@Override
		public boolean isResourceBlank() {
			return getResource().isBlank();
		}

		@Override
		public ItemVariant getResource() {
			return ItemVariant.of(getStack());
		}

		@Override
		public long getAmount() {
			class_1799 stack = getStack();
			return stack.method_7960() ? 0 : stack.method_7947();
		}

		@Override
		public long getCapacity() {
			class_1799 stack = getStack();
			return stack.method_7960() ? behaviour.maxStackSize.get() : Math.min(stack.method_7914(), behaviour.maxStackSize.get());
		}

		public class_1799 getStack() {
			TransportedItemStack held = behaviour.heldItem;
			if (held == null || held.stack == null || held.stack.method_7960())
				return class_1799.field_8037;
			return held.stack;
		}
	}
}
