package com.simibubi.create.content.logistics.depot;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import javax.annotation.Nullable;

import com.jozufozu.flywheel.util.transform.TransformStack;
import com.simibubi.create.AllBlocks;
import com.simibubi.create.AllPackets;
import com.simibubi.create.content.kinetics.base.KineticBlockEntity;
import com.simibubi.create.content.kinetics.belt.behaviour.DirectBeltInputBehaviour;
import com.simibubi.create.content.kinetics.belt.transport.TransportedItemStack;
import com.simibubi.create.content.logistics.funnel.AbstractFunnelBlock;
import com.simibubi.create.content.logistics.funnel.FunnelBlock;
import com.simibubi.create.foundation.blockEntity.behaviour.BlockEntityBehaviour;
import com.simibubi.create.foundation.blockEntity.behaviour.ValueBoxTransform;
import com.simibubi.create.foundation.blockEntity.behaviour.scrollValue.ScrollValueBehaviour;
import com.simibubi.create.foundation.item.ItemHelper;
import com.simibubi.create.foundation.utility.AngleHelper;
import com.simibubi.create.foundation.utility.Iterate;
import com.simibubi.create.foundation.utility.Lang;
import com.simibubi.create.foundation.utility.LongAttached;
import com.simibubi.create.foundation.utility.NBTHelper;
import com.simibubi.create.foundation.utility.Pair;
import com.simibubi.create.foundation.utility.VecHelper;
import com.simibubi.create.foundation.utility.animation.LerpedFloat;
import com.simibubi.create.foundation.utility.animation.LerpedFloat.Chaser;
import com.simibubi.create.infrastructure.config.AllConfigs;

import io.github.fabricators_of_create.porting_lib.transfer.TransferUtil;
import io.github.fabricators_of_create.porting_lib.transfer.item.ItemStackHandler;
import io.github.fabricators_of_create.porting_lib.util.NBTSerializer;
import net.fabricmc.api.EnvType;
import net.fabricmc.api.Environment;
import net.fabricmc.fabric.api.transfer.v1.item.ItemVariant;
import net.fabricmc.fabric.api.transfer.v1.storage.Storage;
import net.fabricmc.fabric.api.transfer.v1.storage.StorageView;
import net.fabricmc.fabric.api.transfer.v1.storage.base.SidedStorageBlockEntity;
import net.fabricmc.fabric.api.transfer.v1.transaction.Transaction;
import net.minecraft.class_1297;
import net.minecraft.class_1299;
import net.minecraft.class_1304;
import net.minecraft.class_1542;
import net.minecraft.class_1657;
import net.minecraft.class_1770;
import net.minecraft.class_1799;
import net.minecraft.class_1937;
import net.minecraft.class_2338;
import net.minecraft.class_2350;
import net.minecraft.class_2350.class_2351;
import net.minecraft.class_238;
import net.minecraft.class_239.class_240;
import net.minecraft.class_2426;
import net.minecraft.class_243;
import net.minecraft.class_2487;
import net.minecraft.class_2512;
import net.minecraft.class_2520;
import net.minecraft.class_2591;
import net.minecraft.class_2680;
import net.minecraft.class_2945;
import net.minecraft.class_3417;
import net.minecraft.class_3419;
import net.minecraft.class_3532;
import net.minecraft.class_3619;
import net.minecraft.class_3959;
import net.minecraft.class_3959.class_242;
import net.minecraft.class_3959.class_3960;
import net.minecraft.class_3965;
import net.minecraft.class_4587;

public class EjectorBlockEntity extends KineticBlockEntity implements SidedStorageBlockEntity {

	List<LongAttached<class_1799>> launchedItems;
	ScrollValueBehaviour maxStackSize;
	DepotBehaviour depotBehaviour;
	EntityLauncher launcher;
	LerpedFloat lidProgress;
	boolean powered;
	boolean launch;
	State state;

	// item collision
	@Nullable
	Pair<class_243, class_2338> earlyTarget;
	float earlyTargetTime;
	// runtime stuff
	int scanCooldown;
	class_1799 trackedItem;

	public enum State {
		CHARGED, LAUNCHING, RETRACTING;
	}

	public EjectorBlockEntity(class_2591<?> typeIn, class_2338 pos, class_2680 state) {
		super(typeIn, pos, state);
		launcher = new EntityLauncher(1, 0);
		lidProgress = LerpedFloat.linear()
			.startWithValue(1);
		this.state = State.RETRACTING;
		launchedItems = new ArrayList<>();
		powered = false;
	}

	@Override
	public void addBehaviours(List<BlockEntityBehaviour> behaviours) {
		super.addBehaviours(behaviours);
		behaviours.add(depotBehaviour = new DepotBehaviour(this));

		maxStackSize =
			new ScrollValueBehaviour(Lang.translateDirect("weighted_ejector.stack_size"), this, new EjectorSlot())
				.between(0, 64)
				.withFormatter(i -> i == 0 ? "*" : String.valueOf(i));
		behaviours.add(maxStackSize);

		depotBehaviour.maxStackSize = () -> maxStackSize.getValue();
		depotBehaviour.canAcceptItems = () -> state == State.CHARGED;
		depotBehaviour.canFunnelsPullFrom = side -> side != getFacing();
		depotBehaviour.enableMerging();
		depotBehaviour.addSubBehaviours(behaviours);
	}

	@Override
	public void initialize() {
		super.initialize();
		updateSignal();
	}

	public void activate() {
		launch = true;
		nudgeEntities();
	}

	protected boolean cannotLaunch() {
		return state != State.CHARGED && !(field_11863.field_9236 && state == State.LAUNCHING);
	}

	public void activateDeferred() {
		if (cannotLaunch())
			return;
		class_2350 facing = getFacing();
		List<class_1297> entities =
			field_11863.method_18467(class_1297.class, new class_238(field_11867).method_1009(-1 / 16f, 0, -1 / 16f));

		// Launch Items
		boolean doLogic = !field_11863.field_9236 || isVirtual();
		if (doLogic)
			launchItems();

		// Launch Entities
		for (class_1297 entity : entities) {
			boolean isPlayerEntity = entity instanceof class_1657;
			if (!entity.method_5805())
				continue;
			if (entity instanceof class_1542)
				continue;
			if (entity.method_5657() == class_3619.field_15975)
				continue;

			entity.method_24830(false);

			if (isPlayerEntity != field_11863.field_9236)
				continue;

			entity.method_5814(field_11867.method_10263() + .5f, field_11867.method_10264() + 1, field_11867.method_10260() + .5f);
			launcher.applyMotion(entity, facing);

			if (!isPlayerEntity)
				continue;

			class_1657 playerEntity = (class_1657) entity;

			if (launcher.getHorizontalDistance() * launcher.getHorizontalDistance()
				+ launcher.getVerticalDistance() * launcher.getVerticalDistance() >= 25 * 25)
				AllPackets.getChannel()
					.sendToServer(new EjectorAwardPacket(field_11867));

			if (!(playerEntity.method_6118(class_1304.field_6174)
				.method_7909() instanceof class_1770))
				continue;

			playerEntity.method_36457(-35);
			playerEntity.method_36456(facing.method_10144());
			playerEntity.method_18799(playerEntity.method_18798()
				.method_1021(.75f));
			deployElytra(playerEntity);
			AllPackets.getChannel()
				.sendToServer(new EjectorElytraPacket(field_11867));
		}

		if (doLogic) {
			lidProgress.chase(1, .8f, Chaser.EXP);
			state = State.LAUNCHING;
			if (!field_11863.field_9236) {
				field_11863.method_8396(null, field_11867, class_3417.field_15080, class_3419.field_15245, .35f, 1f);
				field_11863.method_8396(null, field_11867, class_3417.field_14982, class_3419.field_15245, .1f, 1.4f);
			}
		}
	}

	public void deployElytra(class_1657 playerEntity) {
		EntityHack.setElytraFlying(playerEntity);
	}

	protected void launchItems() {
		class_1799 heldItemStack = depotBehaviour.getHeldItemStack();
		class_2350 funnelFacing = getFacing().method_10153();

		if (AbstractFunnelBlock.getFunnelFacing(field_11863.method_8320(field_11867.method_10084())) == funnelFacing) {
			DirectBeltInputBehaviour directOutput = getBehaviour(DirectBeltInputBehaviour.TYPE);

			if (depotBehaviour.heldItem != null) {
				class_1799 remainder = directOutput.tryExportingToBeltFunnel(heldItemStack, funnelFacing, false);
				if (remainder == null)
					;
				else if (remainder.method_7960())
					depotBehaviour.removeHeldItem();
				else if (remainder.method_7947() != heldItemStack.method_7947())
					depotBehaviour.heldItem.stack = remainder;
			}

			for (Iterator<TransportedItemStack> iterator = depotBehaviour.incoming.iterator(); iterator.hasNext();) {
				TransportedItemStack transportedItemStack = iterator.next();
				class_1799 stack = transportedItemStack.stack;
				class_1799 remainder = directOutput.tryExportingToBeltFunnel(stack, funnelFacing, false);
				if (remainder == null)
					;
				else if (remainder.method_7960())
					iterator.remove();
				else if (!class_1799.method_7984(remainder, stack))
					transportedItemStack.stack = remainder;
			}

			ItemStackHandler outputs = depotBehaviour.processingOutputBuffer;
			for (int i = 0; i < outputs.getSlotCount(); i++) {
				class_1799 stack = outputs.getStackInSlot(i);
				if (stack.method_7960())
					continue;
				class_1799 remainder = directOutput.tryExportingToBeltFunnel(stack, funnelFacing, false);
				if (remainder != null)
					outputs.setStackInSlot(i, remainder);
			}
			return;
		}

		if (!field_11863.field_9236)
			for (class_2350 d : Iterate.directions) {
				class_2680 blockState = field_11863.method_8320(field_11867.method_10093(d));
				if (!(blockState.method_26204() instanceof class_2426))
					continue;
				if (blockState.method_11654(class_2426.field_10927) != d.method_10153())
					continue;
				blockState.method_26191(d.method_10153(), blockState, field_11863, field_11867.method_10093(d), field_11867);
			}

		if (depotBehaviour.heldItem != null) {
			addToLaunchedItems(heldItemStack);
			depotBehaviour.removeHeldItem();
		}

		for (TransportedItemStack transportedItemStack : depotBehaviour.incoming)
			addToLaunchedItems(transportedItemStack.stack);
		depotBehaviour.incoming.clear();

		ItemStackHandler outputs = depotBehaviour.processingOutputBuffer;
		try (Transaction t = TransferUtil.getTransaction()) {
			for (StorageView<ItemVariant> view : outputs.nonEmptyViews()) {
				ItemVariant var = view.getResource();
				long extracted = view.extract(view.getResource(), 64, t);
				if (extracted != 0)
					addToLaunchedItems(var.toStack(ItemHelper.truncateLong(extracted)));
			}
			t.commit();
		}
	}

	protected boolean addToLaunchedItems(class_1799 stack) {
		if ((!field_11863.field_9236 || isVirtual()) && trackedItem == null && scanCooldown == 0) {
			scanCooldown = AllConfigs.server().kinetics.ejectorScanInterval.get();
			trackedItem = stack;
		}
		return launchedItems.add(LongAttached.withZero(stack));
	}

	protected class_2350 getFacing() {
		class_2680 blockState = method_11010();
		if (!AllBlocks.WEIGHTED_EJECTOR.has(blockState))
			return class_2350.field_11036;
		class_2350 facing = blockState.method_11654(EjectorBlock.HORIZONTAL_FACING);
		return facing;
	}

	@Override
	public void tick() {
		super.tick();

		boolean doLogic = !field_11863.field_9236 || isVirtual();
		State prevState = state;
		float totalTime = Math.max(3, (float) launcher.getTotalFlyingTicks());

		if (scanCooldown > 0)
			scanCooldown--;

		if (launch) {
			launch = false;
			activateDeferred();
		}

		for (Iterator<LongAttached<class_1799>> iterator = launchedItems.iterator(); iterator.hasNext();) {
			LongAttached<class_1799> LongAttached = iterator.next();
			boolean hit = false;
			if (LongAttached.getSecond() == trackedItem)
				hit = scanTrajectoryForObstacles(LongAttached.getFirst());
			float maxTime = earlyTarget != null ? Math.min(earlyTargetTime, totalTime) : totalTime;
			if (hit || LongAttached.exceeds((int) maxTime)) {
				placeItemAtTarget(doLogic, maxTime, LongAttached);
				iterator.remove();
			}
			LongAttached.increment();
		}

		if (state == State.LAUNCHING) {
			lidProgress.chase(1, .8f, Chaser.EXP);
			lidProgress.tickChaser();
			if (lidProgress.getValue() > 1 - 1 / 16f && doLogic) {
				state = State.RETRACTING;
				lidProgress.setValue(1);
			}
		}

		if (state == State.CHARGED) {
			lidProgress.setValue(0);
			lidProgress.updateChaseSpeed(0);
			if (doLogic)
				ejectIfTriggered();
		}

		if (state == State.RETRACTING) {
			if (lidProgress.getChaseTarget() == 1 && !lidProgress.settled()) {
				lidProgress.tickChaser();
			} else {
				lidProgress.updateChaseTarget(0);
				lidProgress.updateChaseSpeed(0);
				if (lidProgress.getValue() == 0 && doLogic) {
					state = State.CHARGED;
					lidProgress.setValue(0);
					sendData();
				}

				float value = class_3532.method_15363(lidProgress.getValue() - getWindUpSpeed(), 0, 1);
				lidProgress.setValue(value);

				int soundRate = (int) (1 / (getWindUpSpeed() * 5)) + 1;
				float volume = .125f;
				float pitch = 1.5f - lidProgress.getValue();
				if (((int) field_11863.method_8510()) % soundRate == 0 && doLogic)
					field_11863.method_8396(null, field_11867, class_3417.field_15105, class_3419.field_15245,
						volume, pitch);
			}
		}

		if (state != prevState)
			notifyUpdate();
	}

	private boolean scanTrajectoryForObstacles(long time) {
		if (time <= 2)
			return false;

		class_243 source = getLaunchedItemLocation(time);
		class_243 target = getLaunchedItemLocation(time + 1);

		class_3965 rayTraceBlocks = field_11863.method_17742(new class_3959(source, target, class_3960.field_17558, class_242.field_1348, null));
		boolean miss = rayTraceBlocks.method_17783() == class_240.field_1333;

		if (!miss && rayTraceBlocks.method_17783() == class_240.field_1332) {
			class_2680 blockState = field_11863.method_8320(rayTraceBlocks.method_17777());
			if (FunnelBlock.isFunnel(blockState) && blockState.method_28498(FunnelBlock.EXTRACTING)
				&& blockState.method_11654(FunnelBlock.EXTRACTING))
				miss = true;
		}

		if (miss) {
			if (earlyTarget != null && earlyTargetTime < time + 1) {
				earlyTarget = null;
				earlyTargetTime = 0;
			}
			return false;
		}

		class_243 vec = rayTraceBlocks.method_17784();
		earlyTarget = Pair.of(vec.method_1019(class_243.method_24954(rayTraceBlocks.method_17780()
			.method_10163())
			.method_1021(.25f)), rayTraceBlocks.method_17777());
		earlyTargetTime = (float) (time + (source.method_1022(vec) / source.method_1022(target)));
		sendData();
		return true;
	}

	protected void nudgeEntities() {
		for (class_1297 entity : field_11863.method_18467(class_1297.class,
			new class_238(field_11867).method_1009(-1 / 16f, 0, -1 / 16f))) {
			if (!entity.method_5805())
				continue;
			if (entity.method_5657() == class_3619.field_15975)
				continue;
			if (!(entity instanceof class_1657))
				entity.method_5814(entity.method_23317(), entity.method_23318() + .125f, entity.method_23321());
		}
	}

	protected void ejectIfTriggered() {
		if (powered)
			return;
		int presentStackSize = depotBehaviour.getPresentStackSize();
		if (presentStackSize == 0)
			return;
		if (presentStackSize < maxStackSize.getValue())
			return;
		if (depotBehaviour.heldItem != null && depotBehaviour.heldItem.beltPosition < .49f)
			return;

		class_2350 funnelFacing = getFacing().method_10153();
		class_1799 held = depotBehaviour.getHeldItemStack();
		if (AbstractFunnelBlock.getFunnelFacing(field_11863.method_8320(field_11867.method_10084())) == funnelFacing) {
			DirectBeltInputBehaviour directOutput = getBehaviour(DirectBeltInputBehaviour.TYPE);
			if (depotBehaviour.heldItem != null) {
				class_1799 tryFunnel = directOutput.tryExportingToBeltFunnel(held, funnelFacing, true);
				if (tryFunnel == null || !tryFunnel.method_7960())
					return;
			}
		}

		DirectBeltInputBehaviour targetOpenInv = getTargetOpenInv();

		// Do not eject if target cannot accept held item
		if (targetOpenInv != null && depotBehaviour.heldItem != null
			&& targetOpenInv.handleInsertion(held, class_2350.field_11036, true)
				.method_7947() == held.method_7947())
			return;

		activate();
		notifyUpdate();
	}

	protected void placeItemAtTarget(boolean doLogic, float maxTime, LongAttached<class_1799> LongAttached) {
		if (!doLogic)
			return;
		if (LongAttached.getSecond() == trackedItem)
			trackedItem = null;

		DirectBeltInputBehaviour targetOpenInv = getTargetOpenInv();
		if (targetOpenInv != null) {
			class_1799 remainder = targetOpenInv.handleInsertion(LongAttached.getValue(), class_2350.field_11036, false);
			LongAttached.setSecond(remainder);
		}

		if (LongAttached.getValue()
			.method_7960())
			return;

		class_243 ejectVec = earlyTarget != null ? earlyTarget.getFirst() : getLaunchedItemLocation(maxTime);
		class_243 ejectMotionVec = getLaunchedItemMotion(maxTime);
		class_1542 item = new class_1542(field_11863, ejectVec.field_1352, ejectVec.field_1351, ejectVec.field_1350, LongAttached.getValue());
		item.method_18799(ejectMotionVec);
		item.method_6988();
		field_11863.method_8649(item);
	}

	public DirectBeltInputBehaviour getTargetOpenInv() {
		class_2338 targetPos = earlyTarget != null ? earlyTarget.getSecond()
			: field_11867.method_10086(launcher.getVerticalDistance())
				.method_10079(getFacing(), Math.max(1, launcher.getHorizontalDistance()));
		return BlockEntityBehaviour.get(field_11863, targetPos, DirectBeltInputBehaviour.TYPE);
	}

	public class_243 getLaunchedItemLocation(float time) {
		return launcher.getGlobalPos(time, getFacing().method_10153(), field_11867);
	}

	public class_243 getLaunchedItemMotion(float time) {
		return launcher.getGlobalVelocity(time, getFacing().method_10153(), field_11867)
			.method_1021(.5f);
	}

	@Override
	public void destroy() {
		super.destroy();
		dropFlyingItems();
	}

	public void dropFlyingItems() {
		for (LongAttached<class_1799> LongAttached : launchedItems) {
			class_243 ejectVec = getLaunchedItemLocation(LongAttached.getFirst());
			class_243 ejectMotionVec = getLaunchedItemMotion(LongAttached.getFirst());
			class_1542 item = new class_1542(field_11863, 0, 0, 0, LongAttached.getValue());
			item.method_23327(ejectVec.field_1352, ejectVec.field_1351, ejectVec.field_1350);
			item.method_18799(ejectMotionVec);
			item.method_6988();
			field_11863.method_8649(item);
		}
		launchedItems.clear();
	}

	public float getWindUpSpeed() {
		int hd = launcher.getHorizontalDistance();
		int vd = launcher.getVerticalDistance();

		float speedFactor = Math.abs(getSpeed()) / 256f;
		float distanceFactor;
		if (hd == 0 && vd == 0)
			distanceFactor = 1;
		else
			distanceFactor = 1 * class_3532.method_15355(hd * hd + vd * vd);
		return speedFactor / distanceFactor;
	}

	@Override
	protected void write(class_2487 compound, boolean clientPacket) {
		super.write(compound, clientPacket);
		compound.method_10569("HorizontalDistance", launcher.getHorizontalDistance());
		compound.method_10569("VerticalDistance", launcher.getVerticalDistance());
		compound.method_10556("Powered", powered);
		NBTHelper.writeEnum(compound, "State", state);
		compound.method_10566("Lid", lidProgress.writeNBT());
		compound.method_10566("LaunchedItems",
			NBTHelper.writeCompoundList(launchedItems, ia -> ia.serializeNBT(NBTSerializer::serializeNBTCompound)));

		if (earlyTarget != null) {
			compound.method_10566("EarlyTarget", VecHelper.writeNBT(earlyTarget.getFirst()));
			compound.method_10566("EarlyTargetPos", class_2512.method_10692(earlyTarget.getSecond()));
			compound.method_10548("EarlyTargetTime", earlyTargetTime);
		}
	}

	@Override
	public void writeSafe(class_2487 compound) {
		super.writeSafe(compound);
		compound.method_10569("HorizontalDistance", launcher.getHorizontalDistance());
		compound.method_10569("VerticalDistance", launcher.getVerticalDistance());
	}

	@Override
	protected void read(class_2487 compound, boolean clientPacket) {
		super.read(compound, clientPacket);
		int horizontalDistance = compound.method_10550("HorizontalDistance");
		int verticalDistance = compound.method_10550("VerticalDistance");

		if (launcher.getHorizontalDistance() != horizontalDistance
			|| launcher.getVerticalDistance() != verticalDistance) {
			launcher.set(horizontalDistance, verticalDistance);
			launcher.clamp(AllConfigs.server().kinetics.maxEjectorDistance.get());
		}

		powered = compound.method_10577("Powered");
		state = NBTHelper.readEnum(compound, "State", State.class);
		lidProgress.readNBT(compound.method_10562("Lid"), false);
		launchedItems = NBTHelper.readCompoundList(compound.method_10554("LaunchedItems", class_2520.field_33260),
			nbt -> LongAttached.read(nbt, class_1799::method_7915));

		earlyTarget = null;
		earlyTargetTime = 0;
		if (compound.method_10545("EarlyTarget")) {
			earlyTarget = Pair.of(VecHelper.readNBT(compound.method_10554("EarlyTarget", class_2520.field_33256)),
				class_2512.method_10691(compound.method_10562("EarlyTargetPos")));
			earlyTargetTime = compound.method_10583("EarlyTargetTime");
		}

		if (compound.method_10545("ForceAngle"))
			lidProgress.startWithValue(compound.method_10583("ForceAngle"));
	}

	public void updateSignal() {
		boolean shoudPower = field_11863.method_49803(field_11867);
		if (shoudPower == powered)
			return;
		powered = shoudPower;
		sendData();
	}

	public void setTarget(int horizontalDistance, int verticalDistance) {
		launcher.set(Math.max(1, horizontalDistance), verticalDistance);
		sendData();
	}

	public class_2338 getTargetPosition() {
		class_2680 blockState = method_11010();
		if (!AllBlocks.WEIGHTED_EJECTOR.has(blockState))
			return field_11867;
		class_2350 facing = blockState.method_11654(EjectorBlock.HORIZONTAL_FACING);
		return field_11867.method_10079(facing, launcher.getHorizontalDistance())
			.method_10086(launcher.getVerticalDistance());
	}

	@Override
	public Storage<ItemVariant> getItemStorage(@Nullable class_2350 face) {
		return depotBehaviour.itemHandler;
	}

	public float getLidProgress(float pt) {
		return lidProgress.getValue(pt);
	}

	public State getState() {
		return state;
	}

	@Override
	@Environment(EnvType.CLIENT)
	public class_238 getRenderBoundingBox() {
		return INFINITE_EXTENT_AABB;
	}

	private static abstract class EntityHack extends class_1297 {

		public EntityHack(class_1299<?> p_i48580_1_, class_1937 p_i48580_2_) {
			super(p_i48580_1_, p_i48580_2_);
		}

		public static void setElytraFlying(class_1297 e) {
			class_2945 data = e.method_5841();
			data.method_12778(field_5990, (byte) (data.method_12789(field_5990) | 1 << 7));
		}

	}

	private class EjectorSlot extends ValueBoxTransform.Sided {

		@Override
		public class_243 getLocalOffset(class_2680 state) {
			if (direction != class_2350.field_11036)
				return super.getLocalOffset(state);
			return new class_243(.5, 10.5 / 16f, .5).method_1019(VecHelper.rotate(VecHelper.voxelSpace(0, 0, -5), angle(state), class_2351.field_11052));
		}

		@Override
		public void rotate(class_2680 state, class_4587 ms) {
			if (direction != class_2350.field_11036) {
				super.rotate(state, ms);
				return;
			}
			TransformStack.cast(ms)
				.rotateY(angle(state))
				.rotateX(90);
		}

		protected float angle(class_2680 state) {
			float horizontalAngle = AllBlocks.WEIGHTED_EJECTOR.has(state)
				? AngleHelper.horizontalAngle(state.method_11654(EjectorBlock.HORIZONTAL_FACING))
				: 0;
			return horizontalAngle;
		}

		@Override
		protected boolean isSideActive(class_2680 state, class_2350 direction) {
			return direction.method_10166() == state.method_11654(EjectorBlock.HORIZONTAL_FACING)
				.method_10166()
				|| direction == class_2350.field_11036 && EjectorBlockEntity.this.state != EjectorBlockEntity.State.CHARGED;
		}

		@Override
		protected class_243 getSouthLocation() {
			return direction == class_2350.field_11036 ? class_243.field_1353 : VecHelper.voxelSpace(8, 6, 15.5);
		}

	}

}
