package com.simibubi.create.content.logistics.filter;

import static com.simibubi.create.foundation.gui.AllGuiTextures.PLAYER_INVENTORY;

import java.util.Collections;
import java.util.List;
import net.minecraft.class_1661;
import net.minecraft.class_1799;
import net.minecraft.class_2561;
import net.minecraft.class_332;
import net.minecraft.class_5250;
import net.minecraft.class_768;
import com.google.common.collect.ImmutableList;
import com.simibubi.create.AllItems;
import com.simibubi.create.AllPackets;
import com.simibubi.create.content.logistics.filter.FilterScreenPacket.Option;
import com.simibubi.create.foundation.gui.AllGuiTextures;
import com.simibubi.create.foundation.gui.AllIcons;
import com.simibubi.create.foundation.gui.element.GuiGameElement;
import com.simibubi.create.foundation.gui.menu.AbstractSimiContainerScreen;
import com.simibubi.create.foundation.gui.widget.IconButton;
import com.simibubi.create.foundation.gui.widget.Indicator;
import com.simibubi.create.foundation.gui.widget.Indicator.State;
import com.simibubi.create.foundation.item.TooltipHelper;
import com.simibubi.create.foundation.item.TooltipHelper.Palette;

import io.github.fabricators_of_create.porting_lib.util.ItemStackUtil;
import io.github.fabricators_of_create.porting_lib.util.PlayerEntityHelper;

public abstract class AbstractFilterScreen<F extends AbstractFilterMenu> extends AbstractSimiContainerScreen<F> {

	protected AllGuiTextures background;
	private List<class_768> extraAreas = Collections.emptyList();

	private IconButton resetButton;
	private IconButton confirmButton;

	protected AbstractFilterScreen(F menu, class_1661 inv, class_2561 title, AllGuiTextures background) {
		super(menu, inv, title);
		this.background = background;
	}

	@Override
	protected void method_25426() {
		setWindowSize(Math.max(background.width, PLAYER_INVENTORY.width),
			background.height + 4 + PLAYER_INVENTORY.height);
		super.method_25426();

		int x = field_2776;
		int y = field_2800;

		resetButton = new IconButton(x + background.width - 62, y + background.height - 24, AllIcons.I_TRASH);
		resetButton.withCallback(() -> {
			field_2797.clearContents();
			contentsCleared();
			field_2797.sendClearPacket();
		});
		confirmButton = new IconButton(x + background.width - 33, y + background.height - 24, AllIcons.I_CONFIRM);
		confirmButton.withCallback(() -> {
			field_22787.field_1724.method_7346();
		});

		method_37063(resetButton);
		method_37063(confirmButton);

		extraAreas = ImmutableList.of(new class_768(x + background.width, y + background.height - 40, 80, 48));
	}

	@Override
	protected void method_2389(class_332 graphics, float partialTicks, int mouseX, int mouseY) {
		int invX = getLeftOfCentered(PLAYER_INVENTORY.width);
		int invY = field_2800 + background.height + 4;
		renderPlayerInventory(graphics, invX, invY);

		int x = field_2776;
		int y = field_2800;

		background.render(graphics, x, y);
		graphics.method_51439(field_22793, field_22785, x + (background.width - 8) / 2 - field_22793.method_27525(field_22785) / 2, y + 4,
			AllItems.FILTER.isIn(field_2797.contentHolder) ? 0x303030 : 0x592424, false);

		GuiGameElement.of(field_2797.contentHolder).<GuiGameElement
			.GuiRenderBuilder>at(x + background.width + 8, y + background.height - 52, -200)
			.scale(4)
			.render(graphics);
	}

	@Override
	protected void method_37432() {
		if (!class_1799.method_7973(field_2797.player.method_6047(), field_2797.contentHolder))
			PlayerEntityHelper.closeScreen(field_2797.player);

		super.method_37432();

		handleTooltips();
		handleIndicators();
	}

	protected void handleTooltips() {
		List<IconButton> tooltipButtons = getTooltipButtons();

		for (IconButton button : tooltipButtons) {
			if (!button.getToolTip()
				.isEmpty()) {
				button.setToolTip(button.getToolTip()
					.get(0));
				button.getToolTip()
					.add(TooltipHelper.holdShift(Palette.YELLOW, method_25442()));
			}
		}

		if (method_25442()) {
			List<class_5250> tooltipDescriptions = getTooltipDescriptions();
			for (int i = 0; i < tooltipButtons.size(); i++)
				fillToolTip(tooltipButtons.get(i), tooltipDescriptions.get(i));
		}
	}

	public void handleIndicators() {
		for (IconButton button : getTooltipButtons())
			button.field_22763 = isButtonEnabled(button);
		for (Indicator indicator : getIndicators())
			indicator.state = isIndicatorOn(indicator) ? State.ON : State.OFF;
	}

	protected abstract boolean isButtonEnabled(IconButton button);

	protected abstract boolean isIndicatorOn(Indicator indicator);

	protected List<IconButton> getTooltipButtons() {
		return Collections.emptyList();
	}

	protected List<class_5250> getTooltipDescriptions() {
		return Collections.emptyList();
	}

	protected List<Indicator> getIndicators() {
		return Collections.emptyList();
	}

	private void fillToolTip(IconButton button, class_2561 tooltip) {
		if (!button.method_25367())
			return;
		List<class_2561> tip = button.getToolTip();
		tip.addAll(TooltipHelper.cutTextComponent(tooltip, Palette.ALL_GRAY));
	}

	protected void contentsCleared() {}

	protected void sendOptionUpdate(Option option) {
		AllPackets.getChannel()
			.sendToServer(new FilterScreenPacket(option));
	}

	@Override
	public List<class_768> getExtraAreas() {
		return extraAreas;
	}

}
