package com.simibubi.create.content.logistics.filter;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.stream.Collectors;
import net.minecraft.class_124;
import net.minecraft.class_1661;
import net.minecraft.class_1799;
import net.minecraft.class_2487;
import net.minecraft.class_2561;
import net.minecraft.class_332;
import net.minecraft.class_4587;
import net.minecraft.class_5250;
import com.simibubi.create.AllPackets;
import com.simibubi.create.content.logistics.filter.AttributeFilterMenu.WhitelistMode;
import com.simibubi.create.content.logistics.filter.FilterScreenPacket.Option;
import com.simibubi.create.foundation.gui.AllGuiTextures;
import com.simibubi.create.foundation.gui.AllIcons;
import com.simibubi.create.foundation.gui.widget.IconButton;
import com.simibubi.create.foundation.gui.widget.Indicator;
import com.simibubi.create.foundation.gui.widget.Label;
import com.simibubi.create.foundation.gui.widget.SelectionScrollInput;
import com.simibubi.create.foundation.utility.Components;
import com.simibubi.create.foundation.utility.Lang;
import com.simibubi.create.foundation.utility.Pair;
import io.github.fabricators_of_create.porting_lib.util.ItemStackUtil;

public class AttributeFilterScreen extends AbstractFilterScreen<AttributeFilterMenu> {

	private static final String PREFIX = "gui.attribute_filter.";

	private class_2561 addDESC = Lang.translateDirect(PREFIX + "add_attribute");
	private class_2561 addInvertedDESC = Lang.translateDirect(PREFIX + "add_inverted_attribute");

	private class_2561 allowDisN = Lang.translateDirect(PREFIX + "allow_list_disjunctive");
	private class_2561 allowDisDESC = Lang.translateDirect(PREFIX + "allow_list_disjunctive.description");
	private class_2561 allowConN = Lang.translateDirect(PREFIX + "allow_list_conjunctive");
	private class_2561 allowConDESC = Lang.translateDirect(PREFIX + "allow_list_conjunctive.description");
	private class_2561 denyN = Lang.translateDirect(PREFIX + "deny_list");
	private class_2561 denyDESC = Lang.translateDirect(PREFIX + "deny_list.description");

	private class_2561 referenceH = Lang.translateDirect(PREFIX + "add_reference_item");
	private class_2561 noSelectedT = Lang.translateDirect(PREFIX + "no_selected_attributes");
	private class_2561 selectedT = Lang.translateDirect(PREFIX + "selected_attributes");

	private IconButton whitelistDis, whitelistCon, blacklist;
	private Indicator whitelistDisIndicator, whitelistConIndicator, blacklistIndicator;
	private IconButton add;
	private IconButton addInverted;

	private class_1799 lastItemScanned = class_1799.field_8037;
	private List<ItemAttribute> attributesOfItem = new ArrayList<>();
	private List<class_2561> selectedAttributes = new ArrayList<>();
	private SelectionScrollInput attributeSelector;
	private Label attributeSelectorLabel;

	public AttributeFilterScreen(AttributeFilterMenu menu, class_1661 inv, class_2561 title) {
		super(menu, inv, title, AllGuiTextures.ATTRIBUTE_FILTER);
	}

	@Override
	protected void method_25426() {
		setWindowOffset(-11, 7);
		super.method_25426();

		int x = field_2776;
		int y = field_2800;

		whitelistDis = new IconButton(x + 47, y + 61, AllIcons.I_WHITELIST_OR);
		whitelistDis.withCallback(() -> {
			field_2797.whitelistMode = WhitelistMode.WHITELIST_DISJ;
			sendOptionUpdate(Option.WHITELIST);
		});
		whitelistDis.setToolTip(allowDisN);
		whitelistCon = new IconButton(x + 65, y + 61, AllIcons.I_WHITELIST_AND);
		whitelistCon.withCallback(() -> {
			field_2797.whitelistMode = WhitelistMode.WHITELIST_CONJ;
			sendOptionUpdate(Option.WHITELIST2);
		});
		whitelistCon.setToolTip(allowConN);
		blacklist = new IconButton(x + 83, y + 61, AllIcons.I_WHITELIST_NOT);
		blacklist.withCallback(() -> {
			field_2797.whitelistMode = WhitelistMode.BLACKLIST;
			sendOptionUpdate(Option.BLACKLIST);
		});
		blacklist.setToolTip(denyN);

		whitelistDisIndicator = new Indicator(x + 47, y + 55, Components.immutableEmpty());
		whitelistConIndicator = new Indicator(x + 65, y + 55, Components.immutableEmpty());
		blacklistIndicator = new Indicator(x + 83, y + 55, Components.immutableEmpty());

		addRenderableWidgets(blacklist, whitelistCon, whitelistDis, blacklistIndicator, whitelistConIndicator,
			whitelistDisIndicator);

		method_37063(add = new IconButton(x + 182, y + 23, AllIcons.I_ADD));
		method_37063(addInverted = new IconButton(x + 200, y + 23, AllIcons.I_ADD_INVERTED_ATTRIBUTE));
		add.withCallback(() -> {
			handleAddedAttibute(false);
		});
		add.setToolTip(addDESC);
		addInverted.withCallback(() -> {
			handleAddedAttibute(true);
		});
		addInverted.setToolTip(addInvertedDESC);

		handleIndicators();

		attributeSelectorLabel = new Label(x + 43, y + 28, Components.immutableEmpty()).colored(0xF3EBDE)
			.withShadow();
		attributeSelector = new SelectionScrollInput(x + 39, y + 23, 137, 18);
		attributeSelector.forOptions(Arrays.asList(Components.immutableEmpty()));
		attributeSelector.removeCallback();
		referenceItemChanged(field_2797.ghostInventory.getStackInSlot(0));

		method_37063(attributeSelector);
		method_37063(attributeSelectorLabel);

		selectedAttributes.clear();
		selectedAttributes.add((field_2797.selectedAttributes.isEmpty() ? noSelectedT : selectedT).method_27662()
			.method_27692(class_124.field_1054));
		field_2797.selectedAttributes.forEach(at -> selectedAttributes.add(Components.literal("- ")
			.method_10852(at.getFirst()
				.format(at.getSecond()))
			.method_27692(class_124.field_1080)));
	}

	private void referenceItemChanged(class_1799 stack) {
		lastItemScanned = stack;

		if (stack.method_7960()) {
			attributeSelector.field_22763 = false;
			attributeSelector.field_22764 = false;
			attributeSelectorLabel.text = referenceH.method_27662()
				.method_27692(class_124.field_1056);
			add.field_22763 = false;
			addInverted.field_22763 = false;
			attributeSelector.calling(s -> {
			});
			return;
		}

		add.field_22763 = true;

		addInverted.field_22763 = true;
		attributeSelector.titled(stack.method_7964()
			.method_27662()
			.method_27693("..."));
		attributesOfItem.clear();
		for (ItemAttribute itemAttribute : ItemAttribute.types)
			attributesOfItem.addAll(itemAttribute.listAttributesOf(stack, field_22787.field_1687));
		List<class_2561> options = attributesOfItem.stream()
			.map(a -> a.format(false))
			.collect(Collectors.toList());
		attributeSelector.forOptions(options);
		attributeSelector.field_22763 = true;
		attributeSelector.field_22764 = true;
		attributeSelector.setState(0);
		attributeSelector.calling(i -> {
			attributeSelectorLabel.setTextAndTrim(options.get(i), true, 112);
			ItemAttribute selected = attributesOfItem.get(i);
			for (Pair<ItemAttribute, Boolean> existing : field_2797.selectedAttributes) {
				class_2487 testTag = new class_2487();
				class_2487 testTag2 = new class_2487();
				existing.getFirst()
					.serializeNBT(testTag);
				selected.serializeNBT(testTag2);
				if (testTag.equals(testTag2)) {
					add.field_22763 = false;
					addInverted.field_22763 = false;
					return;
				}
			}
			add.field_22763 = true;
			addInverted.field_22763 = true;
		});
		attributeSelector.onChanged();
	}

	@Override
	public void renderForeground(class_332 graphics, int mouseX, int mouseY, float partialTicks) {
		class_1799 stack = field_2797.ghostInventory.getStackInSlot(1);
		class_4587 matrixStack = graphics.method_51448();
		matrixStack.method_22903();
		matrixStack.method_46416(0, 0, 150);
		graphics.method_51432(field_22793, stack, field_2776 + 22, field_2800 + 59,
			String.valueOf(selectedAttributes.size() - 1));
		matrixStack.method_22909();

		super.renderForeground(graphics, mouseX, mouseY, partialTicks);
	}

	@Override
	protected void method_37432() {
		super.method_37432();
		class_1799 stackInSlot = field_2797.ghostInventory.getStackInSlot(0);
		if (!class_1799.method_7973(stackInSlot, lastItemScanned))
			referenceItemChanged(stackInSlot);
	}

	@Override
	protected void method_2380(class_332 graphics, int mouseX, int mouseY) {
		if (this.field_2797.method_34255().method_7960() && this.field_2787 != null && this.field_2787.method_7681()) {
			if (this.field_2787.field_7874 == 37) {
				graphics.method_51434(field_22793, selectedAttributes, mouseX, mouseY);
				return;
			}
			graphics.method_51446(field_22793, this.field_2787.method_7677(), mouseX, mouseY);
		}
		super.method_2380(graphics, mouseX, mouseY);
	}

	@Override
	protected List<IconButton> getTooltipButtons() {
		return Arrays.asList(blacklist, whitelistCon, whitelistDis);
	}

	@Override
	protected List<class_5250> getTooltipDescriptions() {
		return Arrays.asList(denyDESC.method_27662(), allowConDESC.method_27662(), allowDisDESC.method_27662());
	}

	@Override
	protected List<Indicator> getIndicators() {
		return Arrays.asList(blacklistIndicator, whitelistConIndicator, whitelistDisIndicator);
	}

	protected boolean handleAddedAttibute(boolean inverted) {
		int index = attributeSelector.getState();
		if (index >= attributesOfItem.size())
			return false;
		add.field_22763 = false;
		addInverted.field_22763 = false;
		class_2487 tag = new class_2487();
		ItemAttribute itemAttribute = attributesOfItem.get(index);
		itemAttribute.serializeNBT(tag);
		AllPackets.getChannel()
			.sendToServer(new FilterScreenPacket(inverted ? Option.ADD_INVERTED_TAG : Option.ADD_TAG, tag));
		field_2797.appendSelectedAttribute(itemAttribute, inverted);
		if (field_2797.selectedAttributes.size() == 1)
			selectedAttributes.set(0, selectedT.method_27662()
				.method_27692(class_124.field_1054));
		selectedAttributes.add(Components.literal("- ").method_10852(itemAttribute.format(inverted))
			.method_27692(class_124.field_1080));
		return true;
	}

	@Override
	protected void contentsCleared() {
		selectedAttributes.clear();
		selectedAttributes.add(noSelectedT.method_27662()
			.method_27692(class_124.field_1054));
		if (!lastItemScanned.method_7960()) {
			add.field_22763 = true;
			addInverted.field_22763 = true;
		}
	}

	@Override
	protected boolean isButtonEnabled(IconButton button) {
		if (button == blacklist)
			return field_2797.whitelistMode != WhitelistMode.BLACKLIST;
		if (button == whitelistCon)
			return field_2797.whitelistMode != WhitelistMode.WHITELIST_CONJ;
		if (button == whitelistDis)
			return field_2797.whitelistMode != WhitelistMode.WHITELIST_DISJ;
		return true;
	}

	@Override
	protected boolean isIndicatorOn(Indicator indicator) {
		if (indicator == blacklistIndicator)
			return field_2797.whitelistMode == WhitelistMode.BLACKLIST;
		if (indicator == whitelistConIndicator)
			return field_2797.whitelistMode == WhitelistMode.WHITELIST_CONJ;
		if (indicator == whitelistDisIndicator)
			return field_2797.whitelistMode == WhitelistMode.WHITELIST_DISJ;
		return false;
	}

}
