package com.simibubi.create.content.logistics.filter;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import javax.annotation.Nonnull;

import com.simibubi.create.AllItems;
import com.simibubi.create.AllKeys;
import com.simibubi.create.content.logistics.filter.AttributeFilterMenu.WhitelistMode;
import com.simibubi.create.foundation.item.ItemHelper;
import com.simibubi.create.foundation.utility.Components;
import com.simibubi.create.foundation.utility.Lang;

import io.github.fabricators_of_create.porting_lib.transfer.item.ItemHandlerHelper;
import io.github.fabricators_of_create.porting_lib.transfer.item.ItemStackHandler;
import io.github.fabricators_of_create.porting_lib.fluids.FluidStack;
import io.github.fabricators_of_create.porting_lib.util.NetworkHooks;
import net.fabricmc.api.EnvType;
import net.fabricmc.api.Environment;
import net.minecraft.class_124;
import net.minecraft.class_1268;
import net.minecraft.class_1269;
import net.minecraft.class_1271;
import net.minecraft.class_1657;
import net.minecraft.class_1661;
import net.minecraft.class_1703;
import net.minecraft.class_1792;
import net.minecraft.class_1799;
import net.minecraft.class_1836;
import net.minecraft.class_1838;
import net.minecraft.class_1937;
import net.minecraft.class_2487;
import net.minecraft.class_2499;
import net.minecraft.class_2520;
import net.minecraft.class_2561;
import net.minecraft.class_3222;
import net.minecraft.class_3908;

public class FilterItem extends class_1792 implements class_3908 {

	private FilterType type;

	private enum FilterType {
		REGULAR, ATTRIBUTE;
	}

	public static FilterItem regular(class_1793 properties) {
		return new FilterItem(FilterType.REGULAR, properties);
	}

	public static FilterItem attribute(class_1793 properties) {
		return new FilterItem(FilterType.ATTRIBUTE, properties);
	}

	private FilterItem(FilterType type, class_1793 properties) {
		super(properties);
		this.type = type;
	}

	@Nonnull
	@Override
	public class_1269 method_7884(class_1838 context) {
		if (context.method_8036() == null)
			return class_1269.field_5811;
		return method_7836(context.method_8045(), context.method_8036(), context.method_20287()).method_5467();
	}

	@Override
	@Environment(EnvType.CLIENT)
	public void method_7851(class_1799 stack, class_1937 worldIn, List<class_2561> tooltip, class_1836 flagIn) {
		if (!AllKeys.shiftDown()) {
			List<class_2561> makeSummary = makeSummary(stack);
			if (makeSummary.isEmpty())
				return;
			tooltip.add(Components.literal(" "));
			tooltip.addAll(makeSummary);
		}
	}

	private List<class_2561> makeSummary(class_1799 filter) {
		List<class_2561> list = new ArrayList<>();
		if (!filter.method_7985())
			return list;

		if (type == FilterType.REGULAR) {
			ItemStackHandler filterItems = getFilterItems(filter);
			boolean blacklist = filter.method_7948()
				.method_10577("Blacklist");

			list.add((blacklist ? Lang.translateDirect("gui.filter.deny_list")
				: Lang.translateDirect("gui.filter.allow_list")).method_27692(class_124.field_1065));
			int count = 0;
			for (int i = 0; i < filterItems.getSlotCount(); i++) {
				if (count > 3) {
					list.add(Components.literal("- ...")
						.method_27692(class_124.field_1063));
					break;
				}

				class_1799 filterStack = filterItems.getStackInSlot(i);
				if (filterStack.method_7960())
					continue;
				list.add(Components.literal("- ")
					.method_10852(filterStack.method_7964())
					.method_27692(class_124.field_1080));
				count++;
			}

			if (count == 0)
				return Collections.emptyList();
		}

		if (type == FilterType.ATTRIBUTE) {
			WhitelistMode whitelistMode = WhitelistMode.values()[filter.method_7948()
				.method_10550("WhitelistMode")];
			list.add((whitelistMode == WhitelistMode.WHITELIST_CONJ
				? Lang.translateDirect("gui.attribute_filter.allow_list_conjunctive")
				: whitelistMode == WhitelistMode.WHITELIST_DISJ
					? Lang.translateDirect("gui.attribute_filter.allow_list_disjunctive")
					: Lang.translateDirect("gui.attribute_filter.deny_list")).method_27692(class_124.field_1065));

			int count = 0;
			class_2499 attributes = filter.method_7948()
				.method_10554("MatchedAttributes", class_2520.field_33260);
			for (class_2520 inbt : attributes) {
				class_2487 compound = (class_2487) inbt;
				ItemAttribute attribute = ItemAttribute.fromNBT(compound);
				if (attribute == null)
					continue;
				boolean inverted = compound.method_10577("Inverted");
				if (count > 3) {
					list.add(Components.literal("- ...")
						.method_27692(class_124.field_1063));
					break;
				}
				list.add(Components.literal("- ")
					.method_10852(attribute.format(inverted)));
				count++;
			}

			if (count == 0)
				return Collections.emptyList();
		}

		return list;
	}

	@Override
	public class_1271<class_1799> method_7836(class_1937 world, class_1657 player, class_1268 hand) {
		class_1799 heldItem = player.method_5998(hand);

		if (!player.method_5715() && hand == class_1268.field_5808) {
			if (!world.field_9236 && player instanceof class_3222)
				NetworkHooks.openScreen((class_3222) player, this, buf -> {
					buf.method_10793(heldItem);
				});
			return class_1271.method_22427(heldItem);
		}
		return class_1271.method_22430(heldItem);
	}

	@Override
	public class_1703 createMenu(int id, class_1661 inv, class_1657 player) {
		class_1799 heldItem = player.method_6047();
		if (type == FilterType.REGULAR)
			return FilterMenu.create(id, inv, heldItem);
		if (type == FilterType.ATTRIBUTE)
			return AttributeFilterMenu.create(id, inv, heldItem);
		return null;
	}

	@Override
	public class_2561 method_5476() {
		return method_7848();
	}

	public static ItemStackHandler getFilterItems(class_1799 stack) {
		ItemStackHandler newInv = new ItemStackHandler(18);
		if (AllItems.FILTER.get() != stack.method_7909())
			throw new IllegalArgumentException("Cannot get filter items from non-filter: " + stack);
		if (!stack.method_7985())
			return newInv;
		class_2487 invNBT = stack.method_7911("Items");
		if (!invNBT.method_33133())
			newInv.deserializeNBT(invNBT);
		return newInv;
	}

	public static boolean testDirect(class_1799 filter, class_1799 stack, boolean matchNBT) {
		if (matchNBT) {
			return ItemHandlerHelper.canItemStacksStack(filter, stack);
		} else {
			return ItemHelper.sameItem(filter, stack);
		}
	}

}
