package com.simibubi.create.content.logistics.filter;

import java.util.ArrayList;
import java.util.List;
import net.minecraft.class_1799;
import net.minecraft.class_1937;
import net.minecraft.class_2487;
import net.minecraft.class_2499;
import net.minecraft.class_2520;
import com.simibubi.create.AllItems;
import com.simibubi.create.content.fluids.transfer.GenericItemEmptying;
import com.simibubi.create.foundation.utility.Pair;

import io.github.fabricators_of_create.porting_lib.transfer.item.ItemStackHandler;
import io.github.fabricators_of_create.porting_lib.fluids.FluidStack;

public class FilterItemStack {

	private class_1799 filterItemStack;
	private boolean fluidExtracted;
	private FluidStack filterFluidStack;

	public static FilterItemStack of(class_1799 filter) {
		if (filter.method_7985()) {
			if (AllItems.FILTER.isIn(filter))
				return new ListFilterItemStack(filter);
			if (AllItems.ATTRIBUTE_FILTER.isIn(filter))
				return new AttributeFilterItemStack(filter);
		}

		return new FilterItemStack(filter);
	}

	public static FilterItemStack of(class_2487 tag) {
		return of(class_1799.method_7915(tag));
	}

	public static FilterItemStack empty() {
		return of(class_1799.field_8037);
	}

	public boolean isEmpty() {
		return filterItemStack.method_7960();
	}

	public class_2487 serializeNBT() {
		class_2487 ret = new class_2487();
		filterItemStack.method_7953(ret);
		return ret;
	}

	public class_1799 item() {
		return filterItemStack;
	}

	public FluidStack fluid(class_1937 level) {
		resolveFluid(level);
		return filterFluidStack;
	}

	public boolean isFilterItem() {
		return filterItemStack.method_7909() instanceof FilterItem;
	}

	//

	public boolean test(class_1937 world, class_1799 stack) {
		return test(world, stack, false);
	}

	public boolean test(class_1937 world, FluidStack stack) {
		return test(world, stack, true);
	}

	public boolean test(class_1937 world, class_1799 stack, boolean matchNBT) {
		if (isEmpty())
			return true;
		return FilterItem.testDirect(filterItemStack, stack, matchNBT);
	}

	public boolean test(class_1937 world, FluidStack stack, boolean matchNBT) {
		if (isEmpty())
			return true;
		if (stack.isEmpty())
			return false;

		resolveFluid(world);

		if (filterFluidStack.isEmpty())
			return false;
		if (!matchNBT)
			return filterFluidStack.getFluid()
				.method_15780(stack.getFluid());
		return filterFluidStack.isFluidEqual(stack);
	}

	//

	private void resolveFluid(class_1937 world) {
		if (!fluidExtracted) {
			fluidExtracted = true;
			if (GenericItemEmptying.canItemBeEmptied(world, filterItemStack))
				filterFluidStack = GenericItemEmptying.emptyItem(world, filterItemStack, true)
				.getFirst();
		}
	}

	protected FilterItemStack(class_1799 filter) {
		filterItemStack = filter;
		filterFluidStack = FluidStack.EMPTY;
		fluidExtracted = false;
	}

	public static class ListFilterItemStack extends FilterItemStack {

		public List<FilterItemStack> containedItems;
		public boolean shouldRespectNBT;
		public boolean isBlacklist;

		protected ListFilterItemStack(class_1799 filter) {
			super(filter);
			boolean defaults = !filter.method_7985();

			containedItems = new ArrayList<>();
			ItemStackHandler items = FilterItem.getFilterItems(filter);
			for (int i = 0; i < items.getSlots().size(); i++) {
				class_1799 stackInSlot = items.getStackInSlot(i);
				if (!stackInSlot.method_7960())
					containedItems.add(FilterItemStack.of(stackInSlot));
			}

			shouldRespectNBT = !defaults ? false
				: filter.method_7969()
					.method_10577("RespectNBT");
			isBlacklist = defaults ? false
				: filter.method_7969()
					.method_10577("Blacklist");
		}

		@Override
		public boolean test(class_1937 world, class_1799 stack, boolean matchNBT) {
			if (containedItems.isEmpty())
				return super.test(world, stack, matchNBT);
			for (FilterItemStack filterItemStack : containedItems)
				if (filterItemStack.test(world, stack, shouldRespectNBT))
					return !isBlacklist;
			return isBlacklist;
		}

		@Override
		public boolean test(class_1937 world, FluidStack stack, boolean matchNBT) {
			for (FilterItemStack filterItemStack : containedItems)
				if (filterItemStack.test(world, stack, shouldRespectNBT))
					return !isBlacklist;
			return isBlacklist;
		}

	}

	public static class AttributeFilterItemStack extends FilterItemStack {

		public enum WhitelistMode {
			WHITELIST_DISJ, WHITELIST_CONJ, BLACKLIST;
		}

		public WhitelistMode whitelistMode;
		public List<Pair<ItemAttribute, Boolean>> attributeTests;

		protected AttributeFilterItemStack(class_1799 filter) {
			super(filter);
			boolean defaults = !filter.method_7985();

			attributeTests = new ArrayList<>();
			whitelistMode = WhitelistMode.values()[defaults ? 0
				: filter.method_7969()
					.method_10550("WhitelistMode")];

			class_2499 attributes = defaults ? new class_2499()
				: filter.method_7969()
					.method_10554("MatchedAttributes", class_2520.field_33260);
			for (class_2520 inbt : attributes) {
				class_2487 compound = (class_2487) inbt;
				ItemAttribute attribute = ItemAttribute.fromNBT(compound);
				if (attribute != null)
					attributeTests.add(Pair.of(attribute, compound.method_10577("Inverted")));
			}
		}

		@Override
		public boolean test(class_1937 world, FluidStack stack, boolean matchNBT) {
			return false;
		}

		@Override
		public boolean test(class_1937 world, class_1799 stack, boolean matchNBT) {
			if (attributeTests.isEmpty())
				return super.test(world, stack, matchNBT);
			for (Pair<ItemAttribute, Boolean> test : attributeTests) {
				ItemAttribute attribute = test.getFirst();
				boolean inverted = test.getSecond();
				boolean matches = attribute.appliesTo(stack, world) != inverted;

				if (matches) {
					switch (whitelistMode) {
					case BLACKLIST:
						return false;
					case WHITELIST_CONJ:
						continue;
					case WHITELIST_DISJ:
						return true;
					}
				} else {
					switch (whitelistMode) {
					case BLACKLIST:
						continue;
					case WHITELIST_CONJ:
						return false;
					case WHITELIST_DISJ:
						continue;
					}
				}
			}

			switch (whitelistMode) {
			case BLACKLIST:
				return true;
			case WHITELIST_CONJ:
				return true;
			case WHITELIST_DISJ:
				return false;
			}

			return false;
		}

	}

}
