package com.simibubi.create.content.logistics.filter;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import java.util.function.BiPredicate;
import java.util.function.Predicate;
import java.util.stream.Collectors;

import org.apache.commons.lang3.StringUtils;
import org.jetbrains.annotations.Nullable;

import com.simibubi.create.AllRecipeTypes;
import com.simibubi.create.content.kinetics.fan.processing.AllFanProcessingTypes;
import com.simibubi.create.content.logistics.filter.attribute.BookAuthorAttribute;
import com.simibubi.create.content.logistics.filter.attribute.BookCopyAttribute;
import com.simibubi.create.content.logistics.filter.attribute.ColorAttribute;
import com.simibubi.create.content.logistics.filter.attribute.EnchantAttribute;
import com.simibubi.create.content.logistics.filter.attribute.FluidContentsAttribute;
import com.simibubi.create.content.logistics.filter.attribute.ItemNameAttribute;
import com.simibubi.create.content.logistics.filter.attribute.ShulkerFillLevelAttribute;
import com.simibubi.create.content.logistics.filter.attribute.astralsorcery.AstralSorceryAmuletAttribute;
import com.simibubi.create.content.logistics.filter.attribute.astralsorcery.AstralSorceryAttunementAttribute;
import com.simibubi.create.content.logistics.filter.attribute.astralsorcery.AstralSorceryCrystalAttribute;
import com.simibubi.create.content.logistics.filter.attribute.astralsorcery.AstralSorceryPerkGemAttribute;
import com.simibubi.create.foundation.utility.Lang;

import io.github.fabricators_of_create.porting_lib.transfer.item.ItemStackHandlerContainer;
import net.fabricmc.api.EnvType;
import net.fabricmc.api.Environment;
import net.fabricmc.fabric.api.transfer.v1.context.ContainerItemContext;
import net.fabricmc.fabric.api.transfer.v1.fluid.FluidStorage;
import net.fabricmc.loader.api.FabricLoader;
import net.fabricmc.loader.api.ModContainer;
import net.minecraft.class_1263;
import net.minecraft.class_1304;
import net.minecraft.class_1309;
import net.minecraft.class_1747;
import net.minecraft.class_1792;
import net.minecraft.class_1799;
import net.minecraft.class_1860;
import net.minecraft.class_1890;
import net.minecraft.class_1937;
import net.minecraft.class_2487;
import net.minecraft.class_2609;
import net.minecraft.class_2960;
import net.minecraft.class_3489;
import net.minecraft.class_3956;
import net.minecraft.class_3962;
import net.minecraft.class_5250;
import net.minecraft.class_6862;
import net.minecraft.class_7923;
import net.minecraft.class_7924;

public interface ItemAttribute {

	static List<ItemAttribute> types = new ArrayList<>();

	static ItemAttribute standard = register(StandardTraits.DUMMY);
	static ItemAttribute inTag = register(new InTag(class_3489.field_15539));
	static ItemAttribute addedBy = register(new AddedBy("dummy"));
	static ItemAttribute hasEnchant = register(EnchantAttribute.EMPTY);
	static ItemAttribute shulkerFillLevel = register(ShulkerFillLevelAttribute.EMPTY);
	static ItemAttribute hasColor = register(ColorAttribute.EMPTY);
	static ItemAttribute hasFluid = register(FluidContentsAttribute.EMPTY);
	static ItemAttribute hasName = register(new ItemNameAttribute("dummy"));
	static ItemAttribute bookAuthor = register(new BookAuthorAttribute("dummy"));
	static ItemAttribute bookCopy = register(new BookCopyAttribute(-1));
	static ItemAttribute astralAmulet = register(new AstralSorceryAmuletAttribute("dummy", -1));
	static ItemAttribute astralAttunement = register(new AstralSorceryAttunementAttribute("dummy"));
	static ItemAttribute astralCrystal = register(new AstralSorceryCrystalAttribute("dummy"));
	static ItemAttribute astralPerkGem = register(new AstralSorceryPerkGemAttribute("dummy"));

	static ItemAttribute register(ItemAttribute attributeType) {
		types.add(attributeType);
		return attributeType;
	}

	@Nullable
	static ItemAttribute fromNBT(class_2487 nbt) {
		for (ItemAttribute itemAttribute : types)
			if (itemAttribute.canRead(nbt))
				return itemAttribute.readNBT(nbt.method_10562(itemAttribute.getNBTKey()));
		return null;
	}

	default boolean appliesTo(class_1799 stack, class_1937 world) {
		return appliesTo(stack);
	}

	boolean appliesTo(class_1799 stack);

	default List<ItemAttribute> listAttributesOf(class_1799 stack, class_1937 world) {
		return listAttributesOf(stack);
	}

	List<ItemAttribute> listAttributesOf(class_1799 stack);

	String getTranslationKey();

	void writeNBT(class_2487 nbt);

	ItemAttribute readNBT(class_2487 nbt);

	default void serializeNBT(class_2487 nbt) {
		class_2487 compound = new class_2487();
		writeNBT(compound);
		nbt.method_10566(getNBTKey(), compound);
	}

	default Object[] getTranslationParameters() {
		return new String[0];
	}

	default boolean canRead(class_2487 nbt) {
		return nbt.method_10545(getNBTKey());
	}

	default String getNBTKey() {
		return getTranslationKey();
	}

	@Environment(value = EnvType.CLIENT)
	default class_5250 format(boolean inverted) {
		return Lang.translateDirect("item_attributes." + getTranslationKey() + (inverted ? ".inverted" : ""),
				getTranslationParameters());
	}

	public static enum StandardTraits implements ItemAttribute {

		DUMMY(s -> false),
		PLACEABLE(s -> s.method_7909() instanceof class_1747),
		CONSUMABLE(class_1799::method_19267),
		FLUID_CONTAINER(s -> ContainerItemContext.withConstant(s).find(FluidStorage.ITEM) != null),
		ENCHANTED(class_1799::method_7942),
		MAX_ENCHANTED(StandardTraits::maxEnchanted),
		RENAMED(class_1799::method_7938),
		DAMAGED(class_1799::method_7986),
		BADLY_DAMAGED(s -> s.method_7986() && (float) s.method_7919() / s.method_7936() > 3 / 4f),
		NOT_STACKABLE(((Predicate<class_1799>) class_1799::method_7946).negate()),
		EQUIPABLE(s -> class_1309.method_32326(s)
				.method_5925() != class_1304.class_1305.field_6177),
		FURNACE_FUEL(class_2609::method_11195),
		WASHABLE(AllFanProcessingTypes.SPLASHING::canProcess),
		HAUNTABLE(AllFanProcessingTypes.HAUNTING::canProcess),
		CRUSHABLE((s, w) -> testRecipe(s, w, AllRecipeTypes.CRUSHING.getType())
				|| testRecipe(s, w, AllRecipeTypes.MILLING.getType())),
		SMELTABLE((s, w) -> testRecipe(s, w, class_3956.field_17546)),
		SMOKABLE((s, w) -> testRecipe(s, w, class_3956.field_17548)),
		BLASTABLE((s, w) -> testRecipe(s, w, class_3956.field_17547)),
		COMPOSTABLE(s -> class_3962.field_17566.containsKey(s.method_7909()));

		private static final class_1263 RECIPE_WRAPPER = new ItemStackHandlerContainer(1);
		private Predicate<class_1799> test;
		private BiPredicate<class_1799, class_1937> testWithWorld;

		private StandardTraits(Predicate<class_1799> test) {
			this.test = test;
		}

		private static boolean testRecipe(class_1799 s, class_1937 w, class_3956<? extends class_1860<class_1263>> type) {
			RECIPE_WRAPPER.method_5447(0, s.method_7972());
			return w.method_8433()
					.method_8132(type, RECIPE_WRAPPER, w)
					.isPresent();
		}

		private static boolean maxEnchanted(class_1799 s) {
			return class_1890.method_8222(s)
					.entrySet()
					.stream()
					.anyMatch(e -> e.getKey()
							.method_8183() <= e.getValue());
		}

		private StandardTraits(BiPredicate<class_1799, class_1937> test) {
			this.testWithWorld = test;
		}

		@Override
		public boolean appliesTo(class_1799 stack, class_1937 world) {
			if (testWithWorld != null)
				return testWithWorld.test(stack, world);
			return appliesTo(stack);
		}

		@Override
		public boolean appliesTo(class_1799 stack) {
			return test.test(stack);
		}

		@Override
		public List<ItemAttribute> listAttributesOf(class_1799 stack, class_1937 world) {
			List<ItemAttribute> attributes = new ArrayList<>();
			for (StandardTraits trait : values())
				if (trait.appliesTo(stack, world))
					attributes.add(trait);
			return attributes;
		}

		@Override
		public List<ItemAttribute> listAttributesOf(class_1799 stack) {
			return null;
		}

		@Override
		public String getTranslationKey() {
			return Lang.asId(name());
		}

		@Override
		public String getNBTKey() {
			return "standard_trait";
		}

		@Override
		public void writeNBT(class_2487 nbt) {
			nbt.method_10556(name(), true);
		}

		@Override
		public ItemAttribute readNBT(class_2487 nbt) {
			for (StandardTraits trait : values())
				if (nbt.method_10545(trait.name()))
					return trait;
			return null;
		}

	}

	public static class InTag implements ItemAttribute {

		public class_6862<class_1792> tag;

		public InTag(class_6862<class_1792> tag) {
			this.tag = tag;
		}

		@Override
		public boolean appliesTo(class_1799 stack) {
			return stack.method_31573(tag);
		}

		@Override
		public List<ItemAttribute> listAttributesOf(class_1799 stack) {
			return stack.method_40133()
					.map(InTag::new)
					.collect(Collectors.toList());
		}

		@Override
		public String getTranslationKey() {
			return "in_tag";
		}

		@Override
		public Object[] getTranslationParameters() {
			return new Object[]{"#" + tag.comp_327()};
		}

		@Override
		public void writeNBT(class_2487 nbt) {
			nbt.method_10582("space", tag.comp_327().method_12836());
			nbt.method_10582("path", tag.comp_327().method_12832());
		}

		@Override
		public ItemAttribute readNBT(class_2487 nbt) {
			return new InTag(class_6862.method_40092(class_7924.field_41197, new class_2960(nbt.method_10558("space"), nbt.method_10558("path"))));
		}

	}

	public static class AddedBy implements ItemAttribute {

		private String modId;

		public AddedBy(String modId) {
			this.modId = modId;
		}

		@Override
		public boolean appliesTo(class_1799 stack) {
			return modId.equals(class_7923.field_41178.method_10221(stack.method_7909()).method_12836());
		}

		@Override
		public List<ItemAttribute> listAttributesOf(class_1799 stack) {
			String id = class_7923.field_41178.method_10221(stack.method_7909()).method_12836();
			return id == null ? Collections.emptyList() : Arrays.asList(new AddedBy(id));
		}

		@Override
		public String getTranslationKey() {
			return "added_by";
		}

		@Override
		public Object[] getTranslationParameters() {
			ModContainer container = FabricLoader.getInstance().getModContainer(modId).orElse(null);
			String name = container == null ? name = StringUtils.capitalize(modId) : container.getMetadata().getName();
			return new Object[]{name};
		}

		@Override
		public void writeNBT(class_2487 nbt) {
			nbt.method_10582("id", modId);
		}

		@Override
		public ItemAttribute readNBT(class_2487 nbt) {
			return new AddedBy(nbt.method_10558("id"));
		}

	}

}
