package com.simibubi.create.content.logistics.filter.attribute;

import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.stream.Collectors;
import net.minecraft.class_1074;
import net.minecraft.class_1767;
import net.minecraft.class_1780;
import net.minecraft.class_1781;
import net.minecraft.class_1799;
import net.minecraft.class_2487;
import net.minecraft.class_2499;
import com.simibubi.create.content.logistics.filter.ItemAttribute;
import com.simibubi.create.foundation.utility.RegisteredObjects;

import io.github.fabricators_of_create.porting_lib.util.TagUtil;

public class ColorAttribute implements ItemAttribute {
	public static final ColorAttribute EMPTY = new ColorAttribute(class_1767.field_7945);

	public final class_1767 color;

	public ColorAttribute(class_1767 color) {
		this.color = color;
	}

	@Override
	public boolean appliesTo(class_1799 itemStack) {
		return findMatchingDyeColors(itemStack).stream().anyMatch(color::equals);
	}

	@Override
	public List<ItemAttribute> listAttributesOf(class_1799 itemStack) {
		return findMatchingDyeColors(itemStack).stream().map(ColorAttribute::new).collect(Collectors.toList());
	}

	private Collection<class_1767> findMatchingDyeColors(class_1799 stack) {
		class_2487 nbt = stack.method_7969();

		class_1767 color = TagUtil.getColorFromStack(stack);
		if (color != null)
			return Collections.singletonList(color);

		Set<class_1767> colors = new HashSet<>();
		if (stack.method_7909() instanceof class_1781 && nbt != null) {
			class_2499 listnbt = nbt.method_10562("Fireworks").method_10554("Explosions", 10);
			for (int i = 0; i < listnbt.size(); i++) {
				colors.addAll(getFireworkStarColors(listnbt.method_10602(i)));
			}
		}

		if (stack.method_7909() instanceof class_1780 && nbt != null) {
			colors.addAll(getFireworkStarColors(nbt.method_10562("Explosion")));
		}

		Arrays.stream(class_1767.values()).filter(c -> RegisteredObjects.getKeyOrThrow(stack.method_7909()).method_12832().startsWith(c.method_7792() + "_")).forEach(colors::add);

		return colors;
	}

	private Collection<class_1767> getFireworkStarColors(class_2487 compound) {
		Set<class_1767> colors = new HashSet<>();
		Arrays.stream(compound.method_10561("Colors")).mapToObj(class_1767::method_7786).forEach(colors::add);
		Arrays.stream(compound.method_10561("FadeColors")).mapToObj(class_1767::method_7786).forEach(colors::add);
		return colors;
	}

	@Override
	public String getTranslationKey() {
		return "color";
	}

	@Override
	public Object[] getTranslationParameters() {
		return new Object[] { class_1074.method_4662("color.minecraft." + color.method_7792()) };
	}

	@Override
	public void writeNBT(class_2487 nbt) {
		nbt.method_10569("id", color.method_7789());
	}

	@Override
	public ItemAttribute readNBT(class_2487 nbt) {
		return nbt.method_10545("id") ?
			new ColorAttribute(class_1767.method_7791(nbt.method_10550("id")))
			: EMPTY;
	}
}
