package com.simibubi.create.content.logistics.funnel;

import org.jetbrains.annotations.Nullable;

import com.simibubi.create.AllBlocks;
import com.simibubi.create.AllShapes;
import com.simibubi.create.content.kinetics.belt.BeltBlock;
import com.simibubi.create.content.kinetics.belt.BeltSlope;
import com.simibubi.create.content.kinetics.belt.behaviour.DirectBeltInputBehaviour;
import com.simibubi.create.content.schematics.requirement.ISpecialBlockItemRequirement;
import com.simibubi.create.content.schematics.requirement.ItemRequirement;
import com.simibubi.create.foundation.advancement.AllAdvancements;
import com.simibubi.create.foundation.block.ProperWaterloggedBlock;
import com.simibubi.create.foundation.blockEntity.behaviour.BlockEntityBehaviour;
import com.simibubi.create.foundation.utility.Lang;
import com.simibubi.create.foundation.utility.VoxelShaper;
import com.tterrag.registrate.util.entry.BlockEntry;

import net.fabricmc.fabric.api.block.BlockPickInteractionAware;
import net.minecraft.class_1269;
import net.minecraft.class_1542;
import net.minecraft.class_1657;
import net.minecraft.class_1750;
import net.minecraft.class_1799;
import net.minecraft.class_1838;
import net.minecraft.class_1922;
import net.minecraft.class_1936;
import net.minecraft.class_1937;
import net.minecraft.class_2248;
import net.minecraft.class_2338;
import net.minecraft.class_2350;
import net.minecraft.class_239;
import net.minecraft.class_2586;
import net.minecraft.class_265;
import net.minecraft.class_2680;
import net.minecraft.class_2689.class_2690;
import net.minecraft.class_2754;
import net.minecraft.class_3542;
import net.minecraft.class_3726;
import net.minecraft.class_3727;
import net.minecraft.class_4538;

public class BeltFunnelBlock extends AbstractHorizontalFunnelBlock implements ISpecialBlockItemRequirement, BlockPickInteractionAware {

	private BlockEntry<? extends FunnelBlock> parent;

	public static final class_2754<Shape> SHAPE = class_2754.method_11850("shape", Shape.class);

	public enum Shape implements class_3542 {
		RETRACTED(AllShapes.BELT_FUNNEL_RETRACTED),
		EXTENDED(AllShapes.BELT_FUNNEL_EXTENDED),
		PUSHING(AllShapes.BELT_FUNNEL_PERPENDICULAR),
		PULLING(AllShapes.BELT_FUNNEL_PERPENDICULAR);

		VoxelShaper shaper;

		private Shape(VoxelShaper shaper) {
			this.shaper = shaper;
		}

		@Override
		public String method_15434() {
			return Lang.asId(name());
		}
	}

	public BeltFunnelBlock(BlockEntry<? extends FunnelBlock> parent, class_2251 p_i48377_1_) {
		super(p_i48377_1_);
		this.parent = parent;
		method_9590(method_9564().method_11657(SHAPE, Shape.RETRACTED));
	}

	@Override
	protected void method_9515(class_2690<class_2248, class_2680> p_206840_1_) {
		super.method_9515(p_206840_1_.method_11667(SHAPE));
	}

	public boolean isOfSameType(FunnelBlock otherFunnel) {
		return parent.get() == otherFunnel;
	}
	
	@Override
	public void method_9536(class_2680 state, class_1937 world, class_2338 pos, class_2680 newState, boolean isMoving) {
		if (newState.method_26204() instanceof FunnelBlock fb && isOfSameType(fb))
			return;
		super.method_9536(state, world, pos, newState, isMoving);
	}
	
	@Override
	public class_265 method_9530(class_2680 state, class_1922 p_220053_2_, class_2338 p_220053_3_,
		class_3726 p_220053_4_) {
		return state.method_11654(SHAPE).shaper.get(state.method_11654(HORIZONTAL_FACING));
	}

	@Override
	public class_265 method_9549(class_2680 p_220071_1_, class_1922 p_220071_2_, class_2338 p_220071_3_,
		class_3726 p_220071_4_) {
		if (p_220071_4_ instanceof class_3727
			&& ((class_3727) p_220071_4_).method_32480() instanceof class_1542
			&& (p_220071_1_.method_11654(SHAPE) == Shape.PULLING || p_220071_1_.method_11654(SHAPE) == Shape.PUSHING))
			return AllShapes.FUNNEL_COLLISION.get(getFacing(p_220071_1_));
		return method_9530(p_220071_1_, p_220071_2_, p_220071_3_, p_220071_4_);
	}

	@Override
	public class_2680 method_9605(class_1750 ctx) {
		class_2680 stateForPlacement = super.method_9605(ctx);
		class_2338 pos = ctx.method_8037();
		class_1937 world = ctx.method_8045();
		class_2350 facing = ctx.method_8038()
			.method_10166()
			.method_10179() ? ctx.method_8038() : ctx.method_8042();

		class_2680 state = stateForPlacement.method_11657(HORIZONTAL_FACING, facing);
		boolean sneaking = ctx.method_8036() != null && ctx.method_8036()
			.method_5715();
		return state.method_11657(SHAPE, getShapeForPosition(world, pos, facing, !sneaking));
	}

	public static Shape getShapeForPosition(class_1922 world, class_2338 pos, class_2350 facing, boolean extracting) {
		class_2338 posBelow = pos.method_10074();
		class_2680 stateBelow = world.method_8320(posBelow);
		Shape perpendicularState = extracting ? Shape.PUSHING : Shape.PULLING;
		if (!AllBlocks.BELT.has(stateBelow))
			return perpendicularState;
		class_2350 movementFacing = stateBelow.method_11654(BeltBlock.HORIZONTAL_FACING);
		return movementFacing.method_10166() != facing.method_10166() ? perpendicularState : Shape.RETRACTED;
	}

	@Override
	public class_1799 getPickedStack(class_2680 state, class_1922 view, class_2338 pos, @Nullable class_1657 player, @Nullable class_239 result) {
		return parent.asStack();
	}

	@Override
	public class_2680 method_9559(class_2680 state, class_2350 direction, class_2680 neighbour, class_1936 world,
		class_2338 pos, class_2338 p_196271_6_) {
		updateWater(world, state, pos);
		if (!isOnValidBelt(state, world, pos)) {
			class_2680 parentState = ProperWaterloggedBlock.withWater(world, parent.getDefaultState(), pos);
			if (state.method_28500(POWERED)
				.orElse(false))
				parentState = parentState.method_11657(POWERED, true);
			if (state.method_11654(SHAPE) == Shape.PUSHING)
				parentState = parentState.method_11657(FunnelBlock.EXTRACTING, true);
			return parentState.method_11657(FunnelBlock.FACING, state.method_11654(HORIZONTAL_FACING));
		}
		Shape updatedShape =
			getShapeForPosition(world, pos, state.method_11654(HORIZONTAL_FACING), state.method_11654(SHAPE) == Shape.PUSHING);
		Shape currentShape = state.method_11654(SHAPE);
		if (updatedShape == currentShape)
			return state;

		// Don't revert wrenched states
		if (updatedShape == Shape.PUSHING && currentShape == Shape.PULLING)
			return state;
		if (updatedShape == Shape.RETRACTED && currentShape == Shape.EXTENDED)
			return state;

		return state.method_11657(SHAPE, updatedShape);
	}

	public static boolean isOnValidBelt(class_2680 state, class_4538 world, class_2338 pos) {
		class_2680 stateBelow = world.method_8320(pos.method_10074());
		if ((stateBelow.method_26204() instanceof BeltBlock))
			return BeltBlock.canTransportObjects(stateBelow);
		DirectBeltInputBehaviour directBeltInputBehaviour =
			BlockEntityBehaviour.get(world, pos.method_10074(), DirectBeltInputBehaviour.TYPE);
		if (directBeltInputBehaviour == null)
			return false;
		return directBeltInputBehaviour.canSupportBeltFunnels();
	}

	@Override
	public class_1269 onWrenched(class_2680 state, class_1838 context) {
		class_1937 world = context.method_8045();
		if (world.field_9236)
			return class_1269.field_5812;

		Shape shape = state.method_11654(SHAPE);
		Shape newShape = shape;
		if (shape == Shape.PULLING)
			newShape = Shape.PUSHING;
		else if (shape == Shape.PUSHING)
			newShape = Shape.PULLING;
		else if (shape == Shape.EXTENDED)
			newShape = Shape.RETRACTED;
		else if (shape == Shape.RETRACTED) {
			class_2680 belt = world.method_8320(context.method_8037()
				.method_10074());
			if (belt.method_26204() instanceof BeltBlock && belt.method_11654(BeltBlock.SLOPE) != BeltSlope.HORIZONTAL)
				newShape = Shape.RETRACTED;
			else
				newShape = Shape.EXTENDED;
		}

		if (newShape == shape)
			return class_1269.field_5812;

		world.method_8501(context.method_8037(), state.method_11657(SHAPE, newShape));

		if (newShape == Shape.EXTENDED) {
			class_2350 facing = state.method_11654(HORIZONTAL_FACING);
			class_2680 opposite = world.method_8320(context.method_8037()
				.method_10093(facing));
			if (opposite.method_26204() instanceof BeltFunnelBlock && opposite.method_11654(SHAPE) == Shape.EXTENDED
				&& opposite.method_11654(HORIZONTAL_FACING) == facing.method_10153())
				AllAdvancements.FUNNEL_KISS.awardTo(context.method_8036());
		}
		return class_1269.field_5812;
	}

	@Override
	public ItemRequirement getRequiredItems(class_2680 state, class_2586 be) {
		return ItemRequirement.of(parent.getDefaultState(), be);
	}

}
