package com.simibubi.create.content.logistics.funnel;

import java.lang.ref.WeakReference;
import java.util.List;

import org.apache.commons.lang3.mutable.MutableBoolean;

import com.jozufozu.flywheel.backend.instancing.InstancedRenderDispatcher;
import com.simibubi.create.AllBlocks;
import com.simibubi.create.AllPackets;
import com.simibubi.create.AllSoundEvents;
import com.simibubi.create.content.equipment.goggles.IHaveHoveringInformation;
import com.simibubi.create.content.kinetics.belt.BeltBlockEntity;
import com.simibubi.create.content.kinetics.belt.BeltHelper;
import com.simibubi.create.content.kinetics.belt.behaviour.DirectBeltInputBehaviour;
import com.simibubi.create.content.kinetics.belt.transport.TransportedItemStack;
import com.simibubi.create.content.logistics.funnel.BeltFunnelBlock.Shape;
import com.simibubi.create.foundation.advancement.AllAdvancements;
import com.simibubi.create.foundation.blockEntity.SmartBlockEntity;
import com.simibubi.create.foundation.blockEntity.behaviour.BlockEntityBehaviour;
import com.simibubi.create.foundation.blockEntity.behaviour.filtering.FilteringBehaviour;
import com.simibubi.create.foundation.blockEntity.behaviour.inventory.InvManipulationBehaviour;
import com.simibubi.create.foundation.blockEntity.behaviour.inventory.VersionedInventoryTrackerBehaviour;
import com.simibubi.create.foundation.item.ItemHelper.ExtractionCountMode;
import com.simibubi.create.foundation.utility.BlockFace;
import com.simibubi.create.foundation.utility.VecHelper;
import com.simibubi.create.foundation.utility.animation.LerpedFloat;
import com.simibubi.create.foundation.utility.animation.LerpedFloat.Chaser;
import com.simibubi.create.infrastructure.config.AllConfigs;

import io.github.fabricators_of_create.porting_lib.util.EnvExecutor;
import net.fabricmc.api.EnvType;
import net.minecraft.class_1542;
import net.minecraft.class_1799;
import net.minecraft.class_2338;
import net.minecraft.class_2350;
import net.minecraft.class_238;
import net.minecraft.class_243;
import net.minecraft.class_2487;
import net.minecraft.class_2591;
import net.minecraft.class_2680;
import net.minecraft.class_2741;

public class FunnelBlockEntity extends SmartBlockEntity implements IHaveHoveringInformation {

	private FilteringBehaviour filtering;
	private InvManipulationBehaviour invManipulation;
	private VersionedInventoryTrackerBehaviour invVersionTracker;
	private int extractionCooldown;

	private WeakReference<class_1542> lastObserved; // In-world Extractors only

	LerpedFloat flap;

	static enum Mode {
		INVALID, PAUSED, COLLECT, PUSHING_TO_BELT, TAKING_FROM_BELT, EXTRACT
	}

	public FunnelBlockEntity(class_2591<?> type, class_2338 pos, class_2680 state) {
		super(type, pos, state);
		extractionCooldown = 0;
		flap = createChasingFlap();
	}

	public Mode determineCurrentMode() {
		class_2680 state = method_11010();
		if (!FunnelBlock.isFunnel(state))
			return Mode.INVALID;
		if (state.method_28500(class_2741.field_12484)
			.orElse(false))
			return Mode.PAUSED;
		if (state.method_26204() instanceof BeltFunnelBlock) {
			Shape shape = state.method_11654(BeltFunnelBlock.SHAPE);
			if (shape == Shape.PULLING)
				return Mode.TAKING_FROM_BELT;
			if (shape == Shape.PUSHING)
				return Mode.PUSHING_TO_BELT;

			BeltBlockEntity belt = BeltHelper.getSegmentBE(field_11863, field_11867.method_10074());
			if (belt != null)
				return belt.getMovementFacing() == state.method_11654(BeltFunnelBlock.HORIZONTAL_FACING)
					? Mode.PUSHING_TO_BELT
					: Mode.TAKING_FROM_BELT;
			return Mode.INVALID;
		}
		if (state.method_26204() instanceof FunnelBlock)
			return state.method_11654(FunnelBlock.EXTRACTING) ? Mode.EXTRACT : Mode.COLLECT;

		return Mode.INVALID;
	}

	@Override
	public void tick() {
		super.tick();
		flap.tickChaser();
		Mode mode = determineCurrentMode();
		if (field_11863.field_9236)
			return;

		// Redstone resets the extraction cooldown
		if (mode == Mode.PAUSED)
			extractionCooldown = 0;
		if (mode == Mode.TAKING_FROM_BELT)
			return;

		if (extractionCooldown > 0) {
			extractionCooldown--;
			return;
		}

		if (mode == Mode.PUSHING_TO_BELT)
			activateExtractingBeltFunnel();
		if (mode == Mode.EXTRACT)
			activateExtractor();
	}

	private void activateExtractor() {
		if (invVersionTracker.stillWaiting(invManipulation))
			return;
		
		class_2680 blockState = method_11010();
		class_2350 facing = AbstractFunnelBlock.getFunnelFacing(blockState);

		if (facing == null)
			return;

		boolean trackingEntityPresent = true;
		class_238 area = getEntityOverflowScanningArea();

		// Check if last item is still blocking the extractor
		if (lastObserved == null) {
			trackingEntityPresent = false;
		} else {
			class_1542 lastEntity = lastObserved.get();
			if (lastEntity == null || !lastEntity.method_5805() || !lastEntity.method_5829()
				.method_994(area)) {
				trackingEntityPresent = false;
				lastObserved = null;
			}
		}

		if (trackingEntityPresent)
			return;

		// Find other entities blocking the extract (only if necessary)
		int amountToExtract = getAmountToExtract();
		ExtractionCountMode mode = getModeToExtract();
		class_1799 stack = invManipulation.simulate()
			.extract(mode, amountToExtract);
		if (stack.method_7960()) {
			invVersionTracker.awaitNewVersion(invManipulation);
			return;
		}
		for (class_1542 itemEntity : field_11863.method_18467(class_1542.class, area)) {
			lastObserved = new WeakReference<>(itemEntity);
			return;
		}

		// Extract
		stack = invManipulation.extract(mode, amountToExtract);
		if (stack.method_7960())
			return;

		flap(false);
		onTransfer(stack);

		class_243 outputPos = VecHelper.getCenterOf(field_11867);
		boolean vertical = facing.method_10166()
			.method_10178();
		boolean up = facing == class_2350.field_11036;

		outputPos = outputPos.method_1019(class_243.method_24954(facing.method_10163())
			.method_1021(vertical ? up ? .15f : .5f : .25f));
		if (!vertical)
			outputPos = outputPos.method_1023(0, .45f, 0);

		class_243 motion = class_243.field_1353;
		if (up)
			motion = new class_243(0, 4 / 16f, 0);

		class_1542 item = new class_1542(field_11863, outputPos.field_1352, outputPos.field_1351, outputPos.field_1350, stack.method_7972());
		item.method_6988();
		item.method_18799(motion);
		field_11863.method_8649(item);
		lastObserved = new WeakReference<>(item);

		startCooldown();
	}

	static final class_238 coreBB = new class_238(VecHelper.CENTER_OF_ORIGIN, VecHelper.CENTER_OF_ORIGIN).method_1014(.75f);

	private class_238 getEntityOverflowScanningArea() {
		class_2350 facing = AbstractFunnelBlock.getFunnelFacing(method_11010());
		class_238 bb = coreBB.method_996(field_11867);
		if (facing == null || facing == class_2350.field_11036)
			return bb;
		return bb.method_1012(0, -1, 0);
	}

	private void activateExtractingBeltFunnel() {
		if (invVersionTracker.stillWaiting(invManipulation))
			return;

		class_2680 blockState = method_11010();
		class_2350 facing = blockState.method_11654(BeltFunnelBlock.HORIZONTAL_FACING);
		DirectBeltInputBehaviour inputBehaviour =
			BlockEntityBehaviour.get(field_11863, field_11867.method_10074(), DirectBeltInputBehaviour.TYPE);

		if (inputBehaviour == null)
			return;
		if (!inputBehaviour.canInsertFromSide(facing))
			return;
		if (inputBehaviour.isOccupied(facing))
			return;

		int amountToExtract = getAmountToExtract();
		ExtractionCountMode mode = getModeToExtract();
		MutableBoolean deniedByInsertion = new MutableBoolean(false);
		class_1799 stack = invManipulation.extract(mode, amountToExtract, s -> {
			class_1799 handleInsertion = inputBehaviour.handleInsertion(s, facing, true);
			if (handleInsertion.method_7960())
				return true;
			deniedByInsertion.setTrue();
			return false;
		});
		if (stack.method_7960()) {
			if (deniedByInsertion.isFalse())
				invVersionTracker.awaitNewVersion(invManipulation.getInventory());
			return;
		}
		flap(false);
		onTransfer(stack);
		inputBehaviour.handleInsertion(stack, facing, false);
		startCooldown();
	}

	public int getAmountToExtract() {
		if (!supportsAmountOnFilter())
			return 64;
		int amountToExtract = invManipulation.getAmountFromFilter();
		if (!filtering.isActive())
			amountToExtract = 1;
		return amountToExtract;
	}

	public ExtractionCountMode getModeToExtract() {
		if (!supportsAmountOnFilter() || !filtering.isActive())
			return ExtractionCountMode.UPTO;
		return invManipulation.getModeFromFilter();
	}

	private int startCooldown() {
		return extractionCooldown = AllConfigs.server().logistics.defaultExtractionTimer.get();
	}

	@Override
	public void addBehaviours(List<BlockEntityBehaviour> behaviours) {
		invManipulation =
			new InvManipulationBehaviour(this, (w, p, s) -> new BlockFace(p, AbstractFunnelBlock.getFunnelFacing(s)
				.method_10153()));
		behaviours.add(invManipulation);
		
		behaviours.add(invVersionTracker = new VersionedInventoryTrackerBehaviour(this));

		filtering = new FilteringBehaviour(this, new FunnelFilterSlotPositioning());
		filtering.showCountWhen(this::supportsAmountOnFilter);
		filtering.onlyActiveWhen(this::supportsFiltering);
		filtering.withCallback($ -> invVersionTracker.reset());
		behaviours.add(filtering);
		
		behaviours.add(new DirectBeltInputBehaviour(this).onlyInsertWhen(this::supportsDirectBeltInput)
			.setInsertionHandler(this::handleDirectBeltInput));
		registerAwardables(behaviours, AllAdvancements.FUNNEL);
	}

	private boolean supportsAmountOnFilter() {
		class_2680 blockState = method_11010();
		boolean beltFunnelsupportsAmount = false;
		if (blockState.method_26204() instanceof BeltFunnelBlock) {
			Shape shape = blockState.method_11654(BeltFunnelBlock.SHAPE);
			if (shape == Shape.PUSHING)
				beltFunnelsupportsAmount = true;
			else
				beltFunnelsupportsAmount = BeltHelper.getSegmentBE(field_11863, field_11867.method_10074()) != null;
		}
		boolean extractor = blockState.method_26204() instanceof FunnelBlock && blockState.method_11654(FunnelBlock.EXTRACTING);
		return beltFunnelsupportsAmount || extractor;
	}

	private boolean supportsDirectBeltInput(class_2350 side) {
		class_2680 blockState = method_11010();
		if (blockState == null)
			return false;
		if (!(blockState.method_26204() instanceof FunnelBlock))
			return false;
		if (blockState.method_11654(FunnelBlock.EXTRACTING))
			return false;
		return FunnelBlock.getFunnelFacing(blockState) == class_2350.field_11036;
	}

	private boolean supportsFiltering() {
		class_2680 blockState = method_11010();
		return AllBlocks.BRASS_BELT_FUNNEL.has(blockState) || AllBlocks.BRASS_FUNNEL.has(blockState);
	}

	private class_1799 handleDirectBeltInput(TransportedItemStack stack, class_2350 side, boolean simulate) {
		class_1799 inserted = stack.stack;
		if (!filtering.test(inserted))
			return inserted;
		if (determineCurrentMode() == Mode.PAUSED)
			return inserted;
		if (simulate)
			invManipulation.simulate();
		if (!simulate)
			onTransfer(inserted);
		return invManipulation.insert(inserted);
	}

	public void flap(boolean inward) {
		if (!field_11863.field_9236) {
			AllPackets.getChannel()
				.sendToClientsTracking(new FunnelFlapPacket(this, inward), this);
		} else {
			flap.setValue(inward ? 1 : -1);
			AllSoundEvents.FUNNEL_FLAP.playAt(field_11863, field_11867, 1, 1, true);
		}
	}

	public boolean hasFlap() {
		class_2680 blockState = method_11010();
		if (!AbstractFunnelBlock.getFunnelFacing(blockState)
			.method_10166()
			.method_10179())
			return false;
		return true;
	}

	public float getFlapOffset() {
		class_2680 blockState = method_11010();
		if (!(blockState.method_26204() instanceof BeltFunnelBlock))
			return -1 / 16f;
		switch (blockState.method_11654(BeltFunnelBlock.SHAPE)) {
		default:
		case RETRACTED:
			return 0;
		case EXTENDED:
			return 8 / 16f;
		case PULLING:
		case PUSHING:
			return -2 / 16f;
		}
	}

	@Override
	protected void write(class_2487 compound, boolean clientPacket) {
		super.write(compound, clientPacket);
		compound.method_10569("TransferCooldown", extractionCooldown);
	}

	@Override
	protected void read(class_2487 compound, boolean clientPacket) {
		super.read(compound, clientPacket);
		extractionCooldown = compound.method_10550("TransferCooldown");

		if (clientPacket)
			EnvExecutor.runWhenOn(EnvType.CLIENT, () -> () -> InstancedRenderDispatcher.enqueueUpdate(this));
	}

	public void onTransfer(class_1799 stack) {
		AllBlocks.SMART_OBSERVER.get()
			.onFunnelTransfer(field_11863, field_11867, stack);
		award(AllAdvancements.FUNNEL);
	}

	private LerpedFloat createChasingFlap() {
		return LerpedFloat.linear()
			.startWithValue(.25f)
			.chase(0, .05f, Chaser.EXP);
	}

}
