package com.simibubi.create.content.logistics.tunnel;

import com.simibubi.create.AllBlockEntityTypes;
import com.simibubi.create.AllBlocks;
import com.simibubi.create.content.equipment.wrench.IWrenchable;
import com.simibubi.create.content.kinetics.belt.BeltBlock;
import com.simibubi.create.content.kinetics.belt.BeltSlope;
import com.simibubi.create.content.kinetics.belt.behaviour.DirectBeltInputBehaviour;
import com.simibubi.create.content.logistics.funnel.BeltFunnelBlock;
import com.simibubi.create.foundation.block.IBE;
import com.simibubi.create.foundation.blockEntity.behaviour.BlockEntityBehaviour;
import com.simibubi.create.foundation.utility.Lang;
import com.simibubi.create.foundation.utility.worldWrappers.WrappedWorld;
import net.minecraft.class_1269;
import net.minecraft.class_1750;
import net.minecraft.class_1838;
import net.minecraft.class_1922;
import net.minecraft.class_1936;
import net.minecraft.class_1937;
import net.minecraft.class_2248;
import net.minecraft.class_2338;
import net.minecraft.class_2350;
import net.minecraft.class_2350.class_2351;
import net.minecraft.class_2350.class_2352;
import net.minecraft.class_2470;
import net.minecraft.class_2586;
import net.minecraft.class_2591;
import net.minecraft.class_265;
import net.minecraft.class_2680;
import net.minecraft.class_2689.class_2690;
import net.minecraft.class_2741;
import net.minecraft.class_2754;
import net.minecraft.class_2769;
import net.minecraft.class_3542;
import net.minecraft.class_3726;
import net.minecraft.class_4538;

public class BeltTunnelBlock extends class_2248 implements IBE<BeltTunnelBlockEntity>, IWrenchable {

	public static final class_2769<Shape> SHAPE = class_2754.method_11850("shape", Shape.class);
	public static final class_2769<class_2351> HORIZONTAL_AXIS = class_2741.field_12529;

	public BeltTunnelBlock(class_2251 properties) {
		super(properties);
		method_9590(method_9564().method_11657(SHAPE, Shape.STRAIGHT));
	}

	public enum Shape implements class_3542 {
		STRAIGHT, WINDOW, CLOSED, T_LEFT, T_RIGHT, CROSS;

		@Override
		public String method_15434() {
			return Lang.asId(name());
		}
	}

	@Override
	public class_265 method_9530(class_2680 state, class_1922 worldIn, class_2338 pos, class_3726 context) {
		return BeltTunnelShapes.getShape(state);
	}

	@Override
	public boolean method_9558(class_2680 state, class_4538 worldIn, class_2338 pos) {
		class_2680 blockState = worldIn.method_8320(pos.method_10074());
		if (!isValidPositionForPlacement(state, worldIn, pos))
			return false;
		if (!blockState.method_11654(BeltBlock.CASING))
			return false;
		return true;
	}

	public boolean isValidPositionForPlacement(class_2680 state, class_4538 worldIn, class_2338 pos) {
		class_2680 blockState = worldIn.method_8320(pos.method_10074());
		if (!AllBlocks.BELT.has(blockState))
			return false;
		if (blockState.method_11654(BeltBlock.SLOPE) != BeltSlope.HORIZONTAL)
			return false;
		return true;
	}

	public static boolean hasWindow(class_2680 state) {
		return state.method_11654(SHAPE) == Shape.WINDOW || state.method_11654(SHAPE) == Shape.CLOSED;
	}

	public static boolean isStraight(class_2680 state) {
		return hasWindow(state) || state.method_11654(SHAPE) == Shape.STRAIGHT;
	}

	public static boolean isJunction(class_2680 state) {
		Shape shape = state.method_11654(SHAPE);
		return shape == Shape.CROSS || shape == Shape.T_LEFT || shape == Shape.T_RIGHT;
	}

	@Override
	public class_2680 method_9605(class_1750 context) {
		return getTunnelState(context.method_8045(), context.method_8037());
	}

	@Override
	public void method_9615(class_2680 state, class_1937 world, class_2338 pos, class_2680 p_220082_4_, boolean p_220082_5_) {
		if (!(world instanceof WrappedWorld) && !world.method_8608())
			withBlockEntityDo(world, pos, BeltTunnelBlockEntity::updateTunnelConnections);
	}

	@Override
	public class_2680 method_9559(class_2680 state, class_2350 facing, class_2680 facingState, class_1936 worldIn,
		class_2338 currentPos, class_2338 facingPos) {
		if (facing.method_10166()
			.method_10178())
			return state;
		if (!(worldIn instanceof WrappedWorld) && !worldIn.method_8608())
			withBlockEntityDo(worldIn, currentPos, BeltTunnelBlockEntity::updateTunnelConnections);
		class_2680 tunnelState = getTunnelState(worldIn, currentPos);
		if (tunnelState.method_11654(HORIZONTAL_AXIS) == state.method_11654(HORIZONTAL_AXIS)) {
			if (hasWindow(tunnelState) == hasWindow(state))
				return state;
		}

		return tunnelState;
	}

	public void updateTunnel(class_1936 world, class_2338 pos) {
		class_2680 tunnel = world.method_8320(pos);
		class_2680 newTunnel = getTunnelState(world, pos);
		if (tunnel != newTunnel && !world.method_8608()) {
			world.method_8652(pos, newTunnel, 3);
			class_2586 be = world.method_8321(pos);
			if (be != null && (be instanceof BeltTunnelBlockEntity))
				((BeltTunnelBlockEntity) be).updateTunnelConnections();
		}
	}

	private class_2680 getTunnelState(class_1922 reader, class_2338 pos) {
		class_2680 state = method_9564();
		class_2680 belt = reader.method_8320(pos.method_10074());
		if (AllBlocks.BELT.has(belt))
			state = state.method_11657(HORIZONTAL_AXIS, belt.method_11654(BeltBlock.HORIZONTAL_FACING)
				.method_10166());
		class_2351 axis = state.method_11654(HORIZONTAL_AXIS);

		// T and Cross
		class_2350 left = class_2350.method_10156(class_2352.field_11056, axis)
			.method_10170();
		boolean onLeft = hasValidOutput(reader, pos.method_10074(), left);
		boolean onRight = hasValidOutput(reader, pos.method_10074(), left.method_10153());

		if (onLeft && onRight)
			state = state.method_11657(SHAPE, Shape.CROSS);
		else if (onLeft)
			state = state.method_11657(SHAPE, Shape.T_LEFT);
		else if (onRight)
			state = state.method_11657(SHAPE, Shape.T_RIGHT);

		if (state.method_11654(SHAPE) == Shape.STRAIGHT) {
			boolean canHaveWindow = canHaveWindow(reader, pos, axis);
			if (canHaveWindow)
				state = state.method_11657(SHAPE, Shape.WINDOW);
		}

		return state;
	}

	protected boolean canHaveWindow(class_1922 reader, class_2338 pos, class_2351 axis) {
		class_2350 fw = class_2350.method_10156(class_2352.field_11056, axis);
		class_2680 blockState1 = reader.method_8320(pos.method_10093(fw));
		class_2680 blockState2 = reader.method_8320(pos.method_10093(fw.method_10153()));

		boolean funnel1 = blockState1.method_26204() instanceof BeltFunnelBlock
			&& blockState1.method_11654(BeltFunnelBlock.SHAPE) == BeltFunnelBlock.Shape.EXTENDED
			&& blockState1.method_11654(BeltFunnelBlock.HORIZONTAL_FACING) == fw.method_10153();
		boolean funnel2 = blockState2.method_26204() instanceof BeltFunnelBlock
			&& blockState2.method_11654(BeltFunnelBlock.SHAPE) == BeltFunnelBlock.Shape.EXTENDED
			&& blockState2.method_11654(BeltFunnelBlock.HORIZONTAL_FACING) == fw;

		boolean valid1 = blockState1.method_26204() instanceof BeltTunnelBlock || funnel1;
		boolean valid2 = blockState2.method_26204() instanceof BeltTunnelBlock || funnel2;
		boolean canHaveWindow = valid1 && valid2 && !(funnel1 && funnel2);
		return canHaveWindow;
	}

	private boolean hasValidOutput(class_1922 world, class_2338 pos, class_2350 side) {
		class_2680 blockState = world.method_8320(pos.method_10093(side));
		if (AllBlocks.BELT.has(blockState))
			return blockState.method_11654(BeltBlock.HORIZONTAL_FACING)
				.method_10166() == side.method_10166();
		DirectBeltInputBehaviour behaviour =
			BlockEntityBehaviour.get(world, pos.method_10093(side), DirectBeltInputBehaviour.TYPE);
		return behaviour != null && behaviour.canInsertFromSide(side);
	}

	@Override
	public class_1269 onWrenched(class_2680 state, class_1838 context) {
		if (!hasWindow(state))
			return class_1269.field_5811;

		// Toggle windows
		Shape shape = state.method_11654(SHAPE);
		shape = shape == Shape.CLOSED ? Shape.WINDOW : Shape.CLOSED;
		class_1937 world = context.method_8045();
		if (!world.field_9236)
			world.method_8652(context.method_8037(), state.method_11657(SHAPE, shape), 2);
		return class_1269.field_5812;
	}

	@Override
	public class_2680 method_9598(class_2680 state, class_2470 rotation) {
		class_2350 fromAxis = class_2350.method_10156(class_2352.field_11056, state.method_11654(HORIZONTAL_AXIS));
		class_2350 rotated = rotation.method_10503(fromAxis);

		return state.method_11657(HORIZONTAL_AXIS, rotated.method_10166());
	}

	@Override
	public void method_9612(class_2680 state, class_1937 worldIn, class_2338 pos, class_2248 blockIn, class_2338 fromPos,
		boolean isMoving) {
		if (worldIn.field_9236)
			return;

		if (fromPos.equals(pos.method_10074())) {
			if (!method_9558(state, worldIn, pos)) {
				worldIn.method_22352(pos, true);
				return;
			}
		}
	}

	@Override
	protected void method_9515(class_2690<class_2248, class_2680> builder) {
		builder.method_11667(HORIZONTAL_AXIS, SHAPE);
		super.method_9515(builder);
	}

	@Override
	public Class<BeltTunnelBlockEntity> getBlockEntityClass() {
		return BeltTunnelBlockEntity.class;
	}

	@Override
	public class_2591<? extends BeltTunnelBlockEntity> getBlockEntityType() {
		return AllBlockEntityTypes.ANDESITE_TUNNEL.get();
	}

}
