package com.simibubi.create.content.logistics.tunnel;

import java.util.EnumMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import com.simibubi.create.content.kinetics.belt.transport.ItemHandlerBeltSegment;

import org.apache.commons.lang3.tuple.Pair;
import org.jetbrains.annotations.Nullable;

import com.jozufozu.flywheel.backend.instancing.InstancedRenderDispatcher;
import com.simibubi.create.AllPackets;
import com.simibubi.create.content.kinetics.belt.BeltBlockEntity;
import com.simibubi.create.content.logistics.funnel.BeltFunnelBlock;
import com.simibubi.create.content.logistics.tunnel.BeltTunnelBlock.Shape;
import com.simibubi.create.foundation.blockEntity.SmartBlockEntity;
import com.simibubi.create.foundation.blockEntity.behaviour.BlockEntityBehaviour;
import com.simibubi.create.foundation.utility.Iterate;
import com.simibubi.create.foundation.utility.animation.LerpedFloat;
import com.simibubi.create.foundation.utility.animation.LerpedFloat.Chaser;
import com.tterrag.registrate.fabric.EnvExecutor;

import io.github.fabricators_of_create.porting_lib.util.StorageProvider;
import net.fabricmc.api.EnvType;
import net.fabricmc.fabric.api.transfer.v1.item.ItemVariant;
import net.fabricmc.fabric.api.transfer.v1.storage.Storage;
import net.fabricmc.fabric.api.transfer.v1.storage.base.SidedStorageBlockEntity;
import net.minecraft.class_1937;
import net.minecraft.class_2338;
import net.minecraft.class_2350;
import net.minecraft.class_2350.class_2351;
import net.minecraft.class_2350.class_2352;
import net.minecraft.class_2487;
import net.minecraft.class_2497;
import net.minecraft.class_2499;
import net.minecraft.class_2520;
import net.minecraft.class_2591;
import net.minecraft.class_2680;
import net.minecraft.class_2741;
import net.minecraft.class_3218;

public class BeltTunnelBlockEntity extends SmartBlockEntity implements SidedStorageBlockEntity {

	public Map<class_2350, LerpedFloat> flaps;
	public Set<class_2350> sides;

	protected StorageProvider<ItemVariant> belowProvider;
	protected List<Pair<class_2350, Boolean>> flapsToSend;

	public BeltTunnelBlockEntity(class_2591<?> type, class_2338 pos, class_2680 state) {
		super(type, pos, state);
		flaps = new EnumMap<>(class_2350.class);
		sides = new HashSet<>();
		flapsToSend = new LinkedList<>();
	}

	@Override
	public void method_31662(class_1937 level) {
		super.method_31662(level);
		belowProvider = StorageProvider.createForItems(level, field_11867.method_10074()).filter(this::isBeltStorage);
	}

	public boolean isBeltStorage(StorageProvider<ItemVariant> provider, Storage<ItemVariant> storage) {
		return storage instanceof ItemHandlerBeltSegment;
	}

	@Override
	public void invalidate() {
		super.invalidate();
	}

	protected void writeFlapsAndSides(class_2487 compound) {
		class_2499 flapsNBT = new class_2499();
		for (class_2350 direction : flaps.keySet())
			flapsNBT.add(class_2497.method_23247(direction.method_10146()));
		compound.method_10566("Flaps", flapsNBT);

		class_2499 sidesNBT = new class_2499();
		for (class_2350 direction : sides)
			sidesNBT.add(class_2497.method_23247(direction.method_10146()));
		compound.method_10566("Sides", sidesNBT);
	}

	@Override
	public void writeSafe(class_2487 tag) {
		writeFlapsAndSides(tag);
		super.writeSafe(tag);
	}

	@Override
	public void write(class_2487 compound, boolean clientPacket) {
		writeFlapsAndSides(compound);
		super.write(compound, clientPacket);
	}

	@Override
	protected void read(class_2487 compound, boolean clientPacket) {
		Set<class_2350> newFlaps = new HashSet<>(6);
		class_2499 flapsNBT = compound.method_10554("Flaps", class_2520.field_33253);
		for (class_2520 inbt : flapsNBT)
			if (inbt instanceof class_2497)
				newFlaps.add(class_2350.method_10143(((class_2497) inbt).method_10701()));

		sides.clear();
		class_2499 sidesNBT = compound.method_10554("Sides", class_2520.field_33253);
		for (class_2520 inbt : sidesNBT)
			if (inbt instanceof class_2497)
				sides.add(class_2350.method_10143(((class_2497) inbt).method_10701()));

		for (class_2350 d : Iterate.directions)
			if (!newFlaps.contains(d))
				flaps.remove(d);
			else if (!flaps.containsKey(d))
				flaps.put(d, createChasingFlap());

		// Backwards compat
		if (!compound.method_10545("Sides") && compound.method_10545("Flaps"))
			sides.addAll(flaps.keySet());
		super.read(compound, clientPacket);
		if (clientPacket)
			EnvExecutor.runWhenOn(EnvType.CLIENT, () -> () -> InstancedRenderDispatcher.enqueueUpdate(this));
	}

	private LerpedFloat createChasingFlap() {
		return LerpedFloat.linear()
			.startWithValue(.25f)
			.chase(0, .05f, Chaser.EXP);
	}

	public void updateTunnelConnections() {
		flaps.clear();
		sides.clear();
		class_2680 tunnelState = method_11010();
		for (class_2350 direction : Iterate.horizontalDirections) {
			if (direction.method_10166() != tunnelState.method_11654(class_2741.field_12529)) {
				boolean positive =
					direction.method_10171() == class_2352.field_11056 ^ direction.method_10166() == class_2351.field_11051;
				Shape shape = tunnelState.method_11654(BeltTunnelBlock.SHAPE);
				if (BeltTunnelBlock.isStraight(tunnelState))
					continue;
				if (positive && shape == Shape.T_LEFT)
					continue;
				if (!positive && shape == Shape.T_RIGHT)
					continue;
			}

			sides.add(direction);

			// Flap might be occluded
			class_2680 nextState = field_11863.method_8320(field_11867.method_10093(direction));
			if (nextState.method_26204() instanceof BeltTunnelBlock)
				continue;
			if (nextState.method_26204() instanceof BeltFunnelBlock)
				if (nextState.method_11654(BeltFunnelBlock.SHAPE) == BeltFunnelBlock.Shape.EXTENDED
					&& nextState.method_11654(BeltFunnelBlock.HORIZONTAL_FACING) == direction.method_10153())
					continue;

			flaps.put(direction, createChasingFlap());
		}
		sendData();
	}

	public void flap(class_2350 side, boolean inward) {
		if (field_11863.field_9236) {
			if (flaps.containsKey(side))
				flaps.get(side)
					.setValue(inward ^ side.method_10166() == class_2351.field_11051 ? -1 : 1);
			return;
		}

		flapsToSend.add(Pair.of(side, inward));
	}

	@Override
	public void initialize() {
		super.initialize();
		updateTunnelConnections();
	}

	@Override
	public void tick() {
		super.tick();
		if (!field_11863.field_9236) {
			if (!flapsToSend.isEmpty())
				sendFlaps();
			return;
		}
		flaps.forEach((d, value) -> value.tickChaser());
	}

	private void sendFlaps() {
		AllPackets.getChannel().sendToClientsTracking(new TunnelFlapPacket(this, flapsToSend), (class_3218) field_11863, method_11016());
		flapsToSend.clear();
	}

	@Override
	public void addBehaviours(List<BlockEntityBehaviour> behaviours) {}

	@Override
	public Storage<ItemVariant> getItemStorage(@Nullable class_2350 face) {
		if (belowProvider == null)
			return null;
		if (belowProvider.findBlockEntity() instanceof BeltBlockEntity)
			return belowProvider.get(class_2350.field_11036);
		return null;
	}
}
