package com.simibubi.create.content.logistics.tunnel;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.IdentityHashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Random;
import java.util.Set;

import javax.annotation.Nullable;

import org.apache.commons.lang3.tuple.Pair;

import com.simibubi.create.AllBlocks;
import com.simibubi.create.content.equipment.goggles.IHaveGoggleInformation;
import com.simibubi.create.content.kinetics.belt.BeltBlockEntity;
import com.simibubi.create.content.kinetics.belt.BeltHelper;
import com.simibubi.create.content.kinetics.belt.behaviour.DirectBeltInputBehaviour;
import com.simibubi.create.content.logistics.funnel.BeltFunnelBlock;
import com.simibubi.create.content.logistics.funnel.BeltFunnelBlock.Shape;
import com.simibubi.create.content.logistics.funnel.FunnelBlock;
import com.simibubi.create.foundation.blockEntity.behaviour.BlockEntityBehaviour;
import com.simibubi.create.foundation.blockEntity.behaviour.filtering.FilteringBehaviour;
import com.simibubi.create.foundation.blockEntity.behaviour.filtering.SidedFilteringBehaviour;
import com.simibubi.create.foundation.blockEntity.behaviour.scrollValue.INamedIconOptions;
import com.simibubi.create.foundation.blockEntity.behaviour.scrollValue.ScrollOptionBehaviour;
import com.simibubi.create.foundation.gui.AllIcons;
import com.simibubi.create.foundation.utility.BlockHelper;
import com.simibubi.create.foundation.utility.Components;
import com.simibubi.create.foundation.utility.Couple;
import com.simibubi.create.foundation.utility.Iterate;
import com.simibubi.create.foundation.utility.Lang;
import com.simibubi.create.foundation.utility.NBTHelper;
import com.simibubi.create.infrastructure.config.AllConfigs;

import io.github.fabricators_of_create.porting_lib.transfer.item.ItemHandlerHelper;
import io.github.fabricators_of_create.porting_lib.util.NBTSerializer;
import net.fabricmc.fabric.api.transfer.v1.item.ItemVariant;
import net.fabricmc.fabric.api.transfer.v1.storage.Storage;
import net.fabricmc.fabric.api.transfer.v1.storage.base.SidedStorageBlockEntity;
import net.fabricmc.fabric.api.transfer.v1.transaction.TransactionContext;
import net.fabricmc.fabric.api.transfer.v1.transaction.base.SnapshotParticipant;
import net.minecraft.class_124;
import net.minecraft.class_1542;
import net.minecraft.class_1799;
import net.minecraft.class_2248;
import net.minecraft.class_2338;
import net.minecraft.class_2350;
import net.minecraft.class_2350.class_2351;
import net.minecraft.class_2350.class_2352;
import net.minecraft.class_243;
import net.minecraft.class_2487;
import net.minecraft.class_2512;
import net.minecraft.class_2520;
import net.minecraft.class_2561;
import net.minecraft.class_2586;
import net.minecraft.class_2591;
import net.minecraft.class_2680;

public class BrassTunnelBlockEntity extends BeltTunnelBlockEntity implements IHaveGoggleInformation, SidedStorageBlockEntity {

	SidedFilteringBehaviour filtering;

	boolean connectedLeft;
	boolean connectedRight;

	class_1799 stackToDistribute;
	class_2350 stackEnteredFrom;

	float distributionProgress;
	int distributionDistanceLeft;
	int distributionDistanceRight;
	int previousOutputIndex;

	// <filtered, non-filtered>
	Couple<List<Pair<class_2338, class_2350>>> distributionTargets;

	private boolean syncedOutputActive;
	private Set<BrassTunnelBlockEntity> syncSet;

	protected ScrollOptionBehaviour<SelectionMode> selectionMode;
	private BrassTunnelItemHandler tunnelCapability;

	public final SnapshotParticipant<Data> snapshotParticipant = new SnapshotParticipant<>() {

		@Override
		protected Data createSnapshot() {
			return new Data(stackToDistribute.method_7972(), distributionProgress, stackEnteredFrom);
		}

		@Override
		protected void readSnapshot(Data snapshot) {
			stackToDistribute = snapshot.stack;
			distributionProgress = snapshot.progress;
			stackEnteredFrom = snapshot.enteredFrom;
		}
	};

	private record Data(class_1799 stack, float progress, class_2350 enteredFrom) {
	}

	public BrassTunnelBlockEntity(class_2591<?> type, class_2338 pos, class_2680 state) {
		super(type, pos, state);
		distributionTargets = Couple.create(ArrayList::new);
		syncSet = new HashSet<>();
		stackToDistribute = class_1799.field_8037;
		stackEnteredFrom = null;
		// fabric: beltCapability moved to cache, initialized on level set
		tunnelCapability = new BrassTunnelItemHandler(this);
		previousOutputIndex = 0;
		syncedOutputActive = false;
	}

	@Override
	public void addBehaviours(List<BlockEntityBehaviour> behaviours) {
		super.addBehaviours(behaviours);
		behaviours.add(selectionMode = new ScrollOptionBehaviour<>(SelectionMode.class,
			Lang.translateDirect("logistics.when_multiple_outputs_available"), this, new BrassTunnelModeSlot()));

		selectionMode.onlyActiveWhen(this::hasDistributionBehaviour);

		// Propagate settings across connected tunnels
		selectionMode.withCallback(setting -> {
			for (boolean side : Iterate.trueAndFalse) {
				if (!isConnected(side))
					continue;
				BrassTunnelBlockEntity adjacent = getAdjacent(side);
				if (adjacent != null)
					adjacent.selectionMode.setValue(setting);
			}
		});
	}

	@Override
	public void tick() {
		super.tick();
		BeltBlockEntity beltBelow = BeltHelper.getSegmentBE(field_11863, field_11867.method_10074());

		if (distributionProgress > 0)
			distributionProgress--;
		if (beltBelow == null || beltBelow.getSpeed() == 0)
			return;
		if (stackToDistribute.method_7960() && !syncedOutputActive)
			return;
		if (field_11863.field_9236 && !isVirtual())
			return;

		if (distributionProgress == -1) {
			distributionTargets.forEach(List::clear);
			distributionDistanceLeft = 0;
			distributionDistanceRight = 0;

			syncSet.clear();
			List<Pair<BrassTunnelBlockEntity, class_2350>> validOutputs = gatherValidOutputs();
			if (selectionMode.get() == SelectionMode.SYNCHRONIZE) {
				boolean allEmpty = true;
				boolean allFull = true;
				for (BrassTunnelBlockEntity be : syncSet) {
					boolean hasStack = !be.stackToDistribute.method_7960();
					allEmpty &= !hasStack;
					allFull &= hasStack;
				}
				final boolean notifySyncedOut = !allEmpty;
				if (allFull || allEmpty)
					syncSet.forEach(be -> be.syncedOutputActive = notifySyncedOut);
			}

			if (validOutputs == null)
				return;
			if (stackToDistribute.method_7960())
				return;

			for (Pair<BrassTunnelBlockEntity, class_2350> pair : validOutputs) {
				BrassTunnelBlockEntity tunnel = pair.getKey();
				class_2350 output = pair.getValue();
				if (insertIntoTunnel(tunnel, output, stackToDistribute, true) == null)
					continue;
				distributionTargets.get(!tunnel.flapFilterEmpty(output))
					.add(Pair.of(tunnel.field_11867, output));
				int distance = tunnel.field_11867.method_10263() + tunnel.field_11867.method_10260() - field_11867.method_10263()
					- field_11867.method_10260();
				if (distance < 0)
					distributionDistanceLeft = Math.max(distributionDistanceLeft, -distance);
				else
					distributionDistanceRight = Math.max(distributionDistanceRight, distance);
			}

			if (distributionTargets.getFirst()
				.isEmpty()
				&& distributionTargets.getSecond()
					.isEmpty())
				return;

			if (selectionMode.get() != SelectionMode.SYNCHRONIZE || syncedOutputActive) {
				distributionProgress = AllConfigs.server().logistics.brassTunnelTimer.get();
				sendData();
			}
			return;
		}

		if (distributionProgress != 0)
			return;

		distributionTargets.forEach(list -> {
			if (stackToDistribute.method_7960())
				return;
			List<Pair<BrassTunnelBlockEntity, class_2350>> validTargets = new ArrayList<>();
			for (Pair<class_2338, class_2350> pair : list) {
				class_2338 tunnelPos = pair.getKey();
				class_2350 output = pair.getValue();
				if (tunnelPos.equals(field_11867) && output == stackEnteredFrom)
					continue;
				class_2586 be = field_11863.method_8321(tunnelPos);
				if (!(be instanceof BrassTunnelBlockEntity))
					continue;
				validTargets.add(Pair.of((BrassTunnelBlockEntity) be, output));
			}
			distribute(validTargets);
			distributionProgress = -1;
		});
	}

	private static Random rand = new Random();
	private static Map<Pair<BrassTunnelBlockEntity, class_2350>, class_1799> distributed = new IdentityHashMap<>();
	private static Set<Pair<BrassTunnelBlockEntity, class_2350>> full = new HashSet<>();

	private void distribute(List<Pair<BrassTunnelBlockEntity, class_2350>> validTargets) {
		int amountTargets = validTargets.size();
		if (amountTargets == 0)
			return;

		distributed.clear();
		full.clear();

		int indexStart = previousOutputIndex % amountTargets;
		SelectionMode mode = selectionMode.get();
		boolean force = mode == SelectionMode.FORCED_ROUND_ROBIN || mode == SelectionMode.FORCED_SPLIT;
		boolean split = mode == SelectionMode.FORCED_SPLIT || mode == SelectionMode.SPLIT;
		boolean robin = mode == SelectionMode.FORCED_ROUND_ROBIN || mode == SelectionMode.ROUND_ROBIN;

		if (mode == SelectionMode.RANDOMIZE)
			indexStart = rand.nextInt(amountTargets);
		if (mode == SelectionMode.PREFER_NEAREST || mode == SelectionMode.SYNCHRONIZE)
			indexStart = 0;

		class_1799 toDistribute = stackToDistribute.method_7972();
		for (boolean distributeAgain : Iterate.trueAndFalse) {
			class_1799 toDistributeThisCycle = null;
			int remainingOutputs = amountTargets;
			int leftovers = 0;

			for (boolean simulate : Iterate.trueAndFalse) {
				if (remainingOutputs == 0)
					break;

				leftovers = 0;
				int index = indexStart;
				int stackSize = toDistribute.method_7947();
				int splitStackSize = stackSize / remainingOutputs;
				int splitRemainder = stackSize % remainingOutputs;
				int visited = 0;

				toDistributeThisCycle = toDistribute.method_7972();
				if (!(force || split) && simulate)
					continue;

				while (visited < amountTargets) {
					Pair<BrassTunnelBlockEntity, class_2350> pair = validTargets.get(index);
					BrassTunnelBlockEntity tunnel = pair.getKey();
					class_2350 side = pair.getValue();
					index = (index + 1) % amountTargets;
					visited++;

					if (full.contains(pair)) {
						if (split && simulate)
							remainingOutputs--;
						continue;
					}

					int count = split ? splitStackSize + (splitRemainder > 0 ? 1 : 0) : stackSize;
					class_1799 toOutput = ItemHandlerHelper.copyStackWithSize(toDistributeThisCycle, count);

					// Grow by 1 to determine if target is full even after a successful transfer
					boolean testWithIncreasedCount = distributed.containsKey(pair);
					int increasedCount = testWithIncreasedCount ? distributed.get(pair)
						.method_7947() : 0;
					if (testWithIncreasedCount)
						toOutput.method_7933(increasedCount);

					class_1799 remainder = insertIntoTunnel(tunnel, side, toOutput, true);

					if (remainder == null || remainder.method_7947() == (testWithIncreasedCount ? count + 1 : count)) {
						if (force)
							return;
						if (split && simulate)
							remainingOutputs--;
						if (!simulate)
							full.add(pair);
						if (robin)
							break;
						continue;
					} else if (!remainder.method_7960() && !simulate) {
						full.add(pair);
					}

					if (!simulate) {
						toOutput.method_7934(remainder.method_7947());
						distributed.put(pair, toOutput);
					}

					leftovers += remainder.method_7947();
					toDistributeThisCycle.method_7934(count);
					if (toDistributeThisCycle.method_7960())
						break;
					splitRemainder--;
					if (!split)
						break;
				}
			}

			toDistribute.method_7939(toDistributeThisCycle.method_7947() + leftovers);
			if (leftovers == 0 && distributeAgain)
				break;
			if (!split)
				break;
		}

		int failedTransferrals = 0;
		for (Entry<Pair<BrassTunnelBlockEntity, class_2350>, class_1799> entry : distributed.entrySet()) {
			Pair<BrassTunnelBlockEntity, class_2350> pair = entry.getKey();
			failedTransferrals += insertIntoTunnel(pair.getKey(), pair.getValue(), entry.getValue(), false).method_7947();
		}

		toDistribute.method_7933(failedTransferrals);
		stackToDistribute = ItemHandlerHelper.copyStackWithSize(stackToDistribute, toDistribute.method_7947());
		if (stackToDistribute.method_7960())
			stackEnteredFrom = null;
		previousOutputIndex++;
		previousOutputIndex %= amountTargets;
		notifyUpdate();
	}

	public void setStackToDistribute(class_1799 stack, @Nullable class_2350 enteredFrom, @Nullable TransactionContext ctx) {
		if (ctx != null) {
			snapshotParticipant.updateSnapshots(ctx);
		}
		stackToDistribute = stack;
		stackEnteredFrom = enteredFrom;
		distributionProgress = -1;
	}

	public class_1799 getStackToDistribute() {
		return stackToDistribute;
	}

	public List<class_1799> grabAllStacksOfGroup(boolean simulate) {
		List<class_1799> list = new ArrayList<>();

		class_1799 own = getStackToDistribute();
		if (!own.method_7960()) {
			list.add(own);
			if (!simulate)
				setStackToDistribute(class_1799.field_8037, null, null);
		}

		for (boolean left : Iterate.trueAndFalse) {
			BrassTunnelBlockEntity adjacent = this;
			while (adjacent != null) {
				if (!field_11863.method_8477(adjacent.method_11016()))
					return null;
				adjacent = adjacent.getAdjacent(left);
				if (adjacent == null)
					continue;
				class_1799 other = adjacent.getStackToDistribute();
				if (other.method_7960())
					continue;
				list.add(other);
				if (!simulate)
					adjacent.setStackToDistribute(class_1799.field_8037, null, null);
			}
		}

		return list;
	}

	@Nullable
	protected class_1799 insertIntoTunnel(BrassTunnelBlockEntity tunnel, class_2350 side, class_1799 stack,
		boolean simulate) {
		if (stack.method_7960())
			return stack;
		if (!tunnel.testFlapFilter(side, stack))
			return null;

		BeltBlockEntity below = BeltHelper.getSegmentBE(field_11863, tunnel.field_11867.method_10074());
		if (below == null)
			return null;
		class_2338 offset = tunnel.method_11016()
			.method_10074()
			.method_10093(side);
		DirectBeltInputBehaviour sideOutput = BlockEntityBehaviour.get(field_11863, offset, DirectBeltInputBehaviour.TYPE);
		if (sideOutput != null) {
			if (!sideOutput.canInsertFromSide(side))
				return null;
			class_1799 result = sideOutput.handleInsertion(stack, side, simulate);
			if (result.method_7960() && !simulate)
				tunnel.flap(side, false);
			return result;
		}

		class_2350 movementFacing = below.getMovementFacing();
		if (side == movementFacing)
			if (!BlockHelper.hasBlockSolidSide(field_11863.method_8320(offset), field_11863, offset, side.method_10153())) {
				BeltBlockEntity controllerBE = below.getControllerBE();
				if (controllerBE == null)
					return null;

				if (!simulate) {
					tunnel.flap(side, true);
					class_1799 ejected = stack;
					float beltMovementSpeed = below.getDirectionAwareBeltMovementSpeed();
					float movementSpeed = Math.max(Math.abs(beltMovementSpeed), 1 / 8f);
					int additionalOffset = beltMovementSpeed > 0 ? 1 : 0;
					class_243 outPos = BeltHelper.getVectorForOffset(controllerBE, below.index + additionalOffset);
					class_243 outMotion = class_243.method_24954(side.method_10163())
						.method_1021(movementSpeed)
						.method_1031(0, 1 / 8f, 0);
					outPos.method_1019(outMotion.method_1029());
					class_1542 entity = new class_1542(field_11863, outPos.field_1352, outPos.field_1351 + 6 / 16f, outPos.field_1350, ejected);
					entity.method_18799(outMotion);
					entity.method_6988();
					entity.field_6037 = true;
					field_11863.method_8649(entity);
				}

				return class_1799.field_8037;
			}

		return null;
	}

	public boolean testFlapFilter(class_2350 side, class_1799 stack) {
		if (filtering == null)
			return false;
		if (filtering.get(side) == null) {
			FilteringBehaviour adjacentFilter =
				BlockEntityBehaviour.get(field_11863, field_11867.method_10093(side), FilteringBehaviour.TYPE);
			if (adjacentFilter == null)
				return true;
			return adjacentFilter.test(stack);
		}
		return filtering.test(side, stack);
	}

	public boolean flapFilterEmpty(class_2350 side) {
		if (filtering == null)
			return false;
		if (filtering.get(side) == null) {
			FilteringBehaviour adjacentFilter =
				BlockEntityBehaviour.get(field_11863, field_11867.method_10093(side), FilteringBehaviour.TYPE);
			if (adjacentFilter == null)
				return true;
			return adjacentFilter.getFilter()
				.method_7960();
		}
		return filtering.getFilter(side)
			.method_7960();
	}

	@Override
	public void initialize() {
		if (filtering == null) {
			filtering = createSidedFilter();
			attachBehaviourLate(filtering);
		}
		super.initialize();
	}

	public boolean canInsert(class_2350 side, class_1799 stack) {
		if (filtering != null && !filtering.test(side, stack))
			return false;
		if (!hasDistributionBehaviour())
			return true;
		if (!stackToDistribute.method_7960())
			return false;
		return true;
	}

	public boolean hasDistributionBehaviour() {
		if (flaps.isEmpty())
			return false;
		if (connectedLeft || connectedRight)
			return true;
		class_2680 blockState = method_11010();
		if (!AllBlocks.BRASS_TUNNEL.has(blockState))
			return false;
		class_2351 axis = blockState.method_11654(BrassTunnelBlock.HORIZONTAL_AXIS);
		for (class_2350 direction : flaps.keySet())
			if (direction.method_10166() != axis)
				return true;
		return false;
	}

	private List<Pair<BrassTunnelBlockEntity, class_2350>> gatherValidOutputs() {
		List<Pair<BrassTunnelBlockEntity, class_2350>> validOutputs = new ArrayList<>();
		boolean synchronize = selectionMode.get() == SelectionMode.SYNCHRONIZE;
		addValidOutputsOf(this, validOutputs);

		for (boolean left : Iterate.trueAndFalse) {
			BrassTunnelBlockEntity adjacent = this;
			while (adjacent != null) {
				if (!field_11863.method_8477(adjacent.method_11016()))
					return null;
				adjacent = adjacent.getAdjacent(left);
				if (adjacent == null)
					continue;
				addValidOutputsOf(adjacent, validOutputs);
			}
		}

		if (!syncedOutputActive && synchronize)
			return null;
		return validOutputs;
	}

	private void addValidOutputsOf(BrassTunnelBlockEntity tunnelBE,
		List<Pair<BrassTunnelBlockEntity, class_2350>> validOutputs) {
		syncSet.add(tunnelBE);
		BeltBlockEntity below = BeltHelper.getSegmentBE(field_11863, tunnelBE.field_11867.method_10074());
		if (below == null)
			return;
		class_2350 movementFacing = below.getMovementFacing();
		class_2680 blockState = method_11010();
		if (!AllBlocks.BRASS_TUNNEL.has(blockState))
			return;

		boolean prioritizeSides = tunnelBE == this;

		for (boolean sidePass : Iterate.trueAndFalse) {
			if (!prioritizeSides && sidePass)
				continue;
			for (class_2350 direction : Iterate.horizontalDirections) {
				if (direction == movementFacing && below.getSpeed() == 0)
					continue;
				if (prioritizeSides && sidePass == (direction.method_10166() == movementFacing.method_10166()))
					continue;
				if (direction == movementFacing.method_10153())
					continue;
				if (!tunnelBE.sides.contains(direction))
					continue;

				class_2338 offset = tunnelBE.field_11867.method_10074()
					.method_10093(direction);

				class_2680 potentialFunnel = field_11863.method_8320(offset.method_10084());
				if (potentialFunnel.method_26204() instanceof BeltFunnelBlock
					&& potentialFunnel.method_11654(BeltFunnelBlock.SHAPE) == Shape.PULLING
					&& FunnelBlock.getFunnelFacing(potentialFunnel) == direction)
					continue;

				DirectBeltInputBehaviour inputBehaviour =
					BlockEntityBehaviour.get(field_11863, offset, DirectBeltInputBehaviour.TYPE);
				if (inputBehaviour == null) {
					if (direction == movementFacing)
						if (!BlockHelper.hasBlockSolidSide(field_11863.method_8320(offset), field_11863, offset,
							direction.method_10153()))
							validOutputs.add(Pair.of(tunnelBE, direction));
					continue;
				}
				if (inputBehaviour.canInsertFromSide(direction))
					validOutputs.add(Pair.of(tunnelBE, direction));
				continue;
			}
		}
	}

	@Override
	public void addBehavioursDeferred(List<BlockEntityBehaviour> behaviours) {
		super.addBehavioursDeferred(behaviours);
		filtering = createSidedFilter();
		behaviours.add(filtering);
	}

	protected SidedFilteringBehaviour createSidedFilter() {
		return new SidedFilteringBehaviour(this, new BrassTunnelFilterSlot(), this::makeFilter,
			this::isValidFaceForFilter);
	}

	private FilteringBehaviour makeFilter(class_2350 side, FilteringBehaviour filter) {
		return filter;
	}

	private boolean isValidFaceForFilter(class_2350 side) {
		return sides.contains(side);
	}

	@Override
	public void write(class_2487 compound, boolean clientPacket) {
		compound.method_10556("SyncedOutput", syncedOutputActive);
		compound.method_10556("ConnectedLeft", connectedLeft);
		compound.method_10556("ConnectedRight", connectedRight);

		compound.method_10566("StackToDistribute", NBTSerializer.serializeNBT(stackToDistribute));
		if (stackEnteredFrom != null)
			NBTHelper.writeEnum(compound, "StackEnteredFrom", stackEnteredFrom);

		compound.method_10548("DistributionProgress", distributionProgress);
		compound.method_10569("PreviousIndex", previousOutputIndex);
		compound.method_10569("DistanceLeft", distributionDistanceLeft);
		compound.method_10569("DistanceRight", distributionDistanceRight);

		for (boolean filtered : Iterate.trueAndFalse) {
			compound.method_10566(filtered ? "FilteredTargets" : "Targets",
				NBTHelper.writeCompoundList(distributionTargets.get(filtered), pair -> {
					class_2487 nbt = new class_2487();
					nbt.method_10566("Pos", class_2512.method_10692(pair.getKey()));
					nbt.method_10569("Face", pair.getValue()
						.method_10146());
					return nbt;
				}));
		}

		super.write(compound, clientPacket);
	}

	@Override
	protected void read(class_2487 compound, boolean clientPacket) {
		boolean wasConnectedLeft = connectedLeft;
		boolean wasConnectedRight = connectedRight;

		syncedOutputActive = compound.method_10577("SyncedOutput");
		connectedLeft = compound.method_10577("ConnectedLeft");
		connectedRight = compound.method_10577("ConnectedRight");

		stackToDistribute = class_1799.method_7915(compound.method_10562("StackToDistribute"));
		stackEnteredFrom =
			compound.method_10545("StackEnteredFrom") ? NBTHelper.readEnum(compound, "StackEnteredFrom", class_2350.class)
				: null;

		distributionProgress = compound.method_10583("DistributionProgress");
		previousOutputIndex = compound.method_10550("PreviousIndex");
		distributionDistanceLeft = compound.method_10550("DistanceLeft");
		distributionDistanceRight = compound.method_10550("DistanceRight");

		for (boolean filtered : Iterate.trueAndFalse) {
			distributionTargets.set(filtered, NBTHelper
				.readCompoundList(compound.method_10554(filtered ? "FilteredTargets" : "Targets", class_2520.field_33260), nbt -> {
					class_2338 pos = class_2512.method_10691(nbt.method_10562("Pos"));
					class_2350 face = class_2350.method_10143(nbt.method_10550("Face"));
					return Pair.of(pos, face);
				}));
		}

		super.read(compound, clientPacket);

		if (!clientPacket)
			return;
		if (wasConnectedLeft != connectedLeft || wasConnectedRight != connectedRight) {
//			requestModelDataUpdate();
			if (method_11002())
				field_11863.method_8413(method_11016(), method_11010(), method_11010(), 16);
		}
		filtering.updateFilterPresence();
	}

	public boolean isConnected(boolean leftSide) {
		return leftSide ? connectedLeft : connectedRight;
	}

	@Override
	public void updateTunnelConnections() {
		super.updateTunnelConnections();
		boolean connectivityChanged = false;
		boolean nowConnectedLeft = determineIfConnected(true);
		boolean nowConnectedRight = determineIfConnected(false);

		if (connectedLeft != nowConnectedLeft) {
			connectedLeft = nowConnectedLeft;
			connectivityChanged = true;
			BrassTunnelBlockEntity adjacent = getAdjacent(true);
			if (adjacent != null && !field_11863.field_9236) {
				adjacent.updateTunnelConnections();
				adjacent.selectionMode.setValue(selectionMode.getValue());
			}
		}

		if (connectedRight != nowConnectedRight) {
			connectedRight = nowConnectedRight;
			connectivityChanged = true;
			BrassTunnelBlockEntity adjacent = getAdjacent(false);
			if (adjacent != null && !field_11863.field_9236) {
				adjacent.updateTunnelConnections();
				adjacent.selectionMode.setValue(selectionMode.getValue());
			}
		}

		if (filtering != null)
			filtering.updateFilterPresence();
		if (connectivityChanged)
			sendData();
	}

	protected boolean determineIfConnected(boolean leftSide) {
		if (flaps.isEmpty())
			return false;
		BrassTunnelBlockEntity adjacentTunnelBE = getAdjacent(leftSide);
		return adjacentTunnelBE != null && !adjacentTunnelBE.flaps.isEmpty();
	}

	@Nullable
	protected BrassTunnelBlockEntity getAdjacent(boolean leftSide) {
		if (!method_11002())
			return null;

		class_2680 blockState = method_11010();
		if (!AllBlocks.BRASS_TUNNEL.has(blockState))
			return null;

		class_2351 axis = blockState.method_11654(BrassTunnelBlock.HORIZONTAL_AXIS);
		class_2350 baseDirection = class_2350.method_10156(class_2352.field_11056, axis);
		class_2350 direction = leftSide ? baseDirection.method_10160() : baseDirection.method_10170();
		class_2338 adjacentPos = field_11867.method_10093(direction);
		class_2680 adjacentBlockState = field_11863.method_8320(adjacentPos);

		if (!AllBlocks.BRASS_TUNNEL.has(adjacentBlockState))
			return null;
		if (adjacentBlockState.method_11654(BrassTunnelBlock.HORIZONTAL_AXIS) != axis)
			return null;
		class_2586 adjacentBE = field_11863.method_8321(adjacentPos);
		if (adjacentBE.method_11015())
			return null;
		if (!(adjacentBE instanceof BrassTunnelBlockEntity))
			return null;
		return (BrassTunnelBlockEntity) adjacentBE;
	}

	@Override
	public void invalidate() {
		super.invalidate();
	}

	@Override
	public void destroy() {
		super.destroy();
		class_2248.method_9577(field_11863, field_11867, stackToDistribute);
		stackEnteredFrom = null;
	}

	@Override
	public Storage<ItemVariant> getItemStorage(@Nullable class_2350 face) {
		return tunnelCapability;
	}

	public Storage<ItemVariant> getBeltCapability() {
		return belowProvider != null ? belowProvider.get(class_2350.field_11036) : null;
	}

	public enum SelectionMode implements INamedIconOptions {
		SPLIT(AllIcons.I_TUNNEL_SPLIT),
		FORCED_SPLIT(AllIcons.I_TUNNEL_FORCED_SPLIT),
		ROUND_ROBIN(AllIcons.I_TUNNEL_ROUND_ROBIN),
		FORCED_ROUND_ROBIN(AllIcons.I_TUNNEL_FORCED_ROUND_ROBIN),
		PREFER_NEAREST(AllIcons.I_TUNNEL_PREFER_NEAREST),
		RANDOMIZE(AllIcons.I_TUNNEL_RANDOMIZE),
		SYNCHRONIZE(AllIcons.I_TUNNEL_SYNCHRONIZE),

		;

		private final String translationKey;
		private final AllIcons icon;

		SelectionMode(AllIcons icon) {
			this.icon = icon;
			this.translationKey = "tunnel.selection_mode." + Lang.asId(name());
		}

		@Override
		public AllIcons getIcon() {
			return icon;
		}

		@Override
		public String getTranslationKey() {
			return translationKey;
		}
	}

	public boolean canTakeItems() {
		return stackToDistribute.method_7960() && !syncedOutputActive;
	}

	@Override
	public boolean addToGoggleTooltip(List<class_2561> tooltip, boolean isPlayerSneaking) {
		List<class_1799> allStacks = grabAllStacksOfGroup(true);
		if (allStacks.isEmpty())
			return false;

		tooltip.add(componentSpacing.method_27662()
			.method_10852(Lang.translateDirect("tooltip.brass_tunnel.contains"))
			.method_27692(class_124.field_1068));
		for (class_1799 item : allStacks) {
			tooltip.add(componentSpacing.method_27662()
				.method_10852(Lang.translateDirect("tooltip.brass_tunnel.contains_entry",
					Components.translatable(item.method_7922())
						.getString(),
					item.method_7947()))
				.method_27692(class_124.field_1080));
		}
		tooltip.add(componentSpacing.method_27662()
			.method_10852(Lang.translateDirect("tooltip.brass_tunnel.retrieve"))
			.method_27692(class_124.field_1063));

		return true;
	}
}
