package com.simibubi.create.content.logistics.vault;

import java.util.List;

import javax.annotation.Nullable;

import com.simibubi.create.AllBlockEntityTypes;
import com.simibubi.create.api.connectivity.ConnectivityHandler;
import com.simibubi.create.foundation.blockEntity.IMultiBlockEntityContainer;
import com.simibubi.create.foundation.blockEntity.SmartBlockEntity;
import com.simibubi.create.foundation.blockEntity.behaviour.BlockEntityBehaviour;
import com.simibubi.create.foundation.blockEntity.behaviour.inventory.VersionedInventoryWrapper;
import com.simibubi.create.infrastructure.config.AllConfigs;
import io.github.fabricators_of_create.porting_lib.transfer.item.ItemStackHandler;
import net.fabricmc.fabric.api.transfer.v1.item.ItemVariant;
import net.fabricmc.fabric.api.transfer.v1.storage.Storage;
import net.fabricmc.fabric.api.transfer.v1.storage.base.CombinedSlottedStorage;
import net.fabricmc.fabric.api.transfer.v1.storage.base.CombinedStorage;
import net.fabricmc.fabric.api.transfer.v1.storage.base.SidedStorageBlockEntity;
import net.minecraft.class_1799;
import net.minecraft.class_2246;
import net.minecraft.class_2338;
import net.minecraft.class_2350;
import net.minecraft.class_2350.class_2351;
import net.minecraft.class_2487;
import net.minecraft.class_2512;
import net.minecraft.class_2586;
import net.minecraft.class_2591;
import net.minecraft.class_2680;

public class ItemVaultBlockEntity extends SmartBlockEntity implements IMultiBlockEntityContainer.Inventory, SidedStorageBlockEntity {

	protected Storage<ItemVariant> itemCapability;

	protected ItemStackHandler inventory;
	protected class_2338 controller;
	protected class_2338 lastKnownPos;
	protected boolean updateConnectivity;
	protected int radius;
	protected int length;
	protected class_2351 axis;

	protected boolean recalculateComparatorsNextTick = false;

	public ItemVaultBlockEntity(class_2591<?> type, class_2338 pos, class_2680 state) {
		super(type, pos, state);

		inventory = new ItemStackHandler(AllConfigs.server().logistics.vaultCapacity.get()) {
			@Override
			protected void onContentsChanged(int slot) {
				super.onContentsChanged(slot);
				recalculateComparatorsNextTick = true;
			}
		};

		itemCapability = null;
		radius = 1;
		length = 1;
	}

	@Override
	public void addBehaviours(List<BlockEntityBehaviour> behaviours) {}

	protected void updateConnectivity() {
		updateConnectivity = false;
		if (field_11863.method_8608())
			return;
		if (!isController())
			return;
		ConnectivityHandler.formMulti(this);
	}

	protected void updateComparators() {
		recalculateComparatorsNextTick = false;

		ItemVaultBlockEntity controllerBE = getControllerBE();
		if (controllerBE == null)
			return;

		field_11863.method_8524(controllerBE.field_11867);

		class_2338 pos = controllerBE.method_11016();
		for (int y = 0; y < controllerBE.radius; y++) {
			for (int z = 0; z < (controllerBE.axis == class_2351.field_11048 ? controllerBE.radius : controllerBE.length); z++) {
				for (int x = 0; x < (controllerBE.axis == class_2351.field_11051 ? controllerBE.radius : controllerBE.length); x++) {
					field_11863.method_8455(pos.method_10069(x, y, z), method_11010().method_26204());
				}
			}
		}
	}

	@Override
	public void tick() {
		super.tick();

		if (lastKnownPos == null)
			lastKnownPos = method_11016();
		else if (!lastKnownPos.equals(field_11867) && field_11867 != null) {
			onPositionChanged();
			return;
		}

		if (updateConnectivity)
			updateConnectivity();

		if (recalculateComparatorsNextTick)
			updateComparators();
	}

	@Override
	public class_2338 getLastKnownPos() {
		return lastKnownPos;
	}

	@Override
	public boolean isController() {
		return controller == null || field_11867.method_10263() == controller.method_10263()
			&& field_11867.method_10264() == controller.method_10264() && field_11867.method_10260() == controller.method_10260();
	}

	private void onPositionChanged() {
		removeController(true);
		lastKnownPos = field_11867;
	}

	@SuppressWarnings("unchecked")
	@Override
	public ItemVaultBlockEntity getControllerBE() {
		if (isController())
			return this;
		class_2586 blockEntity = field_11863.method_8321(controller);
		if (blockEntity instanceof ItemVaultBlockEntity)
			return (ItemVaultBlockEntity) blockEntity;
		return null;
	}

	public void removeController(boolean keepContents) {
		if (field_11863.method_8608())
			return;
		updateConnectivity = true;
		controller = null;
		radius = 1;
		length = 1;

		class_2680 state = method_11010();
		if (ItemVaultBlock.isVault(state)) {
			state = state.method_11657(ItemVaultBlock.LARGE, false);
			method_10997().method_8652(field_11867, state, 22);
		}

		itemCapability = null;
		method_5431();
		sendData();
	}

	@Override
	public void setController(class_2338 controller) {
		if (field_11863.field_9236 && !isVirtual())
			return;
		if (controller.equals(this.controller))
			return;
		this.controller = controller;
		itemCapability = null;
		method_5431();
		sendData();
	}

	@Override
	public class_2338 getController() {
		return isController() ? field_11867 : controller;
	}

	@Override
	protected void read(class_2487 compound, boolean clientPacket) {
		super.read(compound, clientPacket);

		class_2338 controllerBefore = controller;
		int prevSize = radius;
		int prevLength = length;

		updateConnectivity = compound.method_10545("Uninitialized");
		controller = null;
		lastKnownPos = null;

		if (compound.method_10545("LastKnownPos"))
			lastKnownPos = class_2512.method_10691(compound.method_10562("LastKnownPos"));
		if (compound.method_10545("Controller"))
			controller = class_2512.method_10691(compound.method_10562("Controller"));

		if (isController()) {
			radius = compound.method_10550("Size");
			length = compound.method_10550("Length");
		}

		if (!clientPacket) {
			inventory.deserializeNBT(compound.method_10562("Inventory"));
			return;
		}

		boolean changeOfController =
			controllerBefore == null ? controller != null : !controllerBefore.equals(controller);
		if (method_11002() && (changeOfController || prevSize != radius || prevLength != length))
			field_11863.method_16109(method_11016(), class_2246.field_10124.method_9564(), method_11010());
	}

	@Override
	protected void write(class_2487 compound, boolean clientPacket) {
		if (updateConnectivity)
			compound.method_10556("Uninitialized", true);
		if (lastKnownPos != null)
			compound.method_10566("LastKnownPos", class_2512.method_10692(lastKnownPos));
		if (!isController())
			compound.method_10566("Controller", class_2512.method_10692(controller));
		if (isController()) {
			compound.method_10569("Size", radius);
			compound.method_10569("Length", length);
		}

		super.write(compound, clientPacket);

		if (!clientPacket) {
			compound.method_10582("StorageType", "CombinedInv");
			compound.method_10566("Inventory", inventory.serializeNBT());
		}
	}

	public ItemStackHandler getInventoryOfBlock() {
		return inventory;
	}

	public void applyInventoryToBlock(ItemStackHandler handler) {
		for (int i = 0; i < inventory.getSlotCount(); i++)
			inventory.setStackInSlot(i, i < handler.getSlotCount() ? handler.getStackInSlot(i) : class_1799.field_8037);
	}

	@Nullable
	@Override
	public Storage<ItemVariant> getItemStorage(@Nullable class_2350 face) {
		initCapability();
		return itemCapability;
	}

	private void initCapability() {
		if (itemCapability != null)
			return;
		if (!isController()) {
			ItemVaultBlockEntity controllerBE = getControllerBE();
			if (controllerBE == null)
				return;
			controllerBE.initCapability();
			itemCapability = controllerBE.itemCapability;
			return;
		}

		boolean alongZ = ItemVaultBlock.getVaultBlockAxis(method_11010()) == class_2351.field_11051;
		ItemStackHandler[] invs = new ItemStackHandler[length * radius * radius];
		for (int yOffset = 0; yOffset < length; yOffset++) {
			for (int xOffset = 0; xOffset < radius; xOffset++) {
				for (int zOffset = 0; zOffset < radius; zOffset++) {
					class_2338 vaultPos = alongZ ? field_11867.method_10069(xOffset, zOffset, yOffset)
						: field_11867.method_10069(yOffset, xOffset, zOffset);
					ItemVaultBlockEntity vaultAt =
						ConnectivityHandler.partAt(AllBlockEntityTypes.ITEM_VAULT.get(), field_11863, vaultPos);
					invs[yOffset * radius * radius + xOffset * radius + zOffset] =
						vaultAt != null ? vaultAt.inventory : new ItemStackHandler();
				}
			}
		}

		Storage<ItemVariant> combinedInvWrapper = new VersionedInventoryWrapper(new CombinedStorage<>(List.of(invs)));
		itemCapability = combinedInvWrapper;
	}

	public static int getMaxLength(int radius) {
		return radius * 3;
	}

	@Override
	public void preventConnectivityUpdate() { updateConnectivity = false; }

	// fabric: see comment in FluidTankItem
	public void queueConnectivityUpdate() {
		updateConnectivity = true;
	}

	@Override
	public void notifyMultiUpdated() {
		class_2680 state = this.method_11010();
		if (ItemVaultBlock.isVault(state)) { // safety
			field_11863.method_8652(method_11016(), state.method_11657(ItemVaultBlock.LARGE, radius > 2), 6);
		}
		itemCapability = null;
		method_5431();
	}

	@Override
	public class_2350.class_2351 getMainConnectionAxis() { return getMainAxisOf(this); }

	@Override
	public int getMaxLength(class_2350.class_2351 longAxis, int width) {
		if (longAxis == class_2350.class_2351.field_11052) return getMaxWidth();
		return getMaxLength(width);
	}

	@Override
	public int getMaxWidth() {
		return 3;
	}

	@Override
	public int getHeight() { return length; }

	@Override
	public int getWidth() { return radius; }

	@Override
	public void setHeight(int height) { this.length = height; }

	@Override
	public void setWidth(int width) { this.radius = width; }

	@Override
	public boolean hasInventory() { return true; }
}
