package com.simibubi.create.content.processing.basin;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Optional;

import org.jetbrains.annotations.Nullable;

import com.google.common.collect.ImmutableList;
import com.simibubi.create.AllParticleTypes;
import com.simibubi.create.AllTags;
import com.simibubi.create.content.equipment.goggles.IHaveGoggleInformation;
import com.simibubi.create.content.fluids.FluidFX;
import com.simibubi.create.content.fluids.particle.FluidParticleData;
import com.simibubi.create.content.kinetics.belt.behaviour.DirectBeltInputBehaviour;
import com.simibubi.create.content.kinetics.mixer.MechanicalMixerBlockEntity;
import com.simibubi.create.content.processing.burner.BlazeBurnerBlock;
import com.simibubi.create.content.processing.burner.BlazeBurnerBlock.HeatLevel;
import com.simibubi.create.foundation.blockEntity.SmartBlockEntity;
import com.simibubi.create.foundation.blockEntity.behaviour.BlockEntityBehaviour;
import com.simibubi.create.foundation.blockEntity.behaviour.ValueBoxTransform;
import com.simibubi.create.foundation.blockEntity.behaviour.filtering.FilteringBehaviour;
import com.simibubi.create.foundation.blockEntity.behaviour.fluid.SmartFluidTankBehaviour;
import com.simibubi.create.foundation.blockEntity.behaviour.fluid.SmartFluidTankBehaviour.TankSegment;
import com.simibubi.create.foundation.blockEntity.behaviour.inventory.InvManipulationBehaviour;
import com.simibubi.create.foundation.fluid.CombinedTankWrapper;
import com.simibubi.create.foundation.item.ItemHelper;
import com.simibubi.create.foundation.item.SmartInventory;
import com.simibubi.create.foundation.utility.AnimationTickHolder;
import com.simibubi.create.foundation.utility.BlockHelper;
import com.simibubi.create.foundation.utility.Components;
import com.simibubi.create.foundation.utility.Couple;
import com.simibubi.create.foundation.utility.Iterate;
import com.simibubi.create.foundation.utility.Lang;
import com.simibubi.create.foundation.utility.LangBuilder;
import com.simibubi.create.foundation.utility.LongAttached;
import com.simibubi.create.foundation.utility.NBTHelper;
import com.simibubi.create.foundation.utility.VecHelper;
import com.simibubi.create.foundation.utility.animation.LerpedFloat;
import com.simibubi.create.foundation.utility.animation.LerpedFloat.Chaser;
import com.simibubi.create.infrastructure.config.AllConfigs;

import io.github.fabricators_of_create.porting_lib.transfer.TransferUtil;
import io.github.fabricators_of_create.porting_lib.fluids.FluidStack;
import io.github.fabricators_of_create.porting_lib.util.FluidTextUtil;
import io.github.fabricators_of_create.porting_lib.util.FluidUnit;
import io.github.fabricators_of_create.porting_lib.util.NBTSerializer;
import io.github.fabricators_of_create.porting_lib.util.StorageProvider;
import it.unimi.dsi.fastutil.Pair;
import net.fabricmc.fabric.api.transfer.v1.fluid.FluidConstants;
import net.fabricmc.fabric.api.transfer.v1.fluid.FluidVariant;
import net.fabricmc.fabric.api.transfer.v1.item.ItemVariant;
import net.fabricmc.fabric.api.transfer.v1.storage.Storage;
import net.fabricmc.fabric.api.transfer.v1.storage.StorageView;
import net.fabricmc.fabric.api.transfer.v1.storage.base.CombinedStorage;
import net.fabricmc.fabric.api.transfer.v1.storage.base.SidedStorageBlockEntity;
import net.fabricmc.fabric.api.transfer.v1.transaction.Transaction;
import net.fabricmc.fabric.api.transfer.v1.transaction.TransactionContext;
import net.fabricmc.fabric.api.transfer.v1.transaction.base.SnapshotParticipant;
import net.minecraft.class_124;
import net.minecraft.class_1799;
import net.minecraft.class_1937;
import net.minecraft.class_2248;
import net.minecraft.class_2338;
import net.minecraft.class_2350;
import net.minecraft.class_2350.class_2351;
import net.minecraft.class_2394;
import net.minecraft.class_243;
import net.minecraft.class_2487;
import net.minecraft.class_2499;
import net.minecraft.class_2519;
import net.minecraft.class_2520;
import net.minecraft.class_2561;
import net.minecraft.class_2586;
import net.minecraft.class_2591;
import net.minecraft.class_2680;
import net.minecraft.class_3532;
import net.minecraft.class_5819;

public class BasinBlockEntity extends SmartBlockEntity implements IHaveGoggleInformation, SidedStorageBlockEntity {

	private boolean needsUpdate; // fabric: need to delay to avoid doing stuff mid-transaction, causing a crash
	private boolean areFluidsMoving;
	LerpedFloat ingredientRotationSpeed;
	LerpedFloat ingredientRotation;

	public BasinInventory inputInventory;
	public SmartFluidTankBehaviour inputTank;
	protected SmartInventory outputInventory;
	protected SmartFluidTankBehaviour outputTank;
	private FilteringBehaviour filtering;
	private boolean contentsChanged;

	private Couple<SmartInventory> invs;
	private Couple<SmartFluidTankBehaviour> tanks;

	protected Storage<ItemVariant> itemCapability;
	protected Storage<FluidVariant> fluidCapability;

	List<class_2350> disabledSpoutputs;
	class_2350 preferredSpoutput;
	protected List<class_1799> spoutputBuffer;
	protected List<FluidStack> spoutputFluidBuffer;
	int recipeBackupCheck;

	public static final int OUTPUT_ANIMATION_TIME = 10;
	List<LongAttached<class_1799>> visualizedOutputItems;
	List<LongAttached<FluidStack>> visualizedOutputFluids;

	// fabric: transfer things

	private final Map<class_2350, Pair<StorageProvider<ItemVariant>, StorageProvider<FluidVariant>>> spoutputOutputs = new HashMap<>();

	SnapshotParticipant<Data> snapshotParticipant = new SnapshotParticipant<>() {
		@Override
		protected Data createSnapshot() {
			return new Data(new ArrayList<>(spoutputBuffer), new ArrayList<>(spoutputFluidBuffer));
		}

		@Override
		protected void readSnapshot(Data snapshot) {
			spoutputBuffer = snapshot.spoutputBuffer;
			spoutputFluidBuffer = snapshot.spoutputFluidBuffer;
		}
	};

	record Data(List<class_1799> spoutputBuffer, List<FluidStack> spoutputFluidBuffer) {
	}

	public BasinBlockEntity(class_2591<?> type, class_2338 pos, class_2680 state) {
		super(type, pos, state);
		inputInventory = new BasinInventory(9, this);
		inputInventory.whenContentsChanged(() -> contentsChanged = true);
		outputInventory = new BasinInventory(9, this).forbidInsertion()
			.withMaxStackSize(64);
		areFluidsMoving = false;
		itemCapability = new CombinedStorage<>(List.of(inputInventory, outputInventory));
		contentsChanged = true;
		ingredientRotation = LerpedFloat.angular()
			.startWithValue(0);
		ingredientRotationSpeed = LerpedFloat.linear()
			.startWithValue(0);

		invs = Couple.create(inputInventory, outputInventory);
		tanks = Couple.create(inputTank, outputTank);
		visualizedOutputItems = Collections.synchronizedList(new ArrayList<>());
		visualizedOutputFluids = Collections.synchronizedList(new ArrayList<>());
		disabledSpoutputs = new ArrayList<>();
		preferredSpoutput = null;
		spoutputBuffer = new ArrayList<>();
		spoutputFluidBuffer = new ArrayList<>();
		recipeBackupCheck = 20;
	}

	@Override
	public void method_31662(class_1937 level) {
		super.method_31662(level);
		spoutputOutputs.clear();
		for (class_2350 direction : Iterate.horizontalDirections) {
			class_2338 pos = method_11016().method_10074().method_10093(direction);
			StorageProvider<ItemVariant> items = StorageProvider.createForItems(level, pos);
			StorageProvider<FluidVariant> fluids = StorageProvider.createForFluids(level, pos);
			spoutputOutputs.put(direction, Pair.of(items, fluids));
		}
	}

	public Storage<ItemVariant> getItemSpoutputOutput(class_2350 facing) {
		Pair<StorageProvider<ItemVariant>, StorageProvider<FluidVariant>> providers = spoutputOutputs.get(facing);
		return providers == null ? null : providers.first().get(facing.method_10153());
	}

	public Storage<FluidVariant> getFluidSpoutputOutput(class_2350 facing) {
		Pair<StorageProvider<ItemVariant>, StorageProvider<FluidVariant>> providers = spoutputOutputs.get(facing);
		return providers == null ? null : providers.second().get(facing.method_10153());
	}

	@Override
	public void addBehaviours(List<BlockEntityBehaviour> behaviours) {
		behaviours.add(new DirectBeltInputBehaviour(this));
		filtering = new FilteringBehaviour(this, new BasinValueBox()).withCallback(newFilter -> contentsChanged = true)
			.forRecipes();
		behaviours.add(filtering);

		inputTank = new SmartFluidTankBehaviour(SmartFluidTankBehaviour.INPUT, this, 2, FluidConstants.BUCKET, true)
			.whenFluidUpdates(() -> contentsChanged = true);
		outputTank = new SmartFluidTankBehaviour(SmartFluidTankBehaviour.OUTPUT, this, 2, FluidConstants.BUCKET, true)
			.whenFluidUpdates(() -> contentsChanged = true)
			.forbidInsertion();
		behaviours.add(inputTank);
		behaviours.add(outputTank);

		fluidCapability = new CombinedTankWrapper(inputTank.getCapability(), outputTank.getCapability());
	}

	@Override
	protected void read(class_2487 compound, boolean clientPacket) {
		super.read(compound, clientPacket);
		inputInventory.deserializeNBT(compound.method_10562("InputItems"));
		outputInventory.deserializeNBT(compound.method_10562("OutputItems"));

		preferredSpoutput = null;
		if (compound.method_10545("PreferredSpoutput"))
			preferredSpoutput = NBTHelper.readEnum(compound, "PreferredSpoutput", class_2350.class);
		disabledSpoutputs.clear();
		class_2499 disabledList = compound.method_10554("DisabledSpoutput", class_2520.field_33258);
		disabledList.forEach(d -> disabledSpoutputs.add(class_2350.valueOf(((class_2519) d).method_10714())));
		spoutputBuffer = NBTHelper.readItemList(compound.method_10554("Overflow", class_2520.field_33260));
		spoutputFluidBuffer = NBTHelper.readCompoundList(compound.method_10554("FluidOverflow", class_2520.field_33260),
			FluidStack::loadFluidStackFromNBT);

		if (!clientPacket)
			return;

		NBTHelper.iterateCompoundList(compound.method_10554("VisualizedItems", class_2520.field_33260),
			c -> visualizedOutputItems.add(LongAttached.with(OUTPUT_ANIMATION_TIME, class_1799.method_7915(c))));
		NBTHelper.iterateCompoundList(compound.method_10554("VisualizedFluids", class_2520.field_33260),
			c -> visualizedOutputFluids
				.add(LongAttached.with(OUTPUT_ANIMATION_TIME, FluidStack.loadFluidStackFromNBT(c))));
	}

	@Override
	public void write(class_2487 compound, boolean clientPacket) {
		super.write(compound, clientPacket);
		compound.method_10566("InputItems", inputInventory.serializeNBT());
		compound.method_10566("OutputItems", outputInventory.serializeNBT());

		if (preferredSpoutput != null)
			NBTHelper.writeEnum(compound, "PreferredSpoutput", preferredSpoutput);
		class_2499 disabledList = new class_2499();
		disabledSpoutputs.forEach(d -> disabledList.add(class_2519.method_23256(d.name())));
		compound.method_10566("DisabledSpoutput", disabledList);
		compound.method_10566("Overflow", NBTHelper.writeItemList(spoutputBuffer));
		compound.method_10566("FluidOverflow",
			NBTHelper.writeCompoundList(spoutputFluidBuffer, fs -> fs.writeToNBT(new class_2487())));

		if (!clientPacket)
			return;

		compound.method_10566("VisualizedItems", NBTHelper.writeCompoundList(visualizedOutputItems, ia -> NBTSerializer.serializeNBTCompound(ia.getValue())));
		compound.method_10566("VisualizedFluids", NBTHelper.writeCompoundList(visualizedOutputFluids, ia -> ia.getValue()
			.writeToNBT(new class_2487())));
		visualizedOutputItems.clear();
		visualizedOutputFluids.clear();
	}

	@Override
	public void destroy() {
		super.destroy();
		ItemHelper.dropContents(field_11863, field_11867, inputInventory);
		ItemHelper.dropContents(field_11863, field_11867, outputInventory);
		spoutputBuffer.forEach(is -> class_2248.method_9577(field_11863, field_11867, is));
	}

	@Override
	public void remove() {
		super.remove();
		onEmptied();
	}

	public void onEmptied() {
		getOperator().ifPresent(be -> be.basinRemoved = true);
	}

	@Override
	public void invalidate() {
		super.invalidate();
	}

//	@Nonnull
//	@Override
//	public <T> LazyOptional<T> getCapability(@Nonnull Capability<T> cap, Direction side) {
//		if (cap == ForgeCapabilities.ITEM_HANDLER)
//			return itemCapability.cast();
//		if (cap == ForgeCapabilities.FLUID_HANDLER)
//			return fluidCapability.cast();
//		return super.getCapability(cap, side);
//	}

	@Override
	public void notifyUpdate() {
		this.needsUpdate = true;
	}

	@Override
	public void lazyTick() {
		super.lazyTick();

		if (!field_11863.field_9236) {
			updateSpoutput();
			if (recipeBackupCheck-- > 0)
				return;
			recipeBackupCheck = 20;
			if (isEmpty())
				return;
			notifyChangeOfContents();
			return;
		}

		class_2586 blockEntity = field_11863.method_8321(field_11867.method_10086(2));
		if (!(blockEntity instanceof MechanicalMixerBlockEntity)) {
			setAreFluidsMoving(false);
			return;
		}

		MechanicalMixerBlockEntity mixer = (MechanicalMixerBlockEntity) blockEntity;
		setAreFluidsMoving(mixer.running && mixer.runningTicks <= 20);
	}

	public boolean isEmpty() {
		return inputInventory.method_5442() && outputInventory.method_5442() && inputTank.isEmpty() && outputTank.isEmpty();
	}

	public void onWrenched(class_2350 face) {
		class_2680 blockState = method_11010();
		class_2350 currentFacing = blockState.method_11654(BasinBlock.FACING);

		disabledSpoutputs.remove(face);
		if (currentFacing == face) {
			if (preferredSpoutput == face)
				preferredSpoutput = null;
			disabledSpoutputs.add(face);
		} else
			preferredSpoutput = face;

		updateSpoutput();
	}

	private void updateSpoutput() {
		class_2680 blockState = method_11010();
		class_2350 currentFacing = blockState.method_11654(BasinBlock.FACING);
		class_2350 newFacing = class_2350.field_11033;
		for (class_2350 test : Iterate.horizontalDirections) {
			boolean canOutputTo = BasinBlock.canOutputTo(field_11863, field_11867, test);
			if (canOutputTo && !disabledSpoutputs.contains(test))
				newFacing = test;
		}

		if (preferredSpoutput != null && BasinBlock.canOutputTo(field_11863, field_11867, preferredSpoutput)
			&& preferredSpoutput != class_2350.field_11036)
			newFacing = preferredSpoutput;

		if (newFacing == currentFacing)
			return;

		field_11863.method_8501(field_11867, blockState.method_11657(BasinBlock.FACING, newFacing));

		if (newFacing.method_10166()
			.method_10178())
			return;

		try (Transaction t = TransferUtil.getTransaction()) {
			for (StorageView<ItemVariant> view : outputInventory.nonEmptyViews()) {
				ItemVariant variant = view.getResource();
				class_1799 stack = variant.toStack(ItemHelper.truncateLong(view.getAmount()));
				if (acceptOutputs(ImmutableList.of(stack), ImmutableList.of(), t)) {
					view.extract(variant, stack.method_7947(), t);
				}
			}
			for (StorageView<FluidVariant> view : outputTank.getCapability().nonEmptyViews()) {
				FluidVariant variant = view.getResource();
				FluidStack stack = new FluidStack(view);
				if (acceptOutputs(ImmutableList.of(), ImmutableList.of(stack), t)) {
					view.extract(variant, stack.getAmount(), t);
				}
			}
			t.commit();
		}

		notifyChangeOfContents();
		notifyUpdate();
	}

	@Override
	public void tick() {
		super.tick();
		if (needsUpdate) {
			needsUpdate = false;
			super.notifyUpdate();
		}
		if (field_11863.field_9236) {
			createFluidParticles();
			tickVisualizedOutputs();
			ingredientRotationSpeed.tickChaser();
			ingredientRotation.setValue(ingredientRotation.getValue() + ingredientRotationSpeed.getValue());
		}

		if ((!spoutputBuffer.isEmpty() || !spoutputFluidBuffer.isEmpty()) && !field_11863.field_9236)
			tryClearingSpoutputOverflow();
		if (!contentsChanged)
			return;

		contentsChanged = false;
		sendData();
		getOperator().ifPresent(be -> be.basinChecker.scheduleUpdate());

		for (class_2350 offset : Iterate.horizontalDirections) {
			class_2338 toUpdate = field_11867.method_10084()
				.method_10093(offset);
			class_2680 stateToUpdate = field_11863.method_8320(toUpdate);
			if (stateToUpdate.method_26204() instanceof BasinBlock
				&& stateToUpdate.method_11654(BasinBlock.FACING) == offset.method_10153()) {
				class_2586 be = field_11863.method_8321(toUpdate);
				if (be instanceof BasinBlockEntity)
					((BasinBlockEntity) be).contentsChanged = true;
			}
		}
	}

	private void tryClearingSpoutputOverflow() {
		class_2680 blockState = method_11010();
		if (!(blockState.method_26204() instanceof BasinBlock))
			return;
		class_2350 direction = blockState.method_11654(BasinBlock.FACING);
		class_2586 be = field_11863.method_8321(field_11867.method_10074()
			.method_10093(direction));

		FilteringBehaviour filter = null;
		InvManipulationBehaviour inserter = null;
		if (be != null) {
			filter = BlockEntityBehaviour.get(field_11863, be.method_11016(), FilteringBehaviour.TYPE);
			inserter = BlockEntityBehaviour.get(field_11863, be.method_11016(), InvManipulationBehaviour.TYPE);
		}

		Storage<ItemVariant> targetInv = getItemSpoutputOutput(direction);
		if (targetInv == null && inserter != null) targetInv = inserter.getInventory();

		Storage<FluidVariant> targetTank = getFluidSpoutputOutput(direction);

		boolean update = false;

		try (Transaction t = TransferUtil.getTransaction()) {
			for (Iterator<class_1799> iterator = spoutputBuffer.iterator(); iterator.hasNext();) {
				class_1799 itemStack = iterator.next();
				// fabric: cleanup for #599
				if (itemStack.method_7960())
					continue;

				if (direction == class_2350.field_11033) {
					class_2248.method_9577(field_11863, field_11867, itemStack);
					iterator.remove();
					update = true;
					continue;
				}

				if (targetInv == null)
					break;
				try (Transaction nested = t.openNested()) {
					long inserted = targetInv.insert(ItemVariant.of(itemStack), itemStack.method_7947(), nested);
					if (itemStack.method_7947() != inserted)
						continue;
					if (filter != null && !filter.test(itemStack))
						continue;

					update = true;
					iterator.remove();
					visualizedOutputItems.add(LongAttached.withZero(itemStack));
					nested.commit();
				}
			}

			for (Iterator<FluidStack> iterator = spoutputFluidBuffer.iterator(); iterator.hasNext();) {
				FluidStack fluidStack = iterator.next();

				if (direction == class_2350.field_11033) {
					iterator.remove();
					update = true;
					continue;
				}

				if (targetTank == null)
					break;

				try (Transaction nested = t.openNested()) {
					long fill = targetTank instanceof SmartFluidTankBehaviour.InternalFluidHandler
							? ((SmartFluidTankBehaviour.InternalFluidHandler) targetTank).forceFill(fluidStack.copy(), nested)
							: targetTank.insert(fluidStack.getType(), fluidStack.getAmount(), nested);
					if (fill != fluidStack.getAmount())
						break;

					update = true;
					iterator.remove();
					visualizedOutputFluids.add(LongAttached.withZero(fluidStack));
					nested.commit();
				}
			}

			if (update) {
				notifyChangeOfContents();
				sendData();
			}
			t.commit();
		}
	}

	public float getTotalFluidUnits(float partialTicks) {
		int renderedFluids = 0;
		float totalUnits = 0;

		for (SmartFluidTankBehaviour behaviour : getTanks()) {
			if (behaviour == null)
				continue;
			for (TankSegment tankSegment : behaviour.getTanks()) {
				if (tankSegment.getRenderedFluid()
					.isEmpty())
					continue;
				float units = tankSegment.getTotalUnits(partialTicks);
				if (units < 1)
					continue;
				totalUnits += units;
				renderedFluids++;
			}
		}

		if (renderedFluids == 0)
			return 0;
		if (totalUnits < 1)
			return 0;
		return totalUnits;
	}

	private Optional<BasinOperatingBlockEntity> getOperator() {
		if (field_11863 == null)
			return Optional.empty();
		class_2586 be = field_11863.method_8321(field_11867.method_10086(2));
		if (be instanceof BasinOperatingBlockEntity)
			return Optional.of((BasinOperatingBlockEntity) be);
		return Optional.empty();
	}

	public FilteringBehaviour getFilter() {
		return filtering;
	}

	public void notifyChangeOfContents() {
		contentsChanged = true;
	}

	public SmartInventory getInputInventory() {
		return inputInventory;
	}

	public SmartInventory getOutputInventory() {
		return outputInventory;
	}

	public boolean canContinueProcessing() {
		return spoutputBuffer.isEmpty() && spoutputFluidBuffer.isEmpty();
	}

	public boolean acceptOutputs(List<class_1799> outputItems, List<FluidStack> outputFluids, TransactionContext ctx) {
		outputInventory.allowInsertion();
		outputTank.allowInsertion();
		boolean acceptOutputsInner = acceptOutputsInner(outputItems, outputFluids, ctx);
		outputInventory.forbidInsertion();
		outputTank.forbidInsertion();
		return acceptOutputsInner;
	}

	private boolean acceptOutputsInner(List<class_1799> outputItems, List<FluidStack> outputFluids, TransactionContext ctx) {
		class_2680 blockState = method_11010();
		if (!(blockState.method_26204() instanceof BasinBlock))
			return false;

		class_2350 direction = blockState.method_11654(BasinBlock.FACING);
		snapshotParticipant.updateSnapshots(ctx);
		if (direction != class_2350.field_11033) {

			class_2586 be = field_11863.method_8321(field_11867.method_10074()
					.method_10093(direction));

			InvManipulationBehaviour inserter =
					be == null ? null : BlockEntityBehaviour.get(field_11863, be.method_11016(), InvManipulationBehaviour.TYPE);
			Storage<ItemVariant> targetInv = getItemSpoutputOutput(direction);
			if (targetInv == null && inserter != null) targetInv = inserter.getInventory();
			Storage<FluidVariant> targetTank = getFluidSpoutputOutput(direction);
			boolean externalTankNotPresent = targetTank == null;

			if (!outputItems.isEmpty() && targetInv == null)
				return false;
			if (!outputFluids.isEmpty() && externalTankNotPresent) {
				// Special case - fluid outputs but output only accepts items
				targetTank = outputTank.getCapability();
				if (targetTank == null)
					return false;
				if (!acceptFluidOutputsIntoBasin(outputFluids, ctx, targetTank))
					return false;
			}

			for (class_1799 itemStack : outputItems) {
				spoutputBuffer.add(itemStack.method_7972());
			}
			if (!externalTankNotPresent)
				for (FluidStack fluidStack : outputFluids)
					spoutputFluidBuffer.add(fluidStack.copy());
			return true;
		}

		Storage<ItemVariant> targetInv = outputInventory;
		Storage<FluidVariant> targetTank = outputTank.getCapability();

		if (targetInv == null && !outputItems.isEmpty())
			return false;
		if (!acceptItemOutputsIntoBasin(outputItems, ctx, targetInv))
			return false;
		if (outputFluids.isEmpty())
			return true;
		if (targetTank == null)
			return false;
		if (!acceptFluidOutputsIntoBasin(outputFluids, ctx, targetTank))
			return false;

		return true;
	}

	private boolean acceptFluidOutputsIntoBasin(List<FluidStack> outputFluids, TransactionContext ctx,
		Storage<FluidVariant> targetTank) {
		for (FluidStack fluidStack : outputFluids) {
			long fill = targetTank instanceof SmartFluidTankBehaviour.InternalFluidHandler
				? ((SmartFluidTankBehaviour.InternalFluidHandler) targetTank).forceFill(fluidStack.copy(), ctx)
				: targetTank.insert(fluidStack.getType(), fluidStack.getAmount(), ctx);
			if (fill != fluidStack.getAmount())
				return false;
		}
		return true;
	}

	private boolean acceptItemOutputsIntoBasin(List<class_1799> outputItems, TransactionContext ctx, Storage<ItemVariant> targetInv) {
		try (Transaction t = ctx.openNested()) {
			for (class_1799 itemStack : outputItems) {
				long inserted = targetInv.insert(ItemVariant.of(itemStack), itemStack.method_7947(), t);
				if (inserted != itemStack.method_7947()) {
					t.commit();
					return false;
				}
			}
			t.commit();
		}
		return true;
	}

	public void readOnlyItems(class_2487 compound) {
		inputInventory.deserializeNBT(compound.method_10562("InputItems"));
		outputInventory.deserializeNBT(compound.method_10562("OutputItems"));
	}

	public static HeatLevel getHeatLevelOf(class_2680 state) {
		if (state.method_28498(BlazeBurnerBlock.HEAT_LEVEL))
			return state.method_11654(BlazeBurnerBlock.HEAT_LEVEL);
		return AllTags.AllBlockTags.PASSIVE_BOILER_HEATERS.matches(state) && BlockHelper.isNotUnheated(state) ? HeatLevel.SMOULDERING : HeatLevel.NONE;
	}

	public Couple<SmartFluidTankBehaviour> getTanks() {
		return tanks;
	}

	public Couple<SmartInventory> getInvs() {
		return invs;
	}

	// client things

	private void tickVisualizedOutputs() {
		visualizedOutputFluids.forEach(LongAttached::decrement);
		visualizedOutputItems.forEach(LongAttached::decrement);
		visualizedOutputFluids.removeIf(LongAttached::isOrBelowZero);
		visualizedOutputItems.removeIf(LongAttached::isOrBelowZero);
	}

	private void createFluidParticles() {
		class_5819 r = field_11863.field_9229;

		if (!visualizedOutputFluids.isEmpty())
			createOutputFluidParticles(r);

		if (!areFluidsMoving && r.method_43057() > 1 / 8f)
			return;

		int segments = 0;
		for (SmartFluidTankBehaviour behaviour : getTanks()) {
			if (behaviour == null)
				continue;
			for (TankSegment tankSegment : behaviour.getTanks())
				if (!tankSegment.isEmpty(0))
					segments++;
		}
		if (segments < 2)
			return;

		float totalUnits = getTotalFluidUnits(0);
		if (totalUnits == 0)
			return;
		float fluidLevel = class_3532.method_15363(totalUnits / 2000, 0, 1);
		float rim = 2 / 16f;
		float space = 12 / 16f;
		float surface = field_11867.method_10264() + rim + space * fluidLevel + 1 / 32f;

		if (areFluidsMoving) {
			createMovingFluidParticles(surface, segments);
			return;
		}

		for (SmartFluidTankBehaviour behaviour : getTanks()) {
			if (behaviour == null)
				continue;
			for (TankSegment tankSegment : behaviour.getTanks()) {
				if (tankSegment.isEmpty(0))
					continue;
				float x = field_11867.method_10263() + rim + space * r.method_43057();
				float z = field_11867.method_10260() + rim + space * r.method_43057();
				field_11863.method_8494(
					new FluidParticleData(AllParticleTypes.BASIN_FLUID.get(), tankSegment.getRenderedFluid()), x,
					surface, z, 0, 0, 0);
			}
		}
	}

	private void createOutputFluidParticles(class_5819 r) {
		class_2680 blockState = method_11010();
		if (!(blockState.method_26204() instanceof BasinBlock))
			return;
		class_2350 direction = blockState.method_11654(BasinBlock.FACING);
		if (direction == class_2350.field_11033)
			return;
		class_243 directionVec = class_243.method_24954(direction.method_10163());
		class_243 outVec = VecHelper.getCenterOf(field_11867)
			.method_1019(directionVec.method_1021(.65)
				.method_1023(0, 1 / 4f, 0));
		class_243 outMotion = directionVec.method_1021(1 / 16f)
			.method_1031(0, -1 / 16f, 0);

		for (int i = 0; i < 2; i++) {
			visualizedOutputFluids.forEach(ia -> {
				FluidStack fluidStack = ia.getValue();
				class_2394 fluidParticle = FluidFX.getFluidParticle(fluidStack);
				class_243 m = VecHelper.offsetRandomly(outMotion, r, 1 / 16f);
				field_11863.method_8494(fluidParticle, outVec.field_1352, outVec.field_1351, outVec.field_1350, m.field_1352, m.field_1351, m.field_1350);
			});
		}
	}

	private void createMovingFluidParticles(float surface, int segments) {
		class_243 pointer = new class_243(1, 0, 0).method_1021(1 / 16f);
		float interval = 360f / segments;
		class_243 centerOf = VecHelper.getCenterOf(field_11867);
		float intervalOffset = (AnimationTickHolder.getTicks() * 18) % 360;

		int currentSegment = 0;
		for (SmartFluidTankBehaviour behaviour : getTanks()) {
			if (behaviour == null)
				continue;
			for (TankSegment tankSegment : behaviour.getTanks()) {
				if (tankSegment.isEmpty(0))
					continue;
				float angle = interval * (1 + currentSegment) + intervalOffset;
				class_243 vec = centerOf.method_1019(VecHelper.rotate(pointer, angle, class_2351.field_11052));
				field_11863.method_8494(
					new FluidParticleData(AllParticleTypes.BASIN_FLUID.get(), tankSegment.getRenderedFluid()), vec.method_10216(),
					surface, vec.method_10215(), 1, 0, 0);
				currentSegment++;
			}
		}
	}

	public boolean areFluidsMoving() {
		return areFluidsMoving;
	}

	public boolean setAreFluidsMoving(boolean areFluidsMoving) {
		this.areFluidsMoving = areFluidsMoving;
		ingredientRotationSpeed.chase(areFluidsMoving ? 20 : 0, .1f, Chaser.EXP);
		return areFluidsMoving;
	}

	@Override
	public boolean addToGoggleTooltip(List<class_2561> tooltip, boolean isPlayerSneaking) {
		Lang.translate("gui.goggles.basin_contents")
			.forGoggles(tooltip);

		boolean isEmpty = true;

		for (SmartInventory inventory : invs) {
			for (int i = 0; i < inventory.getSlotCount(); i++) {
				class_1799 stackInSlot = inventory.getStackInSlot(i);
				if (stackInSlot.method_7960())
					continue;
				Lang.text("")
						.add(Components.translatable(stackInSlot.method_7922())
								.method_27692(class_124.field_1080))
						.add(Lang.text(" x" + stackInSlot.method_7947())
								.style(class_124.field_1060))
						.forGoggles(tooltip, 1);
				isEmpty = false;
			}
		}

		FluidUnit unit = AllConfigs.client().fluidUnitType.get();
		LangBuilder unitSuffix = Lang.translate(unit.getTranslationKey());
		boolean simplify = AllConfigs.client().simplifyFluidUnit.get();
		for (SmartFluidTankBehaviour behaviour : tanks) {
			for (TankSegment tank : behaviour.getTanks()) {
				FluidStack fluidStack = tank.getRenderedFluid();
				if (fluidStack.isEmpty()) {
					continue;
				}
				Lang.text("")
						.add(Lang.fluidName(fluidStack)
								.add(Lang.text(" "))
								.style(class_124.field_1080)
								.add(Lang.text(FluidTextUtil.getUnicodeMillibuckets(fluidStack.getAmount(), unit, simplify))
										.add(unitSuffix)
										.style(class_124.field_1078)))
						.forGoggles(tooltip, 1);
				isEmpty = false;
			}
		}


		if (isEmpty)
			tooltip.remove(0);

		return true;
	}

	@Nullable
	@Override
	public Storage<FluidVariant> getFluidStorage(@Nullable class_2350 face) {
		return fluidCapability;
	}

	@Nullable
	@Override
	public Storage<ItemVariant> getItemStorage(@Nullable class_2350 face) {
		return itemCapability;
	}

	class BasinValueBox extends ValueBoxTransform.Sided {

		@Override
		protected class_243 getSouthLocation() {
			return VecHelper.voxelSpace(8, 12, 16.05);
		}

		@Override
		protected boolean isSideActive(class_2680 state, class_2350 direction) {
			return direction.method_10166()
				.method_10179();
		}

	}
}
