package com.simibubi.create.content.processing.basin;

import java.util.ArrayList;
import java.util.Collections;
import java.util.LinkedList;
import java.util.List;

import javax.annotation.Nonnull;

import com.simibubi.create.AllRecipeTypes;
import com.simibubi.create.content.processing.burner.BlazeBurnerBlock.HeatLevel;
import com.simibubi.create.content.processing.recipe.ProcessingRecipe;
import com.simibubi.create.content.processing.recipe.ProcessingRecipeBuilder;
import com.simibubi.create.content.processing.recipe.ProcessingRecipeBuilder.ProcessingRecipeParams;
import com.simibubi.create.foundation.blockEntity.behaviour.filtering.FilteringBehaviour;
import com.simibubi.create.foundation.blockEntity.behaviour.fluid.SmartFluidTankBehaviour;
import com.simibubi.create.foundation.blockEntity.behaviour.fluid.SmartFluidTankBehaviour.TankSegment;
import com.simibubi.create.foundation.fluid.FluidIngredient;
import com.simibubi.create.foundation.item.SmartInventory;
import com.simibubi.create.foundation.recipe.DummyCraftingContainer;
import com.simibubi.create.foundation.recipe.IRecipeTypeInfo;
import com.simibubi.create.foundation.utility.Iterate;

import io.github.fabricators_of_create.porting_lib.transfer.TransferUtil;
import io.github.fabricators_of_create.porting_lib.transfer.callbacks.TransactionCallback;
import io.github.fabricators_of_create.porting_lib.fluids.FluidStack;
import net.fabricmc.fabric.api.transfer.v1.fluid.FluidVariant;
import net.fabricmc.fabric.api.transfer.v1.item.ItemVariant;
import net.fabricmc.fabric.api.transfer.v1.storage.Storage;
import net.fabricmc.fabric.api.transfer.v1.storage.StorageView;
import net.fabricmc.fabric.api.transfer.v1.transaction.Transaction;
import net.minecraft.class_1799;
import net.minecraft.class_1856;
import net.minecraft.class_1860;
import net.minecraft.class_1937;
import net.minecraft.class_2371;
import net.minecraft.class_310;
import net.minecraft.class_3955;

public class BasinRecipe extends ProcessingRecipe<SmartInventory> {

	public static boolean match(BasinBlockEntity basin, class_1860<?> recipe) {
		FilteringBehaviour filter = basin.getFilter();
		if (filter == null)
			return false;

		boolean filterTest = filter.test(recipe.method_8110(basin.method_10997()
			.method_30349()));
		if (recipe instanceof BasinRecipe) {
			BasinRecipe basinRecipe = (BasinRecipe) recipe;
			if (basinRecipe.getRollableResults()
				.isEmpty()
				&& !basinRecipe.getFluidResults()
					.isEmpty())
				filterTest = filter.test(basinRecipe.getFluidResults()
					.get(0));
		}

		if (!filterTest)
			return false;

		return apply(basin, recipe, true);
	}

	public static boolean apply(BasinBlockEntity basin, class_1860<?> recipe) {
		return apply(basin, recipe, false);
	}

	private static boolean apply(BasinBlockEntity basin, class_1860<?> recipe, boolean test) {
		boolean isBasinRecipe = recipe instanceof BasinRecipe;
		Storage<ItemVariant> availableItems = basin.getItemStorage(null);
		Storage<FluidVariant> availableFluids = basin.getFluidStorage(null);

		if (availableItems == null || availableFluids == null)
			return false;

		HeatLevel heat = BasinBlockEntity.getHeatLevelOf(basin.method_10997()
			.method_8320(basin.method_11016()
				.method_10087(1)));
		if (isBasinRecipe && !((BasinRecipe) recipe).getRequiredHeat()
			.testBlazeBurner(heat))
			return false;

		List<class_1799> recipeOutputItems = new ArrayList<>();
		List<FluidStack> recipeOutputFluids = new ArrayList<>();

		List<class_1856> ingredients = new LinkedList<>(recipe.method_8117());
		List<FluidIngredient> fluidIngredients =
			isBasinRecipe ? ((BasinRecipe) recipe).getFluidIngredients() : Collections.emptyList();

		// fabric: track consumed items to get remainders later
		class_2371<class_1799> consumedItems = class_2371.method_10211();

		try (Transaction t = TransferUtil.getTransaction()) {
			Ingredients: for (class_1856 ingredient : ingredients) {
				for (StorageView<ItemVariant> view : availableItems.nonEmptyViews()) {
					ItemVariant var = view.getResource();
					class_1799 stack = var.toStack();
					if (!ingredient.method_8093(stack)) continue;
					// Catalyst items are never consumed
					class_1799 remainder = stack.getRecipeRemainder();
					if (!remainder.method_7960() && class_1799.method_7984(remainder, stack))
						continue Ingredients;
					long extracted = view.extract(var, 1, t);
					if (extracted == 0) continue;
					consumedItems.add(stack);
					continue Ingredients;
				}
				// something wasn't found
				return false;
			}

			boolean fluidsAffected = false;
			FluidIngredients: for (FluidIngredient fluidIngredient : fluidIngredients) {
			long amountRequired = fluidIngredient.getRequiredAmount();
				for (StorageView<FluidVariant> view : availableFluids.nonEmptyViews()) {
					FluidStack fluidStack = new FluidStack(view);
					if (!fluidIngredient.test(fluidStack)) continue;
					long drainedAmount = Math.min(amountRequired, fluidStack.getAmount());
					if (view.extract(fluidStack.getType(), drainedAmount, t) == drainedAmount) {
						fluidsAffected = true;
						amountRequired -= drainedAmount;
						if (amountRequired != 0) continue;
						continue FluidIngredients;
					}
				}
				// something wasn't found
				return false;
			}

			if (fluidsAffected) {
				TransactionCallback.onSuccess(t, () -> {
					basin.getBehaviour(SmartFluidTankBehaviour.INPUT)
							.forEach(TankSegment::onFluidStackChanged);
					basin.getBehaviour(SmartFluidTankBehaviour.OUTPUT)
							.forEach(TankSegment::onFluidStackChanged);
				});
			}

			if (recipe instanceof BasinRecipe basinRecipe) {
				recipeOutputItems.addAll(basinRecipe.rollResults());
				recipeOutputFluids.addAll(basinRecipe.getFluidResults());
				recipeOutputItems.addAll(basinRecipe.method_8111(basin.getInputInventory()));
			} else {
				recipeOutputItems.add(recipe.method_8110(basin.method_10997()
						.method_30349()));

				if (recipe instanceof class_3955 craftingRecipe) {
					recipeOutputItems.addAll(craftingRecipe.method_8111(new DummyCraftingContainer(consumedItems)));
				}
			}

			// fabric: bad
			recipeOutputItems.removeIf(class_1799::method_7960);

			if (!basin.acceptOutputs(recipeOutputItems, recipeOutputFluids, t))
				return false;

			if (!test)
				t.commit();
			return true;
		}
	}

	public static BasinRecipe convertShapeless(class_1860<?> recipe) {
		BasinRecipe basinRecipe =
			new ProcessingRecipeBuilder<>(BasinRecipe::new, recipe.method_8114()).withItemIngredients(recipe.method_8117())
				.withSingleItemOutput(recipe.method_8110(class_310.method_1551().field_1687.method_30349()))
				.build();
		return basinRecipe;
	}

	protected BasinRecipe(IRecipeTypeInfo type, ProcessingRecipeParams params) {
		super(type, params);
	}

	public BasinRecipe(ProcessingRecipeParams params) {
		this(AllRecipeTypes.BASIN, params);
	}

	@Override
	protected int getMaxInputCount() {
		return 9;
	}

	@Override
	protected int getMaxOutputCount() {
		return 4;
	}

	@Override
	protected int getMaxFluidInputCount() {
		return 2;
	}

	@Override
	protected int getMaxFluidOutputCount() {
		return 2;
	}

	@Override
	protected boolean canRequireHeat() {
		return true;
	}

	@Override
	protected boolean canSpecifyDuration() {
		return true;
	}

	@Override
	public boolean matches(SmartInventory inv, @Nonnull class_1937 worldIn) {
		return false;
	}

}
