package com.simibubi.create.content.processing.burner;

import java.util.List;

import com.simibubi.create.AllBlocks;
import com.simibubi.create.AllItems;
import com.simibubi.create.AllTags.AllItemTags;
import com.simibubi.create.content.fluids.tank.FluidTankBlock;
import com.simibubi.create.content.processing.burner.BlazeBurnerBlock.HeatLevel;
import com.simibubi.create.foundation.blockEntity.SmartBlockEntity;
import com.simibubi.create.foundation.blockEntity.behaviour.BlockEntityBehaviour;
import com.simibubi.create.foundation.utility.AngleHelper;
import com.simibubi.create.foundation.utility.VecHelper;
import com.simibubi.create.foundation.utility.animation.LerpedFloat;
import com.simibubi.create.foundation.utility.animation.LerpedFloat.Chaser;

import io.github.fabricators_of_create.porting_lib.transfer.callbacks.TransactionCallback;
import net.fabricmc.api.EnvType;
import net.fabricmc.api.Environment;
import net.fabricmc.fabric.api.registry.FuelRegistry;
import net.fabricmc.fabric.api.transfer.v1.transaction.TransactionContext;
import net.minecraft.class_1799;
import net.minecraft.class_2338;
import net.minecraft.class_2350;
import net.minecraft.class_2398;
import net.minecraft.class_243;
import net.minecraft.class_2487;
import net.minecraft.class_2591;
import net.minecraft.class_2680;
import net.minecraft.class_310;
import net.minecraft.class_3417;
import net.minecraft.class_3419;
import net.minecraft.class_3532;
import net.minecraft.class_5819;
import net.minecraft.class_746;

public class BlazeBurnerBlockEntity extends SmartBlockEntity {

	public static final int MAX_HEAT_CAPACITY = 10000;
	public static final int INSERTION_THRESHOLD = 500;

	protected FuelType activeFuel;
	protected int remainingBurnTime;
	protected LerpedFloat headAnimation;
	protected LerpedFloat headAngle;
	protected boolean isCreative;
	protected boolean goggles;
	protected boolean hat;

	public BlazeBurnerBlockEntity(class_2591<?> type, class_2338 pos, class_2680 state) {
		super(type, pos, state);
		activeFuel = FuelType.NONE;
		remainingBurnTime = 0;
		headAnimation = LerpedFloat.linear();
		headAngle = LerpedFloat.angular();
		isCreative = false;
		goggles = false;

		headAngle.startWithValue((AngleHelper.horizontalAngle(state.method_28500(BlazeBurnerBlock.field_11177)
			.orElse(class_2350.field_11035)) + 180) % 360);
	}

	public FuelType getActiveFuel() {
		return activeFuel;
	}

	public int getRemainingBurnTime() {
		return remainingBurnTime;
	}

	public boolean isCreative() {
		return isCreative;
	}

	@Override
	public void tick() {
		super.tick();

		if (field_11863.field_9236) {
			tickAnimation();
			if (!isVirtual())
				spawnParticles(getHeatLevelFromBlock(), 1);
			return;
		}

		if (isCreative)
			return;

		if (remainingBurnTime > 0)
			remainingBurnTime--;

		if (activeFuel == FuelType.NORMAL)
			updateBlockState();
		if (remainingBurnTime > 0)
			return;

		if (activeFuel == FuelType.SPECIAL) {
			activeFuel = FuelType.NORMAL;
			remainingBurnTime = MAX_HEAT_CAPACITY / 2;
		} else
			activeFuel = FuelType.NONE;

		updateBlockState();
	}

	@Environment(EnvType.CLIENT)
	private void tickAnimation() {
		boolean active = getHeatLevelFromBlock().isAtLeast(HeatLevel.FADING) && isValidBlockAbove();

		if (!active) {
			float target = 0;
			class_746 player = class_310.method_1551().field_1724;
			if (player != null && !player.method_5767()) {
				double x;
				double z;
				if (isVirtual()) {
					x = -4;
					z = -10;
				} else {
					x = player.method_23317();
					z = player.method_23321();
				}
				double dx = x - (method_11016().method_10263() + 0.5);
				double dz = z - (method_11016().method_10260() + 0.5);
				target = AngleHelper.deg(-class_3532.method_15349(dz, dx)) - 90;
			}
			target = headAngle.getValue() + AngleHelper.getShortestAngleDiff(headAngle.getValue(), target);
			headAngle.chase(target, .25f, Chaser.exp(5));
			headAngle.tickChaser();
		} else {
			headAngle.chase((AngleHelper.horizontalAngle(method_11010().method_28500(BlazeBurnerBlock.field_11177)
				.orElse(class_2350.field_11035)) + 180) % 360, .125f, Chaser.EXP);
			headAngle.tickChaser();
		}

		headAnimation.chase(active ? 1 : 0, .25f, Chaser.exp(.25f));
		headAnimation.tickChaser();
	}

	@Override
	public void addBehaviours(List<BlockEntityBehaviour> behaviours) {}

	@Override
	public void write(class_2487 compound, boolean clientPacket) {
		if (!isCreative) {
			compound.method_10569("fuelLevel", activeFuel.ordinal());
			compound.method_10569("burnTimeRemaining", remainingBurnTime);
		} else
			compound.method_10556("isCreative", true);
		if (goggles)
			compound.method_10556("Goggles", true);
		if (hat)
			compound.method_10556("TrainHat", true);
		super.write(compound, clientPacket);
	}

	@Override
	protected void read(class_2487 compound, boolean clientPacket) {
		activeFuel = FuelType.values()[compound.method_10550("fuelLevel")];
		remainingBurnTime = compound.method_10550("burnTimeRemaining");
		isCreative = compound.method_10577("isCreative");
		goggles = compound.method_10545("Goggles");
		hat = compound.method_10545("TrainHat");
		super.read(compound, clientPacket);
	}

	public BlazeBurnerBlock.HeatLevel getHeatLevelFromBlock() {
		return BlazeBurnerBlock.getHeatLevelOf(method_11010());
	}

	public void updateBlockState() {
		setBlockHeat(getHeatLevel());
	}

	protected void setBlockHeat(HeatLevel heat) {
		HeatLevel inBlockState = getHeatLevelFromBlock();
		if (inBlockState == heat)
			return;
		field_11863.method_8501(field_11867, method_11010().method_11657(BlazeBurnerBlock.HEAT_LEVEL, heat));
		notifyUpdate();
	}

	/**
	 * @return true if the heater updated its burn time and an item should be
	 *         consumed
	 */
	protected boolean tryUpdateFuel(class_1799 itemStack, boolean forceOverflow, TransactionContext ctx) {
		if (isCreative)
			return false;

		FuelType newFuel = FuelType.NONE;
		int newBurnTime;

		if (AllItemTags.BLAZE_BURNER_FUEL_SPECIAL.matches(itemStack)) {
			newBurnTime = 3200;
			newFuel = FuelType.SPECIAL;
		} else {
			Integer fuel = FuelRegistry.INSTANCE.get(itemStack.method_7909());
			newBurnTime = fuel == null ? 0 : fuel;
			if (newBurnTime > 0) {
				newFuel = FuelType.NORMAL;
			} else if (AllItemTags.BLAZE_BURNER_FUEL_REGULAR.matches(itemStack)) {
				newBurnTime = 1600; // Same as coal
				newFuel = FuelType.NORMAL;
			}
		}

		if (newFuel == FuelType.NONE)
			return false;
		if (newFuel.ordinal() < activeFuel.ordinal())
			return false;

		if (newFuel == activeFuel) {
			if (remainingBurnTime <= INSERTION_THRESHOLD) {
				newBurnTime += remainingBurnTime;
			} else if (forceOverflow && newFuel == FuelType.NORMAL) {
				if (remainingBurnTime < MAX_HEAT_CAPACITY) {
					newBurnTime = Math.min(remainingBurnTime + newBurnTime, MAX_HEAT_CAPACITY);
				} else {
					newBurnTime = remainingBurnTime;
				}
			} else {
				return false;
			}
		}

		FuelType finalNewFuel = newFuel;
		int finalNewBurnTime = newBurnTime;
		TransactionCallback.onSuccess(ctx, () -> {
			activeFuel = finalNewFuel;
			remainingBurnTime = finalNewBurnTime;
			if (field_11863.field_9236) {
				spawnParticleBurst(activeFuel == FuelType.SPECIAL);
				return;
			}
			HeatLevel prev = getHeatLevelFromBlock();
			playSound();
			updateBlockState();

			if (prev != getHeatLevelFromBlock())
				field_11863.method_8396(null, field_11867, class_3417.field_14991, class_3419.field_15245,
						.125f + field_11863.field_9229.method_43057() * .125f, 1.15f - field_11863.field_9229.method_43057() * .25f);
		});

		return true;
	}

	protected void applyCreativeFuel() {
		activeFuel = FuelType.NONE;
		remainingBurnTime = 0;
		isCreative = true;

		HeatLevel next = getHeatLevelFromBlock().nextActiveLevel();

		if (field_11863.field_9236) {
			spawnParticleBurst(next.isAtLeast(HeatLevel.SEETHING));
			return;
		}

		playSound();
		if (next == HeatLevel.FADING)
			next = next.nextActiveLevel();
		setBlockHeat(next);
	}

	public boolean isCreativeFuel(class_1799 stack) {
		return AllItems.CREATIVE_BLAZE_CAKE.isIn(stack);
	}

	public boolean isValidBlockAbove() {
		if (isVirtual())
			return false;
		class_2680 blockState = field_11863.method_8320(field_11867.method_10084());
		return AllBlocks.BASIN.has(blockState) || blockState.method_26204() instanceof FluidTankBlock;
	}

	protected void playSound() {
		field_11863.method_8396(null, field_11867, class_3417.field_14970, class_3419.field_15245,
			.125f + field_11863.field_9229.method_43057() * .125f, .75f - field_11863.field_9229.method_43057() * .25f);
	}

	protected HeatLevel getHeatLevel() {
		HeatLevel level = HeatLevel.SMOULDERING;
		switch (activeFuel) {
		case SPECIAL:
			level = HeatLevel.SEETHING;
			break;
		case NORMAL:
			boolean lowPercent = (double) remainingBurnTime / MAX_HEAT_CAPACITY < 0.0125;
			level = lowPercent ? HeatLevel.FADING : HeatLevel.KINDLED;
			break;
		default:
		case NONE:
			break;
		}
		return level;
	}

	protected void spawnParticles(HeatLevel heatLevel, double burstMult) {
		if (field_11863 == null)
			return;
		if (heatLevel == BlazeBurnerBlock.HeatLevel.NONE)
			return;

		class_5819 r = field_11863.method_8409();

		class_243 c = VecHelper.getCenterOf(field_11867);
		class_243 v = c.method_1019(VecHelper.offsetRandomly(class_243.field_1353, r, .125f)
			.method_18805(1, 0, 1));

		if (r.method_43048(4) != 0)
			return;

		boolean empty = field_11863.method_8320(field_11867.method_10084())
			.method_26220(field_11863, field_11867.method_10084())
			.method_1110();

		if (empty || r.method_43048(8) == 0)
			field_11863.method_8406(class_2398.field_11237, v.field_1352, v.field_1351, v.field_1350, 0, 0, 0);

		double yMotion = empty ? .0625f : r.method_43058() * .0125f;
		class_243 v2 = c.method_1019(VecHelper.offsetRandomly(class_243.field_1353, r, .5f)
			.method_18805(1, .25f, 1)
			.method_1029()
			.method_1021((empty ? .25f : .5) + r.method_43058() * .125f))
			.method_1031(0, .5, 0);

		if (heatLevel.isAtLeast(HeatLevel.SEETHING)) {
			field_11863.method_8406(class_2398.field_22246, v2.field_1352, v2.field_1351, v2.field_1350, 0, yMotion, 0);
		} else if (heatLevel.isAtLeast(HeatLevel.FADING)) {
			field_11863.method_8406(class_2398.field_11240, v2.field_1352, v2.field_1351, v2.field_1350, 0, yMotion, 0);
		}
		return;
	}

	public void spawnParticleBurst(boolean soulFlame) {
		class_243 c = VecHelper.getCenterOf(field_11867);
		class_5819 r = field_11863.field_9229;
		for (int i = 0; i < 20; i++) {
			class_243 offset = VecHelper.offsetRandomly(class_243.field_1353, r, .5f)
				.method_18805(1, .25f, 1)
				.method_1029();
			class_243 v = c.method_1019(offset.method_1021(.5 + r.method_43058() * .125f))
				.method_1031(0, .125, 0);
			class_243 m = offset.method_1021(1 / 32f);

			field_11863.method_8406(soulFlame ? class_2398.field_22246 : class_2398.field_11240, v.field_1352, v.field_1351, v.field_1350, m.field_1352, m.field_1351,
				m.field_1350);
		}
	}

	public enum FuelType {
		NONE, NORMAL, SPECIAL
	}

}
