package com.simibubi.create.content.processing.recipe;

import java.util.ArrayList;
import java.util.List;
import java.util.function.Supplier;
import java.util.stream.Collectors;

import javax.annotation.ParametersAreNonnullByDefault;
import net.minecraft.class_1263;
import net.minecraft.class_1799;
import net.minecraft.class_1856;
import net.minecraft.class_1860;
import net.minecraft.class_1865;
import net.minecraft.class_2371;
import net.minecraft.class_2540;
import net.minecraft.class_2960;
import net.minecraft.class_3956;
import net.minecraft.class_5455;
import net.minecraft.class_6328;
import org.slf4j.Logger;

import com.google.gson.JsonObject;
import com.simibubi.create.Create;
import com.simibubi.create.content.processing.recipe.ProcessingRecipeBuilder.ProcessingRecipeParams;
import com.simibubi.create.foundation.fluid.FluidIngredient;
import com.simibubi.create.foundation.recipe.IRecipeTypeInfo;
import io.github.fabricators_of_create.porting_lib.fluids.FluidStack;

@class_6328
@ParametersAreNonnullByDefault
public abstract class ProcessingRecipe<T extends class_1263> implements class_1860<T> {

	protected class_2960 id;
	protected class_2371<class_1856> ingredients;
	protected class_2371<ProcessingOutput> results;
	protected class_2371<FluidIngredient> fluidIngredients;
	protected class_2371<FluidStack> fluidResults;
	protected int processingDuration;
	protected HeatCondition requiredHeat;

	private class_3956<?> type;
	private class_1865<?> serializer;
	private IRecipeTypeInfo typeInfo;
	private Supplier<class_1799> forcedResult;

	public ProcessingRecipe(IRecipeTypeInfo typeInfo, ProcessingRecipeParams params) {
		this.forcedResult = null;
		this.typeInfo = typeInfo;
		this.processingDuration = params.processingDuration;
		this.fluidIngredients = params.fluidIngredients;
		this.fluidResults = params.fluidResults;
		this.serializer = typeInfo.getSerializer();
		this.requiredHeat = params.requiredHeat;
		this.ingredients = params.ingredients;
		this.type = typeInfo.getType();
		this.results = params.results;
		this.id = params.id;

		validate(typeInfo.getId());
	}

	// Recipe type options:

	protected abstract int getMaxInputCount();

	protected abstract int getMaxOutputCount();

	protected boolean canRequireHeat() {
		return false;
	}

	protected boolean canSpecifyDuration() {
		return false;
	}

	protected int getMaxFluidInputCount() {
		return 0;
	}

	protected int getMaxFluidOutputCount() {
		return 0;
	}

	//

	private void validate(class_2960 recipeTypeId) {
		String messageHeader = "Your custom " + recipeTypeId + " recipe (" + id.toString() + ")";
		Logger logger = Create.LOGGER;
		int ingredientCount = ingredients.size();
		int outputCount = results.size();

		if (ingredientCount > getMaxInputCount())
			logger.warn(messageHeader + " has more item inputs (" + ingredientCount + ") than supported ("
				+ getMaxInputCount() + ").");

		if (outputCount > getMaxOutputCount())
			logger.warn(messageHeader + " has more item outputs (" + outputCount + ") than supported ("
				+ getMaxOutputCount() + ").");

		if (processingDuration > 0 && !canSpecifyDuration())
			logger.warn(messageHeader + " specified a duration. Durations have no impact on this type of recipe.");

		if (requiredHeat != HeatCondition.NONE && !canRequireHeat())
			logger.warn(
				messageHeader + " specified a heat condition. Heat conditions have no impact on this type of recipe.");

		ingredientCount = fluidIngredients.size();
		outputCount = fluidResults.size();

		if (ingredientCount > getMaxFluidInputCount())
			logger.warn(messageHeader + " has more fluid inputs (" + ingredientCount + ") than supported ("
				+ getMaxFluidInputCount() + ").");

		if (outputCount > getMaxFluidOutputCount())
			logger.warn(messageHeader + " has more fluid outputs (" + outputCount + ") than supported ("
				+ getMaxFluidOutputCount() + ").");
	}

	@Override
	public class_2371<class_1856> method_8117() {
		return ingredients;
	}

	public class_2371<FluidIngredient> getFluidIngredients() {
		return fluidIngredients;
	}

	public List<ProcessingOutput> getRollableResults() {
		return results;
	}

	public class_2371<FluidStack> getFluidResults() {
		return fluidResults;
	}

	public List<class_1799> getRollableResultsAsItemStacks() {
		return getRollableResults().stream()
			.map(ProcessingOutput::getStack)
			.collect(Collectors.toList());
	}

	public void enforceNextResult(Supplier<class_1799> stack) {
		forcedResult = stack;
	}

	public List<class_1799> rollResults() {
		return rollResults(this.getRollableResults());
	}

	public List<class_1799> rollResults(List<ProcessingOutput> rollableResults) {
		List<class_1799> results = new ArrayList<>();
		for (int i = 0; i < rollableResults.size(); i++) {
			ProcessingOutput output = rollableResults.get(i);
			class_1799 stack = i == 0 && forcedResult != null ? forcedResult.get() : output.rollOutput();
			if (!stack.method_7960())
				results.add(stack);
		}
		return results;
	}

	public int getProcessingDuration() {
		return processingDuration;
	}

	public HeatCondition getRequiredHeat() {
		return requiredHeat;
	}

	// IRecipe<> paperwork

	@Override
	public class_1799 method_8116(T inv, class_5455 registryAccess) {
		return method_8110(registryAccess);
	}

	@Override
	public boolean method_8113(int width, int height) {
		return true;
	}

	@Override
	public class_1799 method_8110(class_5455 registryAccess) {
		return getRollableResults().isEmpty() ? class_1799.field_8037
			: getRollableResults().get(0)
				.getStack();
	}

	@Override
	public boolean method_8118() {
		return true;
	}

	// Processing recipes do not show up in the recipe book
	@Override
	public String method_8112() {
		return "processing";
	}

	@Override
	public class_2960 method_8114() {
		return id;
	}

	@Override
	public class_1865<?> method_8119() {
		return serializer;
	}

	@Override
	public class_3956<?> method_17716() {
		return type;
	}

	public IRecipeTypeInfo getTypeInfo() {
		return typeInfo;
	}

	// Additional Data added by subtypes

	public void readAdditional(JsonObject json) {}

	public void readAdditional(class_2540 buffer) {}

	public void writeAdditional(JsonObject json) {}

	public void writeAdditional(class_2540 buffer) {}

}
