package com.simibubi.create.content.processing.recipe;

import javax.annotation.ParametersAreNonnullByDefault;
import net.minecraft.class_1856;
import net.minecraft.class_1865;
import net.minecraft.class_2371;
import net.minecraft.class_2540;
import net.minecraft.class_2960;
import net.minecraft.class_3518;
import net.minecraft.class_6328;
import com.google.gson.JsonArray;
import com.google.gson.JsonElement;
import com.google.gson.JsonObject;
import com.simibubi.create.content.processing.recipe.ProcessingRecipeBuilder.ProcessingRecipeFactory;
import com.simibubi.create.foundation.fluid.FluidHelper;
import com.simibubi.create.foundation.fluid.FluidIngredient;
import io.github.fabricators_of_create.porting_lib.fluids.FluidStack;

@class_6328
@ParametersAreNonnullByDefault
public class ProcessingRecipeSerializer<T extends ProcessingRecipe<?>> implements class_1865<T> {

	private final ProcessingRecipeFactory<T> factory;

	public ProcessingRecipeSerializer(ProcessingRecipeFactory<T> factory) {
		this.factory = factory;
	}

	protected void writeToJson(JsonObject json, T recipe) {
		JsonArray jsonIngredients = new JsonArray();
		JsonArray jsonOutputs = new JsonArray();

		recipe.ingredients.forEach(i -> jsonIngredients.add(i.method_8089()));
		recipe.fluidIngredients.forEach(i -> jsonIngredients.add(i.serialize()));

		recipe.results.forEach(o -> jsonOutputs.add(o.serialize()));
		recipe.fluidResults.forEach(o -> jsonOutputs.add(FluidHelper.serializeFluidStack(o)));

		json.add("ingredients", jsonIngredients);
		json.add("results", jsonOutputs);

		int processingDuration = recipe.getProcessingDuration();
		if (processingDuration > 0)
			json.addProperty("processingTime", processingDuration);

		HeatCondition requiredHeat = recipe.getRequiredHeat();
		if (requiredHeat != HeatCondition.NONE)
			json.addProperty("heatRequirement", requiredHeat.serialize());

		recipe.writeAdditional(json);
	}

	protected T readFromJson(class_2960 recipeId, JsonObject json) {
		ProcessingRecipeBuilder<T> builder = new ProcessingRecipeBuilder<>(factory, recipeId);
		class_2371<class_1856> ingredients = class_2371.method_10211();
		class_2371<FluidIngredient> fluidIngredients = class_2371.method_10211();
		class_2371<ProcessingOutput> results = class_2371.method_10211();
		class_2371<FluidStack> fluidResults = class_2371.method_10211();

		for (JsonElement je : class_3518.method_15261(json, "ingredients")) {
			if (FluidIngredient.isFluidIngredient(je))
				fluidIngredients.add(FluidIngredient.deserialize(je));
			else
				ingredients.add(class_1856.method_52177(je));
		}

		for (JsonElement je : class_3518.method_15261(json, "results")) {
			JsonObject jsonObject = je.getAsJsonObject();
			if (class_3518.method_15294(jsonObject, "fluid"))
				fluidResults.add(FluidHelper.deserializeFluidStack(jsonObject));
			else
				results.add(ProcessingOutput.deserialize(je));
		}

		builder.withItemIngredients(ingredients)
			.withItemOutputs(results)
			.withFluidIngredients(fluidIngredients)
			.withFluidOutputs(fluidResults);

		if (class_3518.method_15294(json, "processingTime"))
			builder.duration(class_3518.method_15260(json, "processingTime"));
		if (class_3518.method_15294(json, "heatRequirement"))
			builder.requiresHeat(HeatCondition.deserialize(class_3518.method_15265(json, "heatRequirement")));

		T recipe = builder.build();
		recipe.readAdditional(json);
		return recipe;
	}

	protected void writeToBuffer(class_2540 buffer, T recipe) {
		class_2371<class_1856> ingredients = recipe.ingredients;
		class_2371<FluidIngredient> fluidIngredients = recipe.fluidIngredients;
		class_2371<ProcessingOutput> outputs = recipe.results;
		class_2371<FluidStack> fluidOutputs = recipe.fluidResults;

		buffer.method_10804(ingredients.size());
		ingredients.forEach(i -> i.method_8088(buffer));
		buffer.method_10804(fluidIngredients.size());
		fluidIngredients.forEach(i -> i.write(buffer));

		buffer.method_10804(outputs.size());
		outputs.forEach(o -> o.write(buffer));
		buffer.method_10804(fluidOutputs.size());
		fluidOutputs.forEach(o -> o.writeToPacket(buffer));

		buffer.method_10804(recipe.getProcessingDuration());
		buffer.method_10804(recipe.getRequiredHeat()
			.ordinal());

		recipe.writeAdditional(buffer);
	}

	protected T readFromBuffer(class_2960 recipeId, class_2540 buffer) {
		class_2371<class_1856> ingredients = class_2371.method_10211();
		class_2371<FluidIngredient> fluidIngredients = class_2371.method_10211();
		class_2371<ProcessingOutput> results = class_2371.method_10211();
		class_2371<FluidStack> fluidResults = class_2371.method_10211();

		int size = buffer.method_10816();
		for (int i = 0; i < size; i++)
			ingredients.add(class_1856.method_8086(buffer));

		size = buffer.method_10816();
		for (int i = 0; i < size; i++)
			fluidIngredients.add(FluidIngredient.read(buffer));

		size = buffer.method_10816();
		for (int i = 0; i < size; i++)
			results.add(ProcessingOutput.read(buffer));

		size = buffer.method_10816();
		for (int i = 0; i < size; i++)
			fluidResults.add(FluidStack.readFromPacket(buffer));

		T recipe = new ProcessingRecipeBuilder<>(factory, recipeId).withItemIngredients(ingredients)
			.withItemOutputs(results)
			.withFluidIngredients(fluidIngredients)
			.withFluidOutputs(fluidResults)
			.duration(buffer.method_10816())
			.requiresHeat(HeatCondition.values()[buffer.method_10816()])
			.build();
		recipe.readAdditional(buffer);
		return recipe;
	}

	public final void write(JsonObject json, T recipe) {
		writeToJson(json, recipe);
	}

	@Override
	public final T method_8121(class_2960 id, JsonObject json) {
		return readFromJson(id, json);
	}

	@Override
	public final void toNetwork(class_2540 buffer, T recipe) {
		writeToBuffer(buffer, recipe);
	}

	@Override
	public final T method_8122(class_2960 id, class_2540 buffer) {
		return readFromBuffer(id, buffer);
	}

	public ProcessingRecipeFactory<T> getFactory() {
		return factory;
	}

}
