package com.simibubi.create.content.processing.sequenced;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Optional;
import java.util.Set;
import java.util.function.Predicate;
import java.util.stream.Stream;

import com.simibubi.create.AllRecipeTypes;
import com.simibubi.create.Create;
import com.simibubi.create.content.processing.recipe.ProcessingOutput;
import com.simibubi.create.content.processing.recipe.ProcessingRecipe;
import com.simibubi.create.foundation.fluid.FluidIngredient;
import com.simibubi.create.foundation.utility.Components;
import com.simibubi.create.foundation.utility.Lang;
import com.simibubi.create.foundation.utility.Pair;

import io.github.fabricators_of_create.porting_lib.transfer.item.ItemHandlerHelper;
import net.fabricmc.api.EnvType;
import net.fabricmc.api.Environment;
import net.minecraft.class_124;
import net.minecraft.class_1263;
import net.minecraft.class_1799;
import net.minecraft.class_1856;
import net.minecraft.class_1860;
import net.minecraft.class_1865;
import net.minecraft.class_1935;
import net.minecraft.class_1937;
import net.minecraft.class_2371;
import net.minecraft.class_2487;
import net.minecraft.class_2561;
import net.minecraft.class_2960;
import net.minecraft.class_310;
import net.minecraft.class_3956;
import net.minecraft.class_5455;

public class SequencedAssemblyRecipe implements class_1860<class_1263> {

	protected class_2960 id;
	protected SequencedAssemblyRecipeSerializer serializer;

	protected class_1856 ingredient;
	protected class_2371<class_1856> ingredientList;
	protected List<SequencedRecipe<?>> sequence;
	protected int loops;
	protected ProcessingOutput transitionalItem;

	public final List<ProcessingOutput> resultPool;

	public SequencedAssemblyRecipe(class_2960 recipeId, SequencedAssemblyRecipeSerializer serializer) {
		this.id = recipeId;
		this.serializer = serializer;
		sequence = new ArrayList<>();
		resultPool = new ArrayList<>();
		loops = 5;
	}

	public static <C extends class_1263, R extends ProcessingRecipe<C>> Optional<R> getRecipe(class_1937 world, C inv,
		class_3956<R> type, Class<R> recipeClass) {
		return getRecipe(world, inv, type, recipeClass, r -> r.method_8115(inv, world));
	}

	public static <C extends class_1263, R extends ProcessingRecipe<C>> Optional<R> getRecipe(class_1937 world, C inv,
		class_3956<R> type, Class<R> recipeClass, Predicate<? super R> recipeFilter) {
		return getRecipes(world, inv.method_5438(0), type, recipeClass).filter(recipeFilter)
			.findFirst();
	}

	public static <R extends ProcessingRecipe<?>> Optional<R> getRecipe(class_1937 world, class_1799 item,
		class_3956<R> type, Class<R> recipeClass) {
		List<SequencedAssemblyRecipe> all = world.method_8433()
			.method_30027(AllRecipeTypes.SEQUENCED_ASSEMBLY.getType());
		for (SequencedAssemblyRecipe sequencedAssemblyRecipe : all) {
			if (!sequencedAssemblyRecipe.appliesTo(item))
				continue;
			SequencedRecipe<?> nextRecipe = sequencedAssemblyRecipe.getNextRecipe(item);
			ProcessingRecipe<?> recipe = nextRecipe.getRecipe();
			if (recipe.method_17716() != type || !recipeClass.isInstance(recipe))
				continue;
			recipe.enforceNextResult(() -> sequencedAssemblyRecipe.advance(item));
			return Optional.of(recipeClass.cast(recipe));
		}
		return Optional.empty();
	}

	public static <R extends ProcessingRecipe<?>> Stream<R> getRecipes(class_1937 world, class_1799 item,
		class_3956<R> type, Class<R> recipeClass) {
		List<SequencedAssemblyRecipe> all = world.method_8433()
			.method_30027(AllRecipeTypes.SEQUENCED_ASSEMBLY.getType());

		return all.stream()
				.filter(it -> it.appliesTo(item))
				.map(it -> Pair.of(it, it.getNextRecipe(item).getRecipe()))
				.filter(it -> it.getSecond()
						.method_17716() == type && recipeClass.isInstance(it.getSecond()))
				.map(it -> {
					it.getSecond()
							.enforceNextResult(() -> it.getFirst().advance(item));
					return it.getSecond();
				})
				.map(recipeClass::cast);
	}

	private class_1799 advance(class_1799 input) {
		int step = getStep(input);
		if ((step + 1) / sequence.size() >= loops)
			return rollResult();

		class_1799 advancedItem = ItemHandlerHelper.copyStackWithSize(getTransitionalItem(), 1);
		class_2487 itemTag = advancedItem.method_7948();
		class_2487 tag = new class_2487();
		tag.method_10582("id", id.toString());
		tag.method_10569("Step", step + 1);
		tag.method_10548("Progress", (step + 1f) / (sequence.size() * loops));
		itemTag.method_10566("SequencedAssembly", tag);
		advancedItem.method_7980(itemTag);
		return advancedItem;
	}

	public int getLoops() {
		return loops;
	}

	public void addAdditionalIngredientsAndMachines(List<class_1856> list) {
		sequence.forEach(sr -> sr.getAsAssemblyRecipe()
			.addAssemblyIngredients(list));
		Set<class_1935> machines = new HashSet<>();
		sequence.forEach(sr -> sr.getAsAssemblyRecipe()
			.addRequiredMachines(machines));
		machines.stream()
			.map(class_1856::method_8091)
			.forEach(list::add);
	}

	public void addAdditionalFluidIngredients(List<FluidIngredient> list) {
		sequence.forEach(sr -> sr.getAsAssemblyRecipe()
			.addAssemblyFluidIngredients(list));
	}

	private class_1799 rollResult() {
		float totalWeight = 0;
		for (ProcessingOutput entry : resultPool)
			totalWeight += entry.getChance();
		float number = Create.RANDOM.nextFloat() * totalWeight;
		for (ProcessingOutput entry : resultPool) {
			number -= entry.getChance();
			if (number < 0)
				return entry.getStack()
					.method_7972();
		}
		return class_1799.field_8037;
	}

	private boolean appliesTo(class_1799 input) {
		if (ingredient.method_8093(input))
			return true;
		if (input.method_7985()) {
			if (getTransitionalItem().method_7909() == input.method_7909()) {
				if (input.method_7969().method_10545("SequencedAssembly")) {
					class_2487 tag = input.method_7969().method_10562("SequencedAssembly");
					String id = tag.method_10558("id");
					return id.equals(this.id.toString());
				}
			}
		}
		return false;
	}

	private SequencedRecipe<?> getNextRecipe(class_1799 input) {
		return sequence.get(getStep(input) % sequence.size());
	}

	private int getStep(class_1799 input) {
		if (!input.method_7985())
			return 0;
		class_2487 tag = input.method_7969();
		if (!tag.method_10545("SequencedAssembly"))
			return 0;
		int step = tag.method_10562("SequencedAssembly")
			.method_10550("Step");
		return step;
	}

	@Override
	public boolean method_8115(class_1263 inv, class_1937 p_77569_2_) {
		return false;
	}

	@Override
	public class_1799 method_8116(class_1263 inv, class_5455 registryAccess) {
		return class_1799.field_8037;
	}

	@Override
	public boolean method_8113(int p_194133_1_, int p_194133_2_) {
		return false;
	}

	@Override
	public class_1799 method_8110(class_5455 registryAccess) {
		return resultPool.get(0)
			.getStack();
	}

	public float getOutputChance() {
		float totalWeight = 0;
		for (ProcessingOutput entry : resultPool)
			totalWeight += entry.getChance();
		return resultPool.get(0)
			.getChance() / totalWeight;
	}

	@Override
	public class_2960 method_8114() {
		return id;
	}

	@Override
	public class_1865<?> method_8119() {
		return serializer;
	}

	@Override
	public boolean method_8118() {
		return true;
	}

	@Override
	public class_3956<?> method_17716() {
		return AllRecipeTypes.SEQUENCED_ASSEMBLY.getType();
	}

	@Environment(EnvType.CLIENT)
	public static void addToTooltip(class_1799 stack, List<class_2561> tooltip) {
		if (!stack.method_7985() || !stack.method_7969()
			.method_10545("SequencedAssembly"))
			return;
		class_2487 compound = stack.method_7969()
			.method_10562("SequencedAssembly");
		class_2960 resourceLocation = new class_2960(compound.method_10558("id"));
		Optional<? extends class_1860<?>> optionalRecipe = class_310.method_1551().field_1687.method_8433()
			.method_8130(resourceLocation);
		if (!optionalRecipe.isPresent())
			return;
		class_1860<?> recipe = optionalRecipe.get();
		if (!(recipe instanceof SequencedAssemblyRecipe))
			return;

		SequencedAssemblyRecipe sequencedAssemblyRecipe = (SequencedAssemblyRecipe) recipe;
		int length = sequencedAssemblyRecipe.sequence.size();
		int step = sequencedAssemblyRecipe.getStep(stack);
		int total = length * sequencedAssemblyRecipe.loops;
		tooltip.add(Components.immutableEmpty());
		tooltip.add(Lang.translateDirect("recipe.sequenced_assembly")
			.method_27692(class_124.field_1080));
		tooltip.add(Lang.translateDirect("recipe.assembly.progress", step, total)
			.method_27692(class_124.field_1063));

		int remaining = total - step;
		for (int i = 0; i < length; i++) {
			if (i >= remaining)
				break;
			SequencedRecipe<?> sequencedRecipe = sequencedAssemblyRecipe.sequence.get((i + step) % length);
			class_2561 textComponent = sequencedRecipe.getAsAssemblyRecipe()
				.getDescriptionForAssembly();
			if (i == 0)
				tooltip.add(Lang.translateDirect("recipe.assembly.next", textComponent)
					.method_27692(class_124.field_1075));
			else
				tooltip.add(Components.literal("-> ").method_10852(textComponent)
					.method_27692(class_124.field_1062));
		}

	}

	public class_1856 getIngredient() {
		return ingredient;
	}

	@Override
	public class_2371<class_1856> method_8117() {
		if (ingredientList == null) {
			ingredientList = class_2371.method_10211();
			ingredientList.add(ingredient);
		}
		return ingredientList;
	}

	public List<SequencedRecipe<?>> getSequence() {
		return sequence;
	}

	public class_1799 getTransitionalItem() {
		return transitionalItem.getStack();
	}

}
