package com.simibubi.create.content.processing.sequenced;

import java.util.ArrayList;
import java.util.List;
import java.util.function.Consumer;
import java.util.function.UnaryOperator;

import com.google.gson.JsonArray;
import com.google.gson.JsonObject;
import com.simibubi.create.AllRecipeTypes;
import com.simibubi.create.content.processing.recipe.ProcessingOutput;
import com.simibubi.create.content.processing.recipe.ProcessingRecipe;
import com.simibubi.create.content.processing.recipe.ProcessingRecipeBuilder;
import com.simibubi.create.content.processing.recipe.ProcessingRecipeBuilder.ProcessingRecipeFactory;

import net.fabricmc.fabric.api.resource.conditions.v1.ConditionJsonProvider;
import net.fabricmc.fabric.api.resource.conditions.v1.ResourceConditions;
import net.minecraft.class_1792;
import net.minecraft.class_1799;
import net.minecraft.class_1856;
import net.minecraft.class_1865;
import net.minecraft.class_1935;
import net.minecraft.class_2444;
import net.minecraft.class_2960;
import net.minecraft.class_6862;

public class SequencedAssemblyRecipeBuilder {

	private SequencedAssemblyRecipe recipe;
	protected List<ConditionJsonProvider> recipeConditions;

	public SequencedAssemblyRecipeBuilder(class_2960 id) {
		recipeConditions = new ArrayList<>();
		this.recipe = new SequencedAssemblyRecipe(id,
			AllRecipeTypes.SEQUENCED_ASSEMBLY.getSerializer());
	}

	public <T extends ProcessingRecipe<?>> SequencedAssemblyRecipeBuilder addStep(ProcessingRecipeFactory<T> factory,
		UnaryOperator<ProcessingRecipeBuilder<T>> builder) {
		ProcessingRecipeBuilder<T> recipeBuilder =
			new ProcessingRecipeBuilder<>(factory, new class_2960("dummy"));
		class_1792 placeHolder = recipe.getTransitionalItem()
			.method_7909();
		recipe.getSequence()
			.add(new SequencedRecipe<>(builder.apply(recipeBuilder.require(placeHolder)
				.output(placeHolder))
				.build()));
		return this;
	}

	public SequencedAssemblyRecipeBuilder require(class_1935 ingredient) {
		return require(class_1856.method_8091(ingredient));
	}

	public SequencedAssemblyRecipeBuilder require(class_6862<class_1792> tag) {
		return require(class_1856.method_8106(tag));
	}

	public SequencedAssemblyRecipeBuilder require(class_1856 ingredient) {
		recipe.ingredient = ingredient;
		return this;
	}

	public SequencedAssemblyRecipeBuilder transitionTo(class_1935 item) {
		recipe.transitionalItem = new ProcessingOutput(new class_1799(item), 1);
		return this;
	}

	public SequencedAssemblyRecipeBuilder loops(int loops) {
		recipe.loops = loops;
		return this;
	}

	public SequencedAssemblyRecipeBuilder addOutput(class_1935 item, float weight) {
		return addOutput(new class_1799(item), weight);
	}

	public SequencedAssemblyRecipeBuilder addOutput(class_1799 item, float weight) {
		recipe.resultPool.add(new ProcessingOutput(item, weight));
		return this;
	}

	public SequencedAssemblyRecipe build() {
		return recipe;
	}

	public void build(Consumer<class_2444> consumer) {
		consumer.accept(new DataGenResult(build(), recipeConditions));
	}

	public static class DataGenResult implements class_2444 {

		private SequencedAssemblyRecipe recipe;
		private List<ConditionJsonProvider> recipeConditions;
		private class_2960 id;
		private SequencedAssemblyRecipeSerializer serializer;

		public DataGenResult(SequencedAssemblyRecipe recipe, List<ConditionJsonProvider> recipeConditions) {
			this.recipeConditions = recipeConditions;
			this.recipe = recipe;
			this.id = new class_2960(recipe.method_8114().method_12836(),
					AllRecipeTypes.SEQUENCED_ASSEMBLY.getId().method_12832() + "/" + recipe.method_8114().method_12832());
			this.serializer = (SequencedAssemblyRecipeSerializer) recipe.method_8119();
		}

		@Override
		public void method_10416(JsonObject json) {
			serializer.write(json, recipe);
			if (recipeConditions.isEmpty())
				return;

			JsonArray conds = new JsonArray();
			recipeConditions.forEach(c -> conds.add(c.toJson()));
			json.add(ResourceConditions.CONDITIONS_KEY, conds);
		}

		@Override
		public class_2960 method_10417() {
			return id;
		}

		@Override
		public class_1865<?> method_17800() {
			return serializer;
		}

		@Override
		public JsonObject method_10415() {
			return null;
		}

		@Override
		public class_2960 method_10418() {
			return null;
		}

	}

}
