package com.simibubi.create.content.redstone.diodes;

import static com.simibubi.create.content.redstone.diodes.BrassDiodeBlock.POWERING;

import java.util.List;
import net.minecraft.class_1657;
import net.minecraft.class_2312;
import net.minecraft.class_2338;
import net.minecraft.class_2350;
import net.minecraft.class_2487;
import net.minecraft.class_2591;
import net.minecraft.class_2680;
import net.minecraft.class_3532;
import com.simibubi.create.content.equipment.clipboard.ClipboardCloneable;
import com.simibubi.create.foundation.blockEntity.SmartBlockEntity;
import com.simibubi.create.foundation.blockEntity.behaviour.BlockEntityBehaviour;
import com.simibubi.create.foundation.blockEntity.behaviour.scrollValue.ScrollValueBehaviour;
import com.simibubi.create.foundation.utility.Lang;

public abstract class BrassDiodeBlockEntity extends SmartBlockEntity implements ClipboardCloneable {

	protected int state;
	ScrollValueBehaviour maxState;

	public BrassDiodeBlockEntity(class_2591<?> type, class_2338 pos, class_2680 state) {
		super(type, pos, state);
	}

	@Override
	public void addBehaviours(List<BlockEntityBehaviour> behaviours) {
		maxState = new BrassDiodeScrollValueBehaviour(Lang.translateDirect("logistics.redstone_interval"), this,
			new BrassDiodeScrollSlot()).between(2, 60 * 20 * 60);
		maxState.withFormatter(this::format);
		maxState.withCallback(this::onMaxDelayChanged);
		maxState.setValue(2);
		behaviours.add(maxState);
	}

	public float getProgress() {
		int max = Math.max(2, maxState.getValue());
		return class_3532.method_15340(state, 0, max) / (float) max;
	}

	public boolean isIdle() {
		return state == 0;
	}

	@Override
	public void tick() {
		super.tick();
		boolean powered = method_11010().method_11654(class_2312.field_10911);
		boolean powering = method_11010().method_11654(POWERING);
		boolean atMax = state >= maxState.getValue();
		boolean atMin = state <= 0;
		updateState(powered, powering, atMax, atMin);
	}

	protected abstract void updateState(boolean powered, boolean powering, boolean atMax, boolean atMin);

	private void onMaxDelayChanged(int newMax) {
		state = class_3532.method_15340(state, 0, newMax);
		sendData();
	}

	@Override
	protected void read(class_2487 compound, boolean clientPacket) {
		state = compound.method_10550("State");
		super.read(compound, clientPacket);
	}

	@Override
	public void write(class_2487 compound, boolean clientPacket) {
		compound.method_10569("State", state);
		super.write(compound, clientPacket);
	}

	private String format(int value) {
		if (value < 60)
			return value + "t";
		if (value < 20 * 60)
			return (value / 20) + "s";
		return (value / 20 / 60) + "m";
	}

	@Override
	public String getClipboardKey() {
		return "Block";
	}
	
	@Override
	public boolean readFromClipboard(class_2487 tag, class_1657 player, class_2350 side, boolean simulate) {
		if (!tag.method_10545("Inverted"))
			return false;
		if (simulate)
			return true;
		class_2680 blockState = method_11010();
		if (blockState.method_11654(BrassDiodeBlock.INVERTED) != tag.method_10577("Inverted"))
			field_11863.method_8501(field_11867, blockState.method_28493(BrassDiodeBlock.INVERTED));
		return true;
	}
	
	@Override
	public boolean writeToClipboard(class_2487 tag, class_2350 side) {
		tag.method_10556("Inverted", method_11010().method_28500(BrassDiodeBlock.INVERTED)
			.orElse(false));
		return true;
	}
	
}
