package com.simibubi.create.content.redstone.displayLink;

import java.util.List;
import net.minecraft.class_2338;
import net.minecraft.class_2350;
import net.minecraft.class_2487;
import net.minecraft.class_2512;
import net.minecraft.class_2591;
import net.minecraft.class_2680;
import net.minecraft.class_2960;
import com.simibubi.create.compat.computercraft.AbstractComputerBehaviour;
import com.simibubi.create.compat.computercraft.ComputerCraftProxy;
import com.simibubi.create.content.redstone.displayLink.source.DisplaySource;
import com.simibubi.create.content.redstone.displayLink.target.DisplayTarget;
import com.simibubi.create.foundation.advancement.AllAdvancements;
import com.simibubi.create.foundation.blockEntity.SmartBlockEntity;
import com.simibubi.create.foundation.blockEntity.behaviour.BlockEntityBehaviour;
import com.simibubi.create.foundation.utility.NBTHelper;
import com.simibubi.create.foundation.utility.animation.LerpedFloat;
import com.simibubi.create.foundation.utility.animation.LerpedFloat.Chaser;

public class DisplayLinkBlockEntity extends SmartBlockEntity {

	protected class_2338 targetOffset;

	public DisplaySource activeSource;
	private class_2487 sourceConfig;

	public DisplayTarget activeTarget;
	public int targetLine;

	public LerpedFloat glow;
	private boolean sendPulse;

	public int refreshTicks;
	public AbstractComputerBehaviour computerBehaviour;

	public DisplayLinkBlockEntity(class_2591<?> type, class_2338 pos, class_2680 state) {
		super(type, pos, state);
		targetOffset = class_2338.field_10980;
		sourceConfig = new class_2487();
		targetLine = 0;
		glow = LerpedFloat.linear()
			.startWithValue(0);
		glow.chase(0, 0.5f, Chaser.EXP);
	}

	@Override
	public void addBehaviours(List<BlockEntityBehaviour> behaviours) {
		behaviours.add(computerBehaviour = ComputerCraftProxy.behaviour(this));
		registerAwardables(behaviours, AllAdvancements.DISPLAY_LINK, AllAdvancements.DISPLAY_BOARD);
	}

	@Override
	public void tick() {
		super.tick();

		if (isVirtual()) {
			glow.tickChaser();
			return;
		}

		if (activeSource == null)
			return;
		if (field_11863.field_9236) {
			glow.tickChaser();
			return;
		}

		refreshTicks++;
		if (refreshTicks < activeSource.getPassiveRefreshTicks() || !activeSource.shouldPassiveReset())
			return;
		tickSource();
	}

	public void tickSource() {
		refreshTicks = 0;
		if (method_11010().method_28500(DisplayLinkBlock.POWERED)
			.orElse(true))
			return;
		if (!field_11863.field_9236)
			updateGatheredData();
	}

	public void onNoLongerPowered() {
		if (activeSource == null)
			return;
		refreshTicks = 0;
		activeSource.onSignalReset(new DisplayLinkContext(field_11863, this));
		updateGatheredData();
	}

	public void updateGatheredData() {
		class_2338 sourcePosition = getSourcePosition();
		class_2338 targetPosition = getTargetPosition();

		if (!field_11863.method_8477(targetPosition) || !field_11863.method_8477(sourcePosition))
			return;

		DisplayTarget target = AllDisplayBehaviours.targetOf(field_11863, targetPosition);
		List<DisplaySource> sources = AllDisplayBehaviours.sourcesOf(field_11863, sourcePosition);
		boolean notify = false;

		if (activeTarget != target) {
			activeTarget = target;
			notify = true;
		}

		if (activeSource != null && !sources.contains(activeSource)) {
			activeSource = null;
			sourceConfig = new class_2487();
			notify = true;
		}

		if (notify)
			notifyUpdate();
		if (activeSource == null || activeTarget == null)
			return;

		DisplayLinkContext context = new DisplayLinkContext(field_11863, this);
		activeSource.transferData(context, activeTarget, targetLine);
		sendPulse = true;
		sendData();

		award(AllAdvancements.DISPLAY_LINK);
	}

	@Override
	public void writeSafe(class_2487 tag) {
		super.writeSafe(tag);
		writeGatheredData(tag);
	}

	@Override
	protected void write(class_2487 tag, boolean clientPacket) {
		super.write(tag, clientPacket);
		writeGatheredData(tag);
		if (clientPacket && activeTarget != null)
			tag.method_10582("TargetType", activeTarget.id.toString());
		if (clientPacket && sendPulse) {
			sendPulse = false;
			NBTHelper.putMarker(tag, "Pulse");
		}
	}

	private void writeGatheredData(class_2487 tag) {
		tag.method_10566("TargetOffset", class_2512.method_10692(targetOffset));
		tag.method_10569("TargetLine", targetLine);

		if (activeSource != null) {
			class_2487 data = sourceConfig.method_10553();
			data.method_10582("Id", activeSource.id.toString());
			tag.method_10566("Source", data);
		}
	}

	@Override
	protected void read(class_2487 tag, boolean clientPacket) {
		super.read(tag, clientPacket);
		targetOffset = class_2512.method_10691(tag.method_10562("TargetOffset"));
		targetLine = tag.method_10550("TargetLine");

		if (clientPacket && tag.method_10545("TargetType"))
			activeTarget = AllDisplayBehaviours.getTarget(new class_2960(tag.method_10558("TargetType")));
		if (clientPacket && tag.method_10545("Pulse"))
			glow.setValue(2);

		if (!tag.method_10545("Source"))
			return;

		class_2487 data = tag.method_10562("Source");
		activeSource = AllDisplayBehaviours.getSource(new class_2960(data.method_10558("Id")));
		sourceConfig = new class_2487();
		if (activeSource != null)
			sourceConfig = data.method_10553();
	}

	public void target(class_2338 targetPosition) {
		this.targetOffset = targetPosition.method_10059(field_11867);
	}

	public class_2338 getSourcePosition() {
		return field_11867.method_10093(getDirection());
	}

	public class_2487 getSourceConfig() {
		return sourceConfig;
	}

	public void setSourceConfig(class_2487 sourceConfig) {
		this.sourceConfig = sourceConfig;
	}

	public class_2350 getDirection() {
		return method_11010().method_28500(DisplayLinkBlock.field_10927)
			.orElse(class_2350.field_11036)
			.method_10153();
	}

	public class_2338 getTargetPosition() {
		return field_11867.method_10081(targetOffset);
	}

}
