package com.simibubi.create.content.redstone.displayLink.source;

import java.util.List;
import java.util.Optional;
import java.util.stream.Stream;

import com.simibubi.create.content.fluids.tank.BoilerData;
import com.simibubi.create.content.fluids.tank.FluidTankBlockEntity;
import com.simibubi.create.content.redstone.displayLink.DisplayLinkContext;
import com.simibubi.create.content.redstone.displayLink.target.DisplayTargetStats;
import com.simibubi.create.content.trains.display.FlapDisplayBlockEntity;
import com.simibubi.create.content.trains.display.FlapDisplayLayout;
import com.simibubi.create.content.trains.display.FlapDisplaySection;
import com.simibubi.create.foundation.utility.Components;
import com.simibubi.create.foundation.utility.Lang;

import joptsimple.internal.Strings;
import net.minecraft.class_124;
import net.minecraft.class_2586;
import net.minecraft.class_3722;
import net.minecraft.class_5250;

public class BoilerDisplaySource extends DisplaySource {

	public static final List<class_5250> notEnoughSpaceSingle =
		List.of(Lang.translateDirect("display_source.boiler.not_enough_space")
			.method_10852(Lang.translateDirect("display_source.boiler.for_boiler_status")));

	public static final List<class_5250> notEnoughSpaceDouble =
		List.of(Lang.translateDirect("display_source.boiler.not_enough_space"),
			Lang.translateDirect("display_source.boiler.for_boiler_status"));

	public static final List<List<class_5250>> notEnoughSpaceFlap =
		List.of(List.of(Lang.translateDirect("display_source.boiler.not_enough_space")),
			List.of(Lang.translateDirect("display_source.boiler.for_boiler_status")));

	@Override
	public List<class_5250> provideText(DisplayLinkContext context, DisplayTargetStats stats) {
		if (stats.maxRows() < 2)
			return notEnoughSpaceSingle;
		else if (stats.maxRows() < 4)
			return notEnoughSpaceDouble;

		boolean isBook = context.getTargetBlockEntity() instanceof class_3722;

		if (isBook) {
			Stream<class_5250> componentList = getComponents(context, false).map(components -> {
				Optional<class_5250> reduce = components.stream()
					.reduce(class_5250::method_10852);
				return reduce.orElse(EMPTY_LINE);
			});

			return List.of(componentList.reduce((comp1, comp2) -> comp1.method_10852(Components.literal("\n"))
				.method_10852(comp2))
				.orElse(EMPTY_LINE));
		}

		return getComponents(context, false).map(components -> {
			Optional<class_5250> reduce = components.stream()
				.reduce(class_5250::method_10852);
			return reduce.orElse(EMPTY_LINE);
		})
			.toList();
	}

	@Override
	public List<List<class_5250>> provideFlapDisplayText(DisplayLinkContext context, DisplayTargetStats stats) {
		if (stats.maxRows() < 4) {
			context.flapDisplayContext = Boolean.FALSE;
			return notEnoughSpaceFlap;
		}

		List<List<class_5250>> components = getComponents(context, true).toList();

		if (stats.maxColumns() * FlapDisplaySection.MONOSPACE < 6 * FlapDisplaySection.MONOSPACE + components.get(1)
			.get(1)
			.getString()
			.length() * FlapDisplaySection.WIDE_MONOSPACE) {
			context.flapDisplayContext = Boolean.FALSE;
			return notEnoughSpaceFlap;
		}

		return components;
	}

	@Override
	public void loadFlapDisplayLayout(DisplayLinkContext context, FlapDisplayBlockEntity flapDisplay,
		FlapDisplayLayout layout, int lineIndex) {
		if (lineIndex == 0 || context.flapDisplayContext instanceof Boolean b && !b) {
			if (layout.isLayout("Default"))
				return;

			layout.loadDefault(flapDisplay.getMaxCharCount());
			return;
		}

		String layoutKey = "Boiler";
		if (layout.isLayout(layoutKey))
			return;

		int labelLength = (int) (labelWidth() * FlapDisplaySection.MONOSPACE);
		float maxSpace = flapDisplay.getMaxCharCount(1) * FlapDisplaySection.MONOSPACE;
		FlapDisplaySection label = new FlapDisplaySection(labelLength, "alphabet", false, true);
		FlapDisplaySection symbols = new FlapDisplaySection(maxSpace - labelLength, "pixel", false, false).wideFlaps();

		layout.configure(layoutKey, List.of(label, symbols));
	}

	private Stream<List<class_5250>> getComponents(DisplayLinkContext context, boolean forFlapDisplay) {
		class_2586 sourceBE = context.getSourceBlockEntity();
		if (!(sourceBE instanceof FluidTankBlockEntity tankBlockEntity))
			return Stream.of(EMPTY);

		tankBlockEntity = tankBlockEntity.getControllerBE();
		if (tankBlockEntity == null)
			return Stream.of(EMPTY);

		BoilerData boiler = tankBlockEntity.boiler;

		int totalTankSize = tankBlockEntity.getTotalTankSize();

		boiler.calcMinMaxForSize(totalTankSize);

		String label = forFlapDisplay ? "boiler.status" : "boiler.status_short";
		class_5250 size = labelOf(forFlapDisplay ? "size" : "");
		class_5250 water = labelOf(forFlapDisplay ? "water" : "");
		class_5250 heat = labelOf(forFlapDisplay ? "heat" : "");

		int lw = labelWidth();
		if (forFlapDisplay) {
			size = Components.literal(Strings.repeat(' ', lw - labelWidthOf("size"))).method_10852(size);
			water = Components.literal(Strings.repeat(' ', lw - labelWidthOf("water"))).method_10852(water);
			heat = Components.literal(Strings.repeat(' ', lw - labelWidthOf("heat"))).method_10852(heat);
		}

		return Stream.of(List.of(Lang.translateDirect(label, boiler.getHeatLevelTextComponent())),
			List.of(size, boiler.getSizeComponent(!forFlapDisplay, forFlapDisplay, class_124.field_1070)),
			List.of(water, boiler.getWaterComponent(!forFlapDisplay, forFlapDisplay, class_124.field_1070)),
			List.of(heat, boiler.getHeatComponent(!forFlapDisplay, forFlapDisplay, class_124.field_1070)));
	}

	private int labelWidth() {
		return Math.max(labelWidthOf("water"), Math.max(labelWidthOf("size"), labelWidthOf("heat")));
	}

	private int labelWidthOf(String label) {
		return labelOf(label).getString()
			.length();
	}

	private class_5250 labelOf(String label) {
		if (label.isBlank())
			return Components.empty();
		return Lang.translateDirect("boiler." + label);
	}

	@Override
	protected String getTranslationKey() {
		return "boiler_status";
	}
}
