package com.simibubi.create.content.redstone.displayLink.source;

import com.simibubi.create.content.redstone.displayLink.DisplayLinkBlock;
import com.simibubi.create.content.redstone.displayLink.DisplayLinkBlockEntity;
import com.simibubi.create.content.redstone.displayLink.DisplayLinkContext;
import com.simibubi.create.content.redstone.displayLink.target.DisplayTargetStats;
import com.simibubi.create.foundation.gui.ModularGuiLineBuilder;
import com.simibubi.create.foundation.utility.Lang;
import net.fabricmc.api.EnvType;
import net.fabricmc.api.Environment;
import net.minecraft.class_2487;
import net.minecraft.class_2494;
import net.minecraft.class_2499;
import net.minecraft.class_2520;
import net.minecraft.class_5250;

public class ItemThroughputDisplaySource extends AccumulatedItemCountDisplaySource {

	static final int POOL_SIZE = 10;

	@Override
	protected class_5250 provideLine(DisplayLinkContext context, DisplayTargetStats stats) {
		class_2487 conf = context.sourceConfig();
		if (conf.method_10545("Inactive"))
			return ZERO.method_27661();

		double interval = 20 * Math.pow(60, conf.method_10550("Interval"));
		double rate = conf.method_10583("Rate") * interval;

		if (rate > 0) {
			long previousTime = conf.method_10537("LastReceived");
			long gameTime = context.blockEntity()
				.method_10997()
				.method_8510();
			int diff = (int) (gameTime - previousTime);
			if (diff > 0) {
				// Too long since last item
				int lastAmount = conf.method_10550("LastReceivedAmount");
				double timeBetweenStacks = lastAmount / rate;
				if (diff > timeBetweenStacks * 2)
					conf.method_10556("Inactive", true);
			}
		}

		return Lang.number(rate)
			.component();
	}

	public void itemReceived(DisplayLinkBlockEntity be, int amount) {
		if (be.method_11010()
			.method_28500(DisplayLinkBlock.POWERED)
			.orElse(true))
			return;

		class_2487 conf = be.getSourceConfig();
		long gameTime = be.method_10997()
			.method_8510();

		if (!conf.method_10545("LastReceived")) {
			conf.method_10544("LastReceived", gameTime);
			return;
		}

		long previousTime = conf.method_10537("LastReceived");
		class_2499 rates = conf.method_10554("PrevRates", class_2520.field_33255);

		if (rates.size() != POOL_SIZE) {
			rates = new class_2499();
			for (int i = 0; i < POOL_SIZE; i++)
				rates.add(class_2494.method_23244(-1));
		}

		int poolIndex = conf.method_10550("Index") % POOL_SIZE;
		rates.method_10606(poolIndex, class_2494.method_23244((float) (amount / (double) (gameTime - previousTime))));

		float rate = 0;
		int validIntervals = 0;
		for (int i = 0; i < POOL_SIZE; i++) {
			float pooledRate = rates.method_10604(i);
			if (pooledRate >= 0) {
				rate += pooledRate;
				validIntervals++;
			}
		}

		conf.method_10551("Rate");
		if (validIntervals > 0) {
			rate /= validIntervals;
			conf.method_10548("Rate", rate);
		}

		conf.method_10551("Inactive");
		conf.method_10569("LastReceivedAmount", amount);
		conf.method_10544("LastReceived", gameTime);
		conf.method_10569("Index", poolIndex + 1);
		conf.method_10566("PrevRates", rates);
		be.updateGatheredData();
	}

	@Override
	@Environment(EnvType.CLIENT)
	public void initConfigurationWidgets(DisplayLinkContext context, ModularGuiLineBuilder builder,
		boolean isFirstLine) {
		super.initConfigurationWidgets(context, builder, isFirstLine);
		if (isFirstLine)
			return;

		builder.addSelectionScrollInput(0, 80, (si, l) -> {
			si.forOptions(Lang.translatedOptions("display_source.item_throughput.interval", "second", "minute", "hour"))
				.titled(Lang.translateDirect("display_source.item_throughput.interval"));
		}, "Interval");
	}

	@Override
	protected String getTranslationKey() {
		return "item_throughput";
	}

}
