package com.simibubi.create.content.redstone.link.controller;

import java.util.List;
import java.util.UUID;

import com.simibubi.create.AllSoundEvents;
import com.simibubi.create.foundation.blockEntity.SmartBlockEntity;
import com.simibubi.create.foundation.blockEntity.behaviour.BlockEntityBehaviour;

import com.simibubi.create.foundation.utility.fabric.ReachUtil;

import io.github.fabricators_of_create.porting_lib.util.EnvExecutor;
import net.fabricmc.api.EnvType;
import net.fabricmc.api.Environment;
import net.minecraft.class_1268;
import net.minecraft.class_1297;
import net.minecraft.class_1542;
import net.minecraft.class_1657;
import net.minecraft.class_1799;
import net.minecraft.class_1937;
import net.minecraft.class_2338;
import net.minecraft.class_2350;
import net.minecraft.class_243;
import net.minecraft.class_2487;
import net.minecraft.class_2591;
import net.minecraft.class_2680;
import net.minecraft.class_310;
import net.minecraft.class_3218;

public class LecternControllerBlockEntity extends SmartBlockEntity {

	private class_1799 controller;
	private UUID user;
	private UUID prevUser;	// used only on client
	private boolean deactivatedThisTick;	// used only on server

	public LecternControllerBlockEntity(class_2591<?> type, class_2338 pos, class_2680 state) {
		super(type, pos, state);
	}

	@Override
	public void addBehaviours(List<BlockEntityBehaviour> behaviours) { }

	@Override
	protected void write(class_2487 compound, boolean clientPacket) {
		super.write(compound, clientPacket);
		compound.method_10566("Controller", controller.method_7953(new class_2487()));
		if (user != null)
			compound.method_25927("User", user);
	}

	@Override
	public void writeSafe(class_2487 compound) {
		super.writeSafe(compound);
		compound.method_10566("Controller", controller.method_7953(new class_2487()));
	}

	@Override
	protected void read(class_2487 compound, boolean clientPacket) {
		super.read(compound, clientPacket);
		controller = class_1799.method_7915(compound.method_10562("Controller"));
		user = compound.method_25928("User") ? compound.method_25926("User") : null;
	}

	public class_1799 getController() {
		return controller;
	}

	public boolean hasUser() { return user != null; }

	public boolean isUsedBy(class_1657 player) {
		return hasUser() && user.equals(player.method_5667());
	}

	public void tryStartUsing(class_1657 player) {
		if (!deactivatedThisTick && !hasUser() && !playerIsUsingLectern(player) && playerInRange(player, field_11863, field_11867))
			startUsing(player);
	}

	public void tryStopUsing(class_1657 player) {
		if (isUsedBy(player))
			stopUsing(player);
	}

	private void startUsing(class_1657 player) {
		user = player.method_5667();
		player.getCustomData().method_10556("IsUsingLecternController", true);
		sendData();
	}

	private void stopUsing(class_1657 player) {
		user = null;
		if (player != null)
			player.getCustomData().method_10551("IsUsingLecternController");
		deactivatedThisTick = true;
		sendData();
	}

	public static boolean playerIsUsingLectern(class_1657 player) {
		return player.getCustomData().method_10545("IsUsingLecternController");
	}

	@Override
	public void tick() {
		super.tick();

		if (field_11863.field_9236) {
			EnvExecutor.runWhenOn(EnvType.CLIENT, () -> this::tryToggleActive);
			prevUser = user;
		}

		if (!field_11863.field_9236) {
			deactivatedThisTick = false;

			if (!(field_11863 instanceof class_3218))
				return;
			if (user == null)
				return;

			class_1297 entity = ((class_3218) field_11863).method_14190(user);
			if (!(entity instanceof class_1657)) {
				stopUsing(null);
				return;
			}

			class_1657 player = (class_1657) entity;
			if (!playerInRange(player, field_11863, field_11867) || !playerIsUsingLectern(player))
				stopUsing(player);
		}
	}

	@Environment(EnvType.CLIENT)
	private void tryToggleActive() {
		if (user == null && class_310.method_1551().field_1724.method_5667().equals(prevUser)) {
			LinkedControllerClientHandler.deactivateInLectern();
		} else if (prevUser == null && class_310.method_1551().field_1724.method_5667().equals(user)) {
			LinkedControllerClientHandler.activateInLectern(field_11867);
		}
	}

	public void setController(class_1799 newController) {
		controller = newController;
		if (newController != null) {
			AllSoundEvents.CONTROLLER_PUT.playOnServer(field_11863, field_11867);
		}
	}

	public void swapControllers(class_1799 stack, class_1657 player, class_1268 hand, class_2680 state) {
		class_1799 newController = stack.method_7972();
		stack.method_7939(0);
		if (player.method_5998(hand).method_7960()) {
			player.method_6122(hand, controller);
		} else {
			dropController(state);
		}
		setController(newController);
	}

	public void dropController(class_2680 state) {
		class_1297 playerEntity = ((class_3218) field_11863).method_14190(user);
		if (playerEntity instanceof class_1657)
			stopUsing((class_1657) playerEntity);

		class_2350 dir = state.method_11654(LecternControllerBlock.field_16404);
		double x = field_11867.method_10263() + 0.5 + 0.25*dir.method_10148();
		double y = field_11867.method_10264() + 1;
		double z = field_11867.method_10260() + 0.5 + 0.25*dir.method_10165();
		class_1542 itementity = new class_1542(field_11863, x, y, z, controller.method_7972());
		itementity.method_6988();
		field_11863.method_8649(itementity);
		controller = null;
	}

	public static boolean playerInRange(class_1657 player, class_1937 world, class_2338 pos) {
		//double modifier = world.isRemote ? 0 : 1.0;
		double reach = 0.4* ReachUtil.reach(player);// + modifier;
		return player.method_5707(class_243.method_24953(pos)) < reach*reach;
	}

}
