package com.simibubi.create.content.redstone.nixieTube;

import java.lang.ref.WeakReference;
import java.util.List;
import java.util.Optional;
import net.minecraft.class_2338;
import net.minecraft.class_2350;
import net.minecraft.class_2487;
import net.minecraft.class_2586;
import net.minecraft.class_2591;
import net.minecraft.class_2680;
import net.minecraft.class_5250;
import com.simibubi.create.content.redstone.displayLink.DisplayLinkBlock;
import com.simibubi.create.content.trains.signal.SignalBlockEntity;
import com.simibubi.create.content.trains.signal.SignalBlockEntity.SignalState;
import com.simibubi.create.foundation.blockEntity.SmartBlockEntity;
import com.simibubi.create.foundation.blockEntity.behaviour.BlockEntityBehaviour;
import com.simibubi.create.foundation.utility.Components;
import com.simibubi.create.foundation.utility.Couple;
import com.simibubi.create.foundation.utility.DynamicComponent;

public class NixieTubeBlockEntity extends SmartBlockEntity {

	private static final Couple<String> EMPTY = Couple.create("", "");

	private int redstoneStrength;
	private Optional<DynamicComponent> customText;
	private int nixieIndex;
	private Couple<String> displayedStrings;

	private WeakReference<SignalBlockEntity> cachedSignalTE;
	public SignalState signalState;

	public NixieTubeBlockEntity(class_2591<?> type, class_2338 pos, class_2680 state) {
		super(type, pos, state);
		customText = Optional.empty();
		redstoneStrength = 0;
		cachedSignalTE = new WeakReference<>(null);
	}

	@Override
	public void tick() {
		super.tick();
		if (!field_11863.field_9236)
			return;

		signalState = null;
		SignalBlockEntity signalBlockEntity = cachedSignalTE.get();

		if (signalBlockEntity == null || signalBlockEntity.method_11015()) {
			class_2350 facing = NixieTubeBlock.getFacing(method_11010());
			class_2586 blockEntity = field_11863.method_8321(field_11867.method_10093(facing.method_10153()));
			if (blockEntity instanceof SignalBlockEntity signal) {
				signalState = signal.getState();
				cachedSignalTE = new WeakReference<>(signal);
			}
			return;
		}

		signalState = signalBlockEntity.getState();
	}

	@Override
	public void initialize() {
		if (field_11863.field_9236)
			updateDisplayedStrings();
	}

	//

	public boolean reactsToRedstone() {
		return customText.isEmpty();
	}

	public Couple<String> getDisplayedStrings() {
		if (displayedStrings == null)
			return EMPTY;
		return displayedStrings;
	}

	public class_5250 getFullText() {
		return customText.map(DynamicComponent::get)
			.orElse(Components.literal("" + redstoneStrength));
	}

	public void updateRedstoneStrength(int signalStrength) {
		clearCustomText();
		redstoneStrength = signalStrength;
		DisplayLinkBlock.notifyGatherers(field_11863, field_11867);
		notifyUpdate();
	}

	public void displayCustomText(String tagElement, int nixiePositionInRow) {
		if (tagElement == null)
			return;
		if (customText.filter(d -> d.sameAs(tagElement))
			.isPresent())
			return;

		DynamicComponent component = customText.orElseGet(DynamicComponent::new);
		component.displayCustomText(field_11863, field_11867, tagElement);
		customText = Optional.of(component);
		nixieIndex = nixiePositionInRow;
		DisplayLinkBlock.notifyGatherers(field_11863, field_11867);
		notifyUpdate();
	}

	public void updateDisplayedStrings() {
		if (signalState != null)
			return;
		customText.map(DynamicComponent::resolve)
			.ifPresentOrElse(
				fullText -> displayedStrings =
					Couple.create(charOrEmpty(fullText, nixieIndex * 2), charOrEmpty(fullText, nixieIndex * 2 + 1)),
				() -> displayedStrings =
					Couple.create(redstoneStrength < 10 ? "0" : "1", String.valueOf(redstoneStrength % 10)));
	}

	public void clearCustomText() {
		nixieIndex = 0;
		customText = Optional.empty();
	}

	public int getRedstoneStrength() {
		return redstoneStrength;
	}

	//

	@Override
	protected void read(class_2487 nbt, boolean clientPacket) {
		super.read(nbt, clientPacket);

		if (nbt.method_10545("CustomText")) {
			DynamicComponent component = customText.orElseGet(DynamicComponent::new);
			component.read(field_11863, field_11867, nbt);

			if (component.isValid()) {
				customText = Optional.of(component);
				nixieIndex = nbt.method_10550("CustomTextIndex");
			} else {
				customText = Optional.empty();
				nixieIndex = 0;
			}
		}

		if (customText.isEmpty())
			redstoneStrength = nbt.method_10550("RedstoneStrength");
		if (clientPacket)
			updateDisplayedStrings();
	}

	@Override
	protected void write(class_2487 nbt, boolean clientPacket) {
		super.write(nbt, clientPacket);

		if (customText.isPresent()) {
			nbt.method_10569("CustomTextIndex", nixieIndex);
			customText.get()
				.write(nbt);
		} else
			nbt.method_10569("RedstoneStrength", redstoneStrength);
	}

	private String charOrEmpty(String string, int index) {
		return string.length() <= index ? " " : string.substring(index, index + 1);
	}

	@Override
	public void addBehaviours(List<BlockEntityBehaviour> behaviours) {}

}
