package com.simibubi.create.content.redstone.smartObserver;

import java.util.List;
import net.minecraft.class_1799;
import net.minecraft.class_2248;
import net.minecraft.class_2338;
import net.minecraft.class_2350;
import net.minecraft.class_2487;
import net.minecraft.class_2591;
import net.minecraft.class_2680;
import com.simibubi.create.content.fluids.FluidTransportBehaviour;
import com.simibubi.create.content.fluids.PipeConnection.Flow;
import com.simibubi.create.content.kinetics.belt.behaviour.TransportedItemStackHandlerBehaviour;
import com.simibubi.create.content.kinetics.belt.behaviour.TransportedItemStackHandlerBehaviour.TransportedResult;
import com.simibubi.create.content.redstone.DirectedDirectionalBlock;
import com.simibubi.create.content.redstone.FilteredDetectorFilterSlot;
import com.simibubi.create.foundation.blockEntity.SmartBlockEntity;
import com.simibubi.create.foundation.blockEntity.behaviour.BlockEntityBehaviour;
import com.simibubi.create.foundation.blockEntity.behaviour.filtering.FilteringBehaviour;
import com.simibubi.create.foundation.blockEntity.behaviour.inventory.CapManipulationBehaviourBase.InterfaceProvider;
import com.simibubi.create.foundation.blockEntity.behaviour.inventory.InvManipulationBehaviour;
import com.simibubi.create.foundation.blockEntity.behaviour.inventory.TankManipulationBehaviour;
import com.simibubi.create.foundation.blockEntity.behaviour.inventory.VersionedInventoryTrackerBehaviour;
import com.simibubi.create.foundation.utility.BlockFace;
import com.simibubi.create.foundation.utility.Iterate;

public class SmartObserverBlockEntity extends SmartBlockEntity {

	private static final int DEFAULT_DELAY = 6;
	private FilteringBehaviour filtering;
	private InvManipulationBehaviour observedInventory;
	private TankManipulationBehaviour observedTank;
	
	private VersionedInventoryTrackerBehaviour invVersionTracker;
	private boolean sustainSignal;
	
	public int turnOffTicks = 0;

	public SmartObserverBlockEntity(class_2591<?> type, class_2338 pos, class_2680 state) {
		super(type, pos, state);
		setLazyTickRate(20);
	}

	@Override
	public void addBehaviours(List<BlockEntityBehaviour> behaviours) {
		behaviours.add(filtering = new FilteringBehaviour(this, new FilteredDetectorFilterSlot(false))
			.withCallback($ -> invVersionTracker.reset()));
		behaviours.add(invVersionTracker = new VersionedInventoryTrackerBehaviour(this));

		InterfaceProvider towardBlockFacing =
			(w, p, s) -> new BlockFace(p, DirectedDirectionalBlock.getTargetDirection(s));

		behaviours.add(observedInventory = new InvManipulationBehaviour(this, towardBlockFacing).bypassSidedness());
		behaviours.add(observedTank = new TankManipulationBehaviour(this, towardBlockFacing).bypassSidedness());
	}

	@Override
	public void tick() {
		super.tick();
		
		if (field_11863.method_8608())
			return;
		
		class_2680 state = method_11010();
		if (turnOffTicks > 0) {
			turnOffTicks--;
			if (turnOffTicks == 0)
				field_11863.method_39279(field_11867, state.method_26204(), 1);
		}

		if (!isActive())
			return;

		class_2338 targetPos = field_11867.method_10093(SmartObserverBlock.getTargetDirection(state));
		class_2248 block = field_11863.method_8320(targetPos)
			.method_26204();

		if (!filtering.getFilter()
			.method_7960() && block.method_8389() != null && filtering.test(new class_1799(block))) {
			activate(3);
			return;
		}

		// Detect items on belt
		TransportedItemStackHandlerBehaviour behaviour =
			BlockEntityBehaviour.get(field_11863, targetPos, TransportedItemStackHandlerBehaviour.TYPE);
		if (behaviour != null) {
			behaviour.handleCenteredProcessingOnAllItems(.45f, stack -> {
				if (!filtering.test(stack.stack) || turnOffTicks == 6)
					return TransportedResult.doNothing();
				activate();
				return TransportedResult.doNothing();
			});
			return;
		}

		// Detect fluids in pipe
		FluidTransportBehaviour fluidBehaviour =
			BlockEntityBehaviour.get(field_11863, targetPos, FluidTransportBehaviour.TYPE);
		if (fluidBehaviour != null) {
			for (class_2350 side : Iterate.directions) {
				Flow flow = fluidBehaviour.getFlow(side);
				if (flow == null || !flow.inbound || !flow.complete)
					continue;
				if (!filtering.test(flow.fluid))
					continue;
				activate();
				return;
			}
			return;
		}

		if (observedInventory.hasInventory()) {
			boolean skipInv = invVersionTracker.stillWaiting(observedInventory);
			invVersionTracker.awaitNewVersion(observedInventory);

			if (skipInv && sustainSignal)
				turnOffTicks = DEFAULT_DELAY;

			if (!skipInv) {
				sustainSignal = false;
				if (!observedInventory.simulate()
					.extract()
					.method_7960()) {
					sustainSignal = true;
					activate();
					return;
				}
			}
		}

		if (!observedTank.simulate()
			.extractAny()
			.isEmpty()) {
			activate();
			return;
		}
	}

	public void activate() {
		activate(DEFAULT_DELAY);
	}

	public void activate(int ticks) {
		class_2680 state = method_11010();
		turnOffTicks = ticks;
		if (state.method_11654(SmartObserverBlock.POWERED))
			return;
		field_11863.method_8501(field_11867, state.method_11657(SmartObserverBlock.POWERED, true));
		field_11863.method_8452(field_11867, state.method_26204());
	}

	private boolean isActive() {
		return true;
	}

	@Override
	public void write(class_2487 compound, boolean clientPacket) {
		compound.method_10569("TurnOff", turnOffTicks);
		super.write(compound, clientPacket);
	}

	@Override
	protected void read(class_2487 compound, boolean clientPacket) {
		super.read(compound, clientPacket);
		turnOffTicks = compound.method_10550("TurnOff");
	}

}
