package com.simibubi.create.content.schematics;

import com.simibubi.create.AllBlocks;
import com.simibubi.create.Create;
import com.simibubi.create.content.contraptions.BlockMovementChecks;
import com.simibubi.create.content.contraptions.StructureTransform;
import com.simibubi.create.content.schematics.cannon.MaterialChecklist;
import com.simibubi.create.content.schematics.requirement.ItemRequirement;
import com.simibubi.create.foundation.blockEntity.IMergeableBE;
import com.simibubi.create.foundation.utility.BBHelper;
import com.simibubi.create.foundation.utility.BlockHelper;
import java.util.LinkedList;
import java.util.List;
import java.util.stream.Collectors;
import net.minecraft.class_1297;
import net.minecraft.class_1799;
import net.minecraft.class_1937;
import net.minecraft.class_2248;
import net.minecraft.class_2338;
import net.minecraft.class_2350;
import net.minecraft.class_2487;
import net.minecraft.class_2499;
import net.minecraft.class_2512;
import net.minecraft.class_2586;
import net.minecraft.class_2680;
import net.minecraft.class_2741;
import net.minecraft.class_2742;
import net.minecraft.class_2756;
import net.minecraft.class_3341;
import net.minecraft.class_3492;
import net.minecraft.class_3499;
import net.minecraft.class_7924;

public class SchematicPrinter {

	public enum PrintStage {
		BLOCKS, DEFERRED_BLOCKS, ENTITIES
	}

	private boolean schematicLoaded;
	private boolean isErrored;
	private SchematicWorld blockReader;
	private class_2338 schematicAnchor;

	private class_2338 currentPos;
	private int printingEntityIndex;
	private PrintStage printStage;
	private List<class_2338> deferredBlocks;

	public SchematicPrinter() {
		printingEntityIndex = -1;
		printStage = PrintStage.BLOCKS;
		deferredBlocks = new LinkedList<>();
	}

	public void fromTag(class_2487 compound, boolean clientPacket) {
		if (compound.method_10545("CurrentPos"))
			currentPos = class_2512.method_10691(compound.method_10562("CurrentPos"));
		if (clientPacket) {
			schematicLoaded = false;
			if (compound.method_10545("Anchor")) {
				schematicAnchor = class_2512.method_10691(compound.method_10562("Anchor"));
				schematicLoaded = true;
			}
		}

		printingEntityIndex = compound.method_10550("EntityProgress");
		printStage = PrintStage.valueOf(compound.method_10558("PrintStage"));
		compound.method_10554("DeferredBlocks", 10).stream()
				.map(p -> class_2512.method_10691((class_2487) p))
				.collect(Collectors.toCollection(() -> deferredBlocks));
	}

	public void write(class_2487 compound) {
		if (currentPos != null)
			compound.method_10566("CurrentPos", class_2512.method_10692(currentPos));
		if (schematicAnchor != null)
			compound.method_10566("Anchor", class_2512.method_10692(schematicAnchor));

		compound.method_10569("EntityProgress", printingEntityIndex);
		compound.method_10582("PrintStage", printStage.name());
		class_2499 tagDeferredBlocks = new class_2499();
		for (class_2338 p : deferredBlocks)
			tagDeferredBlocks.add(class_2512.method_10692(p));
		compound.method_10566("DeferredBlocks", tagDeferredBlocks);
	}

	public void loadSchematic(class_1799 blueprint, class_1937 originalWorld, boolean processNBT) {
		if (!blueprint.method_7985() || !blueprint.method_7969().method_10577("Deployed"))
			return;

		class_3499 activeTemplate =
			SchematicItem.loadSchematic(originalWorld.method_45448(class_7924.field_41254), blueprint);
		class_3492 settings = SchematicItem.getSettings(blueprint, processNBT);

		schematicAnchor = class_2512.method_10691(blueprint.method_7969()
				.method_10562("Anchor"));
		blockReader = new SchematicWorld(schematicAnchor, originalWorld);

		try {
			activeTemplate.method_15172(blockReader, schematicAnchor, schematicAnchor, settings,
				blockReader.method_8409(), class_2248.field_31028);
		} catch (Exception e) {
			Create.LOGGER.error("Failed to load Schematic for Printing", e);
			schematicLoaded = true;
			isErrored = true;
			return;
		}

		class_2338 extraBounds = class_3499.method_15171(settings, new class_2338(activeTemplate.method_15160())
				.method_10069(-1, -1, -1));
		blockReader.bounds = BBHelper.encapsulate(blockReader.bounds, extraBounds);

		StructureTransform transform = new StructureTransform(settings.method_15134(), class_2350.class_2351.field_11052,
				settings.method_15113(), settings.method_15114());
		for (class_2586 be : blockReader.getBlockEntities())
			transform.apply(be);

		printingEntityIndex = -1;
		printStage = PrintStage.BLOCKS;
		deferredBlocks.clear();
		class_3341 bounds = blockReader.getBounds();
		currentPos = new class_2338(bounds.method_35415() - 1, bounds.method_35416(), bounds.method_35417());
		schematicLoaded = true;
	}

	public void resetSchematic() {
		schematicLoaded = false;
		schematicAnchor = null;
		isErrored = false;
		currentPos = null;
		blockReader = null;
		printingEntityIndex = -1;
		printStage = PrintStage.BLOCKS;
		deferredBlocks.clear();
	}

	public boolean isLoaded() {
		return schematicLoaded;
	}

	public boolean isErrored() {
		return isErrored;
	}

	public class_2338 getCurrentTarget() {
		if (!isLoaded() || isErrored())
			return null;
		return schematicAnchor.method_10081(currentPos);
	}

	public PrintStage getPrintStage() {
		return printStage;
	}

	public class_2338 getAnchor() {
		return schematicAnchor;
	}

	public boolean isWorldEmpty() {
		return blockReader.getAllPositions().isEmpty();
		//return blockReader.getBounds().getLength().equals(new Vector3i(0,0,0));
	}

	@FunctionalInterface
	public interface BlockTargetHandler {
		void handle(class_2338 target, class_2680 blockState, class_2586 blockEntity);
	}

	@FunctionalInterface
	public interface EntityTargetHandler {
		void handle(class_2338 target, class_1297 entity);
	}

	public void handleCurrentTarget(BlockTargetHandler blockHandler, EntityTargetHandler entityHandler) {
		class_2338 target = getCurrentTarget();

		if (printStage == PrintStage.ENTITIES) {
			class_1297 entity = blockReader.getEntityStream()
					.collect(Collectors.toList())
					.get(printingEntityIndex);
			entityHandler.handle(target, entity);
		} else {
			class_2680 blockState = BlockHelper.setZeroAge(blockReader.method_8320(target));
			class_2586 blockEntity = blockReader.method_8321(target);
			blockHandler.handle(target, blockState, blockEntity);
		}
	}

	@FunctionalInterface
	public interface PlacementPredicate {
		boolean shouldPlace(class_2338 target, class_2680 blockState, class_2586 blockEntity,
							class_2680 toReplace, class_2680 toReplaceOther, boolean isNormalCube);
	}

	public boolean shouldPlaceCurrent(class_1937 world) {
		return shouldPlaceCurrent(world, (a, b, c, d, e, f) -> true);
	}

	public boolean shouldPlaceCurrent(class_1937 world, PlacementPredicate predicate) {
		if (world == null)
			return false;

		if (printStage == PrintStage.ENTITIES)
			return true;

		return shouldPlaceBlock(world, predicate, getCurrentTarget());
	}

	public boolean shouldPlaceBlock(class_1937 world, PlacementPredicate predicate, class_2338 pos) {
		class_2680 state = BlockHelper.setZeroAge(blockReader.method_8320(pos));
		class_2586 blockEntity = blockReader.method_8321(pos);

		class_2680 toReplace = world.method_8320(pos);
		class_2586 toReplaceBE = world.method_8321(pos);
		class_2680 toReplaceOther = null;

		if (state.method_28498(class_2741.field_12483) && state.method_28498(class_2741.field_12481)
				&& state.method_11654(class_2741.field_12483) == class_2742.field_12557)
			toReplaceOther = world.method_8320(pos.method_10093(state.method_11654(class_2741.field_12481)));
		if (state.method_28498(class_2741.field_12533)
				&& state.method_11654(class_2741.field_12533) == class_2756.field_12607)
			toReplaceOther = world.method_8320(pos.method_10084());

		boolean mergeTEs = blockEntity != null && toReplaceBE instanceof IMergeableBE mergeBE && toReplaceBE.method_11017()
				.equals(blockEntity.method_11017());

		if (!world.method_8477(pos))
			return false;
		if (!world.method_8621().method_11952(pos))
			return false;
		if (toReplace == state && !mergeTEs)
			return false;
		if (toReplace.method_26214(world, pos) == -1
				|| (toReplaceOther != null && toReplaceOther.method_26214(world, pos) == -1))
			return false;

		boolean isNormalCube = state.method_26212(blockReader, currentPos);
		return predicate.shouldPlace(pos, state, blockEntity, toReplace, toReplaceOther, isNormalCube);
	}

	public ItemRequirement getCurrentRequirement() {
		if (printStage == PrintStage.ENTITIES)
			return ItemRequirement.of(blockReader.getEntityStream()
					.collect(Collectors.toList())
					.get(printingEntityIndex));

		class_2338 target = getCurrentTarget();
		class_2680 blockState = BlockHelper.setZeroAge(blockReader.method_8320(target));
		class_2586 blockEntity = blockReader.method_8321(target);
		return ItemRequirement.of(blockState, blockEntity);
	}

	public int markAllBlockRequirements(MaterialChecklist checklist, class_1937 world, PlacementPredicate predicate) {
		int blocksToPlace = 0;
		for (class_2338 pos : blockReader.getAllPositions()) {
			class_2338 relPos = pos.method_10081(schematicAnchor);
			class_2680 required = blockReader.method_8320(relPos);
			class_2586 requiredBE = blockReader.method_8321(relPos);

			if (!world.method_8477(pos.method_10081(schematicAnchor))) {
				checklist.warnBlockNotLoaded();
				continue;
			}
			if (!shouldPlaceBlock(world, predicate, relPos))
				continue;
			ItemRequirement requirement = ItemRequirement.of(required, requiredBE);
			if (requirement.isEmpty())
				continue;
			if (requirement.isInvalid())
				continue;
			checklist.require(requirement);
			blocksToPlace++;
		}
		return blocksToPlace;
	}

	public void markAllEntityRequirements(MaterialChecklist checklist) {
		blockReader.getEntityStream()
				.forEach(entity -> {
					ItemRequirement requirement = ItemRequirement.of(entity);
					if (requirement.isEmpty())
						return;
					if (requirement.isInvalid())
						return;
					checklist.require(requirement);
				});
	}

	public boolean advanceCurrentPos() {
		List<class_1297> entities = blockReader.getEntityStream().collect(Collectors.toList());

		do {
			if (printStage == PrintStage.BLOCKS) {
				while (tryAdvanceCurrentPos()) {
					deferredBlocks.add(currentPos);
				}
			}

			if (printStage == PrintStage.DEFERRED_BLOCKS) {
				if (deferredBlocks.isEmpty()) {
					printStage = PrintStage.ENTITIES;
				} else {
					currentPos = deferredBlocks.remove(0);
				}
			}

			if (printStage == PrintStage.ENTITIES) {
				if (printingEntityIndex + 1 < entities.size()) {
					printingEntityIndex++;
					currentPos = entities.get(printingEntityIndex).method_24515().method_10059(schematicAnchor);
				} else {
					// Reached end of printing
					return false;
				}
			}
		} while (!blockReader.getBounds().method_14662(currentPos));

		// More things available to print
		return true;
	}

	public boolean tryAdvanceCurrentPos() {
		currentPos = currentPos.method_10093(class_2350.field_11034);
		class_3341 bounds = blockReader.getBounds();
		class_2338 posInBounds = currentPos.method_10069(-bounds.method_35415(), -bounds.method_35416(), -bounds.method_35417());

		if (posInBounds.method_10263() > bounds.method_35414())
			currentPos = new class_2338(bounds.method_35415(), currentPos.method_10264(), currentPos.method_10260() + 1).method_10067();
		if (posInBounds.method_10260() > bounds.method_14663())
			currentPos = new class_2338(currentPos.method_10263(), currentPos.method_10264() + 1, bounds.method_35417()).method_10067();

		// End of blocks reached
		if (currentPos.method_10264() > bounds.method_14660()) {
			printStage = PrintStage.DEFERRED_BLOCKS;
			return false;
		}

		return shouldDeferBlock(blockReader.method_8320(getCurrentTarget()));
	}

	public static boolean shouldDeferBlock(class_2680 state) {
		return AllBlocks.GANTRY_CARRIAGE.has(state) || AllBlocks.MECHANICAL_ARM.has(state)
				|| BlockMovementChecks.isBrittle(state);
	}

}
