package com.simibubi.create.content.schematics.cannon;

import java.util.Arrays;
import java.util.Optional;
import net.minecraft.class_1297;
import net.minecraft.class_1299;
import net.minecraft.class_1799;
import net.minecraft.class_1937;
import net.minecraft.class_2248;
import net.minecraft.class_2338;
import net.minecraft.class_2350.class_2351;
import net.minecraft.class_2487;
import net.minecraft.class_2512;
import net.minecraft.class_2520;
import net.minecraft.class_2680;
import net.minecraft.class_3532;
import net.minecraft.class_7871;
import com.simibubi.create.AllBlocks;
import com.simibubi.create.content.kinetics.belt.BeltBlock;
import com.simibubi.create.content.kinetics.belt.BeltBlockEntity;
import com.simibubi.create.content.kinetics.belt.BeltBlockEntity.CasingType;
import com.simibubi.create.content.kinetics.belt.BeltPart;
import com.simibubi.create.content.kinetics.belt.BeltSlope;
import com.simibubi.create.content.kinetics.belt.item.BeltConnectorItem;
import com.simibubi.create.content.kinetics.simpleRelays.AbstractSimpleShaftBlock;
import com.simibubi.create.foundation.utility.BlockHelper;
import io.github.fabricators_of_create.porting_lib.util.NBTSerializer;

public abstract class LaunchedItem {

	public int totalTicks;
	public int ticksRemaining;
	public class_2338 target;
	public class_1799 stack;

	private LaunchedItem(class_2338 start, class_2338 target, class_1799 stack) {
		this(target, stack, ticksForDistance(start, target), ticksForDistance(start, target));
	}

	private static int ticksForDistance(class_2338 start, class_2338 target) {
		return (int) (Math.max(10, Math.sqrt(Math.sqrt(target.method_10262(start))) * 4f));
	}

	LaunchedItem() {}

	private LaunchedItem(class_2338 target, class_1799 stack, int ticksLeft, int total) {
		this.target = target;
		this.stack = stack;
		this.totalTicks = total;
		this.ticksRemaining = ticksLeft;
	}

	public boolean update(class_1937 world) {
		if (ticksRemaining > 0) {
			ticksRemaining--;
			return false;
		}
		if (world.field_9236)
			return false;

		place(world);
		return true;
	}

	public class_2487 serializeNBT() {
		class_2487 c = new class_2487();
		c.method_10569("TotalTicks", totalTicks);
		c.method_10569("TicksLeft", ticksRemaining);
		c.method_10566("Stack", NBTSerializer.serializeNBT(stack));
		c.method_10566("Target", class_2512.method_10692(target));
		return c;
	}

	public static LaunchedItem fromNBT(class_2487 c, class_7871<class_2248> holderGetter) {
		LaunchedItem launched = c.method_10545("Length") ? new LaunchedItem.ForBelt()
			: c.method_10545("BlockState") ? new LaunchedItem.ForBlockState() : new LaunchedItem.ForEntity();
		launched.readNBT(c, holderGetter);
		return launched;
	}

	abstract void place(class_1937 world);

	void readNBT(class_2487 c, class_7871<class_2248> holderGetter) {
		target = class_2512.method_10691(c.method_10562("Target"));
		ticksRemaining = c.method_10550("TicksLeft");
		totalTicks = c.method_10550("TotalTicks");
		stack = class_1799.method_7915(c.method_10562("Stack"));
	}

	public static class ForBlockState extends LaunchedItem {
		public class_2680 state;
		public class_2487 data;

		ForBlockState() {}

		public ForBlockState(class_2338 start, class_2338 target, class_1799 stack, class_2680 state, class_2487 data) {
			super(start, target, stack);
			this.state = state;
			this.data = data;
		}

		@Override
		public class_2487 serializeNBT() {
			class_2487 serializeNBT = super.serializeNBT();
			serializeNBT.method_10566("BlockState", class_2512.method_10686(state));
			if (data != null) {
				data.method_10551("x");
				data.method_10551("y");
				data.method_10551("z");
				data.method_10551("id");
				serializeNBT.method_10566("Data", data);
			}
			return serializeNBT;
		}

		@Override
		void readNBT(class_2487 nbt, class_7871<class_2248> holderGetter) {
			super.readNBT(nbt, holderGetter);
			state = class_2512.method_10681(holderGetter, nbt.method_10562("BlockState"));
			if (nbt.method_10573("Data", class_2520.field_33260)) {
				data = nbt.method_10562("Data");
			}
		}

		@Override
		void place(class_1937 world) {
			BlockHelper.placeSchematicBlock(world, state, target, stack, data);
		}

	}

	public static class ForBelt extends ForBlockState {
		public int length;
		public CasingType[] casings;

		public ForBelt() {}

		@Override
		public class_2487 serializeNBT() {
			class_2487 serializeNBT = super.serializeNBT();
			serializeNBT.method_10569("Length", length);
			serializeNBT.method_10572("Casing", Arrays.stream(casings)
				.map(CasingType::ordinal)
				.toList());
			return serializeNBT;
		}

		@Override
		void readNBT(class_2487 nbt, class_7871<class_2248> holderGetter) {
			length = nbt.method_10550("Length");
			int[] intArray = nbt.method_10561("Casing");
			casings = new CasingType[length];
			for (int i = 0; i < casings.length; i++)
				casings[i] = i >= intArray.length ? CasingType.NONE
					: CasingType.values()[class_3532.method_15340(intArray[i], 0, CasingType.values().length - 1)];
			super.readNBT(nbt, holderGetter);
		}

		public ForBelt(class_2338 start, class_2338 target, class_1799 stack, class_2680 state, CasingType[] casings) {
			super(start, target, stack, state, null);
			this.casings = casings;
			this.length = casings.length;
		}

		@Override
		void place(class_1937 world) {
			boolean isStart = state.method_11654(BeltBlock.PART) == BeltPart.START;
			class_2338 offset = BeltBlock.nextSegmentPosition(state, class_2338.field_10980, isStart);
			int i = length - 1;
			class_2351 axis = state.method_11654(BeltBlock.SLOPE) == BeltSlope.SIDEWAYS ? class_2351.field_11052
				: state.method_11654(BeltBlock.HORIZONTAL_FACING)
					.method_10170()
					.method_10166();
			world.method_8501(target, AllBlocks.SHAFT.getDefaultState()
				.method_11657(AbstractSimpleShaftBlock.AXIS, axis));
			BeltConnectorItem.createBelts(world, target,
				target.method_10069(offset.method_10263() * i, offset.method_10264() * i, offset.method_10260() * i));

			for (int segment = 0; segment < length; segment++) {
				if (casings[segment] == CasingType.NONE)
					continue;
				class_2338 casingTarget =
					target.method_10069(offset.method_10263() * segment, offset.method_10264() * segment, offset.method_10260() * segment);
				if (world.method_8321(casingTarget) instanceof BeltBlockEntity bbe)
					bbe.setCasingType(casings[segment]);
			}
		}

	}

	public static class ForEntity extends LaunchedItem {
		public class_1297 entity;
		private class_2487 deferredTag;

		ForEntity() {}

		public ForEntity(class_2338 start, class_2338 target, class_1799 stack, class_1297 entity) {
			super(start, target, stack);
			this.entity = entity;
		}

		@Override
		public boolean update(class_1937 world) {
			if (deferredTag != null && entity == null) {
				try {
					Optional<class_1297> loadEntityUnchecked = class_1299.method_5892(deferredTag, world);
					if (!loadEntityUnchecked.isPresent())
						return true;
					entity = loadEntityUnchecked.get();
				} catch (Exception var3) {
					return true;
				}
				deferredTag = null;
			}
			return super.update(world);
		}

		@Override
		public class_2487 serializeNBT() {
			class_2487 serializeNBT = super.serializeNBT();
			if (entity != null)
				serializeNBT.method_10566("Entity", NBTSerializer.serializeNBT(entity));
			return serializeNBT;
		}

		@Override
		void readNBT(class_2487 nbt, class_7871<class_2248> holderGetter) {
			super.readNBT(nbt, holderGetter);
			if (nbt.method_10545("Entity"))
				deferredTag = nbt.method_10562("Entity");
		}

		@Override
		void place(class_1937 world) {
			if (entity != null)
				world.method_8649(entity);
		}

	}

}
