package com.simibubi.create.content.schematics.cannon;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Locale;

import com.google.common.collect.Sets;
import com.simibubi.create.AllBlocks;
import com.simibubi.create.content.equipment.clipboard.ClipboardEntry;
import com.simibubi.create.content.equipment.clipboard.ClipboardOverrides;
import com.simibubi.create.content.equipment.clipboard.ClipboardOverrides.ClipboardType;
import com.simibubi.create.content.schematics.requirement.ItemRequirement;
import com.simibubi.create.content.schematics.requirement.ItemRequirement.ItemUseType;
import com.simibubi.create.foundation.utility.Components;
import com.simibubi.create.foundation.utility.Lang;

import io.github.fabricators_of_create.porting_lib.transfer.TransferUtil;
import it.unimi.dsi.fastutil.objects.Object2IntArrayMap;
import it.unimi.dsi.fastutil.objects.Object2IntMap;
import net.fabricmc.fabric.api.transfer.v1.item.ItemVariant;
import net.fabricmc.fabric.api.transfer.v1.storage.StorageView;
import net.minecraft.class_124;
import net.minecraft.class_1792;
import net.minecraft.class_1799;
import net.minecraft.class_1802;
import net.minecraft.class_2487;
import net.minecraft.class_2499;
import net.minecraft.class_2519;
import net.minecraft.class_2561;
import net.minecraft.class_2568;
import net.minecraft.class_2583;
import net.minecraft.class_5250;

public class MaterialChecklist {

	public static final int MAX_ENTRIES_PER_PAGE = 5;
	public static final int MAX_ENTRIES_PER_CLIPBOARD_PAGE = 7;

	public Object2IntMap<class_1792> gathered = new Object2IntArrayMap<>();
	public Object2IntMap<class_1792> required = new Object2IntArrayMap<>();
	public Object2IntMap<class_1792> damageRequired = new Object2IntArrayMap<>();
	public boolean blocksNotLoaded;

	public void warnBlockNotLoaded() {
		blocksNotLoaded = true;
	}

	public void require(ItemRequirement requirement) {
		if (requirement.isEmpty())
			return;
		if (requirement.isInvalid())
			return;

		for (ItemRequirement.StackRequirement stack : requirement.getRequiredItems()) {
			if (stack.usage == ItemUseType.DAMAGE)
				putOrIncrement(damageRequired, stack.stack);
			if (stack.usage == ItemUseType.CONSUME)
				putOrIncrement(required, stack.stack);
		}
	}

	private void putOrIncrement(Object2IntMap<class_1792> map, class_1799 stack) {
		class_1792 item = stack.method_7909();
		if (item == class_1802.field_8162)
			return;
		if (map.containsKey(item))
			map.put(item, map.getInt(item) + stack.method_7947());
		else
			map.put(item, stack.method_7947());
	}

	public void collect(class_1799 stack) {
		class_1792 item = stack.method_7909();
		if (required.containsKey(item) || damageRequired.containsKey(item))
			if (gathered.containsKey(item))
				gathered.put(item, gathered.getInt(item) + stack.method_7947());
			else
				gathered.put(item, stack.method_7947());
	}

	public void collect(StorageView<ItemVariant> view) {
		if (view.isResourceBlank())
			return;
		int amount = TransferUtil.truncateLong(view.getAmount());
		class_1799 stack = view.getResource().toStack(amount);
		collect(stack);
	}

	public class_1799 createWrittenBook() {
		class_1799 book = new class_1799(class_1802.field_8360);

		class_2487 tag = book.method_7948();
		class_2499 pages = new class_2499();

		int itemsWritten = 0;
		class_5250 textComponent;

		if (blocksNotLoaded) {
			textComponent = Components.literal("\n" + class_124.field_1061);
			textComponent = textComponent.method_10852(Lang.translateDirect("materialChecklist.blocksNotLoaded"));
			pages.add(class_2519.method_23256(class_2561.class_2562.method_10867(textComponent)));
		}

		List<class_1792> keys = new ArrayList<>(Sets.union(required.keySet(), damageRequired.keySet()));
		Collections.sort(keys, (item1, item2) -> {
			Locale locale = Locale.ENGLISH;
			String name1 = item1.method_7848()
				.getString()
				.toLowerCase(locale);
			String name2 = item2.method_7848()
				.getString()
				.toLowerCase(locale);
			return name1.compareTo(name2);
		});

		textComponent = Components.empty();
		List<class_1792> completed = new ArrayList<>();
		for (class_1792 item : keys) {
			int amount = getRequiredAmount(item);
			if (gathered.containsKey(item))
				amount -= gathered.getInt(item);

			if (amount <= 0) {
				completed.add(item);
				continue;
			}

			if (itemsWritten == MAX_ENTRIES_PER_PAGE) {
				itemsWritten = 0;
				textComponent.method_10852(Components.literal("\n >>>")
					.method_27692(class_124.field_1078));
				pages.add(class_2519.method_23256(class_2561.class_2562.method_10867(textComponent)));
				textComponent = Components.empty();
			}

			itemsWritten++;
			textComponent.method_10852(entry(new class_1799(item), amount, true, true));
		}

		for (class_1792 item : completed) {
			if (itemsWritten == MAX_ENTRIES_PER_PAGE) {
				itemsWritten = 0;
				textComponent.method_10852(Components.literal("\n >>>")
					.method_27692(class_124.field_1077));
				pages.add(class_2519.method_23256(class_2561.class_2562.method_10867(textComponent)));
				textComponent = Components.empty();
			}

			itemsWritten++;
			textComponent.method_10852(entry(new class_1799(item), getRequiredAmount(item), false, true));
		}

		pages.add(class_2519.method_23256(class_2561.class_2562.method_10867(textComponent)));

		tag.method_10566("pages", pages);
		tag.method_10556("readonly", true);
		tag.method_10582("author", "Schematicannon");
		tag.method_10582("title", class_124.field_1078 + "Material Checklist");
		textComponent = Lang.translateDirect("materialChecklist")
			.method_10862(class_2583.field_24360.method_10977(class_124.field_1078)
				.method_10978(Boolean.FALSE));
		book.method_7911("display")
			.method_10582("Name", class_2561.class_2562.method_10867(textComponent));
		book.method_7980(tag);

		return book;
	}

	public class_1799 createWrittenClipboard() {
		class_1799 clipboard = AllBlocks.CLIPBOARD.asStack();
		class_2487 tag = clipboard.method_7948();
		int itemsWritten = 0;

		List<List<ClipboardEntry>> pages = new ArrayList<>();
		List<ClipboardEntry> currentPage = new ArrayList<>();

		if (blocksNotLoaded) {
			currentPage.add(new ClipboardEntry(false, Lang.translateDirect("materialChecklist.blocksNotLoaded")
				.method_27692(class_124.field_1061)));
		}

		List<class_1792> keys = new ArrayList<>(Sets.union(required.keySet(), damageRequired.keySet()));
		Collections.sort(keys, (item1, item2) -> {
			Locale locale = Locale.ENGLISH;
			String name1 = item1.method_7848()
				.getString()
				.toLowerCase(locale);
			String name2 = item2.method_7848()
				.getString()
				.toLowerCase(locale);
			return name1.compareTo(name2);
		});

		List<class_1792> completed = new ArrayList<>();
		for (class_1792 item : keys) {
			int amount = getRequiredAmount(item);
			if (gathered.containsKey(item))
				amount -= gathered.getInt(item);

			if (amount <= 0) {
				completed.add(item);
				continue;
			}

			if (itemsWritten == MAX_ENTRIES_PER_CLIPBOARD_PAGE) {
				itemsWritten = 0;
				currentPage.add(new ClipboardEntry(false, Components.literal(">>>")
					.method_27692(class_124.field_1063)));
				pages.add(currentPage);
				currentPage = new ArrayList<>();
			}

			itemsWritten++;
			currentPage.add(new ClipboardEntry(false, entry(new class_1799(item), amount, true, false))
				.displayItem(new class_1799(item)));
		}

		for (class_1792 item : completed) {
			if (itemsWritten == MAX_ENTRIES_PER_CLIPBOARD_PAGE) {
				itemsWritten = 0;
				currentPage.add(new ClipboardEntry(true, Components.literal(">>>")
					.method_27692(class_124.field_1077)));
				pages.add(currentPage);
				currentPage = new ArrayList<>();
			}

			itemsWritten++;
			currentPage.add(new ClipboardEntry(true, entry(new class_1799(item), getRequiredAmount(item), false, false))
				.displayItem(new class_1799(item)));
		}

		pages.add(currentPage);
		ClipboardEntry.saveAll(pages, clipboard);
		ClipboardOverrides.switchTo(ClipboardType.WRITTEN, clipboard);
		clipboard.method_7911("display")
			.method_10582("Name", class_2561.class_2562.method_10867(Lang.translateDirect("materialChecklist")
				.method_10862(class_2583.field_24360.method_10978(Boolean.FALSE))));
		tag.method_10556("Readonly", true);
		clipboard.method_7980(tag);
		return clipboard;
	}

	public int getRequiredAmount(class_1792 item) {
		int amount = required.getOrDefault(item, 0);
		if (damageRequired.containsKey(item))
			amount += Math.ceil(damageRequired.getInt(item) / (float) new class_1799(item).method_7936());
		return amount;
	}

	private class_5250 entry(class_1799 item, int amount, boolean unfinished, boolean forBook) {
		int stacks = amount / 64;
		int remainder = amount % 64;
		class_5250 tc = Components.empty();
		tc.method_10852(Components.translatable(item.method_7922())
			.method_10862(class_2583.field_24360
				.method_10949(new class_2568(class_2568.class_5247.field_24343, new class_2568.class_5249(item)))));

		if (!unfinished && forBook)
			tc.method_27693(" \u2714");
		if (!unfinished || forBook)
			tc.method_27692(unfinished ? class_124.field_1078 : class_124.field_1077);
		return tc.method_10852(Components.literal("\n" + " x" + amount)
			.method_27692(class_124.field_1074))
			.method_10852(Components.literal(" | " + stacks + "\u25A4 +" + remainder + (forBook ? "\n" : ""))
				.method_27692(class_124.field_1080));
	}

}
