package com.simibubi.create.content.schematics.cannon;

import java.util.Arrays;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import javax.annotation.Nullable;

import com.simibubi.create.AllBlocks;
import com.simibubi.create.AllItems;
import com.simibubi.create.AllSoundEvents;
import com.simibubi.create.content.kinetics.belt.BeltBlock;
import com.simibubi.create.content.kinetics.belt.BeltBlockEntity;
import com.simibubi.create.content.kinetics.belt.BeltBlockEntity.CasingType;
import com.simibubi.create.content.kinetics.belt.BeltPart;
import com.simibubi.create.content.kinetics.belt.BeltSlope;
import com.simibubi.create.content.kinetics.simpleRelays.AbstractSimpleShaftBlock;
import com.simibubi.create.content.schematics.SchematicPrinter;
import com.simibubi.create.content.schematics.requirement.ItemRequirement;
import com.simibubi.create.content.schematics.requirement.ItemRequirement.ItemUseType;
import com.simibubi.create.foundation.blockEntity.SmartBlockEntity;
import com.simibubi.create.foundation.blockEntity.behaviour.BlockEntityBehaviour;
import com.simibubi.create.foundation.utility.BlockHelper;
import com.simibubi.create.foundation.utility.Iterate;
import com.simibubi.create.foundation.utility.Lang;
import com.simibubi.create.infrastructure.config.AllConfigs;
import com.simibubi.create.infrastructure.config.CSchematics;

import io.github.fabricators_of_create.porting_lib.block.CustomRenderBoundingBoxBlockEntity;
import io.github.fabricators_of_create.porting_lib.transfer.TransferUtil;
import io.github.fabricators_of_create.porting_lib.util.NBTSerializer;
import io.github.fabricators_of_create.porting_lib.util.StorageProvider;
import net.fabricmc.api.EnvType;
import net.fabricmc.api.Environment;
import net.fabricmc.fabric.api.transfer.v1.item.ItemVariant;
import net.fabricmc.fabric.api.transfer.v1.storage.Storage;
import net.fabricmc.fabric.api.transfer.v1.storage.base.ResourceAmount;
import net.fabricmc.fabric.api.transfer.v1.transaction.Transaction;
import net.fabricmc.fabric.api.transfer.v1.transaction.TransactionContext;
import net.minecraft.class_1297;
import net.minecraft.class_1657;
import net.minecraft.class_1661;
import net.minecraft.class_1703;
import net.minecraft.class_1747;
import net.minecraft.class_1792;
import net.minecraft.class_1799;
import net.minecraft.class_1802;
import net.minecraft.class_1937;
import net.minecraft.class_2246;
import net.minecraft.class_2338;
import net.minecraft.class_2350;
import net.minecraft.class_2350.class_2351;
import net.minecraft.class_2350.class_2352;
import net.minecraft.class_238;
import net.minecraft.class_2487;
import net.minecraft.class_2499;
import net.minecraft.class_2540;
import net.minecraft.class_2561;
import net.minecraft.class_2586;
import net.minecraft.class_2591;
import net.minecraft.class_2671;
import net.minecraft.class_2680;
import net.minecraft.class_2741;
import net.minecraft.class_2742;
import net.minecraft.class_2756;
import net.minecraft.class_3908;

public class SchematicannonBlockEntity extends SmartBlockEntity implements class_3908, CustomRenderBoundingBoxBlockEntity {

	public static final int NEIGHBOUR_CHECKING = 100;
	public static final int MAX_ANCHOR_DISTANCE = 256;

	// Inventory
	public SchematicannonInventory inventory;

	public boolean sendUpdate;
	// Sync
	public boolean dontUpdateChecklist;
	public int neighbourCheckCooldown;

	// Printer
	public SchematicPrinter printer;
	public class_1799 missingItem;
	public boolean positionNotLoaded;
	public boolean hasCreativeCrate;
	private int printerCooldown;
	private int skipsLeft;
	private boolean blockSkipped;

	public class_2338 previousTarget;
	public List<LaunchedItem> flyingBlocks;
	public MaterialChecklist checklist;

	// Gui information
	public float fuelLevel;
	public float bookPrintingProgress;
	public float schematicProgress;
	public String statusMsg;
	public State state;
	public int blocksPlaced;
	public int blocksToPlace;

	// Settings
	public int replaceMode;
	public boolean skipMissing;
	public boolean replaceBlockEntities;

	// Render
	public boolean firstRenderTick;
	public float defaultYaw;

	// fabric: transfer
	private final Map<class_2350, StorageProvider<ItemVariant>> storages = new HashMap<>();

	public SchematicannonBlockEntity(class_2591<?> type, class_2338 pos, class_2680 state) {
		super(type, pos, state);
		setLazyTickRate(30);
		flyingBlocks = new LinkedList<>();
		inventory = new SchematicannonInventory(this);
		statusMsg = "idle";
		this.state = State.STOPPED;
		replaceMode = 2;
		checklist = new MaterialChecklist();
		printer = new SchematicPrinter();
	}

	@Override
	public void method_31662(class_1937 level) {
		super.method_31662(level);
		for (class_2350 direction : Iterate.directions) {
			class_2338 pos = field_11867.method_10093(direction);
			StorageProvider<ItemVariant> provider = StorageProvider.createForItems(level, pos);
			storages.put(direction, provider);
		}
	}

	// fabric: storages not stored, only check for creative crate
	public void findInventories() {
		hasCreativeCrate = false;

		for (class_2350 facing : Iterate.directions) {
			class_2338 pos = field_11867.method_10093(facing);

			if (!field_11863.method_8477(pos))
				continue;

			if (AllBlocks.CREATIVE_CRATE.has(field_11863.method_8320(pos)))
				hasCreativeCrate = true;
		}
	}

	@Override
	protected void read(class_2487 compound, boolean clientPacket) {
		if (!clientPacket) {
			inventory.deserializeNBT(compound.method_10562("Inventory"));
		}

		// Gui information
		statusMsg = compound.method_10558("Status");
		schematicProgress = compound.method_10583("Progress");
		bookPrintingProgress = compound.method_10583("PaperProgress");
		fuelLevel = compound.method_10583("Fuel");
		String stateString = compound.method_10558("State");
		state = stateString.isEmpty() ? State.STOPPED : State.valueOf(compound.method_10558("State"));
		blocksPlaced = compound.method_10550("AmountPlaced");
		blocksToPlace = compound.method_10550("AmountToPlace");

		missingItem = null;
		if (compound.method_10545("MissingItem"))
			missingItem = class_1799.method_7915(compound.method_10562("MissingItem"));

		// Settings
		class_2487 options = compound.method_10562("Options");
		replaceMode = options.method_10550("ReplaceMode");
		skipMissing = options.method_10577("SkipMissing");
		replaceBlockEntities = options.method_10577("ReplaceTileEntities");

		// Printer & Flying Blocks
		if (compound.method_10545("Printer"))
			printer.fromTag(compound.method_10562("Printer"), clientPacket);
		if (compound.method_10545("FlyingBlocks"))
			readFlyingBlocks(compound);

		defaultYaw = compound.method_10583("DefaultYaw");

		super.read(compound, clientPacket);
	}

	protected void readFlyingBlocks(class_2487 compound) {
		class_2499 tagBlocks = compound.method_10554("FlyingBlocks", 10);
		if (tagBlocks.isEmpty())
			flyingBlocks.clear();

		boolean pastDead = false;

		for (int i = 0; i < tagBlocks.size(); i++) {
			class_2487 c = tagBlocks.method_10602(i);
			LaunchedItem launched = LaunchedItem.fromNBT(c, blockHolderGetter());
			class_2338 readBlockPos = launched.target;

			// Always write to Server block entity
			if (field_11863 == null || !field_11863.field_9236) {
				flyingBlocks.add(launched);
				continue;
			}

			// Delete all Client side blocks that are now missing on the server
			while (!pastDead && !flyingBlocks.isEmpty() && !flyingBlocks.get(0).target.equals(readBlockPos)) {
				flyingBlocks.remove(0);
			}

			pastDead = true;

			// Add new server side blocks
			if (i >= flyingBlocks.size()) {
				flyingBlocks.add(launched);
				continue;
			}

			// Don't do anything with existing
		}
	}

	@Override
	public void write(class_2487 compound, boolean clientPacket) {
		if (!clientPacket) {
			compound.method_10566("Inventory", inventory.serializeNBT());
			if (state == State.RUNNING) {
				compound.method_10556("Running", true);
			}
		}

		// Gui information
		compound.method_10548("Progress", schematicProgress);
		compound.method_10548("PaperProgress", bookPrintingProgress);
		compound.method_10548("Fuel", fuelLevel);
		compound.method_10582("Status", statusMsg);
		compound.method_10582("State", state.name());
		compound.method_10569("AmountPlaced", blocksPlaced);
		compound.method_10569("AmountToPlace", blocksToPlace);

		if (missingItem != null)
			compound.method_10566("MissingItem", NBTSerializer.serializeNBT(missingItem));

		// Settings
		class_2487 options = new class_2487();
		options.method_10569("ReplaceMode", replaceMode);
		options.method_10556("SkipMissing", skipMissing);
		options.method_10556("ReplaceTileEntities", replaceBlockEntities);
		compound.method_10566("Options", options);

		// Printer & Flying Blocks
		class_2487 printerData = new class_2487();
		printer.write(printerData);
		compound.method_10566("Printer", printerData);

		class_2499 tagFlyingBlocks = new class_2499();
		for (LaunchedItem b : flyingBlocks)
			tagFlyingBlocks.add(b.serializeNBT());
		compound.method_10566("FlyingBlocks", tagFlyingBlocks);

		compound.method_10548("DefaultYaw", defaultYaw);

		super.write(compound, clientPacket);
	}

	@Override
	public void tick() {
		super.tick();

		if (state != State.STOPPED && neighbourCheckCooldown-- <= 0) {
			neighbourCheckCooldown = NEIGHBOUR_CHECKING;
			findInventories();
		}

		firstRenderTick = true;
		previousTarget = printer.getCurrentTarget();
		tickFlyingBlocks();

		if (field_11863.field_9236)
			return;

		// Update Fuel and Paper
		tickPaperPrinter();
		refillFuelIfPossible();

		// Update Printer
		skipsLeft = 1000;
		blockSkipped = true;

		while (blockSkipped && skipsLeft-- > 0)
			tickPrinter();

		schematicProgress = 0;
		if (blocksToPlace > 0)
			schematicProgress = (float) blocksPlaced / blocksToPlace;

		// Update Client block entity
		if (sendUpdate) {
			sendUpdate = false;
			field_11863.method_8413(field_11867, method_11010(), method_11010(), 6);
		}
	}

	public CSchematics config() {
		return AllConfigs.server().schematics;
	}

	protected void tickPrinter() {
		class_1799 blueprint = inventory.getStackInSlot(0);
		blockSkipped = false;

		if (blueprint.method_7960() && !statusMsg.equals("idle") && inventory.getStackInSlot(1)
			.method_7960()) {
			state = State.STOPPED;
			statusMsg = "idle";
			sendUpdate = true;
			return;
		}

		// Skip if not Active
		if (state == State.STOPPED) {
			if (printer.isLoaded())
				resetPrinter();
			return;
		}

		if (state == State.PAUSED && !positionNotLoaded && missingItem == null && fuelLevel > getFuelUsageRate())
			return;

		// Initialize Printer
		if (!printer.isLoaded()) {
			initializePrinter(blueprint);
			return;
		}

		// Cooldown from last shot
		if (printerCooldown > 0) {
			printerCooldown--;
			return;
		}

		// Check Fuel
		if (fuelLevel <= 0 && !hasCreativeCrate) {
			fuelLevel = 0;
			state = State.PAUSED;
			statusMsg = "noGunpowder";
			sendUpdate = true;
			return;
		}

		if (hasCreativeCrate) {
			if (missingItem != null) {
				missingItem = null;
				state = State.RUNNING;
			}
		}

		// Update Target
		if (missingItem == null && !positionNotLoaded) {
			if (!printer.advanceCurrentPos()) {
				finishedPrinting();
				return;
			}
			sendUpdate = true;
		}

		// Check block
		if (!method_10997().method_8477(printer.getCurrentTarget())) {
			positionNotLoaded = true;
			statusMsg = "targetNotLoaded";
			state = State.PAUSED;
			return;
		} else {
			if (positionNotLoaded) {
				positionNotLoaded = false;
				state = State.RUNNING;
			}
		}

		// Get item requirement
		ItemRequirement requirement = printer.getCurrentRequirement();
		if (requirement.isInvalid() || !printer.shouldPlaceCurrent(field_11863, this::shouldPlace)) {
			sendUpdate = !statusMsg.equals("searching");
			statusMsg = "searching";
			blockSkipped = true;
			return;
		}

		// Find item
		List<ItemRequirement.StackRequirement> requiredItems = requirement.getRequiredItems();
		if (!requirement.isEmpty()) {
			try (Transaction t = TransferUtil.getTransaction()) {
				for (ItemRequirement.StackRequirement required : requiredItems) {
					if (!grabItemsFromAttachedInventories(required, t)) {
						if (skipMissing) {
							statusMsg = "skipping";
							blockSkipped = true;
							if (missingItem != null) {
								missingItem = null;
								state = State.RUNNING;
							}
							return;
						}

						missingItem = required.stack;
						state = State.PAUSED;
						statusMsg = "missingBlock";
						return;
					}
				}

				t.commit();
			}
		}

		// Success
		state = State.RUNNING;
		class_1799 icon = requirement.isEmpty() || requiredItems.isEmpty() ? class_1799.field_8037 : requiredItems.get(0).stack;
		printer.handleCurrentTarget((target, blockState, blockEntity) -> {
			// Launch block
			statusMsg = blockState.method_26204() != class_2246.field_10124 ? "placing" : "clearing";
			launchBlockOrBelt(target, icon, blockState, blockEntity);
		}, (target, entity) -> {
			// Launch entity
			statusMsg = "placing";
			launchEntity(target, icon, entity);
		});

		printerCooldown = config().schematicannonDelay.get();
		fuelLevel -= getFuelUsageRate();
		sendUpdate = true;
		missingItem = null;
	}

	public double getFuelUsageRate() {
		return hasCreativeCrate ? 0 : config().schematicannonFuelUsage.get() / 100f;
	}

	protected void initializePrinter(class_1799 blueprint) {
		if (!blueprint.method_7985()) {
			state = State.STOPPED;
			statusMsg = "schematicInvalid";
			sendUpdate = true;
			return;
		}

		if (!blueprint.method_7969()
				.method_10577("Deployed")) {
			state = State.STOPPED;
			statusMsg = "schematicNotPlaced";
			sendUpdate = true;
			return;
		}

		// Load blocks into reader
		printer.loadSchematic(blueprint, field_11863, true);

		if (printer.isErrored()) {
			state = State.STOPPED;
			statusMsg = "schematicErrored";
			inventory.setStackInSlot(0, class_1799.field_8037);
			inventory.setStackInSlot(1, new class_1799(AllItems.EMPTY_SCHEMATIC.get()));
			printer.resetSchematic();
			sendUpdate = true;
			return;
		}

		if (printer.isWorldEmpty()) {
			state = State.STOPPED;
			statusMsg = "schematicExpired";
			inventory.setStackInSlot(0, class_1799.field_8037);
			inventory.setStackInSlot(1, new class_1799(AllItems.EMPTY_SCHEMATIC.get()));
			printer.resetSchematic();
			sendUpdate = true;
			return;
		}

		if (!printer.getAnchor()
				.method_19771(method_11016(), MAX_ANCHOR_DISTANCE)) {
			state = State.STOPPED;
			statusMsg = "targetOutsideRange";
			printer.resetSchematic();
			sendUpdate = true;
			return;
		}

		state = State.PAUSED;
		statusMsg = "ready";
		updateChecklist();
		sendUpdate = true;
		blocksToPlace += blocksPlaced;
	}

	protected class_1799 getItemForBlock(class_2680 blockState) {
		class_1792 item = class_1747.field_8003.getOrDefault(blockState.method_26204(), class_1802.field_8162);
		return item == class_1802.field_8162 ? class_1799.field_8037 : new class_1799(item);
	}

	protected boolean grabItemsFromAttachedInventories(ItemRequirement.StackRequirement required, TransactionContext transaction) {
		if (hasCreativeCrate)
			return true;

		class_1799 stack = required.stack;
		ItemUseType usage = required.usage;

		// Find and apply damage
		// fabric - can't modify directly, extract and re-insert
		if (usage == ItemUseType.DAMAGE) {
			if (!stack.method_7963())
				return false;

			for (Entry<class_2350, StorageProvider<ItemVariant>> entry : storages.entrySet()) {
				Storage<ItemVariant> storage = entry.getValue().get(entry.getKey().method_10153());
				if (storage == null)
					continue;
				try (Transaction t = transaction.openNested()) {
					ResourceAmount<ItemVariant> resource = TransferUtil.extractMatching(storage, required::matches, 1, t);
					if (resource == null || resource.amount() != 1)
						continue; // failed, skip
					ItemVariant variant = resource.resource();
					class_1799 newStack = variant.toStack();
					newStack.method_7974(newStack.method_7919() + 1);
					if (stack.method_7919() < stack.method_7936()) {
						// stack not broken, re-insert
						ItemVariant newVariant = ItemVariant.of(newStack);
						long inserted = storage.insert(newVariant, 1, t);
						if (inserted != 1)
							continue; // failed to re-insert, cancel this whole attempt
					}
					t.commit();
					return true;
				}
			}
			// could not find in any storage
			return false;
		}

		// Find and remove
		int toExtract = stack.method_7947();
		try (Transaction t = transaction.openNested()) {
			for (Entry<class_2350, StorageProvider<ItemVariant>> entry : storages.entrySet()) {
				Storage<ItemVariant> storage = entry.getValue().get(entry.getKey().method_10153());
				if (storage == null)
					continue;

				ResourceAmount<ItemVariant> resource = TransferUtil.extractMatching(storage, required::matches, stack.method_7947(), t);
				if (resource == null)
					continue;
				toExtract -= resource.amount();
				if (toExtract > 0) // still need to extract more
					continue;
				// extracted enough
				t.commit();
				return true;
			}
		}
		// if we get here we didn't find enough
		return false;
	}

	public void finishedPrinting() {
		inventory.setStackInSlot(0, class_1799.field_8037);
		inventory.setStackInSlot(1, new class_1799(AllItems.EMPTY_SCHEMATIC.get(), inventory.getStackInSlot(1)
				.method_7947() + 1));
		state = State.STOPPED;
		statusMsg = "finished";
		resetPrinter();
		AllSoundEvents.SCHEMATICANNON_FINISH.playOnServer(field_11863, field_11867);
		sendUpdate = true;
	}

	protected void resetPrinter() {
		printer.resetSchematic();
		missingItem = null;
		sendUpdate = true;
		schematicProgress = 0;
		blocksPlaced = 0;
		blocksToPlace = 0;
	}

	protected boolean shouldPlace(class_2338 pos, class_2680 state, class_2586 be, class_2680 toReplace,
								  class_2680 toReplaceOther, boolean isNormalCube) {
		if (pos.method_19771(method_11016(), 2f))
			return false;
		if (!replaceBlockEntities
				&& (toReplace.method_31709() || (toReplaceOther != null && toReplaceOther.method_31709())))
			return false;

		if (shouldIgnoreBlockState(state, be))
			return false;

		boolean placingAir = state.method_26215();

		if (replaceMode == 3)
			return true;
		if (replaceMode == 2 && !placingAir)
			return true;
		if (replaceMode == 1 && (isNormalCube || (!toReplace.method_26212(field_11863, pos)
				&& (toReplaceOther == null || !toReplaceOther.method_26212(field_11863, pos)))) && !placingAir)
			return true;
		if (replaceMode == 0 && !toReplace.method_26212(field_11863, pos)
				&& (toReplaceOther == null || !toReplaceOther.method_26212(field_11863, pos)) && !placingAir)
			return true;

		return false;
	}

	protected boolean shouldIgnoreBlockState(class_2680 state, class_2586 be) {
		// Block doesn't have a mapping (Water, lava, etc)
		if (state.method_26204() == class_2246.field_10369)
			return true;

		ItemRequirement requirement = ItemRequirement.of(state, be);
		if (requirement.isEmpty())
			return false;
		if (requirement.isInvalid())
			return false;

		// Block doesn't need to be placed twice (Doors, beds, double plants)
		if (state.method_28498(class_2741.field_12533)
				&& state.method_11654(class_2741.field_12533) == class_2756.field_12609)
			return true;
		if (state.method_28498(class_2741.field_12483)
				&& state.method_11654(class_2741.field_12483) == class_2742.field_12560)
			return true;
		if (state.method_26204() instanceof class_2671)
			return true;
		if (AllBlocks.BELT.has(state))
			return state.method_11654(BeltBlock.PART) == BeltPart.MIDDLE;

		return false;
	}

	protected void tickFlyingBlocks() {
		List<LaunchedItem> toRemove = new LinkedList<>();
		for (LaunchedItem b : flyingBlocks)
			if (b.update(field_11863))
				toRemove.add(b);
		flyingBlocks.removeAll(toRemove);
	}

	protected void refillFuelIfPossible() {
		if (hasCreativeCrate)
			return;
		if (1 - fuelLevel + 1 / 128f < getFuelAddedByGunPowder())
			return;
		if (inventory.getStackInSlot(4)
				.method_7960())
			return;

		inventory.getStackInSlot(4)
				.method_7934(1);
		fuelLevel += getFuelAddedByGunPowder();
		if (statusMsg.equals("noGunpowder")) {
			if (blocksPlaced > 0)
				state = State.RUNNING;
			statusMsg = "ready";
		}
		sendUpdate = true;
	}

	public double getFuelAddedByGunPowder() {
		return config().schematicannonGunpowderWorth.get() / 100f;
	}

	protected void tickPaperPrinter() {
		int BookInput = 2;
		int BookOutput = 3;

		class_1799 blueprint = inventory.getStackInSlot(0);
		class_1799 paper = inventory.getStackInSlot(BookInput).method_7972();
		paper.method_7939(1);
		boolean outputFull = inventory.getStackInSlot(BookOutput)
				.method_7947() == inventory.getSlotLimit(BookOutput);
if (printer.isErrored())
			return;

		if (!printer.isLoaded()) {
			if (!blueprint.method_7960())
				initializePrinter(blueprint);
			return;
		}

		if (paper.method_7960() || outputFull) {
			if (bookPrintingProgress != 0)
				sendUpdate = true;
			bookPrintingProgress = 0;
			dontUpdateChecklist = false;
			return;
		}

		if (bookPrintingProgress >= 1) {
			bookPrintingProgress = 0;

			if (!dontUpdateChecklist)
				updateChecklist();

			dontUpdateChecklist = true;
			class_1799 extractItem = inventory.getStackInSlot(BookInput);
			// non-empty, would early exit above
			TransferUtil.extract(inventory, ItemVariant.of(extractItem), 1);
			class_1799 stack = AllBlocks.CLIPBOARD.isIn(extractItem) ? checklist.createWrittenClipboard()
				: checklist.createWrittenBook();
			stack.method_7939(inventory.getStackInSlot(BookOutput)
					.method_7947() + 1);
			inventory.setStackInSlot(BookOutput, stack);
			sendUpdate = true;
			return;
		}

		bookPrintingProgress += 0.05f;
		sendUpdate = true;
	}

	public static class_2680 stripBeltIfNotLast(class_2680 blockState) {
		BeltPart part = blockState.method_11654(BeltBlock.PART);
		if (part == BeltPart.MIDDLE)
			return class_2246.field_10124.method_9564();

		// is highest belt?
		boolean isLastSegment = false;
		class_2350 facing = blockState.method_11654(BeltBlock.HORIZONTAL_FACING);
		BeltSlope slope = blockState.method_11654(BeltBlock.SLOPE);
		boolean positive = facing.method_10171() == class_2352.field_11056;
		boolean start = part == BeltPart.START;
		boolean end = part == BeltPart.END;

		switch (slope) {
			case DOWNWARD:
				isLastSegment = start;
				break;
			case UPWARD:
				isLastSegment = end;
				break;
			default:
				isLastSegment = positive && end || !positive && start;
		}
		if (isLastSegment)
			return blockState;

		return AllBlocks.SHAFT.getDefaultState()
				.method_11657(AbstractSimpleShaftBlock.AXIS, slope == BeltSlope.SIDEWAYS ? class_2351.field_11052
						:facing.method_10170()
								.method_10166());
	}

	protected void launchBlockOrBelt(class_2338 target, class_1799 icon, class_2680 blockState, class_2586 blockEntity) {
		if (AllBlocks.BELT.has(blockState)) {
			blockState = stripBeltIfNotLast(blockState);
			if (blockEntity instanceof BeltBlockEntity bbe && AllBlocks.BELT.has(blockState)) {
				CasingType[] casings = new CasingType[bbe.beltLength];
				Arrays.fill(casings, CasingType.NONE);
				class_2338 currentPos = target;
				for (int i = 0; i < bbe.beltLength; i++) {
					class_2680 currentState = bbe.method_10997()
						.method_8320(currentPos);
					if (!(currentState.method_26204() instanceof BeltBlock))
						break;
					if (!(bbe.method_10997()
						.method_8321(currentPos) instanceof BeltBlockEntity beltAtSegment))
						break;
					casings[i] = beltAtSegment.casing;
					currentPos = BeltBlock.nextSegmentPosition(currentState, currentPos,
						blockState.method_11654(BeltBlock.PART) != BeltPart.END);
				}
				launchBelt(target, blockState, bbe.beltLength, casings);
			} else if (blockState != class_2246.field_10124.method_9564())
				launchBlock(target, icon, blockState, null);
			return;
		}

		class_2487 data = BlockHelper.prepareBlockEntityData(blockState, blockEntity);
		launchBlock(target, icon, blockState, data);
	}

	protected void launchBelt(class_2338 target, class_2680 state, int length, CasingType[] casings) {
		blocksPlaced++;
		class_1799 connector = AllItems.BELT_CONNECTOR.asStack();
		flyingBlocks.add(new LaunchedItem.ForBelt(this.method_11016(), target, connector, state, casings));
		playFiringSound();
	}

	protected void launchBlock(class_2338 target, class_1799 stack, class_2680 state, @Nullable class_2487 data) {
		if (!state.method_26215())
			blocksPlaced++;
		flyingBlocks.add(new LaunchedItem.ForBlockState(this.method_11016(), target, stack, state, data));
		playFiringSound();
	}

	protected void launchEntity(class_2338 target, class_1799 stack, class_1297 entity) {
		blocksPlaced++;
		flyingBlocks.add(new LaunchedItem.ForEntity(this.method_11016(), target, stack, entity));
		playFiringSound();
	}

	public void playFiringSound() {
		AllSoundEvents.SCHEMATICANNON_LAUNCH_BLOCK.playOnServer(field_11863, field_11867);
	}

	public void sendToMenu(class_2540 buffer) {
		buffer.method_10807(method_11016());
		buffer.method_10794(method_16887());
	}

	@Override
	public class_1703 createMenu(int id, class_1661 inv, class_1657 player) {
		return SchematicannonMenu.create(id, inv, this);
	}

	@Override
	public class_2561 method_5476() {
		return Lang.translateDirect("gui.schematicannon.title");
	}

	public void updateChecklist() {
		checklist.required.clear();
		checklist.damageRequired.clear();
		checklist.blocksNotLoaded = false;

		if (printer.isLoaded() && !printer.isErrored()) {
			blocksToPlace = blocksPlaced;
			blocksToPlace += printer.markAllBlockRequirements(checklist, field_11863, this::shouldPlace);
			printer.markAllEntityRequirements(checklist);
		}

		checklist.gathered.clear();
		try (Transaction t = TransferUtil.getTransaction()) {
			for (Entry<class_2350, StorageProvider<ItemVariant>> entry : storages.entrySet()) {
				Storage<ItemVariant> storage = entry.getValue().get(entry.getKey().method_10153());
				if (storage == null)
					continue;
				storage.nonEmptyViews().forEach(checklist::collect);
			}
		}
		sendUpdate = true;
	}

	@Override
	public void addBehaviours(List<BlockEntityBehaviour> behaviours) {
	}

	@Override
	public void lazyTick() {
		super.lazyTick();
		findInventories();
	}

	@Override
	@Environment(EnvType.CLIENT)
	public class_238 getRenderBoundingBox() {
		return INFINITE_EXTENT_AABB;
	}

	public enum State {
		STOPPED, PAUSED, RUNNING;
	}

}
