package com.simibubi.create.content.schematics.cannon;

import com.jozufozu.flywheel.backend.Backend;
import com.jozufozu.flywheel.core.model.ModelUtil;
import com.jozufozu.flywheel.core.virtual.VirtualEmptyBlockGetter;
import com.jozufozu.flywheel.fabric.model.DefaultLayerFilteringBakedModel;
import com.simibubi.create.AllBlocks;
import com.simibubi.create.AllPartialModels;
import com.simibubi.create.content.schematics.cannon.LaunchedItem.ForBelt;
import com.simibubi.create.content.schematics.cannon.LaunchedItem.ForBlockState;
import com.simibubi.create.content.schematics.cannon.LaunchedItem.ForEntity;
import com.simibubi.create.foundation.blockEntity.renderer.SafeBlockEntityRenderer;
import com.simibubi.create.foundation.render.CachedBufferer;
import com.simibubi.create.foundation.render.SuperByteBuffer;

import io.github.fabricators_of_create.porting_lib.mixin.accessors.client.accessor.BlockRenderDispatcherAccessor;
import io.github.fabricators_of_create.porting_lib.models.virtual.FixedLightBakedModel;
import net.minecraft.class_1087;
import net.minecraft.class_1799;
import net.minecraft.class_1921;
import net.minecraft.class_2338;
import net.minecraft.class_2350;
import net.minecraft.class_2398;
import net.minecraft.class_243;
import net.minecraft.class_2680;
import net.minecraft.class_310;
import net.minecraft.class_3532;
import net.minecraft.class_4587;
import net.minecraft.class_4588;
import net.minecraft.class_4597;
import net.minecraft.class_4696;
import net.minecraft.class_5614;
import net.minecraft.class_5819;
import net.minecraft.class_776;
import net.minecraft.class_7833;
import net.minecraft.class_811;

public class SchematicannonRenderer extends SafeBlockEntityRenderer<SchematicannonBlockEntity> {

	public SchematicannonRenderer(class_5614.class_5615 context) {}

	@Override
	protected void renderSafe(SchematicannonBlockEntity blockEntity, float partialTicks, class_4587 ms,
		class_4597 buffer, int light, int overlay) {

		boolean blocksLaunching = !blockEntity.flyingBlocks.isEmpty();
		if (blocksLaunching)
			renderLaunchedBlocks(blockEntity, partialTicks, ms, buffer, light, overlay);

		if (Backend.canUseInstancing(blockEntity.method_10997()))
			return;

		class_2338 pos = blockEntity.method_11016();
		class_2680 state = blockEntity.method_11010();

		double[] cannonAngles = getCannonAngles(blockEntity, pos, partialTicks);

		double yaw = cannonAngles[0];
		double pitch = cannonAngles[1];

		double recoil = getRecoil(blockEntity, partialTicks);

		ms.method_22903();

		class_4588 vb = buffer.getBuffer(class_1921.method_23577());

		SuperByteBuffer connector = CachedBufferer.partial(AllPartialModels.SCHEMATICANNON_CONNECTOR, state);
		connector.translate(.5f, 0, .5f);
		connector.rotate(class_2350.field_11036, (float) ((yaw + 90) / 180 * Math.PI));
		connector.translate(-.5f, 0, -.5f);
		connector.light(light)
			.renderInto(ms, vb);

		SuperByteBuffer pipe = CachedBufferer.partial(AllPartialModels.SCHEMATICANNON_PIPE, state);
		pipe.translate(.5f, 15 / 16f, .5f);
		pipe.rotate(class_2350.field_11036, (float) ((yaw + 90) / 180 * Math.PI));
		pipe.rotate(class_2350.field_11035, (float) (pitch / 180 * Math.PI));
		pipe.translate(-.5f, -15 / 16f, -.5f);
		pipe.translate(0, -recoil / 100, 0);
		pipe.light(light)
			.renderInto(ms, vb);

		ms.method_22909();
	}

	public static double[] getCannonAngles(SchematicannonBlockEntity blockEntity, class_2338 pos, float partialTicks) {
		double yaw;
		double pitch;

		class_2338 target = blockEntity.printer.getCurrentTarget();
		if (target != null) {

			// Calculate Angle of Cannon
			class_243 diff = class_243.method_24954(target.method_10059(pos));
			if (blockEntity.previousTarget != null) {
				diff = (class_243.method_24954(blockEntity.previousTarget)
					.method_1019(class_243.method_24954(target.method_10059(blockEntity.previousTarget))
						.method_1021(partialTicks))).method_1020(class_243.method_24954(pos));
			}

			double diffX = diff.method_10216();
			double diffZ = diff.method_10215();
			yaw = class_3532.method_15349(diffX, diffZ);
			yaw = yaw / Math.PI * 180;

			float distance = class_3532.method_15355((float) (diffX * diffX + diffZ * diffZ));
			double yOffset = 0 + distance * 2f;
			pitch = class_3532.method_15349(distance, diff.method_10214() * 3 + yOffset);
			pitch = pitch / Math.PI * 180 + 10;

		} else {
			yaw = blockEntity.defaultYaw;
			pitch = 40;
		}

		return new double[] { yaw, pitch };
	}

	public static double getRecoil(SchematicannonBlockEntity blockEntity, float partialTicks) {
		double recoil = 0;

		for (LaunchedItem launched : blockEntity.flyingBlocks) {

			if (launched.ticksRemaining == 0)
				continue;

			// Apply Recoil if block was just launched
			if ((launched.ticksRemaining + 1 - partialTicks) > launched.totalTicks - 10)
				recoil = Math.max(recoil, (launched.ticksRemaining + 1 - partialTicks) - launched.totalTicks + 10);
		}

		return recoil;
	}

	private static void renderLaunchedBlocks(SchematicannonBlockEntity blockEntity, float partialTicks, class_4587 ms,
		class_4597 buffer, int light, int overlay) {
		for (LaunchedItem launched : blockEntity.flyingBlocks) {

			if (launched.ticksRemaining == 0)
				continue;

			// Calculate position of flying block
			class_243 start = class_243.method_24953(blockEntity.method_11016()
				.method_10084());
			class_243 target = class_243.method_24953(launched.target);
			class_243 distance = target.method_1020(start);

			double yDifference = target.field_1351 - start.field_1351;
			double throwHeight = Math.sqrt(distance.method_1027()) * .6f + yDifference;
			class_243 cannonOffset = distance.method_1031(0, throwHeight, 0)
				.method_1029()
				.method_1021(2);
			start = start.method_1019(cannonOffset);
			yDifference = target.field_1351 - start.field_1351;

			float progress =
				((float) launched.totalTicks - (launched.ticksRemaining + 1 - partialTicks)) / launched.totalTicks;
			class_243 blockLocationXZ = target.method_1020(start)
				.method_1021(progress)
				.method_18805(1, 0, 1);

			// Height is determined through a bezier curve
			float t = progress;
			double yOffset = 2 * (1 - t) * t * throwHeight + t * t * yDifference;
			class_243 blockLocation = blockLocationXZ.method_1031(0.5, yOffset + 1.5, 0.5)
				.method_1019(cannonOffset);

			// Offset to position
			ms.method_22903();
			ms.method_22904(blockLocation.field_1352, blockLocation.field_1351, blockLocation.field_1350);

			ms.method_46416(.125f, .125f, .125f);
			ms.method_22907(class_7833.field_40716.rotationDegrees(360 * t));
			ms.method_22907(class_7833.field_40714.rotationDegrees(360 * t));
			ms.method_46416(-.125f, -.125f, -.125f);

			if (launched instanceof ForBlockState) {
				// Render the Block
				class_2680 state;
				if (launched instanceof ForBelt) {
					// Render a shaft instead of the belt
					state = AllBlocks.SHAFT.getDefaultState();
				} else {
					state = ((ForBlockState) launched).state;
				}
				float scale = .3f;
				ms.method_22905(scale, scale, scale);
//				Minecraft.getInstance()
//					.getBlockRenderer()
//					.renderSingleBlock(state, ms, buffer, light, overlay);
				class_776 dispatcher = class_310.method_1551()
					.method_1541();
				switch (state.method_26217()) {
					case field_11458 -> {
						class_1087 model = dispatcher.method_3349(state);
						model = DefaultLayerFilteringBakedModel.wrap(model);
						model = FixedLightBakedModel.wrap(model, light);
						dispatcher.method_3350()
								.method_3374(VirtualEmptyBlockGetter.INSTANCE, model, state, class_2338.field_10980, ms, buffer.getBuffer(class_4696.method_23683(state, false)), false, class_5819.method_43047(), 42L, overlay);
					}
					case field_11456 -> ((BlockRenderDispatcherAccessor) dispatcher).getBlockEntityRenderer().method_3166(new class_1799(state.method_26204()), class_811.field_4315, ms, buffer, light, overlay);
				}
			} else if (launched instanceof ForEntity) {
				// Render the item
				float scale = 1.2f;
				ms.method_22905(scale, scale, scale);
				class_310.method_1551()
					.method_1480()
					.method_23178(launched.stack, class_811.field_4318, light, overlay, ms, buffer, blockEntity.method_10997(), 0);
			}

			ms.method_22909();

			// Render particles for launch
			if (launched.ticksRemaining == launched.totalTicks && blockEntity.firstRenderTick) {
				start = start.method_1023(.5, .5, .5);
				blockEntity.firstRenderTick = false;
				for (int i = 0; i < 10; i++) {
					class_5819 r = blockEntity.method_10997()
						.method_8409();
					double sX = cannonOffset.field_1352 * .01f;
					double sY = (cannonOffset.field_1351 + 1) * .01f;
					double sZ = cannonOffset.field_1350 * .01f;
					double rX = r.method_43057() - sX * 40;
					double rY = r.method_43057() - sY * 40;
					double rZ = r.method_43057() - sZ * 40;
					blockEntity.method_10997()
						.method_8406(class_2398.field_11204, start.field_1352 + rX, start.field_1351 + rY, start.field_1350 + rZ, sX, sY, sZ);
				}
			}

		}
	}

	@Override
	public boolean shouldRenderOffScreen(SchematicannonBlockEntity blockEntity) {
		return true;
	}

	@Override
	public int method_33893() {
		return 128;
	}

}
