package com.simibubi.create.content.schematics.client;

import java.util.List;
import net.minecraft.class_1799;
import net.minecraft.class_2338;
import net.minecraft.class_2415;
import net.minecraft.class_2470;
import net.minecraft.class_2512;
import net.minecraft.class_2561;
import net.minecraft.class_332;
import net.minecraft.class_342;
import net.minecraft.class_3492;
import com.simibubi.create.AllItems;
import com.simibubi.create.CreateClient;
import com.simibubi.create.foundation.gui.AbstractSimiScreen;
import com.simibubi.create.foundation.gui.AllGuiTextures;
import com.simibubi.create.foundation.gui.AllIcons;
import com.simibubi.create.foundation.gui.element.GuiGameElement;
import com.simibubi.create.foundation.gui.widget.IconButton;
import com.simibubi.create.foundation.gui.widget.Label;
import com.simibubi.create.foundation.gui.widget.ScrollInput;
import com.simibubi.create.foundation.gui.widget.SelectionScrollInput;
import com.simibubi.create.foundation.utility.Components;
import com.simibubi.create.foundation.utility.Lang;

public class SchematicEditScreen extends AbstractSimiScreen {

	private final List<class_2561> rotationOptions =
		Lang.translatedOptions("schematic.rotation", "none", "cw90", "cw180", "cw270");
	private final List<class_2561> mirrorOptions =
		Lang.translatedOptions("schematic.mirror", "none", "leftRight", "frontBack");
	private final class_2561 rotationLabel = Lang.translateDirect("schematic.rotation");
	private final class_2561 mirrorLabel = Lang.translateDirect("schematic.mirror");

	private AllGuiTextures background;

	private class_342 xInput;
	private class_342 yInput;
	private class_342 zInput;
	private IconButton confirmButton;

	private ScrollInput rotationArea;
	private ScrollInput mirrorArea;
	private SchematicHandler handler;

	public SchematicEditScreen() {
		background = AllGuiTextures.SCHEMATIC;
		handler = CreateClient.SCHEMATIC_HANDLER;
	}

	@Override
	protected void method_25426() {
		setWindowSize(background.width, background.height);
		setWindowOffset(-6, 0);
		super.method_25426();

		int x = guiLeft;
		int y = guiTop;

		xInput = new class_342(field_22793, x + 50, y + 26, 34, 10, Components.immutableEmpty());
		yInput = new class_342(field_22793, x + 90, y + 26, 34, 10, Components.immutableEmpty());
		zInput = new class_342(field_22793, x + 130, y + 26, 34, 10, Components.immutableEmpty());

		class_2338 anchor = handler.getTransformation()
				.getAnchor();
		if (handler.isDeployed()) {
			xInput.method_1852("" + anchor.method_10263());
			yInput.method_1852("" + anchor.method_10264());
			zInput.method_1852("" + anchor.method_10260());
		} else {
			class_2338 alt = field_22787.field_1724.method_24515();
			xInput.method_1852("" + alt.method_10263());
			yInput.method_1852("" + alt.method_10264());
			zInput.method_1852("" + alt.method_10260());
		}

		for (class_342 widget : new class_342[] { xInput, yInput, zInput }) {
			widget.method_1880(6);
			widget.method_1858(false);
			widget.method_1868(0xFFFFFF);
			widget.method_25365(false);
			widget.method_25402(0, 0, 0);
			widget.method_1890(s -> {
				if (s.isEmpty() || s.equals("-"))
					return true;
				try {
					Integer.parseInt(s);
					return true;
				} catch (NumberFormatException e) {
					return false;
				}
			});
		}

		class_3492 settings = handler.getTransformation()
			.toSettings();
		Label labelR = new Label(x + 50, y + 48, Components.immutableEmpty()).withShadow();
		rotationArea = new SelectionScrollInput(x + 45, y + 43, 118, 18).forOptions(rotationOptions)
			.titled(rotationLabel.method_27662())
			.setState(settings.method_15113()
				.ordinal())
			.writingTo(labelR);

		Label labelM = new Label(x + 50, y + 70, Components.immutableEmpty()).withShadow();
		mirrorArea = new SelectionScrollInput(x + 45, y + 65, 118, 18).forOptions(mirrorOptions)
			.titled(mirrorLabel.method_27662())
			.setState(settings.method_15114()
				.ordinal())
			.writingTo(labelM);

		addRenderableWidgets(xInput, yInput, zInput);
		addRenderableWidgets(labelR, labelM, rotationArea, mirrorArea);

		confirmButton =
			new IconButton(x + background.width - 33, y + background.height - 24, AllIcons.I_CONFIRM);
		confirmButton.withCallback(() -> {
			method_25419();
		});
		method_37063(confirmButton);
	}

	@Override
	public boolean method_25404(int code, int p_keyPressed_2_, int p_keyPressed_3_) {
		if (method_25437(code)) {
			String coords = field_22787.field_1774.method_1460();
			if (coords != null && !coords.isEmpty()) {
				coords.replaceAll(" ", "");
				String[] split = coords.split(",");
				if (split.length == 3) {
					boolean valid = true;
					for (String s : split) {
						try {
							Integer.parseInt(s);
						} catch (NumberFormatException e) {
							valid = false;
						}
					}
					if (valid) {
						xInput.method_1852(split[0]);
						yInput.method_1852(split[1]);
						zInput.method_1852(split[2]);
						return true;
					}
				}
			}
		}

		return super.method_25404(code, p_keyPressed_2_, p_keyPressed_3_);
	}

	@Override
	protected void renderWindow(class_332 graphics, int mouseX, int mouseY, float partialTicks) {
		int x = guiLeft;
		int y = guiTop;

		background.render(graphics, x, y);
		String title = handler.getCurrentSchematicName();
		graphics.method_25300(field_22793, title, x + (background.width - 8) / 2, y + 3, 0xFFFFFF);

		GuiGameElement.of(AllItems.SCHEMATIC.asStack())
				.<GuiGameElement.GuiRenderBuilder>at(x + background.width + 6, y + background.height - 40, -200)
				.scale(3)
				.render(graphics);
	}

	@Override
	public void method_25432() {
		boolean validCoords = true;
		class_2338 newLocation = null;
		try {
			newLocation = new class_2338(Integer.parseInt(xInput.method_1882()), Integer.parseInt(yInput.method_1882()),
				Integer.parseInt(zInput.method_1882()));
		} catch (NumberFormatException e) {
			validCoords = false;
		}

		class_3492 settings = new class_3492();
		settings.method_15123(class_2470.values()[rotationArea.getState()]);
		settings.method_15125(class_2415.values()[mirrorArea.getState()]);

		if (validCoords && newLocation != null) {
			class_1799 item = handler.getActiveSchematicItem();
			if (item != null) {
				item.method_7969()
					.method_10556("Deployed", true);
				item.method_7969()
					.method_10566("Anchor", class_2512.method_10692(newLocation));
			}

			handler.getTransformation()
				.init(newLocation, settings, handler.getBounds());
			handler.markDirty();
			handler.deploy();
		}
	}

}
