package com.simibubi.create.content.schematics.client;

import java.util.LinkedHashMap;
import java.util.Map;
import net.minecraft.class_1087;
import net.minecraft.class_1921;
import net.minecraft.class_2338;
import net.minecraft.class_2464;
import net.minecraft.class_2680;
import net.minecraft.class_287;
import net.minecraft.class_290;
import net.minecraft.class_293;
import net.minecraft.class_310;
import net.minecraft.class_3341;
import net.minecraft.class_4587;
import net.minecraft.class_4608;
import net.minecraft.class_4696;
import net.minecraft.class_5819;
import net.minecraft.class_776;
import net.minecraft.class_778;
import com.jozufozu.flywheel.core.model.ModelUtil;
import com.jozufozu.flywheel.core.model.ShadeSeparatedBufferedData;
import com.jozufozu.flywheel.core.model.ShadeSeparatingVertexConsumer;
import com.jozufozu.flywheel.fabric.model.LayerFilteringBakedModel;
import com.simibubi.create.content.schematics.SchematicWorld;
import com.simibubi.create.foundation.render.BlockEntityRenderHelper;
import com.simibubi.create.foundation.render.SuperByteBuffer;
import com.simibubi.create.foundation.render.SuperRenderTypeBuffer;

public class SchematicRenderer {

	private static final ThreadLocal<ThreadLocalObjects> THREAD_LOCAL_OBJECTS = ThreadLocal.withInitial(ThreadLocalObjects::new);

	private final Map<class_1921, SuperByteBuffer> bufferCache = new LinkedHashMap<>();
	private boolean active;
	private boolean changed;
	protected SchematicWorld schematic;
	private class_2338 anchor;

	public SchematicRenderer() {
		changed = false;
	}

	public void display(SchematicWorld world) {
		this.anchor = world.anchor;
		this.schematic = world;
		this.active = true;
		this.changed = true;
	}

	public void setActive(boolean active) {
		this.active = active;
	}

	public void update() {
		changed = true;
	}

	public void tick() {
		if (!active)
			return;
		class_310 mc = class_310.method_1551();
		if (mc.field_1687 == null || mc.field_1724 == null || !changed)
			return;

		redraw();
		changed = false;
	}

	public void render(class_4587 ms, SuperRenderTypeBuffer buffers) {
		if (!active)
			return;
		bufferCache.forEach((layer, buffer) -> {
			buffer.renderInto(ms, buffers.getBuffer(layer));
		});
		BlockEntityRenderHelper.renderBlockEntities(schematic, schematic.getRenderedBlockEntities(), ms, buffers);
	}

	protected void redraw() {
		bufferCache.forEach((layer, sbb) -> sbb.delete());
		bufferCache.clear();

		for (class_1921 layer : class_1921.method_22720()) {
			SuperByteBuffer buffer = drawLayer(layer);
			if (!buffer.isEmpty())
				bufferCache.put(layer, buffer);
			else
				buffer.delete();
		}
	}

	protected SuperByteBuffer drawLayer(class_1921 layer) {
		class_776 dispatcher = class_310.method_1551().method_1541();
		class_778 renderer = dispatcher.method_3350();
		ThreadLocalObjects objects = THREAD_LOCAL_OBJECTS.get();

		class_4587 poseStack = objects.poseStack;
		class_5819 random = objects.random;
		class_2338.class_2339 mutableBlockPos = objects.mutableBlockPos;
		SchematicWorld renderWorld = schematic;
		renderWorld.renderMode = true;
		class_3341 bounds = renderWorld.getBounds();

		ShadeSeparatingVertexConsumer shadeSeparatingWrapper = objects.shadeSeparatingWrapper;
		class_287 shadedBuilder = objects.shadedBuilder;
		class_287 unshadedBuilder = objects.unshadedBuilder;

		shadedBuilder.method_1328(class_293.class_5596.field_27382, class_290.field_1590);
		unshadedBuilder.method_1328(class_293.class_5596.field_27382, class_290.field_1590);
		shadeSeparatingWrapper.prepare(shadedBuilder, unshadedBuilder);

		class_778.method_20544();
		for (class_2338 localPos : class_2338.method_10094(bounds.method_35415(), bounds.method_35416(), bounds.method_35417(), bounds.method_35418(), bounds.method_35419(), bounds.method_35420())) {
			class_2338 pos = mutableBlockPos.method_35831(localPos, anchor);
			class_2680 state = renderWorld.method_8320(pos);

			if (state.method_26217() == class_2464.field_11458) {
				class_1087 model = dispatcher.method_3349(state);
				long seed = state.method_26190(pos);
				random.method_43052(seed);
				if (model.isVanillaAdapter()) {
					if (class_4696.method_23679(state) != layer) {
						continue;
					}
				} else {
					model = LayerFilteringBakedModel.wrap(model, layer);
				}
				model = shadeSeparatingWrapper.wrapModel(model);

				poseStack.method_22903();
				poseStack.method_46416(localPos.method_10263(), localPos.method_10264(), localPos.method_10260());

				renderer.method_3374(renderWorld, model, state, pos, poseStack, shadeSeparatingWrapper, true, random,
						seed, class_4608.field_21444);

				poseStack.method_22909();
			}
		}
		class_778.method_20545();

		shadeSeparatingWrapper.clear();
		ShadeSeparatedBufferedData bufferedData = ModelUtil.endAndCombine(shadedBuilder, unshadedBuilder);

		renderWorld.renderMode = false;

		SuperByteBuffer sbb = new SuperByteBuffer(bufferedData);
		bufferedData.release();
		return sbb;
	}

	// fabric: calling chunkBufferLayers early causes issues (#612), let the map handle its size on its own
//	private static int getLayerCount() {
//		return RenderType.chunkBufferLayers()
//			.size();
//	}

	private static class ThreadLocalObjects {
		public final class_4587 poseStack = new class_4587();
		public final class_5819 random = class_5819.method_43053();
		public final class_2338.class_2339 mutableBlockPos = new class_2338.class_2339();
		public final ShadeSeparatingVertexConsumer shadeSeparatingWrapper = new ShadeSeparatingVertexConsumer();
		public final class_287 shadedBuilder = new class_287(512);
		public final class_287 unshadedBuilder = new class_287(512);
	}

}
