package com.simibubi.create.content.schematics.client;

import java.util.List;
import java.util.function.Consumer;
import net.minecraft.class_1041;
import net.minecraft.class_2561;
import net.minecraft.class_310;
import net.minecraft.class_332;
import net.minecraft.class_437;
import net.minecraft.class_4587;
import com.mojang.blaze3d.systems.RenderSystem;
import com.simibubi.create.AllKeys;
import com.simibubi.create.content.schematics.client.tools.ToolType;
import com.simibubi.create.foundation.gui.AllGuiTextures;
import com.simibubi.create.foundation.utility.Components;
import com.simibubi.create.foundation.utility.Lang;

public class ToolSelectionScreen extends class_437 {

	public final String scrollToCycle = Lang.translateDirect("gui.toolmenu.cycle")
		.getString();
	public final String holdToFocus = "gui.toolmenu.focusKey";

	protected List<ToolType> tools;
	protected Consumer<ToolType> callback;
	public boolean focused;
	private float yOffset;
	protected int selection;
	private boolean initialized;

	protected int w;
	protected int h;

	public ToolSelectionScreen(List<ToolType> tools, Consumer<ToolType> callback) {
		super(Components.literal("Tool Selection"));
		this.field_22787 = class_310.method_1551();
		this.tools = tools;
		this.callback = callback;
		focused = false;
		yOffset = 0;
		selection = 0;
		initialized = false;

		callback.accept(tools.get(selection));

		w = Math.max(tools.size() * 50 + 30, 220);
		h = 30;
	}

	public void setSelectedElement(ToolType tool) {
		if (!tools.contains(tool))
			return;
		selection = tools.indexOf(tool);
	}

	public void cycle(int direction) {
		selection += (direction < 0) ? 1 : -1;
		selection = (selection + tools.size()) % tools.size();
	}

	private void draw(class_332 graphics, float partialTicks) {
		class_4587 matrixStack = graphics.method_51448();
		class_1041 mainWindow = field_22787.method_22683();
		if (!initialized)
			method_25423(field_22787, mainWindow.method_4486(), mainWindow.method_4502());

		int x = (mainWindow.method_4486() - w) / 2 + 15;
		int y = mainWindow.method_4502() - h - 75;

		matrixStack.method_22903();
		matrixStack.method_46416(0, -yOffset, focused ? 100 : 0);

		AllGuiTextures gray = AllGuiTextures.HUD_BACKGROUND;
		RenderSystem.enableBlend();
		RenderSystem.setShaderColor(1, 1, 1, focused ? 7 / 8f : 1 / 2f);

		graphics.method_25290(gray.location, x - 15, y, gray.startX, gray.startY, w, h, gray.width, gray.height);

		float toolTipAlpha = yOffset / 10;
		List<class_2561> toolTip = tools.get(selection)
			.getDescription();
		int stringAlphaComponent = ((int) (toolTipAlpha * 0xFF)) << 24;

		if (toolTipAlpha > 0.25f) {
			RenderSystem.setShaderColor(.7f, .7f, .8f, toolTipAlpha);
			graphics.method_25290(gray.location, x - 15, y + 33, gray.startX, gray.startY, w, h + 22, gray.width, gray.height);
			RenderSystem.setShaderColor(1, 1, 1, 1);

			if (toolTip.size() > 0)
				graphics.method_51439(field_22793, toolTip.get(0), x - 10, y + 38, 0xEEEEEE + stringAlphaComponent, false);
			if (toolTip.size() > 1)
				graphics.method_51439(field_22793, toolTip.get(1), x - 10, y + 50, 0xCCDDFF + stringAlphaComponent, false);
			if (toolTip.size() > 2)
				graphics.method_51439(field_22793, toolTip.get(2), x - 10, y + 60, 0xCCDDFF + stringAlphaComponent, false);
			if (toolTip.size() > 3)
				graphics.method_51439(field_22793, toolTip.get(3), x - 10, y + 72, 0xCCCCDD + stringAlphaComponent, false);
		}

		RenderSystem.setShaderColor(1, 1, 1, 1);
		if (tools.size() > 1) {
			String keyName = AllKeys.TOOL_MENU.getBoundKey();
			int width = field_22787.method_22683()
				.method_4486();
			if (!focused)
				graphics.method_27534(field_22787.field_1772, Lang.translateDirect(holdToFocus, keyName), width / 2,
					y - 10, 0xCCDDFF);
			else
				graphics.method_25300(field_22787.field_1772, scrollToCycle, width / 2, y - 10, 0xCCDDFF);
		} else {
			x += 65;
		}
		

		for (int i = 0; i < tools.size(); i++) {
			RenderSystem.enableBlend();
			matrixStack.method_22903();

			float alpha = focused ? 1 : .2f;
			if (i == selection) {
				matrixStack.method_46416(0, -10, 0);
				RenderSystem.setShaderColor(1, 1, 1, 1);
				graphics.method_25300(field_22787.field_1772, tools.get(i)
					.getDisplayName()
					.getString(), x + i * 50 + 24, y + 28, 0xCCDDFF);
				alpha = 1;
			}
			RenderSystem.setShaderColor(0, 0, 0, alpha);
			tools.get(i)
				.getIcon()
				.render(graphics, x + i * 50 + 16, y + 12);
			RenderSystem.setShaderColor(1, 1, 1, alpha);
			tools.get(i)
				.getIcon()
				.render(graphics, x + i * 50 + 16, y + 11);

			matrixStack.method_22909();
		}

		RenderSystem.disableBlend();
		matrixStack.method_22909();
	}

	public void update() {
		if (focused)
			yOffset += (10 - yOffset) * .1f;
		else
			yOffset *= .9f;
	}

	public void renderPassive(class_332 graphics, float partialTicks) {
		draw(graphics, partialTicks);
	}

	@Override
	public void method_25419() {
		callback.accept(tools.get(selection));
	}

	@Override
	protected void method_25426() {
		super.method_25426();
		initialized = true;
	}
}
