package com.simibubi.create.content.schematics.requirement;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import com.simibubi.create.foundation.utility.NBTProcessors;

import net.fabricmc.fabric.api.transfer.v1.item.ItemVariant;
import net.minecraft.class_1297;
import net.minecraft.class_1531;
import net.minecraft.class_1533;
import net.minecraft.class_1792;
import net.minecraft.class_1799;
import net.minecraft.class_1802;
import net.minecraft.class_2185;
import net.minecraft.class_2246;
import net.minecraft.class_2248;
import net.minecraft.class_2344;
import net.minecraft.class_2369;
import net.minecraft.class_2472;
import net.minecraft.class_2488;
import net.minecraft.class_2542;
import net.minecraft.class_2573;
import net.minecraft.class_2586;
import net.minecraft.class_2680;
import net.minecraft.class_2741;
import net.minecraft.class_2771;

public class ItemRequirement {
	public static final ItemRequirement NONE = new ItemRequirement(Collections.emptyList());
	public static final ItemRequirement INVALID = new ItemRequirement(Collections.emptyList());

	protected List<StackRequirement> requiredItems;

	public ItemRequirement(List<StackRequirement> requiredItems) {
		this.requiredItems = requiredItems;
	}

	public ItemRequirement(StackRequirement stackRequirement) {
		this(List.of(stackRequirement));
	}

	public ItemRequirement(ItemUseType usage, class_1799 stack) {
		this(new StackRequirement(stack, usage));
	}

	public ItemRequirement(ItemUseType usage, class_1792 item) {
		this(usage, new class_1799(item));
	}

	public ItemRequirement(ItemUseType usage, List<class_1799> requiredItems) {
		this(requiredItems.stream()
			.map(req -> new StackRequirement(req, usage))
			.collect(Collectors.toList()));
	}

	public static ItemRequirement of(class_2680 state, class_2586 be) {
		class_2248 block = state.method_26204();

		ItemRequirement requirement;
		if (block instanceof ISpecialBlockItemRequirement specialBlock) {
			requirement = specialBlock.getRequiredItems(state, be);
		} else {
			requirement = defaultOf(state, be);
		}

		if (be instanceof ISpecialBlockEntityItemRequirement specialBE)
			requirement = requirement.union(specialBE.getRequiredItems(state));

		return requirement;
	}

	private static ItemRequirement defaultOf(class_2680 state, class_2586 be) {
		class_2248 block = state.method_26204();
		if (block == class_2246.field_10124)
			return NONE;

		class_1792 item = block.method_8389();
		if (item == class_1802.field_8162)
			return INVALID;

		// double slab needs two items
		if (state.method_28498(class_2741.field_12485)
			&& state.method_11654(class_2741.field_12485) == class_2771.field_12682)
			return new ItemRequirement(ItemUseType.CONSUME, new class_1799(item, 2));
		if (block instanceof class_2542)
			return new ItemRequirement(ItemUseType.CONSUME, new class_1799(item, state.method_11654(class_2542.field_11710)
				.intValue()));
		if (block instanceof class_2472)
			return new ItemRequirement(ItemUseType.CONSUME, new class_1799(item, state.method_11654(class_2472.field_11472)
				.intValue()));
		if (block instanceof class_2488)
			return new ItemRequirement(ItemUseType.CONSUME, new class_1799(item, state.method_11654(class_2488.field_11518)
				.intValue()));
		if (block instanceof class_2344 || block instanceof class_2369)
			return new ItemRequirement(ItemUseType.CONSUME, class_1802.field_8831);
		if (block instanceof class_2185 && be instanceof class_2573 bannerBE)
			return new ItemRequirement(new StrictNbtStackRequirement(bannerBE.method_10907(), ItemUseType.CONSUME));

		return new ItemRequirement(ItemUseType.CONSUME, item);
	}

	public static ItemRequirement of(class_1297 entity) {
		if (entity instanceof ISpecialEntityItemRequirement specialEntity)
			return specialEntity.getRequiredItems();

		if (entity instanceof class_1533 itemFrame) {
			class_1799 frame = new class_1799(class_1802.field_8143);
			class_1799 displayedItem = NBTProcessors.withUnsafeNBTDiscarded(itemFrame.method_6940());
			if (displayedItem.method_7960())
				return new ItemRequirement(ItemUseType.CONSUME, class_1802.field_8143);
			return new ItemRequirement(List.of(new ItemRequirement.StackRequirement(frame, ItemUseType.CONSUME),
				new ItemRequirement.StrictNbtStackRequirement(displayedItem, ItemUseType.CONSUME)));
		}

		if (entity instanceof class_1531 armorStand) {
			List<StackRequirement> requirements = new ArrayList<>();
			requirements.add(new StackRequirement(new class_1799(class_1802.field_8694), ItemUseType.CONSUME));
			armorStand.method_5743()
				.forEach(s -> requirements
					.add(new StrictNbtStackRequirement(NBTProcessors.withUnsafeNBTDiscarded(s), ItemUseType.CONSUME)));
			return new ItemRequirement(requirements);
		}

		return INVALID;
	}

	public boolean isEmpty() {
		return NONE == this;
	}

	public boolean isInvalid() {
		return INVALID == this;
	}

	public List<StackRequirement> getRequiredItems() {
		return requiredItems;
	}

	public ItemRequirement union(ItemRequirement other) {
		if (this.isInvalid() || other.isInvalid())
			return INVALID;
		if (this.isEmpty())
			return other;
		if (other.isEmpty())
			return this;

		return new ItemRequirement(Stream.concat(requiredItems.stream(), other.requiredItems.stream())
			.collect(Collectors.toList()));
	}

	public enum ItemUseType {
		CONSUME, DAMAGE
	}

	public static class StackRequirement {
		public final class_1799 stack;
		public final ItemUseType usage;

		public StackRequirement(class_1799 stack, ItemUseType usage) {
			this.stack = stack;
			this.usage = usage;
		}

		public boolean matches(class_1799 other) {
			return class_1799.method_7984(stack, other);
		}

		public boolean matches(ItemVariant variant) {
			return variant.getItem() == stack.method_7909();
		}
	}

	public static class StrictNbtStackRequirement extends StackRequirement {
		public StrictNbtStackRequirement(class_1799 stack, ItemUseType usage) {
			super(stack, usage);
		}

		@Override
		public boolean matches(class_1799 other) {
			return class_1799.method_31577(stack, other);
		}

		@Override
		public boolean matches(ItemVariant variant) {
			return variant.matches(stack);
		}
	}
}
