package com.simibubi.create.content.schematics.table;

import static com.simibubi.create.foundation.gui.AllGuiTextures.PLAYER_INVENTORY;
import static com.simibubi.create.foundation.gui.AllGuiTextures.SCHEMATIC_TABLE_PROGRESS;

import java.nio.file.Paths;
import java.util.Collections;
import java.util.List;
import net.minecraft.class_156;
import net.minecraft.class_1661;
import net.minecraft.class_1799;
import net.minecraft.class_2561;
import net.minecraft.class_332;
import net.minecraft.class_3532;
import net.minecraft.class_768;
import com.google.common.collect.ImmutableList;
import com.simibubi.create.AllBlocks;
import com.simibubi.create.CreateClient;
import com.simibubi.create.content.schematics.client.ClientSchematicLoader;
import com.simibubi.create.foundation.gui.AllGuiTextures;
import com.simibubi.create.foundation.gui.AllIcons;
import com.simibubi.create.foundation.gui.element.GuiGameElement;
import com.simibubi.create.foundation.gui.menu.AbstractSimiContainerScreen;
import com.simibubi.create.foundation.gui.widget.IconButton;
import com.simibubi.create.foundation.gui.widget.Label;
import com.simibubi.create.foundation.gui.widget.ScrollInput;
import com.simibubi.create.foundation.gui.widget.SelectionScrollInput;
import com.simibubi.create.foundation.utility.Components;
import com.simibubi.create.foundation.utility.Lang;

public class SchematicTableScreen extends AbstractSimiContainerScreen<SchematicTableMenu> {

	private final class_2561 uploading = Lang.translateDirect("gui.schematicTable.uploading");
	private final class_2561 finished = Lang.translateDirect("gui.schematicTable.finished");
	private final class_2561 refresh = Lang.translateDirect("gui.schematicTable.refresh");
	private final class_2561 folder = Lang.translateDirect("gui.schematicTable.open_folder");
	private final class_2561 noSchematics = Lang.translateDirect("gui.schematicTable.noSchematics");
	private final class_2561 availableSchematicsTitle = Lang.translateDirect("gui.schematicTable.availableSchematics");

	protected AllGuiTextures background;

	private ScrollInput schematicsArea;
	private IconButton confirmButton;
	private IconButton folderButton;
	private IconButton refreshButton;
	private Label schematicsLabel;

	private float progress;
	private float chasingProgress;
	private float lastChasingProgress;

	private final class_1799 renderedItem = AllBlocks.SCHEMATIC_TABLE.asStack();

	private List<class_768> extraAreas = Collections.emptyList();

	public SchematicTableScreen(SchematicTableMenu menu, class_1661 playerInventory,
		class_2561 title) {
		super(menu, playerInventory, title);
		background = AllGuiTextures.SCHEMATIC_TABLE;
	}

	@Override
	protected void method_25426() {
		setWindowSize(background.width, background.height + 4 + AllGuiTextures.PLAYER_INVENTORY.height);
		setWindowOffset(-11, 8);
		super.method_25426();

		CreateClient.SCHEMATIC_SENDER.refresh();
		List<class_2561> availableSchematics = CreateClient.SCHEMATIC_SENDER.getAvailableSchematics();

		int x = field_2776;
		int y = field_2800;

		schematicsLabel = new Label(x + 49, y + 26, Components.immutableEmpty()).withShadow();
		schematicsLabel.text = Components.immutableEmpty();
		if (!availableSchematics.isEmpty()) {
			schematicsArea =
				new SelectionScrollInput(x + 45, y + 21, 139, 18).forOptions(availableSchematics)
					.titled(availableSchematicsTitle.method_27662())
					.writingTo(schematicsLabel);
			method_37063(schematicsArea);
			method_37063(schematicsLabel);
		}

		confirmButton = new IconButton(x + 44, y + 56, AllIcons.I_CONFIRM);
		confirmButton.withCallback(() -> {
			if (field_2797.canWrite() && schematicsArea != null) {
				ClientSchematicLoader schematicSender = CreateClient.SCHEMATIC_SENDER;
				lastChasingProgress = chasingProgress = progress = 0;
				List<class_2561> availableSchematics1 = schematicSender.getAvailableSchematics();
				class_2561 schematic = availableSchematics1.get(schematicsArea.getState());
				schematicSender.startNewUpload(schematic.getString());
			}
		});

		folderButton = new IconButton(x + 21, y + 21, AllIcons.I_OPEN_FOLDER);
		folderButton.withCallback(() -> {
			class_156.method_668()
				.method_672(Paths.get("schematics/")
					.toFile());
		});
		folderButton.setToolTip(folder);
		refreshButton = new IconButton(x + 207, y + 21, AllIcons.I_REFRESH);
		refreshButton.withCallback(() -> {
			ClientSchematicLoader schematicSender = CreateClient.SCHEMATIC_SENDER;
			schematicSender.refresh();
			List<class_2561> availableSchematics1 = schematicSender.getAvailableSchematics();
			method_37066(schematicsArea);

			if (!availableSchematics1.isEmpty()) {
				schematicsArea = new SelectionScrollInput(field_2776 + 45, field_2800 + 21, 139, 18)
					.forOptions(availableSchematics1)
					.titled(availableSchematicsTitle.method_27662())
					.writingTo(schematicsLabel);
				schematicsArea.onChanged();
				method_37063(schematicsArea);
			} else {
				schematicsArea = null;
				schematicsLabel.text = Components.immutableEmpty();
			}
		});
		refreshButton.setToolTip(refresh);

		method_37063(confirmButton);
		method_37063(folderButton);
		method_37063(refreshButton);

		extraAreas = ImmutableList.of(
			new class_768(x + background.width, y + background.height - 40, 48, 48),
			new class_768(refreshButton.method_46426(), refreshButton.method_46427(), refreshButton.method_25368(), refreshButton.method_25364())
		);
	}

	@Override
	protected void method_2389(class_332 graphics, float partialTicks, int mouseX, int mouseY) {
		int invX = getLeftOfCentered(PLAYER_INVENTORY.width);
		int invY = field_2800 + background.height + 4;
		renderPlayerInventory(graphics, invX, invY);

		int x = field_2776;
		int y = field_2800;

		background.render(graphics, x, y);

		class_2561 titleText;
		if (field_2797.contentHolder.isUploading)
			titleText = uploading;
		else if (field_2797.method_7611(1)
			.method_7681())
			titleText = finished;
		else
			titleText = field_22785;
		graphics.method_27534(field_22793, titleText, x + (background.width - 8) / 2, y + 3, 0xFFFFFF);

		if (schematicsArea == null)
			graphics.method_27535(field_22793, noSchematics, x + 54, y + 26, 0xD3D3D3);

		GuiGameElement.of(renderedItem)
			.<GuiGameElement.GuiRenderBuilder>at(x + background.width, y + background.height - 40, -200)
			.scale(3)
			.render(graphics);

		int width = (int) (SCHEMATIC_TABLE_PROGRESS.width
			* class_3532.method_16439(partialTicks, lastChasingProgress, chasingProgress));
		int height = SCHEMATIC_TABLE_PROGRESS.height;
		graphics.method_25302(SCHEMATIC_TABLE_PROGRESS.location, x + 70, y + 57, SCHEMATIC_TABLE_PROGRESS.startX,
			SCHEMATIC_TABLE_PROGRESS.startY, width, height);
	}

	@Override
	protected void method_37432() {
		super.method_37432();

		boolean finished = field_2797.method_7611(1)
			.method_7681();

		if (field_2797.contentHolder.isUploading || finished) {
			if (finished) {
				chasingProgress = lastChasingProgress = progress = 1;
			} else {
				lastChasingProgress = chasingProgress;
				progress = field_2797.contentHolder.uploadingProgress;
				chasingProgress += (progress - chasingProgress) * .5f;
			}
			confirmButton.field_22763 = false;

			if (schematicsLabel != null) {
				schematicsLabel.colored(0xCCDDFF);
				String uploadingSchematic = field_2797.contentHolder.uploadingSchematic;
				schematicsLabel.text = uploadingSchematic == null ? null : Components.literal(uploadingSchematic);
			}
			if (schematicsArea != null)
				schematicsArea.field_22764 = false;

		} else {
			progress = 0;
			chasingProgress = lastChasingProgress = 0;
			confirmButton.field_22763 = true;

			if (schematicsLabel != null)
				schematicsLabel.colored(0xFFFFFF);
			if (schematicsArea != null) {
				schematicsArea.writingTo(schematicsLabel);
				schematicsArea.field_22764 = true;
			}
		}
	}

	@Override
	public List<class_768> getExtraAreas() {
		return extraAreas;
	}

}
