package com.simibubi.create.content.trains.bogey;

import static com.simibubi.create.content.trains.entity.CarriageBogey.UPSIDE_DOWN_KEY;

import org.jetbrains.annotations.NotNull;

import com.simibubi.create.AllBogeyStyles;
import com.simibubi.create.foundation.blockEntity.CachedRenderBBBlockEntity;
import com.simibubi.create.foundation.utility.NBTHelper;
import com.simibubi.create.foundation.utility.animation.LerpedFloat;
import net.minecraft.class_1937;
import net.minecraft.class_2338;
import net.minecraft.class_238;
import net.minecraft.class_2487;
import net.minecraft.class_2591;
import net.minecraft.class_2680;
import net.minecraft.class_2960;

public abstract class AbstractBogeyBlockEntity extends CachedRenderBBBlockEntity {
	public static final String BOGEY_STYLE_KEY = "BogeyStyle";
	public static final String BOGEY_DATA_KEY = "BogeyData";

	private class_2487 bogeyData;

	public AbstractBogeyBlockEntity(class_2591<?> type, class_2338 pos, class_2680 state) {
		super(type, pos, state);
	}

	public abstract BogeyStyle getDefaultStyle();

	public class_2487 getBogeyData() {
		if (this.bogeyData == null || !this.bogeyData.method_10545(BOGEY_STYLE_KEY))
			this.bogeyData = this.createBogeyData();
		return this.bogeyData;
	}

	public void setBogeyData(@NotNull class_2487 newData) {
		if (!newData.method_10545(BOGEY_STYLE_KEY)) {
			class_2960 style = getDefaultStyle().name;
			NBTHelper.writeResourceLocation(newData, BOGEY_STYLE_KEY, style);
		}
		this.bogeyData = newData;
	}

	public void setBogeyStyle(@NotNull BogeyStyle style) {
		class_2960 location = style.name;
		class_2487 data = this.getBogeyData();
		NBTHelper.writeResourceLocation(data, BOGEY_STYLE_KEY, location);
		markUpdated();
	}

	@NotNull
	public BogeyStyle getStyle() {
		class_2487 data = this.getBogeyData();
		class_2960 currentStyle = NBTHelper.readResourceLocation(data, BOGEY_STYLE_KEY);
		BogeyStyle style = AllBogeyStyles.BOGEY_STYLES.get(currentStyle);
		if (style == null) {
			setBogeyStyle(getDefaultStyle());
			return getStyle();
		}
		return style;
	}

	@Override
	protected void method_11007(@NotNull class_2487 pTag) {
		class_2487 data = this.getBogeyData();
		if (data != null) pTag.method_10566(BOGEY_DATA_KEY, data); // Now contains style
		super.method_11007(pTag);
	}

	@Override
	public void method_11014(class_2487 pTag) {
		if (pTag.method_10545(BOGEY_DATA_KEY))
			this.bogeyData = pTag.method_10562(BOGEY_DATA_KEY);
		else
			this.bogeyData = this.createBogeyData();
		super.method_11014(pTag);
	}

	private class_2487 createBogeyData() {
		class_2487 nbt = new class_2487();
		NBTHelper.writeResourceLocation(nbt, BOGEY_STYLE_KEY, getDefaultStyle().name);
		boolean upsideDown = false;
		if (method_11010().method_26204() instanceof AbstractBogeyBlock<?> bogeyBlock)
			upsideDown = bogeyBlock.isUpsideDown(method_11010());
		nbt.method_10556(UPSIDE_DOWN_KEY, upsideDown);
		return nbt;
	}

	@Override
	protected class_238 createRenderBoundingBox() {
		return super.createRenderBoundingBox().method_1014(2);
	}

	// Ponder
	LerpedFloat virtualAnimation = LerpedFloat.angular();

	public float getVirtualAngle(float partialTicks) {
		return virtualAnimation.getValue(partialTicks);
	}

	public void animate(float distanceMoved) {
		class_2680 blockState = method_11010();
		if (!(blockState.method_26204() instanceof AbstractBogeyBlock<?> type))
			return;
		double angleDiff = 360 * distanceMoved / (Math.PI * 2 * type.getWheelRadius());
		double newWheelAngle = (virtualAnimation.getValue() - angleDiff) % 360;
		virtualAnimation.setValue(newWheelAngle);
	}

	private void markUpdated() {
		method_5431();
		class_1937 level = method_10997();
		if (level != null)
			method_10997().method_8413(method_11016(), method_11010(), method_11010(), 3);
	}
}
