package com.simibubi.create.content.trains.display;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import net.minecraft.class_1767;
import net.minecraft.class_2338;
import net.minecraft.class_2338.class_2339;
import net.minecraft.class_2350;
import net.minecraft.class_238;
import net.minecraft.class_2382;
import net.minecraft.class_2487;
import net.minecraft.class_2561;
import net.minecraft.class_2586;
import net.minecraft.class_2591;
import net.minecraft.class_2680;
import net.minecraft.class_3417;
import net.minecraft.class_3419;
import net.minecraft.class_3532;
import com.google.gson.JsonElement;
import com.simibubi.create.AllSoundEvents;
import com.simibubi.create.content.kinetics.base.KineticBlockEntity;
import com.simibubi.create.foundation.blockEntity.behaviour.BlockEntityBehaviour;
import com.simibubi.create.foundation.utility.Components;
import com.simibubi.create.foundation.utility.DyeHelper;
import com.simibubi.create.foundation.utility.DynamicComponent;
import com.simibubi.create.foundation.utility.NBTHelper;

public class FlapDisplayBlockEntity extends KineticBlockEntity {

	public List<FlapDisplayLayout> lines;
	public boolean isController;
	public boolean isRunning;
	public int xSize, ySize;
	public class_1767[] colour;
	public boolean[] glowingLines;
	public boolean[] manualLines;

	public FlapDisplayBlockEntity(class_2591<?> type, class_2338 pos, class_2680 state) {
		super(type, pos, state);
		setLazyTickRate(10);
		isController = false;
		xSize = 1;
		ySize = 1;
		colour = new class_1767[2];
		manualLines = new boolean[2];
		glowingLines = new boolean[2];
	}

	@Override
	public void initialize() {
		super.initialize();
	}

	@Override
	public void lazyTick() {
		super.lazyTick();
		updateControllerStatus();
	}

	public void updateControllerStatus() {
		if (field_11863.field_9236)
			return;

		class_2680 blockState = method_11010();
		if (!(blockState.method_26204() instanceof FlapDisplayBlock))
			return;

		class_2350 leftDirection = blockState.method_11654(FlapDisplayBlock.HORIZONTAL_FACING)
			.method_10170();
		boolean shouldBeController = !blockState.method_11654(FlapDisplayBlock.UP)
			&& field_11863.method_8320(field_11867.method_10093(leftDirection)) != blockState;

		int newXSize = 1;
		int newYSize = 1;

		if (shouldBeController) {
			for (int xOffset = 1; xOffset < 32; xOffset++) {
				if (field_11863.method_8320(field_11867.method_10079(leftDirection.method_10153(), xOffset)) != blockState)
					break;
				newXSize++;
			}
			for (int yOffset = 0; yOffset < 32; yOffset++) {
				if (!field_11863.method_8320(field_11867.method_10079(class_2350.field_11033, yOffset))
					.method_28500(FlapDisplayBlock.DOWN)
					.orElse(false))
					break;
				newYSize++;
			}
		}

		if (isController == shouldBeController && newXSize == xSize && newYSize == ySize)
			return;

		isController = shouldBeController;
		xSize = newXSize;
		ySize = newYSize;
		colour = Arrays.copyOf(colour, ySize * 2);
		glowingLines = Arrays.copyOf(glowingLines, ySize * 2);
		manualLines = new boolean[ySize * 2];
		lines = null;
		sendData();
	}

	@Override
	public void tick() {
		super.tick();
		isRunning = super.isSpeedRequirementFulfilled();
		if ((!field_11863.field_9236 || !isRunning) && !isVirtual())
			return;
		int activeFlaps = 0;
		boolean instant = getSpeed() > 128;
		for (FlapDisplayLayout line : lines)
			for (FlapDisplaySection section : line.getSections())
				activeFlaps += section.tick(instant);
		if (activeFlaps == 0)
			return;

		float volume = class_3532.method_15363(activeFlaps / 20f, 0.25f, 1.5f);
		float bgVolume = class_3532.method_15363(activeFlaps / 40f, 0.25f, 1f);
		class_2338 middle = field_11867.method_10079(getDirection().method_10170(), xSize / 2)
			.method_10079(class_2350.field_11033, ySize / 2);
		AllSoundEvents.SCROLL_VALUE.playAt(field_11863, middle, volume, 0.56f, false);
		field_11863.method_8486(middle.method_10263(), middle.method_10264(), middle.method_10260(), class_3417.field_26951, class_3419.field_15245,
			.35f * bgVolume, 1.95f, false);
	}

	@Override
	protected boolean isNoisy() {
		return false;
	}

	@Override
	public boolean isSpeedRequirementFulfilled() {
		return isRunning;
	}

	public void applyTextManually(int lineIndex, String rawComponentText) {
		List<FlapDisplayLayout> lines = getLines();
		if (lineIndex >= lines.size())
			return;
		
		FlapDisplayLayout layout = lines.get(lineIndex);
		if (!layout.isLayout("Default"))
			layout.loadDefault(getMaxCharCount());
		List<FlapDisplaySection> sections = layout.getSections();

		FlapDisplaySection flapDisplaySection = sections.get(0);
		if (rawComponentText == null) {
			manualLines[lineIndex] = false;
			flapDisplaySection.setText(Components.immutableEmpty());
			notifyUpdate();
			return;
		}

		JsonElement json = DynamicComponent.getJsonFromString(rawComponentText);
		if (json == null)
			return;

		manualLines[lineIndex] = true;
		class_2561 text = isVirtual() ? class_2561.class_2562.method_10877(rawComponentText)
			: DynamicComponent.parseCustomText(field_11863, field_11867, json);
		flapDisplaySection.setText(text);
		if (isVirtual())
			flapDisplaySection.refresh(true);
		else
			notifyUpdate();
	}

	public void setColour(int lineIndex, class_1767 color) {
		colour[lineIndex] = color == class_1767.field_7952 ? null : color;
		notifyUpdate();
	}
	
	public void setGlowing(int lineIndex) {
		glowingLines[lineIndex] = true;
		notifyUpdate();
	}

	public List<FlapDisplayLayout> getLines() {
		if (lines == null)
			initDefaultSections();
		return lines;
	}

	public void initDefaultSections() {
		lines = new ArrayList<>();
		for (int i = 0; i < ySize * 2; i++)
			lines.add(new FlapDisplayLayout(getMaxCharCount()));
	}

	public int getMaxCharCount() {
		return getMaxCharCount(0);
	}

	public int getMaxCharCount(int gaps) {
		return (int) ((xSize * 16f - 2f - 4f * gaps) / 3.5f);
	}

	@Override
	protected void write(class_2487 tag, boolean clientPacket) {
		super.write(tag, clientPacket);

		tag.method_10556("Controller", isController);
		tag.method_10569("XSize", xSize);
		tag.method_10569("YSize", ySize);

		for (int j = 0; j < manualLines.length; j++)
			if (manualLines[j])
				NBTHelper.putMarker(tag, "CustomLine" + j);
		
		for (int j = 0; j < glowingLines.length; j++)
			if (glowingLines[j])
				NBTHelper.putMarker(tag, "GlowingLine" + j);

		for (int j = 0; j < colour.length; j++)
			if (colour[j] != null)
				NBTHelper.writeEnum(tag, "Dye" + j, colour[j]);

		List<FlapDisplayLayout> lines = getLines();
		for (int i = 0; i < lines.size(); i++)
			tag.method_10566("Display" + i, lines.get(i)
				.write());
	}

	@Override
	protected void read(class_2487 tag, boolean clientPacket) {
		super.read(tag, clientPacket);
		boolean wasActive = isController;
		int prevX = xSize;
		int prevY = ySize;

		isController = tag.method_10577("Controller");
		xSize = tag.method_10550("XSize");
		ySize = tag.method_10550("YSize");

		manualLines = new boolean[ySize * 2];
		for (int i = 0; i < ySize * 2; i++)
			manualLines[i] = tag.method_10545("CustomLine" + i);
		
		glowingLines = new boolean[ySize * 2];
		for (int i = 0; i < ySize * 2; i++)
			glowingLines[i] = tag.method_10545("GlowingLine" + i);

		colour = new class_1767[ySize * 2];
		for (int i = 0; i < ySize * 2; i++)
			colour[i] = tag.method_10545("Dye" + i) ? NBTHelper.readEnum(tag, "Dye" + i, class_1767.class) : null;

		if (clientPacket && wasActive != isController || prevX != xSize || prevY != ySize) {
			invalidateRenderBoundingBox();
			lines = null;
		}

		List<FlapDisplayLayout> lines = getLines();
		for (int i = 0; i < lines.size(); i++)
			lines.get(i)
				.read(tag.method_10562("Display" + i));
	}

	public int getLineIndexAt(double yCoord) {
		return (int) class_3532.method_15350(Math.floor(2 * (field_11867.method_10264() - yCoord + 1)), 0, ySize * 2);
	}

	public FlapDisplayBlockEntity getController() {
		if (isController)
			return this;

		class_2680 blockState = method_11010();
		if (!(blockState.method_26204() instanceof FlapDisplayBlock))
			return null;

		class_2339 pos = method_11016().method_25503();
		class_2350 side = blockState.method_11654(FlapDisplayBlock.HORIZONTAL_FACING)
			.method_10170();

		for (int i = 0; i < 64; i++) {
			class_2680 other = field_11863.method_8320(pos);

			if (other.method_28500(FlapDisplayBlock.UP)
				.orElse(false)) {
				pos.method_10098(class_2350.field_11036);
				continue;
			}

			if (!field_11863.method_8320(pos.method_10093(side))
				.method_28500(FlapDisplayBlock.UP)
				.orElse(true)) {
				pos.method_10098(side);
				continue;
			}

			class_2586 found = field_11863.method_8321(pos);
			if (found instanceof FlapDisplayBlockEntity flap && flap.isController)
				return flap;

			break;
		}

		return null;
	}

	@Override
	protected class_238 createRenderBoundingBox() {
		class_238 aabb = new class_238(field_11867);
		if (!isController)
			return aabb;
		class_2382 normal = getDirection().method_10170()
			.method_10163();
		return aabb.method_1012(normal.method_10263() * xSize, -ySize, normal.method_10260() * xSize);
	}

	public class_2350 getDirection() {
		return method_11010().method_28500(FlapDisplayBlock.HORIZONTAL_FACING)
			.orElse(class_2350.field_11035)
			.method_10153();
	}

	@Override
	public void addBehaviours(List<BlockEntityBehaviour> behaviours) {}

	public int getLineColor(int line) {
		class_1767 color = colour[line];
		return color == null ? 0xFF_D3C6BA
			: DyeHelper.DYE_TABLE.get(color)
				.getFirst() | 0xFF_000000;
	}
	
	public boolean isLineGlowing(int line) {
		return glowingLines[line];
	}
	
}
