package com.simibubi.create.content.trains.entity;

import java.util.ArrayList;
import java.util.HashMap;

import javax.annotation.Nullable;
import net.minecraft.class_2338;
import net.minecraft.class_2428;
import net.minecraft.class_2487;
import net.minecraft.class_2512;
import net.minecraft.class_2520;
import net.minecraft.class_2680;
import net.minecraft.class_3414;
import net.minecraft.class_3417;
import net.minecraft.class_3419;
import net.minecraft.class_3499.class_3501;
import net.minecraft.class_3709;
import com.google.common.collect.Multimap;
import com.google.common.collect.Multimaps;
import com.simibubi.create.AllBlocks;
import com.simibubi.create.AllSoundEvents;
import com.simibubi.create.content.contraptions.Contraption;
import com.simibubi.create.content.decoration.steamWhistle.WhistleBlock;
import com.simibubi.create.content.decoration.steamWhistle.WhistleBlock.WhistleSize;
import com.simibubi.create.foundation.utility.NBTHelper;
import com.simibubi.create.foundation.utility.Pair;

public class ArrivalSoundQueue {

	public int offset;
	int min, max;
	Multimap<Integer, class_2338> sources;

	public ArrivalSoundQueue() {
		sources = Multimaps.newMultimap(new HashMap<>(), ArrayList::new);
		min = Integer.MAX_VALUE;
		max = Integer.MIN_VALUE;
	}

	@Nullable
	public Integer firstTick() {
		return sources.isEmpty() ? null : min + offset;
	}

	@Nullable
	public Integer lastTick() {
		return sources.isEmpty() ? null : max + offset;
	}

	public boolean tick(CarriageContraptionEntity entity, int tick, boolean backwards) {
		tick = tick - offset;
		if (!sources.containsKey(tick))
			return backwards ? tick > min : tick < max;
		Contraption contraption = entity.getContraption();
		for (class_2338 blockPos : sources.get(tick))
			play(entity, contraption.getBlocks()
				.get(blockPos));
		return backwards ? tick > min : tick < max;
	}

	public Pair<Boolean, Integer> getFirstWhistle(CarriageContraptionEntity entity) {
		Integer firstTick = firstTick();
		Integer lastTick = lastTick();
		if (firstTick == null || lastTick == null || firstTick > lastTick)
			return null;
		for (int i = firstTick; i <= lastTick; i++) {
			if (!sources.containsKey(i - offset))
				continue;
			Contraption contraption = entity.getContraption();
			for (class_2338 blockPos : sources.get(i - offset)) {
				class_3501 info = contraption.getBlocks()
					.get(blockPos);
				if (info == null)
					continue;
				class_2680 state = info.comp_1342();
				if (state.method_26204() instanceof WhistleBlock && info.comp_1343() != null) {
					int pitch = info.comp_1343().method_10550("Pitch");
					WhistleSize size = state.method_11654(WhistleBlock.SIZE);
					return Pair.of(size == WhistleSize.LARGE, (size == WhistleSize.SMALL ? 12 : 0) - pitch);
				}
			}
		}
		return null;
	}

	public void serialize(class_2487 tagIn) {
		class_2487 tag = new class_2487();
		tag.method_10569("Offset", offset);
		tag.method_10566("Sources", NBTHelper.writeCompoundList(sources.entries(), e -> {
			class_2487 c = new class_2487();
			c.method_10569("Tick", e.getKey());
			c.method_10566("Pos", class_2512.method_10692(e.getValue()));
			return c;
		}));
		tagIn.method_10566("SoundQueue", tag);
	}

	public void deserialize(class_2487 tagIn) {
		class_2487 tag = tagIn.method_10562("SoundQueue");
		offset = tag.method_10550("Offset");
		NBTHelper.iterateCompoundList(tag.method_10554("Sources", class_2520.field_33260),
			c -> add(c.method_10550("Tick"), class_2512.method_10691(c.method_10562("Pos"))));
	}

	public void add(int offset, class_2338 localPos) {
		sources.put(offset, localPos);
		min = Math.min(offset, min);
		max = Math.max(offset, max);
	}

	public static boolean isPlayable(class_2680 state) {
		if (state.method_26204() instanceof class_3709)
			return true;
		if (state.method_26204() instanceof class_2428)
			return true;
		if (state.method_26204() instanceof WhistleBlock)
			return true;
		return false;
	}

	public static void play(CarriageContraptionEntity entity, class_3501 info) {
		if (info == null)
			return;
		class_2680 state = info.comp_1342();

		if (state.method_26204() instanceof class_3709) {
			if (AllBlocks.HAUNTED_BELL.has(state))
				playSimple(entity, AllSoundEvents.HAUNTED_BELL_USE.getMainEvent(), 1, 1);
			else
				playSimple(entity, class_3417.field_17265, 1, 1);
		}

		if (state.method_26204() instanceof class_2428 nb) {
			float f = (float) Math.pow(2, (state.method_11654(class_2428.field_11324) - 12) / 12.0);
			playSimple(entity, state.method_11654(class_2428.field_11325)
				.method_11886()
				.comp_349(), 1, f);
		}

		if (state.method_26204() instanceof WhistleBlock && info.comp_1343() != null) {
			int pitch = info.comp_1343().method_10550("Pitch");
			WhistleSize size = state.method_11654(WhistleBlock.SIZE);
			float f = (float) Math.pow(2, ((size == WhistleSize.SMALL ? 12 : 0) - pitch) / 12.0);
			playSimple(entity,
				(size == WhistleSize.LARGE ? AllSoundEvents.WHISTLE_TRAIN_LOW : AllSoundEvents.WHISTLE_TRAIN)
					.getMainEvent(),
				1, f);
//			playSimple(entity, AllSoundEvents.WHISTLE_CHIFF.getMainEvent(), .75f,
//				size == WhistleSize.SMALL ? f + .75f : f);
		}

	}

	private static void playSimple(CarriageContraptionEntity entity, class_3414 event, float volume, float pitch) {
		entity.method_37908().method_43129(null, entity, event, class_3419.field_15254, 5 * volume, pitch);
	}

}
