package com.simibubi.create.content.trains.entity;

import java.lang.ref.WeakReference;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Objects;
import java.util.Set;
import java.util.UUID;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.function.Consumer;
import java.util.function.Function;

import javax.annotation.Nullable;

import org.apache.commons.lang3.mutable.MutableDouble;

import com.simibubi.create.content.contraptions.Contraption;
import com.simibubi.create.content.contraptions.minecart.TrainCargoManager;
import com.simibubi.create.content.trains.entity.TravellingPoint.IEdgePointListener;
import com.simibubi.create.content.trains.entity.TravellingPoint.ITrackSelector;
import com.simibubi.create.content.trains.graph.DimensionPalette;
import com.simibubi.create.content.trains.graph.TrackGraph;
import com.simibubi.create.content.trains.graph.TrackNodeLocation;
import com.simibubi.create.foundation.advancement.AllAdvancements;
import com.simibubi.create.foundation.utility.Couple;
import com.simibubi.create.foundation.utility.Iterate;
import com.simibubi.create.foundation.utility.NBTHelper;
import com.simibubi.create.foundation.utility.VecHelper;
import com.tterrag.registrate.fabric.EnvExecutor;

import io.github.fabricators_of_create.porting_lib.util.NBTSerializer;
import net.fabricmc.api.EnvType;
import net.fabricmc.api.Environment;
import net.minecraft.class_1297;
import net.minecraft.class_1299;
import net.minecraft.class_1657;
import net.minecraft.class_1937;
import net.minecraft.class_2338;
import net.minecraft.class_243;
import net.minecraft.class_2487;
import net.minecraft.class_2520;
import net.minecraft.class_3218;
import net.minecraft.class_3222;
import net.minecraft.class_3532;
import net.minecraft.class_5321;
import net.minecraft.server.MinecraftServer;

public class Carriage {

	public static final AtomicInteger netIdGenerator = new AtomicInteger();

	public Train train;
	public int id;
	public boolean blocked;
	public boolean stalled;
	public Couple<Boolean> presentConductors;

	public int bogeySpacing;
	public Couple<CarriageBogey> bogeys;
	public TrainCargoManager storage;

	class_2487 serialisedEntity;
	Map<Integer, class_2487> serialisedPassengers;

	private Map<class_5321<class_1937>, DimensionalCarriageEntity> entities;

	static final int FIRST = 0, MIDDLE = 1, LAST = 2, BOTH = 3;

	public Carriage(CarriageBogey bogey1, @Nullable CarriageBogey bogey2, int bogeySpacing) {
		this.bogeySpacing = bogeySpacing;
		this.bogeys = Couple.create(bogey1, bogey2);
		this.id = netIdGenerator.incrementAndGet();
		this.serialisedEntity = new class_2487();
		this.presentConductors = Couple.create(false, false);
		this.serialisedPassengers = new HashMap<>();
		this.entities = new HashMap<>();
		this.storage = new TrainCargoManager();

		bogey1.setLeading();
		bogey1.carriage = this;
		if (bogey2 != null)
			bogey2.carriage = this;
	}

	public boolean isOnIncompatibleTrack() {
		return leadingBogey().type.isOnIncompatibleTrack(this, true)
				|| trailingBogey().type.isOnIncompatibleTrack(this, false);
	}

	public void setTrain(Train train) {
		this.train = train;
	}

	public boolean presentInMultipleDimensions() {
		return entities.size() > 1;
	}

	public void setContraption(class_1937 level, CarriageContraption contraption) {
		this.storage = null;
		CarriageContraptionEntity entity = CarriageContraptionEntity.create(level, contraption);
		entity.setCarriage(this);
		contraption.startMoving(level);
		contraption.onEntityInitialize(level, entity);
		updateContraptionAnchors();

		DimensionalCarriageEntity dimensional = getDimensional(level);
		dimensional.alignEntity(entity);
		dimensional.removeAndSaveEntity(entity, true);
	}

	public DimensionalCarriageEntity getDimensional(class_1937 level) {
		return getDimensional(level.method_27983());
	}

	public DimensionalCarriageEntity getDimensional(class_5321<class_1937> dimension) {
		return entities.computeIfAbsent(dimension, $ -> new DimensionalCarriageEntity());
	}

	@Nullable
	public DimensionalCarriageEntity getDimensionalIfPresent(class_5321<class_1937> dimension) {
		return entities.get(dimension);
	}

	public double travel(class_1937 level, TrackGraph graph, double distance, TravellingPoint toFollowForward,
		TravellingPoint toFollowBackward, int type) {

		Function<TravellingPoint, ITrackSelector> forwardControl =
			toFollowForward == null ? train.navigation::control : mp -> mp.follow(toFollowForward);
		Function<TravellingPoint, ITrackSelector> backwardControl =
			toFollowBackward == null ? train.navigation::control : mp -> mp.follow(toFollowBackward);

		boolean onTwoBogeys = isOnTwoBogeys();
		double stress = train.derailed ? 0 : onTwoBogeys ? bogeySpacing - getAnchorDiff() : 0;
		blocked = false;

		MutableDouble distanceMoved = new MutableDouble(distance);
		boolean iterateFromBack = distance < 0;

		for (boolean firstBogey : Iterate.trueAndFalse) {
			if (!firstBogey && !onTwoBogeys)
				continue;

			boolean actuallyFirstBogey = !onTwoBogeys || (firstBogey ^ iterateFromBack);
			CarriageBogey bogey = bogeys.get(actuallyFirstBogey);
			double bogeyCorrection = stress * (actuallyFirstBogey ? 0.5d : -0.5d);
			double bogeyStress = bogey.getStress();

			for (boolean firstWheel : Iterate.trueAndFalse) {
				boolean actuallyFirstWheel = firstWheel ^ iterateFromBack;
				TravellingPoint point = bogey.points.get(actuallyFirstWheel);
				TravellingPoint prevPoint = !actuallyFirstWheel ? bogey.points.getFirst()
					: !actuallyFirstBogey && onTwoBogeys ? bogeys.getFirst().points.getSecond() : null;
				TravellingPoint nextPoint = actuallyFirstWheel ? bogey.points.getSecond()
					: actuallyFirstBogey && onTwoBogeys ? bogeys.getSecond().points.getFirst() : null;

				double correction = bogeyStress * (actuallyFirstWheel ? 0.5d : -0.5d);
				double toMove = distanceMoved.getValue();

				ITrackSelector frontTrackSelector =
					prevPoint == null ? forwardControl.apply(point) : point.follow(prevPoint);
				ITrackSelector backTrackSelector =
					nextPoint == null ? backwardControl.apply(point) : point.follow(nextPoint);

				boolean atFront = (type == FIRST || type == BOTH) && actuallyFirstWheel && actuallyFirstBogey;
				boolean atBack =
					(type == LAST || type == BOTH) && !actuallyFirstWheel && (!actuallyFirstBogey || !onTwoBogeys);

				IEdgePointListener frontListener = train.frontSignalListener();
				IEdgePointListener backListener = train.backSignalListener();
				IEdgePointListener passiveListener = point.ignoreEdgePoints();

				toMove += correction + bogeyCorrection;

				ITrackSelector trackSelector = toMove > 0 ? frontTrackSelector : backTrackSelector;
				IEdgePointListener signalListener =
					toMove > 0 ? atFront ? frontListener : atBack ? backListener : passiveListener
						: atFront ? backListener : atBack ? frontListener : passiveListener;

				double moved = point.travel(graph, toMove, trackSelector, signalListener, point.ignoreTurns(), c -> {
					for (DimensionalCarriageEntity dce : entities.values())
						if (c.either(tnl -> tnl.equalsIgnoreDim(dce.pivot)))
							return false;
					if (entities.size() > 1) {
						train.status.doublePortal();
						return true;
					}
					return false;
				});

				blocked |= point.blocked;

				distanceMoved.setValue(moved);
			}
		}

		updateContraptionAnchors();
		manageEntities(level);
		return distanceMoved.getValue();
	}

	public double getAnchorDiff() {
		double diff = 0;
		int entries = 0;

		TravellingPoint leadingPoint = getLeadingPoint();
		TravellingPoint trailingPoint = getTrailingPoint();
		if (leadingPoint.node1 != null && trailingPoint.node1 != null)
			if (!leadingPoint.node1.getLocation().dimension.equals(trailingPoint.node1.getLocation().dimension))
				return bogeySpacing;

		for (DimensionalCarriageEntity dce : entities.values())
			if (dce.leadingAnchor() != null && dce.trailingAnchor() != null) {
				entries++;
				diff += dce.leadingAnchor()
					.method_1022(dce.trailingAnchor());
			}

		if (entries == 0)
			return bogeySpacing;
		return diff / entries;
	}

	public void updateConductors() {
		if (anyAvailableEntity() == null || entities.size() > 1 || serialisedPassengers.size() > 0)
			return;
		presentConductors.replace($ -> false);
		for (DimensionalCarriageEntity dimensionalCarriageEntity : entities.values()) {
			CarriageContraptionEntity entity = dimensionalCarriageEntity.entity.get();
			if (entity != null && entity.method_5805())
				presentConductors.replaceWithParams((current, checked) -> current || checked, entity.checkConductors());
		}
	}

	private Set<class_5321<class_1937>> currentlyTraversedDimensions = new HashSet<>();

	public void manageEntities(class_1937 level) {
		currentlyTraversedDimensions.clear();

		bogeys.forEach(cb -> {
			if (cb == null)
				return;
			cb.points.forEach(tp -> {
				if (tp.node1 == null)
					return;
				currentlyTraversedDimensions.add(tp.node1.getLocation().dimension);
			});
		});

		for (Iterator<Entry<class_5321<class_1937>, DimensionalCarriageEntity>> iterator = entities.entrySet()
			.iterator(); iterator.hasNext();) {
			Entry<class_5321<class_1937>, DimensionalCarriageEntity> entry = iterator.next();

			boolean discard =
				!currentlyTraversedDimensions.isEmpty() && !currentlyTraversedDimensions.contains(entry.getKey());

			MinecraftServer server = level.method_8503();
			if (server == null)
				continue;
			class_3218 currentLevel = server.method_3847(entry.getKey());
			if (currentLevel == null)
				continue;

			DimensionalCarriageEntity dimensionalCarriageEntity = entry.getValue();
			CarriageContraptionEntity entity = dimensionalCarriageEntity.entity.get();

			if (entity == null) {
				if (discard)
					iterator.remove();
				else if (dimensionalCarriageEntity.positionAnchor != null && CarriageEntityHandler
					.isActiveChunk(currentLevel, class_2338.method_49638(dimensionalCarriageEntity.positionAnchor)))
					dimensionalCarriageEntity.createEntity(currentLevel, anyAvailableEntity() == null);

			} else {
				if (discard) {
					discard = dimensionalCarriageEntity.discardTicks > 3;
					dimensionalCarriageEntity.discardTicks++;
				} else
					dimensionalCarriageEntity.discardTicks = 0;

				CarriageEntityHandler.validateCarriageEntity(entity);
				if (!entity.method_5805() || entity.leftTickingChunks || discard) {
					dimensionalCarriageEntity.removeAndSaveEntity(entity, discard);
					if (discard)
						iterator.remove();
					continue;
				}
			}

			entity = dimensionalCarriageEntity.entity.get();
			if (entity != null && dimensionalCarriageEntity.positionAnchor != null) {
				dimensionalCarriageEntity.alignEntity(entity);
				entity.syncCarriage();
			}
		}

	}

	public void updateContraptionAnchors() {
		CarriageBogey leadingBogey = leadingBogey();
		if (leadingBogey.points.either(t -> t.edge == null))
			return;
		CarriageBogey trailingBogey = trailingBogey();
		if (trailingBogey.points.either(t -> t.edge == null))
			return;

		class_5321<class_1937> leadingBogeyDim = leadingBogey.getDimension();
		class_5321<class_1937> trailingBogeyDim = trailingBogey.getDimension();
		double leadingWheelSpacing = leadingBogey.type.getWheelPointSpacing();
		double trailingWheelSpacing = trailingBogey.type.getWheelPointSpacing();

		boolean leadingUpsideDown = leadingBogey.isUpsideDown();
		boolean trailingUpsideDown = trailingBogey.isUpsideDown();

		for (boolean leading : Iterate.trueAndFalse) {
			TravellingPoint point = leading ? getLeadingPoint() : getTrailingPoint();
			TravellingPoint otherPoint = !leading ? getLeadingPoint() : getTrailingPoint();
			class_5321<class_1937> dimension = point.node1.getLocation().dimension;
			class_5321<class_1937> otherDimension = otherPoint.node1.getLocation().dimension;

			if (dimension.equals(otherDimension) && leading) {
				getDimensional(dimension).discardPivot();
				continue;
			}

			DimensionalCarriageEntity dce = getDimensional(dimension);

			dce.positionAnchor = dimension.equals(leadingBogeyDim) ? leadingBogey.getAnchorPosition()
				: pivoted(dce, dimension, point,
					leading ? leadingWheelSpacing / 2 : bogeySpacing + trailingWheelSpacing / 2,
					leadingUpsideDown, trailingUpsideDown);

			boolean backAnchorFlip = trailingBogey.isUpsideDown() ^ leadingBogey.isUpsideDown();

			if (isOnTwoBogeys()) {
				dce.rotationAnchors.setFirst(dimension.equals(leadingBogeyDim) ? leadingBogey.getAnchorPosition()
					: pivoted(dce, dimension, point,
						leading ? leadingWheelSpacing / 2 : bogeySpacing + trailingWheelSpacing / 2,
						leadingUpsideDown, trailingUpsideDown));
				dce.rotationAnchors.setSecond(dimension.equals(trailingBogeyDim) ? trailingBogey.getAnchorPosition(backAnchorFlip)
					: pivoted(dce, dimension, point,
						leading ? leadingWheelSpacing / 2 + bogeySpacing : trailingWheelSpacing / 2,
						leadingUpsideDown, trailingUpsideDown));

			} else {
				if (dimension.equals(otherDimension)) {
					dce.rotationAnchors = leadingBogey.points.map(tp -> tp.getPosition(train.graph));
				} else {
					dce.rotationAnchors.setFirst(leadingBogey.points.getFirst() == point
						? point.getPosition(train.graph)
						: pivoted(dce, dimension, point, leadingWheelSpacing, leadingUpsideDown, trailingUpsideDown));
					dce.rotationAnchors.setSecond(leadingBogey.points.getSecond() == point
						? point.getPosition(train.graph)
						: pivoted(dce, dimension, point, leadingWheelSpacing, leadingUpsideDown, trailingUpsideDown));
				}
			}

			int prevmin = dce.minAllowedLocalCoord();
			int prevmax = dce.maxAllowedLocalCoord();

			dce.updateCutoff(leading);

			if (prevmin != dce.minAllowedLocalCoord() || prevmax != dce.maxAllowedLocalCoord()) {
				dce.updateRenderedCutoff();
				dce.updatePassengerLoadout();
			}
		}

	}

	private class_243 pivoted(DimensionalCarriageEntity dce, class_5321<class_1937> dimension, TravellingPoint start,
		double offset, boolean leadingUpsideDown, boolean trailingUpsideDown) {
		if (train.graph == null)
			return dce.pivot == null ? null : dce.pivot.getLocation();
		TrackNodeLocation pivot = dce.findPivot(dimension, start == getLeadingPoint());
		if (pivot == null)
			return null;
		boolean flipped = start != getLeadingPoint() && (leadingUpsideDown != trailingUpsideDown);
		class_243 startVec = start.getPosition(train.graph, flipped);
		class_243 portalVec = pivot.getLocation()
			.method_1031(0, leadingUpsideDown ? -1.0 : 1.0, 0);
		return VecHelper.lerp((float) (offset / startVec.method_1022(portalVec)), startVec, portalVec);
	}

	public void alignEntity(class_1937 level) {
		DimensionalCarriageEntity dimensionalCarriageEntity = entities.get(level.method_27983());
		if (dimensionalCarriageEntity != null) {
			CarriageContraptionEntity entity = dimensionalCarriageEntity.entity.get();
			if (entity != null)
				dimensionalCarriageEntity.alignEntity(entity);
		}
	}

	public TravellingPoint getLeadingPoint() {
		return leadingBogey().leading();
	}

	public TravellingPoint getTrailingPoint() {
		return trailingBogey().trailing();
	}

	public CarriageBogey leadingBogey() {
		return bogeys.getFirst();
	}

	public CarriageBogey trailingBogey() {
		return isOnTwoBogeys() ? bogeys.getSecond() : leadingBogey();
	}

	public boolean isOnTwoBogeys() {
		return bogeys.getSecond() != null;
	}

	public CarriageContraptionEntity anyAvailableEntity() {
		for (DimensionalCarriageEntity dimensionalCarriageEntity : entities.values()) {
			CarriageContraptionEntity entity = dimensionalCarriageEntity.entity.get();
			if (entity != null)
				return entity;
		}
		return null;
	}

	public void forEachPresentEntity(Consumer<CarriageContraptionEntity> callback) {
		for (DimensionalCarriageEntity dimensionalCarriageEntity : entities.values()) {
			CarriageContraptionEntity entity = dimensionalCarriageEntity.entity.get();
			if (entity != null)
				callback.accept(entity);
		}
	}

	public class_2487 write(DimensionPalette dimensions) {
		class_2487 tag = new class_2487();
		tag.method_10566("FirstBogey", bogeys.getFirst()
			.write(dimensions));
		if (isOnTwoBogeys())
			tag.method_10566("SecondBogey", bogeys.getSecond()
				.write(dimensions));
		tag.method_10569("Spacing", bogeySpacing);
		tag.method_10556("FrontConductor", presentConductors.getFirst());
		tag.method_10556("BackConductor", presentConductors.getSecond());
		tag.method_10556("Stalled", stalled);

		Map<Integer, class_2487> passengerMap = new HashMap<>();

		for (DimensionalCarriageEntity dimensionalCarriageEntity : entities.values()) {
			CarriageContraptionEntity entity = dimensionalCarriageEntity.entity.get();
			if (entity == null)
				continue;
			serialize(entity);
			Contraption contraption = entity.getContraption();
			if (contraption == null)
				continue;
			Map<UUID, Integer> mapping = contraption.getSeatMapping();
			for (class_1297 passenger : entity.method_5685())
				if (mapping.containsKey(passenger.method_5667()))
					passengerMap.put(mapping.get(passenger.method_5667()), NBTSerializer.serializeNBTCompound(passenger));
		}

		tag.method_10566("Entity", serialisedEntity.method_10553());

		class_2487 passengerTag = new class_2487();
		passengerMap.putAll(serialisedPassengers);
		passengerMap.forEach((seat, nbt) -> passengerTag.method_10566("Seat" + seat, nbt.method_10553()));
		tag.method_10566("Passengers", passengerTag);

		tag.method_10566("EntityPositioning", NBTHelper.writeCompoundList(entities.entrySet(), e -> {
			class_2487 c = e.getValue()
				.write();
			c.method_10569("Dim", dimensions.encode(e.getKey()));
			return c;
		}));

		return tag;
	}

	private void serialize(class_1297 entity) {
		serialisedEntity = NBTSerializer.serializeNBTCompound(entity);
		serialisedEntity.method_10551("Passengers");
		serialisedEntity.method_10562("Contraption")
			.method_10551("Passengers");
	}

	public static Carriage read(class_2487 tag, TrackGraph graph, DimensionPalette dimensions) {
		CarriageBogey bogey1 = CarriageBogey.read(tag.method_10562("FirstBogey"), graph, dimensions);
		CarriageBogey bogey2 =
			tag.method_10545("SecondBogey") ? CarriageBogey.read(tag.method_10562("SecondBogey"), graph, dimensions) : null;

		Carriage carriage = new Carriage(bogey1, bogey2, tag.method_10550("Spacing"));

		carriage.stalled = tag.method_10577("Stalled");
		carriage.presentConductors = Couple.create(tag.method_10577("FrontConductor"), tag.method_10577("BackConductor"));
		carriage.serialisedEntity = tag.method_10562("Entity")
			.method_10553();

		NBTHelper.iterateCompoundList(tag.method_10554("EntityPositioning", class_2520.field_33260),
			c -> carriage.getDimensional(dimensions.decode(c.method_10550("Dim")))
				.read(c));

		class_2487 passengersTag = tag.method_10562("Passengers");
		passengersTag.method_10541()
			.forEach(key -> carriage.serialisedPassengers.put(Integer.valueOf(key.substring(4)),
				passengersTag.method_10562(key)));

		return carriage;
	}

	private TravellingPoint portalScout = new TravellingPoint();

	public class DimensionalCarriageEntity {
		public class_243 positionAnchor;
		public Couple<class_243> rotationAnchors;
		public WeakReference<CarriageContraptionEntity> entity;

		public TrackNodeLocation pivot;
		int discardTicks;

		// 0 == whole, 0..1 = fading out, -1..0 = fading in
		public float cutoff;

		// client
		public boolean pointsInitialised;

		public DimensionalCarriageEntity() {
			this.entity = new WeakReference<>(null);
			this.rotationAnchors = Couple.create(null, null);
			this.pointsInitialised = false;
		}

		public void discardPivot() {
			int prevmin = minAllowedLocalCoord();
			int prevmax = maxAllowedLocalCoord();

			cutoff = 0;
			pivot = null;

			if ((!serialisedPassengers.isEmpty() && entity.get() != null) || prevmin != minAllowedLocalCoord()
				|| prevmax != maxAllowedLocalCoord()) {
				updatePassengerLoadout();
				updateRenderedCutoff();
			}
		}

		public void updateCutoff(boolean leadingIsCurrent) {
			class_243 leadingAnchor = rotationAnchors.getFirst();
			class_243 trailingAnchor = rotationAnchors.getSecond();

			if (leadingAnchor == null || trailingAnchor == null)
				return;
			if (pivot == null) {
				cutoff = 0;
				return;
			}

			class_243 pivotLoc = pivot.getLocation()
				.method_1031(0, 1, 0);

			double leadingSpacing = leadingBogey().type.getWheelPointSpacing() / 2;
			double trailingSpacing = trailingBogey().type.getWheelPointSpacing() / 2;
			double anchorSpacing = leadingSpacing + bogeySpacing + trailingSpacing;

			if (isOnTwoBogeys()) {
				class_243 diff = trailingAnchor.method_1020(leadingAnchor)
					.method_1029();
				trailingAnchor = trailingAnchor.method_1019(diff.method_1021(trailingSpacing));
				leadingAnchor = leadingAnchor.method_1019(diff.method_1021(-leadingSpacing));
			}

			double leadingDiff = leadingAnchor.method_1022(pivotLoc);
			double trailingDiff = trailingAnchor.method_1022(pivotLoc);

			leadingDiff /= anchorSpacing;
			trailingDiff /= anchorSpacing;

			if (leadingIsCurrent && leadingDiff > trailingDiff && leadingDiff > 1)
				cutoff = 0;
			else if (leadingIsCurrent && leadingDiff < trailingDiff && trailingDiff > 1)
				cutoff = 1;
			else if (!leadingIsCurrent && leadingDiff > trailingDiff && leadingDiff > 1)
				cutoff = -1;
			else if (!leadingIsCurrent && leadingDiff < trailingDiff && trailingDiff > 1)
				cutoff = 0;
			else
				cutoff = (float) class_3532.method_15350(1 - (leadingIsCurrent ? leadingDiff : trailingDiff), 0, 1)
					* (leadingIsCurrent ? 1 : -1);
		}

		public TrackNodeLocation findPivot(class_5321<class_1937> dimension, boolean leading) {
			if (pivot != null)
				return pivot;

			TravellingPoint start = leading ? getLeadingPoint() : getTrailingPoint();
			TravellingPoint end = !leading ? getLeadingPoint() : getTrailingPoint();

			portalScout.node1 = start.node1;
			portalScout.node2 = start.node2;
			portalScout.edge = start.edge;
			portalScout.position = start.position;

			ITrackSelector trackSelector = portalScout.follow(end);
			int distance = bogeySpacing + 10;
			int direction = leading ? -1 : 1;

			portalScout.travel(train.graph, direction * distance, trackSelector, portalScout.ignoreEdgePoints(),
				portalScout.ignoreTurns(), nodes -> {
					for (boolean b : Iterate.trueAndFalse)
						if (nodes.get(b).dimension.equals(dimension))
							pivot = nodes.get(b);
					return true;
				});

			return pivot;
		}

		public class_2487 write() {
			class_2487 tag = new class_2487();
			tag.method_10548("Cutoff", cutoff);
			tag.method_10569("DiscardTicks", discardTicks);
			storage.write(tag, false);
			if (pivot != null)
				tag.method_10566("Pivot", pivot.write(null));
			if (positionAnchor != null)
				tag.method_10566("PositionAnchor", VecHelper.writeNBT(positionAnchor));
			if (rotationAnchors.both(Objects::nonNull))
				tag.method_10566("RotationAnchors", rotationAnchors.serializeEach(VecHelper::writeNBTCompound));
			return tag;
		}

		public void read(class_2487 tag) {
			cutoff = tag.method_10583("Cutoff");
			discardTicks = tag.method_10550("DiscardTicks");
			storage.read(tag, null, false);
			if (tag.method_10545("Pivot"))
				pivot = TrackNodeLocation.read(tag.method_10562("Pivot"), null);
			if (positionAnchor != null)
				return;
			if (tag.method_10545("PositionAnchor"))
				positionAnchor = VecHelper.readNBT(tag.method_10554("PositionAnchor", class_2520.field_33256));
			if (tag.method_10545("RotationAnchors"))
				rotationAnchors = Couple.deserializeEach(tag.method_10554("RotationAnchors", class_2520.field_33260),
					VecHelper::readNBTCompound);
		}

		public class_243 leadingAnchor() {
			return isOnTwoBogeys() ? rotationAnchors.getFirst() : positionAnchor;
		}

		public class_243 trailingAnchor() {
			return isOnTwoBogeys() ? rotationAnchors.getSecond() : positionAnchor;
		}

		public int minAllowedLocalCoord() {
			if (cutoff <= 0)
				return Integer.MIN_VALUE;
			if (cutoff >= 1)
				return Integer.MAX_VALUE;
			return class_3532.method_15375(-bogeySpacing + -1 + (2 + bogeySpacing) * cutoff);
		}

		public int maxAllowedLocalCoord() {
			if (cutoff >= 0)
				return Integer.MAX_VALUE;
			if (cutoff <= -1)
				return Integer.MIN_VALUE;
			return class_3532.method_15386(-bogeySpacing + -1 + (2 + bogeySpacing) * (cutoff + 1));
		}

		public void updatePassengerLoadout() {
			class_1297 entity = this.entity.get();
			if (!(entity instanceof CarriageContraptionEntity cce))
				return;
			if (!(entity.method_37908() instanceof class_3218 sLevel))
				return;

			Set<Integer> loadedPassengers = new HashSet<>();
			int min = minAllowedLocalCoord();
			int max = maxAllowedLocalCoord();

			for (Entry<Integer, class_2487> entry : serialisedPassengers.entrySet()) {
				Integer seatId = entry.getKey();
				List<class_2338> seats = cce.getContraption()
					.getSeats();
				if (seatId >= seats.size())
					continue;

				class_2338 localPos = seats.get(seatId);
				if (!cce.isLocalCoordWithin(localPos, min, max))
					continue;

				class_2487 tag = entry.getValue();
				class_1297 passenger = null;

				if (tag.method_10545("PlayerPassenger")) {
					passenger = sLevel.method_8503()
						.method_3760()
						.method_14602(tag.method_25926("PlayerPassenger"));

				} else {
					passenger = class_1299.method_17842(tag, entity.method_37908(), e -> {
						e.method_29495(positionAnchor);
						return e;
					});
					if (passenger != null)
						sLevel.method_30736(passenger);
				}

				if (passenger != null) {
					class_5321<class_1937> passengerDimension = passenger.method_37908().method_27983();
					if (!passengerDimension.equals(sLevel.method_27983()) && passenger instanceof class_3222 sp)
						continue;
					cce.addSittingPassenger(passenger, seatId);
				}

				loadedPassengers.add(seatId);
			}

			loadedPassengers.forEach(serialisedPassengers::remove);

			Map<UUID, Integer> mapping = cce.getContraption()
				.getSeatMapping();
			for (class_1297 passenger : entity.method_5685()) {
				class_2338 localPos = cce.getContraption()
					.getSeatOf(passenger.method_5667());
				if (cce.isLocalCoordWithin(localPos, min, max))
					continue;
				if (!mapping.containsKey(passenger.method_5667()))
					continue;

				Integer seat = mapping.get(passenger.method_5667());
				if ((passenger instanceof class_3222 sp)) {
					dismountPlayer(sLevel, sp, seat, true);
					continue;
				}

				serialisedPassengers.put(seat, NBTSerializer.serializeNBTCompound(passenger));
				passenger.method_31472();
			}

		}

		private void dismountPlayer(class_3218 sLevel, class_3222 sp, Integer seat, boolean capture) {
			if (!capture) {
				sp.method_5848();
				return;
			}

			class_2487 tag = new class_2487();
			tag.method_25927("PlayerPassenger", sp.method_5667());
			serialisedPassengers.put(seat, tag);
			sp.method_5848();
			sp.getCustomData()
				.method_10551("ContraptionDismountLocation");

			for (Entry<class_5321<class_1937>, DimensionalCarriageEntity> other : entities.entrySet()) {
				DimensionalCarriageEntity otherDce = other.getValue();
				if (otherDce == this)
					continue;
				if (sp.method_37908().method_27983()
					.equals(other.getKey()))
					continue;
				class_243 loc = otherDce.pivot == null ? otherDce.positionAnchor : otherDce.pivot.getLocation();
				if (loc == null)
					continue;
				class_3218 level = sLevel.method_8503()
					.method_3847(other.getKey());
				sp.method_14251(level, loc.field_1352, loc.field_1351, loc.field_1350, sp.method_36454(), sp.method_36455());
				sp.method_30229();
				AllAdvancements.TRAIN_PORTAL.awardTo(sp);
			}
		}

		public void updateRenderedCutoff() {
			class_1297 entity = this.entity.get();
			if (!(entity instanceof CarriageContraptionEntity cce))
				return;
			Contraption contraption = cce.getContraption();
			if (!(contraption instanceof CarriageContraption cc))
				return;
			cc.portalCutoffMin = minAllowedLocalCoord();
			cc.portalCutoffMax = maxAllowedLocalCoord();
			if (!entity.method_37908().method_8608())
				return;
			EnvExecutor.runWhenOn(EnvType.CLIENT, () -> () -> invalidate(cce));
		}

		@Environment(EnvType.CLIENT)
		private void invalidate(CarriageContraptionEntity entity) {
			entity.getContraption().deferInvalidate = true;
			entity.updateRenderedPortalCutoff();
		}

		private void createEntity(class_1937 level, boolean loadPassengers) {
			class_1297 entity = class_1299.method_5892(serialisedEntity, level)
				.orElse(null);

			if (!(entity instanceof CarriageContraptionEntity cce)) {
				train.invalid = true;
				return;
			}

			entity.method_29495(positionAnchor);
			this.entity = new WeakReference<>(cce);

			cce.setCarriage(Carriage.this);
			cce.syncCarriage();

			if (level instanceof class_3218 sl)
				sl.method_8649(entity);

			updatePassengerLoadout();
		}

		private void removeAndSaveEntity(CarriageContraptionEntity entity, boolean portal) {
			Contraption contraption = entity.getContraption();
			if (contraption != null) {
				Map<UUID, Integer> mapping = contraption.getSeatMapping();
				for (class_1297 passenger : entity.method_5685()) {
					if (!mapping.containsKey(passenger.method_5667()))
						continue;

					Integer seat = mapping.get(passenger.method_5667());

					if (passenger instanceof class_3222 sp) {
						dismountPlayer(sp.method_51469(), sp, seat, portal);
						continue;
					}

					serialisedPassengers.put(seat, NBTSerializer.serializeNBTCompound(passenger));
				}
			}

			for (class_1297 passenger : entity.method_5685())
				if (!(passenger instanceof class_1657))
					passenger.method_31472();

			serialize(entity);
			entity.method_31472();
			this.entity.clear();
		}

		public void alignEntity(CarriageContraptionEntity entity) {
			if (rotationAnchors.either(Objects::isNull))
				return;

			class_243 positionVec = rotationAnchors.getFirst();
			class_243 coupledVec = rotationAnchors.getSecond();

			double diffX = positionVec.field_1352 - coupledVec.field_1352;
			double diffY = positionVec.field_1351 - coupledVec.field_1351;
			double diffZ = positionVec.field_1350 - coupledVec.field_1350;

			entity.prevYaw = entity.yaw;
			entity.prevPitch = entity.pitch;

			if (!entity.method_37908().method_8608()) {
				class_243 lookahead = positionAnchor.method_1019(positionAnchor.method_1020(entity.method_19538())
					.method_1029()
					.method_1021(16));

				for (class_1297 e : entity.method_5685()) {
					if (!(e instanceof class_1657))
						continue;
					if (e.method_5858(entity) > 32 * 32)
						continue;
					if (CarriageEntityHandler.isActiveChunk(entity.method_37908(), class_2338.method_49638(lookahead)))
						break;
					train.carriageWaitingForChunks = id;
					return;
				}

				if (entity.method_5685()
					.stream()
					.anyMatch(p -> p instanceof class_1657)
					) {
				}

				if (train.carriageWaitingForChunks == id)
					train.carriageWaitingForChunks = -1;

				entity.setServerSidePrevPosition();
			}

			entity.method_33574(positionAnchor);
			entity.yaw = (float) (class_3532.method_15349(diffZ, diffX) * 180 / Math.PI) + 180;
			entity.pitch = (float) (Math.atan2(diffY, Math.sqrt(diffX * diffX + diffZ * diffZ)) * 180 / Math.PI) * -1;

			if (!entity.firstPositionUpdate)
				return;

			entity.field_6014 = entity.method_23317();
			entity.field_6036 = entity.method_23318();
			entity.field_5969 = entity.method_23321();
			entity.prevYaw = entity.yaw;
			entity.prevPitch = entity.pitch;
		}
	}

}
