package com.simibubi.create.content.trains.entity;

import static com.simibubi.create.content.trains.bogey.AbstractBogeyBlockEntity.BOGEY_DATA_KEY;
import static com.simibubi.create.content.trains.bogey.AbstractBogeyBlockEntity.BOGEY_STYLE_KEY;

import javax.annotation.Nullable;
import net.minecraft.class_1937;
import net.minecraft.class_2248;
import net.minecraft.class_2350.class_2351;
import net.minecraft.class_243;
import net.minecraft.class_2487;
import net.minecraft.class_2520;
import net.minecraft.class_2960;
import net.minecraft.class_3532;
import net.minecraft.class_5321;
import net.minecraft.class_7923;
import com.jozufozu.flywheel.api.MaterialManager;
import com.simibubi.create.AllBlocks;
import com.simibubi.create.AllBogeyStyles;
import com.simibubi.create.Create;
import com.simibubi.create.content.trains.bogey.AbstractBogeyBlock;
import com.simibubi.create.content.trains.bogey.AbstractBogeyBlockEntity;
import com.simibubi.create.content.trains.bogey.BogeyInstance;
import com.simibubi.create.content.trains.bogey.BogeyStyle;
import com.simibubi.create.content.trains.graph.DimensionPalette;
import com.simibubi.create.content.trains.graph.TrackGraph;
import com.simibubi.create.foundation.utility.AngleHelper;
import com.simibubi.create.foundation.utility.Couple;
import com.simibubi.create.foundation.utility.Iterate;
import com.simibubi.create.foundation.utility.NBTHelper;
import com.simibubi.create.foundation.utility.RegisteredObjects;
import com.simibubi.create.foundation.utility.VecHelper;
import com.simibubi.create.foundation.utility.animation.LerpedFloat;

public class CarriageBogey {

	public static final String UPSIDE_DOWN_KEY = "UpsideDown";

	public Carriage carriage;
	boolean isLeading;

	public class_2487 bogeyData;

	AbstractBogeyBlock<?> type;
	boolean upsideDown;
	Couple<TravellingPoint> points;

	LerpedFloat wheelAngle;
	LerpedFloat yaw;
	LerpedFloat pitch;

	public Couple<class_243> couplingAnchors;

	int derailAngle;

	public CarriageBogey(AbstractBogeyBlock<?> type, boolean upsideDown, class_2487 bogeyData, TravellingPoint point, TravellingPoint point2) {
		this.type = type;
		this.upsideDown = type.canBeUpsideDown() && upsideDown;
		point.upsideDown = this.upsideDown;
		point2.upsideDown = this.upsideDown;
		if (bogeyData == null || bogeyData.method_33133())
			bogeyData = this.createBogeyData(); // Prevent Crash When Updating
		bogeyData.method_10556(UPSIDE_DOWN_KEY, upsideDown);
		this.bogeyData = bogeyData;
		points = Couple.create(point, point2);
		wheelAngle = LerpedFloat.angular();
		yaw = LerpedFloat.angular();
		pitch = LerpedFloat.angular();
		derailAngle = Create.RANDOM.nextInt(60) - 30;
		couplingAnchors = Couple.create(null, null);
	}

	public class_5321<class_1937> getDimension() {
		TravellingPoint leading = leading();
		TravellingPoint trailing = trailing();
		if (leading.edge == null || trailing.edge == null)
			return null;
		if (leading.edge.isInterDimensional() || trailing.edge.isInterDimensional())
			return null;
		class_5321<class_1937> dimension1 = leading.node1.getLocation().dimension;
		class_5321<class_1937> dimension2 = trailing.node1.getLocation().dimension;
		if (dimension1.equals(dimension2))
			return dimension1;
		return null;
	}

	public void updateAngles(CarriageContraptionEntity entity, double distanceMoved) {
		double angleDiff = 360 * distanceMoved / (Math.PI * 2 * type.getWheelRadius());

		float xRot = 0;
		float yRot = 0;

		if (leading().edge == null || carriage.train.derailed) {
			yRot = -90 + entity.yaw - derailAngle;
		} else if (!entity.method_37908().method_27983()
				.equals(getDimension())) {
			yRot = -90 + entity.yaw;
			xRot = 0;
		} else {
			class_243 positionVec = leading().getPosition(carriage.train.graph);
			class_243 coupledVec = trailing().getPosition(carriage.train.graph);
			double diffX = positionVec.field_1352 - coupledVec.field_1352;
			double diffY = positionVec.field_1351 - coupledVec.field_1351;
			double diffZ = positionVec.field_1350 - coupledVec.field_1350;
			yRot = AngleHelper.deg(class_3532.method_15349(diffZ, diffX)) + 90;
			xRot = AngleHelper.deg(Math.atan2(diffY, Math.sqrt(diffX * diffX + diffZ * diffZ)));
		}

		double newWheelAngle = (wheelAngle.getValue() - angleDiff) % 360;

		for (boolean twice : Iterate.trueAndFalse) {
			if (twice && !entity.firstPositionUpdate)
				continue;
			wheelAngle.setValue(newWheelAngle);
			pitch.setValue(xRot);
			yaw.setValue(-yRot);
		}
	}

	public TravellingPoint leading() {
		TravellingPoint point = points.getFirst();
		point.upsideDown = isUpsideDown();
		return point;
	}

	public TravellingPoint trailing() {
		TravellingPoint point = points.getSecond();
		point.upsideDown = isUpsideDown();
		return point;
	}

	public double getStress() {
		if (getDimension() == null)
			return 0;
		if (carriage.train.derailed)
			return 0;
		return type.getWheelPointSpacing() - leading().getPosition(carriage.train.graph)
				.method_1022(trailing().getPosition(carriage.train.graph));
	}

	@Nullable
	public class_243 getAnchorPosition() {
		return getAnchorPosition(false);
	}

	@Nullable
	public class_243 getAnchorPosition(boolean flipUpsideDown) {
		if (leading().edge == null)
			return null;
		return points.getFirst()
				.getPosition(carriage.train.graph, flipUpsideDown)
				.method_1019(points.getSecond()
						.getPosition(carriage.train.graph, flipUpsideDown))
				.method_1021(.5);
	}

	public void updateCouplingAnchor(class_243 entityPos, float entityXRot, float entityYRot, int bogeySpacing,
									 float partialTicks, boolean leading) {
		boolean selfUpsideDown = isUpsideDown();
		boolean leadingUpsideDown = carriage.leadingBogey().isUpsideDown();
		class_243 thisOffset = type.getConnectorAnchorOffset(selfUpsideDown);
		thisOffset = thisOffset.method_18805(1, 1, leading ? -1 : 1);

		thisOffset = VecHelper.rotate(thisOffset, pitch.getValue(partialTicks), class_2351.field_11048);
		thisOffset = VecHelper.rotate(thisOffset, yaw.getValue(partialTicks), class_2351.field_11052);
		thisOffset = VecHelper.rotate(thisOffset, -entityYRot - 90, class_2351.field_11052);
		thisOffset = VecHelper.rotate(thisOffset, entityXRot, class_2351.field_11048);
		thisOffset = VecHelper.rotate(thisOffset, -180, class_2351.field_11052);
		thisOffset = thisOffset.method_1031(0, 0, leading ? 0 : -bogeySpacing);
		thisOffset = VecHelper.rotate(thisOffset, 180, class_2351.field_11052);
		thisOffset = VecHelper.rotate(thisOffset, -entityXRot, class_2351.field_11048);
		thisOffset = VecHelper.rotate(thisOffset, entityYRot + 90, class_2351.field_11052);
		if (selfUpsideDown != leadingUpsideDown)
			thisOffset = thisOffset.method_1031(0, selfUpsideDown ? -2 : 2, 0);

		couplingAnchors.set(leading, entityPos.method_1019(thisOffset));
	}

	public class_2487 write(DimensionPalette dimensions) {
		class_2487 tag = new class_2487();
		tag.method_10582("Type", RegisteredObjects.getKeyOrThrow((class_2248) type)
				.toString());
		tag.method_10566("Points", points.serializeEach(tp -> tp.write(dimensions)));
		tag.method_10556("UpsideDown", upsideDown);
		bogeyData.method_10556(UPSIDE_DOWN_KEY, upsideDown);
		NBTHelper.writeResourceLocation(bogeyData, BOGEY_STYLE_KEY, getStyle().name);
		tag.method_10566(BOGEY_DATA_KEY, bogeyData);
		return tag;
	}

	public static CarriageBogey read(class_2487 tag, TrackGraph graph, DimensionPalette dimensions) {
		class_2960 location = new class_2960(tag.method_10558("Type"));
		class_2248 block = class_7923.field_41175.method_10223(location);
		AbstractBogeyBlock<?> type = block instanceof AbstractBogeyBlock<?> bogey ? bogey : AllBlocks.SMALL_BOGEY.get();
		boolean upsideDown = tag.method_10577("UpsideDown");
		Couple<TravellingPoint> points = Couple.deserializeEach(tag.method_10554("Points", class_2520.field_33260),
				c -> TravellingPoint.read(c, graph, dimensions));
		class_2487 data = tag.method_10562(AbstractBogeyBlockEntity.BOGEY_DATA_KEY);
		return new CarriageBogey(type, upsideDown, data, points.getFirst(), points.getSecond());
	}

	public BogeyInstance createInstance(MaterialManager materialManager) {
		return this.getStyle().createInstance(this, type.getSize(), materialManager);
	}

	public BogeyStyle getStyle() {
		class_2960 location = NBTHelper.readResourceLocation(this.bogeyData, BOGEY_STYLE_KEY);
		BogeyStyle style = AllBogeyStyles.BOGEY_STYLES.get(location);
		return style != null ? style : AllBogeyStyles.STANDARD; // just for safety
	}

	private class_2487 createBogeyData() {
		BogeyStyle style = type != null ? type.getDefaultStyle() : AllBogeyStyles.STANDARD;
		class_2487 nbt = style.defaultData != null ? style.defaultData : new class_2487();
		NBTHelper.writeResourceLocation(nbt, BOGEY_STYLE_KEY, style.name);
		nbt.method_10556(UPSIDE_DOWN_KEY, isUpsideDown());
		return nbt;
	}

	void setLeading() {
		isLeading = true;
	}

	public boolean isUpsideDown() {
		return type.canBeUpsideDown() && upsideDown;
	}
}
