package com.simibubi.create.content.trains.entity;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;

import com.simibubi.create.foundation.fluid.CombinedTankWrapper;

import io.github.fabricators_of_create.porting_lib.transfer.fluid.FluidTank;
import io.github.fabricators_of_create.porting_lib.transfer.item.ItemStackHandler;
import io.github.fabricators_of_create.porting_lib.fluids.FluidStack;

import org.apache.commons.lang3.tuple.Pair;

import com.simibubi.create.AllBlocks;
import com.simibubi.create.content.contraptions.AbstractContraptionEntity;
import com.simibubi.create.content.contraptions.AssemblyException;
import com.simibubi.create.content.contraptions.Contraption;
import com.simibubi.create.content.contraptions.ContraptionType;
import com.simibubi.create.content.contraptions.MountedStorageManager;
import com.simibubi.create.content.contraptions.actors.trainControls.ControlsBlock;
import com.simibubi.create.content.contraptions.minecart.TrainCargoManager;
import com.simibubi.create.content.contraptions.render.ContraptionLighter;
import com.simibubi.create.content.contraptions.render.NonStationaryLighter;
import com.simibubi.create.content.processing.burner.BlazeBurnerBlock;
import com.simibubi.create.content.processing.burner.BlazeBurnerBlock.HeatLevel;
import com.simibubi.create.content.trains.bogey.AbstractBogeyBlock;
import com.simibubi.create.foundation.utility.Couple;
import com.simibubi.create.foundation.utility.Iterate;
import com.simibubi.create.foundation.utility.Lang;
import com.simibubi.create.foundation.utility.NBTHelper;
import com.simibubi.create.foundation.utility.VecHelper;
import net.fabricmc.api.EnvType;
import net.fabricmc.api.Environment;
import net.minecraft.class_1937;
import net.minecraft.class_2246;
import net.minecraft.class_2338;
import net.minecraft.class_2350;
import net.minecraft.class_2350.class_2351;
import net.minecraft.class_238;
import net.minecraft.class_2487;
import net.minecraft.class_2499;
import net.minecraft.class_2512;
import net.minecraft.class_2520;
import net.minecraft.class_2586;
import net.minecraft.class_2680;
import net.minecraft.class_3499.class_3501;

public class CarriageContraption extends Contraption {

	private class_2350 assemblyDirection;
	private boolean forwardControls;
	private boolean backwardControls;

	public Couple<Boolean> blazeBurnerConductors;
	public Map<class_2338, Couple<Boolean>> conductorSeats;
	public ArrivalSoundQueue soundQueue;

	protected MountedStorageManager storageProxy;

	// during assembly only
	private int bogeys;
	private boolean sidewaysControls;
	private class_2338 secondBogeyPos;
	private List<class_2338> assembledBlazeBurners;

	// render
	public int portalCutoffMin;
	public int portalCutoffMax;

	static final ContraptionInvWrapper fallbackItems = new ContraptionInvWrapper();
	static final CombinedTankWrapper fallbackFluids = new CombinedTankWrapper();

	public CarriageContraption() {
		conductorSeats = new HashMap<>();
		assembledBlazeBurners = new ArrayList<>();
		blazeBurnerConductors = Couple.create(false, false);
		soundQueue = new ArrivalSoundQueue();
		portalCutoffMin = Integer.MIN_VALUE;
		portalCutoffMax = Integer.MAX_VALUE;
		storage = new TrainCargoManager();
	}

	public void setSoundQueueOffset(int offset) {
		soundQueue.offset = offset;
	}

	public CarriageContraption(class_2350 assemblyDirection) {
		this();
		this.assemblyDirection = assemblyDirection;
		this.bogeys = 0;
	}

	@Override
	public boolean assemble(class_1937 world, class_2338 pos) throws AssemblyException {
		if (!searchMovedStructure(world, pos, null))
			return false;
		if (blocks.size() <= 1)
			return false;
		if (bogeys == 0)
			return false;
		if (bogeys > 2)
			throw new AssemblyException(Lang.translateDirect("train_assembly.too_many_bogeys", bogeys));
		if (sidewaysControls)
			throw new AssemblyException(Lang.translateDirect("train_assembly.sideways_controls"));

		for (class_2338 blazePos : assembledBlazeBurners)
			for (class_2350 direction : Iterate.directionsInAxis(assemblyDirection.method_10166()))
				if (inControl(blazePos, direction))
					blazeBurnerConductors.set(direction != assemblyDirection, true);
		for (class_2338 seatPos : getSeats())
			for (class_2350 direction : Iterate.directionsInAxis(assemblyDirection.method_10166()))
				if (inControl(seatPos, direction))
					conductorSeats.computeIfAbsent(seatPos, p -> Couple.create(false, false))
						.set(direction != assemblyDirection, true);

		return true;
	}

	public boolean inControl(class_2338 pos, class_2350 direction) {
		class_2338 controlsPos = pos.method_10093(direction);
		if (!blocks.containsKey(controlsPos))
			return false;
		class_3501 info = blocks.get(controlsPos);
		if (!AllBlocks.TRAIN_CONTROLS.has(info.comp_1342()))
			return false;
		return info.comp_1342().method_11654(ControlsBlock.field_11177) == direction.method_10153();
	}

	public void swapStorageAfterAssembly(CarriageContraptionEntity cce) {
		// Ensure that the entity does not hold its inventory data, because the global
		// carriage manages it instead
		Carriage carriage = cce.getCarriage();
		if (carriage.storage == null) {
			carriage.storage = (TrainCargoManager) storage;
			storage = new MountedStorageManager();
		}
		storageProxy = carriage.storage;
	}

	public void returnStorageForDisassembly(MountedStorageManager storage) {
		this.storage = storage;
	}

	@Override
	protected boolean isAnchoringBlockAt(class_2338 pos) {
		return false;
	}

	@Override
	protected Pair<class_3501, class_2586> capture(class_1937 world, class_2338 pos) {
		class_2680 blockState = world.method_8320(pos);

		if (ArrivalSoundQueue.isPlayable(blockState)) {
			int anchorCoord = VecHelper.getCoordinate(anchor, assemblyDirection.method_10166());
			int posCoord = VecHelper.getCoordinate(pos, assemblyDirection.method_10166());
			soundQueue.add((posCoord - anchorCoord) * assemblyDirection.method_10171()
				.method_10181(), toLocalPos(pos));
		}

		if (blockState.method_26204() instanceof AbstractBogeyBlock<?> bogey) {
			boolean captureBE = bogey.captureBlockEntityForTrain();
			bogeys++;
			if (bogeys == 2)
				secondBogeyPos = pos;
			return Pair.of(new class_3501(pos, blockState, captureBE ? getBlockEntityNBT(world, pos) : null),
				captureBE ? world.method_8321(pos) : null);
		}

		if (AllBlocks.BLAZE_BURNER.has(blockState)
			&& blockState.method_11654(BlazeBurnerBlock.HEAT_LEVEL) != HeatLevel.NONE)
			assembledBlazeBurners.add(toLocalPos(pos));

		if (AllBlocks.TRAIN_CONTROLS.has(blockState)) {
			class_2350 facing = blockState.method_11654(ControlsBlock.field_11177);
			if (facing.method_10166() != assemblyDirection.method_10166())
				sidewaysControls = true;
			else {
				boolean forwards = facing == assemblyDirection;
				if (forwards)
					forwardControls = true;
				else
					backwardControls = true;
			}
		}

		return super.capture(world, pos);
	}

	@Override
	public class_2487 writeNBT(boolean spawnPacket) {
		class_2487 tag = super.writeNBT(spawnPacket);
		NBTHelper.writeEnum(tag, "AssemblyDirection", getAssemblyDirection());
		tag.method_10556("FrontControls", forwardControls);
		tag.method_10556("BackControls", backwardControls);
		tag.method_10556("FrontBlazeConductor", blazeBurnerConductors.getFirst());
		tag.method_10556("BackBlazeConductor", blazeBurnerConductors.getSecond());
		class_2499 list = NBTHelper.writeCompoundList(conductorSeats.entrySet(), e -> {
			class_2487 compoundTag = new class_2487();
			compoundTag.method_10566("Pos", class_2512.method_10692(e.getKey()));
			compoundTag.method_10556("Forward", e.getValue()
				.getFirst());
			compoundTag.method_10556("Backward", e.getValue()
				.getSecond());
			return compoundTag;
		});
		tag.method_10566("ConductorSeats", list);
		soundQueue.serialize(tag);
		return tag;
	}

	@Override
	public void readNBT(class_1937 world, class_2487 nbt, boolean spawnData) {
		assemblyDirection = NBTHelper.readEnum(nbt, "AssemblyDirection", class_2350.class);
		forwardControls = nbt.method_10577("FrontControls");
		backwardControls = nbt.method_10577("BackControls");
		blazeBurnerConductors =
			Couple.create(nbt.method_10577("FrontBlazeConductor"), nbt.method_10577("BackBlazeConductor"));
		conductorSeats.clear();
		NBTHelper.iterateCompoundList(nbt.method_10554("ConductorSeats", class_2520.field_33260),
			c -> conductorSeats.put(class_2512.method_10691(c.method_10562("Pos")),
				Couple.create(c.method_10577("Forward"), c.method_10577("Backward"))));
		soundQueue.deserialize(nbt);
		super.readNBT(world, nbt, spawnData);
	}

	@Override
	public boolean canBeStabilized(class_2350 facing, class_2338 localPos) {
		return false;
	}

	@Override
	protected MountedStorageManager getStorageForSpawnPacket() {
		return storageProxy;
	}

	@Override
	public ContraptionType getType() {
		return ContraptionType.CARRIAGE;
	}

	@Override
	@Environment(EnvType.CLIENT)
	public ContraptionLighter<?> makeLighter() {
		return new NonStationaryLighter<>(this);
	}

	public class_2350 getAssemblyDirection() {
		return assemblyDirection;
	}

	public boolean hasForwardControls() {
		return forwardControls;
	}

	public boolean hasBackwardControls() {
		return backwardControls;
	}

	public class_2338 getSecondBogeyPos() {
		return secondBogeyPos;
	}

	private Collection<class_2586> specialRenderedBEsOutsidePortal = new ArrayList<>();

	@Override
	public Collection<class_3501> getRenderedBlocks() {
		if (notInPortal())
			return super.getRenderedBlocks();

		specialRenderedBEsOutsidePortal = new ArrayList<>();
		specialRenderedBlockEntities.stream()
			.filter(be -> !isHiddenInPortal(be.method_11016()))
			.forEach(specialRenderedBEsOutsidePortal::add);

		Collection<class_3501> values = new ArrayList<>();
		for (Entry<class_2338, class_3501> entry : blocks.entrySet()) {
			class_2338 pos = entry.getKey();
			if (withinVisible(pos))
				values.add(entry.getValue());
			else if (atSeam(pos))
				values.add(new class_3501(pos, class_2246.field_10399.method_9564(), null));
		}
		return values;
	}

	@Override
	public Collection<class_2586> getSpecialRenderedBEs() {
		if (notInPortal())
			return super.getSpecialRenderedBEs();
		return specialRenderedBEsOutsidePortal;
	}

	@Override
	public Optional<List<class_238>> getSimplifiedEntityColliders() {
		if (notInPortal())
			return super.getSimplifiedEntityColliders();
		return Optional.empty();
	}

	@Override
	public boolean isHiddenInPortal(class_2338 localPos) {
		if (notInPortal())
			return super.isHiddenInPortal(localPos);
		return !withinVisible(localPos) || atSeam(localPos);
	}

	public boolean notInPortal() {
		return portalCutoffMin == Integer.MIN_VALUE && portalCutoffMax == Integer.MAX_VALUE;
	}

	public boolean atSeam(class_2338 localPos) {
		class_2350 facing = assemblyDirection;
		class_2351 axis = facing.method_10170()
			.method_10166();
		int coord = axis.method_10173(localPos.method_10260(), localPos.method_10264(), localPos.method_10263()) * -facing.method_10171()
			.method_10181();
		return coord == portalCutoffMin || coord == portalCutoffMax;
	}

	public boolean withinVisible(class_2338 localPos) {
		class_2350 facing = assemblyDirection;
		class_2351 axis = facing.method_10170()
			.method_10166();
		int coord = axis.method_10173(localPos.method_10260(), localPos.method_10264(), localPos.method_10263()) * -facing.method_10171()
			.method_10181();
		return coord > portalCutoffMin && coord < portalCutoffMax;
	}

	@Override
	public ContraptionInvWrapper getSharedInventory() {
		return storageProxy == null ? fallbackItems : storageProxy.getItems();
	}

	@Override
	public CombinedTankWrapper getSharedFluidTanks() {
		return storageProxy == null ? fallbackFluids : storageProxy.getFluids();
	}

	public void handleContraptionFluidPacket(class_2338 localPos, FluidStack containedFluid) {
		storage.updateContainedFluid(localPos, containedFluid);
	}

	@Override
	public void tickStorage(AbstractContraptionEntity entity) {
		if (entity.method_37908().field_9236)
			storage.entityTick(entity);
		else if (storageProxy != null)
			storageProxy.entityTick(entity);
	}

}
