package com.simibubi.create.content.trains.entity;

import java.util.Map.Entry;
import net.minecraft.class_243;
import net.minecraft.class_2487;
import net.minecraft.class_2520;
import net.minecraft.class_3532;
import com.simibubi.create.content.trains.graph.DimensionPalette;
import com.simibubi.create.content.trains.graph.TrackEdge;
import com.simibubi.create.content.trains.graph.TrackGraph;
import com.simibubi.create.content.trains.graph.TrackGraphLocation;
import com.simibubi.create.content.trains.graph.TrackNode;
import com.simibubi.create.content.trains.graph.TrackNodeLocation;
import com.simibubi.create.foundation.utility.Couple;
import com.simibubi.create.foundation.utility.VecHelper;

public class TrainMigration {

	Couple<TrackNodeLocation> locations;
	double positionOnOldEdge;
	boolean curve;
	class_243 fallback;

	public TrainMigration() {}

	public TrainMigration(TravellingPoint point) {
		double t = point.position / point.edge.getLength();
		fallback = point.edge.getPosition(null, t);
		curve = point.edge.isTurn();
		positionOnOldEdge = point.position;
		locations = Couple.create(point.node1.getLocation(), point.node2.getLocation());
	}

	public TrackGraphLocation tryMigratingTo(TrackGraph graph) {
		TrackNode node1 = graph.locateNode(locations.getFirst());
		TrackNode node2 = graph.locateNode(locations.getSecond());
		if (node1 != null && node2 != null) {
			TrackEdge edge = graph.getConnectionsFrom(node1)
				.get(node2);
			if (edge != null) {
				TrackGraphLocation graphLocation = new TrackGraphLocation();
				graphLocation.graph = graph;
				graphLocation.edge = locations;
				graphLocation.position = positionOnOldEdge;
				return graphLocation;
			}
		}

		if (curve)
			return null;

		class_243 prevDirection = locations.getSecond()
			.getLocation()
			.method_1020(locations.getFirst()
				.getLocation())
			.method_1029();

		for (TrackNodeLocation loc : graph.getNodes()) {
			class_243 nodeVec = loc.getLocation();
			if (nodeVec.method_1025(fallback) > 32 * 32)
				continue;

			TrackNode newNode1 = graph.locateNode(loc);
			for (Entry<TrackNode, TrackEdge> entry : graph.getConnectionsFrom(newNode1)
				.entrySet()) {
				TrackEdge edge = entry.getValue();
				if (edge.isTurn())
					continue;
				TrackNode newNode2 = entry.getKey();
				float radius = 1 / 64f;
				class_243 direction = edge.getDirection(true);
				if (!class_3532.method_20390(direction.method_1026(prevDirection), 1))
					continue;
				class_243 intersectSphere = VecHelper.intersectSphere(nodeVec, direction, fallback, radius);
				if (intersectSphere == null)
					continue;
				if (!class_3532.method_20390(direction.method_1026(intersectSphere.method_1020(nodeVec)
					.method_1029()), 1))
					continue;
				double edgeLength = edge.getLength();
				double position = intersectSphere.method_1022(nodeVec) - radius;
				if (Double.isNaN(position))
					continue;
				if (position < 0)
					continue;
				if (position > edgeLength)
					continue;

				TrackGraphLocation graphLocation = new TrackGraphLocation();
				graphLocation.graph = graph;
				graphLocation.edge = Couple.create(loc, newNode2.getLocation());
				graphLocation.position = position;
				return graphLocation;
			}
		}

		return null;
	}

	public class_2487 write(DimensionPalette dimensions) {
		class_2487 tag = new class_2487();
		tag.method_10556("Curve", curve);
		tag.method_10566("Fallback", VecHelper.writeNBT(fallback));
		tag.method_10549("Position", positionOnOldEdge);
		tag.method_10566("Nodes", locations.serializeEach(l -> l.write(dimensions)));
		return tag;
	}

	public static TrainMigration read(class_2487 tag, DimensionPalette dimensions) {
		TrainMigration trainMigration = new TrainMigration();
		trainMigration.curve = tag.method_10577("Curve");
		trainMigration.fallback = VecHelper.readNBT(tag.method_10554("Fallback", class_2520.field_33256));
		trainMigration.positionOnOldEdge = tag.method_10574("Position");
		trainMigration.locations =
			Couple.deserializeEach(tag.method_10554("Nodes", class_2520.field_33260), c -> TrackNodeLocation.read(c, dimensions));
		return trainMigration;
	}

}