package com.simibubi.create.content.trains.entity;

import java.util.UUID;
import net.minecraft.class_124;
import net.minecraft.class_1297;
import net.minecraft.class_2338;
import net.minecraft.class_243;
import net.minecraft.class_2540;
import net.minecraft.class_3222;
import com.simibubi.create.AllPackets;
import com.simibubi.create.Create;
import com.simibubi.create.content.contraptions.ContraptionRelocationPacket;
import com.simibubi.create.content.trains.track.BezierTrackPointLocation;
import com.simibubi.create.foundation.networking.SimplePacketBase;
import com.simibubi.create.foundation.utility.Lang;
import com.simibubi.create.foundation.utility.VecHelper;
import com.simibubi.create.infrastructure.config.AllConfigs;

public class TrainRelocationPacket extends SimplePacketBase {

	UUID trainId;
	class_2338 pos;
	class_243 lookAngle;
	int entityId;
	private boolean direction;
	private BezierTrackPointLocation hoveredBezier;

	public TrainRelocationPacket(class_2540 buffer) {
		trainId = buffer.method_10790();
		pos = buffer.method_10811();
		lookAngle = VecHelper.read(buffer);
		entityId = buffer.readInt();
		direction = buffer.readBoolean();
		if (buffer.readBoolean())
			hoveredBezier = new BezierTrackPointLocation(buffer.method_10811(), buffer.readInt());
	}

	public TrainRelocationPacket(UUID trainId, class_2338 pos, BezierTrackPointLocation hoveredBezier, boolean direction,
		class_243 lookAngle, int entityId) {
		this.trainId = trainId;
		this.pos = pos;
		this.hoveredBezier = hoveredBezier;
		this.direction = direction;
		this.lookAngle = lookAngle;
		this.entityId = entityId;
	}

	@Override
	public void write(class_2540 buffer) {
		buffer.method_10797(trainId);
		buffer.method_10807(pos);
		VecHelper.write(lookAngle, buffer);
		buffer.writeInt(entityId);
		buffer.writeBoolean(direction);
		buffer.writeBoolean(hoveredBezier != null);
		if (hoveredBezier != null) {
			buffer.method_10807(hoveredBezier.curveTarget());
			buffer.writeInt(hoveredBezier.segment());
		}
	}

	@Override
	public boolean handle(Context context) {
		context.enqueueWork(() -> {
			class_3222 sender = context.getSender();
			Train train = Create.RAILWAYS.trains.get(trainId);
			class_1297 entity = sender.method_37908().method_8469(entityId);

			String messagePrefix = sender.method_5477()
				.getString() + " could not relocate Train ";

			if (train == null || !(entity instanceof CarriageContraptionEntity cce)) {
				Create.LOGGER.warn(messagePrefix + train.id.toString()
					.substring(0, 5) + ": not present on server");
				return;
			}

			if (!train.id.equals(cce.trainId))
				return;

			int verifyDistance = AllConfigs.server().trains.maxTrackPlacementLength.get() * 2;
			if (!sender.method_19538()
				.method_24802(class_243.method_24953(pos), verifyDistance)) {
				Create.LOGGER.warn(messagePrefix + train.name.getString() + ": player too far from clicked pos");
				return;
			}
			if (!sender.method_19538()
				.method_24802(cce.method_19538(), verifyDistance + cce.method_5829()
					.method_17939() / 2)) {
				Create.LOGGER.warn(messagePrefix + train.name.getString() + ": player too far from carriage entity");
				return;
			}

			if (TrainRelocator.relocate(train, sender.method_37908(), pos, hoveredBezier, direction, lookAngle, false)) {
				sender.method_7353(Lang.translateDirect("train.relocate.success")
					.method_27692(class_124.field_1060), true);
				train.carriages.forEach(c -> c.forEachPresentEntity(e -> {
					e.nonDamageTicks = 10;
					AllPackets.getChannel().sendToClientsTracking(new ContraptionRelocationPacket(e.method_5628()), e);
				}));
				return;
			}

			Create.LOGGER.warn(messagePrefix + train.name.getString() + ": relocation failed server-side");
		});
		return true;
	}

}
