package com.simibubi.create.content.trains.graph;

import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;
import java.util.UUID;
import net.minecraft.class_2487;
import net.minecraft.class_2499;
import net.minecraft.class_2520;
import com.simibubi.create.Create;
import com.simibubi.create.content.trains.signal.TrackEdgePoint;
import com.simibubi.create.foundation.utility.NBTHelper;

public class EdgePointStorage {

	private Map<EdgePointType<?>, Map<UUID, TrackEdgePoint>> pointsByType;

	public EdgePointStorage() {
		pointsByType = new HashMap<>();
	}

	public <T extends TrackEdgePoint> void put(EdgePointType<T> type, TrackEdgePoint point) {
		getMap(type).put(point.getId(), point);
	}

	@SuppressWarnings("unchecked")
	public <T extends TrackEdgePoint> T get(EdgePointType<T> type, UUID id) {
		return (T) getMap(type).get(id);
	}

	@SuppressWarnings("unchecked")
	public <T extends TrackEdgePoint> T remove(EdgePointType<T> type, UUID id) {
		return (T) getMap(type).remove(id);
	}

	@SuppressWarnings("unchecked")
	public <T extends TrackEdgePoint> Collection<T> values(EdgePointType<T> type) {
		return getMap(type).values()
			.stream()
			.map(e -> (T) e)
			.toList();
	}

	public Map<UUID, TrackEdgePoint> getMap(EdgePointType<? extends TrackEdgePoint> type) {
		return pointsByType.computeIfAbsent(type, t -> new HashMap<>());
	}

	public void tick(TrackGraph graph, boolean preTrains) {
		pointsByType.values()
			.forEach(map -> map.values()
				.forEach(p -> p.tick(graph, preTrains)));
	}

	public void transferAll(TrackGraph target, EdgePointStorage other) {
		pointsByType.forEach((type, map) -> {
			other.getMap(type)
				.putAll(map);
			map.values()
				.forEach(ep -> Create.RAILWAYS.sync.pointAdded(target, ep));
		});
		pointsByType.clear();
	}

	public class_2487 write(DimensionPalette dimensions) {
		class_2487 nbt = new class_2487();
		for (Entry<EdgePointType<?>, Map<UUID, TrackEdgePoint>> entry : pointsByType.entrySet()) {
			EdgePointType<?> type = entry.getKey();
			class_2499 list = NBTHelper.writeCompoundList(entry.getValue()
				.values(), edgePoint -> {
					class_2487 tag = new class_2487();
					edgePoint.write(tag, dimensions);
					return tag;
				});
			nbt.method_10566(type.getId()
				.toString(), list);
		}
		return nbt;
	}

	public void read(class_2487 nbt, DimensionPalette dimensions) {
		for (EdgePointType<?> type : EdgePointType.TYPES.values()) {
			class_2499 list = nbt.method_10554(type.getId()
				.toString(), class_2520.field_33260);
			Map<UUID, TrackEdgePoint> map = getMap(type);
			NBTHelper.iterateCompoundList(list, tag -> {
				TrackEdgePoint edgePoint = type.create();
				edgePoint.read(tag, false, dimensions);
				map.put(edgePoint.getId(), edgePoint);
			});
		}
	}

}
