package com.simibubi.create.content.trains.graph;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Objects;
import java.util.UUID;
import net.minecraft.class_243;
import net.minecraft.class_2540;
import com.simibubi.create.Create;
import com.simibubi.create.content.trains.GlobalRailwayManager;
import com.simibubi.create.content.trains.signal.TrackEdgePoint;
import com.simibubi.create.content.trains.track.BezierConnection;
import com.simibubi.create.content.trains.track.TrackMaterial;
import com.simibubi.create.foundation.utility.Couple;
import com.simibubi.create.foundation.utility.Pair;
import com.simibubi.create.foundation.utility.VecHelper;

public class TrackGraphSyncPacket extends TrackGraphPacket {

	Map<Integer, Pair<TrackNodeLocation, class_243>> addedNodes;
	List<Pair<Pair<Couple<Integer>, TrackMaterial>, BezierConnection>> addedEdges;
	List<Integer> removedNodes;
	List<TrackEdgePoint> addedEdgePoints;
	List<UUID> removedEdgePoints;
	Map<Integer, Pair<Integer, UUID>> splitSubGraphs;
	Map<Couple<Integer>, Pair<Integer, List<UUID>>> updatedEdgeData;

	boolean fullWipe;

	static final int NULL_GROUP = 0, PASSIVE_GROUP = 1, GROUP = 2;

	public TrackGraphSyncPacket(UUID graphId, int netId) {
		this.graphId = graphId;
		this.netId = netId;
		addedNodes = new HashMap<>();
		addedEdges = new ArrayList<>();
		removedNodes = new ArrayList<>();
		addedEdgePoints = new ArrayList<>();
		removedEdgePoints = new ArrayList<>();
		updatedEdgeData = new HashMap<>();
		splitSubGraphs = new HashMap<>();
		packetDeletesGraph = false;
	}

	public TrackGraphSyncPacket(class_2540 buffer) {
		int size;

		graphId = buffer.method_10790();
		netId = buffer.readInt();
		packetDeletesGraph = buffer.readBoolean();
		fullWipe = buffer.readBoolean();

		if (packetDeletesGraph)
			return;

		DimensionPalette dimensions = DimensionPalette.receive(buffer);

		addedNodes = new HashMap<>();
		addedEdges = new ArrayList<>();
		addedEdgePoints = new ArrayList<>();
		removedEdgePoints = new ArrayList<>();
		removedNodes = new ArrayList<>();
		splitSubGraphs = new HashMap<>();
		updatedEdgeData = new HashMap<>();

		size = buffer.method_10816();
		for (int i = 0; i < size; i++)
			removedNodes.add(buffer.method_10816());

		size = buffer.method_10816();
		for (int i = 0; i < size; i++)
			addedNodes.put(buffer.method_10816(),
				Pair.of(TrackNodeLocation.receive(buffer, dimensions), VecHelper.read(buffer)));

		size = buffer.method_10816();
		for (int i = 0; i < size; i++)
			addedEdges.add(
				Pair.of(Pair.of(Couple.create(buffer::method_10816), TrackMaterial.deserialize(buffer.method_19772())), buffer.readBoolean() ? new BezierConnection(buffer) : null));

		size = buffer.method_10816();
		for (int i = 0; i < size; i++)
			addedEdgePoints.add(EdgePointType.read(buffer, dimensions));

		size = buffer.method_10816();
		for (int i = 0; i < size; i++)
			removedEdgePoints.add(buffer.method_10790());

		size = buffer.method_10816();
		for (int i = 0; i < size; i++) {
			ArrayList<UUID> list = new ArrayList<>();
			Couple<Integer> key = Couple.create(buffer::readInt);
			Pair<Integer, List<UUID>> entry = Pair.of(buffer.method_10816(), list);
			int size2 = buffer.method_10816();
			for (int j = 0; j < size2; j++)
				list.add(buffer.method_10790());
			updatedEdgeData.put(key, entry);
		}

		size = buffer.method_10816();
		for (int i = 0; i < size; i++)
			splitSubGraphs.put(buffer.method_10816(), Pair.of(buffer.readInt(), buffer.method_10790()));
	}

	@Override
	public void write(class_2540 buffer) {
		buffer.method_10797(graphId);
		buffer.writeInt(netId);
		buffer.writeBoolean(packetDeletesGraph);
		buffer.writeBoolean(fullWipe);

		if (packetDeletesGraph)
			return;

		// Populate and send palette ahead of time
		DimensionPalette dimensions = new DimensionPalette();
		addedNodes.forEach((node, loc) -> dimensions.encode(loc.getFirst().dimension));
		addedEdgePoints.forEach(ep -> ep.edgeLocation.forEach(loc -> dimensions.encode(loc.dimension)));
		dimensions.send(buffer);

		buffer.method_10804(removedNodes.size());
		removedNodes.forEach(buffer::method_10804);

		buffer.method_10804(addedNodes.size());
		addedNodes.forEach((node, loc) -> {
			buffer.method_10804(node);
			loc.getFirst()
				.send(buffer, dimensions);
			VecHelper.write(loc.getSecond(), buffer);
		});

		buffer.method_10804(addedEdges.size());
		addedEdges.forEach(pair -> {
			pair.getFirst().getFirst()
				.forEach(buffer::method_10804);
			buffer.method_10814(pair.getFirst().getSecond().id.toString());
			BezierConnection turn = pair.getSecond();
			buffer.writeBoolean(turn != null);
			if (turn != null)
				turn.write(buffer);
		});

		buffer.method_10804(addedEdgePoints.size());
		addedEdgePoints.forEach(ep -> ep.write(buffer, dimensions));

		buffer.method_10804(removedEdgePoints.size());
		removedEdgePoints.forEach(buffer::method_10797);

		buffer.method_10804(updatedEdgeData.size());
		for (Entry<Couple<Integer>, Pair<Integer, List<UUID>>> entry : updatedEdgeData.entrySet()) {
			entry.getKey()
				.forEach(buffer::writeInt);
			Pair<Integer, List<UUID>> pair = entry.getValue();
			buffer.method_10804(pair.getFirst());
			List<UUID> list = pair.getSecond();
			buffer.method_10804(list.size());
			list.forEach(buffer::method_10797);
		}

		buffer.method_10804(splitSubGraphs.size());
		splitSubGraphs.forEach((node, p) -> {
			buffer.method_10804(node);
			buffer.writeInt(p.getFirst());
			buffer.method_10797(p.getSecond());
		});
	}

	@Override
	protected void handle(GlobalRailwayManager manager, TrackGraph graph) {
		if (packetDeletesGraph) {
			manager.removeGraph(graph);
			return;
		}

		if (fullWipe) {
			manager.removeGraph(graph);
			graph = Create.RAILWAYS.sided(null)
				.getOrCreateGraph(graphId, netId);
		}

		for (int nodeId : removedNodes) {
			TrackNode node = graph.getNode(nodeId);
			if (node != null)
				graph.removeNode(null, node.getLocation());
		}

		for (Entry<Integer, Pair<TrackNodeLocation, class_243>> entry : addedNodes.entrySet()) {
			Integer nodeId = entry.getKey();
			Pair<TrackNodeLocation, class_243> nodeLocation = entry.getValue();
			graph.loadNode(nodeLocation.getFirst(), nodeId, nodeLocation.getSecond());
		}

		for (Pair<Pair<Couple<Integer>, TrackMaterial>, BezierConnection> pair : addedEdges) {
			Couple<TrackNode> nodes = pair.getFirst().getFirst()
				.map(graph::getNode);
			TrackNode node1 = nodes.getFirst();
			TrackNode node2 = nodes.getSecond();
			if (node1 != null && node2 != null)
				graph.putConnection(node1, node2, new TrackEdge(node1, node2, pair.getSecond(), pair.getFirst().getSecond()));
		}

		for (TrackEdgePoint edgePoint : addedEdgePoints)
			graph.edgePoints.put(edgePoint.getType(), edgePoint);

		for (UUID uuid : removedEdgePoints)
			for (EdgePointType<?> type : EdgePointType.TYPES.values())
				graph.edgePoints.remove(type, uuid);

		handleEdgeData(manager, graph);

		if (!splitSubGraphs.isEmpty())
			graph.findDisconnectedGraphs(null, splitSubGraphs)
				.forEach(manager::putGraph);
	}

	protected void handleEdgeData(GlobalRailwayManager manager, TrackGraph graph) {
		for (Entry<Couple<Integer>, Pair<Integer, List<UUID>>> entry : updatedEdgeData.entrySet()) {
			List<UUID> idList = entry.getValue()
				.getSecond();
			int groupType = entry.getValue()
				.getFirst();

			Couple<TrackNode> nodes = entry.getKey()
				.map(graph::getNode);
			if (nodes.either(Objects::isNull))
				continue;
			TrackEdge edge = graph.getConnectionsFrom(nodes.getFirst())
				.get(nodes.getSecond());
			if (edge == null)
				continue;

			EdgeData edgeData = new EdgeData(edge);
			if (groupType == NULL_GROUP)
				edgeData.setSingleSignalGroup(null, null);
			else if (groupType == PASSIVE_GROUP)
				edgeData.setSingleSignalGroup(null, EdgeData.passiveGroup);
			else
				edgeData.setSingleSignalGroup(null, idList.get(0));

			List<TrackEdgePoint> points = edgeData.getPoints();
			edge.edgeData = edgeData;

			for (int i = groupType == GROUP ? 1 : 0; i < idList.size(); i++) {
				UUID uuid = idList.get(i);
				for (EdgePointType<?> type : EdgePointType.TYPES.values()) {
					TrackEdgePoint point = graph.edgePoints.get(type, uuid);
					if (point == null)
						continue;
					points.add(point);
					break;
				}
			}
		}
	}

	public void syncEdgeData(TrackNode node1, TrackNode node2, TrackEdge edge) {
		Couple<Integer> key = Couple.create(node1.getNetId(), node2.getNetId());
		List<UUID> list = new ArrayList<>();
		EdgeData edgeData = edge.getEdgeData();
		int groupType = edgeData.hasSignalBoundaries() ? NULL_GROUP
			: EdgeData.passiveGroup.equals(edgeData.getSingleSignalGroup()) ? PASSIVE_GROUP : GROUP;
		if (groupType == GROUP)
			list.add(edgeData.getSingleSignalGroup());
		for (TrackEdgePoint point : edgeData.getPoints())
			list.add(point.getId());
		updatedEdgeData.put(key, Pair.of(groupType, list));
	}

}
