package com.simibubi.create.content.trains.graph;

import java.util.Collection;
import java.util.HashSet;
import java.util.Objects;
import java.util.Set;
import net.minecraft.class_1937;
import net.minecraft.class_2338;
import net.minecraft.class_2382;
import net.minecraft.class_243;
import net.minecraft.class_2487;
import net.minecraft.class_2512;
import net.minecraft.class_2540;
import net.minecraft.class_3532;
import net.minecraft.class_5321;
import com.simibubi.create.content.trains.track.BezierConnection;
import com.simibubi.create.content.trains.track.TrackMaterial;
import com.simibubi.create.foundation.utility.Iterate;

public class TrackNodeLocation extends class_2382 {

	public class_5321<class_1937> dimension;
	public int yOffsetPixels;

	public TrackNodeLocation(class_243 vec) {
		this(vec.field_1352, vec.field_1351, vec.field_1350);
	}

	public TrackNodeLocation(double x, double y, double z) {
		super(class_3532.method_15375(Math.round(x * 2)), class_3532.method_15357(y) * 2, class_3532.method_15375(Math.round(z * 2)));
	}

	public TrackNodeLocation in(class_1937 level) {
		return in(level.method_27983());
	}

	public TrackNodeLocation in(class_5321<class_1937> dimension) {
		this.dimension = dimension;
		return this;
	}

	private static TrackNodeLocation fromPackedPos(class_2338 bufferPos) {
		return new TrackNodeLocation(bufferPos);
	}

	private TrackNodeLocation(class_2338 readBlockPos) {
		super(readBlockPos.method_10263(), readBlockPos.method_10264(), readBlockPos.method_10260());
	}

	public class_243 getLocation() {
		return new class_243(method_10263() / 2.0, method_10264() / 2.0 + yOffsetPixels / 16.0, method_10260() / 2.0);
	}

	public class_5321<class_1937> getDimension() {
		return dimension;
	}

	@Override
	public boolean equals(Object pOther) {
		return equalsIgnoreDim(pOther) && pOther instanceof TrackNodeLocation tnl
			&& Objects.equals(tnl.dimension, dimension);
	}

	public boolean equalsIgnoreDim(Object pOther) {
		return super.equals(pOther) && pOther instanceof TrackNodeLocation tnl && tnl.yOffsetPixels == yOffsetPixels;
	}

	@Override
	public int hashCode() {
		return (method_10264() + ((method_10260() + yOffsetPixels * 31) * 31 + dimension.hashCode()) * 31) * 31 + method_10263();
	}

	public class_2487 write(DimensionPalette dimensions) {
		class_2487 c = class_2512.method_10692(new class_2338(this));
		if (dimensions != null)
			c.method_10569("D", dimensions.encode(dimension));
		if (yOffsetPixels != 0)
			c.method_10569("YO", yOffsetPixels);
		return c;
	}

	public static TrackNodeLocation read(class_2487 tag, DimensionPalette dimensions) {
		TrackNodeLocation location = fromPackedPos(class_2512.method_10691(tag));
		if (dimensions != null)
			location.dimension = dimensions.decode(tag.method_10550("D"));
		location.yOffsetPixels = tag.method_10550("YO");
		return location;
	}

	public void send(class_2540 buffer, DimensionPalette dimensions) {
		buffer.method_10804(method_10263());
		buffer.writeShort(method_10264());
		buffer.method_10804(method_10260());
		buffer.method_10804(yOffsetPixels);
		buffer.method_10804(dimensions.encode(dimension));
	}

	public static TrackNodeLocation receive(class_2540 buffer, DimensionPalette dimensions) {
		TrackNodeLocation location = fromPackedPos(new class_2338(
				buffer.method_10816(),
				buffer.readShort(),
				buffer.method_10816()
		));
		location.yOffsetPixels = buffer.method_10816();
		location.dimension = dimensions.decode(buffer.method_10816());
		return location;
	}

	public Collection<class_2338> allAdjacent() {
		Set<class_2338> set = new HashSet<>();
		class_243 vec3 = getLocation().method_1023(0, yOffsetPixels / 16.0, 0);
		double step = 1 / 8f;
		for (int x : Iterate.positiveAndNegative)
			for (int y : Iterate.positiveAndNegative)
				for (int z : Iterate.positiveAndNegative)
					set.add(class_2338.method_49638(vec3.method_1031(x * step, y * step, z * step)));
		return set;
	}

	public static class DiscoveredLocation extends TrackNodeLocation {

		BezierConnection turn = null;
		boolean forceNode = false;
		class_243 direction;
		class_243 normal;
		TrackMaterial materialA;
		TrackMaterial materialB;

		public DiscoveredLocation(class_1937 level, double x, double y, double z) {
			super(x, y, z);
			in(level);
		}

		public DiscoveredLocation(class_5321<class_1937> dimension, class_243 vec) {
			super(vec);
			in(dimension);
		}

		public DiscoveredLocation(class_1937 level, class_243 vec) {
			this(level.method_27983(), vec);
		}

		public DiscoveredLocation materialA(TrackMaterial material) {
			this.materialA = material;
			return this;
		}

		public DiscoveredLocation materialB(TrackMaterial material) {
			this.materialB = material;
			return this;
		}

		public DiscoveredLocation materials(TrackMaterial materialA, TrackMaterial materialB) {
			this.materialA = materialA;
			this.materialB = materialB;
			return this;
		}

		public DiscoveredLocation viaTurn(BezierConnection turn) {
			this.turn = turn;
			if (turn != null)
				forceNode();
			return this;
		}

		public DiscoveredLocation forceNode() {
			forceNode = true;
			return this;
		}

		public DiscoveredLocation withNormal(class_243 normal) {
			this.normal = normal;
			return this;
		}
		
		public DiscoveredLocation withYOffset(int yOffsetPixels) {
			this.yOffsetPixels = yOffsetPixels;
			return this;
		}

		public DiscoveredLocation withDirection(class_243 direction) {
			this.direction = direction == null ? null : direction.method_1029();
			return this;
		}

		public boolean connectedViaTurn() {
			return turn != null;
		}

		public BezierConnection getTurn() {
			return turn;
		}

		public boolean shouldForceNode() {
			return forceNode;
		}

		public boolean differentMaterials() {
			return materialA != materialB;
		}

		public boolean notInLineWith(class_243 direction) {
			return this.direction != null
				&& Math.max(direction.method_1026(this.direction), direction.method_1026(this.direction.method_1021(-1))) < 7 / 8f;
		}

		public class_243 getDirection() {
			return direction;
		}

	}

}
