package com.simibubi.create.content.trains.observer;

import java.util.UUID;
import net.minecraft.class_1799;
import net.minecraft.class_1937;
import net.minecraft.class_2487;
import net.minecraft.class_2540;
import net.minecraft.class_2586;
import com.simibubi.create.Create;
import com.simibubi.create.content.logistics.filter.FilterItemStack;
import com.simibubi.create.content.trains.entity.Train;
import com.simibubi.create.content.trains.graph.DimensionPalette;
import com.simibubi.create.content.trains.graph.TrackEdge;
import com.simibubi.create.content.trains.graph.TrackGraph;
import com.simibubi.create.content.trains.signal.SignalPropagator;
import com.simibubi.create.content.trains.signal.SingleBlockEntityEdgePoint;
import com.simibubi.create.foundation.blockEntity.behaviour.BlockEntityBehaviour;
import com.simibubi.create.foundation.blockEntity.behaviour.filtering.FilteringBehaviour;

import io.github.fabricators_of_create.porting_lib.util.NBTSerializer;

public class TrackObserver extends SingleBlockEntityEdgePoint {

	private int activated;
	private FilterItemStack filter;
	private UUID currentTrain;

	public TrackObserver() {
		activated = 0;
		filter = FilterItemStack.empty();
		currentTrain = null;
	}

	@Override
	public void blockEntityAdded(class_2586 blockEntity, boolean front) {
		super.blockEntityAdded(blockEntity, front);
		FilteringBehaviour filteringBehaviour = BlockEntityBehaviour.get(blockEntity, FilteringBehaviour.TYPE);
		if (filteringBehaviour != null)
			setFilterAndNotify(blockEntity.method_10997(), filteringBehaviour.getFilter());
	}

	@Override
	public void tick(TrackGraph graph, boolean preTrains) {
		super.tick(graph, preTrains);
		if (isActivated())
			activated--;
		if (!isActivated())
			currentTrain = null;
	}

	public void setFilterAndNotify(class_1937 level, class_1799 filter) {
		this.filter = FilterItemStack.of(filter.method_7972());
		notifyTrains(level);
	}

	private void notifyTrains(class_1937 level) {
		TrackGraph graph = Create.RAILWAYS.sided(level)
			.getGraph(level, edgeLocation.getFirst());
		if (graph == null)
			return;
		TrackEdge edge = graph.getConnection(edgeLocation.map(graph::locateNode));
		if (edge == null)
			return;
		SignalPropagator.notifyTrains(graph, edge);
	}

	public FilterItemStack getFilter() {
		return filter;
	}

	public UUID getCurrentTrain() {
		return currentTrain;
	}

	public boolean isActivated() {
		return activated > 0;
	}

	public void keepAlive(Train train) {
		activated = 8;
		currentTrain = train.id;
	}

	@Override
	public void read(class_2487 nbt, boolean migration, DimensionPalette dimensions) {
		super.read(nbt, migration, dimensions);
		activated = nbt.method_10550("Activated");
		filter = FilterItemStack.of(nbt.method_10562("Filter"));
		if (nbt.method_10545("TrainId"))
			currentTrain = nbt.method_25926("TrainId");
	}

	@Override
	public void read(class_2540 buffer, DimensionPalette dimensions) {
		super.read(buffer, dimensions);
	}

	@Override
	public void write(class_2487 nbt, DimensionPalette dimensions) {
		super.write(nbt, dimensions);
		nbt.method_10569("Activated", activated);
		nbt.method_10566("Filter", NBTSerializer.serializeNBTCompound(filter));
		if (currentTrain != null)
			nbt.method_25927("TrainId", currentTrain);
	}

	@Override
	public void write(class_2540 buffer, DimensionPalette dimensions) {
		super.write(buffer, dimensions);
	}

}
