package com.simibubi.create.content.trains.schedule.condition;

import java.util.Arrays;
import java.util.List;

import com.google.common.collect.ImmutableList;
import com.simibubi.create.content.trains.entity.Carriage;
import com.simibubi.create.content.trains.entity.Train;
import com.simibubi.create.foundation.gui.ModularGuiLineBuilder;
import com.simibubi.create.foundation.utility.Components;
import com.simibubi.create.foundation.utility.Lang;
import com.simibubi.create.foundation.utility.Pair;
import net.fabricmc.api.EnvType;
import net.fabricmc.api.Environment;
import net.minecraft.class_124;
import net.minecraft.class_1799;
import net.minecraft.class_1937;
import net.minecraft.class_2487;
import net.minecraft.class_2561;

public abstract class CargoThresholdCondition extends LazyTickedScheduleCondition {
	public static enum Ops {
		GREATER(">"), LESS("<"), EQUAL("=");

		public String formatted;

		private Ops(String formatted) {
			this.formatted = formatted;
		}

		public boolean test(long current, long target) {
			return switch (this) {
			case GREATER -> current > target;
			case EQUAL -> current == target;
			case LESS -> current < target;
			default -> throw new IllegalArgumentException("Unexpected value: " + this);
			};
		}

		public static List<? extends class_2561> translatedOptions() {
			return Arrays.stream(values())
				.map(op -> Lang.translateDirect("schedule.condition.threshold." + Lang.asId(op.name())))
				.toList();
		}
	}

	public CargoThresholdCondition() {
		super(20);
		data.method_10582("Threshold", "10");
	}

	@Override
	public boolean lazyTickCompletion(class_1937 level, Train train, class_2487 context) {
		int lastChecked = context.method_10545("LastChecked") ? context.method_10550("LastChecked") : -1;
		int status = 0;
		for (Carriage carriage : train.carriages)
			status += carriage.storage.getVersion();
		if (status == lastChecked)
			return false;
		context.method_10569("LastChecked", status);
		return test(level, train, context);
	}

	protected void requestStatusToUpdate(long amount, class_2487 context) {
		context.method_10544("CurrentDisplay", amount);
		super.requestStatusToUpdate(context);
	};

	protected long getLastDisplaySnapshot(class_2487 context) {
		if (!context.method_10545("CurrentDisplay"))
			return -1;
		return context.method_10537("CurrentDisplay");
	}

	protected abstract boolean test(class_1937 level, Train train, class_2487 context);

	protected abstract class_2561 getUnit();

	protected abstract class_1799 getIcon();

	@Override
	public Pair<class_1799, class_2561> getSummary() {
		return Pair.of(getIcon(), Components.literal(getOperator().formatted + " " + getThreshold()).method_10852(getUnit()));
	}

	@Override
	public int slotsTargeted() {
		return 1;
	}

	public Ops getOperator() {
		return enumData("Operator", Ops.class);
	}

	public long getThreshold() {
		try {
			return Long.parseLong(textData("Threshold"));
		} catch (NumberFormatException e) {
			data.method_10582("Threshold", "0");
		}
		return 0;
	}

	public int getMeasure() {
		return intData("Measure");
	}

	@Override
	public List<class_2561> getSecondLineTooltip(int slot) {
		return ImmutableList.of(Lang.translateDirect("schedule.condition.threshold.place_item"),
			Lang.translateDirect("schedule.condition.threshold.place_item_2")
				.method_27692(class_124.field_1080),
			Lang.translateDirect("schedule.condition.threshold.place_item_3")
				.method_27692(class_124.field_1080));
	}

	@Override
	@Environment(EnvType.CLIENT)
	public void initConfigurationWidgets(ModularGuiLineBuilder builder) {
		builder.addSelectionScrollInput(0, 24, (i, l) -> {
			i.forOptions(Ops.translatedOptions())
				.titled(Lang.translateDirect("schedule.condition.threshold.train_holds", ""))
				.format(state -> Components.literal(" " + Ops.values()[state].formatted));
		}, "Operator");
		builder.addIntegerTextInput(29, 41, (e, t) -> {
		}, "Threshold");
	}

}
