package com.simibubi.create.content.trains.schedule.condition;

import java.util.List;

import com.google.common.collect.ImmutableList;
import com.simibubi.create.Create;
import com.simibubi.create.content.logistics.filter.FilterItemStack;
import com.simibubi.create.content.trains.entity.Carriage;
import com.simibubi.create.content.trains.entity.Train;
import com.simibubi.create.foundation.fluid.CombinedTankWrapper;
import com.simibubi.create.foundation.gui.ModularGuiLineBuilder;
import com.simibubi.create.foundation.utility.Components;
import com.simibubi.create.foundation.utility.Lang;

import io.github.fabricators_of_create.porting_lib.transfer.TransferUtil;
import io.github.fabricators_of_create.porting_lib.fluids.FluidStack;
import io.github.fabricators_of_create.porting_lib.util.NBTSerializer;
import net.fabricmc.fabric.api.transfer.v1.fluid.FluidConstants;
import net.fabricmc.fabric.api.transfer.v1.fluid.FluidVariant;
import net.fabricmc.fabric.api.transfer.v1.storage.StorageView;
import net.fabricmc.fabric.api.transfer.v1.transaction.Transaction;
import net.minecraft.class_124;
import net.minecraft.class_1799;
import net.minecraft.class_1937;
import net.minecraft.class_2487;
import net.minecraft.class_2561;
import net.minecraft.class_2960;
import net.minecraft.class_310;
import net.minecraft.class_5250;
import net.fabricmc.api.EnvType;
import net.fabricmc.api.Environment;

public class FluidThresholdCondition extends CargoThresholdCondition {

	private FilterItemStack compareStack = FilterItemStack.empty();

	@Override
	protected class_2561 getUnit() {
		return Components.literal("b");
	}

	@Override
	protected class_1799 getIcon() {
		return compareStack.item();
	}

	@Override
	protected boolean test(class_1937 level, Train train, class_2487 context) {
		Ops operator = getOperator();
		long target = getThreshold();

		int foundFluid = 0;
		for (Carriage carriage : train.carriages) {
			CombinedTankWrapper fluids = carriage.storage.getFluids();
			try (Transaction t = TransferUtil.getTransaction()) {
				for (StorageView<FluidVariant> view : fluids.nonEmptyViews()) {
					FluidStack fluidInTank = new FluidStack(view);
					if (!compareStack.test(level, fluidInTank))
						continue;
					foundFluid += fluidInTank.getAmount();
				}
			}
		}

		requestStatusToUpdate(foundFluid / FluidConstants.BUCKET, context);
		return operator.test(foundFluid, target * FluidConstants.BUCKET);
	}

	@Override
	protected void writeAdditional(class_2487 tag) {
		super.writeAdditional(tag);
		tag.method_10566("Bucket", NBTSerializer.serializeNBTCompound(compareStack));
	}

	@Override
	protected void readAdditional(class_2487 tag) {
		super.readAdditional(tag);
		if (tag.method_10545("Bucket"))
			compareStack = FilterItemStack.of(tag.method_10562("Bucket"));
	}

	@Override
	public boolean tickCompletion(class_1937 level, Train train, class_2487 context) {
		return super.tickCompletion(level, train, context);
	}

	@Environment(EnvType.CLIENT)
	private FluidStack loadFluid() {
		return compareStack.fluid(class_310.method_1551().field_1687);
	}

	@Override
	public List<class_2561> getTitleAs(String type) {
		return ImmutableList.of(
			Lang.translateDirect("schedule.condition.threshold.train_holds",
				Lang.translateDirect("schedule.condition.threshold." + Lang.asId(getOperator().name()))),
			Lang.translateDirect("schedule.condition.threshold.x_units_of_item", getThreshold(),
				Lang.translateDirect("schedule.condition.threshold.buckets"),
				compareStack.isEmpty() ? Lang.translateDirect("schedule.condition.threshold.anything")
					: compareStack.isFilterItem()
						? Lang.translateDirect("schedule.condition.threshold.matching_content")
						: loadFluid().getDisplayName())
				.method_27692(class_124.field_1062));
	}

	@Override
	public void setItem(int slot, class_1799 stack) {
		compareStack = FilterItemStack.of(stack);
	}

	@Override
	public class_1799 getItem(int slot) {
		return compareStack.item();
	}

	@Override
	public class_2960 getId() {
		return Create.asResource("fluid_threshold");
	}

	@Override
	@Environment(EnvType.CLIENT)
	public void initConfigurationWidgets(ModularGuiLineBuilder builder) {
		super.initConfigurationWidgets(builder);
		builder.addSelectionScrollInput(71, 50, (i, l) -> {
			i.forOptions(ImmutableList.of(Lang.translateDirect("schedule.condition.threshold.buckets")))
				.titled(null);
		}, "Measure");
	}

	@Override
	public class_5250 getWaitingStatus(class_1937 level, Train train, class_2487 tag) {
		long lastDisplaySnapshot = getLastDisplaySnapshot(tag);
		if (lastDisplaySnapshot == -1)
			return Components.empty();
		int offset = getOperator() == Ops.LESS ? -1 : getOperator() == Ops.GREATER ? 1 : 0;
		return Lang.translateDirect("schedule.condition.threshold.status", lastDisplaySnapshot,
			Math.max(0, getThreshold() + offset), Lang.translateDirect("schedule.condition.threshold.buckets"));
	}

}
