package com.simibubi.create.content.trains.schedule.condition;

import java.util.List;

import com.google.common.collect.ImmutableList;
import com.simibubi.create.AllBlocks;
import com.simibubi.create.Create;
import com.simibubi.create.content.redstone.link.RedstoneLinkNetworkHandler.Frequency;
import com.simibubi.create.content.trains.entity.Train;
import com.simibubi.create.foundation.gui.ModularGuiLineBuilder;
import com.simibubi.create.foundation.utility.Components;
import com.simibubi.create.foundation.utility.Couple;
import com.simibubi.create.foundation.utility.Lang;
import com.simibubi.create.foundation.utility.Pair;

import io.github.fabricators_of_create.porting_lib.util.NBTSerializer;
import net.fabricmc.api.EnvType;
import net.fabricmc.api.Environment;
import net.minecraft.class_124;
import net.minecraft.class_1799;
import net.minecraft.class_1937;
import net.minecraft.class_2487;
import net.minecraft.class_2520;
import net.minecraft.class_2561;
import net.minecraft.class_2960;
import net.minecraft.class_5250;

public class RedstoneLinkCondition extends ScheduleWaitCondition {

	public Couple<Frequency> freq;

	public RedstoneLinkCondition() {
		freq = Couple.create(() -> Frequency.EMPTY);
	}

	@Override
	public int slotsTargeted() {
		return 2;
	}

	@Override
	public Pair<class_1799, class_2561> getSummary() {
		return Pair.of(AllBlocks.REDSTONE_LINK.asStack(),
			lowActivation() ? Lang.translateDirect("schedule.condition.redstone_link_off")
				: Lang.translateDirect("schedule.condition.redstone_link_on"));
	}

	@Override
	public List<class_2561> getSecondLineTooltip(int slot) {
		return ImmutableList.of(Lang.translateDirect(slot == 0 ? "logistics.firstFrequency" : "logistics.secondFrequency")
			.method_27692(class_124.field_1061));
	}

	@Override
	public List<class_2561> getTitleAs(String type) {
		return ImmutableList.of(
			Lang.translateDirect("schedule.condition.redstone_link.frequency_" + (lowActivation() ? "unpowered" : "powered")),
			Components.literal(" #1 ").method_27692(class_124.field_1080)
				.method_10852(freq.getFirst()
					.getStack()
					.method_7964()
					.method_27661()
					.method_27692(class_124.field_1062)),
			Components.literal(" #2 ").method_27692(class_124.field_1080)
				.method_10852(freq.getSecond()
					.getStack()
					.method_7964()
					.method_27661()
					.method_27692(class_124.field_1062)));
	}

	@Override
	public boolean tickCompletion(class_1937 level, Train train, class_2487 context) {
		int lastChecked = context.method_10545("LastChecked") ? context.method_10550("LastChecked") : -1;
		int status = Create.REDSTONE_LINK_NETWORK_HANDLER.globalPowerVersion.get();
		if (status == lastChecked)
			return false;
		context.method_10569("LastChecked", status);
		return Create.REDSTONE_LINK_NETWORK_HANDLER.hasAnyLoadedPower(freq) != lowActivation();
	}

	@Override
	public void setItem(int slot, class_1799 stack) {
		freq.set(slot == 0, Frequency.of(stack));
		super.setItem(slot, stack);
	}

	@Override
	public class_1799 getItem(int slot) {
		return freq.get(slot == 0)
			.getStack();
	}

	@Override
	public class_2960 getId() {
		return Create.asResource("redstone_link");
	}

	@Override
	protected void writeAdditional(class_2487 tag) {
		tag.method_10566("Frequency", freq.serializeEach(f -> NBTSerializer.serializeNBTCompound(f.getStack())));
	}

	public boolean lowActivation() {
		return intData("Inverted") == 1;
	}

	@Override
	protected void readAdditional(class_2487 tag) {
		if (tag.method_10545("Frequency"))
			freq = Couple.deserializeEach(tag.method_10554("Frequency", class_2520.field_33260), c -> Frequency.of(class_1799.method_7915(c)));
	}

	@Override
	@Environment(EnvType.CLIENT)
	public void initConfigurationWidgets(ModularGuiLineBuilder builder) {
		builder.addSelectionScrollInput(20, 101,
			(i, l) -> i.forOptions(Lang.translatedOptions("schedule.condition.redstone_link", "powered", "unpowered"))
				.titled(Lang.translateDirect("schedule.condition.redstone_link.frequency_state")),
			"Inverted");
	}

	@Override
	public class_5250 getWaitingStatus(class_1937 level, Train train, class_2487 tag) {
		return Lang.translateDirect("schedule.condition.redstone_link.status");
	}

}
