package com.simibubi.create.content.trains.schedule.condition;

import java.util.List;

import com.google.common.collect.ImmutableList;
import com.simibubi.create.content.trains.entity.Train;
import com.simibubi.create.foundation.gui.ModularGuiLineBuilder;
import com.simibubi.create.foundation.utility.Components;
import com.simibubi.create.foundation.utility.Lang;
import net.fabricmc.api.EnvType;
import net.fabricmc.api.Environment;
import net.minecraft.class_124;
import net.minecraft.class_1799;
import net.minecraft.class_1802;
import net.minecraft.class_1937;
import net.minecraft.class_2487;
import net.minecraft.class_2561;
import net.minecraft.class_5250;

public abstract class TimedWaitCondition extends ScheduleWaitCondition {

	public static enum TimeUnit {
		TICKS(1, "t", "generic.unit.ticks"),
		SECONDS(20, "s", "generic.unit.seconds"),
		MINUTES(20 * 60, "min", "generic.unit.minutes");

		public int ticksPer;
		public String suffix;
		public String key;

		private TimeUnit(int ticksPer, String suffix, String key) {
			this.ticksPer = ticksPer;
			this.suffix = suffix;
			this.key = key;
		}

		public static List<class_2561> translatedOptions() {
			return Lang.translatedOptions(null, TICKS.key, SECONDS.key, MINUTES.key);
		}
	}

	protected void requestDisplayIfNecessary(class_2487 context, int time) {
		int ticksUntilDeparture = totalWaitTicks() - time;
		if (ticksUntilDeparture < 20 * 60 && ticksUntilDeparture % 100 == 0)
			requestStatusToUpdate(context);
		if (ticksUntilDeparture >= 20 * 60 && ticksUntilDeparture % (20 * 60) == 0)
			requestStatusToUpdate(context);
	}

	public int totalWaitTicks() {
		return getValue() * getUnit().ticksPer;
	}

	public TimedWaitCondition() {
		data.method_10569("Value", 5);
		data.method_10569("TimeUnit", TimeUnit.SECONDS.ordinal());
	}

	protected class_2561 formatTime(boolean compact) {
		if (compact)
			return Components.literal(getValue() + getUnit().suffix);
		return Components.literal(getValue() + " ").method_10852(Lang.translateDirect(getUnit().key));
	}

	@Override
	public List<class_2561> getTitleAs(String type) {
		return ImmutableList.of(
			Components.translatable(getId().method_12836() + ".schedule." + type + "." + getId().method_12832()),
			Lang.translateDirect("schedule.condition.for_x_time", formatTime(false))
				.method_27692(class_124.field_1062));
	}

	@Override
	public class_1799 getSecondLineIcon() {
		return new class_1799(class_1802.field_8619);
	}

	@Override
	public List<class_2561> getSecondLineTooltip(int slot) {
		return ImmutableList.of(Lang.translateDirect("generic.duration"));
	}

	public int getValue() {
		return intData("Value");
	}

	public TimeUnit getUnit() {
		return enumData("TimeUnit", TimeUnit.class);
	}

	@Override
	@Environment(EnvType.CLIENT)
	public void initConfigurationWidgets(ModularGuiLineBuilder builder) {
		builder.addScrollInput(0, 31, (i, l) -> {
			i.titled(Lang.translateDirect("generic.duration"))
				.withShiftStep(15)
				.withRange(0, 121);
			i.lockedTooltipX = -15;
			i.lockedTooltipY = 35;
		}, "Value");

		builder.addSelectionScrollInput(36, 85, (i, l) -> {
			i.forOptions(TimeUnit.translatedOptions())
				.titled(Lang.translateDirect("generic.timeUnit"));
		}, "TimeUnit");
	}

	@Override
	public class_5250 getWaitingStatus(class_1937 level, Train train, class_2487 tag) {
		int time = tag.method_10550("Time");
		int ticksUntilDeparture = totalWaitTicks() - time;
		boolean showInMinutes = ticksUntilDeparture >= 20 * 60;
		int num =
			(int) (showInMinutes ? Math.floor(ticksUntilDeparture / (20 * 60f)) : Math.ceil(ticksUntilDeparture / 100f) * 5);
		String key = "generic." + (showInMinutes ? num == 1 ? "daytime.minute" : "unit.minutes"
			: num == 1 ? "daytime.second" : "unit.seconds");
		return Lang.translateDirect("schedule.condition." + getId().method_12832() + ".status",
			Components.literal(num + " ").method_10852(Lang.translateDirect(key)));
	}

}
