package com.simibubi.create.content.trains.signal;

import java.util.UUID;
import net.minecraft.class_1936;
import net.minecraft.class_2338;
import net.minecraft.class_2487;
import net.minecraft.class_2520;
import net.minecraft.class_2540;
import net.minecraft.class_2586;
import com.simibubi.create.Create;
import com.simibubi.create.content.trains.graph.DimensionPalette;
import com.simibubi.create.content.trains.graph.EdgePointType;
import com.simibubi.create.content.trains.graph.TrackEdge;
import com.simibubi.create.content.trains.graph.TrackGraph;
import com.simibubi.create.content.trains.graph.TrackNode;
import com.simibubi.create.content.trains.graph.TrackNodeLocation;
import com.simibubi.create.content.trains.track.TrackTargetingBehaviour;
import com.simibubi.create.foundation.blockEntity.behaviour.BlockEntityBehaviour;
import com.simibubi.create.foundation.utility.Couple;

public abstract class TrackEdgePoint {

	public UUID id;
	public Couple<TrackNodeLocation> edgeLocation;
	public double position;
	private EdgePointType<?> type;

	public void setId(UUID id) {
		this.id = id;
	}

	public UUID getId() {
		return id;
	}

	public void setType(EdgePointType<?> type) {
		this.type = type;
	}

	public EdgePointType<?> getType() {
		return type;
	}

	public abstract boolean canMerge();

	public boolean canCoexistWith(EdgePointType<?> otherType, boolean front) {
		return false;
	}

	public abstract void invalidate(class_1936 level);

	protected void invalidateAt(class_1936 level, class_2338 blockEntityPos) {
		TrackTargetingBehaviour<?> behaviour = BlockEntityBehaviour.get(level, blockEntityPos, TrackTargetingBehaviour.TYPE);
		if (behaviour == null)
			return;
		class_2487 migrationData = new class_2487();
		DimensionPalette dimensions = new DimensionPalette();
		write(migrationData, dimensions);
		dimensions.write(migrationData);
		behaviour.invalidateEdgePoint(migrationData);
	}

	public abstract void blockEntityAdded(class_2586 blockEntity, boolean front);

	public abstract void blockEntityRemoved(class_2338 blockEntityPos, boolean front);

	public void onRemoved(TrackGraph graph) {}

	public void setLocation(Couple<TrackNodeLocation> nodes, double position) {
		this.edgeLocation = nodes;
		this.position = position;
	}

	public double getLocationOn(TrackEdge edge) {
		return isPrimary(edge.node1) ? edge.getLength() - position : position;
	}

	public boolean canNavigateVia(TrackNode side) {
		return true;
	}

	public boolean isPrimary(TrackNode node1) {
		return edgeLocation.getSecond()
			.equals(node1.getLocation());
	}

	public void read(class_2487 nbt, boolean migration, DimensionPalette dimensions) {
		if (migration)
			return;

		id = nbt.method_25926("Id");
		position = nbt.method_10574("Position");
		edgeLocation = Couple.deserializeEach(nbt.method_10554("Edge", class_2520.field_33260),
			tag -> TrackNodeLocation.read(tag, dimensions));
	}

	public void read(class_2540 buffer, DimensionPalette dimensions) {
		id = buffer.method_10790();
		edgeLocation = Couple.create(() -> TrackNodeLocation.receive(buffer, dimensions));
		position = buffer.readDouble();
	}

	public void write(class_2487 nbt, DimensionPalette dimensions) {
		nbt.method_25927("Id", id);
		nbt.method_10549("Position", position);
		nbt.method_10566("Edge", edgeLocation.serializeEach(loc -> loc.write(dimensions)));
	}

	public void write(class_2540 buffer, DimensionPalette dimensions) {
		buffer.method_10812(type.getId());
		buffer.method_10797(id);
		edgeLocation.forEach(loc -> loc.send(buffer, dimensions));
		buffer.writeDouble(position);
	}

	public void tick(TrackGraph graph, boolean preTrains) {}

	protected void removeFromAllGraphs() {
		for (TrackGraph trackGraph : Create.RAILWAYS.trackNetworks.values())
			if (trackGraph.removePoint(getType(), id) != null)
				return;
	}

}
