package com.simibubi.create.content.trains.station;

import java.lang.ref.WeakReference;

import javax.annotation.Nullable;
import net.minecraft.class_2487;
import net.minecraft.class_2540;
import net.minecraft.class_2586;
import net.minecraft.class_2680;
import com.simibubi.create.content.trains.entity.Train;
import com.simibubi.create.content.trains.graph.DimensionPalette;
import com.simibubi.create.content.trains.graph.TrackNode;
import com.simibubi.create.content.trains.signal.SingleBlockEntityEdgePoint;

public class GlobalStation extends SingleBlockEntityEdgePoint {

	public String name;
	public WeakReference<Train> nearestTrain;
	public boolean assembling;

	public GlobalStation() {
		name = "Track Station";
		nearestTrain = new WeakReference<Train>(null);
	}

	@Override
	public void blockEntityAdded(class_2586 blockEntity, boolean front) {
		super.blockEntityAdded(blockEntity, front);
		class_2680 state = blockEntity.method_11010();
		assembling =
			state != null && state.method_28498(StationBlock.ASSEMBLING) && state.method_11654(StationBlock.ASSEMBLING);
	}

	@Override
	public void read(class_2487 nbt, boolean migration, DimensionPalette dimensions) {
		super.read(nbt, migration, dimensions);
		name = nbt.method_10558("Name");
		assembling = nbt.method_10577("Assembling");
		nearestTrain = new WeakReference<Train>(null);
	}

	@Override
	public void read(class_2540 buffer, DimensionPalette dimensions) {
		super.read(buffer, dimensions);
		name = buffer.method_19772();
		assembling = buffer.readBoolean();
		if (buffer.readBoolean())
			blockEntityPos = buffer.method_10811();
	}

	@Override
	public void write(class_2487 nbt, DimensionPalette dimensions) {
		super.write(nbt, dimensions);
		nbt.method_10582("Name", name);
		nbt.method_10556("Assembling", assembling);
	}

	@Override
	public void write(class_2540 buffer, DimensionPalette dimensions) {
		super.write(buffer, dimensions);
		buffer.method_10814(name);
		buffer.writeBoolean(assembling);
		buffer.writeBoolean(blockEntityPos != null);
		if (blockEntityPos != null)
			buffer.method_10807(blockEntityPos);
	}

	public boolean canApproachFrom(TrackNode side) {
		return isPrimary(side) && !assembling;
	}

	@Override
	public boolean canNavigateVia(TrackNode side) {
		return super.canNavigateVia(side) && !assembling;
	}

	public void reserveFor(Train train) {
		Train nearestTrain = getNearestTrain();
		if (nearestTrain == null
			|| nearestTrain.navigation.distanceToDestination > train.navigation.distanceToDestination)
			this.nearestTrain = new WeakReference<>(train);
	}

	public void cancelReservation(Train train) {
		if (nearestTrain.get() == train)
			nearestTrain = new WeakReference<>(null);
	}

	public void trainDeparted(Train train) {
		cancelReservation(train);
	}

	@Nullable
	public Train getPresentTrain() {
		Train nearestTrain = getNearestTrain();
		if (nearestTrain == null || nearestTrain.getCurrentStation() != this)
			return null;
		return nearestTrain;
	}

	@Nullable
	public Train getImminentTrain() {
		Train nearestTrain = getNearestTrain();
		if (nearestTrain == null)
			return nearestTrain;
		if (nearestTrain.getCurrentStation() == this)
			return nearestTrain;
		if (!nearestTrain.navigation.isActive())
			return null;
		if (nearestTrain.navigation.distanceToDestination > 30)
			return null;
		return nearestTrain;
	}

	@Nullable
	public Train getNearestTrain() {
		return this.nearestTrain.get();
	}

}
