package com.simibubi.create.content.trains.station;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Objects;
import java.util.UUID;
import java.util.function.Consumer;

import javax.annotation.Nullable;

import com.simibubi.create.content.trains.graph.DiscoveredPath;

import com.simibubi.create.AllBlocks;
import com.simibubi.create.AllItems;
import com.simibubi.create.AllPackets;
import com.simibubi.create.AllSoundEvents;
import com.simibubi.create.Create;
import com.simibubi.create.compat.computercraft.AbstractComputerBehaviour;
import com.simibubi.create.compat.computercraft.ComputerCraftProxy;
import com.simibubi.create.content.contraptions.AssemblyException;
import com.simibubi.create.content.contraptions.ITransformableBlockEntity;
import com.simibubi.create.content.contraptions.StructureTransform;
import com.simibubi.create.content.decoration.slidingDoor.DoorControlBehaviour;
import com.simibubi.create.content.logistics.depot.DepotBehaviour;
import com.simibubi.create.content.redstone.displayLink.DisplayLinkBlock;
import com.simibubi.create.content.trains.bogey.AbstractBogeyBlock;
import com.simibubi.create.content.trains.bogey.AbstractBogeyBlockEntity;
import com.simibubi.create.content.trains.entity.Carriage;
import com.simibubi.create.content.trains.entity.CarriageBogey;
import com.simibubi.create.content.trains.entity.CarriageContraption;
import com.simibubi.create.content.trains.entity.Train;
import com.simibubi.create.content.trains.entity.TrainPacket;
import com.simibubi.create.content.trains.entity.TravellingPoint;
import com.simibubi.create.content.trains.graph.EdgePointType;
import com.simibubi.create.content.trains.graph.TrackEdge;
import com.simibubi.create.content.trains.graph.TrackGraph;
import com.simibubi.create.content.trains.graph.TrackGraphLocation;
import com.simibubi.create.content.trains.graph.TrackNode;
import com.simibubi.create.content.trains.graph.TrackNodeLocation;
import com.simibubi.create.content.trains.graph.TrackNodeLocation.DiscoveredLocation;
import com.simibubi.create.content.trains.schedule.Schedule;
import com.simibubi.create.content.trains.schedule.ScheduleItem;
import com.simibubi.create.content.trains.track.ITrackBlock;
import com.simibubi.create.content.trains.track.TrackTargetingBehaviour;
import com.simibubi.create.foundation.advancement.AllAdvancements;
import com.simibubi.create.foundation.block.ProperWaterloggedBlock;
import com.simibubi.create.foundation.blockEntity.SmartBlockEntity;
import com.simibubi.create.foundation.blockEntity.behaviour.BlockEntityBehaviour;
import com.simibubi.create.foundation.utility.Iterate;
import com.simibubi.create.foundation.utility.Lang;
import com.simibubi.create.foundation.utility.NBTHelper;
import com.simibubi.create.foundation.utility.VecHelper;
import com.simibubi.create.foundation.utility.WorldAttached;
import com.simibubi.create.foundation.utility.animation.LerpedFloat;
import com.simibubi.create.foundation.utility.animation.LerpedFloat.Chaser;
import com.simibubi.create.infrastructure.config.AllConfigs;

import net.fabricmc.api.EnvType;
import net.fabricmc.api.Environment;
import net.fabricmc.fabric.api.transfer.v1.item.ItemVariant;
import net.fabricmc.fabric.api.transfer.v1.storage.Storage;
import net.fabricmc.fabric.api.transfer.v1.storage.base.SidedStorageBlockEntity;
import net.minecraft.class_124;
import net.minecraft.class_1268;
import net.minecraft.class_1542;
import net.minecraft.class_1657;
import net.minecraft.class_1799;
import net.minecraft.class_2338;
import net.minecraft.class_2338.class_2339;
import net.minecraft.class_2350;
import net.minecraft.class_2350.class_2351;
import net.minecraft.class_2350.class_2352;
import net.minecraft.class_238;
import net.minecraft.class_2398;
import net.minecraft.class_243;
import net.minecraft.class_2487;
import net.minecraft.class_2498;
import net.minecraft.class_2561;
import net.minecraft.class_2586;
import net.minecraft.class_2591;
import net.minecraft.class_2680;
import net.minecraft.class_3218;
import net.minecraft.class_3222;
import net.minecraft.class_3341;
import net.minecraft.class_3419;
import net.minecraft.class_3532;

public class StationBlockEntity extends SmartBlockEntity implements ITransformableBlockEntity, SidedStorageBlockEntity {

	public TrackTargetingBehaviour<GlobalStation> edgePoint;
	public DoorControlBehaviour doorControls;
	public LerpedFloat flag;

	protected int failedCarriageIndex;
	protected AssemblyException lastException;
	protected DepotBehaviour depotBehaviour;
	public AbstractComputerBehaviour computerBehaviour;

	// for display
	UUID imminentTrain;
	boolean trainPresent;
	boolean trainBackwards;
	boolean trainCanDisassemble;
	boolean trainHasSchedule;
	boolean trainHasAutoSchedule;

	int flagYRot = -1;
	boolean flagFlipped;

	public class_2561 lastDisassembledTrainName;

	public StationBlockEntity(class_2591<?> type, class_2338 pos, class_2680 state) {
		super(type, pos, state);
		setLazyTickRate(20);
		lastException = null;
		failedCarriageIndex = -1;
		flag = LerpedFloat.linear()
			.startWithValue(0);
	}

	@Override
	public void addBehaviours(List<BlockEntityBehaviour> behaviours) {
		behaviours.add(edgePoint = new TrackTargetingBehaviour<>(this, EdgePointType.STATION));
		behaviours.add(doorControls = new DoorControlBehaviour(this));
		behaviours.add(depotBehaviour = new DepotBehaviour(this).onlyAccepts(AllItems.SCHEDULE::isIn)
			.withCallback(s -> applyAutoSchedule()));
		depotBehaviour.addSubBehaviours(behaviours);
		registerAwardables(behaviours, AllAdvancements.CONTRAPTION_ACTORS, AllAdvancements.TRAIN,
			AllAdvancements.LONG_TRAIN, AllAdvancements.CONDUCTOR);
		behaviours.add(computerBehaviour = ComputerCraftProxy.behaviour(this));
	}

	@Override
	protected void read(class_2487 tag, boolean clientPacket) {
		lastException = AssemblyException.read(tag);
		failedCarriageIndex = tag.method_10550("FailedCarriageIndex");
		super.read(tag, clientPacket);
		invalidateRenderBoundingBox();

		if (tag.method_10545("ForceFlag"))
			trainPresent = tag.method_10577("ForceFlag");
		if (tag.method_10545("PrevTrainName"))
			lastDisassembledTrainName = class_2561.class_2562.method_10877(tag.method_10558("PrevTrainName"));

		if (!clientPacket)
			return;
		if (!tag.method_10545("ImminentTrain")) {
			imminentTrain = null;
			trainPresent = false;
			trainCanDisassemble = false;
			trainBackwards = false;
			return;
		}

		imminentTrain = tag.method_25926("ImminentTrain");
		trainPresent = tag.method_10545("TrainPresent");
		trainCanDisassemble = tag.method_10545("TrainCanDisassemble");
		trainBackwards = tag.method_10545("TrainBackwards");
		trainHasSchedule = tag.method_10545("TrainHasSchedule");
		trainHasAutoSchedule = tag.method_10545("TrainHasAutoSchedule");
	}

	@Override
	protected void write(class_2487 tag, boolean clientPacket) {
		AssemblyException.write(tag, lastException);
		tag.method_10569("FailedCarriageIndex", failedCarriageIndex);

		if (lastDisassembledTrainName != null)
			tag.method_10582("PrevTrainName", class_2561.class_2562.method_10867(lastDisassembledTrainName));

		super.write(tag, clientPacket);

		if (!clientPacket)
			return;
		if (imminentTrain == null)
			return;

		tag.method_25927("ImminentTrain", imminentTrain);

		if (trainPresent)
			NBTHelper.putMarker(tag, "TrainPresent");
		if (trainCanDisassemble)
			NBTHelper.putMarker(tag, "TrainCanDisassemble");
		if (trainBackwards)
			NBTHelper.putMarker(tag, "TrainBackwards");
		if (trainHasSchedule)
			NBTHelper.putMarker(tag, "TrainHasSchedule");
		if (trainHasAutoSchedule)
			NBTHelper.putMarker(tag, "TrainHasAutoSchedule");
	}

	@Nullable
	public GlobalStation getStation() {
		return edgePoint.getEdgePoint();
	}

	// Train Assembly

	public static WorldAttached<Map<class_2338, class_3341>> assemblyAreas = new WorldAttached<>(w -> new HashMap<>());

	class_2350 assemblyDirection;
	int assemblyLength;
	int[] bogeyLocations;
	AbstractBogeyBlock<?>[] bogeyTypes;
	boolean[] upsideDownBogeys;
	int bogeyCount;

	@Override
	public void lazyTick() {
		if (isAssembling() && !field_11863.field_9236)
			refreshAssemblyInfo();
		super.lazyTick();
	}

	@Override
	public void tick() {
		if (isAssembling() && field_11863.field_9236)
			refreshAssemblyInfo();
		super.tick();

		if (field_11863.field_9236) {
			float currentTarget = flag.getChaseTarget();
			if (currentTarget == 0 || flag.settled()) {
				int target = trainPresent || isAssembling() ? 1 : 0;
				if (target != currentTarget) {
					flag.chase(target, 0.1f, Chaser.LINEAR);
					if (target == 1)
						AllSoundEvents.CONTRAPTION_ASSEMBLE.playAt(field_11863, field_11867, 1, 2, true);
				}
			}
			boolean settled = flag.getValue() > .15f;
			flag.tickChaser();
			if (currentTarget == 0 && settled != flag.getValue() > .15f)
				AllSoundEvents.CONTRAPTION_DISASSEMBLE.playAt(field_11863, field_11867, 0.75f, 1.5f, true);
			return;
		}

		GlobalStation station = getStation();
		if (station == null)
			return;

		Train imminentTrain = station.getImminentTrain();
		boolean trainPresent = imminentTrain != null && imminentTrain.getCurrentStation() == station;
		boolean canDisassemble = trainPresent && imminentTrain.canDisassemble();
		UUID imminentID = imminentTrain != null ? imminentTrain.id : null;
		boolean trainHasSchedule = trainPresent && imminentTrain.runtime.getSchedule() != null;
		boolean trainHasAutoSchedule = trainHasSchedule && imminentTrain.runtime.isAutoSchedule;
		boolean newlyArrived = this.trainPresent != trainPresent;

		if (trainPresent && imminentTrain.runtime.displayLinkUpdateRequested) {
			DisplayLinkBlock.notifyGatherers(field_11863, field_11867);
			imminentTrain.runtime.displayLinkUpdateRequested = false;
		}

		if (newlyArrived)
			applyAutoSchedule();

		if (newlyArrived || this.trainCanDisassemble != canDisassemble
			|| !Objects.equals(imminentID, this.imminentTrain) || this.trainHasSchedule != trainHasSchedule
			|| this.trainHasAutoSchedule != trainHasAutoSchedule) {

			this.imminentTrain = imminentID;
			this.trainPresent = trainPresent;
			this.trainCanDisassemble = canDisassemble;
			this.trainBackwards = imminentTrain != null && imminentTrain.currentlyBackwards;
			this.trainHasSchedule = trainHasSchedule;
			this.trainHasAutoSchedule = trainHasAutoSchedule;

			notifyUpdate();
		}
	}

	public boolean trackClicked(class_1657 player, class_1268 hand, ITrackBlock track, class_2680 state,
		class_2338 pos) {
		refreshAssemblyInfo();
		class_3341 bb = assemblyAreas.get(field_11863)
			.get(field_11867);
		if (bb == null || !bb.method_14662(pos))
			return false;

		class_2338 up = class_2338.method_49638(track.getUpNormal(field_11863, pos, state));
		class_2338 down = class_2338.method_49638(track.getUpNormal(field_11863, pos, state).method_1021(-1));
		int bogeyOffset = pos.method_19455(edgePoint.getGlobalPosition()) - 1;

		if (!isValidBogeyOffset(bogeyOffset)) {
			for (boolean upsideDown : Iterate.falseAndTrue) {
				for (int i = -1; i <= 1; i++) {
					class_2338 bogeyPos = pos.method_10079(assemblyDirection, i)
						.method_10081(upsideDown ? down : up);
					class_2680 blockState = field_11863.method_8320(bogeyPos);
					if (!(blockState.method_26204() instanceof AbstractBogeyBlock<?> bogey))
						continue;
					class_2586 be = field_11863.method_8321(bogeyPos);
					if (!(be instanceof AbstractBogeyBlockEntity oldBE))
						continue;
					class_2487 oldData = oldBE.getBogeyData();
					class_2680 newBlock = bogey.getNextSize(oldBE);
					if (newBlock.method_26204() == bogey)
						player.method_7353(Lang.translateDirect("bogey.style.no_other_sizes")
							.method_27692(class_124.field_1061), true);
					field_11863.method_8652(bogeyPos, newBlock, 3);
					class_2586 newEntity = field_11863.method_8321(bogeyPos);
					if (!(newEntity instanceof AbstractBogeyBlockEntity newBE))
						continue;
					newBE.setBogeyData(oldData);
					bogey.playRotateSound(field_11863, bogeyPos);
					return true;
				}
			}

			return false;
		}

		class_1799 handItem = player.method_5998(hand);
		if (!player.method_7337() && !AllBlocks.RAILWAY_CASING.isIn(handItem)) {
			player.method_7353(Lang.translateDirect("train_assembly.requires_casing"), true);
			return false;
		}

		boolean upsideDown = (player.method_5695(1.0F) < 0 && (track.getBogeyAnchor(field_11863, pos, state)).method_26204() instanceof AbstractBogeyBlock<?> bogey && bogey.canBeUpsideDown());

		class_2338 targetPos = upsideDown ? pos.method_10081(down) : pos.method_10081(up);
		if (field_11863.method_8320(targetPos)
			.method_26214(field_11863, targetPos) == -1) {
			return false;
		}

		field_11863.method_22352(targetPos, true);

		class_2680 bogeyAnchor = track.getBogeyAnchor(field_11863, pos, state);
		if (bogeyAnchor.method_26204() instanceof AbstractBogeyBlock<?> bogey) {
			bogeyAnchor = bogey.getVersion(bogeyAnchor, upsideDown);
		}
		bogeyAnchor = ProperWaterloggedBlock.withWater(field_11863, bogeyAnchor, pos);
		field_11863.method_8652(targetPos, bogeyAnchor, 3);
		player.method_7353(Lang.translateDirect("train_assembly.bogey_created"), true);
		class_2498 soundtype = bogeyAnchor.method_26204()
			.method_9573(state);
		field_11863.method_8396(null, pos, soundtype.method_10598(), class_3419.field_15245, (soundtype.method_10597() + 1.0F) / 2.0F,
			soundtype.method_10599() * 0.8F);

		if (!player.method_7337()) {
			class_1799 itemInHand = player.method_5998(hand);
			itemInHand.method_7934(1);
			if (itemInHand.method_7960())
				player.method_6122(hand, class_1799.field_8037);
		}

		return true;
	}

	public boolean enterAssemblyMode(@Nullable class_3222 sender) {
		if (isAssembling())
			return false;

		tryDisassembleTrain(sender);
		if (!tryEnterAssemblyMode())
			return false;

		class_2680 newState = method_11010().method_11657(StationBlock.ASSEMBLING, true);
		field_11863.method_8652(method_11016(), newState, 3);
		refreshBlockState();
		refreshAssemblyInfo();

		updateStationState(station -> station.assembling = true);
		GlobalStation station = getStation();
		if (station != null) {
			for (Train train : Create.RAILWAYS.sided(field_11863).trains.values()) {
				if (train.navigation.destination != station)
					continue;

				DiscoveredPath preferredPath = train.runtime.startCurrentInstruction();
				train.navigation.startNavigation(preferredPath != null ? preferredPath : train.navigation.findPathTo(station, Double.MAX_VALUE));
			}
		}

		return true;
	}

	public boolean exitAssemblyMode() {
		if (!isAssembling())
			return false;

		cancelAssembly();
		class_2680 newState = method_11010().method_11657(StationBlock.ASSEMBLING, false);
		field_11863.method_8652(method_11016(), newState, 3);
		refreshBlockState();

		return updateStationState(station -> station.assembling = false);
	}

	public boolean tryDisassembleTrain(@Nullable class_3222 sender) {
		GlobalStation station = getStation();
		if (station == null)
			return false;

		Train train = station.getPresentTrain();
		if (train == null)
			return false;

		class_2338 trackPosition = edgePoint.getGlobalPosition();
		if (!train.disassemble(getAssemblyDirection(), trackPosition.method_10084()))
			return false;

		dropSchedule(sender);
		return true;
	}

	public boolean isAssembling() {
		class_2680 state = method_11010();
		return state.method_28498(StationBlock.ASSEMBLING) && state.method_11654(StationBlock.ASSEMBLING);
	}

	public boolean tryEnterAssemblyMode() {
		if (!edgePoint.hasValidTrack())
			return false;

		class_2338 targetPosition = edgePoint.getGlobalPosition();
		class_2680 trackState = edgePoint.getTrackBlockState();
		ITrackBlock track = edgePoint.getTrack();
		class_243 trackAxis = track.getTrackAxes(field_11863, targetPosition, trackState)
			.get(0);

		boolean axisFound = false;
		for (class_2351 axis : Iterate.axes) {
			if (trackAxis.method_18043(axis) == 0)
				continue;
			if (axisFound)
				return false;
			axisFound = true;
		}

		return true;
	}

	public void dropSchedule(@Nullable class_3222 sender) {
		GlobalStation station = getStation();
		if (station == null)
			return;

		Train train = station.getPresentTrain();
		if (train == null)
			return;

		class_1799 schedule = train.runtime.returnSchedule();
		if (schedule.method_7960())
			return;
		if (sender != null && sender.method_6047().method_7960()) {
			sender.method_31548()
					.method_7398(schedule);
			return;
		}

		class_243 v = VecHelper.getCenterOf(method_11016());
		class_1542 itemEntity = new class_1542(method_10997(), v.field_1352, v.field_1351, v.field_1350, schedule);
		itemEntity.method_18799(class_243.field_1353);
		method_10997().method_8649(itemEntity);
	}

	private boolean updateStationState(Consumer<GlobalStation> updateState) {
		GlobalStation station = getStation();
		TrackGraphLocation graphLocation = edgePoint.determineGraphLocation();
		if (station == null || graphLocation == null)
			return false;

		updateState.accept(station);
		Create.RAILWAYS.sync.pointAdded(graphLocation.graph, station);
		Create.RAILWAYS.markTracksDirty();
		return true;
	}

	public void refreshAssemblyInfo() {
		if (!edgePoint.hasValidTrack())
			return;

		if (!isVirtual()) {
			GlobalStation station = getStation();
			if (station == null || station.getPresentTrain() != null)
				return;
		}

		int prevLength = assemblyLength;
		class_2338 targetPosition = edgePoint.getGlobalPosition();
		class_2680 trackState = edgePoint.getTrackBlockState();
		ITrackBlock track = edgePoint.getTrack();
		getAssemblyDirection();

		class_2339 currentPos = targetPosition.method_25503();
		currentPos.method_10098(assemblyDirection);

		class_2338 bogeyOffset = class_2338.method_49638(track.getUpNormal(field_11863, targetPosition, trackState));

		int MAX_LENGTH = AllConfigs.server().trains.maxAssemblyLength.get();
		int MAX_BOGEY_COUNT = AllConfigs.server().trains.maxBogeyCount.get();

		int bogeyIndex = 0;
		int maxBogeyCount = MAX_BOGEY_COUNT;
		if (bogeyLocations == null)
			bogeyLocations = new int[maxBogeyCount];
		if (bogeyTypes == null)
			bogeyTypes = new AbstractBogeyBlock[maxBogeyCount];
		if (upsideDownBogeys == null)
			upsideDownBogeys = new boolean[maxBogeyCount];
		Arrays.fill(bogeyLocations, -1);
		Arrays.fill(bogeyTypes, null);
		Arrays.fill(upsideDownBogeys, false);

		for (int i = 0; i < MAX_LENGTH; i++) {
			if (i == MAX_LENGTH - 1) {
				assemblyLength = i;
				break;
			}
			if (!track.trackEquals(trackState, field_11863.method_8320(currentPos))) {
				assemblyLength = Math.max(0, i - 1);
				break;
			}

			class_2680 potentialBogeyState = field_11863.method_8320(bogeyOffset.method_10081(currentPos));
			class_2338 upsideDownBogeyOffset = new class_2338(bogeyOffset.method_10263(), bogeyOffset.method_10264()*-1, bogeyOffset.method_10260());
			if (bogeyIndex < bogeyLocations.length) {
				if (potentialBogeyState.method_26204() instanceof AbstractBogeyBlock<?> bogey && !bogey.isUpsideDown(potentialBogeyState)) {
					bogeyTypes[bogeyIndex] = bogey;
					bogeyLocations[bogeyIndex] = i;
					upsideDownBogeys[bogeyIndex] = false;
					bogeyIndex++;
				} else if ((potentialBogeyState = field_11863.method_8320(upsideDownBogeyOffset.method_10081(currentPos))).method_26204() instanceof AbstractBogeyBlock<?> bogey && bogey.isUpsideDown(potentialBogeyState)) {
					bogeyTypes[bogeyIndex] = bogey;
					bogeyLocations[bogeyIndex] = i;
					upsideDownBogeys[bogeyIndex] = true;
					bogeyIndex++;
				}
			}

			currentPos.method_10098(assemblyDirection);
		}

		bogeyCount = bogeyIndex;

		if (field_11863.field_9236)
			return;
		if (prevLength == assemblyLength)
			return;
		if (isVirtual())
			return;

		Map<class_2338, class_3341> map = assemblyAreas.get(field_11863);
		class_2338 startPosition = targetPosition.method_10093(assemblyDirection);
		class_2338 trackEnd = startPosition.method_10079(assemblyDirection, assemblyLength - 1);
		map.put(field_11867, class_3341.method_34390(startPosition, trackEnd));
	}

	public boolean updateName(String name) {
		if (!updateStationState(station -> station.name = name))
			return false;
		notifyUpdate();

		return true;
	}

	public boolean isValidBogeyOffset(int i) {
		if ((i < 3 || bogeyCount == 0) && i != 0)
			return false;
		for (int j : bogeyLocations) {
			if (j == -1)
				break;
			if (i >= j - 2 && i <= j + 2)
				return false;
		}
		return true;
	}

	public class_2350 getAssemblyDirection() {
		if (assemblyDirection != null)
			return assemblyDirection;
		if (!edgePoint.hasValidTrack())
			return null;
		class_2338 targetPosition = edgePoint.getGlobalPosition();
		class_2680 trackState = edgePoint.getTrackBlockState();
		ITrackBlock track = edgePoint.getTrack();
		class_2352 axisDirection = edgePoint.getTargetDirection();
		class_243 axis = track.getTrackAxes(field_11863, targetPosition, trackState)
			.get(0)
			.method_1029()
			.method_1021(axisDirection.method_10181());
		return assemblyDirection = class_2350.method_10142(axis.field_1352, axis.field_1351, axis.field_1350);
	}

	@Override
	public void remove() {
		assemblyAreas.get(field_11863)
			.remove(field_11867);
		super.remove();
	}

	public void assemble(UUID playerUUID) {
		refreshAssemblyInfo();

		if (bogeyLocations == null)
			return;

		if (bogeyLocations[0] != 0) {
			exception(new AssemblyException(Lang.translateDirect("train_assembly.frontmost_bogey_at_station")), -1);
			return;
		}

		if (!edgePoint.hasValidTrack())
			return;

		class_2338 trackPosition = edgePoint.getGlobalPosition();
		class_2680 trackState = edgePoint.getTrackBlockState();
		ITrackBlock track = edgePoint.getTrack();
		class_2338 bogeyOffset = class_2338.method_49638(track.getUpNormal(field_11863, trackPosition, trackState));

		TrackNodeLocation location = null;
		class_243 centre = class_243.method_24955(trackPosition)
			.method_1031(0, track.getElevationAtCenter(field_11863, trackPosition, trackState), 0);
		Collection<DiscoveredLocation> ends = track.getConnected(field_11863, trackPosition, trackState, true, null);
		class_243 targetOffset = class_243.method_24954(assemblyDirection.method_10163());
		for (DiscoveredLocation end : ends)
			if (class_3532.method_20390(0, targetOffset.method_1025(end.getLocation()
				.method_1020(centre)
				.method_1029())))
				location = end;
		if (location == null)
			return;

		List<Double> pointOffsets = new ArrayList<>();
		int iPrevious = -100;
		for (int i = 0; i < bogeyLocations.length; i++) {
			int loc = bogeyLocations[i];
			if (loc == -1)
				break;

			if (loc - iPrevious < 3) {
				exception(new AssemblyException(Lang.translateDirect("train_assembly.bogeys_too_close", i, i + 1)), -1);
				return;
			}

			double bogeySize = bogeyTypes[i].getWheelPointSpacing();
			pointOffsets.add(Double.valueOf(loc + .5 - bogeySize / 2));
			pointOffsets.add(Double.valueOf(loc + .5 + bogeySize / 2));
			iPrevious = loc;
		}

		List<TravellingPoint> points = new ArrayList<>();
		class_243 directionVec = class_243.method_24954(assemblyDirection.method_10163());
		TrackGraph graph = null;
		TrackNode secondNode = null;

		for (int j = 0; j < assemblyLength * 2 + 40; j++) {
			double i = j / 2d;
			if (points.size() == pointOffsets.size())
				break;

			TrackNodeLocation currentLocation = location;
			location = new TrackNodeLocation(location.getLocation()
				.method_1019(directionVec.method_1021(.5))).in(location.dimension);

			if (graph == null)
				graph = Create.RAILWAYS.getGraph(field_11863, currentLocation);
			if (graph == null)
				continue;
			TrackNode node = graph.locateNode(currentLocation);
			if (node == null)
				continue;

			for (int pointIndex = points.size(); pointIndex < pointOffsets.size(); pointIndex++) {
				double offset = pointOffsets.get(pointIndex);
				if (offset > i)
					break;
				double positionOnEdge = i - offset;

				Map<TrackNode, TrackEdge> connectionsFromNode = graph.getConnectionsFrom(node);

				if (secondNode == null)
					for (Entry<TrackNode, TrackEdge> entry : connectionsFromNode.entrySet()) {
						TrackEdge edge = entry.getValue();
						TrackNode otherNode = entry.getKey();
						if (edge.isTurn())
							continue;
						class_243 edgeDirection = edge.getDirection(true);
						if (class_3532.method_20390(edgeDirection.method_1029()
							.method_1026(directionVec), -1d))
							secondNode = otherNode;
					}

				if (secondNode == null) {
					Create.LOGGER.warn("Cannot assemble: No valid starting node found");
					return;
				}

				TrackEdge edge = connectionsFromNode.get(secondNode);

				if (edge == null) {
					Create.LOGGER.warn("Cannot assemble: Missing graph edge");
					return;
				}

				points.add(new TravellingPoint(node, secondNode, edge, positionOnEdge, false));
			}

			secondNode = node;
		}

		if (points.size() != pointOffsets.size()) {
			Create.LOGGER.warn("Cannot assemble: Not all Points created");
			return;
		}

		if (points.size() == 0) {
			exception(new AssemblyException(Lang.translateDirect("train_assembly.no_bogeys")), -1);
			return;
		}

		List<CarriageContraption> contraptions = new ArrayList<>();
		List<Carriage> carriages = new ArrayList<>();
		List<Integer> spacing = new ArrayList<>();
		boolean atLeastOneForwardControls = false;

		for (int bogeyIndex = 0; bogeyIndex < bogeyCount; bogeyIndex++) {
			int pointIndex = bogeyIndex * 2;
			if (bogeyIndex > 0)
				spacing.add(bogeyLocations[bogeyIndex] - bogeyLocations[bogeyIndex - 1]);
			CarriageContraption contraption = new CarriageContraption(assemblyDirection);
			class_2338 bogeyPosOffset = trackPosition.method_10081(bogeyOffset);
			class_2338 upsideDownBogeyPosOffset = trackPosition.method_10081(new class_2338(bogeyOffset.method_10263(), bogeyOffset.method_10264() * -1, bogeyOffset.method_10260()));

			try {
				int offset = bogeyLocations[bogeyIndex] + 1;
				boolean success = contraption.assemble(field_11863, upsideDownBogeys[bogeyIndex] ? upsideDownBogeyPosOffset.method_10079(assemblyDirection, offset) : bogeyPosOffset.method_10079(assemblyDirection, offset));
				atLeastOneForwardControls |= contraption.hasForwardControls();
				contraption.setSoundQueueOffset(offset);
				if (!success) {
					exception(new AssemblyException(Lang.translateDirect("train_assembly.nothing_attached", bogeyIndex + 1)),
						-1);
					return;
				}
			} catch (AssemblyException e) {
				exception(e, contraptions.size() + 1);
				return;
			}

			AbstractBogeyBlock<?> typeOfFirstBogey = bogeyTypes[bogeyIndex];
			boolean firstBogeyIsUpsideDown = upsideDownBogeys[bogeyIndex];
			class_2338 firstBogeyPos = contraption.anchor;
			AbstractBogeyBlockEntity firstBogeyBlockEntity = (AbstractBogeyBlockEntity) field_11863.method_8321(firstBogeyPos);
			CarriageBogey firstBogey =
				new CarriageBogey(typeOfFirstBogey, firstBogeyIsUpsideDown, firstBogeyBlockEntity.getBogeyData(), points.get(pointIndex), points.get(pointIndex + 1));
			CarriageBogey secondBogey = null;
			class_2338 secondBogeyPos = contraption.getSecondBogeyPos();
			int bogeySpacing = 0;

			if (secondBogeyPos != null) {
				if (bogeyIndex == bogeyCount - 1 || !secondBogeyPos
					.equals((upsideDownBogeys[bogeyIndex + 1] ? upsideDownBogeyPosOffset : bogeyPosOffset).method_10079(assemblyDirection, bogeyLocations[bogeyIndex + 1] + 1))) {
					exception(new AssemblyException(Lang.translateDirect("train_assembly.not_connected_in_order")),
						contraptions.size() + 1);
					return;
				}
				AbstractBogeyBlockEntity secondBogeyBlockEntity =
						(AbstractBogeyBlockEntity) field_11863.method_8321(secondBogeyPos);
				bogeySpacing = bogeyLocations[bogeyIndex + 1] - bogeyLocations[bogeyIndex];
				secondBogey = new CarriageBogey(bogeyTypes[bogeyIndex + 1], upsideDownBogeys[bogeyIndex + 1], secondBogeyBlockEntity.getBogeyData(),
						points.get(pointIndex + 2), points.get(pointIndex + 3));
				bogeyIndex++;

			} else if (!typeOfFirstBogey.allowsSingleBogeyCarriage()) {
				exception(new AssemblyException(Lang.translateDirect("train_assembly.single_bogey_carriage")),
					contraptions.size() + 1);
				return;
			}

			contraptions.add(contraption);
			carriages.add(new Carriage(firstBogey, secondBogey, bogeySpacing));
		}

		if (!atLeastOneForwardControls) {
			exception(new AssemblyException(Lang.translateDirect("train_assembly.no_controls")), -1);
			return;
		}

		for (CarriageContraption contraption : contraptions) {
			contraption.removeBlocksFromWorld(field_11863, class_2338.field_10980);
			contraption.expandBoundsAroundAxis(class_2351.field_11052);
		}

		Train train = new Train(UUID.randomUUID(), playerUUID, graph, carriages, spacing, contraptions.stream()
			.anyMatch(CarriageContraption::hasBackwardControls));

		if (lastDisassembledTrainName != null) {
			train.name = lastDisassembledTrainName;
			lastDisassembledTrainName = null;
		}

		for (int i = 0; i < contraptions.size(); i++) {
			CarriageContraption contraption = contraptions.get(i);
			Carriage carriage = carriages.get(i);
			carriage.setContraption(field_11863, contraption);
			if (contraption.containsBlockBreakers())
				award(AllAdvancements.CONTRAPTION_ACTORS);
		}

		GlobalStation station = getStation();
		if (station != null) {
			train.setCurrentStation(station);
			station.reserveFor(train);
		}

		train.collectInitiallyOccupiedSignalBlocks();
		Create.RAILWAYS.addTrain(train);
		AllPackets.getChannel().sendToClientsInServer(new TrainPacket(train, true), field_11863.method_8503());
		clearException();

		award(AllAdvancements.TRAIN);
		if (contraptions.size() >= 6)
			award(AllAdvancements.LONG_TRAIN);
	}

	public void cancelAssembly() {
		assemblyLength = 0;
		assemblyAreas.get(field_11863)
			.remove(field_11867);
		clearException();
	}

	private void clearException() {
		exception(null, -1);
	}

	private void exception(AssemblyException exception, int carriage) {
		failedCarriageIndex = carriage;
		lastException = exception;
		sendData();
	}

	@Override
	@Environment(EnvType.CLIENT)
	public class_238 getRenderBoundingBox() {
		if (isAssembling())
			return INFINITE_EXTENT_AABB;
		return super.getRenderBoundingBox();
	}

	@Override
	protected class_238 createRenderBoundingBox() {
		return new class_238(field_11867, edgePoint.getGlobalPosition()).method_1014(2);
	}

	public class_1799 getAutoSchedule() {
		return depotBehaviour.getHeldItemStack();
	}

	@Nullable
	@Override
	public Storage<ItemVariant> getItemStorage(@Nullable class_2350 face) {
		return depotBehaviour.itemHandler;
	}

	private void applyAutoSchedule() {
		class_1799 stack = getAutoSchedule();
		if (!AllItems.SCHEDULE.isIn(stack))
			return;
		Schedule schedule = ScheduleItem.getSchedule(stack);
		if (schedule == null || schedule.entries.isEmpty())
			return;
		GlobalStation station = getStation();
		if (station == null)
			return;
		Train imminentTrain = station.getImminentTrain();
		if (imminentTrain == null || imminentTrain.getCurrentStation() != station)
			return;

		award(AllAdvancements.CONDUCTOR);
		imminentTrain.runtime.setSchedule(schedule, true);
		AllSoundEvents.CONFIRM.playOnServer(field_11863, field_11867, 1, 1);

		if (!(field_11863 instanceof class_3218 server))
			return;

		class_243 v = class_243.method_24955(field_11867.method_10084());
		server.method_14199(class_2398.field_11211, v.field_1352, v.field_1351, v.field_1350, 8, 0.35, 0.05, 0.35, 1);
		server.method_14199(class_2398.field_11207, v.field_1352, v.field_1351 + .25f, v.field_1350, 10, 0.05, 1, 0.05, 0.005f);
	}

	public boolean resolveFlagAngle() {
		if (flagYRot != -1)
			return true;

		class_2680 target = edgePoint.getTrackBlockState();
		if (!(target.method_26204() instanceof ITrackBlock def))
			return false;

		class_243 axis = null;
		class_2338 trackPos = edgePoint.getGlobalPosition();
		for (class_243 vec3 : def.getTrackAxes(field_11863, trackPos, target))
			axis = vec3.method_1021(edgePoint.getTargetDirection()
				.method_10181());
		if (axis == null)
			return false;

		class_2350 nearest = class_2350.method_10142(axis.field_1352, 0, axis.field_1350);
		flagYRot = (int) (-nearest.method_10144() - 90);

		class_243 diff = class_243.method_24954(trackPos.method_10059(field_11867))
			.method_18805(1, 0, 1);
		if (diff.method_1027() == 0)
			return true;

		flagFlipped = diff.method_1026(class_243.method_24954(nearest.method_10170()
			.method_10163())) > 0;

		return true;
	}

	@Override
	public void transform(StructureTransform transform) {
		edgePoint.transform(transform);
	}

}
