package com.simibubi.create.content.trains.station;

import com.simibubi.create.content.decoration.slidingDoor.DoorControl;
import com.simibubi.create.foundation.networking.BlockEntityConfigurationPacket;
import net.minecraft.class_1937;
import net.minecraft.class_2338;
import net.minecraft.class_2540;
import net.minecraft.class_2680;
import net.minecraft.class_3222;
import net.minecraft.class_3532;

public class StationEditPacket extends BlockEntityConfigurationPacket<StationBlockEntity> {

	boolean dropSchedule;
	boolean assemblyMode;
	Boolean tryAssemble;
	DoorControl doorControl;
	String name;

	public static StationEditPacket dropSchedule(class_2338 pos) {
		StationEditPacket packet = new StationEditPacket(pos);
		packet.dropSchedule = true;
		return packet;
	}

	public static StationEditPacket tryAssemble(class_2338 pos) {
		StationEditPacket packet = new StationEditPacket(pos);
		packet.tryAssemble = true;
		return packet;
	}

	public static StationEditPacket tryDisassemble(class_2338 pos) {
		StationEditPacket packet = new StationEditPacket(pos);
		packet.tryAssemble = false;
		return packet;
	}

	public static StationEditPacket configure(class_2338 pos, boolean assemble, String name, DoorControl doorControl) {
		StationEditPacket packet = new StationEditPacket(pos);
		packet.assemblyMode = assemble;
		packet.tryAssemble = null;
		packet.name = name;
		packet.doorControl = doorControl;
		return packet;
	}

	public StationEditPacket(class_2540 buffer) {
		super(buffer);
	}

	public StationEditPacket(class_2338 pos) {
		super(pos);
	}

	@Override
	protected void writeSettings(class_2540 buffer) {
		buffer.writeBoolean(dropSchedule);
		if (dropSchedule)
			return;
		buffer.writeBoolean(doorControl != null);
		if (doorControl != null)
			buffer.method_10804(doorControl.ordinal());
		buffer.writeBoolean(tryAssemble != null);
		if (tryAssemble != null) {
			buffer.writeBoolean(tryAssemble);
			return;
		}
		buffer.writeBoolean(assemblyMode);
		buffer.method_10814(name);
	}

	@Override
	protected void readSettings(class_2540 buffer) {
		if (buffer.readBoolean()) {
			dropSchedule = true;
			return;
		}
		if (buffer.readBoolean())
			doorControl = DoorControl.values()[class_3532.method_15340(buffer.method_10816(), 0, DoorControl.values().length)];
		name = "";
		if (buffer.readBoolean()) {
			tryAssemble = buffer.readBoolean();
			return;
		}
		assemblyMode = buffer.readBoolean();
		name = buffer.method_10800(256);
	}

	@Override
	protected void applySettings(class_3222 player, StationBlockEntity be) {
		class_1937 level = be.method_10997();
		class_2338 blockPos = be.method_11016();
		class_2680 blockState = level.method_8320(blockPos);

		if (dropSchedule) {
			be.dropSchedule(player);
			return;
		}
		
		if (doorControl != null)
			be.doorControls.set(doorControl);

		if (!name.isBlank())
			be.updateName(name);

		if (!(blockState.method_26204() instanceof StationBlock))
			return;

		Boolean isAssemblyMode = blockState.method_11654(StationBlock.ASSEMBLING);
		boolean assemblyComplete = false;

		if (tryAssemble != null) {
			if (!isAssemblyMode)
				return;
			if (tryAssemble) {
				be.assemble(player.method_5667());
				assemblyComplete = be.getStation() != null && be.getStation()
					.getPresentTrain() != null;
			} else {
				if (be.tryDisassembleTrain(player) && be.tryEnterAssemblyMode())
					be.refreshAssemblyInfo();
			}
			if (!assemblyComplete)
				return;
		}

		if (assemblyMode)
			be.enterAssemblyMode(player);
		else
			be.exitAssemblyMode();
	}

	@Override
	protected void applySettings(StationBlockEntity be) {}

}
