package com.simibubi.create.content.trains.track;

import java.util.function.Function;
import java.util.function.UnaryOperator;

import com.simibubi.create.content.contraptions.glue.SuperGlueEntity;
import com.simibubi.create.foundation.utility.AttachedRegistry;
import com.simibubi.create.foundation.utility.BlockFace;
import com.simibubi.create.foundation.utility.Pair;

import io.github.fabricators_of_create.porting_lib.entity.ITeleporter;
import net.minecraft.class_1937;
import net.minecraft.class_2246;
import net.minecraft.class_2248;
import net.minecraft.class_2338;
import net.minecraft.class_2350;
import net.minecraft.class_238;
import net.minecraft.class_2680;
import net.minecraft.class_2741;
import net.minecraft.class_2960;
import net.minecraft.class_3218;
import net.minecraft.class_5321;
import net.minecraft.class_5454;
import net.minecraft.class_7923;
import net.minecraft.server.MinecraftServer;

public class AllPortalTracks {

	// Portals must be entered from the side and must lead to a different dimension
	// than the one entered from

	@FunctionalInterface
	public interface PortalTrackProvider extends UnaryOperator<Pair<class_3218, BlockFace>> {
	};

	private static final AttachedRegistry<class_2248, PortalTrackProvider> PORTAL_BEHAVIOURS =
		new AttachedRegistry<>(class_7923.field_41175);

	public static void registerIntegration(class_2960 block, PortalTrackProvider provider) {
		PORTAL_BEHAVIOURS.register(block, provider);
	}

	public static void registerIntegration(class_2248 block, PortalTrackProvider provider) {
		PORTAL_BEHAVIOURS.register(block, provider);
	}

	public static boolean isSupportedPortal(class_2680 state) {
		return PORTAL_BEHAVIOURS.get(state.method_26204()) != null;
	}

	public static Pair<class_3218, BlockFace> getOtherSide(class_3218 level, BlockFace inboundTrack) {
		class_2338 portalPos = inboundTrack.getConnectedPos();
		class_2680 portalState = level.method_8320(portalPos);
		PortalTrackProvider provider = PORTAL_BEHAVIOURS.get(portalState.method_26204());
		return provider == null ? null : provider.apply(Pair.of(level, inboundTrack));
	}

	// Builtin handlers

	public static void registerDefaults() {
		registerIntegration(class_2246.field_10316, AllPortalTracks::nether);
	}

	private static Pair<class_3218, BlockFace> nether(Pair<class_3218, BlockFace> inbound) {
		return standardPortalProvider(inbound, class_1937.field_25179, class_1937.field_25180, class_3218::method_14173);
	}

	public static Pair<class_3218, BlockFace> standardPortalProvider(Pair<class_3218, BlockFace> inbound,
		class_5321<class_1937> firstDimension, class_5321<class_1937> secondDimension,
		Function<class_3218, ITeleporter> customPortalForcer) {
		class_3218 level = inbound.getFirst();
		class_5321<class_1937> resourcekey = level.method_27983() == secondDimension ? firstDimension : secondDimension;
		MinecraftServer minecraftserver = level.method_8503();
		class_3218 otherLevel = minecraftserver.method_3847(resourcekey);

		if (otherLevel == null || !minecraftserver.method_3839())
			return null;

		BlockFace inboundTrack = inbound.getSecond();
		class_2338 portalPos = inboundTrack.getConnectedPos();
		class_2680 portalState = level.method_8320(portalPos);
		ITeleporter teleporter = customPortalForcer.apply(otherLevel);

		SuperGlueEntity probe = new SuperGlueEntity(level, new class_238(portalPos));
		probe.method_36456(inboundTrack.getFace()
			.method_10144());
		probe.setPortalEntrancePos();

		class_5454 portalinfo = teleporter.getPortalInfo(probe, otherLevel, probe::method_30329);
		if (portalinfo == null)
			return null;

		class_2338 otherPortalPos = class_2338.method_49638(portalinfo.field_25879);
		class_2680 otherPortalState = otherLevel.method_8320(otherPortalPos);
		if (otherPortalState.method_26204() != portalState.method_26204())
			return null;

		class_2350 targetDirection = inboundTrack.getFace();
		if (targetDirection.method_10166() == otherPortalState.method_11654(class_2741.field_12529))
			targetDirection = targetDirection.method_10170();
		class_2338 otherPos = otherPortalPos.method_10093(targetDirection);
		return Pair.of(otherLevel, new BlockFace(otherPos, targetDirection.method_10153()));
	}

}
