package com.simibubi.create.content.trains.track;

import java.util.Iterator;

import com.jozufozu.flywheel.util.transform.TransformStack;
import com.simibubi.create.AllBlocks;
import com.simibubi.create.foundation.utility.Couple;
import com.simibubi.create.foundation.utility.Iterate;
import com.simibubi.create.foundation.utility.NBTHelper;
import com.simibubi.create.foundation.utility.VecHelper;
import net.fabricmc.api.EnvType;
import net.fabricmc.api.Environment;
import net.minecraft.class_1542;
import net.minecraft.class_1657;
import net.minecraft.class_1661;
import net.minecraft.class_1799;
import net.minecraft.class_1928;
import net.minecraft.class_1937;
import net.minecraft.class_2338;
import net.minecraft.class_2350.class_2351;
import net.minecraft.class_238;
import net.minecraft.class_2388;
import net.minecraft.class_2398;
import net.minecraft.class_243;
import net.minecraft.class_2487;
import net.minecraft.class_2512;
import net.minecraft.class_2520;
import net.minecraft.class_2540;
import net.minecraft.class_3218;
import net.minecraft.class_3532;
import net.minecraft.class_4587;
import net.minecraft.class_4587.class_4665;

public class BezierConnection implements Iterable<BezierConnection.Segment> {

	public Couple<class_2338> tePositions;
	public Couple<class_243> starts;
	public Couple<class_243> axes;
	public Couple<class_243> normals;
	public Couple<Integer> smoothing;
	public boolean primary;
	public boolean hasGirder;
	protected TrackMaterial trackMaterial;

	// runtime

	class_243 finish1;
	class_243 finish2;
	private boolean resolved;
	private double length;
	private float[] stepLUT;
	private int segments;

	private double radius;
	private double handleLength;

	private class_238 bounds;

	public BezierConnection(Couple<class_2338> positions, Couple<class_243> starts, Couple<class_243> axes, Couple<class_243> normals,
		boolean primary, boolean girder, TrackMaterial material) {
		tePositions = positions;
		this.starts = starts;
		this.axes = axes;
		this.normals = normals;
		this.primary = primary;
		this.hasGirder = girder;
		this.trackMaterial = material;
		resolved = false;
	}

	public BezierConnection secondary() {
		BezierConnection bezierConnection = new BezierConnection(tePositions.swap(), starts.swap(), axes.swap(),
			normals.swap(), !primary, hasGirder, trackMaterial);
		if (smoothing != null)
			bezierConnection.smoothing = smoothing.swap();
		return bezierConnection;
	}

	public BezierConnection clone() {
		return secondary().secondary();
	}

	private static boolean coupleEquals(Couple<?> a, Couple<?> b) {
		return (a.getFirst()
			.equals(b.getFirst())
			&& a.getSecond()
				.equals(b.getSecond()))
			|| (a.getFirst() instanceof class_243 aFirst && a.getSecond() instanceof class_243 aSecond
				&& b.getFirst() instanceof class_243 bFirst && b.getSecond() instanceof class_243 bSecond
				&& aFirst.method_24802(bFirst, 1e-6) && aSecond.method_24802(bSecond, 1e-6));
	}

	public boolean equalsSansMaterial(BezierConnection other) {
		return equalsSansMaterialInner(other) || equalsSansMaterialInner(other.secondary());
	}

	private boolean equalsSansMaterialInner(BezierConnection other) {
		return this == other || (other != null && coupleEquals(this.tePositions, other.tePositions)
			&& coupleEquals(this.starts, other.starts) && coupleEquals(this.axes, other.axes)
			&& coupleEquals(this.normals, other.normals) && this.hasGirder == other.hasGirder);
	}

	public BezierConnection(class_2487 compound, class_2338 localTo) {
		this(Couple.deserializeEach(compound.method_10554("Positions", class_2520.field_33260), class_2512::method_10691)
			.map(b -> b.method_10081(localTo)),
			Couple.deserializeEach(compound.method_10554("Starts", class_2520.field_33260), VecHelper::readNBTCompound)
				.map(v -> v.method_1019(class_243.method_24954(localTo))),
			Couple.deserializeEach(compound.method_10554("Axes", class_2520.field_33260), VecHelper::readNBTCompound),
			Couple.deserializeEach(compound.method_10554("Normals", class_2520.field_33260), VecHelper::readNBTCompound),
			compound.method_10577("Primary"), compound.method_10577("Girder"), TrackMaterial.deserialize(compound.method_10558("Material")));

		if (compound.method_10545("Smoothing"))
			smoothing =
				Couple.deserializeEach(compound.method_10554("Smoothing", class_2520.field_33260), NBTHelper::intFromCompound);
	}

	public class_2487 write(class_2338 localTo) {
		Couple<class_2338> tePositions = this.tePositions.map(b -> b.method_10059(localTo));
		Couple<class_243> starts = this.starts.map(v -> v.method_1020(class_243.method_24954(localTo)));

		class_2487 compound = new class_2487();
		compound.method_10556("Girder", hasGirder);
		compound.method_10556("Primary", primary);
		compound.method_10566("Positions", tePositions.serializeEach(class_2512::method_10692));
		compound.method_10566("Starts", starts.serializeEach(VecHelper::writeNBTCompound));
		compound.method_10566("Axes", axes.serializeEach(VecHelper::writeNBTCompound));
		compound.method_10566("Normals", normals.serializeEach(VecHelper::writeNBTCompound));
		compound.method_10582("Material", getMaterial().id.toString());

		if (smoothing != null)
			compound.method_10566("Smoothing", smoothing.serializeEach(NBTHelper::intToCompound));

		return compound;
	}

	public BezierConnection(class_2540 buffer) {
		this(Couple.create(buffer::method_10811), Couple.create(() -> VecHelper.read(buffer)),
			Couple.create(() -> VecHelper.read(buffer)), Couple.create(() -> VecHelper.read(buffer)),
			buffer.readBoolean(), buffer.readBoolean(), TrackMaterial.deserialize(buffer.method_19772()));
		if (buffer.readBoolean())
			smoothing = Couple.create(buffer::method_10816);
	}

	public void write(class_2540 buffer) {
		tePositions.forEach(buffer::method_10807);
		starts.forEach(v -> VecHelper.write(v, buffer));
		axes.forEach(v -> VecHelper.write(v, buffer));
		normals.forEach(v -> VecHelper.write(v, buffer));
		buffer.writeBoolean(primary);
		buffer.writeBoolean(hasGirder);
		buffer.method_10814(getMaterial().id.toString());
		buffer.writeBoolean(smoothing != null);
		if (smoothing != null)
			smoothing.forEach(buffer::method_10804);
	}

	public class_2338 getKey() {
		return tePositions.getSecond();
	}

	public boolean isPrimary() {
		return primary;
	}

	public int yOffsetAt(class_243 end) {
		if (smoothing == null)
			return 0;
		if (TrackBlockEntityTilt.compareHandles(starts.getFirst(), end))
			return smoothing.getFirst();
		if (TrackBlockEntityTilt.compareHandles(starts.getSecond(), end))
			return smoothing.getSecond();
		return 0;
	}

	// Runtime information

	public double getLength() {
		resolve();
		return length;
	}

	public float[] getStepLUT() {
		resolve();
		return stepLUT;
	}

	public int getSegmentCount() {
		resolve();
		return segments;
	}

	public class_243 getPosition(double t) {
		resolve();
		return VecHelper.bezier(starts.getFirst(), starts.getSecond(), finish1, finish2, (float) t);
	}

	public double getRadius() {
		resolve();
		return radius;
	}

	public double getHandleLength() {
		resolve();
		return handleLength;
	}

	public float getSegmentT(int index) {
		return index == segments ? 1 : index * stepLUT[index] / segments;
	}

	public double incrementT(double currentT, double distance) {
		resolve();
		double dx =
			VecHelper.bezierDerivative(starts.getFirst(), starts.getSecond(), finish1, finish2, (float) currentT)
				.method_1033() / getLength();
		return currentT + distance / dx;

	}

	public class_238 getBounds() {
		resolve();
		return bounds;
	}

	public class_243 getNormal(double t) {
		resolve();
		class_243 end1 = starts.getFirst();
		class_243 end2 = starts.getSecond();
		class_243 fn1 = normals.getFirst();
		class_243 fn2 = normals.getSecond();

		class_243 derivative = VecHelper.bezierDerivative(end1, end2, finish1, finish2, (float) t)
			.method_1029();
		class_243 faceNormal = fn1.equals(fn2) ? fn1 : VecHelper.slerp((float) t, fn1, fn2);
		class_243 normal = faceNormal.method_1036(derivative)
			.method_1029();
		return derivative.method_1036(normal);
	}

	private void resolve() {
		if (resolved)
			return;
		resolved = true;

		class_243 end1 = starts.getFirst();
		class_243 end2 = starts.getSecond();
		class_243 axis1 = axes.getFirst()
			.method_1029();
		class_243 axis2 = axes.getSecond()
			.method_1029();

		determineHandles(end1, end2, axis1, axis2);

		finish1 = axis1.method_1021(handleLength)
			.method_1019(end1);
		finish2 = axis2.method_1021(handleLength)
			.method_1019(end2);

		int scanCount = 16;
		length = 0;

		{
			class_243 previous = end1;
			for (int i = 0; i <= scanCount; i++) {
				float t = i / (float) scanCount;
				class_243 result = VecHelper.bezier(end1, end2, finish1, finish2, t);
				if (previous != null)
					length += result.method_1022(previous);
				previous = result;
			}
		}

		segments = (int) (length * 2);
		stepLUT = new float[segments + 1];
		stepLUT[0] = 1;
		float combinedDistance = 0;

		bounds = new class_238(end1, end2);

		// determine step lut
		{
			class_243 previous = end1;
			for (int i = 0; i <= segments; i++) {
				float t = i / (float) segments;
				class_243 result = VecHelper.bezier(end1, end2, finish1, finish2, t);
				bounds = bounds.method_991(new class_238(result, result));
				if (i > 0) {
					combinedDistance += result.method_1022(previous) / length;
					stepLUT[i] = (float) (t / combinedDistance);
				}
				previous = result;
			}
		}

		bounds = bounds.method_1014(1.375f);
	}

	private void determineHandles(class_243 end1, class_243 end2, class_243 axis1, class_243 axis2) {
		class_243 cross1 = axis1.method_1036(new class_243(0, 1, 0));
		class_243 cross2 = axis2.method_1036(new class_243(0, 1, 0));

		radius = 0;
		double a1 = class_3532.method_15349(-axis2.field_1350, -axis2.field_1352);
		double a2 = class_3532.method_15349(axis1.field_1350, axis1.field_1352);
		double angle = a1 - a2;

		float circle = 2 * class_3532.field_29844;
		angle = (angle + circle) % circle;
		if (Math.abs(circle - angle) < Math.abs(angle))
			angle = circle - angle;

		if (class_3532.method_20390(angle, 0)) {
			double[] intersect = VecHelper.intersect(end1, end2, axis1, cross2, class_2351.field_11052);
			if (intersect != null) {
				double t = Math.abs(intersect[0]);
				double u = Math.abs(intersect[1]);
				double min = Math.min(t, u);
				double max = Math.max(t, u);

				if (min > 1.2 && max / min > 1 && max / min < 3) {
					handleLength = (max - min);
					return;
				}
			}

			handleLength = end2.method_1022(end1) / 3;
			return;
		}

		double n = circle / angle;
		double factor = 4 / 3d * Math.tan(Math.PI / (2 * n));
		double[] intersect = VecHelper.intersect(end1, end2, cross1, cross2, class_2351.field_11052);

		if (intersect == null) {
			handleLength = end2.method_1022(end1) / 3;
			return;
		}

		radius = Math.abs(intersect[1]);
		handleLength = radius * factor;
		if (class_3532.method_20390(handleLength, 0))
			handleLength = 1;
	}

	@Override
	public Iterator<Segment> iterator() {
		resolve();
		var offset = class_243.method_24954(tePositions.getFirst())
			.method_1021(-1)
			.method_1031(0, 3 / 16f, 0);
		return new Bezierator(this, offset);
	}

	public void addItemsToPlayer(class_1657 player) {
		class_1661 inv = player.method_31548();
		int tracks = getTrackItemCost();
		while (tracks > 0) {
			inv.method_7398(new class_1799(getMaterial().getBlock(), Math.min(64, tracks)));
			tracks -= 64;
		}
		int girders = getGirderItemCost();
		while (girders > 0) {
			inv.method_7398(AllBlocks.METAL_GIRDER.asStack(Math.min(64, girders)));
			girders -= 64;
		}
	}

	public int getGirderItemCost() {
		return hasGirder ? getTrackItemCost() * 2 : 0;
	}

	public int getTrackItemCost() {
		return (getSegmentCount() + 1) / 2;
	}

	public void spawnItems(class_1937 level) {
		if (!level.method_8450()
			.method_8355(class_1928.field_19392))
			return;
		class_243 origin = class_243.method_24954(tePositions.getFirst());
		for (Segment segment : this) {
			if (segment.index % 2 != 0 || segment.index == getSegmentCount())
				continue;
			class_243 v = VecHelper.offsetRandomly(segment.position, level.field_9229, .125f)
				.method_1019(origin);
			class_1542 entity = new class_1542(level, v.field_1352, v.field_1351, v.field_1350, getMaterial().asStack());
			entity.method_6988();
			level.method_8649(entity);
			if (!hasGirder)
				continue;
			for (int i = 0; i < 2; i++) {
				entity = new class_1542(level, v.field_1352, v.field_1351, v.field_1350, AllBlocks.METAL_GIRDER.asStack());
				entity.method_6988();
				level.method_8649(entity);
			}
		}
	}

	public void spawnDestroyParticles(class_1937 level) {
		class_2388 data = new class_2388(class_2398.field_11217, getMaterial().getBlock().method_9564());
		class_2388 girderData =
			new class_2388(class_2398.field_11217, AllBlocks.METAL_GIRDER.getDefaultState());
		if (!(level instanceof class_3218 slevel))
			return;
		class_243 origin = class_243.method_24954(tePositions.getFirst());
		for (Segment segment : this) {
			for (int offset : Iterate.positiveAndNegative) {
				class_243 v = segment.position.method_1019(segment.normal.method_1021(14 / 16f * offset))
					.method_1019(origin);
				slevel.method_14199(data, v.field_1352, v.field_1351, v.field_1350, 1, 0, 0, 0, 0);
				if (!hasGirder)
					continue;
				slevel.method_14199(girderData, v.field_1352, v.field_1351 - .5f, v.field_1350, 1, 0, 0, 0, 0);
			}
		}
	}

	public TrackMaterial getMaterial() {
		return trackMaterial;
	}

	public void setMaterial(TrackMaterial material) {
		trackMaterial = material;
	}

	public static class Segment {

		public int index;
		public class_243 position;
		public class_243 derivative;
		public class_243 faceNormal;
		public class_243 normal;

	}

	private static class Bezierator implements Iterator<Segment> {

		private final BezierConnection bc;
		private final Segment segment;
		private final class_243 end1;
		private final class_243 end2;
		private final class_243 finish1;
		private final class_243 finish2;
		private final class_243 faceNormal1;
		private final class_243 faceNormal2;

		private Bezierator(BezierConnection bc, class_243 offset) {
			bc.resolve();
			this.bc = bc;

			end1 = bc.starts.getFirst()
				.method_1019(offset);
			end2 = bc.starts.getSecond()
				.method_1019(offset);

			finish1 = bc.axes.getFirst()
				.method_1021(bc.handleLength)
				.method_1019(end1);
			finish2 = bc.axes.getSecond()
				.method_1021(bc.handleLength)
				.method_1019(end2);

			faceNormal1 = bc.normals.getFirst();
			faceNormal2 = bc.normals.getSecond();
			segment = new Segment();
			segment.index = -1; // will get incremented to 0 in #next()
		}

		@Override
		public boolean hasNext() {
			return segment.index + 1 <= bc.segments;
		}

		@Override
		public Segment next() {
			segment.index++;
			float t = this.bc.getSegmentT(segment.index);
			segment.position = VecHelper.bezier(end1, end2, finish1, finish2, t);
			segment.derivative = VecHelper.bezierDerivative(end1, end2, finish1, finish2, t)
				.method_1029();
			segment.faceNormal =
				faceNormal1.equals(faceNormal2) ? faceNormal1 : VecHelper.slerp(t, faceNormal1, faceNormal2);
			segment.normal = segment.faceNormal.method_1036(segment.derivative)
				.method_1029();
			return segment;
		}
	}

	private SegmentAngles[] bakedSegments;
	private GirderAngles[] bakedGirders;

	@Environment(EnvType.CLIENT)
	public static class SegmentAngles {

		public class_4665 tieTransform;
		public Couple<class_4665> railTransforms;
		public class_2338 lightPosition;

	}

	@Environment(EnvType.CLIENT)
	public static class GirderAngles {

		public Couple<class_4665> beams;
		public Couple<Couple<class_4665>> beamCaps;
		public class_2338 lightPosition;

	}

	@Environment(EnvType.CLIENT)
	public SegmentAngles[] getBakedSegments() {
		if (bakedSegments != null)
			return bakedSegments;

		int segmentCount = getSegmentCount();
		bakedSegments = new SegmentAngles[segmentCount + 1];
		Couple<class_243> previousOffsets = null;

		for (BezierConnection.Segment segment : this) {
			int i = segment.index;
			boolean end = i == 0 || i == segmentCount;

			SegmentAngles angles = bakedSegments[i] = new SegmentAngles();
			Couple<class_243> railOffsets = Couple.create(segment.position.method_1019(segment.normal.method_1021(.965f)),
				segment.position.method_1020(segment.normal.method_1021(.965f)));
			class_243 railMiddle = railOffsets.getFirst()
				.method_1019(railOffsets.getSecond())
				.method_1021(.5);

			if (previousOffsets == null) {
				previousOffsets = railOffsets;
				continue;
			}

			// Tie
			class_243 prevMiddle = previousOffsets.getFirst()
				.method_1019(previousOffsets.getSecond())
				.method_1021(.5);
			class_243 tieAngles = TrackRenderer.getModelAngles(segment.normal, railMiddle.method_1020(prevMiddle));
			angles.lightPosition = class_2338.method_49638(railMiddle);
			angles.railTransforms = Couple.create(null, null);

			class_4587 poseStack = new class_4587();
			TransformStack.cast(poseStack)
				.translate(prevMiddle)
				.rotateYRadians(tieAngles.field_1351)
				.rotateXRadians(tieAngles.field_1352)
				.rotateZRadians(tieAngles.field_1350)
				.translate(-1 / 2f, -2 / 16f - 1 / 256f, 0);
			angles.tieTransform = poseStack.method_23760();

			// Rails
			float scale = end ? 2.2f : 2.1f;
			for (boolean first : Iterate.trueAndFalse) {
				class_243 railI = railOffsets.get(first);
				class_243 prevI = previousOffsets.get(first);
				class_243 diff = railI.method_1020(prevI);
				class_243 anglesI = TrackRenderer.getModelAngles(segment.normal, diff);

				poseStack = new class_4587();
				TransformStack.cast(poseStack)
					.translate(prevI)
					.rotateYRadians(anglesI.field_1351)
					.rotateXRadians(anglesI.field_1352)
					.rotateZRadians(anglesI.field_1350)
					.translate(0, -2 / 16f - 1 / 256f, -1 / 32f)
					.scale(1, 1, (float) diff.method_1033() * scale);
				angles.railTransforms.set(first, poseStack.method_23760());
			}

			previousOffsets = railOffsets;
		}

		return bakedSegments;
	}

	@Environment(EnvType.CLIENT)
	public GirderAngles[] getBakedGirders() {
		if (bakedGirders != null)
			return bakedGirders;

		int segmentCount = getSegmentCount();
		bakedGirders = new GirderAngles[segmentCount + 1];
		Couple<Couple<class_243>> previousOffsets = null;

		for (BezierConnection.Segment segment : this) {
			int i = segment.index;
			boolean end = i == 0 || i == segmentCount;
			GirderAngles angles = bakedGirders[i] = new GirderAngles();

			class_243 leftGirder = segment.position.method_1019(segment.normal.method_1021(.965f));
			class_243 rightGirder = segment.position.method_1020(segment.normal.method_1021(.965f));
			class_243 upNormal = segment.derivative.method_1029()
				.method_1036(segment.normal);
			class_243 firstGirderOffset = upNormal.method_1021(-8 / 16f);
			class_243 secondGirderOffset = upNormal.method_1021(-10 / 16f);
			class_243 leftTop = segment.position.method_1019(segment.normal.method_1021(1))
				.method_1019(firstGirderOffset);
			class_243 rightTop = segment.position.method_1020(segment.normal.method_1021(1))
				.method_1019(firstGirderOffset);
			class_243 leftBottom = leftTop.method_1019(secondGirderOffset);
			class_243 rightBottom = rightTop.method_1019(secondGirderOffset);

			angles.lightPosition = class_2338.method_49638(leftGirder.method_1019(rightGirder)
				.method_1021(.5));

			Couple<Couple<class_243>> offsets =
				Couple.create(Couple.create(leftTop, rightTop), Couple.create(leftBottom, rightBottom));

			if (previousOffsets == null) {
				previousOffsets = offsets;
				continue;
			}

			angles.beams = Couple.create(null, null);
			angles.beamCaps = Couple.create(Couple.create(null, null), Couple.create(null, null));
			float scale = end ? 2.3f : 2.2f;

			for (boolean first : Iterate.trueAndFalse) {

				// Middle
				class_243 currentBeam = offsets.getFirst()
					.get(first)
					.method_1019(offsets.getSecond()
						.get(first))
					.method_1021(.5);
				class_243 previousBeam = previousOffsets.getFirst()
					.get(first)
					.method_1019(previousOffsets.getSecond()
						.get(first))
					.method_1021(.5);
				class_243 beamDiff = currentBeam.method_1020(previousBeam);
				class_243 beamAngles = TrackRenderer.getModelAngles(segment.normal, beamDiff);

				class_4587 poseStack = new class_4587();
				TransformStack.cast(poseStack)
					.translate(previousBeam)
					.rotateYRadians(beamAngles.field_1351)
					.rotateXRadians(beamAngles.field_1352)
					.rotateZRadians(beamAngles.field_1350)
					.translate(0, 2 / 16f + (segment.index % 2 == 0 ? 1 : -1) / 2048f - 1 / 1024f, -1 / 32f)
					.scale(1, 1, (float) beamDiff.method_1033() * scale);
				angles.beams.set(first, poseStack.method_23760());

				// Caps
				for (boolean top : Iterate.trueAndFalse) {
					class_243 current = offsets.get(top)
						.get(first);
					class_243 previous = previousOffsets.get(top)
						.get(first);
					class_243 diff = current.method_1020(previous);
					class_243 capAngles = TrackRenderer.getModelAngles(segment.normal, diff);

					poseStack = new class_4587();
					TransformStack.cast(poseStack)
						.translate(previous)
						.rotateYRadians(capAngles.field_1351)
						.rotateXRadians(capAngles.field_1352)
						.rotateZRadians(capAngles.field_1350)
						.translate(0, 2 / 16f + (segment.index % 2 == 0 ? 1 : -1) / 2048f - 1 / 1024f, -1 / 32f)
						.rotateZ(top ? 0 : 0)
						.scale(1, 1, (float) diff.method_1033() * scale);
					angles.beamCaps.get(top)
						.set(first, poseStack.method_23760());
				}
			}

			previousOffsets = offsets;

		}

		return bakedGirders;
	}

}
