package com.simibubi.create.content.trains.track;

import org.apache.commons.lang3.mutable.MutableObject;

import com.simibubi.create.AllBlocks;
import com.simibubi.create.AllSoundEvents;
import com.simibubi.create.content.trains.graph.EdgePointType;
import com.simibubi.create.content.trains.track.TrackTargetingBlockItem.OverlapResult;
import com.simibubi.create.foundation.networking.BlockEntityConfigurationPacket;
import com.simibubi.create.foundation.utility.Lang;
import net.minecraft.class_124;
import net.minecraft.class_1799;
import net.minecraft.class_2338;
import net.minecraft.class_2487;
import net.minecraft.class_2512;
import net.minecraft.class_2540;
import net.minecraft.class_3222;

public class CurvedTrackSelectionPacket extends BlockEntityConfigurationPacket<TrackBlockEntity> {

	private class_2338 targetPos;
	private boolean front;
	private int segment;
	private int slot;

	public CurvedTrackSelectionPacket(class_2338 pos, class_2338 targetPos, int segment, boolean front, int slot) {
		super(pos);
		this.targetPos = targetPos;
		this.segment = segment;
		this.front = front;
		this.slot = slot;
	}

	public CurvedTrackSelectionPacket(class_2540 buffer) {
		super(buffer);
	}

	@Override
	protected void writeSettings(class_2540 buffer) {
		buffer.method_10807(targetPos);
		buffer.method_10804(segment);
		buffer.writeBoolean(front);
		buffer.method_10804(slot);
	}

	@Override
	protected void readSettings(class_2540 buffer) {
		targetPos = buffer.method_10811();
		segment = buffer.method_10816();
		front = buffer.readBoolean();
		slot = buffer.method_10816();
	}

	@Override
	protected void applySettings(class_3222 player, TrackBlockEntity be) {
		if (player.method_31548().field_7545 != slot)
			return;
		class_1799 stack = player.method_31548()
			.method_5438(slot);
		if (!(stack.method_7909() instanceof TrackTargetingBlockItem))
			return;
		if (player.method_5715() && stack.method_7985()) {
			player.method_7353(Lang.translateDirect("track_target.clear"), true);
			stack.method_7980(null);
			AllSoundEvents.CONTROLLER_CLICK.play(player.method_37908(), null, pos, 1, .5f);
			return;
		}

		EdgePointType<?> type = AllBlocks.TRACK_SIGNAL.isIn(stack) ? EdgePointType.SIGNAL : EdgePointType.STATION;
		MutableObject<OverlapResult> result = new MutableObject<>(null);
		TrackTargetingBlockItem.withGraphLocation(player.method_37908(), pos, front,
			new BezierTrackPointLocation(targetPos, segment), type, (overlap, location) -> result.setValue(overlap));

		if (result.getValue().feedback != null) {
			player.method_7353(Lang.translateDirect(result.getValue().feedback)
				.method_27692(class_124.field_1061), true);
			AllSoundEvents.DENY.play(player.method_37908(), null, pos, .5f, 1);
			return;
		}

		class_2487 stackTag = stack.method_7948();
		stackTag.method_10566("SelectedPos", class_2512.method_10692(pos));
		stackTag.method_10556("SelectedDirection", front);

		class_2487 bezierNbt = new class_2487();
		bezierNbt.method_10569("Segment", segment);
		bezierNbt.method_10566("Key", class_2512.method_10692(targetPos));
		bezierNbt.method_10556("FromStack", true);
		stackTag.method_10566("Bezier", bezierNbt);

		player.method_7353(Lang.translateDirect("track_target.set"), true);
		stack.method_7980(stackTag);
		AllSoundEvents.CONTROLLER_CLICK.play(player.method_37908(), null, pos, 1, 1);
	}

	@Override
	protected int maxRange() {
		return 64;
	}

	@Override
	protected void applySettings(TrackBlockEntity be) {}

}
