package com.simibubi.create.content.trains.track;

import static com.simibubi.create.AllShapes.TRACK_ASC;
import static com.simibubi.create.AllShapes.TRACK_CROSS;
import static com.simibubi.create.AllShapes.TRACK_CROSS_DIAG;
import static com.simibubi.create.AllShapes.TRACK_CROSS_DIAG_ORTHO;
import static com.simibubi.create.AllShapes.TRACK_CROSS_ORTHO_DIAG;
import static com.simibubi.create.AllShapes.TRACK_DIAG;
import static com.simibubi.create.AllShapes.TRACK_ORTHO;
import static com.simibubi.create.AllShapes.TRACK_ORTHO_LONG;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import org.jetbrains.annotations.Nullable;

import com.google.common.base.Predicates;
import com.jozufozu.flywheel.core.PartialModel;
import com.jozufozu.flywheel.util.transform.TransformStack;
import com.simibubi.create.AllBlockEntityTypes;
import com.simibubi.create.AllBlocks;
import com.simibubi.create.AllPartialModels;
import com.simibubi.create.AllShapes;
import com.simibubi.create.AllTags;
import com.simibubi.create.content.decoration.girder.GirderBlock;
import com.simibubi.create.content.equipment.wrench.IWrenchable;
import com.simibubi.create.content.schematics.requirement.ISpecialBlockItemRequirement;
import com.simibubi.create.content.schematics.requirement.ItemRequirement;
import com.simibubi.create.content.schematics.requirement.ItemRequirement.ItemUseType;
import com.simibubi.create.content.trains.CubeParticleData;
import com.simibubi.create.content.trains.graph.TrackNodeLocation;
import com.simibubi.create.content.trains.graph.TrackNodeLocation.DiscoveredLocation;
import com.simibubi.create.content.trains.station.StationBlockEntity;
import com.simibubi.create.content.trains.track.TrackTargetingBehaviour.RenderedTrackOverlayType;
import com.simibubi.create.foundation.block.IBE;
import com.simibubi.create.foundation.block.ProperWaterloggedBlock;
import com.simibubi.create.foundation.block.render.MultiPosDestructionHandler;
import com.simibubi.create.foundation.block.render.ReducedDestroyEffects;
import com.simibubi.create.foundation.utility.AngleHelper;
import com.simibubi.create.foundation.utility.BlockFace;
import com.simibubi.create.foundation.utility.Components;
import com.simibubi.create.foundation.utility.Iterate;
import com.simibubi.create.foundation.utility.Lang;
import com.simibubi.create.foundation.utility.Pair;
import com.simibubi.create.foundation.utility.VecHelper;

import it.unimi.dsi.fastutil.objects.Object2IntArrayMap;
import it.unimi.dsi.fastutil.objects.Object2IntMap;
import net.fabricmc.api.EnvType;
import net.fabricmc.api.Environment;
import net.fabricmc.fabric.api.registry.LandPathNodeTypesRegistry;
import net.minecraft.class_124;
import net.minecraft.class_1268;
import net.minecraft.class_1269;
import net.minecraft.class_1309;
import net.minecraft.class_1657;
import net.minecraft.class_1750;
import net.minecraft.class_1799;
import net.minecraft.class_1838;
import net.minecraft.class_1922;
import net.minecraft.class_1936;
import net.minecraft.class_1937;
import net.minecraft.class_2246;
import net.minecraft.class_2248;
import net.minecraft.class_2338;
import net.minecraft.class_2350;
import net.minecraft.class_2350.class_2351;
import net.minecraft.class_2350.class_2352;
import net.minecraft.class_2415;
import net.minecraft.class_243;
import net.minecraft.class_2470;
import net.minecraft.class_2586;
import net.minecraft.class_259;
import net.minecraft.class_2591;
import net.minecraft.class_265;
import net.minecraft.class_2680;
import net.minecraft.class_2689.class_2690;
import net.minecraft.class_2741;
import net.minecraft.class_2746;
import net.minecraft.class_2754;
import net.minecraft.class_3218;
import net.minecraft.class_3341;
import net.minecraft.class_3532;
import net.minecraft.class_3610;
import net.minecraft.class_3726;
import net.minecraft.class_3965;
import net.minecraft.class_4538;
import net.minecraft.class_4587;
import net.minecraft.class_5250;
import net.minecraft.class_5321;
import net.minecraft.class_5819;
import net.minecraft.class_638;
import net.minecraft.class_6756;
import net.minecraft.class_7;

public class TrackBlock extends class_2248 implements IBE<TrackBlockEntity>, IWrenchable, ITrackBlock, ISpecialBlockItemRequirement, ProperWaterloggedBlock, ReducedDestroyEffects, MultiPosDestructionHandler {

	public static final class_2754<TrackShape> SHAPE = class_2754.method_11850("shape", TrackShape.class);
	public static final class_2746 HAS_BE = class_2746.method_11825("turn");

	protected final TrackMaterial material;

	public TrackBlock(class_2251 p_49795_, TrackMaterial material) {
		super(p_49795_);
		method_9590(method_9564().method_11657(SHAPE, TrackShape.ZO)
			.method_11657(HAS_BE, false)
			.method_11657(WATERLOGGED, false));
		this.material = material;
		LandPathNodeTypesRegistry.register(this, class_7.field_21, null);
	}

	@Override
	protected void method_9515(class_2690<class_2248, class_2680> p_49915_) {
		super.method_9515(p_49915_.method_11667(SHAPE, HAS_BE, WATERLOGGED));
	}

	@Override
	public class_3610 method_9545(class_2680 state) {
		return fluidState(state);
	}

	@Override
	public class_2680 method_9605(class_1750 ctx) {
		class_2680 stateForPlacement = withWater(super.method_9605(ctx), ctx);

		if (ctx.method_8036() == null)
			return stateForPlacement;

		class_243 lookAngle = ctx.method_8036()
			.method_5720();
		lookAngle = lookAngle.method_18805(1, 0, 1);
		if (class_3532.method_20390(lookAngle.method_1033(), 0))
			lookAngle = VecHelper.rotate(new class_243(0, 0, 1), -ctx.method_8036()
				.method_36454(), class_2351.field_11052);

		lookAngle = lookAngle.method_1029();

		TrackShape best = TrackShape.ZO;
		double bestValue = Float.MAX_VALUE;
		for (TrackShape shape : TrackShape.values()) {
			if (shape.isJunction() || shape.isPortal())
				continue;
			class_243 axis = shape.getAxes()
				.get(0);
			double distance = Math.min(axis.method_1025(lookAngle), axis.method_1029()
				.method_1021(-1)
				.method_1025(lookAngle));
			if (distance > bestValue)
				continue;
			bestValue = distance;
			best = shape;
		}

		class_1937 level = ctx.method_8045();
		class_243 bestAxis = best.getAxes()
			.get(0);
		if (bestAxis.method_1027() == 1)
			for (boolean neg : Iterate.trueAndFalse) {
				class_2338 offset = ctx.method_8037()
					.method_10081(class_2338.method_49638(bestAxis.method_1021(neg ? -1 : 1)));

				if (level.method_8320(offset)
					.method_26206(level, offset, class_2350.field_11036)
					&& !level.method_8320(offset.method_10084())
						.method_26206(level, offset, class_2350.field_11033)) {
					if (best == TrackShape.XO)
						best = neg ? TrackShape.AW : TrackShape.AE;
					if (best == TrackShape.ZO)
						best = neg ? TrackShape.AN : TrackShape.AS;
				}
			}

		return stateForPlacement.method_11657(SHAPE, best);
	}

	@Override
	public void method_9576(class_1937 pLevel, class_2338 pPos, class_2680 pState, class_1657 pPlayer) {
		super.method_9576(pLevel, pPos, pState, pPlayer);
		if (pLevel.method_8608())
			return;
		if (!pPlayer.method_7337())
			return;
		withBlockEntityDo(pLevel, pPos, be -> {
			be.cancelDrops = true;
			be.removeInboundConnections(true);
		});
	}

	@Override
	public void method_9615(class_2680 pState, class_1937 pLevel, class_2338 pPos, class_2680 pOldState, boolean pIsMoving) {
		if (pOldState.method_26204() == this && pState.method_11657(HAS_BE, true) == pOldState.method_11657(HAS_BE, true))
			return;
		if (pLevel.field_9236)
			return;
		class_6756<class_2248> blockTicks = pLevel.method_8397();
		if (!blockTicks.method_8674(pPos, this))
			pLevel.method_39279(pPos, this, 1);
		updateGirders(pState, pLevel, pPos, blockTicks);
	}

	@Override
	public void method_9567(class_1937 pLevel, class_2338 pPos, class_2680 pState, class_1309 pPlacer, class_1799 pStack) {
		super.method_9567(pLevel, pPos, pState, pPlacer, pStack);
		withBlockEntityDo(pLevel, pPos, TrackBlockEntity::validateConnections);
	}

	@Override
	public void method_9588(class_2680 state, class_3218 level, class_2338 pos, class_5819 p_60465_) {
		TrackPropagator.onRailAdded(level, pos, state);
		withBlockEntityDo(level, pos, tbe -> tbe.tilt.undoSmoothing());
		if (!state.method_11654(SHAPE)
			.isPortal())
			connectToPortal(level, pos, state);
	}

	protected void connectToPortal(class_3218 level, class_2338 pos, class_2680 state) {
		TrackShape shape = state.method_11654(TrackBlock.SHAPE);
		class_2351 portalTest = shape == TrackShape.XO ? class_2351.field_11048 : shape == TrackShape.ZO ? class_2351.field_11051 : null;
		if (portalTest == null)
			return;

		boolean pop = false;
		String fail = null;
		class_2338 failPos = null;

		for (class_2350 d : Iterate.directionsInAxis(portalTest)) {
			class_2338 portalPos = pos.method_10093(d);
			class_2680 portalState = level.method_8320(portalPos);
			if (!AllPortalTracks.isSupportedPortal(portalState))
				continue;

			pop = true;
			Pair<class_3218, BlockFace> otherSide = AllPortalTracks.getOtherSide(level, new BlockFace(pos, d));
			if (otherSide == null) {
				fail = "missing";
				continue;
			}

			class_3218 otherLevel = otherSide.getFirst();
			BlockFace otherTrack = otherSide.getSecond();
			class_2338 otherTrackPos = otherTrack.getPos();
			class_2680 existing = otherLevel.method_8320(otherTrackPos);
			if (!existing.method_45474()) {
				fail = "blocked";
				failPos = otherTrackPos;
				continue;
			}

			level.method_8652(pos, state.method_11657(SHAPE, TrackShape.asPortal(d))
				.method_11657(HAS_BE, true), 3);
			class_2586 be = level.method_8321(pos);
			if (be instanceof TrackBlockEntity tbe)
				tbe.bind(otherLevel.method_27983(), otherTrackPos);

			otherLevel.method_8652(otherTrackPos, state.method_11657(SHAPE, TrackShape.asPortal(otherTrack.getFace()))
				.method_11657(HAS_BE, true), 3);
			class_2586 otherBE = otherLevel.method_8321(otherTrackPos);
			if (otherBE instanceof TrackBlockEntity tbe)
				tbe.bind(level.method_27983(), pos);

			pop = false;
		}

		if (!pop)
			return;

		level.method_22352(pos, true);

		if (fail == null)
			return;
		class_1657 player = level.method_8604(pos.method_10263(), pos.method_10264(), pos.method_10260(), 10, Predicates.alwaysTrue());
		if (player == null)
			return;
		player.method_7353(Components.literal("<!> ")
			.method_10852(Lang.translateDirect("portal_track.failed"))
			.method_27692(class_124.field_1065), false);
		class_5250 component = failPos != null
			? Lang.translateDirect("portal_track." + fail, failPos.method_10263(), failPos.method_10264(), failPos.method_10260())
			: Lang.translateDirect("portal_track." + fail);
		player.method_7353(Components.literal(" - ")
			.method_27692(class_124.field_1080)
			.method_10852(component.method_27694(st -> st.method_36139(0xFFD3B4))), false);
	}

	@Override
	public class_2680 method_9559(class_2680 state, class_2350 pDirection, class_2680 pNeighborState,
		class_1936 level, class_2338 pCurrentPos, class_2338 pNeighborPos) {
		updateWater(level, state, pCurrentPos);
		TrackShape shape = state.method_11654(SHAPE);
		if (!shape.isPortal())
			return state;

		for (class_2350 d : Iterate.horizontalDirections) {
			if (TrackShape.asPortal(d) != state.method_11654(SHAPE))
				continue;
			if (pDirection != d)
				continue;

			class_2338 portalPos = pCurrentPos.method_10093(d);
			class_2680 portalState = level.method_8320(portalPos);
			if (!AllPortalTracks.isSupportedPortal(portalState))
				return class_2246.field_10124.method_9564();
		}

		return state;
	}

	@Override
	public int getYOffsetAt(class_1922 world, class_2338 pos, class_2680 state, class_243 end) {
		return getBlockEntityOptional(world, pos).map(tbe -> tbe.tilt.getYOffsetForAxisEnd(end))
			.orElse(0);
	}

	@Override
	public Collection<DiscoveredLocation> getConnected(class_1922 worldIn, class_2338 pos, class_2680 state,
		boolean linear, TrackNodeLocation connectedTo) {
		Collection<DiscoveredLocation> list;
		class_1922 world = connectedTo != null && worldIn instanceof class_3218 sl ? sl.method_8503()
			.method_3847(connectedTo.dimension) : worldIn;

		if (getTrackAxes(world, pos, state).size() > 1) {
			class_243 center = class_243.method_24955(pos)
				.method_1031(0, getElevationAtCenter(world, pos, state), 0);
			TrackShape shape = state.method_11654(TrackBlock.SHAPE);
			list = new ArrayList<>();
			for (class_243 axis : getTrackAxes(world, pos, state))
				for (boolean fromCenter : Iterate.trueAndFalse)
					ITrackBlock.addToListIfConnected(connectedTo, list,
						(d, b) -> axis.method_1021(b ? 0 : fromCenter ? -d : d)
							.method_1019(center),
						b -> shape.getNormal(), b -> world instanceof class_1937 l ? l.method_27983() : class_1937.field_25179, v -> 0,
						axis, null, (b, v) -> ITrackBlock.getMaterialSimple(world, v));
		} else
			list = ITrackBlock.super.getConnected(world, pos, state, linear, connectedTo);

		if (!state.method_11654(HAS_BE))
			return list;
		if (linear)
			return list;

		class_2586 blockEntity = world.method_8321(pos);
		if (!(blockEntity instanceof TrackBlockEntity trackBE))
			return list;

		Map<class_2338, BezierConnection> connections = trackBE.getConnections();
		connections.forEach((connectedPos, bc) -> ITrackBlock.addToListIfConnected(connectedTo, list,
			(d, b) -> d == 1 ? class_243.method_24954(bc.tePositions.get(b)) : bc.starts.get(b), bc.normals::get,
			b -> world instanceof class_1937 l ? l.method_27983() : class_1937.field_25179, bc::yOffsetAt, null, bc,
			(b, v) -> ITrackBlock.getMaterialSimple(world, v, bc.getMaterial())));

		if (trackBE.boundLocation == null || !(world instanceof class_3218 level))
			return list;

		class_5321<class_1937> otherDim = trackBE.boundLocation.getFirst();
		class_3218 otherLevel = level.method_8503()
			.method_3847(otherDim);
		if (otherLevel == null)
			return list;
		class_2338 boundPos = trackBE.boundLocation.getSecond();
		class_2680 boundState = otherLevel.method_8320(boundPos);
		if (!AllTags.AllBlockTags.TRACKS.matches(boundState))
			return list;

		class_243 center = class_243.method_24955(pos)
			.method_1031(0, getElevationAtCenter(world, pos, state), 0);
		class_243 boundCenter = class_243.method_24955(boundPos)
			.method_1031(0, getElevationAtCenter(otherLevel, boundPos, boundState), 0);
		TrackShape shape = state.method_11654(TrackBlock.SHAPE);
		TrackShape boundShape = boundState.method_11654(TrackBlock.SHAPE);
		class_243 boundAxis = getTrackAxes(otherLevel, boundPos, boundState).get(0);

		getTrackAxes(world, pos, state).forEach(axis -> {
			ITrackBlock.addToListIfConnected(connectedTo, list, (d, b) -> (b ? axis : boundAxis).method_1021(d)
				.method_1019(b ? center : boundCenter), b -> (b ? shape : boundShape).getNormal(),
				b -> b ? level.method_27983() : otherLevel.method_27983(), v -> 0, axis, null,
				(b, v) -> ITrackBlock.getMaterialSimple(b ? level : otherLevel, v));
		});

		return list;
	}

	public void method_9496(class_2680 pState, class_1937 pLevel, class_2338 pPos, class_5819 pRand) {
		if (!pState.method_11654(SHAPE)
			.isPortal())
			return;
		class_243 v = class_243.method_24954(pPos)
			.method_1023(.125f, 0, .125f);
		CubeParticleData data =
			new CubeParticleData(1, pRand.method_43057(), 1, .0125f + .0625f * pRand.method_43057(), 30, false);
		pLevel.method_8406(data, v.field_1352 + pRand.method_43057() * 1.5f, v.field_1351 + .25f, v.field_1350 + pRand.method_43057() * 1.5f, 0.0D,
			0.04D, 0.0D);
	}

	@Override
	public void method_9536(class_2680 pState, class_1937 pLevel, class_2338 pPos, class_2680 pNewState, boolean pIsMoving) {
		boolean removeBE = false;
		if (pState.method_11654(HAS_BE) && (!pState.method_27852(pNewState.method_26204()) || !pNewState.method_11654(HAS_BE))) {
			class_2586 blockEntity = pLevel.method_8321(pPos);
			if (blockEntity instanceof TrackBlockEntity tbe && !pLevel.field_9236) {
				tbe.cancelDrops |= pNewState.method_26204() == this;
				tbe.removeInboundConnections(true);
			}
			removeBE = true;
		}

		if (pNewState.method_26204() != this || pState.method_11657(HAS_BE, true) != pNewState.method_11657(HAS_BE, true))
			TrackPropagator.onRailRemoved(pLevel, pPos, pState);
		if (removeBE)
			pLevel.method_8544(pPos);
		if (!pLevel.field_9236)
			updateGirders(pState, pLevel, pPos, pLevel.method_8397());
	}

	@Override
	public class_1269 method_9534(class_2680 state, class_1937 world, class_2338 pos, class_1657 player, class_1268 hand,
		class_3965 hit) {

		if (world.field_9236)
			return class_1269.field_5812;
		for (Entry<class_2338, class_3341> entry : StationBlockEntity.assemblyAreas.get(world)
			.entrySet()) {
			if (!entry.getValue()
				.method_14662(pos))
				continue;
			if (world.method_8321(entry.getKey()) instanceof StationBlockEntity station)
				if (station.trackClicked(player, hand, this, state, pos))
					return class_1269.field_5812;
		}

		return class_1269.field_5811;
	}

	private void updateGirders(class_2680 pState, class_1937 pLevel, class_2338 pPos, class_6756<class_2248> blockTicks) {
		for (class_243 vec3 : getTrackAxes(pLevel, pPos, pState)) {
			if (vec3.method_1033() > 1 || vec3.field_1351 != 0)
				continue;
			for (int side : Iterate.positiveAndNegative) {
				class_2338 girderPos = pPos.method_10074()
					.method_10081(class_2338.method_49637(vec3.field_1350 * side, 0, vec3.field_1352 * side));
				class_2680 girderState = pLevel.method_8320(girderPos);
				if (girderState.method_26204() instanceof GirderBlock girderBlock
					&& !blockTicks.method_8674(girderPos, girderBlock))
					pLevel.method_39279(girderPos, girderBlock, 1);
			}
		}
	}

	@Override
	public boolean method_9558(class_2680 state, class_4538 reader, class_2338 pos) {
		return reader.method_8320(pos.method_10074())
			.method_26204() != this;
	}

	@Override
	public class_265 method_9530(class_2680 state, class_1922 p_60556_, class_2338 p_60557_, class_3726 p_60558_) {
		return getFullShape(state);
	}

	@Override
	public class_265 method_9584(class_2680 state, class_1922 pLevel, class_2338 pPos) {
		return getFullShape(state);
	}

	private class_265 getFullShape(class_2680 state) {
		switch (state.method_11654(SHAPE)) {
		case AE:
			return TRACK_ASC.get(class_2350.field_11034);
		case AW:
			return TRACK_ASC.get(class_2350.field_11039);
		case AN:
			return TRACK_ASC.get(class_2350.field_11043);
		case AS:
			return TRACK_ASC.get(class_2350.field_11035);
		case CR_D:
			return TRACK_CROSS_DIAG;
		case CR_NDX:
			return TRACK_CROSS_ORTHO_DIAG.get(class_2350.field_11035);
		case CR_NDZ:
			return TRACK_CROSS_DIAG_ORTHO.get(class_2350.field_11035);
		case CR_O:
			return TRACK_CROSS;
		case CR_PDX:
			return TRACK_CROSS_DIAG_ORTHO.get(class_2350.field_11034);
		case CR_PDZ:
			return TRACK_CROSS_ORTHO_DIAG.get(class_2350.field_11034);
		case ND:
			return TRACK_DIAG.get(class_2350.field_11035);
		case PD:
			return TRACK_DIAG.get(class_2350.field_11034);
		case XO:
			return TRACK_ORTHO.get(class_2350.field_11034);
		case ZO:
			return TRACK_ORTHO.get(class_2350.field_11035);
		case TE:
			return TRACK_ORTHO_LONG.get(class_2350.field_11034);
		case TW:
			return TRACK_ORTHO_LONG.get(class_2350.field_11039);
		case TS:
			return TRACK_ORTHO_LONG.get(class_2350.field_11035);
		case TN:
			return TRACK_ORTHO_LONG.get(class_2350.field_11043);
		case NONE:
		default:
		}
		return AllShapes.TRACK_FALLBACK;
	}

	@Override
	public class_265 method_9549(class_2680 pState, class_1922 pLevel, class_2338 pPos,
		class_3726 pContext) {
		switch (pState.method_11654(SHAPE)) {
		case AE, AW, AN, AS:
			return class_259.method_1073();
		default:
			return AllShapes.TRACK_COLLISION;
		}
	}

	@Override
	public class_2586 method_10123(class_2338 p_153215_, class_2680 state) {
		if (!state.method_11654(HAS_BE))
			return null;
		return AllBlockEntityTypes.TRACK.create(p_153215_, state);
	}

	@Override
	public Class<TrackBlockEntity> getBlockEntityClass() {
		return TrackBlockEntity.class;
	}

	@Override
	public class_2591<? extends TrackBlockEntity> getBlockEntityType() {
		return AllBlockEntityTypes.TRACK.get();
	}

	@Override
	public class_243 getUpNormal(class_1922 world, class_2338 pos, class_2680 state) {
		return state.method_11654(SHAPE)
			.getNormal();
	}

	@Override
	public List<class_243> getTrackAxes(class_1922 world, class_2338 pos, class_2680 state) {
		return state.method_11654(SHAPE)
			.getAxes();
	}

	@Override
	public class_243 getCurveStart(class_1922 world, class_2338 pos, class_2680 state, class_243 axis) {
		boolean vertical = axis.field_1351 != 0;
		return VecHelper.getCenterOf(pos)
			.method_1031(0, (vertical ? 0 : -.5f), 0)
			.method_1019(axis.method_1021(.5));
	}

	@Override
	public class_1269 onWrenched(class_2680 state, class_1838 context) {
		return class_1269.field_5812;
	}

	@Override
	public class_1269 onSneakWrenched(class_2680 state, class_1838 context) {
		class_1657 player = context.method_8036();
		class_1937 level = context.method_8045();
		if (!level.field_9236 && !player.method_7337() && state.method_11654(HAS_BE)) {
			class_2586 blockEntity = level.method_8321(context.method_8037());
			if (blockEntity instanceof TrackBlockEntity trackBE) {
				trackBE.cancelDrops = true;
				trackBE.connections.values()
					.forEach(bc -> bc.addItemsToPlayer(player));
			}
		}

		return IWrenchable.super.onSneakWrenched(state, context);
	}

	@Override
	public class_2680 overlay(class_1922 world, class_2338 pos, class_2680 existing, class_2680 placed) {
		if (placed.method_26204() != this)
			return existing;

		TrackShape existingShape = existing.method_11654(SHAPE);
		TrackShape placedShape = placed.method_11654(SHAPE);
		TrackShape combinedShape = null;

		for (boolean flip : Iterate.trueAndFalse) {
			TrackShape s1 = flip ? existingShape : placedShape;
			TrackShape s2 = flip ? placedShape : existingShape;
			if (s1 == TrackShape.XO && s2 == TrackShape.ZO)
				combinedShape = TrackShape.CR_O;
			if (s1 == TrackShape.PD && s2 == TrackShape.ND)
				combinedShape = TrackShape.CR_D;
			if (s1 == TrackShape.XO && s2 == TrackShape.PD)
				combinedShape = TrackShape.CR_PDX;
			if (s1 == TrackShape.ZO && s2 == TrackShape.PD)
				combinedShape = TrackShape.CR_PDZ;
			if (s1 == TrackShape.XO && s2 == TrackShape.ND)
				combinedShape = TrackShape.CR_NDX;
			if (s1 == TrackShape.ZO && s2 == TrackShape.ND)
				combinedShape = TrackShape.CR_NDZ;
		}

		if (combinedShape != null)
			existing = existing.method_11657(SHAPE, combinedShape);
		return existing;
	}

	@Override
	public class_2680 method_9598(class_2680 state, class_2470 pRotation) {
		return state.method_11657(SHAPE, state.method_11654(SHAPE)
			.rotate(pRotation));
	}

	@Override
	public class_2680 method_9569(class_2680 state, class_2415 pMirror) {
		return state.method_11657(SHAPE, state.method_11654(SHAPE)
			.mirror(pMirror));
	}

	@Override
	public class_2680 getBogeyAnchor(class_1922 world, class_2338 pos, class_2680 state) {
		return AllBlocks.SMALL_BOGEY.getDefaultState()
			.method_11657(class_2741.field_12529, state.method_11654(SHAPE) == TrackShape.XO ? class_2351.field_11048 : class_2351.field_11051);
	}

	@Override
	@Environment(EnvType.CLIENT)
	public PartialModel prepareAssemblyOverlay(class_1922 world, class_2338 pos, class_2680 state, class_2350 direction,
		class_4587 ms) {
		TransformStack.cast(ms)
			.rotateCentered(class_2350.field_11036, AngleHelper.rad(AngleHelper.horizontalAngle(direction)));
		return AllPartialModels.TRACK_ASSEMBLING_OVERLAY;
	}

	@Override
	@Environment(EnvType.CLIENT)
	public PartialModel prepareTrackOverlay(class_1922 world, class_2338 pos, class_2680 state,
		BezierTrackPointLocation bezierPoint, class_2352 direction, class_4587 ms, RenderedTrackOverlayType type) {
		TransformStack msr = TransformStack.cast(ms);

		class_243 axis = null;
		class_243 diff = null;
		class_243 normal = null;
		class_243 offset = null;

		if (bezierPoint != null && world.method_8321(pos) instanceof TrackBlockEntity trackBE) {
			BezierConnection bc = trackBE.connections.get(bezierPoint.curveTarget());
			if (bc != null) {
				double length = class_3532.method_15357(bc.getLength() * 2);
				int seg = bezierPoint.segment() + 1;
				double t = seg / length;
				double tpre = (seg - 1) / length;
				double tpost = (seg + 1) / length;

				offset = bc.getPosition(t);
				normal = bc.getNormal(t);
				diff = bc.getPosition(tpost)
					.method_1020(bc.getPosition(tpre))
					.method_1029();

				msr.translate(offset.method_1020(class_243.method_24955(pos)));
				msr.translate(0, -4 / 16f, 0);
			} else
				return null;
		}

		if (normal == null) {
			axis = state.method_11654(SHAPE)
				.getAxes()
				.get(0);
			diff = axis.method_1021(direction.method_10181())
				.method_1029();
			normal = getUpNormal(world, pos, state);
		}

		class_243 angles = TrackRenderer.getModelAngles(normal, diff);

		msr.centre()
			.rotateYRadians(angles.field_1351)
			.rotateXRadians(angles.field_1352)
			.unCentre();

		if (axis != null)
			msr.translate(0, axis.field_1351 != 0 ? 7 / 16f : 0, axis.field_1351 != 0 ? direction.method_10181() * 2.5f / 16f : 0);
		else {
			msr.translate(0, 4 / 16f, 0);
			if (direction == class_2352.field_11060)
				msr.rotateCentered(class_2350.field_11036, class_3532.field_29844);
		}

		if (bezierPoint == null && world.method_8321(pos) instanceof TrackBlockEntity trackTE
			&& trackTE.isTilted()) {
			double yOffset = 0;
			for (BezierConnection bc : trackTE.connections.values())
				yOffset += bc.starts.getFirst().field_1351 - pos.method_10264();
			msr.centre()
				.rotateX(-direction.method_10181() * trackTE.tilt.smoothingAngle.get())
				.unCentre()
				.translate(0, yOffset / 2, 0);
		}

		return switch (type) {
		case DUAL_SIGNAL -> AllPartialModels.TRACK_SIGNAL_DUAL_OVERLAY;
		case OBSERVER -> AllPartialModels.TRACK_OBSERVER_OVERLAY;
		case SIGNAL -> AllPartialModels.TRACK_SIGNAL_OVERLAY;
		case STATION -> AllPartialModels.TRACK_STATION_OVERLAY;
		};
	}

	@Override
	public boolean trackEquals(class_2680 state1, class_2680 state2) {
		return state1.method_26204() == this && state2.method_26204() == this
			&& state1.method_11657(HAS_BE, false) == state2.method_11657(HAS_BE, false);
	}

	@Override
	public ItemRequirement getRequiredItems(class_2680 state, class_2586 be) {
		int sameTypeTrackAmount = 1;
		Object2IntMap<TrackMaterial> otherTrackAmounts = new Object2IntArrayMap<>();
		int girderAmount = 0;

		if (be instanceof TrackBlockEntity track) {
			for (BezierConnection bezierConnection : track.getConnections()
				.values()) {
				if (!bezierConnection.isPrimary())
					continue;
				TrackMaterial material = bezierConnection.getMaterial();
				if (material == getMaterial()) {
					sameTypeTrackAmount += bezierConnection.getTrackItemCost();
				} else {
					otherTrackAmounts.put(material, otherTrackAmounts.getOrDefault(material, 0) + 1);
				}
				girderAmount += bezierConnection.getGirderItemCost();
			}
		}

		List<class_1799> stacks = new ArrayList<>();
		while (sameTypeTrackAmount > 0) {
			stacks.add(new class_1799(state.method_26204(), Math.min(sameTypeTrackAmount, 64)));
			sameTypeTrackAmount -= 64;
		}
		for (TrackMaterial material : otherTrackAmounts.keySet()) {
			int amt = otherTrackAmounts.getOrDefault(material, 0);
			while (amt > 0) {
				stacks.add(material.asStack(Math.min(amt, 64)));
				amt -= 64;
			}
		}
		while (girderAmount > 0) {
			stacks.add(AllBlocks.METAL_GIRDER.asStack(Math.min(girderAmount, 64)));
			girderAmount -= 64;
		}

		return new ItemRequirement(ItemUseType.CONSUME, stacks);
	}

	@Override
	public TrackMaterial getMaterial() {
		return material;
	}

//	public static class RenderProperties extends ReducedDestroyEffects implements MultiPosDestructionHandler {
		@Override
		@Nullable
		@Environment(EnvType.CLIENT)
		public Set<class_2338> getExtraPositions(class_638 level, class_2338 pos, class_2680 blockState,
											   int progress) {
			class_2586 blockEntity = level.method_8321(pos);
			if (blockEntity instanceof TrackBlockEntity track) {
				return new HashSet<>(track.connections.keySet());
			}
			return null;
		}
//	}

}
