package com.simibubi.create.content.trains.track;

import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.Set;

import org.jetbrains.annotations.Nullable;

import com.jozufozu.flywheel.backend.instancing.InstancedRenderDispatcher;
import com.simibubi.create.AllBlocks;
import com.simibubi.create.AllPackets;
import com.simibubi.create.AllTags;
import com.simibubi.create.content.contraptions.ITransformableBlockEntity;
import com.simibubi.create.content.contraptions.StructureTransform;
import com.simibubi.create.content.trains.graph.TrackNodeLocation;
import com.simibubi.create.foundation.block.ProperWaterloggedBlock;
import com.simibubi.create.foundation.blockEntity.IMergeableBE;
import com.simibubi.create.foundation.blockEntity.RemoveBlockEntityPacket;
import com.simibubi.create.foundation.blockEntity.SmartBlockEntity;
import com.simibubi.create.foundation.blockEntity.behaviour.BlockEntityBehaviour;
import com.simibubi.create.foundation.utility.Pair;
import com.simibubi.create.foundation.utility.VecHelper;
import com.tterrag.registrate.fabric.EnvExecutor;

import net.fabricmc.api.EnvType;
import net.fabricmc.api.Environment;
import net.fabricmc.fabric.api.rendering.data.v1.RenderAttachmentBlockEntity;
import net.minecraft.class_1937;
import net.minecraft.class_2338;
import net.minecraft.class_2350.class_2351;
import net.minecraft.class_238;
import net.minecraft.class_243;
import net.minecraft.class_2487;
import net.minecraft.class_2499;
import net.minecraft.class_2512;
import net.minecraft.class_2520;
import net.minecraft.class_2586;
import net.minecraft.class_2591;
import net.minecraft.class_2680;
import net.minecraft.class_2960;
import net.minecraft.class_3218;
import net.minecraft.class_3532;
import net.minecraft.class_3610;
import net.minecraft.class_3612;
import net.minecraft.class_5321;
import net.minecraft.class_7924;

public class TrackBlockEntity extends SmartBlockEntity implements ITransformableBlockEntity, IMergeableBE, RenderAttachmentBlockEntity {

	Map<class_2338, BezierConnection> connections;
	boolean cancelDrops;

	public Pair<class_5321<class_1937>, class_2338> boundLocation;
	public TrackBlockEntityTilt tilt;

	public TrackBlockEntity(class_2591<?> type, class_2338 pos, class_2680 state) {
		super(type, pos, state);
		connections = new HashMap<>();
		setLazyTickRate(100);
		tilt = new TrackBlockEntityTilt(this);
	}

	public Map<class_2338, BezierConnection> getConnections() {
		return connections;
	}

	@Override
	public void initialize() {
		super.initialize();
		if (!field_11863.field_9236 && hasInteractableConnections())
			registerToCurveInteraction();
	}

	@Override
	public void tick() {
		super.tick();
		tilt.undoSmoothing();
	}

	@Override
	public void lazyTick() {
		for (BezierConnection connection : connections.values())
			if (connection.isPrimary())
				manageFakeTracksAlong(connection, false);
	}

	public void validateConnections() {
		Set<class_2338> invalid = new HashSet<>();

		for (Entry<class_2338, BezierConnection> entry : connections.entrySet()) {
			class_2338 key = entry.getKey();
			BezierConnection bc = entry.getValue();

			if (!key.equals(bc.getKey()) || !field_11867.equals(bc.tePositions.getFirst())) {
				invalid.add(key);
				continue;
			}

			class_2680 blockState = field_11863.method_8320(key);
			if (blockState.method_26204()instanceof ITrackBlock trackBlock && !blockState.method_11654(TrackBlock.HAS_BE))
				for (class_243 v : trackBlock.getTrackAxes(field_11863, key, blockState)) {
					class_243 bcEndAxis = bc.axes.getSecond();
					if (v.method_1022(bcEndAxis) < 1 / 1024f || v.method_1022(bcEndAxis.method_1021(-1)) < 1 / 1024f)
						field_11863.method_8652(key, blockState.method_11657(TrackBlock.HAS_BE, true), 3);
				}

			class_2586 blockEntity = field_11863.method_8321(key);
			if (!(blockEntity instanceof TrackBlockEntity trackBE) || blockEntity.method_11015()) {
				invalid.add(key);
				continue;
			}

			if (!trackBE.connections.containsKey(field_11867)) {
				trackBE.addConnection(bc.secondary());
				trackBE.tilt.tryApplySmoothing();
			}
		}

		for (class_2338 blockPos : invalid)
			removeConnection(blockPos);
	}

	public void addConnection(BezierConnection connection) {
		// don't replace existing connections with different materials
		if (connections.containsKey(connection.getKey()) && connection.equalsSansMaterial(connections.get(connection.getKey())))
			return;
		connections.put(connection.getKey(), connection);
		field_11863.method_39279(field_11867, method_11010().method_26204(), 1);
		notifyUpdate();

		if (connection.isPrimary())
			manageFakeTracksAlong(connection, false);
	}

	public void removeConnection(class_2338 target) {
		if (isTilted())
			tilt.captureSmoothingHandles();

		BezierConnection removed = connections.remove(target);
		notifyUpdate();

		if (removed != null)
			manageFakeTracksAlong(removed, true);

		if (!connections.isEmpty() || method_11010().method_28500(TrackBlock.SHAPE)
			.orElse(TrackShape.NONE)
			.isPortal())
			return;

		class_2680 blockState = field_11863.method_8320(field_11867);
		if (blockState.method_28498(TrackBlock.HAS_BE))
			field_11863.method_8501(field_11867, blockState.method_11657(TrackBlock.HAS_BE, false));
		if (field_11863 instanceof class_3218 serverLevel)
			AllPackets.getChannel().sendToClientsTracking(new RemoveBlockEntityPacket(field_11867), serverLevel, field_11867);
	}

	public void removeInboundConnections(boolean dropAndDiscard) {
		for (BezierConnection bezierConnection : connections.values()) {
			if (!(field_11863.method_8321(bezierConnection.getKey())instanceof TrackBlockEntity tbe))
				return;
			tbe.removeConnection(bezierConnection.tePositions.getFirst());
			if (!dropAndDiscard)
				continue;
			if (!cancelDrops)
				bezierConnection.spawnItems(field_11863);
			bezierConnection.spawnDestroyParticles(field_11863);
		}
		if (dropAndDiscard)
			AllPackets.getChannel().sendToClientsTracking(new RemoveBlockEntityPacket(field_11867), this);
	}

	public void bind(class_5321<class_1937> boundDimension, class_2338 boundLocation) {
		this.boundLocation = Pair.of(boundDimension, boundLocation);
		method_5431();
	}

	public boolean isTilted() {
		return tilt.smoothingAngle.isPresent();
	}

	@Override
	public void writeSafe(class_2487 tag) {
		super.writeSafe(tag);
		writeTurns(tag, true);
	}

	@Override
	protected void write(class_2487 tag, boolean clientPacket) {
		super.write(tag, clientPacket);
		writeTurns(tag, false);
		if (isTilted())
			tag.method_10549("Smoothing", tilt.smoothingAngle.get());
		if (boundLocation == null)
			return;
		tag.method_10566("BoundLocation", class_2512.method_10692(boundLocation.getSecond()));
		tag.method_10582("BoundDimension", boundLocation.getFirst()
			.method_29177()
			.toString());
	}

	private void writeTurns(class_2487 tag, boolean restored) {
		class_2499 listTag = new class_2499();
		for (BezierConnection bezierConnection : connections.values())
			listTag.add((restored ? tilt.restoreToOriginalCurve(bezierConnection.clone()) : bezierConnection)
				.write(field_11867));
		tag.method_10566("Connections", listTag);
	}

	@Override
	protected void read(class_2487 tag, boolean clientPacket) {
		super.read(tag, clientPacket);
		connections.clear();
		for (class_2520 t : tag.method_10554("Connections", class_2520.field_33260)) {
			if (!(t instanceof class_2487))
				return;
			BezierConnection connection = new BezierConnection((class_2487) t, field_11867);
			connections.put(connection.getKey(), connection);
		}

		boolean smoothingPreviously = tilt.smoothingAngle.isPresent();
		tilt.smoothingAngle = Optional.ofNullable(tag.method_10545("Smoothing") ? tag.method_10574("Smoothing") : null);
		if (smoothingPreviously != tilt.smoothingAngle.isPresent() && clientPacket) {
			// fabric: no need for requestModelDataUpdate
			field_11863.method_8413(field_11867, method_11010(), method_11010(), 16);
		}

		EnvExecutor.runWhenOn(EnvType.CLIENT, () -> () -> InstancedRenderDispatcher.enqueueUpdate(this));

		if (hasInteractableConnections())
			registerToCurveInteraction();
		else
			removeFromCurveInteraction();

		if (tag.method_10545("BoundLocation"))
			boundLocation = Pair.of(
				class_5321.method_29179(class_7924.field_41223, new class_2960(tag.method_10558("BoundDimension"))),
				class_2512.method_10691(tag.method_10562("BoundLocation")));
	}

	@Override
	@Environment(EnvType.CLIENT)
	public class_238 getRenderBoundingBox() {
		return INFINITE_EXTENT_AABB;
	}

	@Override
	public void addBehaviours(List<BlockEntityBehaviour> behaviours) {}

	@Override
	public void accept(class_2586 other) {
		if (other instanceof TrackBlockEntity track)
			connections.putAll(track.connections);
		validateConnections();
		field_11863.method_39279(field_11867, method_11010().method_26204(), 1);
	}

	public boolean hasInteractableConnections() {
		for (BezierConnection connection : connections.values())
			if (connection.isPrimary())
				return true;
		return false;
	}

	@Override
	public void transform(StructureTransform transform) {
		Map<class_2338, BezierConnection> restoredConnections = new HashMap<>();
		for (Entry<class_2338, BezierConnection> entry : connections.entrySet())
			restoredConnections.put(entry.getKey(),
				tilt.restoreToOriginalCurve(tilt.restoreToOriginalCurve(entry.getValue()
					.secondary())
					.secondary()));
		connections = restoredConnections;
		tilt.smoothingAngle = Optional.empty();

		if (transform.rotationAxis != class_2351.field_11052)
			return;

		Map<class_2338, BezierConnection> transformedConnections = new HashMap<>();
		for (Entry<class_2338, BezierConnection> entry : connections.entrySet()) {
			BezierConnection newConnection = entry.getValue();
			newConnection.normals.replace(transform::applyWithoutOffsetUncentered);
			newConnection.axes.replace(transform::applyWithoutOffsetUncentered);

			class_2338 diff = newConnection.tePositions.getSecond()
				.method_10059(newConnection.tePositions.getFirst());
			newConnection.tePositions
				.setSecond(class_2338.method_49638(class_243.method_24953(newConnection.tePositions.getFirst())
					.method_1019(transform.applyWithoutOffsetUncentered(class_243.method_24954(diff)))));

			class_243 beVec = class_243.method_24954(field_11867);
			class_243 teCenterVec = beVec.method_1031(0.5, 0.5, 0.5);
			class_243 start = newConnection.starts.getFirst();
			class_243 startToBE = start.method_1020(teCenterVec);
			class_243 endToStart = newConnection.starts.getSecond()
				.method_1020(start);
			startToBE = transform.applyWithoutOffsetUncentered(startToBE)
				.method_1019(teCenterVec);
			endToStart = transform.applyWithoutOffsetUncentered(endToStart)
				.method_1019(startToBE);

			newConnection.starts.setFirst(new TrackNodeLocation(startToBE).getLocation());
			newConnection.starts.setSecond(new TrackNodeLocation(endToStart).getLocation());

			class_2338 newTarget = newConnection.getKey();
			transformedConnections.put(newTarget, newConnection);
		}

		connections = transformedConnections;
	}

	@Override
	public void invalidate() {
		super.invalidate();
		if (field_11863.field_9236)
			removeFromCurveInteraction();
	}

	@Override
	public void remove() {
		super.remove();

		for (BezierConnection connection : connections.values())
			manageFakeTracksAlong(connection, true);

		if (boundLocation != null && field_11863 instanceof class_3218) {
			class_3218 otherLevel = field_11863.method_8503()
				.method_3847(boundLocation.getFirst());
			if (otherLevel == null)
				return;
			if (AllTags.AllBlockTags.TRACKS.matches(otherLevel.method_8320(boundLocation.getSecond())))
				otherLevel.method_22352(boundLocation.getSecond(), false);
		}
	}

	private void registerToCurveInteraction() {
		EnvExecutor.runWhenOn(EnvType.CLIENT, () -> this::registerToCurveInteractionUnsafe);
	}

	private void removeFromCurveInteraction() {
		EnvExecutor.runWhenOn(EnvType.CLIENT, () -> this::removeFromCurveInteractionUnsafe);
	}

	@Override
	@Nullable
	public Double getRenderAttachmentData() {
		if (!isTilted())
			return null;
		return tilt.smoothingAngle.get();
	}

	@Environment(EnvType.CLIENT)
	private void registerToCurveInteractionUnsafe() {
		TrackBlockOutline.TRACKS_WITH_TURNS.get(field_11863)
			.put(field_11867, this);
	}

	@Environment(EnvType.CLIENT)
	private void removeFromCurveInteractionUnsafe() {
		TrackBlockOutline.TRACKS_WITH_TURNS.get(field_11863)
			.remove(field_11867);
	}

	public void manageFakeTracksAlong(BezierConnection bc, boolean remove) {
		Map<Pair<Integer, Integer>, Double> yLevels = new HashMap<>();
		class_2338 tePosition = bc.tePositions.getFirst();
		class_243 end1 = bc.starts.getFirst()
			.method_1020(class_243.method_24954(tePosition))
			.method_1031(0, 3 / 16f, 0);
		class_243 end2 = bc.starts.getSecond()
			.method_1020(class_243.method_24954(tePosition))
			.method_1031(0, 3 / 16f, 0);
		class_243 axis1 = bc.axes.getFirst();
		class_243 axis2 = bc.axes.getSecond();

		double handleLength = bc.getHandleLength();

		class_243 finish1 = axis1.method_1021(handleLength)
			.method_1019(end1);
		class_243 finish2 = axis2.method_1021(handleLength)
			.method_1019(end2);

		class_243 faceNormal1 = bc.normals.getFirst();
		class_243 faceNormal2 = bc.normals.getSecond();

		int segCount = bc.getSegmentCount();
		float[] lut = bc.getStepLUT();

		for (int i = 0; i < segCount; i++) {
			float t = i == segCount ? 1 : i * lut[i] / segCount;
			t += 0.5f / segCount;

			class_243 result = VecHelper.bezier(end1, end2, finish1, finish2, t);
			class_243 derivative = VecHelper.bezierDerivative(end1, end2, finish1, finish2, t)
				.method_1029();
			class_243 faceNormal =
				faceNormal1.equals(faceNormal2) ? faceNormal1 : VecHelper.slerp(t, faceNormal1, faceNormal2);
			class_243 normal = faceNormal.method_1036(derivative)
				.method_1029();
			class_243 below = result.method_1019(faceNormal.method_1021(-.25f));
			class_243 rail1 = below.method_1019(normal.method_1021(.05f));
			class_243 rail2 = below.method_1020(normal.method_1021(.05f));
			class_243 railMiddle = rail1.method_1019(rail2)
				.method_1021(.5);

			for (class_243 vec : new class_243[] { railMiddle }) {
				class_2338 pos = class_2338.method_49638(vec);
				Pair<Integer, Integer> key = Pair.of(pos.method_10263(), pos.method_10260());
				if (!yLevels.containsKey(key) || yLevels.get(key) > vec.field_1351)
					yLevels.put(key, vec.field_1351);
			}
		}

		for (Entry<Pair<Integer, Integer>, Double> entry : yLevels.entrySet()) {
			double yValue = entry.getValue();
			int floor = class_3532.method_15357(yValue);
			class_2338 targetPos = new class_2338(entry.getKey()
				.getFirst(), floor,
				entry.getKey()
					.getSecond());
			targetPos = targetPos.method_10081(tePosition)
				.method_10086(1);

			class_2680 stateAtPos = field_11863.method_8320(targetPos);
			boolean present = AllBlocks.FAKE_TRACK.has(stateAtPos);

			if (remove) {
				if (present)
					field_11863.method_8650(targetPos, false);
				continue;
			}

			class_3610 fluidState = stateAtPos.method_26227();
			if (!fluidState.method_15769() && !fluidState.method_33659(class_3612.field_15910))
				continue;

			if (!present && stateAtPos.method_45474())
				field_11863.method_8652(targetPos,
					ProperWaterloggedBlock.withWater(field_11863, AllBlocks.FAKE_TRACK.getDefaultState(), targetPos), 3);
			FakeTrackBlock.keepAlive(field_11863, targetPos);
		}
	}

}
