package com.simibubi.create.content.trains.track;

import java.util.ArrayList;
import java.util.List;
import java.util.Objects;

import javax.annotation.Nullable;
import net.minecraft.class_1921;
import net.minecraft.class_2338;
import net.minecraft.class_4587;
import net.minecraft.class_4587.class_4665;
import I;
import com.jozufozu.flywheel.api.Material;
import com.jozufozu.flywheel.api.MaterialManager;
import com.jozufozu.flywheel.backend.instancing.blockentity.BlockEntityInstance;
import com.jozufozu.flywheel.core.Materials;
import com.jozufozu.flywheel.core.materials.model.ModelData;
import com.jozufozu.flywheel.light.LightUpdater;
import com.jozufozu.flywheel.util.box.GridAlignedBB;
import com.jozufozu.flywheel.util.box.ImmutableBox;
import com.jozufozu.flywheel.util.transform.TransformStack;
import com.simibubi.create.AllPartialModels;
import com.simibubi.create.content.trains.track.BezierConnection.GirderAngles;
import com.simibubi.create.content.trains.track.BezierConnection.SegmentAngles;
import com.simibubi.create.foundation.utility.Couple;
import com.simibubi.create.foundation.utility.Iterate;

public class TrackInstance extends BlockEntityInstance<TrackBlockEntity> {

	private List<BezierTrackInstance> instances;

	public TrackInstance(MaterialManager materialManager, TrackBlockEntity track) {
		super(materialManager, track);

		update();
	}

	@Override
	public void update() {
		if (blockEntity.connections.isEmpty())
			return;

		remove();
		instances = blockEntity.connections.values()
			.stream()
			.map(this::createInstance)
			.filter(Objects::nonNull)
			.toList();
		LightUpdater.get(world)
			.addListener(this);
	}

	@Override
	public ImmutableBox getVolume() {
		List<class_2338> out = new ArrayList<>();
		out.addAll(blockEntity.connections.keySet());
		out.addAll(blockEntity.connections.keySet());
		return GridAlignedBB.containingAll(out);
	}

	@Override
	public void updateLight() {
		if (instances == null)
			return;
		instances.forEach(BezierTrackInstance::updateLight);
	}

	@Nullable
	private BezierTrackInstance createInstance(BezierConnection bc) {
		if (!bc.isPrimary())
			return null;
		return new BezierTrackInstance(bc);
	}

	@Override
	public void remove() {
		if (instances == null)
			return;
		instances.forEach(BezierTrackInstance::delete);
	}

	private class BezierTrackInstance {

		private final ModelData[] ties;
		private final ModelData[] left;
		private final ModelData[] right;
		private final class_2338[] tiesLightPos;
		private final class_2338[] leftLightPos;
		private final class_2338[] rightLightPos;

		private @Nullable GirderInstance girder;

		private BezierTrackInstance(BezierConnection bc) {
			class_2338 tePosition = bc.tePositions.getFirst();
			girder = bc.hasGirder ? new GirderInstance(bc) : null;

			class_4587 pose = new class_4587();
			TransformStack.cast(pose)
				.translate(getInstancePosition());

			var mat = materialManager.cutout(class_1921.method_23579())
				.material(Materials.TRANSFORMED);

			int segCount = bc.getSegmentCount();
			ties = new ModelData[segCount];
			left = new ModelData[segCount];
			right = new ModelData[segCount];
			tiesLightPos = new class_2338[segCount];
			leftLightPos = new class_2338[segCount];
			rightLightPos = new class_2338[segCount];

			TrackMaterial.TrackModelHolder modelHolder = bc.getMaterial().getModelHolder();

			mat.getModel(modelHolder.tie())
				.createInstances(ties);
			mat.getModel(modelHolder.segment_left())
				.createInstances(left);
			mat.getModel(modelHolder.segment_right())
				.createInstances(right);

			SegmentAngles[] segments = bc.getBakedSegments();
			for (int i = 1; i < segments.length; i++) {
				SegmentAngles segment = segments[i];
				var modelIndex = i - 1;

				ties[modelIndex].setTransform(pose)
					.mulPose(segment.tieTransform.method_23761())
					.mulNormal(segment.tieTransform.method_23762());
				tiesLightPos[modelIndex] = segment.lightPosition.method_10081(tePosition);

				for (boolean first : Iterate.trueAndFalse) {
					class_4665 transform = segment.railTransforms.get(first);
					(first ? this.left : this.right)[modelIndex].setTransform(pose)
						.mulPose(transform.method_23761())
						.mulNormal(transform.method_23762());
					(first ? leftLightPos : rightLightPos)[modelIndex] = segment.lightPosition.method_10081(tePosition);
				}
			}

			updateLight();
		}

		void delete() {
			for (ModelData d : ties)
				d.delete();
			for (ModelData d : left)
				d.delete();
			for (ModelData d : right)
				d.delete();
			if (girder != null)
				girder.delete();
		}

		void updateLight() {
			for (int i = 0; i < ties.length; i++)
				ties[i].updateLight(world, tiesLightPos[i]);
			for (int i = 0; i < left.length; i++)
				left[i].updateLight(world, leftLightPos[i]);
			for (int i = 0; i < right.length; i++)
				right[i].updateLight(world, rightLightPos[i]);
			if (girder != null)
				girder.updateLight();
		}

		private class GirderInstance {

			private final Couple<ModelData[]> beams;
			private final Couple<Couple<ModelData[]>> beamCaps;
			private final class_2338[] lightPos;

			private GirderInstance(BezierConnection bc) {
				class_2338 tePosition = bc.tePositions.getFirst();
				class_4587 pose = new class_4587();
				TransformStack.cast(pose)
					.translate(getInstancePosition())
					.nudge((int) bc.tePositions.getFirst()
						.method_10063());

				var mat = materialManager.cutout(class_1921.method_23579())
					.material(Materials.TRANSFORMED);

				int segCount = bc.getSegmentCount();
				beams = Couple.create(() -> new ModelData[segCount]);
				beamCaps = Couple.create(() -> Couple.create(() -> new ModelData[segCount]));
				lightPos = new class_2338[segCount];
				beams.forEach(mat.getModel(AllPartialModels.GIRDER_SEGMENT_MIDDLE)::createInstances);
				beamCaps.forEachWithContext((c, top) -> c.forEach(mat.getModel(top ? AllPartialModels.GIRDER_SEGMENT_TOP
					: AllPartialModels.GIRDER_SEGMENT_BOTTOM)::createInstances));

				GirderAngles[] bakedGirders = bc.getBakedGirders();
				for (int i = 1; i < bakedGirders.length; i++) {
					GirderAngles segment = bakedGirders[i];
					var modelIndex = i - 1;
					lightPos[modelIndex] = segment.lightPosition.method_10081(tePosition);

					for (boolean first : Iterate.trueAndFalse) {
						class_4665 beamTransform = segment.beams.get(first);
						beams.get(first)[modelIndex].setTransform(pose)
							.mulPose(beamTransform.method_23761())
							.mulNormal(beamTransform.method_23762());
						for (boolean top : Iterate.trueAndFalse) {
							class_4665 beamCapTransform = segment.beamCaps.get(top)
								.get(first);
							beamCaps.get(top)
								.get(first)[modelIndex].setTransform(pose)
									.mulPose(beamCapTransform.method_23761())
									.mulNormal(beamCapTransform.method_23762());
						}
					}
				}

				updateLight();
			}

			void delete() {
				beams.forEach(arr -> {
					for (ModelData d : arr)
						d.delete();
				});
				beamCaps.forEach(c -> c.forEach(arr -> {
					for (ModelData d : arr)
						d.delete();
				}));
			}

			void updateLight() {
				beams.forEach(arr -> {
					for (int i = 0; i < arr.length; i++)
						arr[i].updateLight(world, lightPos[i]);
				});
				beamCaps.forEach(c -> c.forEach(arr -> {
					for (int i = 0; i < arr.length; i++)
						arr[i].updateLight(world, lightPos[i]);
				}));
			}

		}

	}
}
