package com.simibubi.create.content.trains.track;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import com.jozufozu.flywheel.util.Color;
import com.simibubi.create.AllSpecialTextures;
import com.simibubi.create.AllTags;
import com.simibubi.create.CreateClient;
import com.simibubi.create.content.equipment.blueprint.BlueprintOverlayRenderer;
import com.simibubi.create.foundation.block.ProperWaterloggedBlock;
import com.simibubi.create.foundation.utility.AngleHelper;
import com.simibubi.create.foundation.utility.BlockHelper;
import com.simibubi.create.foundation.utility.Couple;
import com.simibubi.create.foundation.utility.Iterate;
import com.simibubi.create.foundation.utility.Lang;
import com.simibubi.create.foundation.utility.Pair;
import com.simibubi.create.foundation.utility.VecHelper;
import com.simibubi.create.foundation.utility.animation.LerpedFloat;
import com.simibubi.create.foundation.utility.animation.LerpedFloat.Chaser;
import com.simibubi.create.infrastructure.config.AllConfigs;

import io.github.fabricators_of_create.porting_lib.transfer.item.ItemHandlerHelper;
import net.fabricmc.api.EnvType;
import net.fabricmc.api.Environment;
import net.minecraft.class_124;
import net.minecraft.class_1268;
import net.minecraft.class_1657;
import net.minecraft.class_1661;
import net.minecraft.class_1747;
import net.minecraft.class_1799;
import net.minecraft.class_1838;
import net.minecraft.class_1937;
import net.minecraft.class_2248;
import net.minecraft.class_2338;
import net.minecraft.class_2343;
import net.minecraft.class_2350;
import net.minecraft.class_2350.class_2351;
import net.minecraft.class_2350.class_2352;
import net.minecraft.class_239;
import net.minecraft.class_239.class_240;
import net.minecraft.class_243;
import net.minecraft.class_2487;
import net.minecraft.class_2512;
import net.minecraft.class_2520;
import net.minecraft.class_2586;
import net.minecraft.class_2680;
import net.minecraft.class_310;
import net.minecraft.class_3532;
import net.minecraft.class_3965;
import net.minecraft.class_746;

public class TrackPlacement {

	public static class PlacementInfo {

		public PlacementInfo(TrackMaterial material) {
			this.trackMaterial = material;
		}

		BezierConnection curve = null;
		boolean valid = false;
		int end1Extent = 0;
		int end2Extent = 0;
		String message = null;

		public int requiredTracks = 0;
		public boolean hasRequiredTracks = false;

		public int requiredPavement = 0;
		public boolean hasRequiredPavement = false;
		public final TrackMaterial trackMaterial;

		// for visualisation
		class_243 end1;
		class_243 end2;
		class_243 normal1;
		class_243 normal2;
		class_243 axis1;
		class_243 axis2;
		class_2338 pos1;
		class_2338 pos2;

		public PlacementInfo withMessage(String message) {
			this.message = "track." + message;
			return this;
		}

		public PlacementInfo tooJumbly() {
			curve = null;
			return this;
		}
	}

	public static PlacementInfo cached;

	static class_2338 hoveringPos;
	static boolean hoveringMaxed;
	static int hoveringAngle;
	static class_1799 lastItem;

	static int extraTipWarmup;

	public static PlacementInfo tryConnect(class_1937 level, class_1657 player, class_2338 pos2, class_2680 state2,
		class_1799 stack, boolean girder, boolean maximiseTurn) {
		class_243 lookVec = player.method_5720();
		int lookAngle = (int) (22.5 + AngleHelper.deg(class_3532.method_15349(lookVec.field_1350, lookVec.field_1352)) % 360) / 8;
		int maxLength = AllConfigs.server().trains.maxTrackPlacementLength.get();

		if (level.field_9236 && cached != null && pos2.equals(hoveringPos) && stack.equals(lastItem)
			&& hoveringMaxed == maximiseTurn && lookAngle == hoveringAngle)
			return cached;

		PlacementInfo info = new PlacementInfo(TrackMaterial.fromItem(stack.method_7909()));
		hoveringMaxed = maximiseTurn;
		hoveringAngle = lookAngle;
		hoveringPos = pos2;
		lastItem = stack;
		cached = info;

		ITrackBlock track = (ITrackBlock) state2.method_26204();
		Pair<class_243, class_2352> nearestTrackAxis = track.getNearestTrackAxis(level, pos2, state2, lookVec);
		class_243 axis2 = nearestTrackAxis.getFirst()
			.method_1021(nearestTrackAxis.getSecond() == class_2352.field_11056 ? -1 : 1);
		class_243 normal2 = track.getUpNormal(level, pos2, state2)
			.method_1029();
		class_243 normedAxis2 = axis2.method_1029();
		class_243 end2 = track.getCurveStart(level, pos2, state2, axis2);

		class_2487 itemTag = stack.method_7969();
		class_2487 selectionTag = itemTag.method_10562("ConnectingFrom");
		class_2338 pos1 = class_2512.method_10691(selectionTag.method_10562("Pos"));
		class_243 axis1 = VecHelper.readNBT(selectionTag.method_10554("Axis", class_2520.field_33256));
		class_243 normedAxis1 = axis1.method_1029();
		class_243 end1 = VecHelper.readNBT(selectionTag.method_10554("End", class_2520.field_33256));
		class_243 normal1 = VecHelper.readNBT(selectionTag.method_10554("Normal", class_2520.field_33256));
		boolean front1 = selectionTag.method_10577("Front");
		class_2680 state1 = level.method_8320(pos1);

		if (level.field_9236) {
			info.end1 = end1;
			info.end2 = end2;
			info.normal1 = normal1;
			info.normal2 = normal2;
			info.axis1 = axis1;
			info.axis2 = axis2;
		}

		if (pos1.equals(pos2))
			return info.withMessage("second_point");
		if (pos1.method_10262(pos2) > maxLength * maxLength)
			return info.withMessage("too_far")
				.tooJumbly();
		if (!state1.method_28498(TrackBlock.HAS_BE))
			return info.withMessage("original_missing");
		if (level.method_8321(pos2) instanceof TrackBlockEntity tbe && tbe.isTilted())
			return info.withMessage("turn_start");

		if (axis1.method_1026(end2.method_1020(end1)) < 0) {
			axis1 = axis1.method_1021(-1);
			normedAxis1 = normedAxis1.method_1021(-1);
			front1 = !front1;
			end1 = track.getCurveStart(level, pos1, state1, axis1);
			if (level.field_9236) {
				info.end1 = end1;
				info.axis1 = axis1;
			}
		}

		double[] intersect = VecHelper.intersect(end1, end2, normedAxis1, normedAxis2, class_2351.field_11052);
		boolean parallel = intersect == null;
		boolean skipCurve = false;

		if ((parallel && normedAxis1.method_1026(normedAxis2) > 0) || (!parallel && (intersect[0] < 0 || intersect[1] < 0))) {
			axis2 = axis2.method_1021(-1);
			normedAxis2 = normedAxis2.method_1021(-1);
			end2 = track.getCurveStart(level, pos2, state2, axis2);
			if (level.field_9236) {
				info.end2 = end2;
				info.axis2 = axis2;
			}
		}

		class_243 cross2 = normedAxis2.method_1036(new class_243(0, 1, 0));

		double a1 = class_3532.method_15349(normedAxis2.field_1350, normedAxis2.field_1352);
		double a2 = class_3532.method_15349(normedAxis1.field_1350, normedAxis1.field_1352);
		double angle = a1 - a2;
		double ascend = end2.method_1020(end1).field_1351;
		double absAscend = Math.abs(ascend);
		boolean slope = !normal1.equals(normal2);

		if (level.field_9236) {
			class_243 offset1 = axis1.method_1021(info.end1Extent);
			class_243 offset2 = axis2.method_1021(info.end2Extent);
			class_2338 targetPos1 = pos1.method_10081(class_2338.method_49638(offset1));
			class_2338 targetPos2 = pos2.method_10081(class_2338.method_49638(offset2));
			info.curve = new BezierConnection(Couple.create(targetPos1, targetPos2),
				Couple.create(end1.method_1019(offset1), end2.method_1019(offset2)), Couple.create(normedAxis1, normedAxis2),
				Couple.create(normal1, normal2), true, girder, TrackMaterial.fromItem(stack.method_7909()));
		}

		// S curve or Straight

		double dist = 0;

		if (parallel) {
			double[] sTest = VecHelper.intersect(end1, end2, normedAxis1, cross2, class_2351.field_11052);
			if (sTest != null) {
				double t = Math.abs(sTest[0]);
				double u = Math.abs(sTest[1]);

				skipCurve = class_3532.method_20390(u, 0);

				if (!skipCurve && sTest[0] < 0)
					return info.withMessage("perpendicular")
						.tooJumbly();

				if (skipCurve) {
					dist = VecHelper.getCenterOf(pos1)
						.method_1022(VecHelper.getCenterOf(pos2));
					info.end1Extent = (int) Math.round((dist + 1) / axis1.method_1033());

				} else {
					if (!class_3532.method_20390(ascend, 0) || normedAxis1.field_1351 != 0)
						return info.withMessage("ascending_s_curve");

					double targetT = u <= 1 ? 3 : u * 2;

					if (t < targetT)
						return info.withMessage("too_sharp");

					// This is for standardising s curve sizes
					if (t > targetT) {
						int correction = (int) ((t - targetT) / axis1.method_1033());
						info.end1Extent = maximiseTurn ? 0 : correction / 2 + (correction % 2);
						info.end2Extent = maximiseTurn ? 0 : correction / 2;
					}
				}
			}
		}

		// Slope

		if (slope) {
			if (!skipCurve)
				return info.withMessage("slope_turn");
			if (class_3532.method_20390(normal1.method_1026(normal2), 0))
				return info.withMessage("opposing_slopes");
			if ((axis1.field_1351 < 0 || axis2.field_1351 > 0) && ascend > 0)
				return info.withMessage("leave_slope_ascending");
			if ((axis1.field_1351 > 0 || axis2.field_1351 < 0) && ascend < 0)
				return info.withMessage("leave_slope_descending");

			skipCurve = false;
			info.end1Extent = 0;
			info.end2Extent = 0;

			class_2351 plane = class_3532.method_20390(axis1.field_1352, 0) ? class_2351.field_11048 : class_2351.field_11051;
			intersect = VecHelper.intersect(end1, end2, normedAxis1, normedAxis2, plane);
			double dist1 = Math.abs(intersect[0] / axis1.method_1033());
			double dist2 = Math.abs(intersect[1] / axis2.method_1033());

			if (dist1 > dist2)
				info.end1Extent = (int) Math.round(dist1 - dist2);
			if (dist2 > dist1)
				info.end2Extent = (int) Math.round(dist2 - dist1);

			double turnSize = Math.min(dist1, dist2);
			if (intersect[0] < 0 || intersect[1] < 0)
				return info.withMessage("too_sharp")
					.tooJumbly();
			if (turnSize < 2)
				return info.withMessage("too_sharp");

			// This is for standardising curve sizes
			if (turnSize > 2 && !maximiseTurn) {
				info.end1Extent += turnSize - 2;
				info.end2Extent += turnSize - 2;
				turnSize = 2;
			}
		}

		// Straight ascend

		if (skipCurve && !class_3532.method_20390(ascend, 0)) {
			int hDistance = info.end1Extent;
			if (axis1.field_1351 == 0 || !class_3532.method_20390(absAscend + 1, dist / axis1.method_1033())) {

				if (axis1.field_1351 != 0 && axis1.field_1351 == -axis2.field_1351)
					return info.withMessage("ascending_s_curve");

				info.end1Extent = 0;
				double minHDistance = Math.max(absAscend < 4 ? absAscend * 4 : absAscend * 3, 6) / axis1.method_1033();
				if (hDistance < minHDistance)
					return info.withMessage("too_steep");
				if (hDistance > minHDistance) {
					int correction = (int) (hDistance - minHDistance);
					info.end1Extent = maximiseTurn ? 0 : correction / 2 + (correction % 2);
					info.end2Extent = maximiseTurn ? 0 : correction / 2;
				}

				skipCurve = false;
			}
		}

		// Turn

		if (!parallel) {
			float absAngle = Math.abs(AngleHelper.deg(angle));
			if (absAngle < 60 || absAngle > 300)
				return info.withMessage("turn_90")
					.tooJumbly();

			intersect = VecHelper.intersect(end1, end2, normedAxis1, normedAxis2, class_2351.field_11052);
			double dist1 = Math.abs(intersect[0]);
			double dist2 = Math.abs(intersect[1]);
			float ex1 = 0;
			float ex2 = 0;

			if (dist1 > dist2)
				ex1 = (float) ((dist1 - dist2) / axis1.method_1033());
			if (dist2 > dist1)
				ex2 = (float) ((dist2 - dist1) / axis2.method_1033());

			double turnSize = Math.min(dist1, dist2) - .1d;
			boolean ninety = (absAngle + .25f) % 90 < 1;

			if (intersect[0] < 0 || intersect[1] < 0)
				return info.withMessage("too_sharp")
					.tooJumbly();

			double minTurnSize = ninety ? 7 : 3.25;
			double turnSizeToFitAscend =
				minTurnSize + (ninety ? Math.max(0, absAscend - 3) * 2f : Math.max(0, absAscend - 1.5f) * 1.5f);

			if (turnSize < minTurnSize)
				return info.withMessage("too_sharp");
			if (turnSize < turnSizeToFitAscend)
				return info.withMessage("too_steep");

			// This is for standardising curve sizes
			if (!maximiseTurn) {
				ex1 += (turnSize - turnSizeToFitAscend) / axis1.method_1033();
				ex2 += (turnSize - turnSizeToFitAscend) / axis2.method_1033();
			}
			info.end1Extent = class_3532.method_15375(ex1);
			info.end2Extent = class_3532.method_15375(ex2);
			turnSize = turnSizeToFitAscend;
		}

		class_243 offset1 = axis1.method_1021(info.end1Extent);
		class_243 offset2 = axis2.method_1021(info.end2Extent);
		class_2338 targetPos1 = pos1.method_10081(class_2338.method_49638(offset1));
		class_2338 targetPos2 = pos2.method_10081(class_2338.method_49638(offset2));

		info.curve = skipCurve ? null
			: new BezierConnection(Couple.create(targetPos1, targetPos2),
				Couple.create(end1.method_1019(offset1), end2.method_1019(offset2)), Couple.create(normedAxis1, normedAxis2),
				Couple.create(normal1, normal2), true, girder, TrackMaterial.fromItem(stack.method_7909()));

		info.valid = true;

		info.pos1 = pos1;
		info.pos2 = pos2;
		info.axis1 = axis1;
		info.axis2 = axis2;

		placeTracks(level, info, state1, state2, targetPos1, targetPos2, true);

		class_1799 offhandItem = player.method_6079()
			.method_7972();
		boolean shouldPave = offhandItem.method_7909() instanceof class_1747;
		if (shouldPave) {
			class_1747 paveItem = (class_1747) offhandItem.method_7909();
			paveTracks(level, info, paveItem, true);
			info.hasRequiredPavement = true;
		}

		info.hasRequiredTracks = true;

		if (!player.method_7337()) {
			for (boolean simulate : Iterate.trueAndFalse) {
				if (level.field_9236 && !simulate)
					break;

				int tracks = info.requiredTracks;
				int pavement = info.requiredPavement;
				int foundTracks = 0;
				int foundPavement = 0;

				class_1661 inv = player.method_31548();
				int size = inv.field_7547.size();
				for (int j = 0; j <= size + 1; j++) {
					int i = j;
					boolean offhand = j == size + 1;
					if (j == size)
						i = inv.field_7545;
					else if (offhand)
						i = 0;
					else if (j == inv.field_7545)
						continue;

					class_1799 stackInSlot = (offhand ? inv.field_7544 : inv.field_7547).get(i);
					boolean isTrack = AllTags.AllBlockTags.TRACKS.matches(stackInSlot) && stackInSlot.method_31574(stack.method_7909());
					if (!isTrack && (!shouldPave || offhandItem.method_7909() != stackInSlot.method_7909()))
						continue;
					if (isTrack ? foundTracks >= tracks : foundPavement >= pavement)
						continue;

					int count = stackInSlot.method_7947();

					if (!simulate) {
						int remainingItems =
							count - Math.min(isTrack ? tracks - foundTracks : pavement - foundPavement, count);
						if (i == inv.field_7545)
							stackInSlot.method_7980(null);
						class_1799 newItem = ItemHandlerHelper.copyStackWithSize(stackInSlot, remainingItems);
						if (offhand)
							player.method_6122(class_1268.field_5810, newItem);
						else
							inv.method_5447(i, newItem);
					}

					if (isTrack)
						foundTracks += count;
					else
						foundPavement += count;
				}

				if (simulate && foundTracks < tracks) {
					info.valid = false;
					info.tooJumbly();
					info.hasRequiredTracks = false;
					return info.withMessage("not_enough_tracks");
				}

				if (simulate && foundPavement < pavement) {
					info.valid = false;
					info.tooJumbly();
					info.hasRequiredPavement = false;
					return info.withMessage("not_enough_pavement");
				}
			}
		}

		if (level.method_8608())
			return info;
		if (shouldPave) {
			class_1747 paveItem = (class_1747) offhandItem.method_7909();
			paveTracks(level, info, paveItem, false);
		}
		return placeTracks(level, info, state1, state2, targetPos1, targetPos2, false);
	}

	private static void paveTracks(class_1937 level, PlacementInfo info, class_1747 blockItem, boolean simulate) {
		class_2248 block = blockItem.method_7711();
		info.requiredPavement = 0;
		if (block == null || block instanceof class_2343 || block.method_9564()
			.method_26220(level, info.pos1)
			.method_1110())
			return;

		Set<class_2338> visited = new HashSet<>();

		for (boolean first : Iterate.trueAndFalse) {
			int extent = (first ? info.end1Extent : info.end2Extent) + (info.curve != null ? 1 : 0);
			class_243 axis = first ? info.axis1 : info.axis2;
			class_2338 pavePos = first ? info.pos1 : info.pos2;
			info.requiredPavement +=
				TrackPaver.paveStraight(level, pavePos.method_10074(), axis, extent, block, simulate, visited);
		}

		if (info.curve != null)
			info.requiredPavement += TrackPaver.paveCurve(level, info.curve, block, simulate, visited);
	}

	private static PlacementInfo placeTracks(class_1937 level, PlacementInfo info, class_2680 state1, class_2680 state2,
		class_2338 targetPos1, class_2338 targetPos2, boolean simulate) {
		info.requiredTracks = 0;

		for (boolean first : Iterate.trueAndFalse) {
			int extent = first ? info.end1Extent : info.end2Extent;
			class_243 axis = first ? info.axis1 : info.axis2;
			class_2338 pos = first ? info.pos1 : info.pos2;
			class_2680 state = first ? state1 : state2;
			if (state.method_28498(TrackBlock.HAS_BE) && !simulate)
				state = state.method_11657(TrackBlock.HAS_BE, false);

			switch (state.method_11654(TrackBlock.SHAPE)) {
			case TE, TW:
				state = state.method_11657(TrackBlock.SHAPE, TrackShape.XO);
				break;
			case TN, TS:
				state = state.method_11657(TrackBlock.SHAPE, TrackShape.ZO);
				break;
			default:
				break;
			}

			for (int i = 0; i < (info.curve != null ? extent + 1 : extent); i++) {
				class_243 offset = axis.method_1021(i);
				class_2338 offsetPos = pos.method_10081(class_2338.method_49638(offset));
				class_2680 stateAtPos = level.method_8320(offsetPos);
				// copy over all shared properties from the shaped state to the correct track material block
				class_2680 toPlace = BlockHelper.copyProperties(state, info.trackMaterial.getBlock().method_9564());

				boolean canPlace = stateAtPos.method_45474();
				if (canPlace)
					info.requiredTracks++;
				if (simulate)
					continue;

				if (stateAtPos.method_26204() instanceof ITrackBlock trackAtPos) {
					toPlace = trackAtPos.overlay(level, offsetPos, stateAtPos, toPlace);
					canPlace = true;
				}

				if (canPlace)
					level.method_8652(offsetPos, ProperWaterloggedBlock.withWater(level, toPlace, offsetPos), 3);
			}
		}

		if (info.curve == null)
			return info;

		if (!simulate) {
			class_2680 onto = info.trackMaterial.getBlock().method_9564();
			class_2680 stateAtPos = level.method_8320(targetPos1);
			level.method_8652(targetPos1, ProperWaterloggedBlock.withWater(level,
					(AllTags.AllBlockTags.TRACKS.matches(stateAtPos) ? stateAtPos : BlockHelper.copyProperties(state1, onto))
							.method_11657(TrackBlock.HAS_BE, true), targetPos1), 3);

			stateAtPos = level.method_8320(targetPos2);
			level.method_8652(targetPos2, ProperWaterloggedBlock.withWater(level,
					(AllTags.AllBlockTags.TRACKS.matches(stateAtPos) ? stateAtPos : BlockHelper.copyProperties(state2, onto))
							.method_11657(TrackBlock.HAS_BE, true), targetPos2), 3);
		}

		class_2586 te1 = level.method_8321(targetPos1);
		class_2586 te2 = level.method_8321(targetPos2);
		int requiredTracksForTurn = (info.curve.getSegmentCount() + 1) / 2;

		if (!(te1 instanceof TrackBlockEntity) || !(te2 instanceof TrackBlockEntity)) {
			info.requiredTracks += requiredTracksForTurn;
			return info;
		}

		TrackBlockEntity tte1 = (TrackBlockEntity) te1;
		TrackBlockEntity tte2 = (TrackBlockEntity) te2;

		if (!tte1.getConnections()
			.containsKey(tte2.method_11016()))
			info.requiredTracks += requiredTracksForTurn;

		if (simulate)
			return info;

		tte1.addConnection(info.curve);
		tte2.addConnection(info.curve.secondary());
		tte1.tilt.tryApplySmoothing();
		tte2.tilt.tryApplySmoothing();
		return info;
	}

	static LerpedFloat animation = LerpedFloat.linear()
		.startWithValue(0);
	static int lastLineCount = 0;

	static class_2338 hintPos;
	static int hintAngle;
	static Couple<List<class_2338>> hints;

	@Environment(EnvType.CLIENT)
	public static void clientTick() {
		class_746 player = class_310.method_1551().field_1724;
		class_1799 stack = player.method_6047();
		class_239 hitResult = class_310.method_1551().field_1765;
		int restoreWarmup = extraTipWarmup;
		extraTipWarmup = 0;

		if (hitResult == null)
			return;
		if (hitResult.method_17783() != class_240.field_1332)
			return;

		class_1268 hand = class_1268.field_5808;
		if (!AllTags.AllBlockTags.TRACKS.matches(stack)) {
			stack = player.method_6079();
			hand = class_1268.field_5810;
			if (!AllTags.AllBlockTags.TRACKS.matches(stack))
				return;
		}

		if (!stack.method_7958())
			return;

		TrackBlockItem blockItem = (TrackBlockItem) stack.method_7909();
		class_1937 level = player.method_37908();
		class_3965 bhr = (class_3965) hitResult;
		class_2338 pos = bhr.method_17777();
		class_2680 hitState = level.method_8320(pos);
		if (!(hitState.method_26204() instanceof TrackBlock) && !hitState.method_45474()) {
			pos = pos.method_10093(bhr.method_17780());
			hitState = blockItem.getPlacementState(new class_1838(player, hand, bhr));
			if (hitState == null)
				return;
		}

		if (!(hitState.method_26204() instanceof TrackBlock))
			return;

		extraTipWarmup = restoreWarmup;
		boolean maxTurns = class_310.method_1551().field_1690.field_1867.method_1434();
		PlacementInfo info = tryConnect(level, player, pos, hitState, stack, false, maxTurns);
		if (extraTipWarmup < 20)
			extraTipWarmup++;
		if (!info.valid || !hoveringMaxed && (info.end1Extent == 0 || info.end2Extent == 0))
			extraTipWarmup = 0;

		if (!player.method_7337() && (info.valid || !info.hasRequiredTracks || !info.hasRequiredPavement))
			BlueprintOverlayRenderer.displayTrackRequirements(info, player.method_6079());

		if (info.valid)
			player.method_7353(Lang.translateDirect("track.valid_connection")
				.method_27692(class_124.field_1060), true);
		else if (info.message != null)
			player.method_7353(Lang.translateDirect(info.message)
				.method_27692(info.message.equals("track.second_point") ? class_124.field_1068 : class_124.field_1061),
				true);

		if (bhr.method_17780() == class_2350.field_11036) {
			class_243 lookVec = player.method_5720();
			int lookAngle = (int) (22.5 + AngleHelper.deg(class_3532.method_15349(lookVec.field_1350, lookVec.field_1352)) % 360) / 8;

			if (!pos.equals(hintPos) || lookAngle != hintAngle) {
				hints = Couple.create(ArrayList::new);
				hintAngle = lookAngle;
				hintPos = pos;

				for (int xOffset = -2; xOffset <= 2; xOffset++) {
					for (int zOffset = -2; zOffset <= 2; zOffset++) {
						class_2338 offset = pos.method_10069(xOffset, 0, zOffset);
						PlacementInfo adjInfo = tryConnect(level, player, offset, hitState, stack, false, maxTurns);
						hints.get(adjInfo.valid)
							.add(offset.method_10074());
					}
				}
			}

			if (hints != null && !hints.either(Collection::isEmpty)) {
				CreateClient.OUTLINER.showCluster("track_valid", hints.getFirst())
					.withFaceTexture(AllSpecialTextures.THIN_CHECKERED)
					.colored(0x95CD41)
					.lineWidth(0);
				CreateClient.OUTLINER.showCluster("track_invalid", hints.getSecond())
					.withFaceTexture(AllSpecialTextures.THIN_CHECKERED)
					.colored(0xEA5C2B)
					.lineWidth(0);
			}
		}

		animation.chase(info.valid ? 1 : 0, 0.25, Chaser.EXP);
		animation.tickChaser();

		if (!info.valid) {
			info.end1Extent = 0;
			info.end2Extent = 0;
		}

		int color = Color.mixColors(0xEA5C2B, 0x95CD41, animation.getValue());
		class_243 up = new class_243(0, 4 / 16f, 0);

		{
			class_243 v1 = info.end1;
			class_243 a1 = info.axis1.method_1029();
			class_243 n1 = info.normal1.method_1036(a1)
				.method_1021(15 / 16f);
			class_243 o1 = a1.method_1021(0.125f);
			class_243 ex1 =
				a1.method_1021((info.end1Extent - (info.curve == null && info.end1Extent > 0 ? 2 : 0)) * info.axis1.method_1033());
			line(1, v1.method_1019(n1)
				.method_1019(up), o1, ex1);
			line(2, v1.method_1020(n1)
				.method_1019(up), o1, ex1);

			class_243 v2 = info.end2;
			class_243 a2 = info.axis2.method_1029();
			class_243 n2 = info.normal2.method_1036(a2)
				.method_1021(15 / 16f);
			class_243 o2 = a2.method_1021(0.125f);
			class_243 ex2 = a2.method_1021(info.end2Extent * info.axis2.method_1033());
			line(3, v2.method_1019(n2)
				.method_1019(up), o2, ex2);
			line(4, v2.method_1020(n2)
				.method_1019(up), o2, ex2);
		}

		BezierConnection bc = info.curve;
		if (bc == null)
			return;

		class_243 previous1 = null;
		class_243 previous2 = null;
		int railcolor = color;
		int segCount = bc.getSegmentCount();

		float s = animation.getValue() * 7 / 8f + 1 / 8f;
		float lw = animation.getValue() * 1 / 16f + 1 / 16f;
		class_243 end1 = bc.starts.getFirst();
		class_243 end2 = bc.starts.getSecond();
		class_243 finish1 = end1.method_1019(bc.axes.getFirst()
			.method_1021(bc.getHandleLength()));
		class_243 finish2 = end2.method_1019(bc.axes.getSecond()
			.method_1021(bc.getHandleLength()));
		String key = "curve";

		for (int i = 0; i <= segCount; i++) {
			float t = i / (float) segCount;
			class_243 result = VecHelper.bezier(end1, end2, finish1, finish2, t);
			class_243 derivative = VecHelper.bezierDerivative(end1, end2, finish1, finish2, t)
				.method_1029();
			class_243 normal = bc.getNormal(t)
				.method_1036(derivative)
				.method_1021(15 / 16f);
			class_243 rail1 = result.method_1019(normal)
				.method_1019(up);
			class_243 rail2 = result.method_1020(normal)
				.method_1019(up);

			if (previous1 != null) {
				class_243 middle1 = rail1.method_1019(previous1)
					.method_1021(0.5f);
				class_243 middle2 = rail2.method_1019(previous2)
					.method_1021(0.5f);
				CreateClient.OUTLINER
					.showLine(Pair.of(key, i * 2), VecHelper.lerp(s, middle1, previous1),
						VecHelper.lerp(s, middle1, rail1))
					.colored(railcolor)
					.disableLineNormals()
					.lineWidth(lw);
				CreateClient.OUTLINER
					.showLine(Pair.of(key, i * 2 + 1), VecHelper.lerp(s, middle2, previous2),
						VecHelper.lerp(s, middle2, rail2))
					.colored(railcolor)
					.disableLineNormals()
					.lineWidth(lw);
			}

			previous1 = rail1;
			previous2 = rail2;
		}

		for (int i = segCount + 1; i <= lastLineCount; i++) {
			CreateClient.OUTLINER.remove(Pair.of(key, i * 2));
			CreateClient.OUTLINER.remove(Pair.of(key, i * 2 + 1));
		}

		lastLineCount = segCount;
	}

	@Environment(EnvType.CLIENT)
	private static void line(int id, class_243 v1, class_243 o1, class_243 ex) {
		int color = Color.mixColors(0xEA5C2B, 0x95CD41, animation.getValue());
		CreateClient.OUTLINER.showLine(Pair.of("start", id), v1.method_1020(o1), v1.method_1019(ex))
			.lineWidth(1 / 8f)
			.disableLineNormals()
			.colored(color);
	}

}
