package com.simibubi.create.content.trains.track;

import static com.simibubi.create.AllPartialModels.GIRDER_SEGMENT_BOTTOM;
import static com.simibubi.create.AllPartialModels.GIRDER_SEGMENT_MIDDLE;
import static com.simibubi.create.AllPartialModels.GIRDER_SEGMENT_TOP;

import com.jozufozu.flywheel.backend.Backend;
import com.simibubi.create.content.trains.track.BezierConnection.GirderAngles;
import com.simibubi.create.content.trains.track.BezierConnection.SegmentAngles;
import com.simibubi.create.foundation.blockEntity.renderer.SafeBlockEntityRenderer;
import com.simibubi.create.foundation.render.CachedBufferer;
import com.simibubi.create.foundation.utility.AngleHelper;
import com.simibubi.create.foundation.utility.Iterate;
import com.simibubi.create.foundation.utility.VecHelper;
import net.minecraft.class_1921;
import net.minecraft.class_1937;
import net.minecraft.class_2246;
import net.minecraft.class_2338;
import net.minecraft.class_2350.class_2351;
import net.minecraft.class_243;
import net.minecraft.class_2680;
import net.minecraft.class_3532;
import net.minecraft.class_4587;
import net.minecraft.class_4587.class_4665;
import net.minecraft.class_4588;
import net.minecraft.class_4597;
import net.minecraft.class_5614;
import net.minecraft.class_761;

public class TrackRenderer extends SafeBlockEntityRenderer<TrackBlockEntity> {

	public TrackRenderer(class_5614.class_5615 context) {}

	@Override
	protected void renderSafe(TrackBlockEntity be, float partialTicks, class_4587 ms, class_4597 buffer, int light,
		int overlay) {
		class_1937 level = be.method_10997();
		if (Backend.canUseInstancing(level))
			return;
		class_4588 vb = buffer.getBuffer(class_1921.method_23579());
		be.connections.values()
			.forEach(bc -> renderBezierTurn(level, bc, ms, vb));
	}

	public static void renderBezierTurn(class_1937 level, BezierConnection bc, class_4587 ms, class_4588 vb) {
		if (!bc.isPrimary())
			return;

		ms.method_22903();
		class_2338 tePosition = bc.tePositions.getFirst();
		class_2680 air = class_2246.field_10124.method_9564();
		SegmentAngles[] segments = bc.getBakedSegments();

		renderGirder(level, bc, ms, vb, tePosition);

		for (int i = 1; i < segments.length; i++) {
			SegmentAngles segment = segments[i];
			int light = class_761.method_23794(level, segment.lightPosition.method_10081(tePosition));

			TrackMaterial.TrackModelHolder modelHolder = bc.getMaterial().getModelHolder();

			CachedBufferer.partial(modelHolder.tie(), air)
				.mulPose(segment.tieTransform.method_23761())
				.mulNormal(segment.tieTransform.method_23762())
				.light(light)
				.renderInto(ms, vb);

			for (boolean first : Iterate.trueAndFalse) {
				class_4665 transform = segment.railTransforms.get(first);
				CachedBufferer.partial(first ? modelHolder.segment_left() : modelHolder.segment_right(), air)
					.mulPose(transform.method_23761())
					.mulNormal(transform.method_23762())
					.light(light)
					.renderInto(ms, vb);
			}
		}

		ms.method_22909();
	}

	private static void renderGirder(class_1937 level, BezierConnection bc, class_4587 ms, class_4588 vb,
		class_2338 tePosition) {
		if (!bc.hasGirder)
			return;

		class_2680 air = class_2246.field_10124.method_9564();
		GirderAngles[] girders = bc.getBakedGirders();

		for (int i = 1; i < girders.length; i++) {
			GirderAngles segment = girders[i];
			int light = class_761.method_23794(level, segment.lightPosition.method_10081(tePosition));

			for (boolean first : Iterate.trueAndFalse) {
				class_4665 beamTransform = segment.beams.get(first);
				CachedBufferer.partial(GIRDER_SEGMENT_MIDDLE, air)
					.mulPose(beamTransform.method_23761())
					.mulNormal(beamTransform.method_23762())
					.light(light)
					.renderInto(ms, vb);

				for (boolean top : Iterate.trueAndFalse) {
					class_4665 beamCapTransform = segment.beamCaps.get(top)
						.get(first);
					CachedBufferer.partial(top ? GIRDER_SEGMENT_TOP : GIRDER_SEGMENT_BOTTOM, air)
						.mulPose(beamCapTransform.method_23761())
						.mulNormal(beamCapTransform.method_23762())
						.light(light)
						.renderInto(ms, vb);
				}
			}
		}
	}

	public static class_243 getModelAngles(class_243 normal, class_243 diff) {
		double diffX = diff.method_10216();
		double diffY = diff.method_10214();
		double diffZ = diff.method_10215();
		double len = class_3532.method_15355((float) (diffX * diffX + diffZ * diffZ));
		double yaw = class_3532.method_15349(diffX, diffZ);
		double pitch = class_3532.method_15349(len, diffY) - Math.PI * .5;

		class_243 yawPitchNormal = VecHelper.rotate(VecHelper.rotate(new class_243(0, 1, 0), AngleHelper.deg(pitch), class_2351.field_11048),
			AngleHelper.deg(yaw), class_2351.field_11052);

		double signum = Math.signum(yawPitchNormal.method_1026(normal));
		if (Math.abs(signum) < 0.5f)
			signum = yawPitchNormal.method_1025(normal) < 0.5f ? -1 : 1;
		double dot = diff.method_1036(normal)
			.method_1029()
			.method_1026(yawPitchNormal);
		double roll = Math.acos(class_3532.method_15350(dot, -1, 1)) * signum;
		return new class_243(pitch, yaw, roll);
	}

	@Override
	public boolean shouldRenderOffScreen(TrackBlockEntity pBlockEntity) {
		return true;
	}

	@Override
	public int method_33893() {
		return 96 * 2;
	}

}
