package com.simibubi.create.content.trains.track;

import java.util.List;
import java.util.UUID;

import javax.annotation.Nullable;

import com.jozufozu.flywheel.core.PartialModel;
import com.simibubi.create.Create;
import com.simibubi.create.content.contraptions.StructureTransform;
import com.simibubi.create.content.schematics.SchematicWorld;
import com.simibubi.create.content.trains.graph.DimensionPalette;
import com.simibubi.create.content.trains.graph.EdgeData;
import com.simibubi.create.content.trains.graph.EdgePointType;
import com.simibubi.create.content.trains.graph.TrackEdge;
import com.simibubi.create.content.trains.graph.TrackGraph;
import com.simibubi.create.content.trains.graph.TrackGraphHelper;
import com.simibubi.create.content.trains.graph.TrackGraphLocation;
import com.simibubi.create.content.trains.graph.TrackNode;
import com.simibubi.create.content.trains.signal.SingleBlockEntityEdgePoint;
import com.simibubi.create.content.trains.signal.TrackEdgePoint;
import com.simibubi.create.foundation.blockEntity.SmartBlockEntity;
import com.simibubi.create.foundation.blockEntity.behaviour.BehaviourType;
import com.simibubi.create.foundation.blockEntity.behaviour.BlockEntityBehaviour;
import com.simibubi.create.foundation.ponder.PonderWorld;
import com.simibubi.create.foundation.render.CachedBufferer;
import com.simibubi.create.foundation.utility.Iterate;
import com.simibubi.create.foundation.utility.VecHelper;
import net.fabricmc.api.EnvType;
import net.fabricmc.api.Environment;
import net.minecraft.class_1921;
import net.minecraft.class_1936;
import net.minecraft.class_1937;
import net.minecraft.class_2248;
import net.minecraft.class_2338;
import net.minecraft.class_2350.class_2351;
import net.minecraft.class_2350.class_2352;
import net.minecraft.class_243;
import net.minecraft.class_2487;
import net.minecraft.class_2512;
import net.minecraft.class_2520;
import net.minecraft.class_2680;
import net.minecraft.class_3532;
import net.minecraft.class_4587;
import net.minecraft.class_4597;
import net.minecraft.class_761;

public class TrackTargetingBehaviour<T extends TrackEdgePoint> extends BlockEntityBehaviour {

	public static final BehaviourType<TrackTargetingBehaviour<?>> TYPE = new BehaviourType<>();

	private class_2338 targetTrack;
	private BezierTrackPointLocation targetBezier;
	private class_2352 targetDirection;
	private UUID id;

	private class_243 prevDirection;
	private class_243 rotatedDirection;

	private class_2487 migrationData;
	private EdgePointType<T> edgePointType;
	private T edgePoint;
	private boolean orthogonal;

	public TrackTargetingBehaviour(SmartBlockEntity be, EdgePointType<T> edgePointType) {
		super(be);
		this.edgePointType = edgePointType;
		targetDirection = class_2352.field_11056;
		targetTrack = class_2338.field_10980;
		id = UUID.randomUUID();
		migrationData = null;
		orthogonal = false;
	}

	@Override
	public boolean isSafeNBT() {
		return true;
	}

	@Override
	public void write(class_2487 nbt, boolean clientPacket) {
		nbt.method_25927("Id", id);
		nbt.method_10566("TargetTrack", class_2512.method_10692(targetTrack));
		nbt.method_10556("Ortho", orthogonal);
		nbt.method_10556("TargetDirection", targetDirection == class_2352.field_11056);
		if (rotatedDirection != null)
			nbt.method_10566("RotatedAxis", VecHelper.writeNBT(rotatedDirection));
		if (prevDirection != null)
			nbt.method_10566("PrevAxis", VecHelper.writeNBT(prevDirection));
		if (migrationData != null && !clientPacket)
			nbt.method_10566("Migrate", migrationData);
		if (targetBezier != null) {
			class_2487 bezierNbt = new class_2487();
			bezierNbt.method_10569("Segment", targetBezier.segment());
			bezierNbt.method_10566("Key", class_2512.method_10692(targetBezier.curveTarget()
				.method_10059(getPos())));
			nbt.method_10566("Bezier", bezierNbt);
		}
		super.write(nbt, clientPacket);
	}

	@Override
	public void read(class_2487 nbt, boolean clientPacket) {
		id = nbt.method_10545("Id") ? nbt.method_25926("Id") : UUID.randomUUID();
		targetTrack = class_2512.method_10691(nbt.method_10562("TargetTrack"));
		targetDirection = nbt.method_10577("TargetDirection") ? class_2352.field_11056 : class_2352.field_11060;
		orthogonal = nbt.method_10577("Ortho");
		if (nbt.method_10545("PrevAxis"))
			prevDirection = VecHelper.readNBT(nbt.method_10554("PrevAxis", class_2520.field_33256));
		if (nbt.method_10545("RotatedAxis"))
			rotatedDirection = VecHelper.readNBT(nbt.method_10554("RotatedAxis", class_2520.field_33256));
		if (nbt.method_10545("Migrate"))
			migrationData = nbt.method_10562("Migrate");
		if (clientPacket)
			edgePoint = null;
		if (nbt.method_10545("Bezier")) {
			class_2487 bezierNbt = nbt.method_10562("Bezier");
			class_2338 key = class_2512.method_10691(bezierNbt.method_10562("Key"));
			targetBezier = new BezierTrackPointLocation(bezierNbt.method_10545("FromStack") ? key : key.method_10081(getPos()),
				bezierNbt.method_10550("Segment"));
		}
		super.read(nbt, clientPacket);
	}

	@Nullable
	public T getEdgePoint() {
		return edgePoint;
	}

	public void invalidateEdgePoint(class_2487 migrationData) {
		this.migrationData = migrationData;
		edgePoint = null;
		blockEntity.sendData();
	}

	@Override
	public void tick() {
		super.tick();
		if (edgePoint == null)
			edgePoint = createEdgePoint();
	}

	@SuppressWarnings("unchecked")
	public T createEdgePoint() {
		class_1937 level = getWorld();
		boolean isClientSide = level.field_9236;
		if (migrationData == null || isClientSide)
			for (TrackGraph trackGraph : Create.RAILWAYS.sided(level).trackNetworks.values()) {
				T point = trackGraph.getPoint(edgePointType, id);
				if (point == null)
					continue;
				return point;
			}

		if (isClientSide)
			return null;
		if (!hasValidTrack())
			return null;
		TrackGraphLocation loc = determineGraphLocation();
		if (loc == null)
			return null;

		TrackGraph graph = loc.graph;
		TrackNode node1 = graph.locateNode(loc.edge.getFirst());
		TrackNode node2 = graph.locateNode(loc.edge.getSecond());
		TrackEdge edge = graph.getConnectionsFrom(node1)
			.get(node2);
		if (edge == null)
			return null;

		T point = edgePointType.create();
		boolean front = getTargetDirection() == class_2352.field_11056;

		prevDirection = edge.getDirectionAt(loc.position)
			.method_1021(front ? -1 : 1);

		if (rotatedDirection != null) {
			double dot = prevDirection.method_1026(rotatedDirection);
			if (dot < -.85f) {
				rotatedDirection = null;
				targetDirection = targetDirection.method_26424();
				return null;
			}

			rotatedDirection = null;
		}

		double length = edge.getLength();
		class_2487 data = migrationData;
		migrationData = null;

		{
			orthogonal = targetBezier == null;
			class_243 direction = edge.getDirection(true);
			int nonZeroComponents = 0;
			for (class_2351 axis : Iterate.axes)
				nonZeroComponents += direction.method_18043(axis) != 0 ? 1 : 0;
			orthogonal &= nonZeroComponents <= 1;
		}

		EdgeData signalData = edge.getEdgeData();
		if (signalData.hasPoints()) {
			for (EdgePointType<?> otherType : EdgePointType.TYPES.values()) {
				TrackEdgePoint otherPoint = signalData.get(otherType, loc.position);
				if (otherPoint == null)
					continue;
				if (otherType != edgePointType) {
					if (!otherPoint.canCoexistWith(edgePointType, front))
						return null;
					continue;
				}
				if (!otherPoint.canMerge())
					return null;
				otherPoint.blockEntityAdded(blockEntity, front);
				id = otherPoint.getId();
				blockEntity.notifyUpdate();
				return (T) otherPoint;
			}
		}

		if (data != null)
			point.read(data, true, DimensionPalette.read(data));

		point.setId(id);
		boolean reverseEdge = front || point instanceof SingleBlockEntityEdgePoint;
		point.setLocation(reverseEdge ? loc.edge : loc.edge.swap(), reverseEdge ? loc.position : length - loc.position);
		point.blockEntityAdded(blockEntity, front);
		loc.graph.addPoint(edgePointType, point);
		blockEntity.sendData();
		return point;
	}

	@Override
	public void destroy() {
		super.destroy();
		if (edgePoint != null && !getWorld().field_9236)
			edgePoint.blockEntityRemoved(getPos(), getTargetDirection() == class_2352.field_11056);
	}

	@Override
	public BehaviourType<?> getType() {
		return TYPE;
	}

	public boolean isOnCurve() {
		return targetBezier != null;
	}

	public boolean isOrthogonal() {
		return orthogonal;
	}

	public boolean hasValidTrack() {
		return getTrackBlockState().method_26204() instanceof ITrackBlock;
	}

	public ITrackBlock getTrack() {
		return (ITrackBlock) getTrackBlockState().method_26204();
	}

	public class_2680 getTrackBlockState() {
		return getWorld().method_8320(getGlobalPosition());
	}

	public class_2338 getGlobalPosition() {
		return targetTrack.method_10081(blockEntity.method_11016());
	}

	public class_2338 getPositionForMapMarker() {
		class_2338 target = targetTrack.method_10081(blockEntity.method_11016());
		if (targetBezier != null && getWorld().method_8321(target) instanceof TrackBlockEntity tbe) {
			BezierConnection bc = tbe.getConnections()
				.get(targetBezier.curveTarget());
			if (bc == null)
				return target;
			double length = class_3532.method_15357(bc.getLength() * 2);
			int seg = targetBezier.segment() + 1;
			double t = seg / length;
			return class_2338.method_49638(bc.getPosition(t));
		}
		return target;
	}

	public class_2352 getTargetDirection() {
		return targetDirection;
	}

	public BezierTrackPointLocation getTargetBezier() {
		return targetBezier;
	}

	public TrackGraphLocation determineGraphLocation() {
		class_1937 level = getWorld();
		class_2338 pos = getGlobalPosition();
		class_2680 state = getTrackBlockState();
		ITrackBlock track = getTrack();
		List<class_243> trackAxes = track.getTrackAxes(level, pos, state);
		class_2352 targetDirection = getTargetDirection();

		return targetBezier != null
			? TrackGraphHelper.getBezierGraphLocationAt(level, pos, targetDirection, targetBezier)
			: TrackGraphHelper.getGraphLocationAt(level, pos, targetDirection, trackAxes.get(0));
	}

	public static enum RenderedTrackOverlayType {
		STATION, SIGNAL, DUAL_SIGNAL, OBSERVER;
	}

	@Environment(EnvType.CLIENT)
	public static void render(class_1936 level, class_2338 pos, class_2352 direction,
		BezierTrackPointLocation bezier, class_4587 ms, class_4597 buffer, int light, int overlay,
		RenderedTrackOverlayType type, float scale) {
		if (level instanceof SchematicWorld && !(level instanceof PonderWorld))
			return;

		class_2680 trackState = level.method_8320(pos);
		class_2248 block = trackState.method_26204();
		if (!(block instanceof ITrackBlock))
			return;

		ms.method_22903();
		ITrackBlock track = (ITrackBlock) block;
		PartialModel partial = track.prepareTrackOverlay(level, pos, trackState, bezier, direction, ms, type);
		if (partial != null)
			CachedBufferer.partial(partial, trackState)
				.translate(.5, 0, .5)
				.scale(scale)
				.translate(-.5, 0, -.5)
				.light(class_761.method_23794(level, pos))
				.renderInto(ms, buffer.getBuffer(class_1921.method_23579()));
		ms.method_22909();
	}

	public void transform(StructureTransform transform) {
		id = UUID.randomUUID();
		targetTrack = transform.applyWithoutOffset(targetTrack);
		if (prevDirection != null)
			rotatedDirection = transform.applyWithoutOffsetUncentered(prevDirection);
		if (targetBezier != null)
			targetBezier = new BezierTrackPointLocation(transform.applyWithoutOffset(targetBezier.curveTarget()
				.method_10059(getPos()))
				.method_10081(getPos()), targetBezier.segment());
		blockEntity.notifyUpdate();
	}

}
