package com.simibubi.create.foundation.config.ui.compat.flywheel;

import java.util.Collections;

import com.electronwill.nightconfig.core.UnmodifiableConfig;
import com.jozufozu.flywheel.config.FlwConfig;
import com.jozufozu.flywheel.config.Option;
import com.simibubi.create.foundation.config.ui.ConfigHelper;
import com.simibubi.create.foundation.config.ui.ConfigScreenList;
import com.simibubi.create.foundation.config.ui.ConfigTextField;
import com.simibubi.create.foundation.config.ui.SubMenuConfigScreen;
import com.simibubi.create.foundation.config.ui.entries.SubMenuEntry;
import com.simibubi.create.foundation.gui.AllIcons;
import com.simibubi.create.foundation.gui.ConfirmationScreen;
import com.simibubi.create.foundation.gui.Theme;
import com.simibubi.create.foundation.gui.UIRenderHelper;
import com.simibubi.create.foundation.gui.element.DelegatedStencilElement;
import com.simibubi.create.foundation.gui.widget.BoxWidget;
import com.simibubi.create.foundation.item.TooltipHelper;
import com.simibubi.create.foundation.item.TooltipHelper.Palette;
import com.simibubi.create.foundation.utility.Color;
import com.simibubi.create.foundation.utility.Couple;
import io.github.fabricators_of_create.porting_lib.mixin.accessors.client.accessor.AbstractSelectionListAccessor;
import net.minecraft.class_124;
import net.minecraft.class_2561;
import net.minecraft.class_437;
import net.minecraft.class_5348;
import net.minecraftforge.fml.config.ModConfig;

public class FlwSubMenuConfigScreen extends SubMenuConfigScreen {

	private final FlwConfig flwConfig;

	public FlwSubMenuConfigScreen(class_437 parent, String title, ModConfig.Type type, FlwConfig flwConfig) {
		super(parent, title, type, null, null);
		this.flwConfig = flwConfig;
	}

	public FlwSubMenuConfigScreen(class_437 parent, ModConfig.Type type, FlwConfig flwConfig) {
		this(parent, "root", type, flwConfig);
	}

	@Override
	protected void saveChanges() {
		flwConfig.save();
	}

	@Override
	protected void resetConfig(UnmodifiableConfig values) {

	}

	@Override
	protected void method_25426() {
		guiLeft = (field_22789 - windowWidth) / 2;
		guiTop = (field_22790 - windowHeight) / 2;
		guiLeft += windowXOffset;
		guiTop += windowYOffset;

		listWidth = Math.min(field_22789 - 80, 300);

		int yCenter = field_22790 / 2;
		int listL = this.field_22789 / 2 - listWidth / 2;
		int listR = this.field_22789 / 2 + listWidth / 2;

		resetAll = new BoxWidget(listR + 10, yCenter - 25, 20, 20)
				.withPadding(2, 2)
				.withCallback((x, y) ->
						new ConfirmationScreen()
								.centered()
								.withText(class_5348.method_29430("Resetting this config is not supported!"))
//								.withAction(success -> {
//									if (success)
//										resetConfig(spec.getValues());
//								})
								.open(this)
				);

		resetAll.showingElement(AllIcons.I_CONFIG_RESET.asStencil().withElementRenderer(BoxWidget.gradientFactory.apply(resetAll)));
		resetAll.getToolTip().add(class_2561.method_43470("Reset All"));
		resetAll.getToolTip().addAll(TooltipHelper.cutStringTextComponent("Click here to reset all settings to their default value.", Palette.ALL_GRAY));

		saveChanges = new BoxWidget(listL - 30, yCenter - 25, 20, 20)
				.withPadding(2, 2)
				.withCallback((x, y) -> {
					if (ConfigHelper.changes.isEmpty())
						return;

					ConfirmationScreen confirm = new ConfirmationScreen()
							.centered()
							.withText(class_5348.method_29430("Saving " + ConfigHelper.changes.size() + " changed value" + (ConfigHelper.changes.size() != 1 ? "s" : "") + ""))
							.withAction(success -> {
								if (success)
									saveChanges();
							});

					addAnnotationsToConfirm(confirm).open(this);
				});
		saveChanges.showingElement(AllIcons.I_CONFIG_SAVE.asStencil().withElementRenderer(BoxWidget.gradientFactory.apply(saveChanges)));
		saveChanges.getToolTip().add(class_2561.method_43470("Save Changes"));
		saveChanges.getToolTip().addAll(TooltipHelper.cutStringTextComponent("Click here to save your current changes.", Palette.ALL_GRAY));

		discardChanges = new BoxWidget(listL - 30, yCenter + 5, 20, 20)
				.withPadding(2, 2)
				.withCallback((x, y) -> {
					if (ConfigHelper.changes.isEmpty())
						return;

					new ConfirmationScreen()
							.centered()
							.withText(class_5348.method_29430("Discarding " + ConfigHelper.changes.size() + " unsaved change" + (ConfigHelper.changes.size() != 1 ? "s" : "") + ""))
							.withAction(success -> {
								if (success)
									clearChanges();
							})
							.open(this);
				});
		discardChanges.showingElement(AllIcons.I_CONFIG_DISCARD.asStencil().withElementRenderer(BoxWidget.gradientFactory.apply(discardChanges)));
		discardChanges.getToolTip().add(class_2561.method_43470("Discard Changes"));
		discardChanges.getToolTip().addAll(TooltipHelper.cutStringTextComponent("Click here to discard all the changes you made.", Palette.ALL_GRAY));

		goBack = new BoxWidget(listL - 30, yCenter + 65, 20, 20)
				.withPadding(2, 2)
				.withCallback(this::attemptBackstep);
		goBack.showingElement(AllIcons.I_CONFIG_BACK.asStencil().withElementRenderer(BoxWidget.gradientFactory.apply(goBack)));
		goBack.getToolTip().add(class_2561.method_43470("Go Back"));

		method_37063(resetAll);
		method_37063(saveChanges);
		method_37063(discardChanges);
		method_37063(goBack);

		list = new ConfigScreenList(field_22787, listWidth, field_22790 - 80, 35, field_22790 - 45, 40);
		list.method_25333(this.field_22789 / 2 - ((AbstractSelectionListAccessor) list).port_lib$getWidth() / 2);

		method_37063(list);

		search = new ConfigTextField(field_22793, field_22789 / 2 - listWidth / 2, field_22790 - 35, listWidth, 20);
		search.method_1863(this::updateFilter);
		search.setHint("Search..");
		search.method_1870();
		method_37063(search);

		flwConfig.getOptionMapView().forEach((key, option) -> {
			String humanKey = toHumanReadable(key);

			Object value = option.get();
			ConfigScreenList.Entry entry = null;
			if (value instanceof Boolean) {
				entry = new FlwBooleanEntry(humanKey, (Option<Boolean>) option);
			} else if (value instanceof Enum) {
				entry = new FlwEnumEntry(humanKey, (Option<Enum<?>>) option);
			}

			if (entry == null)
				entry = new ConfigScreenList.LabeledEntry("Impl missing - " + option.get().getClass().getSimpleName() + "  " + humanKey + " : " + value);

			if (highlights.contains(key))
				entry.annotations.put("highlight", ":)");

			list.method_25396().add(entry);
		});

		Collections.sort(list.method_25396(),
				(e, e2) -> {
					int group = (e2 instanceof SubMenuEntry ? 1 : 0) - (e instanceof SubMenuEntry ? 1 : 0);
					if (group == 0 && e instanceof ConfigScreenList.LabeledEntry && e2 instanceof ConfigScreenList.LabeledEntry) {
						ConfigScreenList.LabeledEntry le = (ConfigScreenList.LabeledEntry) e;
						ConfigScreenList.LabeledEntry le2 = (ConfigScreenList.LabeledEntry) e2;
						return le.label.getComponent()
								.getString()
								.compareTo(le2.label.getComponent()
										.getString());
					}
					return group;
				});

		list.search(highlights.stream().findFirst().orElse(""));

		//extras for server configs
		if (type != ModConfig.Type.SERVER)
			return;
		if (field_22787.method_1496())
			return;

		boolean canEdit = field_22787 != null && field_22787.field_1724 != null && field_22787.field_1724.method_5687(2);

		Couple<Color> red = Theme.p(Theme.Key.BUTTON_FAIL);
		Couple<Color> green = Theme.p(Theme.Key.BUTTON_SUCCESS);

		DelegatedStencilElement stencil = new DelegatedStencilElement();

		serverLocked = new BoxWidget(listR + 10, yCenter + 5, 20, 20)
				.withPadding(2, 2)
				.showingElement(stencil);


		if (!canEdit) {
			list.method_25396().forEach(e -> e.setEditable(false));
			resetAll.field_22763 = false;
			stencil.withStencilRenderer((ms, w, h, alpha) -> AllIcons.I_CONFIG_LOCKED.render(ms, 0, 0));
			stencil.withElementRenderer((ms, w, h, alpha) -> UIRenderHelper.angledGradient(ms, 90, 8, 0, 16, 16, red));
			serverLocked.withBorderColors(red);
			serverLocked.getToolTip().add(class_2561.method_43470("Locked").method_27692(class_124.field_1067));
			serverLocked.getToolTip().addAll(TooltipHelper.cutStringTextComponent("You do not have enough permissions to edit the server config. You can still look at the current values here though.", Palette.ALL_GRAY));
		} else {
			stencil.withStencilRenderer((ms, w, h, alpha) -> AllIcons.I_CONFIG_UNLOCKED.render(ms, 0, 0));
			stencil.withElementRenderer((ms, w, h, alpha) -> UIRenderHelper.angledGradient(ms, 90, 8, 0, 16, 16, green));
			serverLocked.withBorderColors(green);
			serverLocked.getToolTip().add(class_2561.method_43470("Unlocked").method_27692(class_124.field_1067));
			serverLocked.getToolTip().addAll(TooltipHelper.cutStringTextComponent("You have enough permissions to edit the server config. Changes you make here will be synced with the server when you save them.", Palette.ALL_GRAY));
		}

		method_37063(serverLocked);
	}
}
