package com.simibubi.create.foundation.data;

import java.io.BufferedReader;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.concurrent.CompletableFuture;
import net.minecraft.class_156;
import net.minecraft.class_2405;
import net.minecraft.class_2960;
import net.minecraft.class_3518;
import net.minecraft.class_7403;
import net.minecraft.class_7784;
import org.apache.commons.lang3.mutable.MutableBoolean;
import org.apache.commons.lang3.mutable.MutableObject;

import com.google.common.hash.HashCode;
import com.google.gson.Gson;
import com.google.gson.GsonBuilder;
import com.google.gson.JsonObject;
import com.simibubi.create.Create;
import com.simibubi.create.foundation.ponder.PonderScene;
import com.tterrag.registrate.providers.RegistrateDataProvider;
import com.tterrag.registrate.AbstractRegistrate;
import com.tterrag.registrate.providers.ProviderType;

/**
 * @deprecated Use {@link AbstractRegistrate#addRawLang} or {@link AbstractRegistrate#addDataGenerator} with {@link ProviderType#LANG} instead.
 */
@Deprecated(forRemoval = true)
public class LangMerger implements class_2405 {

	static final Gson GSON = new GsonBuilder().setPrettyPrinting()
			.disableHtmlEscaping()
			.create();
	private static final String CATEGORY_HEADER =
		"\t\"_\": \"->------------------------]  %s  [------------------------<-\",";

	private final String modid;
	private final String displayName;
	@SuppressWarnings("removal")
	private final LangPartial[] langPartials;

	private List<Object> mergedLangData;
	private List<String> langIgnore;
	private class_7784 output;

	@SuppressWarnings("removal")
	public <T extends LangPartial> LangMerger(class_7784 output, String modid, String displayName,
		T[] allLangPartials) {
		this.output = output;
		this.modid = modid;
		this.displayName = displayName;
		this.langPartials = allLangPartials;
		this.mergedLangData = new ArrayList<>();
		this.langIgnore = new ArrayList<>();
		populateLangIgnore();
	}

	protected void populateLangIgnore() {
		// Key prefixes added here will NOT be transferred to lang templates
		langIgnore.add("create.ponder.debug_"); // Ponder debug scene text
		langIgnore.add("create.gui.chromatic_projector");
	}

	private boolean shouldIgnore(String key) {
		for (String string : langIgnore)
			if (key.startsWith(string))
				return true;
		return false;
	}

	@Override
	public String method_10321() {
		return displayName + "'s lang merger";
	}

	@Override
	public CompletableFuture<?> method_10319(class_7403 pOutput) {
		Path path = output.method_45973(class_7784.class_7490.field_39368, "lang")
				.method_44107(new class_2960(modid, "en_us"));

		return CompletableFuture.runAsync(() -> {
			try {
				collectExistingEntries(path);
				collectEntries();
				if (mergedLangData.isEmpty())
					return;
				save(pOutput, mergedLangData, path,
					"Merging en_us.json with hand-written lang entries...");

			} catch (IOException ioexception) {
				field_40831.error("Failed to run LangMerger", ioexception);
			}

		}, class_156.method_18349());
	}

	private void collectExistingEntries(Path path) throws IOException {
		if (!Files.exists(path)) {
			Create.LOGGER.warn("Nothing to merge! It appears no lang was generated before me.");
			return;
		}

		try (BufferedReader reader = Files.newBufferedReader(path)) {
			JsonObject jsonobject = class_3518.method_15276(GSON, reader, JsonObject.class);

			/*
			 * Erase additional sections from previous lang in case registrate did not
			 * create a new one (this assumes advancements to be the first section after
			 * game elements)
			 */
			Set<String> keysToRemove = new HashSet<>();
			MutableBoolean startErasing = new MutableBoolean();
			jsonobject.entrySet()
				.stream()
				.forEachOrdered(entry -> {
					String key = entry.getKey();
					if (key.startsWith("advancement"))
						startErasing.setTrue();
					if (startErasing.isFalse())
						return;
					keysToRemove.add(key);
				});
			jsonobject.remove("_");
			keysToRemove.forEach(jsonobject::remove);

			addAll("Game Elements", jsonobject);
			reader.close();
		}
	}

	protected void addAll(String header, JsonObject jsonobject) {
		if (jsonobject == null)
			return;
		header = String.format(CATEGORY_HEADER, header);

		writeData("\n");
		writeData(header);
		writeData("\n\n");

		MutableObject<String> previousKey = new MutableObject<>("");
		jsonobject.entrySet()
				.stream()
				.forEachOrdered(entry -> {
					String key = entry.getKey();
					if (shouldIgnore(key))
						return;
					String value = entry.getValue()
							.getAsString();
					if (!previousKey.getValue()
							.isEmpty() && shouldAddLineBreak(key, previousKey.getValue()))
						writeData("\n");
					writeEntry(key, value);
					previousKey.setValue(key);
				});

		writeData("\n");
	}

	private void writeData(String data) {
		mergedLangData.add(data);
	}

	private void writeEntry(String key, String value) {
		mergedLangData.add(new LangEntry(key, value));
	}

	protected boolean shouldAddLineBreak(String key, String previousKey) {
		// Always put tooltips and ponder scenes in their own paragraphs
		if (key.endsWith(".tooltip"))
			return true;
		if (key.startsWith(modid + ".ponder") && key.endsWith(PonderScene.TITLE_KEY))
			return true;

		key = key.replaceFirst("\\.", "");
		previousKey = previousKey.replaceFirst("\\.", "");

		String[] split = key.split("\\.");
		String[] split2 = previousKey.split("\\.");
		if (split.length == 0 || split2.length == 0)
			return false;

		// Start new paragraph if keys before second point do not match
		return !split[0].equals(split2[0]);
	}

	@SuppressWarnings("removal")
	private void collectEntries() {
		for (LangPartial partial : langPartials)
			addAll(partial.getDisplayName(), partial.provide()
					.getAsJsonObject());
	}

	private void save(class_7403 cache, List<Object> dataIn, Path target, String message) throws IOException {
		ByteArrayOutputStream bytearrayoutputstream = new ByteArrayOutputStream();
		Writer writer = new OutputStreamWriter(bytearrayoutputstream, StandardCharsets.UTF_8);
		writer.append(createString(dataIn));
		writer.close();

		class_7403.field_39439.method_43346(target, bytearrayoutputstream.toByteArray(), HashCode.fromInt(0));
		Create.LOGGER.info(message);
	}

	protected String createString(List<Object> data) {
		StringBuilder builder = new StringBuilder();
		builder.append("{\n");
		data.forEach(builder::append);
		builder.append("\t\"_\": \"Thank you for translating ")
			.append(displayName)
			.append("!\"\n\n");
		builder.append("}");
		return builder.toString();
	}

}
