package com.simibubi.create.foundation.data.recipe;

import static com.simibubi.create.foundation.data.recipe.CompatMetals.ALUMINUM;
import static com.simibubi.create.foundation.data.recipe.CompatMetals.LEAD;
import static com.simibubi.create.foundation.data.recipe.CompatMetals.NICKEL;
import static com.simibubi.create.foundation.data.recipe.CompatMetals.OSMIUM;
import static com.simibubi.create.foundation.data.recipe.CompatMetals.PLATINUM;
import static com.simibubi.create.foundation.data.recipe.CompatMetals.QUICKSILVER;
import static com.simibubi.create.foundation.data.recipe.CompatMetals.SILVER;
import static com.simibubi.create.foundation.data.recipe.CompatMetals.TIN;
import static com.simibubi.create.foundation.data.recipe.CompatMetals.URANIUM;

import java.util.function.Supplier;
import java.util.function.UnaryOperator;

import com.simibubi.create.AllBlocks;
import com.simibubi.create.AllItems;
import com.simibubi.create.AllRecipeTypes;
import com.simibubi.create.AllTags;
import com.simibubi.create.content.decoration.palettes.AllPaletteStoneTypes;
import com.simibubi.create.content.processing.recipe.ProcessingRecipe;
import com.simibubi.create.content.processing.recipe.ProcessingRecipeBuilder;
import com.simibubi.create.foundation.utility.Lang;

import net.fabricmc.fabric.api.datagen.v1.FabricDataGenerator;
import net.fabricmc.fabric.api.datagen.v1.FabricDataOutput;
import net.fabricmc.fabric.api.resource.conditions.v1.DefaultResourceConditions;
import net.minecraft.class_1802;
import net.minecraft.class_1935;
import net.minecraft.class_2246;
import net.minecraft.class_3489;
import net.minecraft.class_3532;

public class CrushingRecipeGen extends ProcessingRecipeGen {

	GeneratedRecipe

	BLAZE_ROD = create(() -> class_1802.field_8894, b -> b.duration(100)
		.output(class_1802.field_8183, 3)
		.output(.25f, class_1802.field_8183, 3)),

		PRISMARINE_CRYSTALS = create(() -> class_1802.field_8434, b -> b.duration(150)
			.output(1f, class_1802.field_8155, 1)
			.output(.5f, class_1802.field_8155, 2)
			.output(.1f, class_1802.field_8601, 2)),

		OBSIDIAN = create(() -> class_2246.field_10540, b -> b.duration(500)
			.output(AllItems.POWDERED_OBSIDIAN.get())
			.output(.75f, class_2246.field_10540)),

		WOOL = create("wool", b -> b.duration(100)
			.require(class_3489.field_15544)
			.output(class_1802.field_8276, 2)
			.output(.5f, class_1802.field_8276)),

		DIORITE = mineralRecycling(AllPaletteStoneTypes.DIORITE, b -> b.duration(350)
			.output(.25f, class_1802.field_8155, 1)),

		CRIMSITE =
			mineralRecycling(AllPaletteStoneTypes.CRIMSITE, AllItems.CRUSHED_IRON::get, () -> class_1802.field_8675, .4f),

		VERIDIUM = mineralRecycling(AllPaletteStoneTypes.VERIDIUM, AllItems.CRUSHED_COPPER::get,
			() -> AllItems.COPPER_NUGGET::get, .8f),

		ASURINE = mineralRecycling(AllPaletteStoneTypes.ASURINE, AllItems.CRUSHED_ZINC::get,
			() -> AllItems.ZINC_NUGGET::get, .3f),

		OCHRUM =
			mineralRecycling(AllPaletteStoneTypes.OCHRUM, AllItems.CRUSHED_GOLD::get, () -> class_1802.field_8397, .2f),

		TUFF = mineralRecycling(AllPaletteStoneTypes.TUFF, b -> b.duration(350)
			.output(.25f, class_1802.field_8145, 1)
			.output(.1f, class_1802.field_8397, 1)
			.output(.1f, AllItems.COPPER_NUGGET.get(), 1)
			.output(.1f, AllItems.ZINC_NUGGET.get(), 1)
			.output(.1f, class_1802.field_8675, 1)),

		COPPER_ORE = stoneOre(() -> class_1802.field_27018, AllItems.CRUSHED_COPPER::get, 5.25f, 250),
		ZINC_ORE = stoneOre(AllBlocks.ZINC_ORE::get, AllItems.CRUSHED_ZINC::get, 1.75f, 250),
		IRON_ORE = stoneOre(() -> class_1802.field_8599, () -> AllItems.CRUSHED_IRON::get, 1.75f, 250),
		GOLD_ORE = stoneOre(() -> class_1802.field_8775, AllItems.CRUSHED_GOLD::get, 1.75f, 250),
		DIAMOND_ORE = stoneOre(() -> class_1802.field_8787, () -> class_1802.field_8477, 1.75f, 350),
		EMERALD_ORE = stoneOre(() -> class_1802.field_8837, () -> class_1802.field_8687, 1.75f, 350),
		COAL_ORE = stoneOre(() -> class_1802.field_8476, () -> class_1802.field_8713, 1.75f, 150),
		REDSTONE_ORE = stoneOre(() -> class_1802.field_8604, () -> class_1802.field_8725, 6.5f, 250),
		LAPIS_ORE = stoneOre(() -> class_1802.field_8809, () -> class_1802.field_8759, 10.5f, 250),

		DEEP_COPPER_ORE = deepslateOre(() -> class_1802.field_29211, AllItems.CRUSHED_COPPER::get, 7.25f, 350),
		DEEP_ZINC_ORE = deepslateOre(AllBlocks.DEEPSLATE_ZINC_ORE::get, AllItems.CRUSHED_ZINC::get, 2.25f, 350),
		DEEP_IRON_ORE = deepslateOre(() -> class_1802.field_29020, AllItems.CRUSHED_IRON::get, 2.25f, 350),
		DEEP_GOLD_ORE = deepslateOre(() -> class_1802.field_29019, AllItems.CRUSHED_GOLD::get, 2.25f, 350),
		DEEP_DIAMOND_ORE = deepslateOre(() -> class_1802.field_29022, () -> class_1802.field_8477, 2.25f, 450),
		DEEP_EMERALD_ORE = deepslateOre(() -> class_1802.field_29216, () -> class_1802.field_8687, 2.25f, 450),
		DEEP_COAL_ORE = deepslateOre(() -> class_1802.field_29212, () -> class_1802.field_8713, 1.75f, 250),
		DEEP_REDSTONE_ORE = deepslateOre(() -> class_1802.field_29023, () -> class_1802.field_8725, 7.5f, 350),
		DEEP_LAPIS_ORE = deepslateOre(() -> class_1802.field_29021, () -> class_1802.field_8759, 12.5f, 350),

		NETHER_GOLD_ORE = netherOre(() -> class_1802.field_23140, () -> class_1802.field_8397, 18, 350),
		NETHER_QUARTZ_ORE = netherOre(() -> class_1802.field_8702, () -> class_1802.field_8155, 2.25f, 350),

		RAW_COPPER_ORE = rawOre(() -> class_1802.field_33401, AllItems.CRUSHED_COPPER::get, 1),
		RAW_ZINC_ORE = rawOre(AllItems.RAW_ZINC::get, AllItems.CRUSHED_ZINC::get, 1),
		RAW_IRON_ORE = rawOre(() -> class_1802.field_33400, AllItems.CRUSHED_IRON::get, 1),
		RAW_GOLD_ORE = rawOre(() -> class_1802.field_33402, AllItems.CRUSHED_GOLD::get, 1),

		OSMIUM_ORE = moddedOre(OSMIUM, AllItems.CRUSHED_OSMIUM::get),
		PLATINUM_ORE = moddedOre(PLATINUM, AllItems.CRUSHED_PLATINUM::get),
		SILVER_ORE = moddedOre(SILVER, AllItems.CRUSHED_SILVER::get),
		TIN_ORE = moddedOre(TIN, AllItems.CRUSHED_TIN::get),
		QUICKSILVER_ORE = moddedOre(QUICKSILVER, AllItems.CRUSHED_QUICKSILVER::get),
		LEAD_ORE = moddedOre(LEAD, AllItems.CRUSHED_LEAD::get),
		ALUMINUM_ORE = moddedOre(ALUMINUM, AllItems.CRUSHED_BAUXITE::get),
		URANIUM_ORE = moddedOre(URANIUM, AllItems.CRUSHED_URANIUM::get),
		NICKEL_ORE = moddedOre(NICKEL, AllItems.CRUSHED_NICKEL::get),

		OSMIUM_RAW_ORE = moddedRawOre(OSMIUM, AllItems.CRUSHED_OSMIUM::get, 1),
		PLATINUM_RAW_ORE = moddedRawOre(PLATINUM, AllItems.CRUSHED_PLATINUM::get, 1),
		SILVER_RAW_ORE = moddedRawOre(SILVER, AllItems.CRUSHED_SILVER::get, 1),
		TIN_RAW_ORE = moddedRawOre(TIN, AllItems.CRUSHED_TIN::get, 1),
		QUICKSILVER_RAW_ORE = moddedRawOre(QUICKSILVER, AllItems.CRUSHED_QUICKSILVER::get, 1),
		LEAD_RAW_ORE = moddedRawOre(LEAD, AllItems.CRUSHED_LEAD::get, 1),
		ALUMINUM_RAW_ORE = moddedRawOre(ALUMINUM, AllItems.CRUSHED_BAUXITE::get, 1),
		URANIUM_RAW_ORE = moddedRawOre(URANIUM, AllItems.CRUSHED_URANIUM::get, 1),
		NICKEL_RAW_ORE = moddedRawOre(NICKEL, AllItems.CRUSHED_NICKEL::get, 1),

		RAW_COPPER_BLOCK = rawOre(() -> class_1802.field_33506, AllItems.CRUSHED_COPPER::get, 9),
		RAW_ZINC_BLOCK = rawOre(AllBlocks.RAW_ZINC_BLOCK::get, AllItems.CRUSHED_ZINC::get, 9),
		RAW_IRON_BLOCK = rawOre(() -> class_1802.field_33505, AllItems.CRUSHED_IRON::get, 9),
		RAW_GOLD_BLOCK = rawOre(() -> class_1802.field_33507, AllItems.CRUSHED_GOLD::get, 9),

		OSMIUM_RAW_BLOCK = moddedRawOre(OSMIUM, AllItems.CRUSHED_OSMIUM::get, 9),
		PLATINUM_RAW_BLOCK = moddedRawOre(PLATINUM, AllItems.CRUSHED_PLATINUM::get, 9),
		SILVER_RAW_BLOCK = moddedRawOre(SILVER, AllItems.CRUSHED_SILVER::get, 9),
		TIN_RAW_BLOCK = moddedRawOre(TIN, AllItems.CRUSHED_TIN::get, 9),
		QUICKSILVER_RAW_BLOCK = moddedRawOre(QUICKSILVER, AllItems.CRUSHED_QUICKSILVER::get, 9),
		LEAD_RAW_BLOCK = moddedRawOre(LEAD, AllItems.CRUSHED_LEAD::get, 9),
		ALUMINUM_RAW_BLOCK = moddedRawOre(ALUMINUM, AllItems.CRUSHED_BAUXITE::get, 9),
		URANIUM_RAW_BLOCK = moddedRawOre(URANIUM, AllItems.CRUSHED_URANIUM::get, 9),
		NICKEL_RAW_BLOCK = moddedRawOre(NICKEL, AllItems.CRUSHED_NICKEL::get, 9),

		NETHER_WART = create("nether_wart_block", b -> b.duration(150)
			.require(class_2246.field_10541)
			.output(.25f, class_1802.field_8790, 1)),

		AMETHYST_CLUSTER = create(() -> class_2246.field_27161, b -> b.duration(150)
			.output(class_1802.field_27063, 7)
			.output(.5f, class_1802.field_27063)),

		GLOWSTONE = create(() -> class_2246.field_10171, b -> b.duration(150)
			.output(class_1802.field_8601, 3)
			.output(.5f, class_1802.field_8601)),

		AMETHYST_BLOCK = create(() -> class_2246.field_27159, b -> b.duration(150)
			.output(class_1802.field_27063, 3)
			.output(.5f, class_1802.field_27063)),

		LEATHER_HORSE_ARMOR = create(() -> class_1802.field_18138, b -> b.duration(200)
			.output(class_1802.field_8745, 2)
			.output(.5f, class_1802.field_8745, 2)),

		IRON_HORSE_ARMOR = create(() -> class_1802.field_8578, b -> b.duration(200)
			.output(class_1802.field_8620, 2)
			.output(.5f, class_1802.field_8745, 1)
			.output(.5f, class_1802.field_8620, 1)
			.output(.25f, class_1802.field_8276, 2)
			.output(.25f, class_1802.field_8675, 4)),

		GOLDEN_HORSE_ARMOR = create(() -> class_1802.field_8560, b -> b.duration(200)
			.output(class_1802.field_8695, 2)
			.output(.5f, class_1802.field_8745, 2)
			.output(.5f, class_1802.field_8695, 2)
			.output(.25f, class_1802.field_8276, 2)
			.output(.25f, class_1802.field_8397, 8)),

		DIAMOND_HORSE_ARMOR = create(() -> class_1802.field_8807, b -> b.duration(200)
			.output(class_1802.field_8477, 1)
			.output(.5f, class_1802.field_8745, 2)
			.output(.1f, class_1802.field_8477, 3)
			.output(.25f, class_1802.field_8276, 2)),

		GRAVEL = create(() -> class_2246.field_10255, b -> b.duration(250)
			.output(class_2246.field_10102)
			.output(.1f, class_1802.field_8145)
			.output(.05f, class_1802.field_8696)),

		NETHERRACK = create(() -> class_2246.field_10515, b -> b.duration(250)
			.output(AllItems.CINDER_FLOUR.get())
			.output(.5f, AllItems.CINDER_FLOUR.get()))

	;

	protected GeneratedRecipe stoneOre(Supplier<class_1935> ore, Supplier<class_1935> raw, float expectedAmount,
		int duration) {
		return ore(class_2246.field_10445, ore, raw, expectedAmount, duration);
	}

	protected GeneratedRecipe deepslateOre(Supplier<class_1935> ore, Supplier<class_1935> raw, float expectedAmount,
		int duration) {
		return ore(class_2246.field_29031, ore, raw, expectedAmount, duration);
	}

	protected GeneratedRecipe netherOre(Supplier<class_1935> ore, Supplier<class_1935> raw, float expectedAmount,
		int duration) {
		return ore(class_2246.field_10515, ore, raw, expectedAmount, duration);
	}

	protected GeneratedRecipe mineralRecycling(AllPaletteStoneTypes type, Supplier<class_1935> crushed,
		Supplier<class_1935> nugget, float chance) {
		return mineralRecycling(type, b -> b.duration(250)
			.output(chance, crushed.get(), 1)
			.output(chance, nugget.get(), 1));
	}

	protected GeneratedRecipe mineralRecycling(AllPaletteStoneTypes type,
		UnaryOperator<ProcessingRecipeBuilder<ProcessingRecipe<?>>> transform) {
		create(Lang.asId(type.name()) + "_recycling", b -> transform.apply(b.require(type.materialTag)));
		return create(type.getBaseBlock()::get, transform);
	}

	protected GeneratedRecipe ore(class_1935 stoneType, Supplier<class_1935> ore, Supplier<class_1935> raw,
		float expectedAmount, int duration) {
		return create(ore, b -> {
			ProcessingRecipeBuilder<ProcessingRecipe<?>> builder = b.duration(duration)
				.output(raw.get(), class_3532.method_15375(expectedAmount));
			float extra = expectedAmount - class_3532.method_15375(expectedAmount);
			if (extra > 0)
				builder.output(extra, raw.get(), 1);
			builder.output(.75f, AllItems.EXP_NUGGET.get(), raw.get() == AllItems.CRUSHED_GOLD.get() ? 2 : 1);
			return builder.output(.125f, stoneType);
		});
	}

	protected GeneratedRecipe rawOre(Supplier<class_1935> input, Supplier<class_1935> result, int amount) {
		return create(input, b -> b.duration(400)
			.output(result.get(), amount)
			.output(.75f, AllItems.EXP_NUGGET.get(), (result.get() == AllItems.CRUSHED_GOLD.get() ? 2 : 1) * amount));
	}

	protected GeneratedRecipe moddedRawOre(CompatMetals metal, Supplier<class_1935> result, int amount) {
		String name = metal.getName();
		return create("raw_" + name + (amount == 1 ? "_ore" : "_block"), b -> {
			String suffix = amount == 1 ? "_ores" : "_blocks";
			return b.duration(400)
				.withCondition(DefaultResourceConditions.tagsPopulated(AllTags.forgeItemTag("raw_" + name + suffix)))
				.require(AllTags.forgeItemTag("raw_" + name + suffix))
				.output(result.get(), amount)
				.output(.75f, AllItems.EXP_NUGGET.get(), amount);
		});
	}

	protected GeneratedRecipe moddedOre(CompatMetals metal, Supplier<class_1935> result) {
		String name = metal.getName();
		return create(name + "_ore", b -> {
			String suffix = "_ores";
			return b.duration(400)
				.withCondition(DefaultResourceConditions.tagsPopulated(AllTags.forgeItemTag(name + suffix)))
				.require(AllTags.forgeItemTag(name + suffix))
				.output(result.get(), 1)
				.output(.75f, result.get(), 1)
				.output(.75f, AllItems.EXP_NUGGET.get(), 1);
		});
	}

	public CrushingRecipeGen(FabricDataOutput output) {
		super(output);
	}

	@Override
	protected AllRecipeTypes getRecipeType() {
		return AllRecipeTypes.CRUSHING;
	}

}
