package com.simibubi.create.foundation.data.recipe;

import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.CompletableFuture;
import java.util.function.Supplier;
import java.util.function.UnaryOperator;

import com.simibubi.create.Create;
import com.simibubi.create.content.processing.recipe.ProcessingRecipe;
import com.simibubi.create.content.processing.recipe.ProcessingRecipeBuilder;
import com.simibubi.create.content.processing.recipe.ProcessingRecipeSerializer;
import com.simibubi.create.foundation.recipe.IRecipeTypeInfo;
import com.simibubi.create.foundation.utility.RegisteredObjects;

import net.fabricmc.fabric.api.datagen.v1.FabricDataGenerator;
import net.fabricmc.fabric.api.datagen.v1.FabricDataOutput;
import net.fabricmc.fabric.api.transfer.v1.fluid.FluidConstants;
import net.minecraft.class_1856;
import net.minecraft.class_1935;
import net.minecraft.class_2405;
import net.minecraft.class_2960;
import net.minecraft.class_7403;

public abstract class ProcessingRecipeGen extends CreateRecipeProvider {

	protected static final List<ProcessingRecipeGen> GENERATORS = new ArrayList<>();
	protected static final long BUCKET = FluidConstants.BUCKET;
	protected static final long BOTTLE = FluidConstants.BOTTLE;

	public static class_2405 registerAll(FabricDataOutput output) {
		GENERATORS.add(new CrushingRecipeGen(output));
		GENERATORS.add(new MillingRecipeGen(output));
		GENERATORS.add(new CuttingRecipeGen(output));
		GENERATORS.add(new WashingRecipeGen(output));
		GENERATORS.add(new PolishingRecipeGen(output));
		GENERATORS.add(new DeployingRecipeGen(output));
		GENERATORS.add(new MixingRecipeGen(output));
		GENERATORS.add(new CompactingRecipeGen(output));
		GENERATORS.add(new PressingRecipeGen(output));
		GENERATORS.add(new FillingRecipeGen(output));
		GENERATORS.add(new EmptyingRecipeGen(output));
		GENERATORS.add(new HauntingRecipeGen(output));
		GENERATORS.add(new ItemApplicationRecipeGen(output));

		return new class_2405() {

			@Override
			public String method_10321() {
				return "Create's Processing Recipes";
			}

			@Override
			public CompletableFuture<?> method_10319(class_7403 dc) {
				return CompletableFuture.allOf(GENERATORS.stream()
					.map(gen -> gen.method_10319(dc))
					.toArray(CompletableFuture[]::new));
			}
		};
	}

	public ProcessingRecipeGen(FabricDataOutput generator) {
		super(generator);
	}

	/**
	 * Create a processing recipe with a single itemstack ingredient, using its id
	 * as the name of the recipe
	 */
	protected <T extends ProcessingRecipe<?>> GeneratedRecipe create(String namespace,
		Supplier<class_1935> singleIngredient, UnaryOperator<ProcessingRecipeBuilder<T>> transform) {
		ProcessingRecipeSerializer<T> serializer = getSerializer();
		GeneratedRecipe generatedRecipe = c -> {
			class_1935 itemLike = singleIngredient.get();
			transform
				.apply(new ProcessingRecipeBuilder<>(serializer.getFactory(),
					new class_2960(namespace, RegisteredObjects.getKeyOrThrow(itemLike.method_8389())
						.method_12832())).withItemIngredients(class_1856.method_8091(itemLike)))
				.build(c);
		};
		all.add(generatedRecipe);
		return generatedRecipe;
	}

	/**
	 * Create a processing recipe with a single itemstack ingredient, using its id
	 * as the name of the recipe
	 */
	<T extends ProcessingRecipe<?>> GeneratedRecipe create(Supplier<class_1935> singleIngredient,
		UnaryOperator<ProcessingRecipeBuilder<T>> transform) {
		return create(Create.ID, singleIngredient, transform);
	}

	protected <T extends ProcessingRecipe<?>> GeneratedRecipe createWithDeferredId(Supplier<class_2960> name,
		UnaryOperator<ProcessingRecipeBuilder<T>> transform) {
		ProcessingRecipeSerializer<T> serializer = getSerializer();
		GeneratedRecipe generatedRecipe =
			c -> transform.apply(new ProcessingRecipeBuilder<>(serializer.getFactory(), name.get()))
				.build(c);
		all.add(generatedRecipe);
		return generatedRecipe;
	}

	/**
	 * Create a new processing recipe, with recipe definitions provided by the
	 * function
	 */
	protected <T extends ProcessingRecipe<?>> GeneratedRecipe create(class_2960 name,
		UnaryOperator<ProcessingRecipeBuilder<T>> transform) {
		return createWithDeferredId(() -> name, transform);
	}

	/**
	 * Create a new processing recipe, with recipe definitions provided by the
	 * function
	 */
	<T extends ProcessingRecipe<?>> GeneratedRecipe create(String name,
		UnaryOperator<ProcessingRecipeBuilder<T>> transform) {
		return create(Create.asResource(name), transform);
	}

	protected abstract IRecipeTypeInfo getRecipeType();

	protected <T extends ProcessingRecipe<?>> ProcessingRecipeSerializer<T> getSerializer() {
		return getRecipeType().getSerializer();
	}

	protected Supplier<class_2960> idWithSuffix(Supplier<class_1935> item, String suffix) {
		return () -> {
			class_2960 registryName = RegisteredObjects.getKeyOrThrow(item.get()
				.method_8389());
			return Create.asResource(registryName.method_12832() + suffix);
		};
	}

	@Override
	public String method_10321() {
		return "Create's Processing Recipes: " + getRecipeType().getId()
			.method_12832();
	}

}
