package com.simibubi.create.foundation.placement;

import java.util.ArrayList;
import java.util.List;
import java.util.stream.Collectors;

import org.joml.Matrix4f;
import com.mojang.blaze3d.systems.RenderSystem;
import com.simibubi.create.foundation.gui.AllGuiTextures;
import com.simibubi.create.foundation.utility.AngleHelper;
import com.simibubi.create.foundation.utility.VecHelper;
import com.simibubi.create.foundation.utility.animation.LerpedFloat;
import com.simibubi.create.foundation.utility.animation.LerpedFloat.Chaser;
import com.simibubi.create.infrastructure.config.AllConfigs;
import com.simibubi.create.infrastructure.config.CClient;

import io.github.fabricators_of_create.porting_lib.event.client.OverlayRenderCallback;
import io.github.fabricators_of_create.porting_lib.event.client.OverlayRenderCallback.Types;
import net.fabricmc.api.EnvType;
import net.fabricmc.api.Environment;
import net.minecraft.class_1041;
import net.minecraft.class_1268;
import net.minecraft.class_1657;
import net.minecraft.class_1799;
import net.minecraft.class_2338;
import net.minecraft.class_243;
import net.minecraft.class_2680;
import net.minecraft.class_287;
import net.minecraft.class_289;
import net.minecraft.class_290;
import net.minecraft.class_293;
import net.minecraft.class_310;
import net.minecraft.class_332;
import net.minecraft.class_3965;
import net.minecraft.class_4587;
import net.minecraft.class_638;
import net.minecraft.class_757;
import net.minecraft.class_7833;

public class PlacementHelpers {

	private static final List<IPlacementHelper> helpers = new ArrayList<>();
	private static int animationTick = 0;
	private static final LerpedFloat angle = LerpedFloat.angular()
		.chase(0, 0.25f, Chaser.EXP);
	private static class_2338 target = null;
	private static class_2338 lastTarget = null;

	public static int register(IPlacementHelper helper) {
		helpers.add(helper);
		return helpers.size() - 1;
	}

	public static IPlacementHelper get(int id) {
		if (id < 0 || id >= helpers.size())
			throw new ArrayIndexOutOfBoundsException("id " + id + " for placement helper not known");

		return helpers.get(id);
	}

	@Environment(EnvType.CLIENT)
	public static void tick() {
		setTarget(null);
		checkHelpers();

		if (target == null) {
			if (animationTick > 0)
				animationTick = Math.max(animationTick - 2, 0);

			return;
		}

		if (animationTick < 10)
			animationTick++;

	}

	@Environment(EnvType.CLIENT)
	private static void checkHelpers() {
		class_310 mc = class_310.method_1551();
		class_638 world = mc.field_1687;

		if (world == null)
			return;

		if (!(mc.field_1765 instanceof class_3965))
			return;

		class_3965 ray = (class_3965) mc.field_1765;

		if (mc.field_1724 == null)
			return;

		if (mc.field_1724.method_5715())// for now, disable all helpers when sneaking TODO add helpers that respect
										// sneaking but still show position
			return;

		for (class_1268 hand : class_1268.values()) {

			class_1799 heldItem = mc.field_1724.method_5998(hand);
			List<IPlacementHelper> filteredForHeldItem = helpers.stream()
				.filter(helper -> helper.matchesItem(heldItem))
				.collect(Collectors.toList());
			if (filteredForHeldItem.isEmpty())
				continue;

			class_2338 pos = ray.method_17777();
			class_2680 state = world.method_8320(pos);

			List<IPlacementHelper> filteredForState = filteredForHeldItem.stream()
				.filter(helper -> helper.matchesState(state))
				.collect(Collectors.toList());
			if (filteredForState.isEmpty())
				continue;

			boolean atLeastOneMatch = false;
			for (IPlacementHelper h : filteredForState) {
				PlacementOffset offset = h.getOffset(mc.field_1724, world, state, pos, ray, heldItem);

				if (offset.isSuccessful()) {
					h.renderAt(pos, state, ray, offset);
					setTarget(offset.getBlockPos());
					atLeastOneMatch = true;
					break;
				}

			}

			// at least one helper activated, no need to check the offhand if we are still
			// in the mainhand
			if (atLeastOneMatch)
				return;

		}
	}

	static void setTarget(class_2338 target) {
		PlacementHelpers.target = target;

		if (target == null)
			return;

		if (lastTarget == null) {
			lastTarget = target;
			return;
		}

		if (!lastTarget.equals(target))
			lastTarget = target;
	}

	@Environment(EnvType.CLIENT)
	public static void afterRenderOverlayLayer(class_332 graphics, float partialTicks, class_1041 res) {
		class_310 mc = class_310.method_1551();
		class_1657 player = mc.field_1724;

		if (player != null && animationTick > 0) {
			float screenY = res.method_4502() / 2f;
			float screenX = res.method_4486() / 2f;
			float progress = getCurrentAlpha();

			drawDirectionIndicator(graphics, partialTicks, screenX, screenY, progress);
		}
	}

	public static float getCurrentAlpha() {
		return Math.min(animationTick / 10f/* + event.getPartialTicks() */, 1f);
	}

	@Environment(EnvType.CLIENT)
	private static void drawDirectionIndicator(class_332 graphics, float partialTicks, float centerX, float centerY,
		float progress) {
		float r = .8f;
		float g = .8f;
		float b = .8f;
		float a = progress * progress;

		class_243 projTarget = VecHelper.projectToPlayerView(VecHelper.getCenterOf(lastTarget), partialTicks);

		class_243 target = new class_243(projTarget.field_1352, projTarget.field_1351, 0);
		if (projTarget.field_1350 > 0)
			target = target.method_22882();

		class_243 norm = target.method_1029();
		class_243 ref = new class_243(0, 1, 0);
		float targetAngle = AngleHelper.deg(Math.acos(norm.method_1026(ref)));

		if (norm.field_1352 < 0)
			targetAngle = 360 - targetAngle;

		if (animationTick < 10)
			angle.setValue(targetAngle);

		angle.chase(targetAngle, .25f, Chaser.EXP);
		angle.tickChaser();

		float snapSize = 22.5f;
		float snappedAngle = (snapSize * Math.round(angle.getValue(0f) / snapSize)) % 360f;

		float length = 10;

		CClient.PlacementIndicatorSetting mode = AllConfigs.client().placementIndicator.get();
		class_4587 ms = graphics.method_51448();
		if (mode == CClient.PlacementIndicatorSetting.TRIANGLE)
			fadedArrow(ms, centerX, centerY, r, g, b, a, length, snappedAngle);
		else if (mode == CClient.PlacementIndicatorSetting.TEXTURE)
			textured(ms, centerX, centerY, a, snappedAngle);
	}

	private static void fadedArrow(class_4587 ms, float centerX, float centerY, float r, float g, float b, float a,
		float length, float snappedAngle) {
//		RenderSystem.disableTexture();
		RenderSystem.enableBlend();
		RenderSystem.defaultBlendFunc();
		RenderSystem.setShader(class_757::method_34540);

		ms.method_22903();
		ms.method_46416(centerX, centerY, 5);
		ms.method_22907(class_7833.field_40718.rotationDegrees(angle.getValue(0)));
		// RenderSystem.rotatef(snappedAngle, 0, 0, 1);
		double scale = AllConfigs.client().indicatorScale.get();
		ms.method_22905((float) scale, (float) scale, 1);

		class_289 tessellator = class_289.method_1348();
		class_287 bufferbuilder = tessellator.method_1349();
		bufferbuilder.method_1328(class_293.class_5596.field_27381, class_290.field_1576);

		Matrix4f mat = ms.method_23760()
			.method_23761();

		bufferbuilder.method_22918(mat, 0, -(10 + length), 0)
			.method_22915(r, g, b, a)
			.method_1344();

		bufferbuilder.method_22918(mat, -9, -3, 0)
			.method_22915(r, g, b, 0f)
			.method_1344();
		bufferbuilder.method_22918(mat, -6, -6, 0)
			.method_22915(r, g, b, 0f)
			.method_1344();
		bufferbuilder.method_22918(mat, -3, -8, 0)
			.method_22915(r, g, b, 0f)
			.method_1344();
		bufferbuilder.method_22918(mat, 0, -8.5f, 0)
			.method_22915(r, g, b, 0f)
			.method_1344();
		bufferbuilder.method_22918(mat, 3, -8, 0)
			.method_22915(r, g, b, 0f)
			.method_1344();
		bufferbuilder.method_22918(mat, 6, -6, 0)
			.method_22915(r, g, b, 0f)
			.method_1344();
		bufferbuilder.method_22918(mat, 9, -3, 0)
			.method_22915(r, g, b, 0f)
			.method_1344();

		tessellator.method_1350();
		RenderSystem.disableBlend();
//		RenderSystem.enableTexture();
		ms.method_22909();
	}

	@Environment(EnvType.CLIENT)
	public static void textured(class_4587 ms, float centerX, float centerY, float alpha, float snappedAngle) {
//		RenderSystem.enableTexture();
		AllGuiTextures.PLACEMENT_INDICATOR_SHEET.bind();
		RenderSystem.enableDepthTest();
		RenderSystem.enableBlend();
		RenderSystem.defaultBlendFunc();
		RenderSystem.setShader(class_757::method_34541);

		ms.method_22903();
		ms.method_46416(centerX, centerY, 50);
		float scale = AllConfigs.client().indicatorScale.get()
			.floatValue() * .75f;
		ms.method_22905(scale, scale, 1);
		ms.method_22905(12, 12, 1);

		float index = snappedAngle / 22.5f;
		float tex_size = 16f / 256f;

		float tx = 0;
		float ty = index * tex_size;
		float tw = 1f;
		float th = tex_size;

		class_289 tessellator = class_289.method_1348();
		class_287 buffer = tessellator.method_1349();
		buffer.method_1328(class_293.class_5596.field_27382, class_290.field_20887);

		Matrix4f mat = ms.method_23760()
			.method_23761();
		buffer.method_22918(mat, -1, -1, 0)
			.method_22915(1f, 1f, 1f, alpha)
			.method_22913(tx, ty)
			.method_1344();
		buffer.method_22918(mat, -1, 1, 0)
			.method_22915(1f, 1f, 1f, alpha)
			.method_22913(tx, ty + th)
			.method_1344();
		buffer.method_22918(mat, 1, 1, 0)
			.method_22915(1f, 1f, 1f, alpha)
			.method_22913(tx + tw, ty + th)
			.method_1344();
		buffer.method_22918(mat, 1, -1, 0)
			.method_22915(1f, 1f, 1f, alpha)
			.method_22913(tx + tw, ty)
			.method_1344();

		tessellator.method_1350();

		RenderSystem.disableBlend();
		ms.method_22909();
	}

}
