package com.simibubi.create.foundation.ponder;

import java.io.BufferedInputStream;
import java.io.DataInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.zip.GZIPInputStream;
import net.minecraft.class_1937;
import net.minecraft.class_2248;
import net.minecraft.class_2338;
import net.minecraft.class_2487;
import net.minecraft.class_2505;
import net.minecraft.class_2507;
import net.minecraft.class_2960;
import net.minecraft.class_310;
import net.minecraft.class_3298;
import net.minecraft.class_3300;
import net.minecraft.class_3492;
import net.minecraft.class_3499;
import net.minecraft.class_7924;
import com.simibubi.create.Create;
import com.simibubi.create.infrastructure.ponder.PonderIndex;
import com.simibubi.create.infrastructure.ponder.SharedText;

import io.github.fabricators_of_create.porting_lib.util.EnvExecutor;

public class PonderRegistry {

	public static final PonderTagRegistry TAGS = new PonderTagRegistry();
	public static final PonderChapterRegistry CHAPTERS = new PonderChapterRegistry();
	// Map from item IDs to storyboard entries
	public static final Map<class_2960, List<PonderStoryBoardEntry>> ALL = new HashMap<>();

	public static void addStoryBoard(PonderStoryBoardEntry entry) {
		synchronized (ALL) {
			List<PonderStoryBoardEntry> list = ALL.computeIfAbsent(entry.getComponent(), $ -> new ArrayList<>());
			synchronized (list) {
				list.add(entry);
			}
		}
	}

	public static List<PonderScene> compile(class_2960 id) {
		List<PonderStoryBoardEntry> list = ALL.get(id);
		if (list == null) {
			return Collections.emptyList();
		}
		return compile(list);
	}

	public static List<PonderScene> compile(PonderChapter chapter) {
		List<PonderStoryBoardEntry> list = CHAPTERS.getStories(chapter);
		if (list == null) {
			return Collections.emptyList();
		}
		return compile(list);
	}

	public static List<PonderScene> compile(List<PonderStoryBoardEntry> entries) {
		if (PonderIndex.editingModeActive()) {
			PonderLocalization.SHARED.clear();
			SharedText.gatherText();
		}

		List<PonderScene> scenes = new ArrayList<>();

		for (int i = 0; i < entries.size(); i++) {
			PonderStoryBoardEntry sb = entries.get(i);
			class_2960 id = sb.getSchematicLocation();
			class_3499 activeTemplate = loadSchematic(id);
			class_1937 level = EnvExecutor.unsafeRunForDist(
					() -> () -> class_310.method_1551().field_1687,
					() -> () -> {
						throw new IllegalStateException("Cannot compile on a server");
					}
			);
			PonderWorld world = new PonderWorld(class_2338.field_10980, level);
			class_3492 settings = FabricPonderProcessing.makePlaceSettings(id);
			activeTemplate.method_15172(world, class_2338.field_10980, class_2338.field_10980, settings, world.field_9229, class_2248.field_31028);
			world.createBackup();
			PonderScene scene = compileScene(i, sb, world);
			scene.begin();
			scenes.add(scene);
		}

		return scenes;
	}

	public static PonderScene compileScene(int i, PonderStoryBoardEntry sb, PonderWorld world) {
		PonderScene scene = new PonderScene(world, sb.getNamespace(), sb.getComponent(), sb.getTags());
		SceneBuilder builder = scene.builder();
		sb.getBoard()
				.program(builder, scene.getSceneBuildingUtil());
		return scene;
	}

	public static class_3499 loadSchematic(class_2960 location) {
		return loadSchematic(class_310.method_1551().method_1478(), location);
	}

	public static class_3499 loadSchematic(class_3300 resourceManager, class_2960 location) {
		String namespace = location.method_12836();
		String path = "ponder/" + location.method_12832() + ".nbt";
		class_2960 location1 = new class_2960(namespace, path);

		Optional<class_3298> optionalResource = resourceManager.method_14486(location1);
		if (optionalResource.isPresent()) {
			class_3298 resource = optionalResource.get();
			try (InputStream inputStream = resource.method_14482()) {
				return loadSchematic(inputStream);
			} catch (IOException e) {
				Create.LOGGER.error("Failed to read ponder schematic: " + location1, e);
			}
		} else {
			Create.LOGGER.error("Ponder schematic missing: " + location1);
		}
		return new class_3499();
	}

	public static class_3499 loadSchematic(InputStream resourceStream) throws IOException {
		class_3499 t = new class_3499();
		DataInputStream stream =
				new DataInputStream(new BufferedInputStream(new GZIPInputStream(resourceStream)));
		class_2487 nbt = class_2507.method_10625(stream, new class_2505(0x20000000L));
		t.method_15183(class_310.method_1551().field_1687.method_45448(class_7924.field_41254), nbt);
		return t;
	}

}
