package com.simibubi.create.foundation.ponder;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.Set;
import net.minecraft.class_1935;
import net.minecraft.class_2960;
import com.google.common.collect.ImmutableSet;
import com.google.common.collect.LinkedHashMultimap;
import com.google.common.collect.Multimap;
import com.simibubi.create.foundation.utility.RegisteredObjects;
import com.tterrag.registrate.util.entry.ItemProviderEntry;

public class PonderTagRegistry {

	private final Multimap<class_2960, PonderTag> tags;
	private final Multimap<PonderChapter, PonderTag> chapterTags;

	private final List<PonderTag> listedTags;

	public PonderTagRegistry() {
		tags = LinkedHashMultimap.create();
		chapterTags = LinkedHashMultimap.create();
		listedTags = new ArrayList<>();
	}

	public Set<PonderTag> getTags(class_2960 item) {
		return ImmutableSet.copyOf(tags.get(item));
	}

	public Set<PonderTag> getTags(PonderChapter chapter) {
		return ImmutableSet.copyOf(chapterTags.get(chapter));
	}

	public Set<class_2960> getItems(PonderTag tag) {
		return tags.entries()
			.stream()
			.filter(e -> e.getValue() == tag)
			.map(Map.Entry::getKey)
			.collect(ImmutableSet.toImmutableSet());
	}

	public Set<PonderChapter> getChapters(PonderTag tag) {
		return chapterTags.entries()
			.stream()
			.filter(e -> e.getValue() == tag)
			.map(Map.Entry::getKey)
			.collect(ImmutableSet.toImmutableSet());
	}

	public List<PonderTag> getListedTags() {
		return listedTags;
	}

	public void listTag(PonderTag tag) {
		listedTags.add(tag);
	}

	public void add(PonderTag tag, class_2960 item) {
		synchronized (tags) {
			tags.put(item, tag);
		}
	}

	public void add(PonderTag tag, PonderChapter chapter) {
		synchronized (chapterTags) {
			chapterTags.put(chapter, tag);
		}
	}

	public ItemBuilder forItems(class_2960... items) {
		return new ItemBuilder(items);
	}

	public TagBuilder forTag(PonderTag tag) {
		return new TagBuilder(tag);
	}

	public class ItemBuilder {

		private final Collection<class_2960> items;

		private ItemBuilder(class_2960... items) {
			this.items = Arrays.asList(items);
		}

		public ItemBuilder add(PonderTag tag) {
			items.forEach(i -> PonderTagRegistry.this.add(tag, i));
			return this;
		}

	}

	public class TagBuilder {

		private final PonderTag tag;

		private TagBuilder(PonderTag tag) {
			this.tag = tag;
		}

		public TagBuilder add(class_2960 item) {
			PonderTagRegistry.this.add(tag, item);
			return this;
		}

		public TagBuilder add(class_1935 item) {
			return add(RegisteredObjects.getKeyOrThrow(item.method_8389()));
		}

		public TagBuilder add(ItemProviderEntry<?> entry) {
			return add(entry.get());
		}

	}

}
