package com.simibubi.create.foundation.ponder;

import com.simibubi.create.foundation.utility.VecHelper;
import net.minecraft.class_2338;
import net.minecraft.class_2350;
import net.minecraft.class_2382;
import net.minecraft.class_243;
import net.minecraft.class_3341;

/**
 * Helpful shortcuts for marking boundaries, points or sections inside the scene
 */
public class SceneBuildingUtil {

	public final SelectionUtil select;
	public final VectorUtil vector;
	public final PositionUtil grid;

	private final class_3341 sceneBounds;

	SceneBuildingUtil(class_3341 sceneBounds) {
		this.sceneBounds = sceneBounds;
		this.select = new SelectionUtil();
		this.vector = new VectorUtil();
		this.grid = new PositionUtil();
	}

	public class PositionUtil {

		public class_2338 at(int x, int y, int z) {
			return new class_2338(x, y, z);
		}

		public class_2338 zero() {
			return at(0, 0, 0);
		}

	}

	public class VectorUtil {

		public class_243 centerOf(int x, int y, int z) {
			return centerOf(grid.at(x, y, z));
		}

		public class_243 centerOf(class_2338 pos) {
			return VecHelper.getCenterOf(pos);
		}

		public class_243 topOf(int x, int y, int z) {
			return blockSurface(grid.at(x, y, z), class_2350.field_11036);
		}

		public class_243 topOf(class_2338 pos) {
			return blockSurface(pos, class_2350.field_11036);
		}

		public class_243 blockSurface(class_2338 pos, class_2350 face) {
			return blockSurface(pos, face, 0);
		}

		public class_243 blockSurface(class_2338 pos, class_2350 face, float margin) {
			return centerOf(pos).method_1019(class_243.method_24954(face.method_10163())
				.method_1021(.5f + margin));
		}

		public class_243 of(double x, double y, double z) {
			return new class_243(x, y, z);
		}

	}

	public class SelectionUtil {

		public Selection everywhere() {
			return Selection.of(sceneBounds);
		}

		public Selection position(int x, int y, int z) {
			return position(grid.at(x, y, z));
		}

		public Selection position(class_2338 pos) {
			return cuboid(pos, class_2338.field_10980);
		}

		public Selection fromTo(int x, int y, int z, int x2, int y2, int z2) {
			return fromTo(new class_2338(x, y, z), new class_2338(x2, y2, z2));
		}

		public Selection fromTo(class_2338 pos1, class_2338 pos2) {
			return cuboid(pos1, pos2.method_10059(pos1));
		}

		public Selection column(int x, int z) {
			return cuboid(new class_2338(x, 1, z), new class_2382(0, sceneBounds.method_14660(), 0));
		}

		public Selection layer(int y) {
			return layers(y, 1);
		}

		public Selection layersFrom(int y) {
			return layers(y, sceneBounds.method_14660() - y);
		}

		public Selection layers(int y, int height) {
			return cuboid(new class_2338(0, y, 0), new class_2382(sceneBounds.method_35414() - 1,
				Math.min(sceneBounds.method_14660() - y, height) - 1, sceneBounds.method_14663() - 1));
		}

		public Selection cuboid(class_2338 origin, class_2382 size) {
			return Selection.of(class_3341.method_34390(origin, origin.method_10081(size)));
		}

	}

}