package com.simibubi.create.foundation.ponder.element;

import java.util.function.Supplier;
import net.minecraft.class_1041;
import net.minecraft.class_1299;
import net.minecraft.class_1453;
import net.minecraft.class_1453.class_7989;
import net.minecraft.class_2338;
import net.minecraft.class_243;
import net.minecraft.class_2487;
import net.minecraft.class_2586;
import net.minecraft.class_310;
import net.minecraft.class_3532;
import net.minecraft.class_4587;
import net.minecraft.class_4597;
import net.minecraft.class_898;
import com.jozufozu.flywheel.util.transform.TransformStack;
import com.simibubi.create.Create;
import com.simibubi.create.content.kinetics.base.KineticBlockEntity;
import com.simibubi.create.foundation.ponder.PonderScene;
import com.simibubi.create.foundation.ponder.PonderWorld;
import com.simibubi.create.foundation.ponder.ui.PonderUI;
import com.simibubi.create.foundation.utility.AngleHelper;

public class ParrotElement extends AnimatedSceneElement {

	private class_243 location;
	private class_1453 entity;
	private ParrotPose pose;
	private boolean deferConductor = false;
	private Supplier<? extends ParrotPose> initialPose;

	public static ParrotElement create(class_243 location, Supplier<? extends ParrotPose> pose) {
		return new ParrotElement(location, pose);
	}

	protected ParrotElement(class_243 location, Supplier<? extends ParrotPose> pose) {
		this.location = location;
		initialPose = pose;
		setPose(initialPose.get());
	}

	@Override
	public void reset(PonderScene scene) {
		super.reset(scene);
		setPose(initialPose.get());
		entity.method_23327(0, 0, 0);
		entity.field_6014 = 0;
		entity.field_6036 = 0;
		entity.field_5969 = 0;
		entity.field_6038 = 0;
		entity.field_5971 = 0;
		entity.field_5989 = 0;
		entity.method_36457(entity.field_6004 = 0);
		entity.method_36456(entity.field_5982 = 180);
		entity.getCustomData()
			.method_10551("TrainHat");
		deferConductor = false;
	}

	@Override
	public void tick(PonderScene scene) {
		super.tick(scene);
		if (entity == null) {
			entity = pose.create(scene.getWorld());
			entity.method_36456(entity.field_5982 = 180);
			if (deferConductor)
				setConductor(deferConductor);
			deferConductor = false;
		}

		entity.field_6012++;
		entity.field_6259 = entity.field_6241;
		entity.field_6827 = entity.field_6819;
		entity.field_6829 = entity.field_6818;
		entity.method_24830(true);

		entity.field_6014 = entity.method_23317();
		entity.field_6036 = entity.method_23318();
		entity.field_5969 = entity.method_23321();
		entity.field_5982 = entity.method_36454();
		entity.field_6004 = entity.method_36455();

		pose.tick(scene, entity, location);

		entity.field_6038 = entity.method_23317();
		entity.field_5971 = entity.method_23318();
		entity.field_5989 = entity.method_23321();
	}

	public void setPositionOffset(class_243 position, boolean immediate) {
		if (entity == null)
			return;
		entity.method_5814(position.field_1352, position.field_1351, position.field_1350);
		if (!immediate)
			return;
		entity.field_6014 = position.field_1352;
		entity.field_6036 = position.field_1351;
		entity.field_5969 = position.field_1350;
	}

	public void setRotation(class_243 eulers, boolean immediate) {
		if (entity == null)
			return;
		entity.method_36457((float) eulers.field_1352);
		entity.method_36456((float) eulers.field_1351);
		if (!immediate)
			return;
		entity.field_6004 = entity.method_36455();
		entity.field_5982 = entity.method_36454();
	}

	public void setConductor(boolean isConductor) {
		if (entity == null) {
			deferConductor = isConductor;
			return;
		}
		class_2487 data = entity.getCustomData();
		if (isConductor)
			data.method_10556("TrainHat", true);
		else
			data.method_10551("TrainHat");
	}

	public class_243 getPositionOffset() {
		return entity != null ? entity.method_19538() : class_243.field_1353;
	}

	public class_243 getRotation() {
		return entity != null ? new class_243(entity.method_36455(), entity.method_36454(), 0) : class_243.field_1353;
	}

	@Override
	protected void renderLast(PonderWorld world, class_4597 buffer, class_4587 ms, float fade, float pt) {
		class_898 entityrenderermanager = class_310.method_1551()
			.method_1561();

		if (entity == null) {
			entity = pose.create(world);
			entity.method_36456(entity.field_5982 = 180);
		}

		ms.method_22903();
		ms.method_22904(location.field_1352, location.field_1351, location.field_1350);
		ms.method_22904(class_3532.method_16436(pt, entity.field_6014, entity.method_23317()), class_3532.method_16436(pt, entity.field_6036, entity.method_23318()),
			class_3532.method_16436(pt, entity.field_5969, entity.method_23321()));

		TransformStack.cast(ms)
			.rotateY(AngleHelper.angleLerp(pt, entity.field_5982, entity.method_36454()));

		entityrenderermanager.method_3954(entity, 0, 0, 0, 0, pt, ms, buffer, lightCoordsFromFade(fade));
		ms.method_22909();
	}

	public void setPose(ParrotPose pose) {
		this.pose = pose;
	}

	public static abstract class ParrotPose {

		abstract void tick(PonderScene scene, class_1453 entity, class_243 location);

		class_1453 create(PonderWorld world) {
			class_1453 entity = new class_1453(class_1299.field_6104, world);
			class_7989[] variants = class_1453.class_7989.values();
			class_1453.class_7989 variant = variants[Create.RANDOM.nextInt(variants.length)];
			entity.method_47848(variant == class_7989.field_41551 ? class_7989.field_41550 : variant); // blue parrots are difficult to see
			return entity;
		}

	}

	public static class DancePose extends ParrotPose {

		@Override
		class_1453 create(PonderWorld world) {
			class_1453 entity = super.create(world);
			entity.method_6006(class_2338.field_10980, true);
			return entity;
		}

		@Override
		void tick(PonderScene scene, class_1453 entity, class_243 location) {
			entity.field_5982 = entity.method_36454();
			entity.method_36456(entity.method_36454() - 2);
		}

	}

	public static class FlappyPose extends ParrotPose {

		@Override
		void tick(PonderScene scene, class_1453 entity, class_243 location) {
			double length = entity.method_19538()
				.method_1023(entity.field_6038, entity.field_5971, entity.field_5989)
				.method_1033();
			entity.method_24830(false);
			double phase = Math.min(length * 15, 8);
			float f = (float) ((PonderUI.ponderTicks % 100) * phase);
			entity.field_6819 = class_3532.method_15374(f) + 1;
			if (length == 0)
				entity.field_6819 = 0;
		}

	}

	public static class SpinOnComponentPose extends ParrotPose {

		private class_2338 componentPos;

		public SpinOnComponentPose(class_2338 componentPos) {
			this.componentPos = componentPos;
		}

		@Override
		void tick(PonderScene scene, class_1453 entity, class_243 location) {
			class_2586 blockEntity = scene.getWorld()
				.method_8321(componentPos);
			if (!(blockEntity instanceof KineticBlockEntity))
				return;
			float rpm = ((KineticBlockEntity) blockEntity).getSpeed();
			entity.field_5982 = entity.method_36454();
			entity.method_36456(entity.method_36454() + (rpm * .3f));
		}

	}

	public static abstract class FaceVecPose extends ParrotPose {

		@Override
		void tick(PonderScene scene, class_1453 entity, class_243 location) {
			class_243 p_200602_2_ = getFacedVec(scene);
			class_243 Vector3d = location.method_1019(entity.method_5836(0));
			double d0 = p_200602_2_.field_1352 - Vector3d.field_1352;
			double d1 = p_200602_2_.field_1351 - Vector3d.field_1351;
			double d2 = p_200602_2_.field_1350 - Vector3d.field_1350;
			double d3 = class_3532.method_15355((float) (d0 * d0 + d2 * d2));
			float targetPitch = class_3532.method_15393((float) -(class_3532.method_15349(d1, d3) * (double) (180F / (float) Math.PI)));
			float targetYaw = class_3532.method_15393((float) -(class_3532.method_15349(d2, d0) * (double) (180F / (float) Math.PI)) + 90);

			entity.method_36457(AngleHelper.angleLerp(.4f, entity.method_36455(), targetPitch));
			entity.method_36456(AngleHelper.angleLerp(.4f, entity.method_36454(), targetYaw));
		}

		protected abstract class_243 getFacedVec(PonderScene scene);

	}

	public static class FacePointOfInterestPose extends FaceVecPose {

		@Override
		protected class_243 getFacedVec(PonderScene scene) {
			return scene.getPointOfInterest();
		}

	}

	public static class FaceCursorPose extends FaceVecPose {

		@Override
		protected class_243 getFacedVec(PonderScene scene) {
			class_310 minecraft = class_310.method_1551();
			class_1041 w = minecraft.method_22683();
			double mouseX = minecraft.field_1729.method_1603() * w.method_4486() / w.method_4480();
			double mouseY = minecraft.field_1729.method_1604() * w.method_4502() / w.method_4507();
			return scene.getTransform()
				.screenToScene(mouseX, mouseY, 300, 0);
		}

	}

}
