package com.simibubi.create.foundation.ponder.element;

import java.util.List;
import java.util.function.Supplier;
import net.minecraft.class_241;
import net.minecraft.class_243;
import net.minecraft.class_2583;
import net.minecraft.class_2960;
import net.minecraft.class_332;
import net.minecraft.class_3532;
import net.minecraft.class_4587;
import net.minecraft.class_5348;
import com.simibubi.create.foundation.gui.Theme;
import com.simibubi.create.foundation.gui.element.BoxElement;
import com.simibubi.create.foundation.ponder.PonderLocalization;
import com.simibubi.create.foundation.ponder.PonderPalette;
import com.simibubi.create.foundation.ponder.PonderScene;
import com.simibubi.create.foundation.ponder.PonderScene.SceneTransform;
import com.simibubi.create.foundation.ponder.ui.PonderUI;
import com.simibubi.create.foundation.utility.Color;

public class TextWindowElement extends AnimatedOverlayElement {

	Supplier<String> textGetter = () -> "(?) No text was provided";
	String bakedText;

	// from 0 to 200
	int y;

	class_243 vec;

	boolean nearScene = false;
	int color = PonderPalette.WHITE.getColor();

	public class Builder {

		private PonderScene scene;

		public Builder(PonderScene scene) {
			this.scene = scene;
		}

		public Builder colored(PonderPalette color) {
			TextWindowElement.this.color = color.getColor();
			return this;
		}

		public Builder pointAt(class_243 vec) {
			TextWindowElement.this.vec = vec;
			return this;
		}

		public Builder independent(int y) {
			TextWindowElement.this.y = y;
			return this;
		}

		public Builder independent() {
			return independent(0);
		}

		public Builder text(String defaultText) {
			textGetter = scene.registerText(defaultText);
			return this;
		}

		public Builder sharedText(class_2960 key) {
			textGetter = () -> PonderLocalization.getShared(key);
			return this;
		}

		public Builder sharedText(String key) {
			return sharedText(new class_2960(scene.getNamespace(), key));
		}

		public Builder placeNearTarget() {
			TextWindowElement.this.nearScene = true;
			return this;
		}

		public Builder attachKeyFrame() {
			scene.builder()
				.addLazyKeyframe();
			return this;
		}

	}

	@Override
	protected void render(PonderScene scene, PonderUI screen, class_332 graphics, float partialTicks, float fade) {
		if (bakedText == null)
			bakedText = textGetter.get();
		if (fade < 1 / 16f)
			return;
		SceneTransform transform = scene.getTransform();
		class_241 sceneToScreen = vec != null ? transform.sceneToScreen(vec, partialTicks)
			: new class_241(screen.field_22789 / 2, (screen.field_22790 - 200) / 2 + y - 8);

		boolean settled = transform.xRotation.settled() && transform.yRotation.settled();
		float pY = settled ? (int) sceneToScreen.field_1342 : sceneToScreen.field_1342;

		float yDiff = (screen.field_22790 / 2f - sceneToScreen.field_1342 - 10) / 100f;
		float targetX = (screen.field_22789 * class_3532.method_16439(yDiff * yDiff, 6f / 8, 5f / 8));

		if (nearScene)
			targetX = Math.min(targetX, sceneToScreen.field_1343 + 50);

		if (settled)
			targetX = (int) targetX;

		int textWidth = (int) Math.min(screen.field_22789 - targetX, 180);

		List<class_5348> lines = screen.getFontRenderer()
			.method_27527()
			.method_27498(bakedText, textWidth, class_2583.field_24360);

		int boxWidth = 0;
		for (class_5348 line : lines)
			boxWidth = Math.max(boxWidth, screen.getFontRenderer()
				.method_27525(line));

		int boxHeight = screen.getFontRenderer()
			.method_1713(bakedText, boxWidth);

		class_4587 ms = graphics.method_51448();
		ms.method_22903();
		ms.method_46416(0, pY, 400);

		new BoxElement().withBackground(Theme.c(Theme.Key.PONDER_BACKGROUND_FLAT))
			.gradientBorder(Theme.p(Theme.Key.TEXT_WINDOW_BORDER))
			.at(targetX - 10, 3, 100)
			.withBounds(boxWidth, boxHeight - 1)
			.render(graphics);

		//PonderUI.renderBox(ms, targetX - 10, 3, boxWidth, boxHeight - 1, 0xaa000000, 0x30eebb00, 0x10eebb00);

		int brighterColor = Color.mixColors(color, 0xFFffffdd, 1 / 2f);
		brighterColor = (0x00ffffff & brighterColor) | 0xff000000;
		if (vec != null) {
			ms.method_22903();
			ms.method_46416(sceneToScreen.field_1343, 0, 0);
			double lineTarget = (targetX - sceneToScreen.field_1343) * fade;
			ms.method_22905((float) lineTarget, 1, 1);
			graphics.method_33284(0, 0, 1, 1, -100, brighterColor, brighterColor);
			graphics.method_33284(0, 1, 1, 2, -100, 0xFF494949, 0xFF393939);
			ms.method_22909();
		}

		ms.method_46416(0, 0, 400);
		for (int i = 0; i < lines.size(); i++) {
			graphics.method_51433(screen.getFontRenderer(), lines.get(i)
				.getString(), (int) (targetX - 10), 3 + 9 * i,
				new Color(brighterColor).scaleAlpha(fade)
					.getRGB(),
				false);
		}
		ms.method_22909();
	}

	public int getColor() {
		return color;
	}

}
