package com.simibubi.create.foundation.ponder.element;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.function.Consumer;

import com.jozufozu.flywheel.core.model.ModelUtil;
import com.jozufozu.flywheel.core.model.ShadeSeparatedBufferedData;
import com.jozufozu.flywheel.core.model.ShadeSeparatingVertexConsumer;
import com.jozufozu.flywheel.fabric.model.LayerFilteringBakedModel;
import com.jozufozu.flywheel.util.transform.TransformStack;
import com.simibubi.create.CreateClient;
import com.simibubi.create.foundation.outliner.AABBOutline;
import com.simibubi.create.foundation.ponder.PonderScene;
import com.simibubi.create.foundation.ponder.PonderWorld;
import com.simibubi.create.foundation.ponder.Selection;
import com.simibubi.create.foundation.render.BlockEntityRenderHelper;
import com.simibubi.create.foundation.render.SuperByteBuffer;
import com.simibubi.create.foundation.render.SuperByteBufferCache;
import com.simibubi.create.foundation.render.SuperRenderTypeBuffer;
import com.simibubi.create.foundation.utility.AnimationTickHolder;
import com.simibubi.create.foundation.utility.Pair;
import com.simibubi.create.foundation.utility.VecHelper;

import net.fabricmc.fabric.api.renderer.v1.model.FabricBakedModel;
import net.minecraft.class_1087;
import net.minecraft.class_1088;
import net.minecraft.class_1921;
import net.minecraft.class_1937;
import net.minecraft.class_2248;
import net.minecraft.class_2338;
import net.minecraft.class_2343;
import net.minecraft.class_2350.class_2351;
import net.minecraft.class_243;
import net.minecraft.class_2464;
import net.minecraft.class_2586;
import net.minecraft.class_265;
import net.minecraft.class_2680;
import net.minecraft.class_287;
import net.minecraft.class_290;
import net.minecraft.class_293;
import net.minecraft.class_310;
import net.minecraft.class_3532;
import net.minecraft.class_3610;
import net.minecraft.class_3726;
import net.minecraft.class_3959;
import net.minecraft.class_3965;
import net.minecraft.class_4583;
import net.minecraft.class_4587;
import net.minecraft.class_4588;
import net.minecraft.class_4597;
import net.minecraft.class_4608;
import net.minecraft.class_4696;
import net.minecraft.class_5558;
import net.minecraft.class_5819;
import net.minecraft.class_776;
import net.minecraft.class_778;

public class WorldSectionElement extends AnimatedSceneElement {

	public static final SuperByteBufferCache.Compartment<Pair<Integer, Integer>> DOC_WORLD_SECTION =
			new SuperByteBufferCache.Compartment<>();

	private static final ThreadLocal<ThreadLocalObjects> THREAD_LOCAL_OBJECTS = ThreadLocal.withInitial(ThreadLocalObjects::new);

	List<class_2586> renderedBlockEntities;
	List<Pair<class_2586, Consumer<class_1937>>> tickableBlockEntities;
	Selection section;
	boolean redraw;

	class_243 prevAnimatedOffset = class_243.field_1353;
	class_243 animatedOffset = class_243.field_1353;
	class_243 prevAnimatedRotation = class_243.field_1353;
	class_243 animatedRotation = class_243.field_1353;
	class_243 centerOfRotation = class_243.field_1353;
	class_243 stabilizationAnchor = null;

	class_2338 selectedBlock;

	public WorldSectionElement() {}

	public WorldSectionElement(Selection section) {
		this.section = section.copy();
		centerOfRotation = section.getCenter();
	}

	public void mergeOnto(WorldSectionElement other) {
		setVisible(false);
		if (other.isEmpty())
			other.set(section);
		else
			other.add(section);
	}

	public void set(Selection selection) {
		applyNewSelection(selection.copy());
	}

	public void add(Selection toAdd) {
		applyNewSelection(this.section.add(toAdd));
	}

	public void erase(Selection toErase) {
		applyNewSelection(this.section.substract(toErase));
	}

	private void applyNewSelection(Selection selection) {
		this.section = selection;
		queueRedraw();
	}

	public void setCenterOfRotation(class_243 center) {
		centerOfRotation = center;
	}

	public void stabilizeRotation(class_243 anchor) {
		stabilizationAnchor = anchor;
	}

	@Override
	public void reset(PonderScene scene) {
		super.reset(scene);
		resetAnimatedTransform();
		resetSelectedBlock();
	}

	public void selectBlock(class_2338 pos) {
		selectedBlock = pos;
	}

	public void resetSelectedBlock() {
		selectedBlock = null;
	}

	public void resetAnimatedTransform() {
		prevAnimatedOffset = class_243.field_1353;
		animatedOffset = class_243.field_1353;
		prevAnimatedRotation = class_243.field_1353;
		animatedRotation = class_243.field_1353;
	}

	public void queueRedraw() {
		redraw = true;
	}

	public boolean isEmpty() {
		return section == null;
	}

	public void setEmpty() {
		section = null;
	}

	public void setAnimatedRotation(class_243 eulerAngles, boolean force) {
		this.animatedRotation = eulerAngles;
		if (force)
			prevAnimatedRotation = animatedRotation;
	}

	public class_243 getAnimatedRotation() {
		return animatedRotation;
	}

	public void setAnimatedOffset(class_243 offset, boolean force) {
		this.animatedOffset = offset;
		if (force)
			prevAnimatedOffset = animatedOffset;
	}

	public class_243 getAnimatedOffset() {
		return animatedOffset;
	}

	@Override
	public boolean isVisible() {
		return super.isVisible() && !isEmpty();
	}

	class WorldSectionRayTraceResult {
		class_243 actualHitVec;
		class_2338 worldPos;
	}

	public Pair<class_243, class_3965> rayTrace(PonderWorld world, class_243 source, class_243 target) {
		world.setMask(this.section);
		class_243 transformedTarget = reverseTransformVec(target);
		class_3965 rayTraceBlocks = world.method_17742(new class_3959(reverseTransformVec(source), transformedTarget,
				class_3959.class_3960.field_17559, class_3959.class_242.field_1348, null));
		world.clearMask();

		if (rayTraceBlocks == null)
			return null;
		if (rayTraceBlocks.method_17784() == null)
			return null;

		double t = rayTraceBlocks.method_17784()
				.method_1020(transformedTarget)
				.method_1027()
				/ source.method_1020(target)
				.method_1027();
		class_243 actualHit = VecHelper.lerp((float) t, target, source);
		return Pair.of(actualHit, rayTraceBlocks);
	}

	private class_243 reverseTransformVec(class_243 in) {
		float pt = AnimationTickHolder.getPartialTicks();
		in = in.method_1020(VecHelper.lerp(pt, prevAnimatedOffset, animatedOffset));
		if (!animatedRotation.equals(class_243.field_1353) || !prevAnimatedRotation.equals(class_243.field_1353)) {
			if (centerOfRotation == null)
				centerOfRotation = section.getCenter();
			double rotX = class_3532.method_16436(pt, prevAnimatedRotation.field_1352, animatedRotation.field_1352);
			double rotZ = class_3532.method_16436(pt, prevAnimatedRotation.field_1350, animatedRotation.field_1350);
			double rotY = class_3532.method_16436(pt, prevAnimatedRotation.field_1351, animatedRotation.field_1351);
			in = in.method_1020(centerOfRotation);
			in = VecHelper.rotate(in, -rotX, class_2351.field_11048);
			in = VecHelper.rotate(in, -rotZ, class_2351.field_11051);
			in = VecHelper.rotate(in, -rotY, class_2351.field_11052);
			in = in.method_1019(centerOfRotation);
			if (stabilizationAnchor != null) {
				in = in.method_1020(stabilizationAnchor);
				in = VecHelper.rotate(in, rotX, class_2351.field_11048);
				in = VecHelper.rotate(in, rotZ, class_2351.field_11051);
				in = VecHelper.rotate(in, rotY, class_2351.field_11052);
				in = in.method_1019(stabilizationAnchor);
			}
		}
		return in;
	}

	public void transformMS(class_4587 ms, float pt) {
		TransformStack.cast(ms)
				.translate(VecHelper.lerp(pt, prevAnimatedOffset, animatedOffset));
		if (!animatedRotation.equals(class_243.field_1353) || !prevAnimatedRotation.equals(class_243.field_1353)) {
			if (centerOfRotation == null)
				centerOfRotation = section.getCenter();
			double rotX = class_3532.method_16436(pt, prevAnimatedRotation.field_1352, animatedRotation.field_1352);
			double rotZ = class_3532.method_16436(pt, prevAnimatedRotation.field_1350, animatedRotation.field_1350);
			double rotY = class_3532.method_16436(pt, prevAnimatedRotation.field_1351, animatedRotation.field_1351);
			TransformStack.cast(ms)
					.translate(centerOfRotation)
					.rotateX(rotX)
					.rotateZ(rotZ)
					.rotateY(rotY)
					.translateBack(centerOfRotation);
			if (stabilizationAnchor != null) {
				TransformStack.cast(ms)
						.translate(stabilizationAnchor)
						.rotateX(-rotX)
						.rotateZ(-rotZ)
						.rotateY(-rotY)
						.translateBack(stabilizationAnchor);
			}
		}
	}

	public void tick(PonderScene scene) {
		prevAnimatedOffset = animatedOffset;
		prevAnimatedRotation = animatedRotation;
		if (!isVisible())
			return;
		loadBEsIfMissing(scene.getWorld());
		renderedBlockEntities.removeIf(be -> scene.getWorld()
				.method_8321(be.method_11016()) != be);
		tickableBlockEntities.removeIf(be -> scene.getWorld()
				.method_8321(be.getFirst()
						.method_11016()) != be.getFirst());
		tickableBlockEntities.forEach(be -> be.getSecond()
				.accept(scene.getWorld()));
	}

	@Override
	public void whileSkipping(PonderScene scene) {
		if (redraw) {
			renderedBlockEntities = null;
			tickableBlockEntities = null;
		}
		redraw = false;
	}

	protected void loadBEsIfMissing(PonderWorld world) {
		if (renderedBlockEntities != null)
			return;
		tickableBlockEntities = new ArrayList<>();
		renderedBlockEntities = new ArrayList<>();
		section.forEach(pos -> {
			class_2586 blockEntity = world.method_8321(pos);
			class_2680 blockState = world.method_8320(pos);
			class_2248 block = blockState.method_26204();
			if (blockEntity == null)
				return;
			if (!(block instanceof class_2343))
				return;
			blockEntity.method_31664(world.method_8320(pos));
			class_5558<?> ticker = ((class_2343) block).method_31645(world, blockState, blockEntity.method_11017());
			if (ticker != null)
				addTicker(blockEntity, ticker);
			renderedBlockEntities.add(blockEntity);
		});
	}

	@SuppressWarnings("unchecked")
	private <T extends class_2586> void addTicker(T blockEntity, class_5558<?> ticker) {
		tickableBlockEntities.add(Pair.of(blockEntity, w -> ((class_5558<T>) ticker).tick(w,
				blockEntity.method_11016(), blockEntity.method_11010(), blockEntity)));
	}

	@Override
	public void renderFirst(PonderWorld world, class_4597 buffer, class_4587 ms, float fade, float pt) {
		int light = -1;
		if (fade != 1)
			light = (int) (class_3532.method_16439(fade, 5, 14));
		if (redraw) {
			renderedBlockEntities = null;
			tickableBlockEntities = null;
		}

		ms.method_22903();
		transformMS(ms, pt);
		world.pushFakeLight(light);
		renderBlockEntities(world, ms, buffer, pt);
		world.popLight();

		Map<class_2338, Integer> blockBreakingProgressions = world.getBlockBreakingProgressions();
		class_4587 overlayMS = null;

		class_776 renderer = class_310.method_1551().method_1541();
		for (Entry<class_2338, Integer> entry : blockBreakingProgressions.entrySet()) {
			class_2338 pos = entry.getKey();
			if (!section.test(pos))
				continue;

			if (overlayMS == null) {
				overlayMS = new class_4587();
				overlayMS.method_23760().method_23761().set(ms.method_23760().method_23761());
				overlayMS.method_23760().method_23762().set(ms.method_23760().method_23762());
			}

			class_4588 builder = new class_4583(
					buffer.getBuffer(class_1088.field_21772.get(entry.getValue())), overlayMS.method_23760()
					.method_23761(),
					overlayMS.method_23760()
							.method_23762(),
				1);

			ms.method_22903();
			ms.method_46416(pos.method_10263(), pos.method_10264(), pos.method_10260());
			renderer
				.method_23071(world.method_8320(pos), pos, world, ms, builder);
			ms.method_22909();
		}

		ms.method_22909();
	}

	@Override
	protected void renderLayer(PonderWorld world, class_4597 buffer, class_1921 type, class_4587 ms, float fade,
							   float pt) {
		SuperByteBufferCache bufferCache = CreateClient.BUFFER_CACHE;

		int code = hashCode() ^ world.hashCode();
		Pair<Integer, Integer> key = Pair.of(code, class_1921.method_22720()
				.indexOf(type));

		if (redraw)
			bufferCache.invalidate(DOC_WORLD_SECTION, key);
		SuperByteBuffer contraptionBuffer =
				bufferCache.get(DOC_WORLD_SECTION, key, () -> buildStructureBuffer(world, type));
		if (contraptionBuffer.isEmpty())
			return;

		transformMS(contraptionBuffer.getTransforms(), pt);
		int light = lightCoordsFromFade(fade);
		contraptionBuffer
				.light(light)
				.renderInto(ms, buffer.getBuffer(type));
	}

	@Override
	protected void renderLast(PonderWorld world, class_4597 buffer, class_4587 ms, float fade, float pt) {
		redraw = false;
		if (selectedBlock == null)
			return;
		class_2680 blockState = world.method_8320(selectedBlock);
		if (blockState.method_26215())
			return;
		class_265 shape =
				blockState.method_26172(world, selectedBlock, class_3726.method_16195(class_310.method_1551().field_1724));
		if (shape.method_1110())
			return;

		ms.method_22903();
		transformMS(ms, pt);
		ms.method_46416(selectedBlock.method_10263(), selectedBlock.method_10264(), selectedBlock.method_10260());

		AABBOutline aabbOutline = new AABBOutline(shape.method_1107());
		aabbOutline.getParams()
				.lineWidth(1 / 64f)
				.colored(0xefefef)
				.disableLineNormals();
		aabbOutline.render(ms, (SuperRenderTypeBuffer) buffer, class_243.field_1353, pt);

		ms.method_22909();
	}

	private void renderBlockEntities(PonderWorld world, class_4587 ms, class_4597 buffer, float pt) {
		loadBEsIfMissing(world);
		BlockEntityRenderHelper.renderBlockEntities(world, renderedBlockEntities, ms, buffer, pt);
	}

	private SuperByteBuffer buildStructureBuffer(PonderWorld world, class_1921 layer) {
		class_776 dispatcher = class_310.method_1551().method_1541();
		ThreadLocalObjects objects = THREAD_LOCAL_OBJECTS.get();

		class_4587 poseStack = objects.poseStack;
		class_5819 random = objects.random;
		ShadeSeparatingVertexConsumer shadeSeparatingWrapper = objects.shadeSeparatingWrapper;
		class_287 shadedBuilder = objects.shadedBuilder;
		class_287 unshadedBuilder = objects.unshadedBuilder;

		shadedBuilder.method_1328(class_293.class_5596.field_27382, class_290.field_1590);
		unshadedBuilder.method_1328(class_293.class_5596.field_27382, class_290.field_1590);
		shadeSeparatingWrapper.prepare(shadedBuilder, unshadedBuilder);

		world.setMask(this.section);
		class_778.method_20544();
		section.forEach(pos -> {
			class_2680 state = world.method_8320(pos);
			class_3610 fluidState = world.method_8316(pos);

			poseStack.method_22903();
			poseStack.method_46416(pos.method_10263(), pos.method_10264(), pos.method_10260());

			if (state.method_26217() == class_2464.field_11458) {
				class_1087 model = dispatcher.method_3349(state);
				if (model.isVanillaAdapter()) {
					if (class_4696.method_23679(state) != layer) {
						model = null;
					}
				} else {
					model = LayerFilteringBakedModel.wrap(model, layer);
				}
				if (model != null) {
					model = shadeSeparatingWrapper.wrapModel(model);
					dispatcher.method_3350()
						.method_3374(world, model, state, pos, poseStack, shadeSeparatingWrapper, true, random, state.method_26190(pos), class_4608.field_21444);
				}
			}

			if (!fluidState.method_15769() && class_4696.method_23680(fluidState) == layer)
				dispatcher.method_3352(pos, world, shadedBuilder, state, fluidState);

			poseStack.method_22909();
		});
		class_778.method_20545();
		world.clearMask();

		shadeSeparatingWrapper.clear();
		ShadeSeparatedBufferedData bufferedData = ModelUtil.endAndCombine(shadedBuilder, unshadedBuilder);

		SuperByteBuffer sbb = new SuperByteBuffer(bufferedData);
		bufferedData.release();
		return sbb;
	}

	private static class ThreadLocalObjects {
		public final class_4587 poseStack = new class_4587();
		public final class_5819 random = class_5819.method_43053();
		public final ShadeSeparatingVertexConsumer shadeSeparatingWrapper = new ShadeSeparatingVertexConsumer();
		public final class_287 shadedBuilder = new class_287(512);
		public final class_287 unshadedBuilder = new class_287(512);
	}

}
