package com.simibubi.create.foundation.ponder.ui;

import com.simibubi.create.foundation.gui.Theme;
import com.simibubi.create.foundation.gui.element.BoxElement;
import com.simibubi.create.foundation.gui.widget.AbstractSimiWidget;
import com.simibubi.create.foundation.ponder.PonderScene;
import com.simibubi.create.foundation.utility.animation.LerpedFloat;

import io.github.fabricators_of_create.porting_lib.util.client.ScreenUtils;
import net.minecraft.class_1144;
import net.minecraft.class_310;
import net.minecraft.class_327;
import net.minecraft.class_332;
import net.minecraft.class_4587;

public class PonderProgressBar extends AbstractSimiWidget {

	LerpedFloat progress;

	PonderUI ponder;

	public PonderProgressBar(PonderUI ponder, int xIn, int yIn, int widthIn, int heightIn) {
		super(xIn, yIn, widthIn, heightIn);

		this.ponder = ponder;
		progress = LerpedFloat.linear()
				.startWithValue(0);
	}

	public void tick() {
		progress.chase(ponder.getActiveScene()
				.getSceneProgress(), .5f, LerpedFloat.Chaser.EXP);
		progress.tickChaser();
	}

	@Override
	protected boolean method_25361(double mouseX, double mouseY) {
		return this.field_22763 && this.field_22764 && ponder.getActiveScene().getKeyframeCount() > 0
				&& mouseX >= (double) this.method_46426() && mouseX < (double) (this.method_46426() + this.field_22758 + 4) && mouseY >= (double) this.method_46427() - 3
				&& mouseY < (double) (this.method_46427() + this.field_22759 + 20);
	}

	@Override
	public void method_25348(double mouseX, double mouseY) {
		PonderScene activeScene = ponder.getActiveScene();

		int keyframeIndex = getHoveredKeyframeIndex(activeScene, mouseX);

		if (keyframeIndex == -1)
			ponder.seekToTime(0);
		else if (keyframeIndex == activeScene.getKeyframeCount())
			ponder.seekToTime(activeScene.getTotalTime());
		else
			ponder.seekToTime(activeScene.getKeyframeTime(keyframeIndex));
	}

	public int getHoveredKeyframeIndex(PonderScene activeScene, double mouseX) {
		int totalTime = activeScene.getTotalTime();
		int clickedAtTime = (int) ((mouseX - method_46426()) / ((double) field_22758 + 4) * totalTime);

		{
			int lastKeyframeTime = activeScene.getKeyframeTime(activeScene.getKeyframeCount() - 1);

			int diffToEnd = totalTime - clickedAtTime;
			int diffToLast = clickedAtTime - lastKeyframeTime;

			if (diffToEnd > 0 && diffToEnd < diffToLast / 2) {
				return activeScene.getKeyframeCount();
			}
		}

		int index = -1;

		for (int i = 0; i < activeScene.getKeyframeCount(); i++) {
			int keyframeTime = activeScene.getKeyframeTime(i);

			if (keyframeTime > clickedAtTime)
				break;

			index = i;
		}

		return index;
	}

	@Override
	public void doRender(class_332 graphics, int mouseX, int mouseY, float partialTicks) {
		class_4587 ms = graphics.method_51448();

		field_22762 = method_25361(mouseX, mouseY);

		new BoxElement()
				.withBackground(Theme.c(Theme.Key.PONDER_BACKGROUND_FLAT))
				.gradientBorder(Theme.p(Theme.Key.PONDER_IDLE))
				.at(method_46426(), method_46427(), 400)
				.withBounds(field_22758, field_22759)
				.render(graphics);

		ms.method_22903();
		ms.method_46416(method_46426() - 2, method_46427() - 2, 100);

		ms.method_22903();
		ms.method_22905((field_22758 + 4) * progress.getValue(partialTicks), 1, 1);
		int c1 = Theme.i(Theme.Key.PONDER_PROGRESSBAR, true);
		int c2 = Theme.i(Theme.Key.PONDER_PROGRESSBAR, false);
		graphics.method_33284(0, 3, 1, 4, 310, c1, c1);
		graphics.method_33284(0, 4, 1, 5, 310, c2, c2);
		ms.method_22909();

		renderKeyframes(graphics, mouseX, partialTicks);

		ms.method_22909();
	}

	private void renderKeyframes(class_332 graphics, int mouseX, float partialTicks) {
		PonderScene activeScene = ponder.getActiveScene();

		int hoverStartColor = Theme.i(Theme.Key.PONDER_HOVER, true) | 0xa0_000000;
		int hoverEndColor = Theme.i(Theme.Key.PONDER_HOVER, false) | 0xa0_000000;
		int idleStartColor = Theme.i(Theme.Key.PONDER_IDLE, true) | 0x40_000000;
		int idleEndColor = Theme.i(Theme.Key.PONDER_IDLE, false) | 0x40_000000;
		int hoverIndex;

		if (field_22762) {
			hoverIndex = getHoveredKeyframeIndex(activeScene, mouseX);
		} else {
			hoverIndex = -2;
		}

		if (hoverIndex == -1)
			drawKeyframe(graphics, activeScene, true, 0, 0, hoverStartColor, hoverEndColor, 8);
		else if (hoverIndex == activeScene.getKeyframeCount())
			drawKeyframe(graphics, activeScene, true, activeScene.getTotalTime(), field_22758 + 4, hoverStartColor, hoverEndColor, 8);

		for (int i = 0; i < activeScene.getKeyframeCount(); i++) {
			int keyframeTime = activeScene.getKeyframeTime(i);
			int keyframePos = (int) (((float) keyframeTime) / ((float) activeScene.getTotalTime()) * (field_22758 + 4));

			boolean selected = i == hoverIndex;
			int startColor = selected ? hoverStartColor : idleStartColor;
			int endColor = selected ? hoverEndColor : idleEndColor;
			int height = selected ? 8 : 4;

			drawKeyframe(graphics, activeScene, selected, keyframeTime, keyframePos, startColor, endColor, height);

		}
	}

	private void drawKeyframe(class_332 graphics, PonderScene activeScene, boolean selected, int keyframeTime,
		int keyframePos, int startColor, int endColor, int height) {
		class_4587 ms = graphics.method_51448();
		if (selected) {
			class_327 font = class_310.method_1551().field_1772;
			graphics.method_33284(keyframePos, 10, keyframePos + 1, 10 + height, 600, endColor, startColor);
			ms.method_22903();
			ms.method_46416(0, 0, 200);
			String text;
			int offset;
			if (activeScene.getCurrentTime() < keyframeTime) {
				text = ">";
				offset = -1 - font.method_1727(text);
			} else {
				text = "<";
				offset = 3;
			}
			graphics.method_51433(font, text, keyframePos + offset, 10, endColor, false);
			ms.method_22909();
		}

		graphics.method_33284(keyframePos, -1, keyframePos + 1, 2 + height, 400, startColor, endColor);
	}

	@Override
	public void method_25354(class_1144 handler) {

	}
}
