package com.simibubi.create.foundation.ponder.ui;

import static com.simibubi.create.foundation.ponder.PonderLocalization.LANG_PREFIX;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.stream.IntStream;

import io.github.fabricators_of_create.porting_lib.mixin.accessors.client.accessor.ScreenAccessor;

import org.joml.Matrix4f;
import com.mojang.blaze3d.systems.RenderSystem;
import com.simibubi.create.Create;
import com.simibubi.create.foundation.gui.AllGuiTextures;
import com.simibubi.create.foundation.gui.AllIcons;
import com.simibubi.create.foundation.gui.ScreenOpener;
import com.simibubi.create.foundation.gui.Theme;
import com.simibubi.create.foundation.gui.UIRenderHelper;
import com.simibubi.create.foundation.gui.element.BoxElement;
import com.simibubi.create.foundation.gui.element.GuiGameElement;
import com.simibubi.create.foundation.ponder.FabricPonderProcessing;
import com.simibubi.create.foundation.item.ItemHelper;
import com.simibubi.create.foundation.ponder.PonderChapter;
import com.simibubi.create.foundation.ponder.PonderRegistry;
import com.simibubi.create.foundation.ponder.PonderScene;
import com.simibubi.create.foundation.ponder.PonderScene.SceneTransform;
import com.simibubi.create.foundation.ponder.PonderStoryBoardEntry;
import com.simibubi.create.foundation.ponder.PonderTag;
import com.simibubi.create.foundation.ponder.PonderWorld;
import com.simibubi.create.foundation.ponder.element.TextWindowElement;
import com.simibubi.create.foundation.render.SuperRenderTypeBuffer;
import com.simibubi.create.foundation.utility.Color;
import com.simibubi.create.foundation.utility.Components;
import com.simibubi.create.foundation.utility.Couple;
import com.simibubi.create.foundation.utility.FontHelper;
import com.simibubi.create.foundation.utility.Iterate;
import com.simibubi.create.foundation.utility.Lang;
import com.simibubi.create.foundation.utility.Pair;
import com.simibubi.create.foundation.utility.Pointing;
import com.simibubi.create.foundation.utility.RegisteredObjects;
import com.simibubi.create.foundation.utility.animation.LerpedFloat;
import com.simibubi.create.foundation.utility.animation.LerpedFloat.Chaser;
import com.simibubi.create.infrastructure.config.AllConfigs;
import com.simibubi.create.infrastructure.ponder.DebugScenes;
import com.simibubi.create.infrastructure.ponder.PonderIndex;

import io.github.fabricators_of_create.porting_lib.util.client.ScreenUtils;
import net.fabricmc.fabric.api.client.keybinding.v1.KeyBindingHelper;
import net.minecraft.class_1041;
import net.minecraft.class_124;
import net.minecraft.class_1799;
import net.minecraft.class_2248;
import net.minecraft.class_2338;
import net.minecraft.class_2350;
import net.minecraft.class_243;
import net.minecraft.class_2960;
import net.minecraft.class_310;
import net.minecraft.class_315;
import net.minecraft.class_327;
import net.minecraft.class_332;
import net.minecraft.class_3341;
import net.minecraft.class_3492;
import net.minecraft.class_3499;
import net.minecraft.class_3532;
import net.minecraft.class_364;
import net.minecraft.class_3674;
import net.minecraft.class_4587;
import net.minecraft.class_5250;
import net.minecraft.class_5819;
import net.minecraft.class_7833;
import net.minecraft.class_7923;
import net.minecraft.class_8251;

public class PonderUI extends NavigatableSimiScreen {

	public static int ponderTicks;
	public static float ponderPartialTicksPaused;

	public static final String PONDERING = LANG_PREFIX + "pondering";
	public static final String IDENTIFY_MODE = LANG_PREFIX + "identify_mode";
	public static final String IN_CHAPTER = LANG_PREFIX + "in_chapter";
	public static final String IDENTIFY = LANG_PREFIX + "identify";
	public static final String PREVIOUS = LANG_PREFIX + "previous";
	public static final String CLOSE = LANG_PREFIX + "close";
	public static final String NEXT = LANG_PREFIX + "next";
	public static final String NEXT_UP = LANG_PREFIX + "next_up";
	public static final String REPLAY = LANG_PREFIX + "replay";
	public static final String SLOW_TEXT = LANG_PREFIX + "slow_text";

	private List<PonderScene> scenes;
	private List<PonderTag> tags;
	private List<PonderButton> tagButtons;
	private List<LerpedFloat> tagFades;
	private LerpedFloat fadeIn;
	class_1799 stack;
	PonderChapter chapter = null;

	private boolean userViewMode;
	private boolean identifyMode;
	private class_1799 hoveredTooltipItem;
	private class_2338 hoveredBlockPos;

	private class_3674 clipboardHelper;
	private class_2338 copiedBlockPos;

	private LerpedFloat finishingFlash;
	private LerpedFloat nextUp;
	private int finishingFlashWarmup = 0;
	private int nextUpWarmup = 0;

	private LerpedFloat lazyIndex;
	private int index = 0;
	private PonderTag referredToByTag;

	private PonderButton left, right, scan, chap, userMode, close, replay, slowMode;
	private int skipCooling = 0;

	private int extendedTickLength = 0;
	private int extendedTickTimer = 0;

	public static PonderUI of(class_2960 id) {
		return new PonderUI(PonderRegistry.compile(id));
	}

	public static PonderUI of(class_1799 item) {
		return new PonderUI(PonderRegistry.compile(RegisteredObjects.getKeyOrThrow(item.method_7909())));
	}

	public static PonderUI of(class_1799 item, PonderTag tag) {
		PonderUI ponderUI = new PonderUI(PonderRegistry.compile(RegisteredObjects.getKeyOrThrow(item.method_7909())));
		ponderUI.referredToByTag = tag;
		return ponderUI;
	}

	public static PonderUI of(PonderChapter chapter) {
		PonderUI ui = new PonderUI(PonderRegistry.compile(chapter));
		ui.chapter = chapter;
		return ui;
	}

	protected PonderUI(List<PonderScene> scenes) {
		class_2960 component = scenes.get(0)
			.getComponent();
		if (class_7923.field_41178.method_10250(component))
			stack = new class_1799(class_7923.field_41178.method_10223(component));
		else
			stack = new class_1799(class_7923.field_41175.method_10223(component));

		tags = new ArrayList<>(PonderRegistry.TAGS.getTags(component));
		this.scenes = scenes;
		if (scenes.isEmpty()) {
			List<PonderStoryBoardEntry> l = Collections.singletonList(new PonderStoryBoardEntry(DebugScenes::empty,
				Create.ID, "debug/scene_1", new class_2960("minecraft", "stick")));
			scenes.addAll(PonderRegistry.compile(l));
		}
		lazyIndex = LerpedFloat.linear()
			.startWithValue(index);
		fadeIn = LerpedFloat.linear()
			.startWithValue(0)
			.chase(1, .1f, Chaser.EXP);
		clipboardHelper = new class_3674();
		finishingFlash = LerpedFloat.linear()
			.startWithValue(0)
			.chase(0, .1f, Chaser.EXP);
		nextUp = LerpedFloat.linear()
			.startWithValue(0)
			.chase(0, .4f, Chaser.EXP);
	}

	@Override
	protected void method_25426() {
		super.method_25426();

		tagButtons = new ArrayList<>();
		tagFades = new ArrayList<>();

		tags.forEach(t -> {
			int i = tagButtons.size();
			int x = 31;
			int y = 81 + i * 30;

			PonderButton b2 = new PonderButton(x, y).showing(t)
				.withCallback((mX, mY) -> {
					centerScalingOn(mX, mY);
					ScreenOpener.transitionTo(new PonderTagScreen(t));
				});

			method_37063(b2);
			tagButtons.add(b2);

			LerpedFloat chase = LerpedFloat.linear()
				.startWithValue(0)
				.chase(0, .05f, Chaser.exp(.1));
			tagFades.add(chase);

		});

		/*
		 * if (chapter != null) { widgets.add(chap = new PonderButton(width - 31 - 24,
		 * 31, () -> { }).showing(chapter)); }
		 */

		class_315 bindings = field_22787.field_1690;
		int spacing = 8;
		int bX = (field_22789 - 20) / 2 - (70 + 2 * spacing);
		int bY = field_22790 - 20 - 31;

		{
			int pX = (field_22789 / 2) - 110;
			int pY = bY + 20 + 4;
			int pW = field_22789 - 2 * pX;
			method_37063(new PonderProgressBar(this, pX, pY, pW, 1));
		}

		method_37063(scan = new PonderButton(bX, bY).withShortcut(bindings.field_1869)
			.showing(AllIcons.I_MTD_SCAN)
			.enableFade(0, 5)
			.withCallback(() -> {
				identifyMode = !identifyMode;
				if (!identifyMode)
					scenes.get(index)
						.deselect();
				else
					ponderPartialTicksPaused = field_22787.method_1488();
			}));
		scan.atZLevel(600);

		method_37063(slowMode = new PonderButton(field_22789 - 20 - 31, bY).showing(AllIcons.I_MTD_SLOW_MODE)
			.enableFade(0, 5)
			.withCallback(() -> setComfyReadingEnabled(!isComfyReadingEnabled())));

		if (PonderIndex.editingModeActive()) {
			method_37063(userMode = new PonderButton(field_22789 - 50 - 31, bY).showing(AllIcons.I_MTD_USER_MODE)
				.enableFade(0, 5)
				.withCallback(() -> userViewMode = !userViewMode));
		}

		bX += 50 + spacing;
		method_37063(left = new PonderButton(bX, bY).withShortcut(bindings.field_1913)
			.showing(AllIcons.I_MTD_LEFT)
			.enableFade(0, 5)
			.withCallback(() -> this.scroll(false)));

		bX += 20 + spacing;
		method_37063(close = new PonderButton(bX, bY).withShortcut(bindings.field_1822)
			.showing(AllIcons.I_MTD_CLOSE)
			.enableFade(0, 5)
			.withCallback(this::method_25419));

		bX += 20 + spacing;
		method_37063(right = new PonderButton(bX, bY).withShortcut(bindings.field_1849)
			.showing(AllIcons.I_MTD_RIGHT)
			.enableFade(0, 5)
			.withCallback(() -> this.scroll(true)));

		bX += 50 + spacing;
		method_37063(replay = new PonderButton(bX, bY).withShortcut(bindings.field_1881)
			.showing(AllIcons.I_MTD_REPLAY)
			.enableFade(0, 5)
			.withCallback(this::replay));
	}

	@Override
	protected void initBackTrackIcon(PonderButton backTrack) {
		backTrack.showing(stack);
	}

	@Override
	public void method_25393() {
		super.method_25393();

		if (skipCooling > 0)
			skipCooling--;

		if (referredToByTag != null) {
			for (int i = 0; i < scenes.size(); i++) {
				PonderScene ponderScene = scenes.get(i);
				if (!ponderScene.getTags()
					.contains(referredToByTag))
					continue;
				if (i == index)
					break;
				scenes.get(index)
					.fadeOut();
				index = i;
				scenes.get(index)
					.begin();
				lazyIndex.chase(index, 1 / 4f, Chaser.EXP);
				identifyMode = false;
				break;
			}
			referredToByTag = null;
		}

		lazyIndex.tickChaser();
		fadeIn.tickChaser();
		finishingFlash.tickChaser();
		nextUp.tickChaser();
		PonderScene activeScene = scenes.get(index);

		extendedTickLength = 0;
		if (isComfyReadingEnabled())
			activeScene.forEachVisible(TextWindowElement.class, twe -> extendedTickLength = 2);

		if (extendedTickTimer == 0) {
			if (!identifyMode) {
				ponderTicks++;
				if (skipCooling == 0)
					activeScene.tick();
			}

			if (!identifyMode) {
				float lazyIndexValue = lazyIndex.getValue();
				if (Math.abs(lazyIndexValue - index) > 1 / 512f)
					scenes.get(lazyIndexValue < index ? index - 1 : index + 1)
						.tick();
			}
			extendedTickTimer = extendedTickLength;
		} else
			extendedTickTimer--;

		if (activeScene.getCurrentTime() == activeScene.getTotalTime() - 1) {
			finishingFlashWarmup = 30;
			nextUpWarmup = 50;
		}

		if (finishingFlashWarmup > 0) {
			finishingFlashWarmup--;
			if (finishingFlashWarmup == 0) {
				finishingFlash.setValue(1);
				finishingFlash.setValue(1);
			}
		}

		if (nextUpWarmup > 0) {
			nextUpWarmup--;
			if (nextUpWarmup == 0)
				nextUp.updateChaseTarget(1);
		}

		updateIdentifiedItem(activeScene);
	}

	public PonderScene getActiveScene() {
		return scenes.get(index);
	}

	public void seekToTime(int time) {
		if (getActiveScene().getCurrentTime() > time)
			replay();

		getActiveScene().seekToTime(time);
		if (time != 0)
			coolDownAfterSkip();
	}

	public void updateIdentifiedItem(PonderScene activeScene) {
		hoveredTooltipItem = class_1799.field_8037;
		hoveredBlockPos = null;
		if (!identifyMode)
			return;

		class_1041 w = field_22787.method_22683();
		double mouseX = field_22787.field_1729.method_1603() * w.method_4486() / w.method_4480();
		double mouseY = field_22787.field_1729.method_1604() * w.method_4502() / w.method_4507();
		SceneTransform t = activeScene.getTransform();
		class_243 vec1 = t.screenToScene(mouseX, mouseY, 1000, 0);
		class_243 vec2 = t.screenToScene(mouseX, mouseY, -100, 0);
		Pair<class_1799, class_2338> pair = activeScene.rayTraceScene(vec1, vec2);
		hoveredTooltipItem = pair.getFirst();
		hoveredBlockPos = pair.getSecond();
	}

	@Override
	public boolean method_25401(double mouseX, double mouseY, double delta) {
		if (scroll(delta > 0))
			return true;
		return super.method_25401(mouseX, mouseY, delta);
	}

	protected void replay() {
		identifyMode = false;
		PonderScene scene = scenes.get(index);

		if (method_25442()) {
			List<PonderStoryBoardEntry> list = PonderRegistry.ALL.get(scene.getComponent());
			PonderStoryBoardEntry sb = list.get(index);
			class_2960 id = sb.getSchematicLocation();
			class_3499 activeTemplate = PonderRegistry.loadSchematic(id);
			PonderWorld world = new PonderWorld(class_2338.field_10980, class_310.method_1551().field_1687);
			class_3492 settings = FabricPonderProcessing.makePlaceSettings(id);
			activeTemplate.method_15172(world, class_2338.field_10980, class_2338.field_10980, settings, class_5819.method_43047(),
				class_2248.field_31028);
			world.createBackup();
			scene = PonderRegistry.compileScene(index, sb, world);
			scene.begin();
			scenes.set(index, scene);
		}

		scene.begin();
	}

	protected boolean scroll(boolean forward) {
		int prevIndex = index;
		index = forward ? index + 1 : index - 1;
		index = class_3532.method_15340(index, 0, scenes.size() - 1);
		if (prevIndex != index) {// && Math.abs(index - lazyIndex.getValue()) < 1.5f) {
			scenes.get(prevIndex)
				.fadeOut();
			scenes.get(index)
				.begin();
			lazyIndex.chase(index, 1 / 4f, Chaser.EXP);
			identifyMode = false;
			return true;
		} else
			index = prevIndex;
		return false;
	}

	@Override
	protected void renderWindow(class_332 graphics, int mouseX, int mouseY, float partialTicks) {
		partialTicks = getPartialTicks();
		RenderSystem.enableBlend();
		renderVisibleScenes(graphics, mouseX, mouseY,
			skipCooling > 0 ? 0 : identifyMode ? ponderPartialTicksPaused : partialTicks);
		renderWidgets(graphics, mouseX, mouseY, identifyMode ? ponderPartialTicksPaused : partialTicks);
	}

	@Override
	public void method_25420(class_332 graphics) {
		super.method_25420(graphics);
	}

	protected void renderVisibleScenes(class_332 graphics, int mouseX, int mouseY, float partialTicks) {
		renderScene(graphics, mouseX, mouseY, index, partialTicks);
		float lazyIndexValue = lazyIndex.getValue(partialTicks);
		if (Math.abs(lazyIndexValue - index) > 1 / 512f)
			renderScene(graphics, mouseX, mouseY, lazyIndexValue < index ? index - 1 : index + 1, partialTicks);
	}

	protected void renderScene(class_332 graphics, int mouseX, int mouseY, int i, float partialTicks) {
		SuperRenderTypeBuffer buffer = SuperRenderTypeBuffer.getInstance();
		PonderScene scene = scenes.get(i);
		double value = lazyIndex.getValue(field_22787.method_1488());
		double diff = i - value;
		double slide = class_3532.method_16436(diff * diff, 200, 600) * diff;

		RenderSystem.enableBlend();
		RenderSystem.enableDepthTest();
		RenderSystem.backupProjectionMatrix();

		// has to be outside of MS transforms, important for vertex sorting
		Matrix4f matrix4f = new Matrix4f(RenderSystem.getProjectionMatrix());
		matrix4f.translate(0, 0, 800);
		RenderSystem.setProjectionMatrix(matrix4f, class_8251.field_43360);

		class_4587 ms = graphics.method_51448();
		ms.method_22903();
		ms.method_46416(0, 0, -800);

		scene.getTransform()
			.updateScreenParams(field_22789, field_22790, slide);
		scene.getTransform()
			.apply(ms, partialTicks);

//		ms.translate(-story.getBasePlateOffsetX() * .5, 0, -story.getBasePlateOffsetZ() * .5);

		scene.getTransform()
			.updateSceneRVE(partialTicks);

		scene.renderScene(buffer, ms, partialTicks);
		buffer.draw();

		class_3341 bounds = scene.getBounds();
		ms.method_22903();

		// kool shadow fx
		if (!scene.shouldHidePlatformShadow()) {
			RenderSystem.enableCull();
			RenderSystem.enableDepthTest();
			ms.method_22903();
			ms.method_46416(scene.getBasePlateOffsetX(), 0, scene.getBasePlateOffsetZ());
			UIRenderHelper.flipForGuiRender(ms);

			float flash = finishingFlash.getValue(partialTicks) * .9f;
			float alpha = flash;
			flash *= flash;
			flash = ((flash * 2) - 1);
			flash *= flash;
			flash = 1 - flash;

			for (int f = 0; f < 4; f++) {
				ms.method_46416(scene.getBasePlateSize(), 0, 0);
				ms.method_22903();
				ms.method_46416(0, 0, -1 / 1024f);
				if (flash > 0) {
					ms.method_22903();
					ms.method_22905(1, .5f + flash * .75f, 1);
					graphics.method_33284(0, -1, -scene.getBasePlateSize(), 0, 0, 0x00_c6ffc9,
						new Color(0xaa_c6ffc9).scaleAlpha(alpha)
							.getRGB());
					ms.method_22909();
				}
				ms.method_46416(0, 0, 2 / 1024f);
				graphics.method_33284(0, 0, -scene.getBasePlateSize(), 4, 0, 0x66_000000, 0x00_000000);
				ms.method_22909();
				ms.method_22907(class_7833.field_40716.rotationDegrees(-90));
			}
			ms.method_22909();
			RenderSystem.disableCull();
			RenderSystem.disableDepthTest();
		}

		// coords for debug
		if (PonderIndex.editingModeActive() && !userViewMode) {

			ms.method_22905(-1, -1, 1);
			ms.method_22905(1 / 16f, 1 / 16f, 1 / 16f);
			ms.method_46416(1, -8, -1 / 64f);

			// X AXIS
			ms.method_22903();
			ms.method_46416(4, -3, 0);
			ms.method_46416(0, 0, -2 / 1024f);
			for (int x = 0; x <= bounds.method_35414(); x++) {
				ms.method_46416(-16, 0, 0);
				graphics.method_51433(field_22793, x == bounds.method_35414() ? "x" : "" + x, 0, 0, 0xFFFFFFFF, false);
			}
			ms.method_22909();

			// Z AXIS
			ms.method_22903();
			ms.method_22905(-1, 1, 1);
			ms.method_46416(0, -3, -4);
			ms.method_22907(class_7833.field_40716.rotationDegrees(-90));
			ms.method_46416(-8, -2, 2 / 64f);
			for (int z = 0; z <= bounds.method_14663(); z++) {
				ms.method_46416(16, 0, 0);
				graphics.method_51433(field_22793, z == bounds.method_14663() ? "z" : "" + z, 0, 0, 0xFFFFFFFF, false);
			}
			ms.method_22909();

			// DIRECTIONS
			ms.method_22903();
			ms.method_46416(bounds.method_35414() * -8, 0, bounds.method_14663() * 8);
			ms.method_22907(class_7833.field_40716.rotationDegrees(-90));
			for (class_2350 d : Iterate.horizontalDirections) {
				ms.method_22907(class_7833.field_40716.rotationDegrees(90));
				ms.method_22903();
				ms.method_46416(0, 0, bounds.method_14663() * 16);
				ms.method_22907(class_7833.field_40714.rotationDegrees(-90));
				graphics.method_51433(field_22793, d.name()
					.substring(0, 1), 0, 0, 0x66FFFFFF, false);
				graphics.method_51433(field_22793, "|", 2, 10, 0x44FFFFFF, false);
				graphics.method_51433(field_22793, ".", 2, 14, 0x22FFFFFF, false);
				ms.method_22909();
			}
			ms.method_22909();
			buffer.draw();
		}

		ms.method_22909();
		ms.method_22909();
		RenderSystem.restoreProjectionMatrix();
	}

	protected void renderWidgets(class_332 graphics, int mouseX, int mouseY, float partialTicks) {
		RenderSystem.disableDepthTest();

		float fade = fadeIn.getValue(partialTicks);
		float lazyIndexValue = lazyIndex.getValue(partialTicks);
		float indexDiff = Math.abs(lazyIndexValue - index);
		PonderScene activeScene = scenes.get(index);
		PonderScene nextScene = scenes.size() > index + 1 ? scenes.get(index + 1) : null;

		boolean noWidgetsHovered = true;
		for (class_364 child : method_25396())
			noWidgetsHovered &= !child.method_25405(mouseX, mouseY);

		int tooltipColor = Theme.i(Theme.Key.TEXT_DARKER);
		renderChapterTitle(graphics, fade, indexDiff, activeScene, tooltipColor);
		renderNavigationMenu(graphics);

		class_4587 ms = graphics.method_51448();

		if (identifyMode) {
			if (noWidgetsHovered && mouseY < field_22790 - 80) {
				ms.method_22903();
				ms.method_46416(mouseX, mouseY, 100);
				if (hoveredTooltipItem.method_7960()) {
					class_5250 text = Lang
						.translateDirect(IDENTIFY_MODE,
							((class_5250) field_22787.field_1690.field_1869.method_16007())
								.method_27692(class_124.field_1068))
						.method_27692(class_124.field_1080);
					graphics.method_51447(field_22793, field_22793.method_1728(text, field_22789 / 3), 0, 0);
				} else
					graphics.method_51446(field_22793, hoveredTooltipItem, 0, 0);
				if (hoveredBlockPos != null && PonderIndex.editingModeActive() && !userViewMode) {
					ms.method_46416(0, -15, 0);
					boolean copied = copiedBlockPos != null && hoveredBlockPos.equals(copiedBlockPos);
					class_5250 coords = Components.literal(
						hoveredBlockPos.method_10263() + ", " + hoveredBlockPos.method_10264() + ", " + hoveredBlockPos.method_10260())
							.method_27692(copied ? class_124.field_1060 : class_124.field_1065);
					graphics.method_51438(field_22793, coords, 0, 0);
				}
				ms.method_22909();
			}
			scan.flash();
		} else {
			scan.dim();
		}

		if (PonderIndex.editingModeActive()) {
			if (userViewMode)
				userMode.flash();
			else
				userMode.dim();
		}

		if (isComfyReadingEnabled())
			slowMode.flash();
		else
			slowMode.dim();

		renderSceneOverlay(graphics, partialTicks, lazyIndexValue, indexDiff);

		boolean finished = activeScene.isFinished();

		if (finished) {
			jumpToNextScene(graphics, partialTicks, nextScene);
		}

		// Widgets
		((ScreenAccessor) this).port_lib$getRenderables().forEach(w -> {
			if (w instanceof PonderButton button) {
				button.fade()
					.startWithValue(fade);
			}
		});

		if (index == 0 || index == 1 && lazyIndexValue < index)
			left.fade()
				.startWithValue(lazyIndexValue);
		if (index == scenes.size() - 1 || index == scenes.size() - 2 && lazyIndexValue > index)
			right.fade()
				.startWithValue(scenes.size() - lazyIndexValue - 1);

		if (finished)
			right.flash();
		else {
			right.dim();
			nextUp.updateChaseTarget(0);
		}

		renderPonderTags(graphics, mouseX, mouseY, partialTicks, fade, activeScene);

		renderHoverTooltips(graphics, tooltipColor);
		RenderSystem.enableDepthTest();
	}

	protected void renderPonderTags(class_332 graphics, int mouseX, int mouseY, float partialTicks, float fade, PonderScene activeScene) {
		class_4587 ms = graphics.method_51448();

		// Tags
		List<PonderTag> sceneTags = activeScene.getTags();
		boolean highlightAll = sceneTags.contains(PonderTag.HIGHLIGHT_ALL);
		double s = class_310.method_1551()
			.method_22683()
			.method_4495();
		IntStream.range(0, tagButtons.size())
			.forEach(i -> {
				ms.method_22903();
				LerpedFloat chase = tagFades.get(i);
				PonderButton button = tagButtons.get(i);
				if (button.method_25405(mouseX, mouseY)) {
					chase.updateChaseTarget(1);
				} else
					chase.updateChaseTarget(0);

				chase.tickChaser();

				if (highlightAll)
					button.flash();
				else
					button.dim();

				int x = button.method_46426() + button.method_25368() + 4;
				int y = button.method_46427() - 2;
				ms.method_46416(x, y + 5 * (1 - fade), 800);

				float fadedWidth = 200 * chase.getValue(partialTicks);
				UIRenderHelper.streak(graphics, 0, 0, 12, 26, (int) fadedWidth);

				RenderSystem.enableScissor((int) (x * s), 0, (int) (fadedWidth * s), (int) (field_22790 * s));

				String tagName = this.tags.get(i)
					.getTitle();
				graphics.method_51433(field_22793, tagName, 3, 8, Theme.i(Theme.Key.TEXT_ACCENT_SLIGHT), false);

				RenderSystem.disableScissor();

				ms.method_22909();
			});
	}

	protected void renderSceneOverlay(class_332 graphics, float partialTicks, float lazyIndexValue, float indexDiff) {
		class_4587 ms = graphics.method_51448();

		// Scene overlay
		float scenePT = skipCooling > 0 ? 0 : partialTicks;
		ms.method_22903();
		ms.method_46416(0, 0, 100);
		renderOverlay(graphics, index, scenePT);
		if (indexDiff > 1 / 512f)
			renderOverlay(graphics, lazyIndexValue < index ? index - 1 : index + 1, scenePT);
		ms.method_22909();
	}

	protected void jumpToNextScene(class_332 graphics, float partialTicks, PonderScene nextScene) {
		class_4587 ms = graphics.method_51448();

		if (nextScene != null && nextUp.getValue() > 1 / 16f && !nextScene.getId()
				.equals(Create.asResource("creative_motor_mojang"))) {
			ms.method_22903();
			ms.method_46416(right.method_46426() + 10, right.method_46427() - 6 + nextUp.getValue(partialTicks) * 5, 400);
			int boxWidth = (Math.max(field_22793.method_1727(nextScene.getTitle()), field_22793.method_27525(Lang.translateDirect(NEXT_UP))) + 5);
			renderSpeechBox(graphics, 0, 0, boxWidth, 20, right.method_25367(), Pointing.DOWN, false);
			ms.method_46416(0, -29, 100);
			graphics.method_27534(field_22793, Lang.translateDirect(NEXT_UP), 0, 0, Theme.i(Theme.Key.TEXT_DARKER));
			graphics.method_25300(field_22793, nextScene.getTitle(), 0, 10, Theme.i(Theme.Key.TEXT));
			ms.method_22909();
		}
	}

	protected void renderHoverTooltips(class_332 graphics, int tooltipColor) {
		class_4587 ms = graphics.method_51448();
		ms.method_22903();
		ms.method_46416(0, 0, 500);
		int tooltipY = field_22790 - 16;
		if (scan.method_25367())
			graphics.method_27534(field_22793, Lang.translateDirect(IDENTIFY), scan.method_46426() + 10, tooltipY, tooltipColor);
		if (index != 0 && left.method_25367())
			graphics.method_27534(field_22793, Lang.translateDirect(PREVIOUS), left.method_46426() + 10, tooltipY, tooltipColor);
		if (close.method_25367())
			graphics.method_27534(field_22793, Lang.translateDirect(CLOSE), close.method_46426() + 10, tooltipY, tooltipColor);
		if (index != scenes.size() - 1 && right.method_25367())
			graphics.method_27534(field_22793, Lang.translateDirect(NEXT), right.method_46426() + 10, tooltipY, tooltipColor);
		if (replay.method_25367())
			graphics.method_27534(field_22793, Lang.translateDirect(REPLAY), replay.method_46426() + 10, tooltipY, tooltipColor);
		if (slowMode.method_25367())
			graphics.method_27534(field_22793, Lang.translateDirect(SLOW_TEXT), slowMode.method_46426() + 5, tooltipY, tooltipColor);
		if (PonderIndex.editingModeActive() && userMode.method_25367())
			graphics.method_25300(field_22793, "Editor View", userMode.method_46426() + 10, tooltipY, tooltipColor);
		ms.method_22909();
	}

	protected void renderChapterTitle(class_332 graphics, float fade, float indexDiff, PonderScene activeScene, int tooltipColor) {
		class_4587 ms = graphics.method_51448();

		// Chapter title
		ms.method_22903();
		ms.method_46416(0, 0, 400);
		int x = 31 + 20 + 8;
		int y = 31;

		String title = activeScene.getTitle();
		int wordWrappedHeight = field_22793.method_1713(title, left.method_46426() - 51);

		int streakHeight = 35 - 9 + wordWrappedHeight;
		UIRenderHelper.streak(graphics, 0, x - 4, y - 12 + streakHeight / 2, streakHeight, (int) (150 * fade));
		UIRenderHelper.streak(graphics, 180, x - 4, y - 12 + streakHeight / 2, streakHeight, (int) (30 * fade));
		new BoxElement().withBackground(Theme.c(Theme.Key.PONDER_BACKGROUND_FLAT))
				.gradientBorder(Theme.p(Theme.Key.PONDER_IDLE))
				.at(21, 21, 100)
				.withBounds(30, 30)
				.render(graphics);

		GuiGameElement.of(stack)
				.scale(2)
				.at(x - 39f, y - 11f)
				.render(graphics);

		graphics.method_51439(field_22793, Lang.translateDirect(PONDERING), x, y - 6, tooltipColor, false);
		y += 8;
		x += 0;
		ms.method_46416(x, y, 0);
		ms.method_22907(class_7833.field_40713.rotationDegrees(indexDiff * -75));
		ms.method_46416(0, 0, 5);
		FontHelper.drawSplitString(ms, field_22793, title, 0, 0, left.method_46426() - 51, Theme.c(Theme.Key.TEXT)
				.scaleAlpha(1 - indexDiff)
				.getRGB());
		ms.method_22909();
		if (chapter != null) {
			ms.method_22903();

			ms.method_46416(chap.method_46426() - 8, chap.method_46427(), 0);
			UIRenderHelper.streak(graphics, 180, 4, 10, 26, (int) (150 * fade));

			drawRightAlignedString(graphics, ms, Lang.translateDirect(IN_CHAPTER)
					.getString(), 0, 0, tooltipColor);
			drawRightAlignedString(graphics, ms, chapter.getTitle(), 0, 12, Theme.i(Theme.Key.TEXT));

			ms.method_22909();
		}
	}

	protected void renderNavigationMenu(class_332 graphics) {
		Color c1 = Theme.c(Theme.Key.PONDER_BACK_ARROW)
			.setAlpha(0x40);
		Color c2 = Theme.c(Theme.Key.PONDER_BACK_ARROW)
			.setAlpha(0x20);
		Color c3 = Theme.c(Theme.Key.PONDER_BACK_ARROW)
			.setAlpha(0x10);
		UIRenderHelper.breadcrumbArrow(graphics, field_22789 / 2 - 20, field_22790 - 51, 0, 20, 20, 5, c1, c2);
		UIRenderHelper.breadcrumbArrow(graphics, field_22789 / 2 + 20, field_22790 - 51, 0, -20, 20, -5, c1, c2);
		UIRenderHelper.breadcrumbArrow(graphics, field_22789 / 2 - 90, field_22790 - 51, 0, 70, 20, 5, c1, c3);
		UIRenderHelper.breadcrumbArrow(graphics, field_22789 / 2 + 90, field_22790 - 51, 0, -70, 20, -5, c1, c3);
	}

	private void renderOverlay(class_332 graphics, int i, float partialTicks) {
		if (identifyMode)
			return;
		class_4587 ms = graphics.method_51448();
		ms.method_22903();
		PonderScene story = scenes.get(i);
		story.renderOverlay(this, graphics, skipCooling > 0 ? 0 : identifyMode ? ponderPartialTicksPaused : partialTicks);
		ms.method_22909();
	}

	@Override
	public boolean method_25402(double x, double y, int button) {
		if (identifyMode && hoveredBlockPos != null && PonderIndex.editingModeActive()) {
			long handle = field_22787.method_22683()
				.method_4490();
			if (copiedBlockPos != null && button == 1) {
				clipboardHelper.method_15979(handle,
					"util.select.fromTo(" + copiedBlockPos.method_10263() + ", " + copiedBlockPos.method_10264() + ", "
						+ copiedBlockPos.method_10260() + ", " + hoveredBlockPos.method_10263() + ", " + hoveredBlockPos.method_10264() + ", "
						+ hoveredBlockPos.method_10260() + ")");
				copiedBlockPos = hoveredBlockPos;
				return true;
			}

			if (method_25442())
				clipboardHelper.method_15979(handle, "util.select.position(" + hoveredBlockPos.method_10263() + ", "
					+ hoveredBlockPos.method_10264() + ", " + hoveredBlockPos.method_10260() + ")");
			else
				clipboardHelper.method_15979(handle, "util.grid.at(" + hoveredBlockPos.method_10263() + ", "
					+ hoveredBlockPos.method_10264() + ", " + hoveredBlockPos.method_10260() + ")");
			copiedBlockPos = hoveredBlockPos;
			return true;
		}

		return super.method_25402(x, y, button);
	}

	@Override
	public boolean method_25404(int code, int p_keyPressed_2_, int p_keyPressed_3_) {
		class_315 settings = class_310.method_1551().field_1690;
		int sCode = KeyBindingHelper.getBoundKeyOf(settings.field_1881)
			.method_1444();
		int aCode = KeyBindingHelper.getBoundKeyOf(settings.field_1913)
			.method_1444();
		int dCode = KeyBindingHelper.getBoundKeyOf(settings.field_1849)
			.method_1444();
		int qCode = KeyBindingHelper.getBoundKeyOf(settings.field_1869)
			.method_1444();

		if (code == sCode) {
			replay();
			return true;
		}

		if (code == aCode) {
			scroll(false);
			return true;
		}

		if (code == dCode) {
			scroll(true);
			return true;
		}

		if (code == qCode) {
			identifyMode = !identifyMode;
			if (!identifyMode)
				scenes.get(index)
					.deselect();
			return true;
		}

		return super.method_25404(code, p_keyPressed_2_, p_keyPressed_3_);
	}

	@Override
	protected String getBreadcrumbTitle() {
		if (chapter != null)
			return chapter.getTitle();

		return stack.method_7909()
			.method_7848()
			.getString();
	}

	public class_327 getFontRenderer() {
		return field_22793;
	}

	protected boolean isMouseOver(double mouseX, double mouseY, int x, int y, int w, int h) {
		boolean hovered = !(mouseX < x || mouseX > x + w);
		hovered &= !(mouseY < y || mouseY > y + h);
		return hovered;
	}

	public static void renderSpeechBox(class_332 graphics, int x, int y, int w, int h, boolean highlighted, Pointing pointing,
		boolean returnWithLocalTransform) {
		class_4587 ms = graphics.method_51448();
		if (!returnWithLocalTransform)
			ms.method_22903();

		int boxX = x;
		int boxY = y;
		int divotX = x;
		int divotY = y;
		int divotRotation = 0;
		int divotSize = 8;
		int distance = 1;
		int divotRadius = divotSize / 2;
		Couple<Color> borderColors = Theme.p(highlighted ? Theme.Key.PONDER_BUTTON_HOVER : Theme.Key.PONDER_IDLE);
		Color c;

		switch (pointing) {
		default:
		case DOWN:
			divotRotation = 0;
			boxX -= w / 2;
			boxY -= h + divotSize + 1 + distance;
			divotX -= divotRadius;
			divotY -= divotSize + distance;
			c = borderColors.getSecond();
			break;
		case LEFT:
			divotRotation = 90;
			boxX += divotSize + 1 + distance;
			boxY -= h / 2;
			divotX += distance;
			divotY -= divotRadius;
			c = Color.mixColors(borderColors, 0.5f);
			break;
		case RIGHT:
			divotRotation = 270;
			boxX -= w + divotSize + 1 + distance;
			boxY -= h / 2;
			divotX -= divotSize + distance;
			divotY -= divotRadius;
			c = Color.mixColors(borderColors, 0.5f);
			break;
		case UP:
			divotRotation = 180;
			boxX -= w / 2;
			boxY += divotSize + 1 + distance;
			divotX -= divotRadius;
			divotY += distance;
			c = borderColors.getFirst();
			break;
		}

		new BoxElement().withBackground(Theme.c(Theme.Key.PONDER_BACKGROUND_FLAT))
			.gradientBorder(borderColors)
			.at(boxX, boxY, 100)
			.withBounds(w, h)
			.render(graphics);

		ms.method_22903();
		ms.method_46416(divotX + divotRadius, divotY + divotRadius, 10);
		ms.method_22907(class_7833.field_40718.rotationDegrees(divotRotation));
		ms.method_46416(-divotRadius, -divotRadius, 0);
		AllGuiTextures.SPEECH_TOOLTIP_BACKGROUND.render(graphics, 0, 0);
		AllGuiTextures.SPEECH_TOOLTIP_COLOR.render(graphics, 0, 0, c);
		ms.method_22909();

		if (returnWithLocalTransform) {
			ms.method_46416(boxX, boxY, 0);
			return;
		}

		ms.method_22909();

	}

	public class_1799 getHoveredTooltipItem() {
		return hoveredTooltipItem;
	}

	public class_1799 getSubject() {
		return stack;
	}

	@Override
	public boolean isEquivalentTo(NavigatableSimiScreen other) {
		if (other instanceof PonderUI)
			return ItemHelper.sameItem(stack, ((PonderUI) other).stack);
		return super.isEquivalentTo(other);
	}

	@Override
	public void shareContextWith(NavigatableSimiScreen other) {
		if (other instanceof PonderUI) {
			PonderUI ponderUI = (PonderUI) other;
			ponderUI.referredToByTag = referredToByTag;
		}
	}

	public static float getPartialTicks() {
		float renderPartialTicks = class_310.method_1551()
			.method_1488();

		if (class_310.method_1551().field_1755 instanceof PonderUI) {
			PonderUI ui = (PonderUI) class_310.method_1551().field_1755;
			if (ui.identifyMode)
				return ponderPartialTicksPaused;

			return (renderPartialTicks + (ui.extendedTickLength - ui.extendedTickTimer)) / (ui.extendedTickLength + 1);
		}

		return renderPartialTicks;
	}

	@Override
	public boolean method_25421() {
		return true;
	}

	public void coolDownAfterSkip() {
		skipCooling = 15;
	}

	@Override
	public void method_25432() {
		super.method_25432();
		hoveredTooltipItem = class_1799.field_8037;
	}

	public void drawRightAlignedString(class_332 graphics, class_4587 ms, String string, int x, int y, int color) {
		graphics.method_51433(field_22793, string, x - field_22793.method_1727(string), y, color, false);
	}

	public boolean isComfyReadingEnabled() {
		return AllConfigs.client().comfyReading.get();
	}

	public void setComfyReadingEnabled(boolean slowTextMode) {
		AllConfigs.client().comfyReading.set(slowTextMode);
	}

}
