package com.simibubi.create.foundation.render;

import java.util.Iterator;

import javax.annotation.Nullable;
import net.minecraft.class_1937;
import net.minecraft.class_2338;
import net.minecraft.class_2586;
import net.minecraft.class_310;
import net.minecraft.class_4587;
import net.minecraft.class_4597;
import net.minecraft.class_4608;
import net.minecraft.class_761;
import net.minecraft.class_827;
import org.joml.Matrix4f;
import org.joml.Vector4f;

import com.jozufozu.flywheel.backend.Backend;
import com.jozufozu.flywheel.backend.instancing.InstancedRenderRegistry;
import com.jozufozu.flywheel.config.BackendType;
import com.jozufozu.flywheel.core.virtual.VirtualRenderWorld;
import com.jozufozu.flywheel.util.transform.TransformStack;
import com.simibubi.create.Create;
import com.simibubi.create.foundation.utility.AnimationTickHolder;
import com.simibubi.create.foundation.utility.RegisteredObjects;
import com.simibubi.create.infrastructure.config.AllConfigs;

public class BlockEntityRenderHelper {

	public static void renderBlockEntities(class_1937 world, Iterable<class_2586> customRenderBEs, class_4587 ms,
			class_4597 buffer) {
		renderBlockEntities(world, null, customRenderBEs, ms, null, buffer);
	}

	public static void renderBlockEntities(class_1937 world, Iterable<class_2586> customRenderBEs, class_4587 ms,
			class_4597 buffer, float pt) {
		renderBlockEntities(world, null, customRenderBEs, ms, null, buffer, pt);
	}

	public static void renderBlockEntities(class_1937 world, @Nullable VirtualRenderWorld renderWorld,
			Iterable<class_2586> customRenderBEs, class_4587 ms, @Nullable Matrix4f lightTransform, class_4597 buffer) {
		renderBlockEntities(world, renderWorld, customRenderBEs, ms, lightTransform, buffer,
			AnimationTickHolder.getPartialTicks());
	}

	public static void renderBlockEntities(class_1937 world, @Nullable VirtualRenderWorld renderWorld,
			Iterable<class_2586> customRenderBEs, class_4587 ms, @Nullable Matrix4f lightTransform, class_4597 buffer,
			float pt) {
		Iterator<class_2586> iterator = customRenderBEs.iterator();
		while (iterator.hasNext()) {
			class_2586 blockEntity = iterator.next();
			if (Backend.getBackendType() == BackendType.INSTANCING && Backend.isFlywheelWorld(renderWorld) && InstancedRenderRegistry.shouldSkipRender(blockEntity))
				continue;

			class_827<class_2586> renderer = class_310.method_1551().method_31975().method_3550(blockEntity);
			if (renderer == null) {
				iterator.remove();
				continue;
			}

			class_2338 pos = blockEntity.method_11016();
			ms.method_22903();
			TransformStack.cast(ms)
				.translate(pos);

			try {
				int worldLight = getCombinedLight(world, getLightPos(lightTransform, pos), renderWorld, pos);

				if (renderWorld != null) {
					// Swap the real world for the render world so that the renderer gets contraption-local information
					blockEntity.method_31662(renderWorld);
					renderer.method_3569(blockEntity, pt, ms, buffer, worldLight, class_4608.field_21444);
					blockEntity.method_31662(world);
				} else {
					renderer.method_3569(blockEntity, pt, ms, buffer, worldLight, class_4608.field_21444);
				}

			} catch (Exception e) {
				iterator.remove();

				String message = "BlockEntity " + RegisteredObjects.getKeyOrThrow(blockEntity.method_11017())
					.toString() + " could not be rendered virtually.";
				if (AllConfigs.client().explainRenderErrors.get())
					Create.LOGGER.error(message, e);
				else
					Create.LOGGER.error(message);
			}

			ms.method_22909();
		}
	}

	private static class_2338 getLightPos(@Nullable Matrix4f lightTransform, class_2338 contraptionPos) {
		if (lightTransform != null) {
			Vector4f lightVec = new Vector4f(contraptionPos.method_10263() + .5f, contraptionPos.method_10264() + .5f, contraptionPos.method_10260() + .5f, 1);
			lightVec.mul(lightTransform);
			return class_2338.method_49637(lightVec.x(), lightVec.y(), lightVec.z());
		} else {
			return contraptionPos;
		}
	}

	public static int getCombinedLight(class_1937 world, class_2338 worldPos, @Nullable VirtualRenderWorld renderWorld,
			class_2338 renderWorldPos) {
		int worldLight = class_761.method_23794(world, worldPos);

		if (renderWorld != null) {
			int renderWorldLight = class_761.method_23794(renderWorld, renderWorldPos);
			return SuperByteBuffer.maxLight(worldLight, renderWorldLight);
		}

		return worldLight;
	}

}
