package com.simibubi.create.foundation.render;

import static net.minecraft.class_2741.field_12525;

import java.util.function.Supplier;
import net.minecraft.class_2350;
import net.minecraft.class_2680;
import net.minecraft.class_4587;
import org.apache.commons.lang3.tuple.Pair;

import com.jozufozu.flywheel.core.PartialModel;
import com.jozufozu.flywheel.util.transform.TransformStack;
import com.simibubi.create.CreateClient;
import com.simibubi.create.foundation.render.SuperByteBufferCache.Compartment;
import com.simibubi.create.foundation.utility.AngleHelper;

public class CachedBufferer {

	public static final Compartment<class_2680> GENERIC_BLOCK = new Compartment<>();
	public static final Compartment<PartialModel> PARTIAL = new Compartment<>();
	public static final Compartment<Pair<class_2350, PartialModel>> DIRECTIONAL_PARTIAL = new Compartment<>();

	public static SuperByteBuffer block(class_2680 toRender) {
		return block(GENERIC_BLOCK, toRender);
	}

	public static SuperByteBuffer block(Compartment<class_2680> compartment, class_2680 toRender) {
		return CreateClient.BUFFER_CACHE.get(compartment, toRender, () -> BakedModelRenderHelper.standardBlockRender(toRender));
	}

	public static SuperByteBuffer partial(PartialModel partial, class_2680 referenceState) {
		return CreateClient.BUFFER_CACHE.get(PARTIAL, partial,
				() -> BakedModelRenderHelper.standardModelRender(partial.get(), referenceState));
	}

	public static SuperByteBuffer partial(PartialModel partial, class_2680 referenceState,
			Supplier<class_4587> modelTransform) {
		return CreateClient.BUFFER_CACHE.get(PARTIAL, partial,
				() -> BakedModelRenderHelper.standardModelRender(partial.get(), referenceState, modelTransform.get()));
	}

	public static SuperByteBuffer partialFacing(PartialModel partial, class_2680 referenceState) {
		class_2350 facing = referenceState.method_11654(field_12525);
		return partialFacing(partial, referenceState, facing);
	}

	public static SuperByteBuffer partialFacing(PartialModel partial, class_2680 referenceState, class_2350 facing) {
		return partialDirectional(partial, referenceState, facing,
			rotateToFace(facing));
	}

	public static SuperByteBuffer partialFacingVertical(PartialModel partial, class_2680 referenceState, class_2350 facing) {
		return partialDirectional(partial, referenceState, facing,
			rotateToFaceVertical(facing));
	}

	public static SuperByteBuffer partialDirectional(PartialModel partial, class_2680 referenceState, class_2350 dir,
			Supplier<class_4587> modelTransform) {
		return CreateClient.BUFFER_CACHE.get(DIRECTIONAL_PARTIAL, Pair.of(dir, partial),
			() -> BakedModelRenderHelper.standardModelRender(partial.get(), referenceState, modelTransform.get()));
	}

	public static Supplier<class_4587> rotateToFace(class_2350 facing) {
		return () -> {
			class_4587 stack = new class_4587();
			TransformStack.cast(stack)
				.centre()
				.rotateY(AngleHelper.horizontalAngle(facing))
				.rotateX(AngleHelper.verticalAngle(facing))
				.unCentre();
			return stack;
		};
	}

	public static Supplier<class_4587> rotateToFaceVertical(class_2350 facing) {
		return () -> {
			class_4587 stack = new class_4587();
			TransformStack.cast(stack)
				.centre()
				.rotateY(AngleHelper.horizontalAngle(facing))
				.rotateX(AngleHelper.verticalAngle(facing) + 90)
				.unCentre();
			return stack;
		};
	}

}
