package com.simibubi.create.foundation.utility;

import java.util.List;

import joptsimple.internal.Strings;
import net.minecraft.class_124;
import net.minecraft.class_1657;
import net.minecraft.class_2561;
import net.minecraft.class_5250;

public class LangBuilder {

	String namespace;
	class_5250 component;

	public LangBuilder(String namespace) {
		this.namespace = namespace;
	}

	public LangBuilder space() {
		return text(" ");
	}

	public LangBuilder newLine() {
		return text("\n");
	}

	/**
	 * Appends a localised component<br>
	 * To add an independently formatted localised component, use add() and a nested
	 * builder
	 * 
	 * @param langKey
	 * @param args
	 * @return
	 */
	public LangBuilder translate(String langKey, Object... args) {
		return add(Components.translatable(namespace + "." + langKey, Lang.resolveBuilders(args)));
	}

	/**
	 * Appends a text component
	 * 
	 * @param literalText
	 * @return
	 */
	public LangBuilder text(String literalText) {
		return add(Components.literal(literalText));
	}

	/**
	 * Appends a colored text component
	 * 
	 * @param format
	 * @param literalText
	 * @return
	 */
	public LangBuilder text(class_124 format, String literalText) {
		return add(Components.literal(literalText).method_27692(format));
	}

	/**
	 * Appends a colored text component
	 * 
	 * @param color
	 * @param literalText
	 * @return
	 */
	public LangBuilder text(int color, String literalText) {
		return add(Components.literal(literalText).method_27694(s -> s.method_36139(color)));
	}

	/**
	 * Appends the contents of another builder
	 * 
	 * @param otherBuilder
	 * @return
	 */
	public LangBuilder add(LangBuilder otherBuilder) {
		return add(otherBuilder.component());
	}

	/**
	 * Appends a component
	 * 
	 * @param customComponent
	 * @return
	 */
	public LangBuilder add(class_5250 customComponent) {
		component = component == null ? customComponent : component.method_10852(customComponent);
		return this;
	}

	//

	/**
	 * Applies the format to all added components
	 * 
	 * @param format
	 * @return
	 */
	public LangBuilder style(class_124 format) {
		assertComponent();
		component = component.method_27692(format);
		return this;
	}

	/**
	 * Applies the color to all added components
	 * 
	 * @param color
	 * @return
	 */
	public LangBuilder color(int color) {
		assertComponent();
		component = component.method_27694(s -> s.method_36139(color));
		return this;
	}

	//

	public class_5250 component() {
		assertComponent();
		return component;
	}

	public String string() {
		return component().getString();
	}

	public String json() {
		return class_2561.class_2562.method_10867(component());
	}

	public void sendStatus(class_1657 player) {
		player.method_7353(component(), true);
	}

	public void sendChat(class_1657 player) {
		player.method_7353(component(), false);
	}

	public void addTo(List<? super class_5250> tooltip) {
		tooltip.add(component());
	}

	public void forGoggles(List<? super class_5250> tooltip) {
		forGoggles(tooltip, 0);
	}

	public void forGoggles(List<? super class_5250> tooltip, int indents) {
		tooltip.add(Lang.builder()
			.text(Strings.repeat(' ', 4 + indents))
			.add(this)
			.component());
	}

	//

	private void assertComponent() {
		if (component == null)
			throw new IllegalStateException("No components were added to builder");
	}

}