package com.simibubi.create.foundation.utility;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.function.UnaryOperator;

import javax.annotation.Nullable;
import net.minecraft.class_1772;
import net.minecraft.class_1799;
import net.minecraft.class_2487;
import net.minecraft.class_2499;
import net.minecraft.class_2520;
import net.minecraft.class_2561;
import net.minecraft.class_2586;
import net.minecraft.class_2591;
import net.minecraft.class_2636;
import com.simibubi.create.AllBlockEntityTypes;

public final class NBTProcessors {

	private static final Map<class_2591<?>, UnaryOperator<class_2487>> processors = new HashMap<>();
	private static final Map<class_2591<?>, UnaryOperator<class_2487>> survivalProcessors = new HashMap<>();

	public static synchronized void addProcessor(class_2591<?> type, UnaryOperator<class_2487> processor) {
		processors.put(type, processor);
	}

	public static synchronized void addSurvivalProcessor(class_2591<?> type,
		UnaryOperator<class_2487> processor) {
		survivalProcessors.put(type, processor);
	}

	static {
		String[] signSides = new String[] { "front_text", "back_text" };
		UnaryOperator<class_2487> signProcessor = data -> {
			for (String side : signSides) {
				if (data.method_10573(side, class_2520.field_33260)) {
					class_2487 sideData = data.method_10562(side);
					if (sideData.method_10573("messages", class_2520.field_33259)) {
						class_2499 messages = sideData.method_10554("messages", class_2520.field_33258);
						for (int i = 0; i < messages.size(); i++) {
							String string = messages.method_10608(i);
							if (textComponentHasClickEvent(string)) {
								return null;
							}
						}
					}
				}
			}
			return data;
		};
		addProcessor(class_2591.field_11911, signProcessor);
		addProcessor(class_2591.field_40330, signProcessor);

		addProcessor(class_2591.field_16412, data -> {
			if (!data.method_10573("Book", class_2520.field_33260))
				return data;
			class_2487 book = data.method_10562("Book");

			if (!book.method_10573("tag", class_2520.field_33260))
				return data;
			class_2487 tag = book.method_10562("tag");

			if (!tag.method_10573("pages", class_2520.field_33259))
				return data;
			class_2499 pages = tag.method_10554("pages", class_2520.field_33258);

			for (class_2520 inbt : pages) {
				if (textComponentHasClickEvent(inbt.method_10714()))
					return null;
			}
			return data;
		});
		addProcessor(AllBlockEntityTypes.CREATIVE_CRATE.get(), itemProcessor("Filter"));
		addProcessor(AllBlockEntityTypes.PLACARD.get(), itemProcessor("Item"));
	}

	public static UnaryOperator<class_2487> itemProcessor(String tagKey) {
		return data -> {
			class_2487 compound = data.method_10562(tagKey);
			if (!compound.method_10573("tag", 10))
				return data;
			class_2487 itemTag = compound.method_10562("tag");
			if (itemTag == null)
				return data;
			HashSet<String> keys = new HashSet<>(itemTag.method_10541());
			for (String key : keys)
				if (isUnsafeItemNBTKey(key))
					itemTag.method_10551(key);
			if (itemTag.method_33133())
				compound.method_10551("tag");
			return data;
		};
	}

	public static class_1799 withUnsafeNBTDiscarded(class_1799 stack) {
		if (stack.method_7969() == null)
			return stack;
		class_1799 copy = stack.method_7972();
		stack.method_7969()
			.method_10541()
			.stream()
			.filter(NBTProcessors::isUnsafeItemNBTKey)
			.forEach(copy::method_7983);
		return copy;
	}

	public static boolean isUnsafeItemNBTKey(String name) {
		if (name.equals(class_1772.field_30874))
			return false;
		if (name.equals("Enchantments"))
			return false;
		if (name.contains("Potion"))
			return false;
		if (name.contains("Damage"))
			return false;
		if (name.equals("display"))
			return false;
		return true;
	}

	public static boolean textComponentHasClickEvent(String json) {
		class_2561 component = class_2561.class_2562.method_10877(json.isEmpty() ? "\"\"" : json);
		return component != null && component.method_10866() != null && component.method_10866()
			.method_10970() != null;
	}

	private NBTProcessors() {}

	@Nullable
	public static class_2487 process(class_2586 blockEntity, class_2487 compound, boolean survival) {
		if (compound == null)
			return null;
		class_2591<?> type = blockEntity.method_11017();
		if (survival && survivalProcessors.containsKey(type))
			compound = survivalProcessors.get(type)
				.apply(compound);
		if (compound != null && processors.containsKey(type))
			return processors.get(type)
				.apply(compound);
		if (blockEntity instanceof class_2636)
			return compound;
		if (blockEntity.method_11011())
			return null;
		return compound;
	}

}
