package com.simibubi.create.foundation.utility.outliner;

import java.util.Optional;

import javax.annotation.Nullable;
import net.minecraft.class_2350;
import net.minecraft.class_2350.class_2351;
import net.minecraft.class_243;
import net.minecraft.class_3532;
import net.minecraft.class_4587;
import net.minecraft.class_4588;
import net.minecraft.class_4608;
import net.minecraft.class_765;
import org.joml.Matrix3f;

import com.jozufozu.flywheel.util.transform.TransformStack;
import com.simibubi.create.AllSpecialTextures;
import com.simibubi.create.foundation.render.RenderTypes;
import com.simibubi.create.foundation.render.SuperRenderTypeBuffer;
import com.simibubi.create.foundation.utility.AngleHelper;
import com.simibubi.create.foundation.utility.Color;
import com.simibubi.create.foundation.utility.VecHelper;

public abstract class Outline {

	protected OutlineParams params;
	protected Matrix3f transformNormals; // TODO: not used?

	public Outline() {
		params = new OutlineParams();
	}

	public abstract void render(class_4587 ms, SuperRenderTypeBuffer buffer, float pt);

	public void tick() {}

	public OutlineParams getParams() {
		return params;
	}

	public void renderCuboidLine(class_4587 ms, SuperRenderTypeBuffer buffer, class_243 start, class_243 end) {
		class_243 diff = end.method_1020(start);
		float hAngle = AngleHelper.deg(class_3532.method_15349(diff.field_1352, diff.field_1350));
		float hDistance = (float) diff.method_18805(1, 0, 1)
			.method_1033();
		float vAngle = AngleHelper.deg(class_3532.method_15349(hDistance, diff.field_1351)) - 90;
		ms.method_22903();
		TransformStack.cast(ms)
			.translate(start)
			.rotateY(hAngle).rotateX(vAngle);
		renderAACuboidLine(ms, buffer, class_243.field_1353, new class_243(0, 0, diff.method_1033()));
		ms.method_22909();
	}

	public void renderAACuboidLine(class_4587 ms, SuperRenderTypeBuffer buffer, class_243 start, class_243 end) {
		float lineWidth = params.getLineWidth();
		if (lineWidth == 0)
			return;

		class_4588 builder = buffer.getBuffer(RenderTypes.getOutlineSolid());

		class_243 diff = end.method_1020(start);
		if (diff.field_1352 + diff.field_1351 + diff.field_1350 < 0) {
			class_243 temp = start;
			start = end;
			end = temp;
			diff = diff.method_1021(-1);
		}

		class_243 extension = diff.method_1029()
			.method_1021(lineWidth / 2);
		class_243 plane = VecHelper.axisAlingedPlaneOf(diff);
		class_2350 face = class_2350.method_10142(diff.field_1352, diff.field_1351, diff.field_1350);
		class_2351 axis = face.method_10166();

		start = start.method_1020(extension);
		end = end.method_1019(extension);
		plane = plane.method_1021(lineWidth / 2);

		class_243 a1 = plane.method_1019(start);
		class_243 b1 = plane.method_1019(end);
		plane = VecHelper.rotate(plane, -90, axis);
		class_243 a2 = plane.method_1019(start);
		class_243 b2 = plane.method_1019(end);
		plane = VecHelper.rotate(plane, -90, axis);
		class_243 a3 = plane.method_1019(start);
		class_243 b3 = plane.method_1019(end);
		plane = VecHelper.rotate(plane, -90, axis);
		class_243 a4 = plane.method_1019(start);
		class_243 b4 = plane.method_1019(end);

		if (params.disableNormals) {
			face = class_2350.field_11036;
			putQuad(ms, builder, b4, b3, b2, b1, face);
			putQuad(ms, builder, a1, a2, a3, a4, face);
			putQuad(ms, builder, a1, b1, b2, a2, face);
			putQuad(ms, builder, a2, b2, b3, a3, face);
			putQuad(ms, builder, a3, b3, b4, a4, face);
			putQuad(ms, builder, a4, b4, b1, a1, face);
			return;
		}

		putQuad(ms, builder, b4, b3, b2, b1, face);
		putQuad(ms, builder, a1, a2, a3, a4, face.method_10153());
		class_243 vec = a1.method_1020(a4);
		face = class_2350.method_10142(vec.field_1352, vec.field_1351, vec.field_1350);
		putQuad(ms, builder, a1, b1, b2, a2, face);
		vec = VecHelper.rotate(vec, -90, axis);
		face = class_2350.method_10142(vec.field_1352, vec.field_1351, vec.field_1350);
		putQuad(ms, builder, a2, b2, b3, a3, face);
		vec = VecHelper.rotate(vec, -90, axis);
		face = class_2350.method_10142(vec.field_1352, vec.field_1351, vec.field_1350);
		putQuad(ms, builder, a3, b3, b4, a4, face);
		vec = VecHelper.rotate(vec, -90, axis);
		face = class_2350.method_10142(vec.field_1352, vec.field_1351, vec.field_1350);
		putQuad(ms, builder, a4, b4, b1, a1, face);
	}

	public void putQuad(class_4587 ms, class_4588 builder, class_243 v1, class_243 v2, class_243 v3, class_243 v4,
		class_2350 normal) {
		putQuadUV(ms, builder, v1, v2, v3, v4, 0, 0, 1, 1, normal);
	}

	public void putQuadUV(class_4587 ms, class_4588 builder, class_243 v1, class_243 v2, class_243 v3, class_243 v4, float minU,
		float minV, float maxU, float maxV, class_2350 normal) {
		putVertex(ms, builder, v1, minU, minV, normal);
		putVertex(ms, builder, v2, maxU, minV, normal);
		putVertex(ms, builder, v3, maxU, maxV, normal);
		putVertex(ms, builder, v4, minU, maxV, normal);
	}

	protected void putVertex(class_4587 ms, class_4588 builder, class_243 pos, float u, float v, class_2350 normal) {
		putVertex(ms.method_23760(), builder, (float) pos.field_1352, (float) pos.field_1351, (float) pos.field_1350, u, v, normal);
	}

	protected void putVertex(class_4587.class_4665 pose, class_4588 builder, float x, float y, float z, float u, float v, class_2350 normal) {
		Color rgb = params.rgb;
		if (transformNormals == null)
			transformNormals = pose.method_23762();

		int xOffset = 0;
		int yOffset = 0;
		int zOffset = 0;

		if (normal != null) {
			xOffset = normal.method_10148();
			yOffset = normal.method_10164();
			zOffset = normal.method_10165();
		}

		builder.method_22918(pose.method_23761(), x, y, z)
			.method_22915(rgb.getRedAsFloat(), rgb.getGreenAsFloat(), rgb.getBlueAsFloat(), rgb.getAlphaAsFloat() * params.alpha)
			.method_22913(u, v)
			.method_22922(class_4608.field_21444)
			.method_22916(params.lightMap)
			.method_23763(pose.method_23762(), xOffset, yOffset, zOffset)
			.method_1344();

		transformNormals = null;
	}

	public static class OutlineParams {
		protected Optional<AllSpecialTextures> faceTexture;
		protected Optional<AllSpecialTextures> hightlightedFaceTexture;
		protected class_2350 highlightedFace;
		protected boolean fadeLineWidth;
		protected boolean disableCull;
		protected boolean disableNormals;
		protected float alpha;
		protected int lightMap;
		protected Color rgb;
		private float lineWidth;

		public OutlineParams() {
			faceTexture = hightlightedFaceTexture = Optional.empty();
			alpha = 1;
			lineWidth = 1 / 32f;
			fadeLineWidth = true;
			rgb = Color.WHITE;
			lightMap = class_765.field_32767;
		}

		// builder

		public OutlineParams colored(int color) {
			rgb = new Color(color, false);
			return this;
		}

		public OutlineParams colored(Color c) {
			rgb = c.copy();
			return this;
		}

		public OutlineParams lightMap(int light) {
			lightMap = light;
			return this;
		}

		public OutlineParams lineWidth(float width) {
			this.lineWidth = width;
			return this;
		}

		public OutlineParams withFaceTexture(AllSpecialTextures texture) {
			this.faceTexture = Optional.ofNullable(texture);
			return this;
		}

		public OutlineParams clearTextures() {
			return this.withFaceTextures(null, null);
		}

		public OutlineParams withFaceTextures(AllSpecialTextures texture, AllSpecialTextures highlightTexture) {
			this.faceTexture = Optional.ofNullable(texture);
			this.hightlightedFaceTexture = Optional.ofNullable(highlightTexture);
			return this;
		}

		public OutlineParams highlightFace(@Nullable class_2350 face) {
			highlightedFace = face;
			return this;
		}

		public OutlineParams disableNormals() {
			disableNormals = true;
			return this;
		}

		public OutlineParams disableCull() {
			disableCull = true;
			return this;
		}

		// getter

		public float getLineWidth() {
			return fadeLineWidth ? alpha * lineWidth : lineWidth;
		}

		public class_2350 getHighlightedFace() {
			return highlightedFace;
		}

	}

}
