package com.simibubi.create.infrastructure.command;

import java.util.Collections;
import java.util.function.Predicate;

import com.mojang.brigadier.CommandDispatcher;
import com.mojang.brigadier.builder.LiteralArgumentBuilder;
import com.mojang.brigadier.tree.CommandNode;
import com.mojang.brigadier.tree.LiteralCommandNode;

import net.fabricmc.api.EnvType;
import net.fabricmc.loader.api.FabricLoader;
import net.minecraft.class_1657;
import net.minecraft.class_2168;
import net.minecraft.class_2170;

public class AllCommands {

	public static final Predicate<class_2168> SOURCE_IS_PLAYER = cs -> cs.method_9228() instanceof class_1657;

	public static void register(CommandDispatcher<class_2168> dispatcher) {

		LiteralCommandNode<class_2168> util = buildUtilityCommands();

		LiteralArgumentBuilder<class_2168> root = class_2170.method_9247("create")
				.requires(cs -> cs.method_9259(0))
				// general purpose
				.then(new ToggleDebugCommand().register())
				.then(FabulousWarningCommand.register())
				.then(OverlayConfigCommand.register())
				.then(DumpRailwaysCommand.register())
				.then(FixLightingCommand.register())
				.then(DebugInfoCommand.register())
				.then(HighlightCommand.register())
				.then(KillTrainCommand.register())
				.then(PassengerCommand.register())
				.then(CouplingCommand.register())
				.then(ConfigCommand.register())
				.then(PonderCommand.register())
				.then(CloneCommand.register())
				.then(GlueCommand.register())


				// utility
				.then(util);

		FabricLoader loader = FabricLoader.getInstance();
		if (loader.isDevelopmentEnvironment() && loader.getEnvironmentType() == EnvType.CLIENT)
			root.then(CreateTestCommand.register());

		LiteralCommandNode<class_2168> createRoot = dispatcher.register(root);

		createRoot.addChild(buildRedirect("u", util));

		CommandNode<class_2168> c = dispatcher.findNode(Collections.singleton("c"));
		if (c != null)
			return;

		dispatcher.getRoot()
			.addChild(buildRedirect("c", createRoot));

	}

	private static LiteralCommandNode<class_2168> buildUtilityCommands() {

		return class_2170.method_9247("util")
				.then(ReplaceInCommandBlocksCommand.register())
				.then(ClearBufferCacheCommand.register())
				.then(CameraDistanceCommand.register())
				.then(CameraAngleCommand.register())
				.then(FlySpeedCommand.register())
				//.then(DebugValueCommand.register())
				//.then(KillTPSCommand.register())
				.build();

	}

	/**
	 * *****
	 * https://github.com/VelocityPowered/Velocity/blob/8abc9c80a69158ebae0121fda78b55c865c0abad/proxy/src/main/java/com/velocitypowered/proxy/util/BrigadierUtils.java#L38
	 * *****
	 * <p>
	 * Returns a literal node that redirects its execution to
	 * the given destination node.
	 *
	 * @param alias       the command alias
	 * @param destination the destination node
	 *
	 * @return the built node
	 */
	public static LiteralCommandNode<class_2168> buildRedirect(final String alias, final LiteralCommandNode<class_2168> destination) {
		// Redirects only work for nodes with children, but break the top argument-less command.
		// Manually adding the root command after setting the redirect doesn't fix it.
		// See https://github.com/Mojang/brigadier/issues/46). Manually clone the node instead.
		LiteralArgumentBuilder<class_2168> builder = LiteralArgumentBuilder
				.<class_2168>literal(alias)
				.requires(destination.getRequirement())
				.forward(destination.getRedirect(), destination.getRedirectModifier(), destination.isFork())
				.executes(destination.getCommand());
		for (CommandNode<class_2168> child : destination.getChildren()) {
			builder.then(child);
		}
		return builder.build();
	}

}
