package com.simibubi.create.infrastructure.command;

import java.util.List;
import net.minecraft.class_2168;
import net.minecraft.class_2170;
import net.minecraft.class_2246;
import net.minecraft.class_2262;
import net.minecraft.class_2338;
import net.minecraft.class_238;
import net.minecraft.class_243;
import net.minecraft.class_2487;
import net.minecraft.class_2586;
import net.minecraft.class_2680;
import net.minecraft.class_2694;
import net.minecraft.class_3218;
import net.minecraft.class_3341;
import net.minecraft.class_3499;
import net.minecraft.class_3829;
import com.google.common.collect.Lists;
import com.mojang.brigadier.Command;
import com.mojang.brigadier.builder.ArgumentBuilder;
import com.mojang.brigadier.exceptions.CommandSyntaxException;
import com.mojang.brigadier.exceptions.Dynamic2CommandExceptionType;
import com.simibubi.create.content.contraptions.glue.SuperGlueEntity;
import com.simibubi.create.foundation.utility.Components;

public class CloneCommand {

	private static final Dynamic2CommandExceptionType CLONE_TOO_BIG_EXCEPTION = new Dynamic2CommandExceptionType(
		(arg1, arg2) -> Components.translatable("commands.clone.toobig", arg1, arg2));

	public static ArgumentBuilder<class_2168, ?> register() {
		return class_2170.method_9247("clone")
			.requires(cs -> cs.method_9259(2))
			.then(class_2170.method_9244("begin", class_2262.method_9698())
				.then(class_2170.method_9244("end", class_2262.method_9698())
					.then(class_2170.method_9244("destination", class_2262.method_9698())
						.then(class_2170.method_9247("skipBlocks")
							.executes(ctx -> doClone(ctx.getSource(), class_2262.method_9696(ctx, "begin"),
								class_2262.method_9696(ctx, "end"),
								class_2262.method_9696(ctx, "destination"), false)))
						.executes(ctx -> doClone(ctx.getSource(), class_2262.method_9696(ctx, "begin"),
							class_2262.method_9696(ctx, "end"),
							class_2262.method_9696(ctx, "destination"), true)))))
			.executes(ctx -> {
				ctx.getSource()
					.method_9226(() -> Components.literal(
						"Clones all blocks as well as super glue from the specified area to the target destination"),
						true);

				return Command.SINGLE_SUCCESS;
			});

	}

	private static int doClone(class_2168 source, class_2338 begin, class_2338 end, class_2338 destination,
		boolean cloneBlocks) throws CommandSyntaxException {
		class_3341 sourceArea = class_3341.method_34390(begin, end);
		class_2338 destinationEnd = destination.method_10081(sourceArea.method_14659());
		class_3341 destinationArea = class_3341.method_34390(destination, destinationEnd);

		int i = sourceArea.method_35414() * sourceArea.method_14660() * sourceArea.method_14663();
		if (i > 32768)
			throw CLONE_TOO_BIG_EXCEPTION.create(32768, i);

		class_3218 world = source.method_9225();

		if (!world.method_22343(begin, end) || !world.method_22343(destination, destinationEnd))
			throw class_2262.field_10703.create();

		class_2338 diffToTarget = new class_2338(destinationArea.method_35415() - sourceArea.method_35415(),
			destinationArea.method_35416() - sourceArea.method_35416(), destinationArea.method_35417() - sourceArea.method_35417());

		int blockPastes = cloneBlocks ? cloneBlocks(sourceArea, world, diffToTarget) : 0;
		int gluePastes = cloneGlue(sourceArea, world, diffToTarget);

		if (cloneBlocks)
			source.method_9226(() -> Components.literal("Successfully cloned " + blockPastes + " Blocks"), true);

		source.method_9226(() -> Components.literal("Successfully applied glue " + gluePastes + " times"), true);
		return blockPastes + gluePastes;

	}

	private static int cloneGlue(class_3341 sourceArea, class_3218 world, class_2338 diffToTarget) {
		int gluePastes = 0;

		class_238 bb = new class_238(sourceArea.method_35415(), sourceArea.method_35416(), sourceArea.method_35417(), sourceArea.method_35418() + 1,
			sourceArea.method_35419() + 1, sourceArea.method_35420() + 1);
		for (SuperGlueEntity g : SuperGlueEntity.collectCropped(world, bb)) {
			g.method_33574(g.method_19538()
				.method_1019(class_243.method_24954(diffToTarget)));
			world.method_8649(g);
			gluePastes++;
		}

		return gluePastes;
	}

	private static int cloneBlocks(class_3341 sourceArea, class_3218 world, class_2338 diffToTarget) {
		int blockPastes = 0;

		List<class_3499.class_3501> blocks = Lists.newArrayList();
		List<class_3499.class_3501> beBlocks = Lists.newArrayList();

		for (int z = sourceArea.method_35417(); z <= sourceArea.method_35420(); ++z) {
			for (int y = sourceArea.method_35416(); y <= sourceArea.method_35419(); ++y) {
				for (int x = sourceArea.method_35415(); x <= sourceArea.method_35418(); ++x) {
					class_2338 currentPos = new class_2338(x, y, z);
					class_2338 newPos = currentPos.method_10081(diffToTarget);
					class_2694 cached = new class_2694(world, currentPos, false);
					class_2680 state = cached.method_11681();
					class_2586 be = world.method_8321(currentPos);
					if (be != null) {
						class_2487 nbt = be.method_38242();
						beBlocks.add(new class_3499.class_3501(newPos, state, nbt));
					} else {
						blocks.add(new class_3499.class_3501(newPos, state, null));
					}
				}
			}
		}

		List<class_3499.class_3501> allBlocks = Lists.newArrayList();
		allBlocks.addAll(blocks);
		allBlocks.addAll(beBlocks);

		List<class_3499.class_3501> reverse = Lists.reverse(allBlocks);

		for (class_3499.class_3501 info : reverse) {
			class_2586 be = world.method_8321(info.comp_1341());
			class_3829.method_16825(be);
			world.method_8652(info.comp_1341(), class_2246.field_10499.method_9564(), 2);
		}

		for (class_3499.class_3501 info : allBlocks) {
			if (world.method_8652(info.comp_1341(), info.comp_1342(), 2))
				blockPastes++;
		}

		for (class_3499.class_3501 info : beBlocks) {
			class_2586 be = world.method_8321(info.comp_1341());
			if (be != null && info.comp_1343() != null) {
				info.comp_1343().method_10569("x", info.comp_1341().method_10263());
				info.comp_1343().method_10569("y", info.comp_1341().method_10264());
				info.comp_1343().method_10569("z", info.comp_1341().method_10260());
				be.method_11014(info.comp_1343());
				be.method_5431();
			}

			// idk why the state is set twice for a be, but its done like this in the
			// original clone command
			world.method_8652(info.comp_1341(), info.comp_1342(), 2);
		}

		for (class_3499.class_3501 info : reverse) {
			world.method_8408(info.comp_1341(), info.comp_1342().method_26204());
		}

		world.method_14196()
			.method_39383(sourceArea, diffToTarget);

		return blockPastes;
	}

}
