package com.simibubi.create.infrastructure.command;

import java.util.ArrayList;
import java.util.Collection;
import java.util.UUID;
import net.minecraft.class_1297;
import net.minecraft.class_1657;
import net.minecraft.class_1688;
import net.minecraft.class_2168;
import net.minecraft.class_2170;
import net.minecraft.class_2186;
import com.google.common.collect.Lists;
import com.mojang.brigadier.Command;
import com.mojang.brigadier.builder.ArgumentBuilder;
import com.mojang.brigadier.exceptions.DynamicCommandExceptionType;
import com.mojang.brigadier.exceptions.SimpleCommandExceptionType;
import com.simibubi.create.content.contraptions.minecart.CouplingHandler;
import com.simibubi.create.content.contraptions.minecart.capability.CapabilityMinecartController;
import com.simibubi.create.content.contraptions.minecart.capability.MinecartController;
import com.simibubi.create.foundation.utility.Components;
import com.simibubi.create.foundation.utility.Iterate;
import io.github.fabricators_of_create.porting_lib.util.LazyOptional;
import io.github.fabricators_of_create.porting_lib.util.MinecartAndRailUtil;

public class CouplingCommand {

	public static final SimpleCommandExceptionType ONLY_MINECARTS_ALLOWED =
		new SimpleCommandExceptionType(Components.literal("Only Minecarts can be coupled"));
	public static final SimpleCommandExceptionType SAME_DIMENSION =
		new SimpleCommandExceptionType(Components.literal("Minecarts have to be in the same Dimension"));
	public static final DynamicCommandExceptionType TWO_CARTS =
		new DynamicCommandExceptionType(a -> Components.literal(
			"Your selector targeted " + a + " entities. You can only couple 2 Minecarts at a time."));

	public static ArgumentBuilder<class_2168, ?> register() {

		return class_2170.method_9247("coupling")
			.requires(cs -> cs.method_9259(2))
			.then(class_2170.method_9247("add")
				.then(class_2170.method_9244("cart1", class_2186.method_9309())
					.then(class_2170.method_9244("cart2", class_2186.method_9309())
						.executes(ctx -> {
							class_1297 cart1 = class_2186.method_9313(ctx, "cart1");
							if (!(cart1 instanceof class_1688))
								throw ONLY_MINECARTS_ALLOWED.create();

							class_1297 cart2 = class_2186.method_9313(ctx, "cart2");
							if (!(cart2 instanceof class_1688))
								throw ONLY_MINECARTS_ALLOWED.create();

							if (!cart1.method_5770()
								.equals(cart2.method_5770()))
								throw SAME_DIMENSION.create();

							class_1297 source = ctx.getSource()
								.method_9228();

							CouplingHandler.tryToCoupleCarts(
								source instanceof class_1657 ? (class_1657) source : null, cart1.method_5770(),
								cart1.method_5628(), cart2.method_5628());

							return Command.SINGLE_SUCCESS;
						})))
				.then(class_2170.method_9244("carts", class_2186.method_9306())
					.executes(ctx -> {
						Collection<? extends class_1297> entities = class_2186.method_9317(ctx, "carts");
						if (entities.size() != 2)
							throw TWO_CARTS.create(entities.size());

						ArrayList<? extends class_1297> eList = Lists.newArrayList(entities);
						class_1297 cart1 = eList.get(0);
						if (!(cart1 instanceof class_1688))
							throw ONLY_MINECARTS_ALLOWED.create();

						class_1297 cart2 = eList.get(1);
						if (!(cart2 instanceof class_1688))
							throw ONLY_MINECARTS_ALLOWED.create();

						if (!cart1.method_5770()
							.equals(cart2.method_5770()))
							throw SAME_DIMENSION.create();

						class_1297 source = ctx.getSource()
							.method_9228();

						CouplingHandler.tryToCoupleCarts(source instanceof class_1657 ? (class_1657) source : null,
							cart1.method_5770(), cart1.method_5628(), cart2.method_5628());

						return Command.SINGLE_SUCCESS;
					})))
			.then(class_2170.method_9247("remove")
				.then(class_2170.method_9244("cart1", class_2186.method_9309())
					.then(class_2170.method_9244("cart2", class_2186.method_9309())
						.executes(ctx -> {
							class_1297 cart1 = class_2186.method_9313(ctx, "cart1");
							if (!(cart1 instanceof class_1688))
								throw ONLY_MINECARTS_ALLOWED.create();

							class_1297 cart2 = class_2186.method_9313(ctx, "cart2");
							if (!(cart2 instanceof class_1688))
								throw ONLY_MINECARTS_ALLOWED.create();

							MinecartController cart1Controller = ((class_1688) cart1).create$getController();

							int cart1Couplings = (cart1Controller.isConnectedToCoupling() ? 1 : 0)
								+ (cart1Controller.isLeadingCoupling() ? 1 : 0);
							if (cart1Couplings == 0) {
								ctx.getSource()
									.method_9226(() -> Components.literal("Minecart has no Couplings Attached"), true);
								return 0;
							}

							for (boolean bool : Iterate.trueAndFalse) {
								UUID coupledCart = cart1Controller.getCoupledCart(bool);
								if (coupledCart == null)
									continue;

								if (coupledCart != cart2.method_5667())
									continue;

								MinecartController cart2Controller =
									CapabilityMinecartController.getIfPresent(cart1.method_5770(), coupledCart);
								if (cart2Controller == null)
									return 0;

								cart1Controller.removeConnection(bool);
								cart2Controller.removeConnection(!bool);
								return Command.SINGLE_SUCCESS;
							}

							ctx.getSource()
								.method_9226(() -> Components.literal("The specified Carts are not coupled"), true);

							return 0;
						}))))
			.then(class_2170.method_9247("removeAll")
				.then(class_2170.method_9244("cart", class_2186.method_9309())
					.executes(ctx -> {
						class_1297 cart = class_2186.method_9313(ctx, "cart");
						if (!(cart instanceof class_1688))
							throw ONLY_MINECARTS_ALLOWED.create();

						MinecartController controller = ((class_1688) cart).create$getController();

						int couplings =
							(controller.isConnectedToCoupling() ? 1 : 0) + (controller.isLeadingCoupling() ? 1 : 0);
						if (couplings == 0) {
							ctx.getSource()
								.method_9226(() -> Components.literal("Minecart has no Couplings Attached"), true);
							return 0;
						}

						controller.decouple();

						ctx.getSource()
							.method_9226(() ->
								Components.literal("Removed " + couplings + " couplings from the Minecart"), true);

						return couplings;
					})));

	}

}
