package com.simibubi.create.infrastructure.gui;

import com.jozufozu.flywheel.util.transform.TransformStack;
import com.mojang.blaze3d.platform.GlStateManager;
import com.mojang.blaze3d.systems.RenderSystem;
import com.simibubi.create.AllBlocks;
import com.simibubi.create.Create;
import com.simibubi.create.compat.Mods;
import com.simibubi.create.foundation.config.ui.BaseConfigScreen;
import com.simibubi.create.foundation.gui.AbstractSimiScreen;
import com.simibubi.create.foundation.gui.AllGuiTextures;
import com.simibubi.create.foundation.gui.ScreenOpener;
import com.simibubi.create.foundation.gui.element.BoxElement;
import com.simibubi.create.foundation.gui.element.GuiGameElement;
import com.simibubi.create.foundation.item.TooltipHelper;
import com.simibubi.create.foundation.item.TooltipHelper.Palette;
import com.simibubi.create.foundation.ponder.ui.PonderTagIndexScreen;
import com.simibubi.create.foundation.utility.Color;
import com.simibubi.create.foundation.utility.Components;
import com.simibubi.create.foundation.utility.Iterate;
import com.simibubi.create.foundation.utility.Lang;
import com.terraformersmc.modmenu.gui.ModsScreen;

import io.github.fabricators_of_create.porting_lib.mixin.accessors.client.accessor.ScreenAccessor;
import io.github.fabricators_of_create.porting_lib.mixin.accessors.client.accessor.TitleScreenAccessor;
import net.minecraft.class_124;
import net.minecraft.class_156;
import net.minecraft.class_2561;
import net.minecraft.class_2960;
import net.minecraft.class_332;
import net.minecraft.class_3532;
import net.minecraft.class_407;
import net.minecraft.class_4185;
import net.minecraft.class_437;
import net.minecraft.class_442;
import net.minecraft.class_4587;
import net.minecraft.class_751;
import net.minecraft.class_766;
import net.minecraft.class_7919;

public class CreateMainMenuScreen extends AbstractSimiScreen {

	public static final class_751 PANORAMA_RESOURCES =
		new class_751(Create.asResource("textures/gui/title/background/panorama"));
	public static final class_2960 PANORAMA_OVERLAY_TEXTURES =
		new class_2960("textures/gui/title/background/panorama_overlay.png");
	public static final class_766 PANORAMA = new class_766(PANORAMA_RESOURCES);

	private static final class_2561 CURSEFORGE_TOOLTIP = Components.literal("CurseForge").method_27694(s -> s.method_36139(0xFC785C).method_10982(true));
	private static final class_2561 MODRINTH_TOOLTIP = Components.literal("Modrinth").method_27694(s -> s.method_36139(0x3FD32B).method_10982(true));

	public static final String CURSEFORGE_LINK = "https://www.curseforge.com/minecraft/mc-mods/create-fabric";
	public static final String MODRINTH_LINK = "https://modrinth.com/mod/create-fabric";
	public static final String ISSUE_TRACKER_LINK = "https://github.com/Fabricators-of-Create/Create/issues";
	public static final String SUPPORT_LINK = "https://github.com/Creators-of-Create/Create/wiki/Supporting-the-Project";

	protected final class_437 parent;
	protected boolean returnOnClose;

	private class_766 vanillaPanorama;
	private long firstRenderTime;
	private class_4185 gettingStarted;

	public final boolean fromTitleOrMods;

	public CreateMainMenuScreen(class_437 parent) {
		this.parent = parent;
		returnOnClose = true;
		fromTitleOrMods = (parent instanceof class_442) || Mods.MODMENU.runIfInstalled(() ->
				() -> parent instanceof ModsScreen).orElse(Boolean.FALSE);
		if (parent instanceof class_442 titleScreen)
			vanillaPanorama = ((TitleScreenAccessor) titleScreen).port_lib$getPanorama();
		else
			vanillaPanorama = new class_766(class_442.field_17774);
	}

	@Override
	public void method_25394(class_332 graphics, int mouseX, int mouseY, float partialTicks) {
		if (firstRenderTime == 0L)
			this.firstRenderTime = class_156.method_658();
		super.method_25394(graphics, mouseX, mouseY, partialTicks);
	}

	@Override
	protected void renderWindow(class_332 graphics, int mouseX, int mouseY, float partialTicks) {
		float f = (float) (class_156.method_658() - this.firstRenderTime) / 1000.0F;
		float alpha = class_3532.method_15363(f, 0.0F, 1.0F);
		float elapsedPartials = field_22787.method_1534();

		if (fromTitleOrMods) {
			if (alpha < 1 && parent instanceof class_442)
				vanillaPanorama.method_3317(elapsedPartials, 1);
			PANORAMA.method_3317(elapsedPartials, alpha);

			RenderSystem.enableBlend();
			RenderSystem.blendFunc(GlStateManager.class_4535.SRC_ALPHA,
				GlStateManager.class_4534.ONE_MINUS_SRC_ALPHA);
			graphics.method_25293(PANORAMA_OVERLAY_TEXTURES, 0, 0, this.field_22789, this.field_22790, 0.0F, 0.0F, 16, 128, 16, 128);
		}

		RenderSystem.enableDepthTest();

		class_4587 ms = graphics.method_51448();

		for (int side : Iterate.positiveAndNegative) {
			ms.method_22903();
			ms.method_46416(field_22789 / 2, 60, 200);
			ms.method_22905(24 * side, 24 * side, 32);
			ms.method_22904(-1.75 * ((alpha * alpha) / 2f + .5f), .25f, 0);
			TransformStack.cast(ms)
				.rotateX(45);
			GuiGameElement.of(AllBlocks.LARGE_COGWHEEL.getDefaultState())
				.rotateBlock(0, class_156.method_658() / 32f * side, 0)
				.render(graphics);
			ms.method_46416(-1, 0, -1);
			GuiGameElement.of(AllBlocks.COGWHEEL.getDefaultState())
				.rotateBlock(0, class_156.method_658() / -16f * side + 22.5f, 0)
				.render(graphics);
			ms.method_22909();
		}

		ms.method_22903();
		ms.method_46416(field_22789 / 2 - 32, 32, -10);
		ms.method_22903();
		ms.method_22905(0.25f, 0.25f, 0.25f);
		AllGuiTextures.LOGO.render(graphics, 0, 0);
		ms.method_22909();
		new BoxElement().withBackground(0x88_000000)
			.flatBorder(new Color(0x01_000000))
			.at(-32, 56, 100)
			.withBounds(128, 11)
			.render(graphics);
		ms.method_22909();

		ms.method_22903();
		ms.method_46416(0, 0, 200);
		graphics.method_27534(field_22793, Components.literal(Create.NAME).method_27692(class_124.field_1067)
			.method_10852(
				Components.literal(" v" + Create.VERSION).method_27695(class_124.field_1067, class_124.field_1068)),
			field_22789 / 2, 89, 0xFF_E4BB67);
		ms.method_22909();

		RenderSystem.disableDepthTest();
	}

	protected void method_25426() {
		super.method_25426();
		returnOnClose = true;
		this.addButtons();
	}

	private void addButtons() {
		int yStart = field_22790 / 4 + 40;
		int center = field_22789 / 2;
		int bHeight = 20;
		int bShortWidth = 98;
		int bLongWidth = 200;

		method_37063(class_4185.method_46430(Lang.translateDirect("menu.return"), $ -> linkTo(parent))
			.method_46434(center - 100, yStart + 92, bLongWidth, bHeight)
			.method_46431());
		method_37063(class_4185.method_46430(Lang.translateDirect("menu.configure"), $ -> linkTo(BaseConfigScreen.forCreate(this)))
			.method_46434(center - 100, yStart + 24 + -16, bLongWidth, bHeight)
			.method_46431());

		gettingStarted = class_4185.method_46430(Lang.translateDirect("menu.ponder_index"), $ -> linkTo(new PonderTagIndexScreen()))
			.method_46434(center + 2, yStart + 48 + -16, bShortWidth, bHeight)
			.method_46431();
		gettingStarted.field_22763 = !fromTitleOrMods;
		method_37063(gettingStarted);

		method_37063(new PlatformIconButton(center - 100, yStart + 48 + -16, bShortWidth / 2, bHeight,
			AllGuiTextures.CURSEFORGE_LOGO, 0.085f,
			b -> linkTo(CURSEFORGE_LINK),
			class_7919.method_47407(CURSEFORGE_TOOLTIP)));
		method_37063(new PlatformIconButton(center - 50, yStart + 48 + -16, bShortWidth / 2, bHeight,
			AllGuiTextures.MODRINTH_LOGO, 0.0575f,
			b -> linkTo(MODRINTH_LINK),
			class_7919.method_47407(MODRINTH_TOOLTIP)));

		method_37063(class_4185.method_46430(Lang.translateDirect("menu.report_bugs"), $ -> linkTo(ISSUE_TRACKER_LINK))
			.method_46434(center + 2, yStart + 68, bShortWidth, bHeight)
			.method_46431());
		method_37063(class_4185.method_46430(Lang.translateDirect("menu.support"), $ -> linkTo(SUPPORT_LINK))
			.method_46434(center - 100, yStart + 68, bShortWidth, bHeight)
			.method_46431());
	}

	@Override
	protected void renderWindowForeground(class_332 graphics, int mouseX, int mouseY, float partialTicks) {
		super.renderWindowForeground(graphics, mouseX, mouseY, partialTicks);
		((ScreenAccessor) this).port_lib$getRenderables().forEach(w -> w.method_25394(graphics, mouseX, mouseY, partialTicks));

		if (fromTitleOrMods) {
			if (mouseX < gettingStarted.method_46426() || mouseX > gettingStarted.method_46426() + 98)
				return;
			if (mouseY < gettingStarted.method_46427() || mouseY > gettingStarted.method_46427() + 20)
				return;
			graphics.method_51434(field_22793,
				TooltipHelper.cutTextComponent(Lang.translateDirect("menu.only_ingame"), Palette.ALL_GRAY), mouseX,
				mouseY);
		}
	}

	private void linkTo(class_437 screen) {
		returnOnClose = false;
		ScreenOpener.open(screen);
	}

	private void linkTo(String url) {
		returnOnClose = false;
		ScreenOpener.open(new class_407((p_213069_2_) -> {
			if (p_213069_2_)
				class_156.method_668()
					.method_670(url);
			this.field_22787.method_1507(this);
		}, url, true));
	}

	@Override
	public boolean method_25421() {
		return true;
	}

	protected static class PlatformIconButton extends class_4185 {
		protected final AllGuiTextures icon;
		protected final float scale;

		public PlatformIconButton(int pX, int pY, int pWidth, int pHeight, AllGuiTextures icon, float scale, class_4241 pOnPress, class_7919 tooltip) {
			super(pX, pY, pWidth, pHeight, Components.immutableEmpty(), pOnPress, field_40754);
			this.icon = icon;
			this.scale = scale;
			method_47400(tooltip);
		}

		@Override
		protected void method_48579(class_332 graphics, int pMouseX, int pMouseY, float pt) {
			super.method_48579(graphics, pMouseX, pMouseY, pt);
			class_4587 pPoseStack = graphics.method_51448();
			pPoseStack.method_22903();
			pPoseStack.method_46416(method_46426() + field_22758 / 2 - (icon.width * scale) / 2, method_46427() + field_22759 / 2 - (icon.height * scale) / 2, 0);
			pPoseStack.method_22905(scale, scale, 1);
			icon.render(graphics, 0, 0);
			pPoseStack.method_22909();
		}
	}

}
